/*****************************************************************************
*
* Copyright (c) 2000 - 2006, The Regents of the University of California
* Produced at the Lawrence Livermore National Laboratory
* LLNL-CODE-400142
* All rights reserved.
*
* This file is  part of VisIt. For  details, see https://visit.llnl.gov/.  The
* full copyright notice is contained in the file COPYRIGHT located at the root
* of the VisIt distribution or at http://www.llnl.gov/visit/copyright.html.
*
* Redistribution  and  use  in  source  and  binary  forms,  with  or  without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of  source code must  retain the above  copyright notice,
*    this list of conditions and the disclaimer below.
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this  list of  conditions  and  the  disclaimer (as noted below)  in  the
*    documentation and/or other materials provided with the distribution.
*  - Neither the name of  the LLNS/LLNL nor the names of  its contributors may
*    be used to endorse or promote products derived from this software without
*    specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT  HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR  IMPLIED WARRANTIES, INCLUDING,  BUT NOT  LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND  FITNESS FOR A PARTICULAR  PURPOSE
* ARE  DISCLAIMED. IN  NO EVENT  SHALL LAWRENCE  LIVERMORE NATIONAL  SECURITY,
* LLC, THE  U.S.  DEPARTMENT OF  ENERGY  OR  CONTRIBUTORS BE  LIABLE  FOR  ANY
* DIRECT,  INDIRECT,   INCIDENTAL,   SPECIAL,   EXEMPLARY,  OR   CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT  LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR
* SERVICES; LOSS OF  USE, DATA, OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER
* CAUSED  AND  ON  ANY  THEORY  OF  LIABILITY,  WHETHER  IN  CONTRACT,  STRICT
* LIABILITY, OR TORT  (INCLUDING NEGLIGENCE OR OTHERWISE)  ARISING IN ANY  WAY
* OUT OF THE  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
*
* History:  
*      7/24/09 - Frain - Major reduction and cleanup of code
*                      - took all knowledge and control of data out of this code
*                      - callbacks are now registered with this file's code
*
*****************************************************************************/

/* SIMPLE SIMULATION SKELETON */
/* DATA ACCESS FUNCTIONS */
#include <VisItDataInterface_V1.h>
#include <VisItControlInterface_V1.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "SimCodeGlobals.h"

#ifdef PARALLEL
  #include <mpi.h>
#endif

#define VISIT_COMMAND_PROCESS 0
#define VISIT_COMMAND_SUCCESS 1
#define VISIT_COMMAND_FAILURE 2

#define VISIT_OKAY 0

/* Is the simulation in run mode (not waiting for VisIt input) */
static int    runFlag = 0;
static int    simcycle = 0;
static double simtime = 0.;
static int    simUpdatePlots = 0;
static int    simDone = 0;

static int par_rank = 0;
static int par_size = 1;


/* Rectilinear mesh */
#define RNX 50
#define RNY 50
int   rmesh_dims[] = {RNX, RNY, 1};
int   rmesh_ndims = 2;

/* Frain added */
static int currentDSID = -99;
static int cycle = 0;
static struct function_ptrs fptrs;

extern void processCustomGUICommand(const char* cmd);

/* end Frain added */
void simulate_one_timestep(void);
void read_input_deck(void) { }
int  simulation_done(void) { return simDone; }


void setCurrentDSID(int id){
   fprintf( stderr, " current id set to %d\n", id);
   currentDSID = id;

   /* Tell VisIt that the timestep changed. */
   VisItTimeStepChanged();

   /* Tell VisIt that we should update the plots. */
   VisItUpdatePlots();     
}

/*****************************************************************************
 * void registerCallbacks
 *
 * args: struct function_ptrs fun_ptrs 
 *
 * purpose: allows for the easy ability to change how the callbacks
 *          are performed.
 * 
 * NOTE: I do not know if VisIt checks the callback locations once at the
 *       start or dynamically checks them.
 *
 *       I am assuming they only get the callback methods once at the beginning
 *       and so I am hard coding the callbacks visit needs, but then in those
 *       methods, the registered callback (if one exists)is then called.  
 *
 * Author: Jon Frain (7/21/09)
 *
 *****************************************************************************/
void registerCallbacks( struct function_ptrs fun_ptrs){
    fptrs.GetMesh       = fun_ptrs.GetMesh;
    fptrs.GetMetaData   = fun_ptrs.GetMetaData;
    fptrs.GetScalar     = fun_ptrs.GetScalar;
    fptrs.GetDomainList = fun_ptrs.GetDomainList;
}


/*****************************************************************************
 * void ControlCommandCallback
 *
 *
 * Callback function for control commands, which are the buttons in the 
 * GUI's Simulation window. This type of command is handled automatically
 * provided that you have registered a command callback such as this.
 *
 *****************************************************************************/
void ControlCommandCallback(const char *cmd,
    int int_data, float float_data,
    const char *string_data)
{
   char cmd2[1024];

  // Frain changed strcmp == 0 to strstr != NULL as these 
    // strings are only substrings
    // - should tokenize string(around ';') to be safer
    if(strstr(cmd, "halt") != NULL)
        runFlag = 0;
    else if(strstr(cmd, "step") != NULL)
        simulate_one_timestep();
    else if(strstr(cmd, "run") != NULL)
        runFlag = 1;  
    else {
        processCustomGUICommand(cmd);
        //VisItProcessCustomCommand(cmd);      
    }
}

/* CHANGE 1 */
#ifdef PARALLEL
   static int visit_broadcast_int_callback(int *value, int sender)
   {
       return MPI_Bcast(value, 1, MPI_INT, sender, MPI_COMM_WORLD);
   }

   static int visit_broadcast_string_callback(char *str, int len, int sender)
   {
      return MPI_Bcast(str, len, MPI_CHAR, sender, MPI_COMM_WORLD);
   }
#endif


/* Helper function for ProcessVisItCommand */
static void BroadcastSlaveCommand(int *command)
{
   #ifdef PARALLEL
      MPI_Bcast(command, 1, MPI_INT, 0, MPI_COMM_WORLD);
   #endif
}

/* Callback involved in command communication. */
void SlaveProcessCallback()
{
   int command = VISIT_COMMAND_PROCESS;
   BroadcastSlaveCommand(&command);
}



/*****************************************************************************
 * void ControlCommandCallback
 *
 * Process commands from viewer on all processors. 
 *
 *****************************************************************************/
int ProcessVisItCommand(void)
{
    int command;
    if (par_rank==0)
    {  
        int success = VisItProcessEngineCommand();

        if (success)
        {
            command = VISIT_COMMAND_SUCCESS;
            BroadcastSlaveCommand(&command);
            return 1;
        }
        else
        {
            command = VISIT_COMMAND_FAILURE;
            BroadcastSlaveCommand(&command);
            return 0;
        }
    }
    else
    {
        /* Note: only through the SlaveProcessCallback callback
         * above can the rank 0 process send a VISIT_COMMAND_PROCESS
         * instruction to the non-rank 0 processes. */
        while (1)
        {
            BroadcastSlaveCommand(&command);
            switch (command)
            {
            case VISIT_COMMAND_PROCESS:
                VisItProcessEngineCommand();
                break;
            case VISIT_COMMAND_SUCCESS:
                return 1;
            case VISIT_COMMAND_FAILURE:
                return 0;
            }
        }
    }
}


/*****************************************************************************
 * void ControlCommandCallback
 *
 * Called to handle case 3 from VisItDetectInput where we have console
 * input that needs to be processed in order to accomplish an action.
 *
 *****************************************************************************/
void
ProcessConsoleCommand()
{
    /* Read A Command */
    char buff[1000];

    if (par_rank == 0)
    {
        int iseof = (fgets(buff, 1000, stdin) == NULL);
        if (iseof)
        {
            sprintf(buff, "quit");
            printf("quit\n");
        }

        if (strlen(buff)>0 && buff[strlen(buff)-1] == '\n')
            buff[strlen(buff)-1] = '\0';
    }

    #ifdef PARALLEL
       /* Broadcast the command to all processors. */
       MPI_Bcast(buff, 1000, MPI_CHAR, 0, MPI_COMM_WORLD);
    #endif

    if(strcmp(buff, "run") == 0)
        runFlag = 1;
    else if(strcmp(buff, "halt") == 0)
        runFlag = 0;
    else if(strcmp(buff, "step") == 0)
        simulate_one_timestep();
    else if(strcmp(buff, "quit") == 0)
    {
        simDone = 1;
        simUpdatePlots = 1;
    }
    else
    {
        printf("  buff = <%s>\n",buff);

        VisItProcessCustomCommand(buff);
    }  
}

/******************************************************************************
 *
 * Function: mainloop
 *
 * Purpose: Handles the program's main event loop and dispatches events to 
 *          other functions for processing.
 *
 * Programmer: Brad Whitlock
 * Date:       Thu Nov 2 17:26:08 PST 2006
 *
 * Modifications:
 *
 *****************************************************************************/

void mainloop(void)
{
    int blocking, visitstate, err = 0;


    if (par_rank == 0)
    {
        fprintf(stderr, "command> ");
        fflush(stderr);
    }

    do
    {
        blocking = runFlag ? 0 : 1;
        /* Get input from VisIt or timeout so the simulation can run. */
        if(par_rank == 0)
        {
          visitstate = VisItDetectInput(blocking, fileno(stdin));
          //visitstate = VisItCheckForNewVisItCommand( );
        }


        #ifdef PARALLEL
           /* Broadcast the return value of VisItDetectInput to all procs. */
           MPI_Bcast(&visitstate, 1, MPI_INT, 0, MPI_COMM_WORLD);
        #endif
        /* Do different things depending on the output from VisItDetectInput. */
        switch(visitstate)
        {
        case 0:
            /* There was no input from VisIt, return control to sim. */
            simulate_one_timestep();
            break;
        case 1:
            /* VisIt is trying to connect to sim. */
            if(VisItAttemptToCompleteConnection())
            {
                simUpdatePlots = 1;
                fprintf(stderr, "VisIt connected\n");

                VisItSetCommandCallback(ControlCommandCallback);
                VisItSetSlaveProcessCallback(SlaveProcessCallback);
            }
            else
                fprintf(stderr, "VisIt did not connect\n");
            break;
        case 2:
            /* VisIt wants to tell the engine something. */
            if(!ProcessVisItCommand())
            {
                fprintf(stderr,"Disconnecting on error!\n");
                /* Disconnect on an error or closed connection. */
                VisItDisconnect();
                /* Start running again if VisIt closes. */
                runFlag = 1;
                simUpdatePlots = 0;
            }
            break;
        case 3:
            /* VisItDetectInput detected console input - do something with it.
             * NOTE: you can't get here unless you pass a file descriptor to
             * VisItDetectInput instead of -1.
             */
            ProcessConsoleCommand();
            if (par_rank == 0)
            {
                fprintf(stderr, "command> ");
                fflush(stderr);
            }
            break;
        default:
            fprintf(stderr, "Can't recover from error %d!\n", visitstate);
            err = 1;
            break;
        }
    } while(!simulation_done() && err == 0);
}



/*****************************************************************************
 * void simulate_one_timestep
 *
 *****************************************************************************/
void simulate_one_timestep(void)
{
    ++simcycle;
    simtime += (M_PI / 10.);

    if(par_rank == 0)
    {
        if(simUpdatePlots == 1)
        {
            /* Tell VisIt that the timestep changed. */
            VisItTimeStepChanged();

            /* Tell VisIt that we should update the plots. */
            VisItUpdatePlots();
        }
    }
}



/******************************************************************************
 *
 * Purpose: This callback function returns metadata with 0 attributes.
 *          For when the user hasn't created any Data Sources
 *
 * Programmer: Jon Frain
 * Date:       7-28-09
 *
 * Modifications: 
 *      
 *
 *****************************************************************************/
VisIt_SimulationMetaData* returnEmptyMeta(void)
{
    VisIt_SimulationMetaData* md;

    md = (VisIt_SimulationMetaData*)malloc(sizeof(VisIt_SimulationMetaData));

    md->currentCycle = 0;
    md->currentTime  = 0;
    md->currentMode  = VISIT_SIMMODE_STOPPED;

    md->numMeshes = 0;
    md->meshes = (VisIt_MeshMetaData*)malloc(sizeof(VisIt_MeshMetaData) 
                 * md->numMeshes);

    md->meshes[0].name = strdup("mesh");
    md->meshes[0].meshType = VISIT_MESHTYPE_RECTILINEAR;
    md->meshes[0].topologicalDimension = 0;
    md->meshes[0].spatialDimension = 0;
    md->meshes[0].numBlocks = 0;
    md->meshes[0].blockTitle = strdup("Domains");
    md->meshes[0].blockPieceName = strdup("domain");
    md->meshes[0].numGroups = 0;
    md->meshes[0].units = NULL;
    md->meshes[0].xLabel = NULL;
    md->meshes[0].yLabel = NULL;
    md->meshes[0].zLabel = NULL;

    /* num attributes */
    md->numScalars = 0;
    md->scalars = (VisIt_ScalarMetaData*)malloc(sizeof(VisIt_ScalarMetaData) 
                  * md->numScalars);
    md->numMaterials   = 0;
    md->numCurves      = 0;
    md->numExpressions = 0;
    //sz = sizeof(VisIt_ExpressionMetaData) * md->numExpressions;
    md->expressions = NULL; //(VisIt_ExpressionMetaData *) malloc(sz);
    //memset(md->expressions, 0, sz); 

    fprintf(stderr, " Returning empty metadata\n");
    /* this will set up the generic and custom commands*/
    VisItInitAllCMD(md, 15);
    return md;
}


/******************************************************************************
 *
 * Purpose: This callback function returns simulation metadata.
 *
 * Programmer: Brad Whitlock
 * Date:       Fri Jan 12 13:37:17 PST 2007
 *
 * Modifications: 
 *       7/24/09 - (Frain) now call registered callback function
 *
 *****************************************************************************/
VisIt_SimulationMetaData *VisItGetMetaData(void)
{
  fprintf(stderr,"visitgetmetadata\n");
  if ( (fptrs.GetMetaData != NULL)  && (currentDSID != -99)) {
     fprintf(stderr, "Calling getmetadata via fptr!\n");
     return (fptrs.GetMetaData(currentDSID));
  }
  else   { 
     fprintf(stderr, " error: returning empty data!\n");
     return returnEmptyMeta();
  }
}

/******************************************************************************
 *
 * Purpose: This callback function returns meshes.
 *
 * Programmer: Brad Whitlock
 * Date:       Fri Jan 12 13:37:17 PST 2007
 *
 * Modifications:
 *       7/24/09 - (Frain) now call registered callback function
 *
 *****************************************************************************/

VisIt_MeshData *VisItGetMesh(int domain, const char *name)
{
  fprintf(stderr,"visitgetmesh\n");
  if ( (fptrs.GetMesh != NULL) && (currentDSID != -99))  {
    return (fptrs.GetMesh(currentDSID,domain, name));
  }
  else 
     return NULL;
}




/*****************************************************************************
* Method: VisItGetScalar
*
* Purpose:
*   This function gets the simulation scalar data.
*
* Arguments:
*   domain: the domain requested.
*   name  : the name of the variable that the data is requested from.
*
* Programmer: Jeremy Meredith
* Creation:   April  4, 2005
*
* Modifications:
*   Shelly Prevost, Tue Sep 12 12:02:21 PDT 2006
*   Enhanced function header compliance.
*
*       7/24/09 - (Frain) now call registered callback function
*
*****************************************************************************/
VisIt_ScalarData *VisItGetScalar(int domain,const char *name)
{
   fprintf(stderr,"visitgetscalar\n");
   if (fptrs.GetScalar != NULL)
     return fptrs.GetScalar(currentDSID, domain, name);
   else
      return NULL;
}




/*****************************************************************************
* Method: VisItGetDomainList
*
* Purpose:
*   This function create a list of domains and return them to the
*   calling function.
*
* Arguments:
*
* Programmer: Jeremy Meredith
* Creation:   April  4, 2005
*
* Modifications:
*   Shelly Prevost, Tue Sep 12 12:02:21 PDT 2006
*   Enhanced function header compliance.
*
*       7/24/09 - (Frain) now call registered callback function
*
*****************************************************************************/
VisIt_DomainList *VisItGetDomainList()
{
  fprintf(stderr,"visitgetdomainlist\n");
  if (( fptrs.GetDomainList != NULL) && (currentDSID != -99)) {
     return (fptrs.GetDomainList(currentDSID));
  }
  else 
     return NULL;
}

void apply(char* cmd){

}


void setupGUIHandler(){
   VisItAddConnection("apply", "clicked()", apply);
}


/*****************************************************************************
 * void runSimCode
 *
 * replaces main method from old code
 *
 *****************************************************************************/
void runSimCode(){
    /* Initialize environment variables. */
    VisItSetupEnvironment();

    /* Setup callbacks for custom GUI*/
      // seem to have an issue with their code
      setupGUIHandler();

    #ifdef PARALLEL
       /* Initialize MPI */
       MPI_Init(&argc, &argv);
       MPI_Comm_rank (MPI_COMM_WORLD, &par_rank);
       MPI_Comm_size (MPI_COMM_WORLD, &par_size);

       /* Install callback functions for global communication. */
       VisItSetBroadcastIntFunction(visit_broadcast_int_callback);
       VisItSetBroadcastStringFunction(visit_broadcast_string_callback);

       /* Tell VSIL whether the simulation is parallel. */
       VisItSetParallel(par_size > 1);
       VisItSetParallelRank(par_rank);
   #endif

   /* Write out .sim file that VisIt uses to connect. Only do it
    * on processor 0.
    */
   /* CHANGE 3 */
   if(par_rank == 0)
   {
        /* Write out .sim file that VisIt uses to connect. */
        VisItInitializeSocketAndDumpSimFile("Granite Sim Code ",
            "Renders Granite data",
            "/path/to/where/sim/was/started",
            NULL, 
            // Frain changed following to .ui file name (was NULL)
            "SimGUI_1.ui",
            NULL);
    }

    /* Call the main loop. */
    mainloop();

    #ifdef PARALLEL
       MPI_Finalize();
    #endif
}




VisIt_SimulationCallback visitCallbacks =
{
    &VisItGetMetaData,
    &VisItGetMesh,
    NULL, /* GetMaterial */
    NULL, /* GetSpecies */
    &VisItGetScalar, /* GetScalar */
    NULL, /* GetCurve */
    NULL, /* GetMixedScalar */
    &VisItGetDomainList  /* GetDomainList */
};


