/* 
    AttributeDescriptor.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
*/

//-------------- Edit history since 11/11/05 ----------------------------------
// 11/11/05 rdb; Fixed byteSize() to handle all supported StorageTypes
//-----------------------------------------------------------------------------

package edu.unh.sdb.common;

/**
 * Represents information about a single attribute of a point.
 */

public class AttributeDescriptor implements Cloneable{

    private static final int StorageTypeMask=0x1F; //  we allow up to 31 storage types. (0 is invalid)
    private static final int StorageTypeMaskBits=5;

    private TypeTable   typeTable;
    private String     	name;
    private int     	    type;
    //private int 		storageType;

    /** Create an AttributeDescriptor from the given type and name.    */
    public AttributeDescriptor( int storageType, String name) 
    {
        this.name      = name;
        this.typeTable = null;

        if ( StorageType.isSimpleStorageType(storageType) 
              && storageType != StorageType.Record)
        {
            this.type = storageType;
        } 
        else 
        {
            new IllegalArgumentError(SDBError.Recoverable,
                "Invalid storage type value: " + storageType + "  for " + name );
            this.type = StorageType.Float; 
        }
    }

    /** Create an unnamed AttributeDescriptor from the given type.    */
    public AttributeDescriptor( int storageType )
    {
       this( storageType, "" );
    }

    /** Create an unnamed AttributeDescriptor from the given type string.    */
    public AttributeDescriptor( String storageType )
    {
       this( storageType, "" );
    }

    /** Create an AttributeDescriptor from the given typestring and name and give TypeTable. 
    */
    public AttributeDescriptor( TypeTable table, String type, String name)
    {
        this.name = name;
        this.typeTable = table;

        this.type = StorageType.typeStringtoInt(type);

        if (this.type==0 )
        {
            int      userTypeCode = this.typeTable.getTypeCode( type );
            UserType uType        =(UserType) (this.typeTable.get(userTypeCode));
            this.type = (userTypeCode << StorageTypeMaskBits ) | uType.getStorageType();
        }
    }

    /** Create an AttributeDescriptor from the given typestring and name.
    */
    public AttributeDescriptor( String storageType, String name)
    {
       this( StorageType.typeStringtoInt(storageType), name );
    }
        
    /** Used by the clone method. */
    private AttributeDescriptor(){}


    /** Clones the object and returns a reference to a cloned AttributeDescriptor.  */
    public AttributeDescriptor copy()
    {
        return (AttributeDescriptor) this.clone();
    }
    
    /** Return a copy of the AttributeDescriptor as an Object. */
    public Object clone(){
    
        AttributeDescriptor n=new AttributeDescriptor();
        
        n.name= this.name;
        n.type= this.type;
        n.typeTable = this.typeTable;
        
        return n;
    }
        
    /** Return the name of the attribute as a string. */
    public String name(){
    
        return this.name;
    }

    /** True if the argument is equal to this AttributeDescriptor, false otherwise.
        Currently, two AttributeDescriptors are equal if their name and type fields
        are equal, and they are using the same TypeTable object.
        */
    public boolean equals(Object o){

        AttributeDescriptor a=(AttributeDescriptor) o;

        return (this.typeTable == a.typeTable) && (this.type==a.type) && (this.name.equals(a.name));
    }


    /** Returns true if this AttributeDescriptor has a user defined type, false otherwise. */
    public boolean isUserType(){

        return (type & ~AttributeDescriptor.StorageTypeMask) !=0;
    }


    
    /** Return the storage type of the attribute as an int constant.*/
    public int getStorageType(){
    
        return this.type & AttributeDescriptor.StorageTypeMask;
    }

    /** Extract and return the storage type encoded in the integer argument.*/
    public static int getStorageType(int type){

        return type & AttributeDescriptor.StorageTypeMask;
    }


    /** Extract and return the user type number for this AttributeDescriptor. The type number can be used
         with a TypeTable to retrieve the UserType definition.
    */
	public int getUserTypeNumber(){

        return (this.type & ~AttributeDescriptor.StorageTypeMask) >> AttributeDescriptor.StorageTypeMaskBits;
    }

    /** Extract and return the user type encoded in the integer argument. The type number can be used
         with a TypeTable to retrieve the UserType definition.
    */
    public static int getUserTypeNumber(int type){

        return (type & ~AttributeDescriptor.StorageTypeMask) >> AttributeDescriptor.StorageTypeMaskBits;
    }


    
    /** Return the type name of this AttributeDescriptor as a String. */
    public String getTypeName(){
       
        
        if(StorageType.isStorageType(this.type)){
            return StorageType.intToTypeString(this.type);
        } else {
            if(this.typeTable!=null){
                return this.typeTable.getTypeName(AttributeDescriptor.getUserTypeNumber(this.type));
            } else {
                new InternalFaultError(SDBError.Recoverable,
                    "The AttributeDescriptor contained a non-storage type, but was constructed without a TypeTable.");
                return new String("<<"+this.type+">>");
            }
        }
    }
    
//    public static final boolean isStorageType(int type){
//    
//        return ((~AttributeDescriptor.StorageTypeMask & type)==0) &&
//                    StorageType.isStorageType(AttributeDescriptor.StorageTypeMask & type);
//    }


    /** Return the number of bytes occupied by an instance of the attribute
        described by this AttributeDescriptor.*/
    public final int byteSize(){
    
        int type=this.getStorageType();
        
        switch(type){
           case StorageType.Byte:
           case StorageType.UnsignedByte:
           case StorageType.Char:
                return 1;
           case StorageType.Short:
           case StorageType.UnsignedShort:
                return 2;
            case StorageType.Float:
            case StorageType.Int:
                return 4;
            case StorageType.Double:
                return 8;
            case StorageType.Record:
                return ((RecordDescriptor)(this.typeTable.get(this.getUserTypeNumber()))).byteSize();
            default:
                new InternalFaultError(SDBError.Recoverable,
                        "Invalid type encountered in AttributeDescriptor.byteSize()");
                return 0;
        }
     }


    /** Returns a String representation of the AttributeDescriptor. */
    public String toString(){
    
        StringBuffer result=new StringBuffer("("+this.name+", ");

        result.append(this.getTypeName());
         
        result.append(")");

        return new String(result);
    }
    
    /** For debugging and testing.*/
    public static void main(String args[]){
    
    
    
        System.out.println("Storage Types Only");

        AttributeDescriptor aBoolean=new AttributeDescriptor(StorageType.Boolean,"Oxygen");
        AttributeDescriptor aByte=new AttributeDescriptor(StorageType.Byte,"Boron");
        AttributeDescriptor anUnsignedByte=new AttributeDescriptor(StorageType.Unsigned_Byte,"Helium");
        AttributeDescriptor aChar=new AttributeDescriptor(StorageType.Char,"Moron");
        AttributeDescriptor aShort=new AttributeDescriptor(StorageType.Short,"Beryllium");
        AttributeDescriptor anUnsignedShort=new AttributeDescriptor(StorageType.Unsigned_Short,"Freon");
        AttributeDescriptor anInt=new AttributeDescriptor(StorageType.Int,"Hydrogen");
        AttributeDescriptor aFloat=new AttributeDescriptor(StorageType.Float,"Carbon");
        AttributeDescriptor aDouble=new AttributeDescriptor(StorageType.Double,"Peon");
        AttributeDescriptor aLong=new AttributeDescriptor(StorageType.Long,"Nitrogen");        
        
        System.out.println(aBoolean);
        System.out.println(aByte);
        System.out.println(anUnsignedByte);
        System.out.println(aChar);
        System.out.println(aShort);
        System.out.println(anUnsignedShort);
        System.out.println(anInt);        
        System.out.println(aFloat);
        System.out.println(aDouble);
        System.out.println(aLong);

        System.out.println();
        System.out.println("With UserTypes:");

        TypeTable ttable=new TypeTable(5);

        RangedType probability=new RangedType("Probability",StorageType.Float,0,1);
        RangedType luminosity=new RangedType("Luminosity",StorageType.Int,1,10);

        ttable.insert(probability);
        ttable.insert(luminosity);

        AttributeDescriptor alpha=new AttributeDescriptor(ttable,"Luminosity","Alpha-Centauri");
        AttributeDescriptor pats=new AttributeDescriptor(ttable,"Probability","Patriots Win?");

        System.out.println("Alpha: "+alpha);
        System.out.println("Pats:"+pats);
        
    }

}

