/*
 * ByteDatum.java
 * 
 * Implements a Datum in which all attributes are of byte type.
 * 
 * ------------------------- Edit history  -----------
 * 10/21/05 rdb: Removed unused variables
 * 01/14/05 rdb: Added getStorageType() and byteSize() and getValues()
 * 01/11/05 rdb: -Redefined semantics of ByteDatum( byte[], int start, int numattr );
 *               -renamed assign methods to setBytes, but changed semantics of
 *                    setBytes( byte[], int start, int numattr ) 
 * 12/12/04 created from ShortDatum.java
 *
 */

/**
 * @author rdb
 *
 */

package edu.unh.sdb.common;
import java.util.*;


/**
*  Represent multiple data values for a single point.
*  This implementation stores values as an array of individual bytes, regardless
*  of the orginal type of an attribute. 
*/

public class ByteDatum extends Datum{

private byte [] values;


/** Creates a null datum with the specified number of values.*/
public ByteDatum( int numvalues ){
    
    this.values=new byte[numvalues];
    Arrays.fill( this.values, (byte)255 );
}

/** Creates a datum with a single attribute containing the argument value. 
 * @deprecated 
 */
public ByteDatum( byte value ){

    this.values=new byte[1];
    this.values[0]=value;    
}

/** Creates a datum with attributes containing the argument values. */
public ByteDatum( byte values[])
{
    this.values =  values.clone();
}

/** Creates a datum by copying a subset of values from an array. 
 * The second argument specifies where to start copying;
 * the 3rd argument specifies how many values to copy.
 *  
 *
	* @deprecated
	*  Not really; just looking for usage because of changed semantics (01/05)
*/
public ByteDatum( byte values[], int startIndex, int numAttributes)
{
    this.values=new byte[numAttributes];
	//rdb int startIndex=datumIndex*numAttributes;
	
	for(int i=0; i < numAttributes; i++)
	{
	   this.values[i] = values[startIndex + i];
	}
}

/** Creates a datum with attributes containing the argument values. */
public ByteDatum( int values[] )
{
  int i, n = values.length;
  this.values = new byte[values.length];

  for ( i = 0; i < n; i++ )
  {
     this.values[i] = (byte) values[i];
  }
}

/** Copy Constructor creates a copy of the argument datum.*/
public ByteDatum(ByteDatum d)
{ 
    this.values = d.values.clone();
}        


/** Copy Constructor creates a copy of the argument datum.*/
public ByteDatum(Datum d)
{ 
	this.values = new byte[ d.size() ];
	
	for( int i=0; i<d.size(); i++)
	{
		this.values[i] = d.getByte(i);
	}
}        

/** This constructor creates a single ByteDatum from an array of datums. 
 *	The Datums in the array are copied in the order they are given.
*/
public ByteDatum( ByteDatum [] datums ) 
{
	int totalSize=0;
	int datumSize;
	int index=0;
	
	for( int i=0; i< datums.length; i++)
	{
		totalSize+=datums[i].size();
	}

	this.values=new byte [totalSize];
	
	for( int i=0; i< datums.length; i++)
	{
		datumSize=datums[i].size();
		for( int j=0; j<datumSize; j++)
		{
		   this.values[index++] = datums[i].values[j];
		}
	}
}

/** Return a copy of this datum. */
public Object clone(){
    
    return new ByteDatum( this );
}    

/** Copy the values of this ByteDatum object into the argument datum. */
public void copyTo(ByteDatum d){

	try{
	
		System.arraycopy(this.values,0,d.values,0,this.values.length);
		
	} catch (ArrayIndexOutOfBoundsException e){
	
		new SDBError(SDBError.Recoverable,"Datum size mismatch.");
	}
	
	catch (NullPointerException e){
	
		if(d == null){
			new IllegalArgumentError(SDBError.Recoverable,"Argument Datum was null or had no data.");
		} else {
			new InternalFaultError(SDBError.Recoverable, "This Datum has no data.");
		}
	}
}


/** Copy the values of this ByteDatum object into the argument datum. */
public void copyTo(Datum d){

	try{
		int i,n=d.size();
		
		for(i=0; i<n; i++){
		
			d.setByte(i,this.values[i]);
		}
	} catch (ArrayIndexOutOfBoundsException e){
	
		new SDBError(SDBError.Recoverable,"Datum size mismatch.");
	}
	
	catch (NullPointerException e){
	
		if(d == null){
			new IllegalArgumentError(SDBError.Recoverable,"Argument Datum was null.");
		} else {
			new InternalFaultError(SDBError.Recoverable, "This Datum has no data.");
		}
	}
}

/** Determine if the argument is a ByteDatum with the same set of data values. */
public boolean equals( Object o )
{
    try 
    {
        ByteDatum d = (ByteDatum) o;
        if ( d.values.length != this.values.length )
           return false;
        for ( int i=0; i < this.values.length; i++ )
        {
			if ( this.values[i] != d.values[i] )
                return false;
		}
    } catch ( Exception e ) // any exception means not equal
      {
          return false;
      }
    return true;
}

/** Set the attribute indicated by the first argument to the value
 *  of the second argument.  
 */    
public void setDouble( int i, double v )
{
   try {  values[i]= (byte) v; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   catch ( ArithmeticException aex )
   { new SDBError( SDBError.Recoverable, "Error casting double to byte" ); }
}

/** Set the attribute indicated by the first argument to the value
 *  of the second argument.  
 */    
public void setFloat( int i, float v )
{
   try {  values[i]= (byte) v; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   catch ( ArithmeticException aex )
   { new SDBError( SDBError.Recoverable, "Error casting float to byte" ); }
}

/** Set the attribute indicated by the first argument to the value
 *  of the second argument.  
 */    
public void setInt( int i, int v )
{
   try {  values[i]= (byte) v; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   catch ( ArithmeticException aex )
   { new SDBError( SDBError.Recoverable, "Error casting int to byte" ); }
}

/** Set the attribute indicated by the first argument to the value
 *  of the second argument.  
 */    
public void setShort( int i, short v )
{
   try {  values[i]=  (byte) v; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   catch ( ArithmeticException aex )
   { new SDBError( SDBError.Recoverable, "Error casting short to byte" ); }
}

/** Set the attribute indicated by the first argument to the value
 *  of the second argument.  
 */    
public void setByte( int i, byte v )
{
   try {  values[i]=  v; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
}

/** Assigns values to an existing datum by copying a subset of values from an array. 
 * The second argument specifies where to start copying;
 * the 3rd argument defines how many values to copy.
 * If this object contains too few attributes to
 * 	hold the new data, an error is created. 
*/
public void setBytes( byte values[], int startIndex, int numAttributes )
{
	try
	{
		for ( int i=0; i < numAttributes; i++ )
		   this.values[i] = values[startIndex + i];
	} catch ( ArrayIndexOutOfBoundsException e )
	  {
	     new IllegalArgumentError( SDBError.Warning, 
		      "Tried to assign to a ByteDatum with too few attributes");
	  }
}

/** Return the field specified by the argument as a double.*/
public double getDouble( int i )
{
   try {  return values[i]; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   return Double.NaN;
}

/** Return the field specified by the argument as a float.*/
public float getFloat( int i )
{
   try {  return values[i]; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   return Float.NaN;
}

/** Return the field specified by the argument as a int.*/
public int getInt( int i )
{
   try {  return values[i]; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   return Integer.MAX_VALUE;
}

/** Return the field specified by the argument as a short.*/
public short getShort( int i )
{
   try {  return values[i]; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   return Byte.MAX_VALUE;
}

/** Return the field specified by the argument as a byte.*/
public byte getByte( int i )
{
   try {  return values[i]; }
   catch ( ArrayIndexOutOfBoundsException ex )
   { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
   return Byte.MAX_VALUE;
}
    
/** Return all attributes of the Datum as a 1-d byte array */
public byte[] getBytes()
{  
   // should we have a reference version of this?
   return  this.values.clone();
}

/** Return all attributes of the Datum as a Values object  */
public Values getValues()
{  
   return new ByteValues( values );
}

/** Return a new datum with the attributes indicated by the RecordSpec. The values of these
  *  attributes are taken from this ByteDatum object. 
  */
public Datum subset(RecordSpec spec){
    
	int i;
	int specsize=spec.size();
	
	ByteDatum newdatum=new ByteDatum(specsize);
	
	for(i=0;i<specsize;i++){
		newdatum.values[i]=this.values[spec.getFieldNumber(i)];
	}
	
	return newdatum;
}


/** Copy the attributes specified by the RecordSpec into the argument datum. */
public void subset(ByteDatum d, RecordSpec spec){
    

	int i;
	int specsize=spec.size();
	
	try{
		for(i=0;i<specsize;i++){
			d.values[i]=this.values[spec.getFieldNumber(i)];
		}
	} catch(ArrayIndexOutOfBoundsException e){
		new IllegalArgumentError(SDBError.Recoverable, "Argument datum was not large enough for RecordSpec.");
	}
}



   /** Copy the attributes specified by the RecordSpec into the argument datum. */
   public void subset( Datum d, RecordSpec spec )
   {
      int specsize = spec.size();

      try
      {
         for ( int i = 0; i < specsize; i++ )
         {
            d.setFloat( i, this.values[spec.getFieldNumber( i )] );
         }
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new IllegalArgumentError( SDBError.Recoverable,
                                       "RecordSpec was too large." );
      }
   }


   /** Return a string representation of the datum. */
   public String toString()
   {
      int i, size = this.values.length;
      String result = new String( "{" );

      for ( i = 0; i < size - 1; i++ )
      {
         result += values[i] + ", ";
      }
      result += values[i] + "}";
      return result;
   }

   /** Return the number of attributes in the datum. */
   public int size()
   {
      return this.values.length;
   }

   /** Return the size in bytes needed to store the data for the attributes. */
   public int byteSize()
   {
      return this.values.length;
   }

   /**
    * Return the StorageType for this Datum
    */
   public int getStorageType()
   {
      return StorageType.Byte;
   }
}
