/* 
    Datum.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    ------------------------- Edit history (started 11/24/04) -----------
    02/01/05 rdb: Added reference versions of array return methods.
    01/14/05 rdb: Added abstract getStorageType() and byteSize(), getValues().
    01/10/05 rdb: Deprecated the "assign" methods of children of Datum
                  Replaced them in the children with set<type>s versions,
                  but changed the semantics of the 3 argument version.
                  
                  Implemented set<type>s methods for all types for all Datums.
                  These are of the form
                  
                  set<type>( <type>[] );            // already existed
                  set<type>( <type>[], int start );
                  set<type>( <type>[], int start, int numAttr );
                  
                  where <type> can be double, float, int, short, byte. The
                  "start" argument is the index of where to start copying in
                  the array. Note: this is not the same as the semantics of
                  the old "assign" method in which the 2nd parameter is multiplied
                  by the 3rd in order to find the start position. The 2 argument
                  version says to copy as many attributes as are defined in the
                  Datum. Defined all versions of the "set<type>" methods at this level;
                  implemented default (not very efficient) versions. The previously
                  implemented assign versions (renamed and with the changed 
                  semantics) will provide efficiency when the data types
                  match, but defining the functionality here makes it easier for
                  applications to avoid having to make special cases.
                  
                  first version expects there to be at least as many entries as attr.
                  2nd begins reading values at "start" index of the array and copies
                     as many values as there are attributes.
                  3rd version only copies the specified number of attributes.
                  
    01/08/05 rdb: added create( RecordDescriptor ) and create( RecordDescriptor, RecordSpec)
    12/15/04 rdb: added static methods to create "appropriate" uniform type children 
                  based on type of input argument (either single variable or array)
    12/12/04 rdb: added getNumAttributes() for consistency with DataCollection terminology
    11/24/04 rdb: added missing complement of set<type> and get<type> methods
                  added array get/set methods and default implementations
             
*/

package edu.unh.sdb.common;


/**
 *  Represents multiple data values (attributes) for a single point in the index space.
 */

public abstract class Datum
{
    /** Default constructor does nothing.*/ 
	public Datum(){ }

	//-------------------- static create methods --------------------
	/**
	 * Create a single attribute Datum composed of 1 double.
	 */
	public static Datum create( double value )
	{
	   double vals[] = new double[1];
	   vals[0] = value;
	   return new DoubleDatum( vals );
	}

	/**
	 * Create a multi attribute Datum composed of double attributes.
	 */
	public static Datum create( double vals[] )
	{
	   return new DoubleDatum( vals );
	}

	/**
	 * Create a single attribute Datum composed of 1 float.
	 */
	public static Datum create( float value )
	{
	   float vals[] = new float[1];
	   vals[0] = value;
	   return new FloatDatum( vals );
	}

	/**
	 * Create a multi attribute Datum composed of double attributes.
	 */
	public static Datum create( float vals[] )
	{
	   return new FloatDatum( vals );
	}

	/**
	 * Create a single attribute Datum composed of 1 int.
	 */
	public static Datum create( int value )
	{
	   int vals[] = new int[1];
	   vals[0] = value;
	   return new IntDatum( vals );
	}

	/**
	 * Create a multi attribute Datum composed of int attributes.
	 */
	public static Datum create( int vals[] )
	{
	   return new IntDatum( vals );
	}

	/**
	 * Create a single attribute Datum composed of 1 short.
	 */
	public static Datum create( short value )
	{
	   short vals[] = new short[1];
	   vals[0] = value;
	   return new ShortDatum( vals );
	}

	/**
	 * Create a multi attribute Datum composed of short attributes.
	 */
	public static Datum create( short vals[] )
	{
	   return new ShortDatum( vals );
	}

	/**
	 * Create a single attribute Datum composed of 1 byte.
	 */
	public static Datum create( byte value )
	{
	   byte vals[] = new byte[1];
	   vals[0] = value;
	   return new ByteDatum( vals );
	}

	/**
	 * Create a multi attribute Datum composed of byte attributes.
	 */
	public static Datum create( byte vals[] )
	{
	   return new ByteDatum( vals );
	}

	/**
	 * Create a multi attribute Datum composed of mixed type attributes.
	 */
	public static Datum create( RecordDescriptor rd )
	{
	   return rd.createDatum();
	}

	/**
	 * Create a multi attribute Datum composed of mixed type attributes.
	 */
	public static Datum create( RecordDescriptor rd, RecordSpec rs )
	{
	   return rd.createDatum( rs );
	}

	/** Return a copy of this datum. */
    public abstract Object clone();    
	
    /** Copy the values of this Datum object into the argument datum. */
    public abstract void copyTo( Datum d );   
    
    //+++++++++++++++++++++++ assign methods ++++++++++++++++++++++++++++++++++
    // The assign methods are convenience functions for assigning values to 
    // Datum fields from arrays. These default implementations are not very
    // efficient. Children of Datum can override to make more efficient and 
    // should override the version in which the types match.
    
    
    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public abstract void setDouble( int i, double v );

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public abstract void setFloat( int i, float v );

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public abstract void setInt( int i, int v );

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public abstract void setShort(int i,short v);
	
    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public abstract void setByte( int i, byte v );
	
	/** Return the field specified by the argument as a double.*/
    public abstract double getDouble( int i );

	/** Return the field specified by the argument as a float.*/
    public abstract float getFloat( int i );

	/** Return the field specified by the argument as a int.*/
    public abstract int getInt( int i );

	/** Return the field specified by the argument as a short.*/
    public abstract short getShort(int i);
		    
	/** Return the field specified by the argument as a byte.*/
    public abstract byte getByte( int i );

    //---------------- get<type> array methods ------------------
    // These are "template methods". They utilize child-specific methods
    // but they work for children -- however, for performance reasons the
    // child classes might want to override.
    
    /** Return all attributes of the Datum as a 1-d double array */
    public double[] getDoubles()
    {
       return getDoubles( null );
    }
    
    /** Return all attributes of the Datum in the passed array.
     *  Passed array is returned for convenience, but if it is null or
     *  not the right length, a new array is created and returned.  */
    public double[] getDoubles( double[] fill )
    {
       if ( fill == null || fill.length != this.size() )
          fill = new double[ this.size() ];
       for ( int i=0; i < this.size(); i++ )
          fill[i] = getDouble( i );
       return fill;
    }
    
    /** Return all attributes of the Datum as a 1-d float array */
    public float[] getFloats()
    {
       return getFloats( null );
    }
    
    /** Return all attributes of the Datum in the passed array.
     *  Passed array is returned for convenience, but if it is null or
     *  not the right length, a new array is created and returned.  */
    public float[] getFloats( float[] fill )
    {
       if ( fill == null || fill.length != this.size() )
          fill = new float[ this.size() ];
       for ( int i=0; i < this.size(); i++ )
          fill[i] = getFloat( i );
       return fill;
    }
    
    /** Return all attributes of the Datum as a 1-d int array */
    public int[] getInts()
    {
       return getInts( null );
    }
    
    /** Return all attributes of the Datum in the passed array.
     *  Passed array is returned for convenience, but if it is null or
     *  not the right length, a new array is created and returned.  */
    public int[] getInts( int[] fill )
    {
       if ( fill == null || fill.length != this.size() )
          fill = new int[ this.size() ];
       for ( int i=0; i < this.size(); i++ )
          fill[i] = getInt( i );
       return fill;
    }
    
    /** Return all attributes of the Datum as a 1-d short array */
    public short[] getShorts()
    {
       return getShorts( null );
    }
    
    /** Return all attributes of the Datum in the passed array.
     *  Passed array is returned for convenience, but if it is null or
     *  not the right length, a new array is created and returned.  */
    public short[] getShorts( short[] fill )
    {
       if ( fill == null || fill.length != this.size() )
          fill = new short[ this.size() ];
       for ( int i=0; i < this.size(); i++ )
          fill[i] = getShort( i );
       return fill;
    }
    
    /** Return all attributes of the Datum as a 1-d byte array */
    public byte[] getBytes()
    {
       return getBytes( null );
    }
    
    /** Return all attributes of the Datum in the passed array.
     *  Passed array is returned for convenience, but if it is null or
     *  not the right length, a new array is created and returned.  */
    public byte[] getBytes( byte[] fill )
    {
       if ( fill == null || fill.length < this.size() )
          fill = new byte[ this.size() ];
       for ( int i=0; i < this.size(); i++ )
          fill[i] = getByte( i );
       return fill;
    }
    
    /**
     * Returns all the attributes of the Datum in the passed array,
     * storing them starting at 'index'. It is an error for 'destination' to
     * be null. 'Index' should be within bounds of 'destination' array.
     */
    public void getBytes (byte[] destination, int index)
    {
        for (int i = 0; i < size (); i++)
            destination [index + i] = getByte (i);
    }
    
    
    /** Return all attributes of the Datum in a Values object */
    abstract public Values getValues();
    
    //---------------- set<type> array methods ------------------
    // These are "template methods". They utilize child-specific methods
    // but they work for children -- however, for performance reasons the
    // child classes might want to override.
    
    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setDoubles( double [] data )
    {
       setDoubles( data, 0, this.size() );
    }

    /** Copy values from a piece of the argument array to this datum. The 2nd 
     * argument indicates where in the array to start the copy. Copy enough 
     * data to fill the Datum. It is an error if there is not enough data.
     */
    public void setDoubles( double [] data, int start )
    {
       setDoubles( data, start, this.size() );
    }

    /** Assigns values to an existing datum by copying a subset of values from an array. 
     * The second argument specifies where to start copying; the 3rd argument defines 
     * how many values to copy.
     * If the object contains too few attributes to hold the new data, an error occurs. 
     */
    public void setDoubles( double values[], int startIndex, int numAttributes )
    {
       try
       {
          for ( int i = 0; i < numAttributes; i++ )
             this.setDouble( i, values[startIndex + i] );
       } catch ( ArrayIndexOutOfBoundsException e )
       {
         new IllegalArgumentError( SDBError.Recoverable, "Datum.assign: data count mismatch." );
       }
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setFloats( float [] data )
    {
       setFloats( data, 0, this.size() );
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setFloats( float [] data, int start )
    {
       setFloats( data, start, this.size() );
    }

    /** Assigns values to an existing datum by copying a subset of values from an array. 
     * The second argument specifies where to start copying; the 3rd argument defines 
     * how many values to copy.
     * If the object contains too few attributes to hold the new data, an error occurs. 
     */
    public void setFloats( float values[], int startIndex, int numAttributes )
    {
       try
       {
          for ( int i = 0; i < numAttributes; i++ )
             this.setFloat( i, values[startIndex + i] );
       } catch ( ArrayIndexOutOfBoundsException e )
       {
         new IllegalArgumentError( SDBError.Recoverable, "Datum.assign: data count mismatch." );
       }
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setInts( int [] data )
    {
       setInts( data, 0, this.size() );
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setInts( int [] data, int start )
    {
       setInts( data, start, this.size() );
    }

    /** Assigns values to an existing datum by copying a subset of values from an array. 
     * The second argument specifies where to start copying; the 3rd argument defines 
     * how many values to copy.
     * If the object contains too few attributes to hold the new data, an error occurs. 
     */
    public void setInts( int values[], int startIndex, int numAttributes )
    {
       try
       {
          for ( int i = 0; i < numAttributes; i++ )
             this.setInt( i, values[startIndex + i] );
       } catch ( ArrayIndexOutOfBoundsException e )
       {
         new IllegalArgumentError( SDBError.Recoverable, "Datum.assign: data count mismatch." );
       }
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setShorts( short [] data )
    {
       setShorts( data, 0, this.size() );
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setShorts( short [] data, int start )
    {
       setShorts( data, start, this.size() );
    }

    /** Assigns values to an existing datum by copying a subset of values from an array. 
     * The second argument specifies where to start copying; the 3rd argument defines 
     * how many values to copy.
     * If the object contains too few attributes to hold the new data, an error occurs.
     */
    public void setShorts( short values[], int startIndex, int numAttributes )
    {
       try
       {
          for ( int i = 0; i < numAttributes; i++ )
             this.setShort( i, values[startIndex + i] );
       } catch ( ArrayIndexOutOfBoundsException e )
       {
         new IllegalArgumentError( SDBError.Recoverable, "Datum.assign: data count mismatch." );
       }
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setBytes( byte [] data )
    {
       setBytes( data, 0, this.size() );
    }

    /** Copy the values in the argument array to this datum. The array length 
     *  should be <= the number of attributes in the datum.
     */
    public void setBytes( byte [] data, int start )
    {
       setBytes( data, start, this.size() );
    }

    /** Assigns values to an existing datum by copying a subset of values from an array. 
     * The second argument specifies where to start copying; the 3rd argument defines 
     * how many values to copy.
     * If the object contains too few attributes to hold the new data, an error occurs. 
     */
    public void setBytes( byte values[], int startIndex, int numAttributes )
    {
       try
       {
          for ( int i = 0; i < numAttributes; i++ )
             this.setByte( i, values[startIndex + i] );
       } catch ( ArrayIndexOutOfBoundsException e )
       {
         new IllegalArgumentError( SDBError.Recoverable, "Datum.assign: data count mismatch." );
       }
    }

    /** Return a new datum with the attributes indicated by the RecordSpec. The values of these
      *  attributes are taken from this Datum object. 
      */
    public abstract Datum subset(RecordSpec spec);

    /** Copy the attributes specified by the RecordSpec into the argument datum. */
    public abstract void subset(Datum d, RecordSpec spec);

    /**
     * Return the StorageType for this Datum.
     */
    abstract public int getStorageType( );
     
    /** Return the number of attributes in the datum. */
    public abstract int size(); 
    
    /** Return the size in bytes required to store the attributes in the datum. */
    public abstract int byteSize(); 
    
    /** Return the number of attributes in the datum. 
     *  Alternative to size() that is consistent with the same method in other
     *  classes.
     */
    final public int getNumAttributes() { return size(); }
    
    /** Returns the number of bytes used to store the data in the Datum. */
        
}
