 /* 
 DoubleDatum.java
 Copyright 2003 University of New Hampshire
 Durham, NH 03824
 ------------------------- Edit history (started 11/24/04) -----------

 01/14/05 rdb: Added getStorageType() and byteSize() and getValues()
 01/11/05 rdb: -Redefined semantics of DoubleDatum( double[], int start, int numattr );
               -implemented
                      setDatums( double[], int start, int numattr ) 
 11/24/04 rdb: 
		 added missing complement of set<type> and get<type> methods
		 added override of Datum.getDoubles() and Datum.setDoubles
		 also added try-catch around attribute index usage.
 */

package edu.unh.sdb.common;

import java.util.*;


/**
 * Represent multiple data values for a single point. This implementation stores
 * values as an array of doubles, regardless of the orginal type of an attribute.
 */

public class DoubleDatum extends Datum
{

   private double[] values;

   /** Creates a null datum with the specified number of values. */
   public DoubleDatum( int numvalues )
   {
      this.values = new double[numvalues];
      Arrays.fill( this.values, Double.NaN );
   }

   /**
    * Creates a datum with a single attribute containing the argument value.
    * 
    * @deprecated
    */
   public DoubleDatum( double value )
   {
      this.values = new double[1];
      this.values[0] = value;
   }

   /** Creates a datum with attributes containing the argument values. */
   public DoubleDatum( double values[] )
   {
      this.values = values.clone();
   }

   /** Creates a datum by copying a subset of values from an array. 
    * The second argument specifies where to start copying;
    * the 3rd argument specifies how many values to copy.
    *
   	* @deprecated
    *  Not really; just looking for usage because of changed semantics (01/05)
   */
   public DoubleDatum( double values[], int startIndex, int numAttributes )
   {
      this.values = new double[ numAttributes ];
      //rdb int startIndex = startIndex * numAttributes;

      for ( int i = 0; i < numAttributes; i++ )
      {
         this.values[i] = values[startIndex + i];
      }
   }

   /** Creates a datum with attributes containing the argument values. */
   public DoubleDatum( int values[] )
   {

      int i, n = values.length;

      this.values = new double[values.length];

      for ( i = 0; i < n; i++ )
      {

         this.values[i] = values[i];
      }

   }

   /** Copy Constructor creates a copy of the argument datum. */
   public DoubleDatum( DoubleDatum d )
   {

      this.values = d.values.clone();
   }

   /** Copy Constructor creates a copy of the argument datum. */
   public DoubleDatum( Datum d )
   {

      int i, n = d.size();

      this.values = new double[n];

      for ( i = 0; i < n; i++ )
      {

         this.values[i] = d.getDouble( i );

      }

      //this.values=(double []) d.values.clone();
   }

   /**
    * This constructor creates a single DoubleDatum from an array of datums. The
    * Datums in the array are copied in the order they are given.
    */
   public DoubleDatum( DoubleDatum[] datums )
   {

      int i, j;
      int totalSize = 0;
      int datumSize;
      int index = 0;

      for ( i = 0; i < datums.length; i++ )
      {

         totalSize += datums[i].size();

      }

      this.values = new double[totalSize];

      for ( i = 0; i < datums.length; i++ )
      {

         datumSize = datums[i].size();

         for ( j = 0; j < datumSize; j++ )
         {
            this.values[index++ ] = datums[i].values[j];
         }
      }

   }

   /** Return a copy of this datum. */
   public Object clone()
   {

      return new DoubleDatum( this );
   }

   /** Copy the values of this DoubleDatum object into the argument datum. */
   public void copyTo( DoubleDatum d )
   {

      try
      {

         System.arraycopy( this.values, 0, d.values, 0, this.values.length );

      } catch ( ArrayIndexOutOfBoundsException e )
      {

         new SDBError( SDBError.Recoverable, "Datum size mismatch." );
      }

      catch ( NullPointerException e )
      {

         if ( d == null )
         {
            new IllegalArgumentError( SDBError.Recoverable,
                  "Argument Datum was null or had no data." );
         }
         else
         {
            new InternalFaultError( SDBError.Recoverable,
                  "This Datum has no data." );
         }
      }
   }

   /** Copy the values of this DoubleDatum object into the argument datum. */
   public void copyTo( Datum d )
   {

      try
      {

         int i, n = d.size();

         for ( i = 0; i < n; i++ )
         {

            d.setDouble( i, this.values[i] );
         }

      } catch ( ArrayIndexOutOfBoundsException e )
      {

         new SDBError( SDBError.Recoverable, "Datum size mismatch." );
      }

      catch ( NullPointerException e )
      {

         if ( d == null )
         {
            new IllegalArgumentError( SDBError.Recoverable,
                  "Argument Datum was null." );
         }
         else
         {
            new InternalFaultError( SDBError.Recoverable,
                  "This Datum has no data." );
         }
      }
   }

   /** Determine if the argument is a DoubleDatum with the same set of data values. */
   public boolean equals( Object o )
   {
       try 
       {
           DoubleDatum d = (DoubleDatum) o;
           if ( d.values.length != this.values.length )
              return false;
           for ( int i=0; i < this.values.length; i++ )
           {
   			if ( this.values[i] != d.values[i] )
                   return false;
   		}
       } catch ( Exception e ) // any exception means not equal
         {
             return false;
         }
       return true;
   }

   /** Set the field specified by the argument to null. */
   public void setNull( int i )
   {
      try
      {
         values[i] = Double.NaN;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setDouble( int i, double v )
   {
      try
      {
         values[i] = v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } 
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setFloat( int i, float v )
   {
      try
      {
         values[i] = v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } 
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setInt( int i, int v )
   {
      try
      {
         values[i] = v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } 
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setShort( int i, short v )
   {
      try
      {
         values[i] = v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setByte( int i, byte v )
   {
      try
      {
         values[i] = v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
   }

	/** Assigns values to an existing datum by copying a subset of values from an array. 
	 * The second argument specifies where to start copying;
	 * the 3rd argument defines how many values to copy.
	 *  If this object contains too few attributes to
	 * 	hold the new data, an error is created. 
	*/
	public void setDoubles( double values[], int startIndex, int numAttributes )
	{
	   try
	   {
           for ( int i=0; i < numAttributes; i++ )
              this.values[i] = values[startIndex + i];
	   } catch ( ArrayIndexOutOfBoundsException e )
		 {
             new IllegalArgumentError( SDBError.Warning, 
			      "Tried to assign to a DoubleDatum with too few attributes");			
		 }
	}

    /** Assigns values to an existing datum by copying a subset of values from an array. 
     *  The second argument specifies where to start copying;
     *  the 3rd argument defines how many values to copy.
     *  If this object contains too few attributes to
     * 	hold the new data, an error is created. 
     */
	public void setShorts(short values[], int startIndex, int numAttributes)
	{		
		try
		{
			for ( int i=0; i < numAttributes; i++ )
			   this.values[i] = values[startIndex + i];
		} catch (ArrayIndexOutOfBoundsException e)
		  {
		     new IllegalArgumentError(SDBError.Recoverable, 
		         "Tried to assign to a ShortDatum with too few attributes");
		  }
	}

   /** Return all attributes of the Datum as a 1-d double array */
   public double[] getDoubles()
   {
      // should we have a reference version of this?
      return this.values.clone();
   }

   /** Return the field specified by the argument as a double. */
   public double getDouble( int i )
   {
      try
      {
         return values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
      return Double.NaN;
   }

   /** Return the field specified by the argument as a double. */
   public float getFloat( int i )
   {
      try
      {
         return (float) values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting double to float" );
      }
      return Float.NaN;
   }

   /** Return the field specified by the argument as a int. */
   public int getInt( int i )
   {
      try
      {
         return (int) values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting double to int" );
      }
      return Integer.MAX_VALUE;
   }

   /** Return the field specified by the argument as a short. */
   public short getShort( int i )
   {
      try
      {
         return (short) values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting double to short" );
      }
      return Short.MAX_VALUE;
   }

   /** Return the field specified by the argument as a byte. */
   public byte getByte( int i )
   {
      try
      {
         return (byte) values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting double to byte" );
      }
      return Byte.MAX_VALUE;
   }

   /** Return all attributes of the Datum as a Values object  */
   public Values getValues()
   {  
      return new DoubleValues( values );
   }

   /**
    * Returns true if the field specified by the argument is null, false
    * otherwise.
    */
   public boolean isNull( int i )
   {
      try
      {
         return Double.isNaN( values[i] );
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
      return true;
   }

   /** Return a new datum with the attributes indicated by the RecordSpec. The values of these
    *  attributes are taken from this DoubleDatum object. 
    */
   public Datum subset( RecordSpec spec )
   {
      int i;
      int specsize = spec.size();

      DoubleDatum newdatum = new DoubleDatum( specsize );

      for ( i = 0; i < specsize; i++ )
      {
         newdatum.values[i] = this.values[spec.getFieldNumber( i )];
      }
      return newdatum;
   }

   /** Copy the attributes specified by the RecordSpec into the argument datum. */
   public void subset( DoubleDatum d, RecordSpec spec )
   {
      int i;
      int specsize = spec.size();

      try
      {
         for ( i = 0; i < specsize; i++ )
         {
            d.values[i] = this.values[spec.getFieldNumber( i )];
         }
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Argument datum was not large enough for RecordSpec." );
      }
   }

   /** Copy the attributes specified by the RecordSpec into the argument datum. */
   public void subset( Datum d, RecordSpec spec )
   {
      int i;
      int specsize = spec.size();
      try
      {
         for ( i = 0; i < specsize; i++ )
         {
            d.setDouble( i, this.values[spec.getFieldNumber( i )] );
         }
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "RecordSpec was too large." );
      }
   }

   /** Return a string representation of the datum. */
   public String toString()
   {
      int i, size = this.values.length;
      double value;

      String result = new String( "{" );
      for ( i = 0; i < size - 1; i++ )
      {
         value = values[i];

         if ( Double.isNaN( value ) )
         {
            result += "null, ";
         }
         else result += values[i] + ", ";
      }
      result += values[i] + "}";
      return result;
   }

   /** Return the number of attributes in the datum. */
   public int size()
   {
      return this.values.length;
   }

   /** Return the size in bytes needed to store the data for the attributes. */
   public int byteSize()
   {
      return this.values.length*8;
   }

   /**
    * Return the StorageType for this Datum
    */
   public int getStorageType()
   {
      return StorageType.Double;
   }

}