/* 
    FloatDatum.java
    Copyright 2003 University of New Hampshire
    Durham, NH 03824
    ------------------------- Edit history (started 11/24/04) -----------
    08/31/05 rdb: Removed phony deprecation on FloatDatum(float[], int, int )
                  (First done on 5/28, but lost).
    01/14/05 rdb: Added getStorageType() and byteSize() and getValues()
    01/11/05 rdb: -Redefined semantics of FloatDatum( float[], int start, int numattr );
                  -deprecated assign( float[], int datumIndex, int numattr )
                    and       assign( float[] )
                  -renamed assign methods to setFloats, but changed semantics of
                      setFloats( float[], int start, int numattr ) 
    
    11/24/04 rdb: 
             -added missing complement of set<type> and get<type> methods
             -added override of Datum.getFloats() and Datum.setFloats
             -also added try-catch around attribute index usage.
*/

package edu.unh.sdb.common;
import java.util.*;


/**
 *  Represent multiple data values for a single point.
 *  This implementation stores values as an array of floats, regardless
 *  of the orginal type of an attribute. 
 */

public class FloatDatum extends Datum
{
   private float [] values;
    
    /** Creates a null datum with the specified number of values.*/
    public FloatDatum( int numvalues )
    {
        this.values=new float[numvalues];
        Arrays.fill(this.values,Float.NaN);
    }
    
	/** Creates a datum with a single attribute containing the argument value. 
	 *  @deprecated
	 */
    public FloatDatum( float value )
    {
        this.values=new float[1];
        this.values[0]=value;    
    }
    
    /** Creates a datum with attributes containing the argument values. */
    public FloatDatum( float values[] )
    {
        this.values= values.clone();
    }
	
	/** Creates a datum by copying a subset of values from an array. 
	 * The second argument specifies where to start copying;
	 * the 3rd argument specifies how many values to copy.
	 *  
     */
	public FloatDatum( float values[], int startIndex, int count )
	{
		this.values=new float[count];
		//rdb int startIndex=startIndex*numAttributes;
		
		for(int i=0; i < count; i++)
		{
			this.values[i] = values[startIndex + i];
		}
	}
	
	/** Creates a datum with attributes containing the argument values. 
	 *  @deprecated
	 */
    public FloatDatum( int values[])
    {
	    int n = values.length;
        this.values=new float[values.length];

		for(int i=0; i<n; i++)
		{
			this.values[i]= values[i];
		}
    }
	
	
    /** Copy Constructor creates a copy of the argument datum.*/
    public FloatDatum( FloatDatum d )
    { 
        this.values= d.values.clone();
    }        


    /** Copy Constructor creates a copy of the argument datum.*/
    public FloatDatum( Datum d )
    { 
		int n = d.size();
		this.values=new float[n];
		
		for ( int i=0; i<n; i++ )
		{
			this.values[i] = d.getFloat(i);
		}
        //this.values=(float []) d.values.clone();
    }        
	
	/** This constructor creates a single FloatDatum from an array of datums. 
	 *	The Datums in the array are copied in the order they are given.
	*/
	public FloatDatum( FloatDatum [] datums )
	{
		int i,j;
		int totalSize=0;
		int datumSize;
		int index=0;
		
		for(i=0; i< datums.length; i++){
			
			totalSize+=datums[i].size();
		
		}
	
		this.values=new float [totalSize];
		
		for(i=0; i< datums.length; i++){
		
			datumSize=datums[i].size();
			
			for(j=0; j<datumSize; j++){
				this.values[index++]=datums[i].values[j];
			}
		}
	}
	
	/** Return a copy of this datum. */
    public Object clone()
    {
        return new FloatDatum( this );
    }    
    
    /** Copy the values of this FloatDatum object into the argument datum. */
    public void copyTo(FloatDatum d){
	
		try{
		
			System.arraycopy(this.values,0,d.values,0,this.values.length);
			
		} catch (ArrayIndexOutOfBoundsException e){
		
			new SDBError(SDBError.Recoverable,"Datum size mismatch.");
		}
		
		catch (NullPointerException e){
		
			if(d == null){
				new IllegalArgumentError(SDBError.Recoverable,"Argument Datum was null or had no data.");
			} else {
				new InternalFaultError(SDBError.Recoverable, "This Datum has no data.");
			}
		}
    }


    /** Copy the values of this FloatDatum object into the argument datum. */
    public void copyTo( Datum d )
    {
		try
		{
			for ( int i=0; i < d.size(); i++ )
			{
				d.setFloat( i, this.values[i] );
			}
		} catch (ArrayIndexOutOfBoundsException e)
		{
			new SDBError(SDBError.Recoverable,"Datum size mismatch.");
		}
		catch (NullPointerException e)
		{
			if(d == null){
				new IllegalArgumentError(SDBError.Recoverable,"Argument Datum was null.");
			} else {
				new InternalFaultError(SDBError.Recoverable, "This Datum has no data.");
			}
		}
    }
    
	/** Determine if the argument is a FloatDatum with the same set of data values. */
	public boolean equals( Object o )
	{
	    try 
	    {
	        FloatDatum d = (FloatDatum) o;
	        if ( d.values.length != this.values.length )
	           return false;
	        for ( int i=0; i < this.values.length; i++ )
	        {
				if ( this.values[i] != d.values[i] )
	                return false;
			}
	    } catch ( Exception e ) // any exception means not equal
	      {
	          return false;
	      }
	    return true;
	}

	/** Set the field specified by the argument to null.*/
    public void setNull( int i )
    {
       try {  values[i] = Float.NaN; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
    }
    
    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setDouble( int i, double v )
    {
       try {  values[i] = (float) v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting double to float" ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setFloat( int i, float v )
    {
       try {  values[i] =  v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setInt( int i, int v )
    {
       try {  values[i]= v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting int to float" ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setShort( int i, short v )
    {
       try {  values[i]= v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
    }

    /** Set the attribute indicated by the first argument to the value
     *  of the second argument.  
     */    
    public void setByte( int i, byte v )
    {
       try {  values[i]= v; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
    }
    
	/** Assigns values to an existing datum by copying a subset of values from an array. 
	 * The second argument specifies where to start copying;
	 * the 3rd argument defines how many values to copy.
 	 *  If this object contains too few attributes to
 	 * 	hold the new data, an error is created. 
	*/
	public void setFloats( float values[], int startIndex, int numAttributes )
	{
	   try
	   {
           for ( int i=0; i < numAttributes; i++ )
              this.values[i] = values[startIndex + i];
	   } catch ( ArrayIndexOutOfBoundsException e )
		 {
			new IllegalArgumentError(SDBError.Recoverable, 
			      "Tried to assign to a FloatDatum with too few attributes");			
		 }
	}

    /** Copy the values in the argument to this datum. 
     * @deprecated
     * Use setFloats( float[] )
     */
    public void assign( float [] data) 
    {
       setFloats( data );
    }

	/** Assigns values to an existing datum by copying a subset of values from an array. 
	 * The second argument specifies where to start copying; the array start index is
	 * 2nd arg * 3rd argument. The 3rd argument defines how many values to copy.
 	 *  If this object contains too few attributes to
 	 * 	hold the new data, an error is created. 
 	 * @deprecated
 	 * Use setFloats( float[], int, int ), but note that the semantics of the start
 	 * index is different in setFloats; the second argument is the array index to
 	 * start the copy.
	*/
	public void assign(float values[], int startIndex, int numAttributes)
	{
	   setFloats( values, startIndex*numAttributes, numAttributes );
	}
	
    /** Return all attributes of the Datum as a 1-d double array */
    public float[] getFloats()
    { 
       // should we have a reference version of this?
       return  values.clone();
    }

    /** Return the field specified by the argument as a double.*/
    public double getDouble( int i )
    {
       try {  return values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       return Double.NaN;
    }

    /** Return the field specified by the argument as a float.*/
    public float getFloat( int i )
    {
       try {  return values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       return Float.NaN;
    }

    /** Return the field specified by the argument as a int.*/
    public int getInt( int i )
    {
       try {  return (int) values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting float to int" ); }
       return Integer.MAX_VALUE;
    }

    /** Return the field specified by the argument as a short.*/
    public short getShort( int i )
    {
       try {  return (short) values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting float to short" ); }
       return Short.MAX_VALUE;
    }

    /** Return the field specified by the argument as a byte.*/
    public byte getByte( int i )
    {
       try {  return (byte) values[i]; }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       catch ( ArithmeticException aex )
       { new SDBError( SDBError.Recoverable, "Error casting float to byte" ); }
       return Byte.MAX_VALUE;
    }
    
    /** Return all attributes of the Datum as a Values object  */
    public Values getValues()
    {  
       return new FloatValues( values );
    }

	/** Returns true if the field specified by the argument is null, false otherwise.*/
    public boolean isNull( int i )
    {
       try {  return Float.isNaN(values[i]); }
       catch ( ArrayIndexOutOfBoundsException ex )
       { new SDBError( SDBError.Recoverable, "Invalid attr index: " + i ); }
       return true;
    }    

    /** Return a new datum with the attributes indicated by the RecordSpec. The values of these
      *  attributes are taken from this FloatDatum object. 
      */
    public Datum subset( RecordSpec spec )
    {
 		int i;
		int specsize=spec.size();
		FloatDatum newdatum = new FloatDatum(specsize);
		
		for(i=0;i<specsize;i++){
			newdatum.values[i]=this.values[spec.getFieldNumber(i)];
		}
		return newdatum;
    }

    /** Copy the attributes specified by the RecordSpec into the argument datum. */
    public void subset(FloatDatum d, RecordSpec spec)
    {
		int i;
		int specsize = spec.size();
		
		try{
			for(i=0;i<specsize;i++){
				d.values[i]=this.values[spec.getFieldNumber(i)];
			}
		} catch(ArrayIndexOutOfBoundsException e){
			new IllegalArgumentError(SDBError.Recoverable, "Argument datum was not large enough for RecordSpec.");
		}
    }



    /** Copy the attributes specified by the RecordSpec into the argument datum. */
    public void subset(Datum d, RecordSpec spec){
        

		int i;
		int specsize=spec.size();
		
		try{
			for(i=0;i<specsize;i++){
				d.setFloat(i, this.values[spec.getFieldNumber(i)]);
			}
		} catch(ArrayIndexOutOfBoundsException e){
			new IllegalArgumentError(SDBError.Recoverable, "RecordSpec was too large.");
		}
    }



    
    /** Return a string representation of the datum. */
    public String toString(){
    
        int i, size=this.values.length;
        float value;
        
        String result=new String("{");
        
        for(i=0;i<size-1;i++){
            value=values[i];
            
            if(Float.isNaN(value)){
                result+="null, ";
            } else
                result+=values[i]+", ";
        }
        
        result+=values[i]+"}";
        
        return result;        
    }
    
    
    /** Return the number of attributes in the datum. */
    public int size(){

        return this.values.length;
    }

    /** Return the size in bytes needed to store the data for the attributes. */
    public int byteSize()
    {
       return this.values.length*4;
    }

    /**
     * Return the StorageType for this Datum
     */
    public int getStorageType()
    {
       return StorageType.Float;
    }
}
