/*
 * IntDatum.java
 * 
 * Implements a Datum in which all attributes are of int type.
 * 
 * ------------------------- Edit history  -----------
 * 10/21/05 rdb: Removed unused variables
 * 01/14/05 rdb: Added getStorageType() and byteSize()
 * 01/10/05 rdb: Redefined semantics of
                 IntDatum( int[], int start, int numattr );
                 assign  ( int[], int start, int numattr );

 * 12/12/04 from ShortDatum.java
 *
 */

/**
 * @author rdb
 *
 */

package edu.unh.sdb.common;

import java.util.*;


/**
 * Represent multiple data values for a single point. This implementation stores
 * values as an array of ints, regardless of the orginal type of an attribute.
 */

public class IntDatum extends Datum
{

   private int[] values;

   /** Creates a null datum with the specified number of values. */
   public IntDatum( int numvalues )
   {
      this.values = new int[numvalues];
      Arrays.fill( this.values, 255 );
   }

   /** Creates a datum with attributes containing the argument values. */
   public IntDatum( int values[] )
   {
      this.values = values.clone();
   }

	/** Creates a datum by copying a subset of values from an array. 
	 * The second argument specifies where to start copying;
	 * the 3rd argument specifies how many values to copy.
	 *  
     * @deprecated
	 *  Not really; just looking for usage because of changed semantics (01/05)
	 */
   public IntDatum( int values[], int startIndex, int numAttributes )
   {
      this.values = new int[numAttributes];
      //rdb int startIndex = datumIndex * numAttributes;

      for ( int i = 0; i < numAttributes; i++ )
      {
         this.values[i] = values[startIndex + i];
      }
   }

   /** Copy Constructor creates a copy of the argument datum. */
   public IntDatum( IntDatum d )
   {
      this.values = d.values.clone();
   }

   /** Copy Constructor creates a copy of the argument datum. */
   public IntDatum( Datum d )
   {
      int n = d.size();
      this.values = new int[n];

      for ( int i = 0; i < n; i++ )
      {
         this.values[i] = d.getInt( i );
      }
   }

   /**
    * This constructor creates a single IntDatum from an array of datums. The
    * Datums in the array are copied in the order they are given.
    */
   public IntDatum( IntDatum[] datums )
   {

      int i, j;
      int totalSize = 0;
      int datumSize;
      int index = 0;

      for ( i = 0; i < datums.length; i++ )
      {
         totalSize += datums[i].size();
      }

      this.values = new int[totalSize];

      for ( i = 0; i < datums.length; i++ )
      {
         datumSize = datums[i].size();
         for ( j = 0; j < datumSize; j++ )
         {
            this.values[index++ ] = datums[i].values[j];
         }
      }
   }

   /** Return a copy of this datum. */
   public Object clone()
   {
      return new IntDatum( this );
   }

   /** Copy the values of this IntDatum object into the argument datum. */
   public void copyTo( IntDatum d )
   {
      try
      {
         System.arraycopy( this.values, 0, d.values, 0, this.values.length );
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new SDBError( SDBError.Recoverable, "Datum size mismatch." );
      }
      catch ( NullPointerException e )
      {
         if ( d == null )
         {
            new IllegalArgumentError( SDBError.Recoverable,
                  "Argument Datum was null or had no data." );
         }
         else
         {
            new InternalFaultError( SDBError.Recoverable,
                  "This Datum has no data." );
         }
      }
   }

   /** Copy the values of this IntDatum object into the argument datum. */
   public void copyTo( Datum d )
   {
      try
      {
         int n = d.size();

         for ( int i = 0; i < n; i++ )
         {
            d.setInt( i, this.values[i] );
         }
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new SDBError( SDBError.Recoverable, "Datum size mismatch." );
      }
      catch ( NullPointerException e )
      {
         if ( d == null )
         {
            new IllegalArgumentError( SDBError.Recoverable,
                  "Argument Datum was null." );
         }
         else
         {
            new InternalFaultError( SDBError.Recoverable,
                  "This Datum has no data." );
         }
      }
   }

   /** Copy the values in the argument to this datum. */
   public void assign( int[] data )
   {
      //System.arraycopy(data,0,this.values,0,data.length);
      int length = data.length;

      for ( int i = 0; i < length; i++ )
      {
         this.values[i] = data[i];
      }
   }

   /** Assigns values to an existing datum by copying a subset of values from an array. 
    * The second argument specifies where to start copying;
    * the 3rd argument defines how many values to copy.
    * If this object contains too few attributes to
    * 	hold the new data, an error is created. 
    * 
    * @deprecated
    *  Not really; just looking for usage because of changed semantics (01/05)
    */
   public void assign( int values[], int startIndex, int numAttributes )
   {
      //rdb int startIndex = startIndex * numAttributes;
      try
      {
         for ( int i = 0; i < numAttributes; i++ )
         {
            this.values[i] = values[startIndex + i];
         }
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Tried to assign to a IntDatum with too few attributes" );
      }
   }

   public boolean equals( Object o )
   {
      try
      {
         IntDatum d = (IntDatum) o;
         int dim = this.values.length;

         for ( int i = 0; i < dim; i++ )
         {
            if ( this.values[i] != d.values[i] )
            {
               return false;
            }
         }
         return true;
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         return false;
      }
      catch ( ClassCastException e )
      {
         return false;
      }
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setDouble( int i, double v )
   {
      try
      {
         values[i] = (int) v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting double to int" );
      }
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setFloat( int i, float v )
   {
      try
      {
         values[i] = (int) v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting float to int" );
      }
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setInt( int i, int v )
   {
      try
      {
         values[i] =  v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } 
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setShort( int i, short v )
   {
      try
      {
         values[i] =  v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
   }

   /**
    * Set the attribute indicated by the first argument to the value of the
    * second argument.
    */
   public void setByte( int i, byte v )
   {
      try
      {
         values[i] = v;
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
   }

   /** Return the field specified by the argument as a double. */
   public double getDouble( int i )
   {
      try
      {
         return values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
      return Double.NaN;
   }

   /** Return the field specified by the argument as a float. */
   public float getFloat( int i )
   {
      try
      {
         return values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
      catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting int to float" );
      }
      return Float.NaN;
   }

   /** Return the field specified by the argument as a int. */
   public int getInt( int i )
   {
      try
      {
         return values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      }
      return Integer.MAX_VALUE;
   }

   /** Return the field specified by the argument as a short. */
   public short getShort( int i )
   {
      try
      {
         return (short) values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } 
      catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting int to short" );
      }
      return Short.MAX_VALUE;
   }

   /** Return the field specified by the argument as a byte. */
   public byte getByte( int i )
   {
      try
      {
         return (byte) values[i];
      } catch ( ArrayIndexOutOfBoundsException ex )
      {
         new SDBError( SDBError.Recoverable, "Invalid attr index: " + i );
      } catch ( ArithmeticException aex )
      {
         new SDBError( SDBError.Recoverable, "Error casting int to byte" );
      }
      return Byte.MAX_VALUE;
   }

   /** set all Attribute values with one array. */
   public void setInts( int[] newVals )
   {
      int count = this.size();
      if ( newVals.length < count )
      {
         count = newVals.length;
         new SDBError( SDBError.Recoverable,
               "Datum.setInts: extra array elements ignored." );
      }
      for ( int i = 0; i < count; i++ )
         this.values[i] = newVals[i];
   }

   /** Return all attributes of the Datum as a 1-d int array */
   public int[] getInts()
   {
      // should we have a reference version of this?
      return values.clone();
   }

   /** Return all attributes of the Datum as a Values object  */
   public Values getValues()
   {  
      return new IntValues( values );
   }

   /**
    * Return a new datum with the attributes indicated by the RecordSpec. The
    * values of these attributes are taken from this IntDatum object.
    */
   public Datum subset( RecordSpec spec )
   {

      int i;
      int specsize = spec.size();

      IntDatum newdatum = new IntDatum( specsize );

      for ( i = 0; i < specsize; i++ )
      {
         newdatum.values[i] = this.values[spec.getFieldNumber( i )];
      }

      return newdatum;
   }

   /** Copy the attributes specified by the RecordSpec into the argument datum. */
   public void subset( IntDatum d, RecordSpec spec )
   {
      int i;
      int specsize = spec.size();

      try
      {
         for ( i = 0; i < specsize; i++ )
         {
            d.values[i] = this.values[spec.getFieldNumber( i )];
         }
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Argument datum was not large enough for RecordSpec." );
      }
   }

   /** Copy the attributes specified by the RecordSpec into the argument datum. */
   public void subset( Datum d, RecordSpec spec )
   {
      int i;
      int specsize = spec.size();

      try
      {
         for ( i = 0; i < specsize; i++ )
         {
            d.setInt( i, this.values[spec.getFieldNumber( i )] );
         }
      } catch ( ArrayIndexOutOfBoundsException e )
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "RecordSpec was too large." );
      }
   }

   /** Return a string representation of the datum. */
   public String toString()
   {
      int i, size = this.values.length;

      String result = new String( "{" );

      for ( i = 0; i < size - 1; i++ )
      {
         result += values[i] + ", ";
      }
      result += values[i] + "}";
      return result;
   }

   /** Return the number of attributes in the datum. */
   public int size()
   {
      return this.values.length;
   }

   /** Return the size in bytes needed to store the data for the attributes. */
   public int byteSize()
   {
      return this.values.length*4;
   }

   /**
    * Return the StorageType for this Datum
    */
   public int getStorageType()
   {
      return StorageType.Int;
   }
}