package edu.unh.sdb.common;

import java.util.*;
import java.io.*;

/**
 * The Metadata class can be used to hold metadata information as {key, value}
 * pairs, which are stored in a hashmap (an instance of the Properties class).
 * The Metadata class also contains a 'name' field which can be thought of as an
 * ID to help distinguish differen Metadata objects.
 * 
 * @author Kevin Rose
 * @date Aug 14, 2008
 */
public class Metadata implements Cloneable {

	private String name = null;

	/**
	 * hashmap to hold metadata key,value pairs
	 */
	private Properties data = new Properties();

	/**
	 * Default constructor. Constructs an empty Metadata object with a null
	 * String for a name.
	 */
	public Metadata() {
	}

	/**
	 * Creates a new Metadata object with the given name.
	 * 
	 * @param name the Metadata object's name.
	 */
	public Metadata(String name) {
		this.name = name;
	}

	/**
	 * creates this object as a copy of the argument.
	 * 
	 * @param m the Metadata to copy.
	 */
	public Metadata(Metadata m) {
		this.name = m.name;
		merge(m);
	}

	@Override
	public Metadata clone() {
		return new Metadata(this);
	}

	/**
	 * adds all the key/values in the argument to this Metadata object. If any
	 * of the keys are the same, use the argument's values.
	 * 
	 * TODO: should it overwrite similar keys? TODO: what if they have different
	 * names? (the Metadata objects)
	 * 
	 * @param m the Metadata to merge with.
	 */
	public void merge(Metadata m) {
		Iterator<String> itr = m.iterator();
		while (itr.hasNext()) {
			String key = itr.next();
			setProperty(key, m.getProperty(key));
		}
	}

	/**
	 * Sets the name of this Metadata object.
	 * 
	 * NOTE: if you are using this Metadata object as part of a
	 * MetadataCollection, changing the name here will not update the key that
	 * this Metadata object is stored under in the MetadataCollection.
	 * 
	 * @param name the name to set.
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * returns the name of this Metadata object.
	 * 
	 * @return returns the name of this Metadata object.
	 */
	public String getName() {
		return name;
	}

	/**
	 * returns a String representation of this Metadata object in XML form.
	 * 
	 * @return returns a String representation of this Metadata object in XML
	 *         form.
	 */
	public String toXML(String indent) {
		String str = "";
		String nm;

		if (name == null) {
			nm = "";
		} else {
			nm = " name= \"" + name + '\"';
		}

		str += indent + "<Metadata" + nm + ">\n";
		Iterator<String> itr = iterator();
		while (itr.hasNext()) {
			String key = itr.next();
			str += indent + "\t" + key + " = " + getProperty(key) + "\n";
		}
		str += indent + "</Metadata>";
		return str;
	}

	/*
	 * Accessors to Properties' methods.
	 */
	/**
	 * @see {@link Properties#load(Reader)}
	 */
	public void load(Reader reader) throws IOException {
		data.load(reader);
	}

	/**
	 * @see {@link Properties#load(InputStream)}
	 */
	public void load(InputStream inStream) throws IOException {
		data.load(inStream);
	}

	/**
	 * Creates a StringReader and uses the load(Reader) method.
	 * 
	 * @see {@link Properties#load(Reader)}
	 */
	public void load(String str) throws IOException {
		data.load(new StringReader(str));
	}

	/**
	 * Returns a Set of all the property names in this object.
	 * 
	 * @return Returns a Set of all the property names in this object.
	 */
	public Set<String> propertyNames() {
		return data.stringPropertyNames();
	}

	/**
	 * Returns an Iterator over all the properties names.
	 * 
	 * @return Returns an Iterator over all the properties names.
	 */
	public Iterator<String> iterator() {
		return data.stringPropertyNames().iterator();
	}

	/**
	 * Returns the property value associated with the key.
	 * 
	 * @param key the property name.
	 * @return Returns the property value associated with the key or null if the
	 *         key isn't present.
	 */
	public String getProperty(String key) {
		return data.getProperty(key);
	}

	/**
	 * Sets the property with given key to the given value.
	 * 
	 * @param key the property's name.
	 * @param value the property's value.
	 * @return the previous value of the specified key, or null if it did not
	 *         have one.
	 */
	public String setProperty(String key, String value) {
		return (String) data.setProperty(key, value);
	}

	/*
	 * Accessors to Hashmap's methods.
	 */
	/**
	 * Returns true if contains a property for the given key.
	 * 
	 * @param key the key to search for.
	 * @return Returns true if contains a property for the given key.
	 */
	public boolean containsKey(String key) {
		return data.containsKey(key);
	}

	/**
	 * Clears this Metadata object of all properties. The name is unchanged.
	 */
	public void clear() {
		data.clear();
	}

	/**
	 * Removes the property associated with the given key.
	 * 
	 * @param key the key to remove.
	 * @return the value to which the key had been mapped to, or null if the key
	 *         did not have a mapping.
	 */
	public String remove(String key) {
		return (String) data.remove(key);
	}

	@Override
	public String toString() {
		return name + ": " + data.toString();
	}
}
