package edu.unh.sdb.common;

import java.util.*;

/**
 * A MetaCollection is a HashMap of Metadata objects, where the name attribute
 * of each Metadata object is the key and the Metadata object itself is the
 * value.
 * 
 * NOTE: global metadata (metadata not given a name attribute when defined) can
 * be accessed by supplying a null string as the name using the 'getMetadata'
 * method.
 * 
 * @author Kevin Rose
 * @date Aug 14, 2008
 */
public class MetadataCollection implements Cloneable {

	/** Holds the Metadata objects using their name as the key */
	private HashMap<String, Metadata> hash = new HashMap<String, Metadata>();

	/**
	 * Constructs an empty MetadataCollection.
	 */
	public MetadataCollection() {
	}

	/**
	 * Constructs a MetadataCollection from an array of Metadata objects.
	 * 
	 * @param metadata the Metadata array.
	 */
	public MetadataCollection(Metadata[] metadata) {
		for (int i = 0; i < metadata.length; i++) {
			update(metadata[i]);
		}
	}

	@Override
	public MetadataCollection clone() {
		MetadataCollection copy = new MetadataCollection();
		Iterator<String> itr = getNames().iterator();
		while (itr.hasNext()) {
			copy.update(getMetadata(itr.next()).clone());
		}
		return copy;
	}

	/**
	 * Returns true if this collection contains a Metadata object with the given
	 * name.
	 * 
	 * @param name the name to search for.
	 * @return returns true if this collection contains a Metadata object with
	 *         the given name.
	 */
	public boolean contains(String name) {
		return hash.containsKey(name);
	}

	/**
	 * Returns the Metadata object with the given name or null if it does not
	 * exist.
	 * 
	 * @param name the Metadata name to search for
	 * @return returns the Metadata object with the given name or null if it
	 *         does not exist.
	 */
	public Metadata getMetadata(String name) {
		return hash.get(name);
	}

	/**
	 * Removes the Metadata object with the given name from the collection.
	 * 
	 * @param name the name of the Metadata object to remove.
	 * @return Removes the Metadata object with the given name from the
	 *         collection.
	 */
	public Metadata removeMetadata(String name) {
		return hash.remove(name);
	}

	/**
	 * Adds the specified Metadata object to this collection, using its name as
	 * the hashmap key. If there was already an entry under this key, it will be
	 * replaced.
	 * 
	 * @param m the Metadata to set
	 * @return the previous value associated with the key, or null if there was
	 *         no mapping for key.
	 */
	public Metadata setMetadata(Metadata m) {
		return hash.put(m.getName(), m);
	}

	/**
	 * If the collection does not already contain a Metadata object with the
	 * same name as argument, then the argument is added to the collection. If
	 * an entry with the same name already exists, then the existing Metadata
	 * object is merged with this one using the method 'merge' in the class
	 * Metadata.
	 * 
	 * @param m the Metadata object to update.
	 */
	public void update(Metadata m) {
		Metadata mine = hash.get(m.getName());
		if (mine != null) {
			mine.merge(m);
		} else {
			hash.put(m.getName(), m);
		}
	}

	/**
	 * Calls the 'update' method for each Metadata object in the given
	 * MetadataCollection. Update will add Metadata objects that are new, or
	 * merge with entries that already exist.
	 * 
	 * @param m the MetadataCollection to merge with.
	 */
	public void merge(MetadataCollection m) {
		Iterator<String> itr = m.getNames().iterator();
		while (itr.hasNext()) {
			this.update(m.getMetadata(itr.next()));
		}
	}

	/**
	 * Returns an iterator over the Metadata objects in this collection
	 * 
	 * @return Returns an iterator over the Metadata objects in this collection
	 */
	public Iterator<Metadata> iterator() {
		return hash.values().iterator();
	}

	/**
	 * Returns a Set containing all the keys (names) contained in this
	 * collection.
	 * 
	 * @return returns a Set containing all the keys (names) contained in this
	 *         collection.
	 */
	public Set<String> getNames() {
		return hash.keySet();
	}

	/**
	 * returns a String representing this metadata collection in XML format.
	 * 
	 * @return returns a String representing this metadata collection in XML
	 *         format.
	 */
	public String toXML(String indent) {
		String str = "";
		Iterator<String> itr = getNames().iterator();
		while (itr.hasNext()) {
			str += getMetadata(itr.next()).toXML(indent) + "\n\n";
		}
		return str;
	}

	@Override
	public String toString() {
		return getNames().toString();
	}
}
