/* 
    RecordDescriptor.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
*/
//---------------------- Edit history (as of 01/04/05) --------------------------
// 07/10/08 kevin: Fixed MixedValues size error in createValues( int )
// 06/05/08 rdb: Made ArrayList generic.
// 11/29/05 rdb: Added convenience functions: String[] getFieldNames() and
//                                            String[] getFieldTypes()
// 11/28/05 rdb: Deleted old deprecated code
// 10/21/05 rdb: Removed unused variables
// 07/16/05 rdb: Corrected createValues() which was not allocating space for all 
//               attributes.
// 01/26/05 rdb: Added getNumAttributes() to be consistent with other classes.
// 01/19/05 rdb: Created main() with some simple tests in it.
//               Modified create( FieldIDMapper ) to check for invalid mappings --
//                  anything that's not a "shuffle" and subset.
// 01/13/05 rdb: Revised constructors and project in order to share code.
// 01/12/05 rdb: Added createDataBlock methods
//               Re-implemented storage type counts code. 
//                 -Created an internal class to hold storage type values and counts.
//                 -Created public methods to access that information.
//                 -deprecated getStorageTypeCounts, replaced with methods:
//                   getNumUsedTypes();          // return # distinct types used
//                   getUsedType( index );      // return type of i-th distinct type
//                   getUsedTypeCount( index ); // return # attr of i-th distinct type
//                 
// 01/10/05 rdb: Added createDatum methods based on FieldIdMapper
// 01/06/05 rdb: Added convenience constructors: 
//               RecordDescriptor( String typecodes ); // for a mixed type 
//               RecordDescriptor( char typecode, int num ); // for single type
//               changed getIndexedAttribute( i ) to getAttribute( i )
// 01/04/05 rdb: Add createDatum to construct an "appropriate" Datum
//-------------------------------------------------------------------------------

package edu.unh.sdb.common;

import edu.unh.sdb.datasource.*;

import java.util.ArrayList;
import java.io.*;

/**
 * A RecordDescriptor is an ordered collection of  AttributeDescriptors. Used to represent 
 * the name and type of all the attributes of a point in the geometric space 
 * (i.e. a Datum).
 */

public class RecordDescriptor extends UserType implements Cloneable
{
   protected AttributeDescriptor[] defs;
   protected TypeTable             types;
   protected int[]                 byteOffsets;
   protected int                   byteSize;
   private int[][]               storageTypeCounts;
   private int[]                 storageTypes;
   private int                   maxByteSize;
   private int[]                 typesUsed;
   private int[]                 typesUsedCount;
	
	protected class TypeCount
	{
	   public int   type;
	   public int   count;
	   
	   TypeCount( int stype, int num )
	   {
	      this.type  = stype;
	      this.count = num;
	   }
	}
	protected ArrayList<TypeCount> typeCounters = new ArrayList<TypeCount>();

    /** Constructs an unnamed  RecordDescriptor from an array of AttributeDescriptors.*/
    public RecordDescriptor( AttributeDescriptor[] attrs )
    {
       this( "", attrs );
    }

    /** Constructs a named RecordDescriptor from an array of AttributeDescriptors.*/
    public RecordDescriptor( String name, AttributeDescriptor[] attrs )
    {
       super( name, StorageType.Record );
       
       this.defs = attrs.clone();
       this.types=null;

       for ( int i=0; i<attrs.length; i++)
       {
           if ( attrs[i].isUserType() )
               new IllegalArgumentError(SDBError.Recoverable,
                     "Attribute " + i + " is a user type and requires a TypeTable.");
       }
       this.makeOffsets();
       this.makeStorageTypeInfo();
    }
    

    /** Constructs an unnamed  RecordDescriptor from an array of AttributeDescriptors, 
     * referring to the given TypeTable.
     */
    public RecordDescriptor (TypeTable types, AttributeDescriptor[] attrs)
    {
       this( "", types, attrs );
    }

    /** Constructs a named RecordDescriptor from an array of AttributeDescriptors, 
     * referring to the given TypeTable.
     */
    public RecordDescriptor( String name, TypeTable types, AttributeDescriptor[] attrs)
    {
        super( name, StorageType.Record );
        
        this.defs = attrs.clone();
        this.types=types;
        this.makeOffsets();
        this.makeStorageTypeInfo();
    }
    
    /** Constructs an unnamed RecordDescriptor from a String in which each character 
     *  identifies a field and its type. The fields are unnamed. 
     *  The legal characters (for now) are:    
     *       d - double
     *       f - float
     *       i - int
     *       s - short
     *       b - byte
     * In the future, might add more. Could also extend to use printf type notation.
     *
     */
    public RecordDescriptor( String typeSpec )
    {
       this( createAttrDescriptorArray( typeSpec ));
    }

    /** Constructs a named RecordDescriptor from a String in which each character 
     *  identifies a field and its type. The fields are unnamed. 
     *  The legal characters (for now) are:    
     *       d - double
     *       f - float
     *       i - int
     *       s - short
     *       b - byte
     * In the future, might add more. Could also extend to use printf type notation.
     *
     */
    public RecordDescriptor( String name, String typeSpec )
    {
       this( name, createAttrDescriptorArray( typeSpec ));
    }

    /** Constructs an unnamed uni-type RecordDescriptor of a specified length
     *  The type is determined by the first argument, the number of fields by the 
     *  second. The fields are unnamed. 
     *  The legal characters (for now) are:    
     *       d - double
     *       f - float
     *       i - int
     *       s - short
     *       b - byte
     * In the future, might add more. Could also extend to use printf type notation.
     *
     */
    public RecordDescriptor( char typeSpec, int numAttr )
    {
       this( createAttrDescriptorArray( typeSpec, numAttr ));
    }

    /** Constructs a named uni-type RecordDescriptor of a specified length
     *  The type is determined by the first argument, the number of fields by the 
     *  second. The fields are unnamed. 
     *  The legal characters (for now) are:    
     *       d - double
     *       f - float
     *       i - int
     *       s - short
     *       b - byte
     * In the future, might add more. Could also extend to use printf type notation.
     *
     */
    public RecordDescriptor( String name, char typeSpec, int numAttr )
    {
       this( name, createAttrDescriptorArray( typeSpec, numAttr ));
    }

	/** Returns a new RecordDescriptor with fields copied from this RecordDescriptor. The
	FieldIDMapper specifies which fields are copied. The location of the fields in the new
	RecordDescriptor is determined by their order in the FieldIDMapper. 
	*/
    public RecordDescriptor project( FieldIDMapper fm )
    {
       if ( fm == null )
          return this;
       
       AttributeDescriptor newAD[] = new AttributeDescriptor[ fm.size() ];
       
       for ( int i = 0; i < fm.size(); i++ )
          newAD[ i ] = this.defs[ fm.toComposite( i ) ];
       return new RecordDescriptor( newAD );
    }

    /** Makes a new RecordDescriptor object from  "this" RecordDescriptor
    and the FieldIDMapper. The new RecordDescriptor will have the 
    same number of fields as the FIDMapper. The componentFields
    of FIDMapper specifys which attributes are taken from "this" 
    RecordDescriptor. The compositeFields of FIDMapper also 
    determines their axisOrdering in the new RecordDescriptor.
    
    This method as currently implemented does not appear to be stable. It 
    makes an array that is the size of the FieldIDMapper, but indexes into
    that array using composite field values that, in general, can be larger
    than the size of the FieldIDMapper.
    
    I am going to assume that the intent is that the FieldIDMapper that is
    passed in is only a "shuffling" FieldIDMapper. That is, there should
    be no indexes that are larger than the size of the component and composite
    arrays. 
    
    The "project" method works if the result is a simple selection (projection).
    
	*/
	public RecordDescriptor create( FieldIDMapper FIDMapper )
	{
	    if ( FIDMapper == null ) 
	       return this;
	    int [] cpnFids = FIDMapper.getComponentFields();
	    int [] cpsFids = FIDMapper.getCompositeFields();
	    AttributeDescriptor [] attri = new AttributeDescriptor[ cpnFids.length ];
	
	    try {
		    for ( int i=0; i < cpnFids.length; i++ )
		        attri[ cpsFids[i] ] = this.defs[ cpnFids[i] ];
	    } catch ( ArrayIndexOutOfBoundsException ex )
		    {
		       new SDBError( SDBError.Recoverable, 
		             "Invalid FieldIDMapper for RecordDescriptor.create." 
		             + " Component field size: " + cpnFids.length
		             + " Composite field size; " + cpsFids.length );
		       return this.project( FIDMapper ); // use component order in result
		    }
	    return new RecordDescriptor(attri);
	}

	/** Makes a new RecordDescriptor object from  "this" RecordDescriptor
	 *  and the RecordSpec. The new RecordDescriptor will have the 
	 *  same number of fields as the RecordSpec. 
	 */
	public RecordDescriptor create( RecordSpec rs )
	{
	   return project( rs );
	}

    /** Returns a new RecordDescriptor with fields copied from this RecordDescriptor. The
	RecordSpec specifies which fields are copied. The location of the fields in the new
	RecordDescriptor is determined by their order in the RecordSpec. 
	*/
	public RecordDescriptor project( RecordSpec rs )
	{
		if ( rs == null )
		   return this;
		
		int [] fields = rs.getFieldNumbers();
		
		AttributeDescriptor [] defs = new AttributeDescriptor[ fields.length ];
		
		for ( int i=0; i < fields.length; i++)
			defs[i] = this.defs[ fields[i] ].copy();
		return new RecordDescriptor( defs );
	}

    /** Create an array of AttributeDescriptors whose types are identified
     *  by the individual characters in the argument String. Valid characters
     *  are d,f,i,s,b for double, float, int, short, byte. We may add some 
     *  additional functionality.
     * 
     * The Attribute descriptors will be unnamed.
     */
    protected static AttributeDescriptor[] createAttrDescriptorArray( String typeSpec )
    {  
       String validChars = "dfisb";
       int validCharCount = 0;
       for ( int i = 0; i < typeSpec.length(); i++ )
          if ( validChars.indexOf( typeSpec.charAt(i)) >= 0 )
             validCharCount++;
       if ( validCharCount < typeSpec.length() )
          new SDBError( SDBError.Recoverable, "Invalid type specification in string: "
                           + typeSpec + ".  Ignoring bad characters." );
          
       AttributeDescriptor ad[] = new AttributeDescriptor[ validCharCount ];
       int valid = 0;
       for ( int i = 0; i < typeSpec.length(); i++ )
       {
          switch ( typeSpec.charAt( i ))
          {
          case 'd': ad[ valid++ ] = new AttributeDescriptor( StorageType.Double, "" );
                    break;
          case 'f': ad[ valid++ ] = new AttributeDescriptor( StorageType.Float, "" );
                    break;
          case 'i': ad[ valid++ ] = new AttributeDescriptor( StorageType.Int, "" );
                    break;
          case 's': ad[ valid++ ] = new AttributeDescriptor( StorageType.Short, "" );
                    break;
          case 'b': ad[ valid++ ] = new AttributeDescriptor( StorageType.Byte, "" );
                    break;
          default:
                    break;
          }
       }
       return ad;
    }
    
    /** Create an array of AttributeDescriptors whose types are identified
     *  by a single char argument indicating type and an attr count. Valid characters
     *  are d,f,i,s,b for double, float, int, short, byte. We may add some 
     *  additional functionality.
     * 
     * The Attribute descriptors will be unnamed.
     */
    protected static AttributeDescriptor[] createAttrDescriptorArray( 
                                            char typeSpec, 
                                            int numAttr )
    {  
       StringBuffer sb = new StringBuffer( numAttr );
       for ( int i = 0; i < numAttr; i++ )
          sb.append( typeSpec );
       return createAttrDescriptorArray( sb.toString() );
    }
    
    /** Constructs an empty RecordDescriptor. */
    protected RecordDescriptor (){

        super(new String(""),StorageType.Record);
        
        this.defs = null;
        this.types = null;
    }

    /** Allocate and fill in the array of byte offsets.
        Also gives a value to this.byteSize.*/
    protected void makeOffsets(){
    
       int i,size=this.defs.length;
       int offset=0;
       int byteSize;
       this.byteOffsets=new int[size];

       this.maxByteSize=0;

       for(i=0;i<size;i++){
          this.byteOffsets[i]=offset;

          byteSize=this.defs[i].byteSize();

          this.maxByteSize= (byteSize > this.maxByteSize) ? byteSize : this.maxByteSize;

          offset+=byteSize;
       }

       this.byteSize=offset;
    }    
    
    /** Make a copy of this RecordDescriptor. If a TypeTable is present, the
        copy will contain a reference to the same TypeTable object.*/
    public RecordDescriptor copy(){
 
        return (RecordDescriptor) this.clone();
    }

    
    /** Make a copy of this RecordDescriptor. If a TypeTable is present, the
        copy will contain a reference to the same TypeTable object.*/    
    public Object clone()
    {
        RecordDescriptor n=new RecordDescriptor();
        
        if (this.defs!=null){
            n.defs = new AttributeDescriptor[this.defs.length];
            System.arraycopy (this.defs, 0, n.defs, 0, this.defs.length);
        }
        if (this.byteOffsets!=null){
            n.byteOffsets = new int[this.byteOffsets.length];
            System.arraycopy (this.byteOffsets, 0, n.byteOffsets, 0, this.byteOffsets.length);
        }
        if(this.storageTypeCounts!=null){
            n.storageTypeCounts = this.storageTypeCounts.clone();
        }
        if(this.storageTypes!=null){            
            n.storageTypes = new int[this.storageTypes.length];
            System.arraycopy (this.storageTypes, 0, n.storageTypes, 0, this.storageTypes.length);
        }
        n.byteSize=this.byteSize;
        n.types=this.types;
        n.maxByteSize = this.maxByteSize; 
   
        n.typesUsed = new int[this.typesUsed.length];
        System.arraycopy (this.typesUsed, 0, n.typesUsed, 0, this.typesUsed.length);
        return n;
    }

    

    /** True if the argument is equal to this AttributeDescriptor, false otherwise.
        Currently, two AttributeDescriptors are equal if their name and type fields
        are equal, and they are using the same TypeTable object.
        */
    public boolean equals( Object o )
    {
        RecordDescriptor r=(RecordDescriptor) o;

        int size = this.defs.length;

        if ( r.size()!=size )
            return false;
        
        for ( int i = 0; i<size; i++ )
        {
            if ( !( this.defs[i].equals( r.defs[i])) )
                return false;
        }

        return true;
    }
    
    /** Return the storage type of an attribute at the specified position.*/
    public final int getStorageType(int index)
    {
       try {
          return this.storageTypes[index];
	   } catch ( ArrayIndexOutOfBoundsException e ) 
		  {
			new SDBError(SDBError.Recoverable, "Invalid attribute index: " + index );
			return 0;
		  }
    }
    
    /** Return an array of storage types; there is one entry per attribute. 
     */
    public final int[] getStorageTypes()
    {
       return storageTypes.clone();
    }
    
    /** Return the underlying storage type for the entire record.
     *  If all attributes are of the same type, the record will be of that type.
     * Normally, mixed types in the record result in StorageType.Record as the
     * storage type. However, it may some day be reasonable to have an 
     * underlying common type even though the fields are mixed. For example,
     * a "smart" RecordDescriptor might choose to promote 1 byte field to short
     * if all the other fields are short since that would provide a more 
     * efficient implementation.
     *    
     * The value returned by this method should match the StorageType associated
     * with the Values object that will be returned by DataCollection.getValues().
     */
    public final int getStorageType()
    {
	   if ( this.typesUsed.length == 0 )
	   {
	      new SDBError( SDBError.Warning, "No fields. Can't determine a type." );
	      return StorageType.Float;
	   }
	   
	   if ( this.typesUsed.length > 1 )
	      return StorageType.Record;
	   else
	      return this.typesUsed[0];
    }  

    /** Return true if the attribute at the specified position is a User
     *  defined type.  Return false otherwise.
     */
    public final boolean isUserType( int index)
    {
        return this.defs[index].isUserType();
    }
 
    /**
     *  Return an array of the field names.
     */
    public String[] getFieldNames()
    {
       String[] names = new String[ this.getNumAttributes() ];
       for ( int i = 0; i < names.length; i++ )
          names[ i ] = this.defs[ i ].name();
       return names;
    }

    /**
     *  Return an array of the field types in string form.
     */
    public String[] getFieldTypes()
    {
       String[] names = new String[ this.getNumAttributes() ];
       for ( int i = 0; i < names.length; i++ )
          names[ i ] = this.defs[ i ].getTypeName();
       return names;
    }

    /** Return the name of a attribute at the specified position.*/
    public String name( int index )
    {
        return this.defs[index].name();
    }

    /** Return the AttributeDescriptor at the specified position.*/
    public final AttributeDescriptor getAttribute( int i )
    {
        return defs[i];
    }

    /** Return the number of Attributes in this collection. */
    public final int size()
    {
        return this.defs.length;
    }

    /** Return the number of Attributes in this collection. */
    public final int getNumAttributes()
    {
        return this.defs.length;
    }

    /** Return the number of bytes occupied by records decribed by this descriptor.        
     *  We are not accounting for padding at the moment.
     */
    public final int byteSize()
    {
        return this.byteSize;
    }


    /** Return the offset of the indicated attribute from the beginning of the record. 
     *  We are not accounting for padding at the moment.
     */
    public final int byteOffset(int attributeNumber)
    {
        return this.byteOffsets[attributeNumber];
    }

    /** Find an entry in list of TypeCount records corresponding to given type 
     */
    private TypeCount findTypeCount( int storageType )
    {
       for ( int i=0; i < this.typeCounters.size(); i++ )
       {
          TypeCount entry = this.typeCounters.get( i );
          if ( entry.type == storageType )
             return entry;
       }
       return null;
    }

    /** Initialize arrays used for quick access to type information.
        <b>storageTypeCounts<\b> is an array of pairs of integers such that the 
        first element of each pair indicates how many attributes with a certain 
        storage type occur in this RecordDescriptor, and the second element indicates 
        the storage type for that count. 
        
        <b>storageTypes<\b> is an array of integers representing the types of each 
        field of the RecordDescriptor.
        */
    protected void makeStorageTypeInfo()
    {
       // rdb: 01/12/05
       // The existing code is poor
       // 1. It is confusing. It uses a 2d array to hold what should be a 1d array of records.
       // 2. It relies heavily on knowledge of the internal implementation of
       //    the StorageType class. This is not good oo methodology.
       // The new code is not as efficient, but it is much simpler and 
       // this is hardly critical path code.
       // To minimize impact on user code, I've left some of the interface the same.
       //   In particular, I left the definition of getStorageType( int index ) as it was.
       
       this.storageTypes = new int[defs.length];
       
       for ( int i = 0; i < defs.length; i++ )
       {
          this.storageTypes[ i ] = defs[i].getStorageType();
          TypeCount entry        = findTypeCount( this.storageTypes[ i ] );
          
          if ( entry == null )
             this.typeCounters.add( new TypeCount( defs[i].getStorageType(), 1));
          else
             entry.count++; // increment counter
       }
       
       this.typesUsed = new int[ typeCounters.size() ];
       this.typesUsedCount = new int[ typeCounters.size() ];
       for ( int i = 0; i < typeCounters.size(); i++ )
       {
          TypeCount entry   = typeCounters.get( i );
          typesUsed[i]      = entry.type;
          typesUsedCount[i] = entry.count;
       }
    
       /********** code replaced 01/12/05 rdb ********************
        int i,j;
        int numStorageTypes=StorageType.getNumStorageTypes();
        int numDistinctTypes=0;
        int [] cnts=new int[numStorageTypes];
        
        this.storageTypes=new int[defs.length];
        
        for(i=0;i<defs.length;i++){
            this.storageTypes[i]=defs[i].getStorageType();
            cnts[this.storageTypes[i]-1]++;
        }
        
        for(i=0;i<numStorageTypes;i++){
            if(cnts[i]!=0){
                numDistinctTypes++;
            }
        }
        
        this.storageTypeCounts=new int[numDistinctTypes][2];
        j=0;
        for(i=0;i<numStorageTypes;i++){
            if(cnts[i]!=0){
                this.storageTypeCounts[j][0]=cnts[i];
                this.storageTypeCounts[j][1]=i+1;
                j++;
            }
        }
        /*********************************************************************/
    }
        
    /** 
     *  Return the number of distinct storage types defined by this RecordDescriptor.
     */
    public int getNumUsedTypes()
    {
       return this.typesUsed.length;
    }
    
    /**
     * Get the an array of all StorageTypes used in this RecordDescriptor.
     * 
     */
    public int[] getUsedTypes( )
    {
       return this.typesUsed.clone();
    }
    
    /**
     * Get the StorageType of the i-th distinct storage type in this RecordDescriptor.
     * 
     */
    public int getUsedType( int index )
    {
       try {
          return this.typesUsed[ index ];
	   } catch ( ArrayIndexOutOfBoundsException e ) 
		  {
			new SDBError(SDBError.Recoverable, "Invalid attribute index: " + index );
			return 0;
		  }
    }

    /**
     * Get the count of attributes of the i-th distinct storage type in this 
     * RecordDescriptor.
     */
    public int getUsedTypeCount( int index )
    {
       try {
          return this.typesUsedCount[ index ];
	   } catch ( ArrayIndexOutOfBoundsException e ) 
		  {
			new SDBError(SDBError.Recoverable, "Invalid attribute index: " + index );
			return 0;
		  }
    }
	
	/** Returns the size in bytes of the largest attribute.
	 */
    public final int getMaxAttributeByteSize(){
	
		return this.maxByteSize;
	}

    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){

        int i, size=this.defs.length;

        StringBuffer result=new StringBuffer(super.toString());
        
        result.append(" {");
        
        for(i=0;i<size-1;i++){
            result.append(this.defs[i]);
            result.append(", ");
        }
        
        if (size>0) {
            result.append(this.defs[i]);
        }
        
        result.append("} >>");
        
        return new String(result);
    }
    
    /** Returns an array of integers indicating the size in
     *  bytes of the correspoding attributes in this RecordDescriptor.
     *
     *	added by Xuan Tang 08/05/03 
     */
    public int[] attributeByteSizes()
    {
       int size = this.defs.length;
       int[]  byteSizes = new int[size];
                       
       for( int i=0 ; i<size; i++)
       {
           byteSizes[i] = this.defs[i].byteSize();   
       }
       return byteSizes;
    } 

    /** Returns an array of integers indicated the byte offsets
     *  of each of the correspoding attributes in this RecordDescriptor.
     *
     *  added by Xuan Tang 08/06/03 */
	public int[] byteOffsets(){ 
      
		return this.byteOffsets;
	}

    //+++++++++++++++++++++++ createDatum methods +++++++++++++++++++++++++++++++
     /** Returns a Datum of an appropriate type. 
      *  If all fields are of the same type, it will create a Datum of
      *  that type; otherwise, it creates a MixedDatum object.
      */
	public Datum createDatum()
	{ 
	   Datum  retDatum = null;
	   switch ( this.getStorageType() )
       {
         case StorageType.Byte:
         case StorageType.UnsignedByte:
         case StorageType.Char:
         case StorageType.Boolean:
            retDatum = new ByteDatum( this.defs.length );
            break;
         case StorageType.Double:
            retDatum = new DoubleDatum( this.defs.length );
            break;
         case StorageType.Float:
            retDatum = new FloatDatum( this.defs.length );
            break;
         case StorageType.Int:
            retDatum = new IntDatum( this.defs.length );
            break;
         case StorageType.Short:
         case StorageType.UnsignedShort:
            retDatum = new ShortDatum( this.defs.length );
            break;
         case StorageType.Record:
            new SDBError( SDBError.Warning, "Beware: Record type not fully supported." );
            retDatum = new MixedDatum( this );
            break;
         default:
            new SDBError( SDBError.Warning, "Type not supported:"
                  + this.typesUsed[0] + " using, Record that is poorly supported." );
            retDatum = new MixedDatum( this );
            break;
	   }
	   return retDatum;
	}
                                                                                                                    
    /** Returns a Datum of an appropriate type as constrained by RecordSpec.
     *  If the RecordSpec argument is null, all fields of the RecordDescriptor 
     *  are used.
     *  If all fields are of the same type, it will create a Datum of
     *  that type; otherwise, it creates a MixedDatum object.
     */
	public Datum createDatum( RecordSpec rs )
	{ 
	   return ( this.project( rs ).createDatum() );
	}
	
    /** Returns a Datum of an appropriate type. Only includes fields defined
     *  by the toComponent results of the FieldIDMapper. If the FieldIDMapper 
	 *  argument is null, all fields of the RecordDescriptor are used.
     *  If all fields are of the same type, it will create a Datum of
     *  that type; otherwise, it creates a MixedDatum object.
     */
	public Datum createDatum( FieldIDMapper fm )
	{ 
	   return ( this.project( fm ).createDatum() );
	}
	
    /** Returns a Datum of an appropriate type, initialized to the values
     *  of the passed in argument Datum. 
     *  If all fields are of the same type, it will create a Datum of
     *  that type; otherwise, it creates a MixedDatum object.
     */
	public Datum createDatum( Datum d )
	{ 
	   Datum retDatum = this.createDatum(); // first get the "right" kind of datum
	   d.copyTo( retDatum );
	   return retDatum;
	}
    
    /** Returns a Datum of an appropriate type, initialized to the values
     *  of the passed in argument Datum. The RecordSpec is intended to
     *  identify the subset of the result Datum. It is NOT applied to the
     *  Datum passed as the argument; this is assumed to match the RecordSpec.
     *  If the RecordSpec argument is null, all fields of the RecordDescriptor 
     *  are used.
     * 
     *  If all fields are of the same type, it will create a Datum of
     *  that type; otherwise, it creates a MixedDatum object.
     */
	public Datum createDatum( Datum d, RecordSpec spec )
	{ 
	   Datum retDatum = this.createDatum( spec ); // first get the "right" kind of datum
	   d.copyTo( retDatum );
	   return retDatum;
	}
    
    /** Returns a Datum of an appropriate type, initialized to the values
     *  of the passed in argument Datum. If the FieldIDMapper argument is
	 * null, all fields of the RecordDescriptor are used.
     * Only includes fields defined by the toComponent results of the FieldIDMapper.
     *  and assumes that the passed in Datum only has those fields!!!!
     *     
     *  If all fields are of the same type, it will create a Datum of
     *  that type; otherwise, it creates a MixedDatum object.
     */
	public Datum createDatum( Datum d, FieldIDMapper fim )
	{ 
	   Datum retDatum = this.createDatum( fim ); // first get the "right" kind of datum
	   d.copyTo( retDatum );
	   return retDatum;
	}
    
    //+++++++++++++++++++++++ createValues method +++++++++++++++++++++++++++++++
    /**
     * Create a Values object of the appropriate type and size for this RecordDescriptor
     * numEntries is the number of data points, so if there are multiple attributes
     * there needs to be space for all of them
     */
	public Values createValues( int numEntries )
	{
	   Values retValues = null;
	   switch ( this.getStorageType() )
       {
         case StorageType.Byte:
         case StorageType.UnsignedByte:
         case StorageType.Char:
         case StorageType.Boolean:
            retValues = new ByteValues( numEntries * getNumAttributes() );
            break;
         case StorageType.Double:
            retValues = new DoubleValues( numEntries * getNumAttributes() );
            break;
         case StorageType.Float:
            retValues = new FloatValues( numEntries * getNumAttributes() );
            break;
         case StorageType.Int:
            retValues = new IntValues( numEntries * getNumAttributes() );
            break;
         case StorageType.Short:
         case StorageType.UnsignedShort:
            retValues = new ShortValues( numEntries * getNumAttributes() );
            break;
         case StorageType.Record:
            new SDBError( SDBError.Warning, "Beware Record type not fully supported." );
            
            //retValues = new MixedValues( this, numEntries * getNumAttributes() );
            retValues = new MixedValues( this, this.byteSize );
            
            break;
         default:
            new SDBError( SDBError.Warning, " Type not supported:"
                  + this.typesUsed[0] + " using Record that is poorly supported.");
            retValues = new MixedValues( this, numEntries * getNumAttributes() );
            break;
	   }
	   return retValues;
	}
	
    //+++++++++++++++++++++++ createDataBlock methods +++++++++++++++++++++++++++++++
	
	/**
	 * Creates a DataBlock that would be efficient to use for this RecordDescriptor.
	 * If the attributes are all of the same data type, creates a BasicBlock.
	 * If the attributes are of mixed types, it creates a CompositeBlock and assigns
	 * all of the attributes of the same type to one of the BasicBlocks inside the
	 * CompositeBlock.
	 */
	public DataBlock createDataBlock( ISBounds bnds )
	{
	   if ( this.typesUsed.length == 0 ) 
	      new SDBError( SDBError.Recoverable, "No fields. Can't create a Datum." );
	   
	   if ( this.typesUsed.length == 1 ) // A simple BasicBlock will do
	      return new BasicBlock( bnds, this );
	   else // need to build a CompositeBlock
	      return new CompositeBlock( bnds, this );
	}

	/**
	 * Creates a DataBlock that would be efficient to use for a subset of this
	 * RecordDescriptor, based on the RecordSpec. If the RecordSpec argument is
	 * null, all fields of the RecordDescriptor are used.
	 * If the attributes are all of the same data type, creates a BasicBlock.
	 * If the attributes are of mixed types, it creates a CompositeBlock and assigns
	 * all of the attributes of the same type to one of the BasicBlocks inside the
	 * CompositeBlock.
	 */
	public DataBlock createDataBlock( ISBounds bnds, RecordSpec rs )
	{
	   return this.project(rs).createDataBlock( bnds );
	}
    
	/**
	 * Creates a DataBlock that would be efficient to use for a subset of this
	 * RecordDescriptor, based on the FieldIDMapper. If the FieldIDMapper argument is
	 * null, all fields of the RecordDescriptor are used.
	 * If the attributes are all of the same data type, creates a BasicBlock.
	 * If the attributes are of mixed types, it creates a CompositeBlock and assigns
	 * all of the attributes of the same type to one of the BasicBlocks inside the
	 * CompositeBlock.
	 */
	public DataBlock createDataBlock( ISBounds bnds, FieldIDMapper fm )
	{
	   return this.project(fm).createDataBlock( bnds );
	}
    
	//----------------------- main --------------------------------
    public static void main( String args[] )
    {
       // testCodes defines what tests are to be executed:
       //  c - RecordDescriptor constructors and create and project methods
       //  d - createDatum methods
       //  b - createBlock methods
       //  t - user types
       //  a - accessors
       
       String testCodes = "cdbta";
       if ( args.length == 0 )
          testCodes = "cudba";
       else if ( args[0].equals( "-h" ) || args[0].equals( "-?" ))
          usage( System.out );
       else 
          testCodes = args[0];
             
       if ( testCodes.indexOf( 'c' ) >= 0 )
          testConstruction();
       if ( testCodes.indexOf( 'b' ) >= 0 )
          testCreateDataBlock();
       if ( testCodes.indexOf( 'd' ) >= 0 )
          testCreateDatum();
       if ( testCodes.indexOf( 't' ) >= 0 )
          testUserTypes(); // rdb moved the old main code to a method
       if ( testCodes.indexOf( 'a' ) >= 0 )
          testAccessors();
    }
    
    private static void usage( PrintStream out )
    {
       out.println( "Usage: java RecordDescriptor [-h] [-?] [testcodes]" );
       out.println( "    where   -h or -? prints this output." );
       out.println( "            testcodes is a string composed of single chars" );
       out.println( "               that identify tests to be run:" );
       out.println( "               c - constructors, create, and project methods." );
       out.println( "               d - createDatum methods." );
       out.println( "               b - createDataBlock methods." );
       out.println( "               t - test UserType code." );
       out.println( "               a - test accessor (set/get) methods.");
       out.println( " No 'testcodes' argument means: do all tests." );
    }

    //------------------------ testConstruction ----------------------------------
    private static void testConstruction()
    {
       System.out.println( "++++++++++++++++++++Constructors++++++++++++++++++++++");
       AttributeDescriptor adInts[] = new AttributeDescriptor[]
			     {
			           new AttributeDescriptor( StorageType.Int, "ientry0" ),
			           new AttributeDescriptor( StorageType.Int, "ientry1" ),
			           new AttributeDescriptor( StorageType.Int, "ientry2" )
			     };
       // Construction methods:
       //      unnamed
       // RecordDescriptor( AttributeDescriptor[] )
       RecordDescriptor rd1 = new RecordDescriptor( adInts );
       System.out.println( "( AD[] ):" + rd1 );
       
       // RecordDescriptor( String )
       RecordDescriptor rd2 = new RecordDescriptor( "ffff" );
       System.out.println( "( \"ffff\" ):" + rd2 );
       
       // RecordDescriptor( char, int )   - homogeneous
       RecordDescriptor rd3 = new RecordDescriptor( 'd', 4 );
       System.out.println( "( 'd', 4 ):" + rd3 );
       
       // RecordDescriptor( TypeTable, AttributeDescriptor[] )
       
       //       named
       // RecordDescriptor( String, AttributeDescriptor[] )
       RecordDescriptor rd4 = new RecordDescriptor( "rd4", adInts );
       System.out.println( "( \"rd4\", AD[] ):" + rd4 );
       
       // RecordDescriptor( String, String )
       RecordDescriptor rd5 = new RecordDescriptor( "rd5", "difs" );
       System.out.println( "( \"rd5\", \"difs\" ):" + rd5 );
       
       // RecordDescriptor( String, char, int )   - homogeneous
       RecordDescriptor rd6 = new RecordDescriptor( "rd6", 'd', 4 );
       System.out.println( "( \"rd6\", 'd', 4 ):" + rd6 );
       
       // RecordDescriptor( String, TypeTable, AttributeDescriptor[] )
       // 
       // XX deprecated XX create( FieldIDMapper )
       // XX deprecated XX create( RecordSpec 
       //
       // project( RecordSpec )
       RecordDescriptor rd7 = rd5.project( new RecordSpec( new int[] {3,1,2} ));
       System.out.println( "rd5.project( rs[3,1,2] ):" + rd7 );
       
       // project( FieldIDMapper )
       FieldIDMapper fidm = new FieldIDMapper( new int[] {0,1,2}, new int[] {3,1,2} );
       RecordDescriptor rd8 = rd5.project( fidm );
       System.out.println( "project( fidm ):" + rd8 );
       
       System.out.println( "+++++++++++++++++++++++++++++++++++++++++++++++++++++++");
       System.out.println( "");
       System.out.println( "");
    }
    //------------------------ testCreateDataBlock ----------------------------------
    private static void testCreateDataBlock()
    {
       
    }
    //------------------------ testAccessors ----------------------------------
    private static void testAccessors()
    {
       
    }
     //------------------------ testCreateDatum----------------------------------
    private static void testCreateDatum()
    {
       // test all int record
       AttributeDescriptor adInts[] = new AttributeDescriptor[]
       {
             new AttributeDescriptor( StorageType.Int, "ientry0" ),
             new AttributeDescriptor( StorageType.Int, "ientry1" ),
             new AttributeDescriptor( StorageType.Int, "ientry2" )
       };
       RecordDescriptor rdInts   = new RecordDescriptor( adInts );
       Datum di = rdInts.createDatum();
       if ( di instanceof IntDatum )
          System.out.println( "---Got IntDatum-----" );
       else
          System.out.println( "*** Error: should be IntDatum. Got: " + di.getClass()  );
                 
       AttributeDescriptor adShorts[] = new AttributeDescriptor[]
       {
          new AttributeDescriptor( StorageType.Short, "sentry0" ),
          new AttributeDescriptor( StorageType.Short, "sentry1" ),
          new AttributeDescriptor( StorageType.Short, "fentry2" )
       };
       RecordDescriptor rdShorts = new RecordDescriptor( adShorts );
       Datum ds = rdShorts.createDatum();
       if ( ds instanceof ShortDatum )
          System.out.println( "---Got ShortDatum-----" );
       else
          System.out.println( "*** Error: should be ShortDatum. Got: " + ds.getClass() );
       
       AttributeDescriptor adFloats[] = new AttributeDescriptor[]
       {
          new AttributeDescriptor( StorageType.Float, "fentry0" ),
          new AttributeDescriptor( StorageType.Float, "fentry1" ),
          new AttributeDescriptor( StorageType.Float, "fentry2" )
       };
       RecordDescriptor rdFloats = new RecordDescriptor( adFloats );
       Datum df = rdFloats.createDatum();
       if ( df instanceof FloatDatum )
          System.out.println( "---Got FloatDatum-----" );
       else
          System.out.println( "*** Error: should be FloatDatumGot: " + df.getClass()  );

       AttributeDescriptor adMixed[] = new AttributeDescriptor[]
       {
          new AttributeDescriptor( StorageType.Short,  "sentry0" ),
          new AttributeDescriptor( StorageType.Int,    "ientry1" ),
          new AttributeDescriptor( StorageType.Double, "dentry2" )
       };
       RecordDescriptor rdMixed  = new RecordDescriptor( adMixed );
       Datum dm = rdMixed.createDatum();
       if ( dm instanceof MixedDatum )
          System.out.println( "---Got MixedDatum-----" );
       else
          System.out.println( "*** Error: should be MixedDatum. Got: " + dm.getClass() );
       dm.setShort( 0, (short)0 );
       dm.setInt( 1, 1 );
       dm.setDouble( 2, 2.0 );
       System.out.println( "MixedDatum.toString(): " + dm.toString() );
       System.out.println( "MixedDatum: < " + dm.getShort( 0 ) + ",  " 
                                          + dm.getInt( 1 )   + ",  "
                                          + dm.getDouble( 2 ) + " > " );
       
       System.out.println( "+++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    }
    private static void testUserTypes() 
    {
        TypeTable ttable=new TypeTable(5);
        int pNum,qNum;
        int i;

        RangedType probability=new RangedType("Probability",StorageType.Double,0,1);
        RangedType quality=new RangedType("Quality",StorageType.Int,1,10);

        qNum=ttable.insert(quality);
        pNum=ttable.insert(probability);
        System.out.println( "testUserTypes: qnum, pnum: " + qNum + " " + pNum );

        AttributeDescriptor [] adArray = new AttributeDescriptor [] {
                           new AttributeDescriptor(ttable,"Probability","chances"),
                           new AttributeDescriptor(ttable,"Quality","rating")
                     };
        RecordDescriptor rd=new RecordDescriptor(ttable,adArray);

        System.out.println(rd);
        
//        int [][] cnts=rd.getStorageTypes();
        int[] used = rd.getUsedTypes();
        
        for( i=0; i<used.length; i++ )
        {
            System.out.println( "number=" + rd.getUsedTypeCount( used[i] )
                              +" type=" + used[i]);
        }
        
        System.out.println("\n");
        
        adArray = new AttributeDescriptor [] 
                               {
                                    new AttributeDescriptor(ttable,"Probability","chances"),
                                    new AttributeDescriptor(ttable,"Quality","rating"),
                                    new AttributeDescriptor(StorageType.Int,"length"),
                                    new AttributeDescriptor(StorageType.Double,"Width"),
                                    new AttributeDescriptor(StorageType.Int,"Height"),
                                    new AttributeDescriptor(StorageType.Byte,"Density"),
                                };
        
        rd=new RecordDescriptor(ttable,adArray);
        
        used = rd.getUsedTypes();
        
        for ( i=0; i<used.length; i++ )
        {
           System.out.println( "number=" + rd.getUsedTypeCount( used[i] )
                                         +" type=" + used[i]);
        }

    }
}
