/* 
    RecordSpec.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    ------------------------------ Edit history as of 1/27/05 ---------------------
    12/07/05 rdb: Added try-catch to getFieldNumber( int )
    03/04/05 rdb: Restored the old implementation (it is faster) while keeping the
                    reference to an equivalent FieldIDMapper. Thus if user explicitly
                    accesses the RecordSpec functions, they will be faster. This
                    also avoids the problem of FieldIDMapper sorting of the components
                    array.
    02/20/05 rdb: Reimplemented with a FieldIDMapper as data member. This makes it
                    much more efficient for the many methods that have to handle either
                    since they don't need to construct a new FieldIDMapper every time.
    02/01/05 rdb: Changed getIndexedFieldNumber to getFieldNumber
    01/27/05 rdb: Added RecordSpec( int ) 
                     which creates an identity mapping of the specified length
*/

package edu.unh.sdb.common;

/**
 * A RecordSpec holds an array of integers, where each integer specifies the
 * field number of a desired attribute. Typically, a RecordSpec will be an
 * argument to a query method, specifying which point attributes should be 
 * returned.
 *      
 */


public class RecordSpec
{
    protected int[] fieldNumbers;
    protected FieldIDMapper fim;
    
    /** Constructor that takes an array of ints */
    public RecordSpec( int specs[] )
    {
       this.fieldNumbers = specs.clone();
       this.fim = new FieldIDMapper( specs );
    }
    
    /** Constructor that creates an identity mapping of the specified length.
     *  This is convenient for code sharing when there is no specified RecordSpec
     */
    public RecordSpec( int len )
    {
       this.fieldNumbers = new int[ len ];
       for ( int i = 0; i < len; i++ )
          this.fieldNumbers[ i ] = i;
       this.fim = new FieldIDMapper( len );
    }

    /** Returns the FieldSpec at the index position.
     * @deprecated 
     * Use getFieldNumber( int )
     */
    final public int getIndexedFieldNumber(int index)
    {
        return getFieldNumber( index );
        //rdb return this.fieldNumbers[index];
    }
    
    /** Returns the FieldIDMapper equivalent to this RecordSpec 
     */
    final public FieldIDMapper getFieldIDMapper()
    {
        return this.fim;
    }
    
    /** Returns the FieldSpec at the index position.
     */
    final public int getFieldNumber( int index )
    {
       try
       {
        //return this.fim.toSource( index );
        return this.fieldNumbers[ index ];
       }
       catch( ArrayIndexOutOfBoundsException ex )
       {
          return -1;
       }
    }
    
    /** get the field number array*/
    public int[] getFieldNumbers()
    {
       return fieldNumbers.clone();
       //rdb the following isn't correct, since the FieldIDMapper reorders the
       //    sources (components).
       //rdb return this.fim.getSources(); 
    }


    /** Returns the number of FieldIDs in the RecordSpec. */
    public int size()
    {
       //return this.fim.size();
       return this.fieldNumbers.length;
    }
}
