/* 
    SDBError.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
*/

//--------------------- Edit history since 11/24/05 -------------------------------
// 11/24/05 rdb: Added logError method, mainly so that could add thrower 
//                 information (class.method) to output message. 
//               Deleted SDBError( int ) constructor; revised methods that called it
//                 to simplify them and call SDBError( int, SDBException ) instead
//               Created SDBError.Warning to indicate a non-fatal message; edited 
//                 all Granite code to replace uses of SDBError.NeverThrow with
//                 SDBError.Warning. The NeverThrow code was not intended as an error
//                 level code; it should refer to system mode indicating what to
//                 to do with exceptions.
//---------------------------------------------------------------------------------
package edu.unh.sdb.common;
//import java.io.*;

/**

 */

public class SDBError
{
    /* Constants for Error Reporting Behavior */
    public static final int AlwaysThrow    = 0;   // throw exceptions always
    public static final int ThrowFatalOnly = 1;   // only throw fatal exceptions
    public static final int NeverThrow     = 2;   // never throw exceptions

    /* Constants for instances of SDBError child classes*/
    public static final int Fatal = 0;
    public static final int Recoverable = 1;
    public static final int Warning = 2;

    /** Determines when to raise an exception for an error.*/
    private   static int   errorReportingLevel = ThrowFatalOnly;
    
    private   int          errorLevel = Recoverable;
    protected SDBException exceptionObject  = null;
    protected int          throwerIndex = 0;   // index in StackTrace of thrower.
    
    /** Constructor sets the error level for a particular error instance. 
        and takes the message to be used in the throw */
    public SDBError( int errorLevel, String message )
    {
       // throwerIndex = 1 since SDBError is the creator of the SDBException 
       //  object, but the caller of SDBError is the "logical" thrower of the
       //  exception.
       this( errorLevel, new SDBException( message ), 1 );
    }
    
    /** Constructor sets the error level for a particular error instance. */
    protected SDBError( int errorLevel, SDBException sdbex)
    {
       // throwerIndex is 0 since we assume that the creator of the exception is
       //   the thrower we want to identify
       this( errorLevel, sdbex, 1 );     
    }
       
    /**
     * private constructor that does all the work.
     */
    private SDBError( int errorLevel, SDBException sdbex, int throwerLevel )
    {
       this.errorLevel = errorLevel;
       this.exceptionObject = sdbex;
       this.throwerIndex = throwerLevel;
       
       logError( sdbex );
       
       if ( shouldThrowException() )
          throw sdbex;
    }
    /** The Error Reporting Level determines when to raise an exception 
        for an error. Currently, the choices are AlwaysThrow, ThrowFatalOnly,
        or NeverThrow. If ThrowFatalOnly is used, then only fatal errors
        generate exceptions. The other two choices should be self explanatory.
    */
    public static void setErrorReportingLevel( int level )
    {
        errorReportingLevel = (level>2) ? 2 : level;
    }
    
    public static int getErrorReportingLevel()
    {
        return errorReportingLevel;
    }
    
    /** Returns true if this instance of SDBError should throw an
        exception. Returns false otherwise.
    */
    protected boolean shouldThrowException()
    {
        switch ( getErrorReportingLevel() )
        {
            case NeverThrow:
                 return false;
            case ThrowFatalOnly:
                 return this.errorLevel == Fatal;
            case AlwaysThrow:
            default:
                 return true;
         }
    }   
    
    /** 
     * Log the error message. This should really be controlled by global logging
     * status information. For now, we just write to standard error.
     */
    protected void logError( SDBException sdbex )
    {
       String throwerPrefix = "";
       StackTraceElement thrower = sdbex.getStackTrace()[ this.throwerIndex ];
       
       throwerPrefix = thrower.getClassName() + "." + thrower.getMethodName() + ": ";

       // Should check logging status before printing
       System.err.println( "Granite error - "
                                     + throwerPrefix + sdbex.getMessage() );
    }
} 