//
//  WorkSpaceManager.java
//  
//
//  Created by Philip Rhodes on Sat Jul 20 2002.
//  Copyright (c) 2001 University of New Hampshire. All rights reserved.
//--------------------- edit history ---------------------------
// 06/06/08 rdb: made HashMap generic

package edu.unh.sdb.common;

import java.util.*;
import edu.unh.sdb.mdb.*;

/** The WorkSpaceManager class is an abstract class meant to be used in a static context only. 
 * 	It maintains a list of of WorkSpaces in its own internal data structures, and maintains a 
 * 	"current" WorkSpace. Since this class is static, the current WorkSpace can be retrieved 
 * 	from anywhere in a program using code like:
 *     
 *         WorkSpaceManager.getCurrentWorkSpace();
 *         
 *     Any other method of this class can be accessed similarly.
 *     
 *     BUGS:  The static mechanism may not work well with multi-threaded programs. 
 */
public abstract class WorkSpaceManager {

    private static HashMap<String,WorkSpace> wsList=new HashMap<String,WorkSpace>();
    private static WorkSpace theCurrentWorkSpace;
    
    /** Set the current WorkSpace to the WorkSpace with the given name. If the manager knows of
 	 *  such a WorkSpace, this method will return true. However, if the manager does not find
 	 *  a WorkSpace with the given name in its internal data structures, it will return false, 
 	 *  and leave the current WorkSpace unchanged. 
     */
    public static boolean setCurrentWorkSpace(String wsName){
    
        WorkSpace newWorkSpace = wsList.get(wsName);
        
        if(newWorkSpace!=null){
            theCurrentWorkSpace=newWorkSpace;
            return true;
        } else {
            return false;
        }
    }
    
    
    /** Set the current WorkSpace to the given WorkSpace. This method will also register an unregistered WorkSpace with
     *   the WorkSpaceManager.
     */
    public static void setCurrentWorkSpace(WorkSpace ws){
    
        registerWorkSpace(ws);
        theCurrentWorkSpace=ws;
    }
    
    /** Return the current WorkSpace, or null if no current WorkSpace has been set. */
    public static WorkSpace getCurrentWorkSpace(){
    
        return theCurrentWorkSpace;
    }
    
    /** Return the WorkSpace with the given name. */
    public static WorkSpace getWorkSpace(URL url) throws Exception{

        WorkSpace ws = wsList.get(url.dbName);

        if(ws == null){
            ws = new WorkSpace(url);
            registerWorkSpace(ws);
        }

        return ws;
	}

    
    /** Enter the given WorkSpace into the WorkSpaceManager's internal data structures. If a WorkSpace with 
     *  the same name is already registered, then no action is taken, and this method returns false. Otherwise,
     *   the given WorkSpace is registered, and this method returns true. 
     */
    public static boolean registerWorkSpace(WorkSpace ws){
    
        WorkSpace newWorkSpace = wsList.get(ws.getName());
        
        if(newWorkSpace==null){
            wsList.put(ws.getName(),ws);
            return true;
        } else {
            return false;
        }
    }
    
    
    /** For debugging and testing. */
    public static void main(String[] args){
    
        WorkSpace ws1=new WorkSpace("foo");
        WorkSpace ws2=new WorkSpace("bar");
        WorkSpace ws3=new WorkSpace("baz");
        WorkSpace ws4=new WorkSpace("boom");
        WorkSpace tempWS;
        
        WorkSpaceManager.setCurrentWorkSpace(ws1);
        WorkSpaceManager.setCurrentWorkSpace(ws2);
        WorkSpaceManager.registerWorkSpace(ws3);

        System.out.println("*******WorkSpaceManager Tests*******");
        
        tempWS=WorkSpaceManager.getCurrentWorkSpace();
        System.out.println("  WorkSpaceManager.getCurrentWorkSpace() should return 'bar'-->"+tempWS.getName());
        
        WorkSpaceManager.setCurrentWorkSpace(ws3);
        tempWS=WorkSpaceManager.getCurrentWorkSpace();
        System.out.println("  WorkSpaceManager.getCurrentWorkSpace() should return 'baz'-->"+tempWS.getName());

        WorkSpaceManager.setCurrentWorkSpace("foo");
        tempWS=WorkSpaceManager.getCurrentWorkSpace();
        System.out.println("  WorkSpaceManager.getCurrentWorkSpace() should return 'foo'-->"+tempWS.getName());
        
        System.out.println("  WorkSpaceManager.setCurrentWorkSpace(\"boom\") should return 'false'-->"+WorkSpaceManager.setCurrentWorkSpace("boom"));
        
        System.out.println("  WorkSpaceManager.registerWorkSpace(ws1) should return 'false'-->"+WorkSpaceManager.registerWorkSpace(ws1));
    
        System.out.println("  WorkSpaceManager.registerWorkSpace(ws4) should return 'true'-->"+WorkSpaceManager.registerWorkSpace(ws4));
    
        System.out.println("******* End WorkSpaceManager Tests*******");
        
    }
    
}
