/** An ARBlock has sub-datablocks as components, which are spatially joined
 * and may have different resolution level.
 *  ARBlocks are used to retrieve and store AR data.
 * 
 */

/*-------- Edit history --------------------------------------------
 * 10/21/05 rdb: Cleaned up unused variables and some reformatting.
 *               Added package empty() method to tell if block has no data
 *
 * 01/30/05 rdb: Deleted Datum datum( ... ) and DataBlock subblock( ... )
 *                        Parents now do it.
 * 11/23/04 rdb: Removed references to DatumBlock which is now deprecated
 *
 *------------------------------------------------------------------*/

package edu.unh.sdb.datasource;
import java.io.*;
import edu.unh.sdb.common.*;
public class ARBlock extends DataBlock {

    private BlockManager manager;
    private boolean hollow=true;
    private int currentBlock = 0 ; // an integer indicating the current block when iterating

    /** Construct a AR datablock with the given ComponentInfo array.
    The ComponentInfo array gives the information that describe the 
    sub-blocks's coverage and resolution.*/    
    public ARBlock( ComponentInfo[] cinfo )
    {
        this.manager = new BlockManager(cinfo);
        this.bounds = this.manager.getBounds();
        this.dimensionality = this.bounds.dim();
        this.recordDesc = this.manager.getRecordDescriptor();
        this.numAttributes = this.recordDesc.size();
    }
    
    /** 
     * Return the number of elements in the datablock. 
     */
    public int volume()
    {
        return this.manager.getBounds().volume();
    }
    
    /** 
     * @deprecated 
     * Return the number of elements in the datablock. 
     */
    public int size()
    {
        return volume();
    }
    
    /** Return by reference the Datum corresponding to the given 
     *  IndexSpaceID.
     */
    public void datum(Datum dtm, IndexSpaceID isid)
    {
        manager.datum(dtm, isid);
        return;
    }
    
    
    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum(Datum dtm, IndexSpaceID isid, FieldIDMapper mp){
        manager.datum(dtm, isid, mp);
        return;
    }
    
    
    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum(Datum dtm, IndexSpaceID isid, RecordSpec recSpec){
        manager.datum(dtm, isid, recSpec);
        return;
    }
       
    /** Set the location specified by the IndexSpaceID to the given datum.*/
    public void setDatum( IndexSpaceID isid, Datum d )
    {
        int length;
        ComponentInfo[] components = this.manager.getComponentInfo();
        length = components.length;
        
        int i = 0;
        for ( ;
            (i<length) && 
              !(isid.containedBy(components[i].getISIDMap().getCompositeExtent()));
            i++)
           ;
        
        DataBlock block = (DataBlock)components[i].getComponent();
        block.setDatum(components[i].getISIDMap().toComponent(isid), d);
    }
    
	/** Get a block of data of the DataSource. The DataBlock is passed by reference*/
    public	void subblock( DataBlock blk, 
                            ISBounds blkisb, ISBounds dsisb, 
                            RecordSpec spec)
    { 
        manager.subblock( blk, blkisb, dsisb, spec );
    }
    
    /** Get a block of data of the DataSource, pass by reference.*/
    public void subblock(DataBlock blk, ISBounds blkisb, ISBounds dsisb){
        manager.subblock(blk, blkisb, dsisb);
    }
    
    /** Pass by reference subblock method taking FieldIDMapper */
    public void subblock(DataBlock blk, ISBounds blkisb, ISBounds dsisb, FieldIDMapper mp) {
        manager.subblock(blk, blkisb, dsisb, mp);
    
    }
    
    /** Fill in this ARBlock with actual data */
    public void fill() {
        this.hollow = false;
    }
    
    /** return true if the DataBlock does not yet have any data 
     * package visibility
     */
    boolean empty()
    {
       return this.hollow;
    }

    
    /** Retrieve the next component block from this ARBlock*/
    public ComponentInfo next() {
        int index = this.currentBlock + 1;
        currentBlock ++ ;
        return this.manager.getComponentInfo(index);
        
    }
    
    /** reset the interating position */
    public void init() {
    
        currentBlock = 0;
        
    }
    
    /* Correctness testing */
    static public void	main(String[] args) throws 	IOException{
	

		AttributeDescriptor[] attribs = new AttributeDescriptor[3];
		attribs[0]=new AttributeDescriptor("float","Carbon");
		attribs[1]=new AttributeDescriptor("float","Oxygen");
		attribs[2]=new AttributeDescriptor("float","Nitrogen");
		RecordDescriptor recordDesc = new RecordDescriptor(attribs);

        /*** rdb: not used
		IntegerIndexSpaceID lower1 = new IntegerIndexSpaceID(0,0);
		IntegerIndexSpaceID upper1 = new IntegerIndexSpaceID(1,3);
		IntegerIndexSpaceID lower2 = new IntegerIndexSpaceID(3,0);
		IntegerIndexSpaceID upper2 = new IntegerIndexSpaceID(4,3);
        /**********/

		ISBounds db1=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(1,3));
		ISBounds db2=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(3,1));
		ISBounds extent1=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(0,1));
		ISBounds extent2=new ISBounds(new IntegerIndexSpaceID(1,0),new IntegerIndexSpaceID(4,1));
        int orderingarray1[]={1,0};
        AxisOrdering ordering1=new AxisOrdering(orderingarray1);

		TextFileDataSource ds1=new TextFileDataSource("",recordDesc, db1,ordering1,"testdata/test-8-3.txt");
		ds1.activate();



  		AttributeDescriptor[] attribs2 = new AttributeDescriptor[2];
		attribs2[0]=new AttributeDescriptor("float","Carbon");
		attribs2[1]=new AttributeDescriptor("float","Oxygen");
        RecordDescriptor recordDesc2 = new RecordDescriptor(attribs2); // 2 attributes

		TextFileDataSource ds2=new TextFileDataSource("",recordDesc2, db2,ordering1,"testdata/test-8-2.txt");
		ds2.activate();


		//FieldID Mapper
		int specarray[] = {0,2};
        RecordSpec rspec1 = new RecordSpec(specarray);
		int fids[] = {0,1};
		FieldIDMapper fidmapper1 = new FieldIDMapper(specarray, fids);
		
        //ISID Mapper
   		ISIDMapper ismapper1 = new ISIDMapper(db1,extent1);
   		ISIDMapper ismapper2 = new ISIDMapper(db2,extent2);
   	    ComponentInfo[] components = new ComponentInfo[2];
		components[0] = new ComponentInfo(ds1,fidmapper1,ismapper1) ;//fidmapper1
		components[1] = new ComponentInfo(ds2,null,ismapper2) ;
		
		BlockedDataSource bds1 = new BlockedDataSource("",components);
		System.out.println(bds1.getBounds());

	    ARBlock ablock = bds1.ARRep(new ISBounds(new IntegerIndexSpaceID(0,0),
	                                new IntegerIndexSpaceID(2,1)));
        Datum dResult, dCorrect;
        IntegerIndexSpaceID id1 = new IntegerIndexSpaceID(0,0);
        System.out.println("****** Correctness Testing -- Datum test ******");
        dResult = ablock.datum(id1);
        dCorrect = ds1.datum(id1,rspec1);
        System.out.println("Result Datum: "+dResult);
        System.out.println("Correct Datum: "+dCorrect);
        System.out.println("");

        id1 = new IntegerIndexSpaceID(0,1);
        dResult = ablock.datum(id1);
        dCorrect = ds1.datum(new IntegerIndexSpaceID(0,2),rspec1);
        System.out.println("Result Datum: "+dResult);
        System.out.println("Correct Datum: "+dCorrect);
        System.out.println("");
        
        id1 = new IntegerIndexSpaceID(2,1);
        dResult = ablock.datum(id1);
        dCorrect = ds2.datum(new IntegerIndexSpaceID(1,1));
        System.out.println("Result Datum: "+dResult);
        System.out.println("Correct Datum: "+dCorrect);
        System.out.println("");
        
        System.out.println("\n****** Correctness Testing -- Subblock test ******");
        /****rdb never used
        ISBounds isb1 = new ISBounds( 
                 new IntegerIndexSpaceID(1,0),new IntegerIndexSpaceID(1,1));
        /***********/
        ISBounds isb2 = new ISBounds(
                  new IntegerIndexSpaceID(0,1),new IntegerIndexSpaceID(2,1));
        BasicBlock blk = (BasicBlock)ablock.subblock( isb2 );
                
        id1 = new IntegerIndexSpaceID(0,0);
        dResult = blk.datum(id1);
        dCorrect = ds1.datum(new IntegerIndexSpaceID(0,2),rspec1);
        System.out.println("Result Datum: "+dResult);
        System.out.println("Correct Datum: "+dCorrect);
        System.out.println("");

        id1 = new IntegerIndexSpaceID(1,0);
        dResult = blk.datum(id1);
        dCorrect = ds2.datum(new IntegerIndexSpaceID(0,1));
        System.out.println("Result Datum: "+dResult);
        System.out.println("Correct Datum: "+dCorrect);
        System.out.println("");

        id1 = new IntegerIndexSpaceID(2,0);
        dResult = blk.datum(id1);
        dCorrect = ds2.datum(new IntegerIndexSpaceID(1,1));
        System.out.println("Result Datum: "+dResult);
        System.out.println("Correct Datum: "+dCorrect);
        System.out.println("");
        
/*		ISIterator iter = new ISIterator(bds1.getBounds());
        while (iter.valid()) {
            System.out.print(iter);
		    System.out.println(bds1.datum(iter));
		    iter.next();
		
		}

		iter = new ISIterator(ablock.getBounds());
        while (iter.valid()) {
            System.out.print(iter);
		    System.out.println(ablock.datum(iter));
		    iter.next();
		
		}
*/
        /*// Test Datum methods
        System.out.println("Test BlockedDataSource.datum(IndexSpaceID) method");
        IntegerIndexSpaceID isid = new IntegerIndexSpaceID(0,0);

        int specarray3[] = {1};
        RecordSpec rspec3 = new RecordSpec(specarray3);
        int specarray2[] = {2};
        RecordSpec rspec2 = new RecordSpec(specarray2);

        Datum d1,d2;
        d1 = bds1.datum(isid);
        d2 = ds1.datum(new IntegerIndexSpaceID(0,0),rspec1);        
        System.out.println(isid+" :");
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        
        isid = new IntegerIndexSpaceID(1,2);
        d1 = bds1.datum(isid);
        d2 = ds1.datum(new IntegerIndexSpaceID(1,2),rspec1);
        System.out.println(isid+" :");
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);

        isid = new IntegerIndexSpaceID(2,3);
        d1 = bds1.datum(isid);
        d2 = new FloatDatum(2);
        d2.setNull(0);
        d2.setNull(1);
        System.out.println(isid+" :");
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        
        isid = new IntegerIndexSpaceID(3,3);
        d1 = bds1.datum(isid);
        d2 = ds2.datum(new IntegerIndexSpaceID(0,3));
        System.out.println(isid+" :");
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        
        isid = new IntegerIndexSpaceID(4,0);
        d1 = bds1.datum(isid);
        d2 = ds2.datum(new IntegerIndexSpaceID(1,0));
        System.out.println(isid+" :");
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        
        
        
        System.out.println("\nTest BlockedDataSource.datum(RecordSpec,IndexSpaceID) method");
        isid = new IntegerIndexSpaceID(4,0);
        d1 = bds1.datum(isid,rspec3);
        d2 = ds2.datum(new IntegerIndexSpaceID(1,0),rspec3);
        System.out.println(isid+" :");
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        
        isid = new IntegerIndexSpaceID(1,2);
        d1 = bds1.datum(isid,rspec3);
        d2 = ds1.datum(new IntegerIndexSpaceID(1,2),rspec2);
        System.out.println(isid+" :");
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        
        //test subblock methods
        System.out.println("\nTest BlockedDataSource.subblock(ISBounds) method");
        ISBounds isb_test = new ISBounds(new IntegerIndexSpaceID(1,1),new IntegerIndexSpaceID(3,2));
        DataBlock block_test = bds1.subblock(isb_test);
        d1 = block_test.datum(new IntegerIndexSpaceID(0,1));
        d2 = ds1.datum(new IntegerIndexSpaceID(1,2),rspec1);
   
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        System.out.println("");
        d1 = block_test.datum(new IntegerIndexSpaceID(2,0));
        d2 = ds2.datum(new IntegerIndexSpaceID(0,1));
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        
        System.out.println("\nTest BlockedDataSource.subblock(ISBounds,ResordSpec) method");
        block_test = bds1.subblock(rspec3,isb_test);
        d1 = block_test.datum(new IntegerIndexSpaceID(0,0));
        d2 = ds1.datum(new IntegerIndexSpaceID(1,1),rspec2);
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        System.out.println("");
        d1 = block_test.datum(new IntegerIndexSpaceID(2,1));
        d2 = ds2.datum(new IntegerIndexSpaceID(0,2),rspec3);
        System.out.println("Result Values:\t "+ d1);
        System.out.println("Correct Values:\t "+ d2);
        */
        
        System.out.println("end");
    }
}



