/*    
    AttributeJoinDataSource.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824

    Edit History
    ------------
    11/29/05 rdb: corrected RecordDescriptor construction to handle attr renaming
    10/21/05 rdb: removed unused variables
    02/20/05 rdb: updated RecordSpec to FieldIDMapper mapping
    01/25/05 rdb: null allowed for RecordSpec and FieldIDMapper args.
    01/06/05 rdb: create Datum objects from RecordDescriptor
    11/23/04 rdb: replaced references to DatumBlock, now deprecated
    last modified 2/3/2002 (or at least last recorded date)
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;
import java.util.*;
//import java.math.*;
 
/** This class creates a new DataSource by joining attributes of
 *  some DataSources.
 */
public class AttributeJoinDataSource extends CompositeDataSource
{
    /*=================================================================*/

    /** Construct a new AJDS from the given name and an array of 
     *  ComponentInfos.     
     */
    public AttributeJoinDataSource( String aname, ComponentInfo [] cinfo ) 
    {
        super( aname, cinfo );
    }
    
    /** Construct a new AJDS from the given name, information of the
     *  components and the ISBounds of this new AJDS. The given arrays
     *  are parallel to the array of components (or ComponentInfos). 
     */
    public AttributeJoinDataSource( String aname, DataSource[] sources, 
                                    FieldIDMapper[] FIDmap,
                                    ISBounds[] componentBounds, 
                                    ISBounds compositeBounds )
    {
        super( aname, 
          createComponentInfos( sources, FIDmap, componentBounds, compositeBounds ));
    }


    /** This constructor joins all the attributes of the given 
     *  congruent DataSources. The attributes of the
     *  component DataSources are concatenated in the sequence of
     *  array index. The given DataSource array is parallel to the
     *  array of components (or ComponentInfos).
     */
    public AttributeJoinDataSource( String aname, DataSource[] sources ) 
    {
        super( aname, createComponentInfos( sources, null, null, null ));
    }

    /** Construct a new AJDS by joining all the attributes of the
     *  subsets specified by the given ISBounds array and the
     *  corresponding DataSource array. The given arrays are parallel to
     *  the array of components (or ComponentInfos).
     */
    public AttributeJoinDataSource( String aname, DataSource[] sources, 
                                   ISBounds[] componentBounds)
    {
        super( aname, 
              createComponentInfos( sources, null, componentBounds, null ));
    }


    /** This constructor joins all the attributes of the given 
     *  congruent distinct DataSources. The attributes of the
     *  component DataSources are concatenated in the sequence of
     *  array index. The given DataSource array is parallel to the
     *  array of components (or ComponentInfos). The ISBounds of
     *  this new AJDS is the given ISBounds.
     */
    public AttributeJoinDataSource( String aname, DataSource[] sources, 
                                    ISBounds compositeBounds )
    {
        super( aname, 
               createComponentInfos(sources, null, null, compositeBounds ));
    }

    /** Construct a new AJDS by joining all the attributes of the
     *  subsets specified by the given ISBounds array and the
     *  corresponding DataSource array. The ISBounds of this new AJDS
     *  is the given ISBounds. The given arrays are parallel to
     *  the array of components (or ComponentInfos).
     */
    public AttributeJoinDataSource(String aname, DataSource[] sources, 
                                   ISBounds[] componentBounds, 
                                   ISBounds compositeBounds) 
    {
        super( aname, 
              createComponentInfos(sources, null, componentBounds, compositeBounds));
    }

    /*================================================================*/


    /** Creates an array of ComponentInfo from the given information for
     *  each component. The given arrays should be parallel to the 
     *  returned array if provided (not null). The elements in the given
     *  DataSource array may have duplicates.
     */
    public static ComponentInfo[] createComponentInfos( 
                               DataSource[] sources,
                               FieldIDMapper[] FIDmap, 
                               ISBounds[] componentBounds,
                               ISBounds compositeBounds )
   {
      int size = sources.length;
      int leng, totAttri = 0;
      int[] cpnFIDs, cpsFIDs;
      ComponentInfo[] cpnInfos = new ComponentInfo[size];

      if ( FIDmap == null )
      {
         /* create FieldIDMappers. */
         FIDmap = new FieldIDMapper[size];
         for ( int i = 0; i < size; i++ )
         {
            leng = sources[i].getNumAttributes();
            cpnFIDs = new int[leng];
            cpsFIDs = new int[leng];
            for ( int j = 0; j < leng; j++ )
            {
               cpnFIDs[j] = j;
               cpsFIDs[j] = totAttri + j;
            }
            FIDmap[i] = new FieldIDMapper( cpnFIDs, cpsFIDs );
            totAttri += leng;
         }
      }

      if ( componentBounds == null )
      {
         /* Fix me: check congruence? */
         componentBounds = new ISBounds[size];
         for ( int i = 0; i < size; i++ )
         {
            componentBounds[i] = sources[i].copyBounds();
         }
      }

      if ( compositeBounds == null )
      {
         compositeBounds = sources[0].copyBounds();
         compositeBounds.moveToOrigin();
      }

      for ( int i = 0; i < size; i++ )
      {
         cpnInfos[i] = new ComponentInfo( sources[i], FIDmap[i],
               new ISIDMapper( componentBounds[i], compositeBounds ) );
      }

      return cpnInfos;
   }

   /** Sets this.recordDesc from this.componentInfos[]. 
    */
   private void setRecordDesc()
   {
      int totAttri = 0, leng;
      int size = this.componentInfos.length;
      int[] cpnFIDs; //component field numbers
      int[] cpsFIDs; //composite field numbers
      //DataSource[] sources = getComponentDataSources();
      FieldIDMapper[] FIDMap = getFIDMappers();
      AttributeDescriptor[] attri;
      RecordDescriptor recDesc;

      /* Find out the total number of attributes for this.recordDesc.
       * That is to find out the maximum number in the compositeFields
       * of FieldIDMapper.
       */
      for ( int i = 0; i < size; i++ )
      {
         cpsFIDs = FIDMap[i].getCompositeFields();
         Arrays.sort( cpsFIDs );
         totAttri = Math.max( totAttri, cpsFIDs[cpsFIDs.length - 1] );
      }
      totAttri++ ;

      attri = new AttributeDescriptor[ totAttri ];
      for ( int i = 0; i < size; i++ )
      {
         cpnFIDs = FIDMap[i].getComponentFields();
         cpsFIDs = FIDMap[i].getCompositeFields();
         
         recDesc = this.componentInfos[ i ].rDesc;
         
         leng = cpnFIDs.length;
         for ( int j = 0; j < leng; j++ )
         {
            attri[cpsFIDs[j]] = recDesc.getAttribute( cpnFIDs[j] );
         }
      }
      this.recordDesc = new RecordDescriptor( attri );
   }

   /** initialize component-related members that are not initialized by super */
   protected void initializeComponents( ComponentInfo[] cpnInfo )
   {
      /* cpnInfo is not used here. It is used in the initialize() of BDS. */
      this.bounds = this.componentInfos[0].getISIDMap().getCompositeExtent();
      setRecordDesc();
      this.numAttributes = this.recordDesc.size();
   }

   /** Check the array of ComponentInfo for AJDS. */
   protected boolean check( ComponentInfo[] cpnInfos )
   {

      int size = cpnInfos.length;
      int max, leng, leng1;
      int[][] cpsIDs;
      int[] allcpsIDs;
      int id;
      DataSource cpnDS;
      FieldIDMapper fIDmap;
      ISIDMapper ISIDmap = cpnInfos[0].getISIDMap();
      ISBounds cpsExtent = ISIDmap.getCompositeExtent();

      /* check compositeExtent */
      for ( int i = 1; i < size; i++ )
      {
         ISIDmap = cpnInfos[i].getISIDMap();
         if ( !cpsExtent.equals( ISIDmap.getCompositeExtent() ) )
         {
            new IllegalArgumentError( SDBError.Recoverable,
                  "ISIDMappers have the same compositeExtent for AJDS" );
            return false;
         }
      }

      /* check congruent componentSubsets */
      /* The compositeExtent and componentSubset should be congruent
       * when the ISIDMapper is created. */

      /* check whether component fields is out of range */
      for ( int i = 0; i < size; i++ )
      {
         fIDmap = cpnInfos[i].getFieldIDMap();
         max = fIDmap.getMaxCpnID();
         cpnDS = (DataSource) cpnInfos[i].getComponent();
         leng = cpnDS.getNumAttributes();
         if ( max >= leng )
         {
            new IllegalArgumentError( SDBError.Recoverable,
                  "Component field number is out of range" );
            return false;
         }
      }

      /* check whether composite fields are distinct */
      cpsIDs = new int[size][];
      leng = 0;
      for ( int i = 0; i < size; i++ )
      {
         cpsIDs[i] = cpnInfos[i].getFieldIDMap().getCompositeFields();
         leng = leng + cpsIDs[i].length;
      }
      allcpsIDs = new int[leng];
      leng = 0;
      for ( int i = 0; i < size; i++ )
      {
         leng1 = cpsIDs[i].length;
         for ( int j = 0; j < leng1; j++ )
         {
            allcpsIDs[leng] = cpsIDs[i][j];
            leng++ ;
         }
      }
      for ( int i = 0; i < leng; i++ )
      {
         id = allcpsIDs[i];
         for ( int j = i + 1; j < leng; j++ )
         {
            if ( id == allcpsIDs[j] )
            {
               System.out.println( "FieldID: " + id + " " + allcpsIDs[j] );
               new IllegalArgumentError( SDBError.Recoverable,
                     "Field IDs must be distict" );
               return false;
            }
         }
      }

      return true;
   }

   /** Overrides from DataSource. */
   public void activate()
   {
      int len = this.componentInfos.length;
      DataSource ds;

      for ( int i = 0; i < len; i++ )
      {
         ds = (DataSource) this.componentInfos[i].getComponent();
         ds.activate();
      }
   }

   /** Overrides from DataSource. 
    */
   public Datum datum( IndexSpaceID did )
   {
      Datum thisDatum = this.recordDesc.createDatum();
      datum( thisDatum, did );
      return thisDatum;
   }

   /** This datum method returns values of a subset of all the 
    *  attributes. This subset is specified by RecordSpec spec.
    */
   public Datum datum( IndexSpaceID did, RecordSpec spec )
   {
      Datum thisDatum = this.recordDesc.createDatum( spec );

      datum( thisDatum, did, spec );
      return thisDatum;
   }

   /** Return the Datum by reference */
   public void datum( Datum dtm, IndexSpaceID did )
   {
      int size = this.componentInfos.length;
      DataSource ds;
      FieldIDMapper fidm;
      ISIDMapper isidm;
      ComponentInfo theCpnInfo;

      for ( int i = 0; i < size; i++ )
      {
         theCpnInfo = this.componentInfos[i];
         ds = (DataSource) theCpnInfo.getComponent();
         fidm = theCpnInfo.getFieldIDMap();
         isidm = theCpnInfo.getISIDMap();
         ds.datum( dtm, isidm.toComponent( did ), fidm );
      }
   }

   /** Fill values in the passed Datum at positions
    *  specified by the corresponding FieldIDMapper.
     *  @deprecated
    */
   public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
   {
      FieldIDMapper fidm;
      ISIDMapper isidm;
      DataSource cds; //a component DS
      ComponentInfo theCpnInfo;
      int size = this.componentInfos.length;

      /* Fill values  in dtm.
       * For each component, create a new FieldIDMapper between itself
       * and dtm, and then pass dtm down.
       */
      for ( int i = 0; i < size; i++ )
      {
         theCpnInfo = this.componentInfos[i];
         cds = (DataSource) theCpnInfo.getComponent();
         isidm = theCpnInfo.getISIDMap();

         fidm = FieldIDMapper.createCrossMap( mp, theCpnInfo.getFieldIDMap() );
         /* pass dtm down */
         if ( fidm.getComponentFields() != null )
         {
            cds.datum( dtm, isidm.toComponent( did ), fidm );
         }
      }
   }

   /** Reads a subset of the DataSource attributes specified by
    *  the RecordSpec argument to the Datum argument defined by
    *  IndexSpaceID did.
    */
   public void datum( Datum dtm, IndexSpaceID did, RecordSpec spec )
   {
      datum( dtm, did, spec.getFieldIDMapper());
   }

   /** Get a block of datums. The lower bound is always origin.
    */
   public DataBlock subblock( ISBounds db )
   {
      DataBlock aBlock;

      ISBounds dstBds = db.copy();
      dstBds.moveToOrigin();

      //FloatDatum[] data = new FloatDatum[volume];
      //for (int i=0; i<volume; i++) {
      //    data[i] = new FloatDatum(datumSize);
      //}
      /* component DataBlocks should have the same AxisOrdering */
      //aBlock = new DatumBlock(data, dstBds, new AxisOrdering(dim()), this.recordDesc);
      aBlock = new BasicBlock( dstBds, this.recordDesc,
            new AxisOrdering( dim() ) );
      subblock( aBlock, dstBds, db );
      return aBlock;
   }

   /** Get a block of datums with a subset of all the attributes
    */
   public DataBlock subblock( ISBounds db, RecordSpec spec )
   {
      int datumSize = spec.size();
      int[] fids = spec.getFieldNumbers();
      DataBlock aBlock;
      AttributeDescriptor attri[] = new AttributeDescriptor[datumSize];
      RecordDescriptor recDes;

      ISBounds dstBds = db.copy();
      dstBds.moveToOrigin();

      /* create a new RecordDescriptor for this block. */
      for ( int i = 0; i < datumSize; i++ )
      {
         attri[i] = this.recordDesc.getAttribute( fids[i] );
      }
      recDes = new RecordDescriptor( attri );

      // component DataBlocks should have the same AxisOrdering 
      /* rdb note: the line below does not seem to use the "same" AxisOrdering
       *           but the default one. 
       */
      aBlock = new BasicBlock( dstBds, recDes, new AxisOrdering( dim() ) );
      subblock( aBlock, dstBds, db, spec.getFieldIDMapper());
      return aBlock;
   }

   /** Reads a subblock defined by srcBounds into the region of the 
    *  Datablock argument defined by dstBounds. 
    */
   public void subblock( DataBlock theBlock, ISBounds dstBounds,
         ISBounds srcBounds )
   {
      subblock( theBlock, dstBounds, srcBounds, (FieldIDMapper) null );
   }

   /** Reads a subblock defined by srcBounds into the region of the 
    *  Datablock argument defined by dstBounds. The FieldIDMapper argument
    *  determines the mapping between datums in the datasource and datums 
    *  in the DataBlock. This argument can be null indicating that all fields
    *  are included.
    *  @deprecated
    */
   public void subblock( DataBlock theBlock, ISBounds dstBounds,
         ISBounds srcBounds, FieldIDMapper mp )
   {
      //@TODO do we need to map srcBounds if ds bounds not at origin???
      //   if components take care of it, we're ok
      DataSource cds; //a component DS
      FieldIDMapper fidm;
      ISIDMapper isidm;
      ComponentInfo theCpnInfo;

    /* For each component, create a new FieldIDMapper between itself
       * and theBlock, and then pass theBlock down.
       */
      for ( int i = 0; i < this.componentInfos.length; i++ )
      {
         theCpnInfo = this.componentInfos[i];
         cds = (DataSource) theCpnInfo.getComponent();
         isidm = theCpnInfo.getISIDMap();
         if ( mp == null )
         {
            fidm = theCpnInfo.getFieldIDMap();
            cds.subblock( theBlock, dstBounds, isidm.toComponent( srcBounds ),
                  fidm );
         }
         else
         {
            fidm =  FieldIDMapper.createCrossMap( mp, theCpnInfo.getFieldIDMap() );
            if ( fidm.getComponentFields() != null ) /* pass theBlock down */
               cds.subblock( theBlock, dstBounds,
                     isidm.toComponent( srcBounds ), fidm );
         }
      }
   }

   /** Reads a subblock defined by srcBounds into the region of the 
    *  Datablock argument defined by dstBounds. The RecordSpec argument 
    *  specifies a subset of the DataSource attributes to be written into 
    *  the DataBlock datums.
    */
   public void subblock( DataBlock theBlock, ISBounds dstBounds,
         ISBounds srcBounds, RecordSpec spec )
   {
      if ( spec == null )
         subblock( theBlock, dstBounds, srcBounds );
      else subblock( theBlock, dstBounds, srcBounds, spec.getFieldIDMapper());
   }
   

   public String toString()
   {

      StringBuffer result = new StringBuffer( "   AttributeJoinDataSource:\n" );
      result.append( super.toString() );
      return new String( result );
   }

   public DataSource subset( RecordDescriptor rd )
   {
      return null;
   }

   public DataSource subset( ISBounds db )
   {
      return null;
   }

   public DataSource subset( RecordDescriptor rd, ISBounds db )
   {
      return null;
   }

   /** Tests AttributeJoinDataSource. Correctness test. */
   public static void test1()
   {
      System.out.println( "***************************************\n" );
      System.out.println( "*****AttributeJoinDataSource Tests******\n" );
      System.out.println( "***************************************\n\n" );

      System.out.println( "=========================================\n" );
      System.out.println( "1. Attribute join on PhysicalDataSources\n" );
      System.out.println( "=========================================\n" );

      //AppMonitor testTime = new AppMonitor();

      PackedIndexSpaceID id0 = new PackedIndexSpaceID( 0, 0, 0 );
      PackedIndexSpaceID id1 = new PackedIndexSpaceID( 3, 3, 3 );
      //PackedIndexSpaceID id2 = new PackedIndexSpaceID( 2, 2, 2 );
      ISBounds isBounds0 = new ISBounds( id0, id1 );
      //ISBounds isBounds1 = new ISBounds( id0, id2 );
      RecordDescriptor recDesc1, recDesc2;
      AttributeDescriptor[] attrDesc1 = new AttributeDescriptor[3];
      AttributeDescriptor[] attrDesc2 = new AttributeDescriptor[4];

      DataSource[] dses;
      ISBounds[] isbs;
      String name = " ";

      attrDesc1[0] = new AttributeDescriptor( "float", "Carbon" );
      attrDesc1[1] = new AttributeDescriptor( "float", "Oxygen" );
      attrDesc1[2] = new AttributeDescriptor( "float", "Nitrogen" );
      recDesc1 = new RecordDescriptor( attrDesc1 );
      attrDesc2[0] = new AttributeDescriptor( "float", "Fe" );
      attrDesc2[1] = new AttributeDescriptor( "float", "Si" );
      attrDesc2[2] = new AttributeDescriptor( "float", "Mg" );
      attrDesc2[3] = new AttributeDescriptor( "float", "S" );
      recDesc2 = new RecordDescriptor( attrDesc2 );
      int orderingarray[] = { 0, 1, 2 };
      AxisOrdering ordering = new AxisOrdering( orderingarray );

      /* First PDS. When testing BoundsError, use isBounds1 */
      TextFileDataSource ds1 = new TextFileDataSource( name, recDesc1,
            isBounds0, ordering, "testdata/file.6.3.txt" );
      ds1.activate();
      System.out.println( "\n****TextFileDataSource ds1****\n" );
      System.out.print( "Attributes: " );
      System.out.println( ds1.getRecordDescriptor() );
      System.out.println( "\n\n" );
      AppMonitor.Init();
      System.out.println( ds1.dumpData() );
      AppMonitor.printStep();
      System.out.println( "\n\n" );

      /* Second DataSource */
      TextFileDataSource ds2 = new TextFileDataSource( name, recDesc2,
            isBounds0, ordering, "testdata/file.6.4.txt" );
      ds2.activate();
      System.out.println( "\n****TextFileDataSource ds2****\n" );
      System.out.print( "Attributes: " );
      System.out.println( ds2.getRecordDescriptor() );
      System.out.println( "\n\n" );
      System.out.println( ds2.dumpData() );
      System.out.println( "\n\n" );

      /* Join ds1 and ds2 */

      dses = new DataSource[2];
      dses[0] = ds1;
      dses[1] = ds2;

      System.out.println( "*******AJDS(String, DataSource[]) ajds1********\n" );
      AttributeJoinDataSource ajds1 = new AttributeJoinDataSource( name, dses );
      System.out.print( "Attributes: " );
      System.out.println( ajds1.getRecordDescriptor() );
      System.out.println( "\n\n" );
      System.out.println( ajds1.dumpData() );
      System.out.println( "\n\n" );

      System.out.println( "*******DataBlock(), all block of ajds1********\n" );
      //DataBlock ajds1block = ajds1.subblock( isBounds1 );
      //DataBlock ajds1block = new DatumBlock(isBounds0, 
      //ajds1.getRecordDescriptor().size());
      //ajds1.subblock(ajds1block, isBounds0);
      //Datum [] blockdata;
      //blockdata = ajds1block.data();
      //int leng = blockdata.length;
      //for (int di=0; di<leng; di++) {
      //    System.out.println(blockdata[di]);
      //}

      System.out.println( "*****AJDS(DataSource[], ISBounds[]) ajds4*******\n" );
      isbs = new ISBounds[2];
      id0 = new PackedIndexSpaceID( 1, 1, 1 );
      id1 = new PackedIndexSpaceID( 2, 2, 2 );
      isbs[0] = new ISBounds( id0, id1 ); /* for ds1 */
      //        id0 = new PackedIndexSpaceID(1, 1, 1);
      //        id1 = new PackedIndexSpaceID(2, 3, 2);
      isbs[1] = new ISBounds( id0, id1 ); /* for ds2 */

      AttributeJoinDataSource ajds4 = new AttributeJoinDataSource( name, dses,
            isbs );
      System.out.print( "Attributes: " );
      System.out.println( ajds4.getRecordDescriptor() );
      System.out.println( "\n\n" );
      System.out.print( "ISBounds: " );
      System.out.println( ajds4.getBounds() );
      System.out.println( ajds4.dumpData() );
      System.out.println( "\n\n" );

      System.out.println( "\n*********ajds5=ajds1+ajds4************\n" );
      dses = new DataSource[2];
      dses[0] = ajds1;
      dses[1] = ajds4;
      AttributeJoinDataSource ajds5 = new AttributeJoinDataSource( name, dses );
      System.out.print( "Attributes: " );
      System.out.println( ajds5.getRecordDescriptor() );
      System.out.println( "\n\n" );
      System.out.println( ajds5.dumpData() );
      System.out.println( "\n\n" );
   }

   /** Test AttributeJoinDataSource
    */
   /* args[0]=A:ASCII, B:binary; args[1]= number of 1-attribute PDSes;
    * args[2]=number of access time
    */
   public static void test2( String argv[] )
   {
      System.out.println( "***************************************\n" );
      System.out.println( "*****AttributeJoinDataSource Tests******\n" );
      System.out.println( "***************************************\n\n" );

      System.out.println( "=========================================\n" );
      System.out.println( "1. Attribute join on PhysicalDataSources\n" );
      System.out.println( "=========================================\n" );

      if ( argv.length == 0 )
      {
         System.out.println( "no arguments" );
      }

      //AppMonitor testTime = new AppMonitor();

      PackedIndexSpaceID id2D0 = new PackedIndexSpaceID( 0, 0 );
      PackedIndexSpaceID id2D1 = new PackedIndexSpaceID( 511, 511 );
      ISBounds isBds2D0 = new ISBounds( id2D0, id2D1 );
      int nPDS = Integer.parseInt( argv[1] );
      int nTimes = Integer.parseInt( argv[2] );
      System.out.println( nPDS );
      System.out.println( nTimes );
      RecordDescriptor[] recDesc1 = new RecordDescriptor[nPDS];
      AttributeDescriptor[] attrDesc = new AttributeDescriptor[1];
      int i;
      DataSource ajds;
      DataSource[] PDSes = new DataSource[nPDS];
      DataBlock ajdsBlock;
      int ordarr2D[] = { 0, 1 };
      AxisOrdering ordering2D = new AxisOrdering( ordarr2D );
      String name = " ";

      for ( i = 0; i < nPDS; i++ )
      {
         attrDesc[0] = new AttributeDescriptor( "float", "a" + i );
         recDesc1[i] = new RecordDescriptor( attrDesc );
      }

      System.out.println( "\n\n*****testing running time for AJDS*****\n" );

      /* create PDSes */
      if ( argv[0].equalsIgnoreCase( "A" ) )
      {
         for ( i = 0; i < nPDS; i++ )
         {
            PDSes[i] = new TextFileDataSource( name, recDesc1[i], isBds2D0,
                  ordering2D, "testdata/tfile" + ( i + 1 ) + "a1.dat" );
         }

         /* Join the PDSes */
         ajds = new AttributeJoinDataSource( name, PDSes );
         System.out.print( "Attributes: " );
         System.out.println( ajds.getRecordDescriptor() );
         System.out.print( "ISBounds: " );
         System.out.println( ajds.getBounds() );
         AppMonitor.Init();
         ajds.activate();
         System.out.println( "After activated:" );
         AppMonitor.printStep();
         //ajdsBlock = new DatumBlock(isBds2D0, ajds.getRecordDescriptor().size());
         for ( i = 1; i < nTimes; i++ )
         {
            //ajds.subblock(ajdsBlock, isBds2D0);
            ajdsBlock = ajds.subblock( isBds2D0 );
            ajdsBlock.getNumAttributes();  // rdb force ref, so no warning
            System.out.println( "Time to get the whole block:" );
            AppMonitor.printStep();
            System.out.println( "\n" );
            //                  ajds.dumpData();
            //                  System.out.println("Traversing time:");
            //                  testTime.PrintStep();
            //                  System.out.println("\n");
         }
      }
      else
      {
         for ( i = 0; i < nPDS; i++ )
         {
            //PDSes[i] = new OldBinaryFileDataSource(name, recDesc1[i], isBds2D0,
            //   ordering2D, "testdata/tfile"+(i+1)+"a1.bin", true);
            System.out.print( "OldBinaryFileDataSource" );
         }

         /* Join the PDSes */
         ajds = new AttributeJoinDataSource( name, PDSes );
         System.out.print( "Attributes: " );
         System.out.println( ajds.getRecordDescriptor() );
         System.out.print( "ISBounds: " );
         System.out.println( ajds.getBounds() );
         AppMonitor.Init();
         for ( i = 1; i <= nTimes; i++ )
         {
            //ajdsBlock = ajds.subblock(isBds2D0);
            System.out.println( ajds.dumpData() );
            System.out.println( "Traversing time:" );
            AppMonitor.printStep();
            System.out.println( "\n" );
         }
      }
   }

   /** AttributeJoinDataSource main test
    */
   static public void main( String[] args )
   {
      Log.init( "errors", Log.LogToFile | Log.LogToStdErr );
      SDBError.setErrorReportingLevel( SDBError.AlwaysThrow );

      test1();
      //test2(args);
   }

    
/*        
    public    Object clone(){

    return new AttributeJoinDataSource(this.recordDesc,this.ISBounds);
    }
*/
}