/* 
    AxisOrdering.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;


/**    The AxisOrdering class is used to specify the axisOrdering
 *    of axes.  Such orderings are  used when a PhysicalDataSource
 *    or DataBlock maps an IndexSpaceID of arbitrary dimensionality
 *    to the 1D offset space of a file or array. They are also
 *    used by ISIterator to determine the order of the iteration.
 *    
 *    An axisOrdering is a tuple of distinct integers in the range 0..d-1, 
 *    where d is the number of elements in the tuple. Of course, d is also
 *    the dimensionality of the axisOrdering. For example, an axisOrdering with
 *    3 elements (A 3-D axisOrdering) must contain the integers 0,1, and 2 
 *    in some arbitrary order.
 *    
 *    If an axisOrdering is of the form {0,1,2...,d-1} where the elements
 *    are in strictly ascending order, then the axisOrdering is a <I>default
 *    axisOrdering</I> for d dimensions.
 *    
 *    Conceptually, an AxisOrdering is a list of axes (dimensions) in which the
 *    position of the axes in the list denotes how the data is composed into rows,
 *    columns, planes, etc. The axes are listed in order of decreasing <i>significance</i>.
 *    For example, the default axisOrdering for a 3D dataset is {0,1,2}. If we take
 *    axis 0 to indicate planes, axis 1 to indicate rows, and axis 2 to indicate
 *    columns, then the default axisOrdering indicates that planes are composed of
 *    rows, which are in turn composed of columns. Another way of looking at this
 *    is to say that the column axis changes most rapidly, followed by the row 
 *    axis, and followed finally by the plane axis.
 *    
 *    If we instantiate an iterator with this axisOrdering and traverse a 3x2x4 space,
 *     we will visit the data in the following order:
 *    
 *<pre>
 *(plane 0)       0  1  2  3
 *                4  5  6  7
 *        
 *(plane 1)       8  9  10 11
 *                12 13 14 15
 *        
 *(plane 2)       16 17 18 19 
 *                20 21 22 23
 *</pre>
 *    
 *    The other important use for AxisOrderings is to specify the mapping between
 *    the 1D space of an array or data file, and the n-D space of a DataSource
 *    or DataBlock. This is closely related to the way AxisOrderings work with
 *    traversals in that populating an n-D space with data from a 1D source 
 *    can <i>conceptually</i> be done by traversing the space and placing each 
 *    consecutive value from the 1D source into each position of the n-D index space
 *    as it is visited.
 *    
 *    For example, if a file containing the integers 0 through 23 in sorted order were
 *    read into a 3x2x4 array using a {0,1,2} axisOrdering, the diagram above would 
 *    correspond to the result.
 */

public class AxisOrdering implements Cloneable
{
    protected int    ordering[];
    
    /** Creates an AxisOrdering of the specified size
      *  and populates it with the default axisOrdering: 0..size-1
	  */
    public AxisOrdering( int size )
    {
        if ( size <= 0 ) 
            new IllegalArgumentError( SDBError.Fatal, 
                  "The size argument must be greater than zero.");
        
        this.ordering = new int[ size ];
        
        for ( int i=0; i<size; i++)
            this.ordering[i]=i;
    }


    /** Private constructor used by copy() */
    private AxisOrdering()
    {
    }

    /** Creates an AxisOrdering from the array of integers. */
    public AxisOrdering( int ordering[] )
    {
        boolean found=true;

        if (ordering == null)
        {
            new IllegalArgumentError(SDBError.Fatal, "Can't create AxisOrdering with null array.");
        }
        
        this.ordering = ordering.clone();
        
        for ( int i=0; i< this.ordering.length && found; i++)
        {
            found=false;
            for ( int j=0; j<this.ordering.length; j++ )
                found=found || this.ordering[j]==i;
        }
        
        if ( !found) {
            new IllegalArgumentError(SDBError.Fatal,"***Error in AxisOrdering(int []). Axis list contains duplicates or is incomplete.");
        }
    }
	
	/**  Sets the Least Significant Axis to the given value by swapping the prior
	  *	 position of the argument axis with the prior Least Significant Axis.
	  */
	public void setLeastSignificantAxis(int axis){
	
		int currentLSAPosition=this.ordering.length-1;
		int currentLSA = this.ordering[currentLSAPosition];
		int i, d = this.ordering.length;
		int axisPosition=-1;
		
		// Find the current position of axis in the axisOrdering array
		for ( i=0; i<d; i++){
		
			if ( this.ordering[i]==axis){
			
				axisPosition = i;
			}
		}
		
		// Now swap the LSA with the position found above.
		int temp=this.ordering[axisPosition];
		this.ordering[axisPosition] = currentLSA;
		this.ordering[currentLSAPosition] = temp;
	}


	/**  Chnages this axisOrdering so that the Most Significant Axis is the
      *  axis specified by the argument by swapping the prior
	  *	 position of the argument axis with the prior Most Significant Axis.
	  */
	public void setMostSignificantAxis(int axis){
	
		int currentMSAPosition=0;
		int currentMSA = this.ordering[currentMSAPosition];
		int i, d = this.ordering.length;
		int axisPosition=-1;
		
		// Find the current position of axis in the axisOrdering array
		for ( i=0; i<d; i++){
		
			if ( this.ordering[i]==axis){
			
				axisPosition = i;
			}
		}
		
		// Now swap the MSA with the position found above.
		int temp=this.ordering[axisPosition];
		this.ordering[axisPosition] = currentMSA;
		this.ordering[currentMSAPosition] = temp;
	}
	

   /**    Returns an array of axis weights corresponding to the axisOrdering and 
	*	bounds parameter.
	*
	*	The axisWeights array contains multipliers or "weights" for the axes.
	*	These weights are used in mapping an n-D space to a 1D space such as
	*	an array or file.The bounds parameter denotes the shape of the n-D
	*	space.
	*	
	*	For any IndexSpaceID, the dot product of the axis values
	*	and the axis weights is the offset of the ID in a 1D space such as a file or array. 
	*	
	*	Taken individually, the axis weight for a dimension tells us the size
	*	of the 1D "step" between elements in the n-D space that are adjacent along
	*	the dimension.          
    *
    *   The bounds argument and this AxisOrdering should have the same dimensionality. If
    *   the bounds dimensionality is greater, extra dimensions are ignored. If shorter,
    *   an error is created.
	*/        
    public long [] makeAxisWeights(ISBounds bounds){

        int i,axis,product;
        int d=this.ordering.length;
        
            
        int[] dimensions=bounds.getDimensions(new int[d]);
                
        long [] axisWeights=new long[this.ordering.length];
            
        try{
        
            product=1;
            for ( i=d-1;i>=0;i--){ // Alternate (works for right-left orderings)
            //for ( i=0;i<d;i++){ // Original  (for left-right orderings)
                axis=this.ordering[i];
                axisWeights[axis]=product;
                product*=dimensions[axis];
            }
            
        } catch(ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Recoverable,"Dimensionalities of ISBounds argument and AxisOrdering do not match.");
        }
    
        return axisWeights;
    }


	/**    Returns an array of axis weights corresponding to the axisOrdering and 
	*	thedimensions parameter. 
	*
	*	The axisWeights array contains multipliers or "weights" for the axes.
	*	These weights are used in mapping an n-D space to a 1D space such as
	*	an array or file. The dimensions parameter denotes the shape of the n-D
	*	space.
	*	
	*	For any IndexSpaceID, the dot product of the axis values
	*	and the axis weights is the offset of the ID in a 1D space such as a file or array.
	*	
	*	Taken individually, the axis weight for a dimension tells us the size
	*	of the 1D "step" between elements in the n-D space that are adjacent along
	*	the dimension. 
    *
    *   The IndexSpaceID argument and this AxisOrdering should have the same dimensionality. If
    *   the IndexSpaceID dimensionality is greater, extra dimensions are ignored. If shorter,
    *   an error is created.
	*/        
    public long [] makeAxisWeights(IndexSpaceID dimensionsID){

        int i,axis,product;
        int d=this.ordering.length;
        
        int [] dimensions=dimensionsID.refCoords();
        long [] axisWeights=new long[this.ordering.length];
            
        try{
        
            product=1;
            for ( i=d-1;i>=0;i--){ // Alternate (works for right-left orderings)
            //for ( i=0;i<d;i++){ // Original  (for left-right orderings)
                axis=this.ordering[i];
                axisWeights[axis]=product;
                product*=dimensions[axis];
            }
            
        } catch(ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Recoverable,"Dimensionalities of IndexSpaceID argument and AxisOrdering do not match.");
        }
    
        return axisWeights;
    }

	/**    Returns an array of axis weights corresponding to the axisOrdering and 
	*	the dimensions parameter. 
	*
	*	The axisWeights array contains multipliers or "weights" for the axes.
	*	These weights are used in mapping an n-D space to a 1D space such as
	*	an array or file. The dimensions parameter denotes the shape of the n-D
	*	space.
	*	
	*	For any IndexSpaceID, the dot product of the axis values
	*	and the axis weights is the offset of the ID in a 1D space such as a file or array.
	*	
	*	Taken individually, the axis weight for a dimension tells us the size
	*	of the 1D "step" between elements in the n-D space that are adjacent along
	*	the dimension.        
    *
    *   The length of the array argument and the dimensinaloty of this AxisOrdering should match. If
    *   the array length is greater, extra dimensions are ignored. If shorter,
    *   an error is created.
	*/        
    public long [] makeAxisWeights(int[] dimensions){

        int i,axis,product;
        int d=this.ordering.length;
        long [] axisWeights=new long[this.ordering.length];
        
        try{        
        
            product = 1;
            for ( i=d-1; i>=0; i--){ // Alternate (works for right-left orderings)
            //for ( i=0;i<d;i++){ // Original  (for left-right orderings)
            
                axis = this.ordering[i];
                axisWeights[axis] = product;
                product *= dimensions[axis];
            }
            
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal,"Dimensionalities of int[] argument and this AxisOrdering do not match.");
        }
    
        return axisWeights;
    }

	/**    Returns an array of axis weights corresponding to the axisOrdering and 
	*	the dimensions parameter. 
	*   
    *   This version returns an array of long ints.
    *
	*	The axisWeights array contains multipliers or "weights" for the axes.
	*	These weights are used in mapping an n-D space to a 1D space such as
	*	an array or file. The dimensions parameter denotes the shape of the n-D
	*	space.
	*	
	*	For any IndexSpaceID, the dot product of the axis values
	*	and the axis weights is the offset of the ID in a 1D space such as a file or array.
	*	
	*	Taken individually, the axis weight for a dimension tells us the size
	*	of the 1D "step" between elements in the n-D space that are adjacent along
	*	the dimension.        
    *
    *   The length of the array argument and the dimensinaloty of this AxisOrdering should match. If
    *   the array length is greater, extra dimensions are ignored. If shorter,
    *   an error is created.
	*/        
    public long [] makeAxisWeights(long [] dimensions){

        int i,axis,product;
        int d=this.ordering.length;
        long [] axisWeights=new long[this.ordering.length];
        
        try{        
        
            product = 1;
            for ( i=d-1; i>=0; i--){ // Alternate (works for right-left orderings)
            //for ( i=0;i<d;i++){ // Original  (for left-right orderings)
            
                axis = this.ordering[i];
                axisWeights[axis] = product;
                product *= dimensions[axis];
            }
            
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal,"Dimensionalities of int[] argument and this AxisOrdering do not match.");
        }
    
        return axisWeights;
    }



//    /**  Map the IndexSpaceID to a single integer that serves as
//      *  as an offset into the physical file or network source. 
//      *  The second argument should be an array of axis weights computed
//      *  the makeAxisWeights() method of this class.
//      *
//      * The dimensionality of the ISID and the length of the axisWeights argument
//      * should match. If the ISID dimensionality is smaller than the axisWeights
//      * length, an error will be created. If it is larger, extra dimensions
//      * will be ignored.
//      *  @deprecated  in favor of the version with long axisWeights
//	  */
//    public static final long computeOffset(IndexSpaceID dsid, int [] axisWeights){
//        
//        int i,d=axisWeights.length;
//        long offset=0;
//        int [] coords = dsid.getCoords();
//                
//        try{
//            
//            for ( i=0;i<d;i++){
//                offset+=(long)coords[i] * axisWeights[i];
//            }
//            
//        } catch(ArrayIndexOutOfBoundsException e){
//        
//            new DimensionalityError(SDBError.Recoverable,"ID argument has wrong dimensionality.");
//        }
//        return offset;
//    }
//
//    /**  Map the coordinates array to a single integer that serves as
//      *  as an offset into the physical file or network source. 
//      *  The second argument should be an array of axis weights computed
//      *  the makeAxisWeights() method of this class.
//      *
//      * The lengths of the two arrays should match. 
//      * If the coords array is shorter than the axisWeights
//      * array, an error will be created. If it is longer, extra dimensions
//      * will be ignored.
//      *  @deprecated  in favor of the version with long axisWeights      
//	  */
//    public static final long computeOffset(int [] coords, int [] axisWeights){
//        
//        int i,d=axisWeights.length;
//        long offset=0;
//        
//        try{
//            
//            for ( i=0;i<d;i++){
//                offset+= (long) coords[i]*axisWeights[i];
//            }
//            
//        } catch(ArrayIndexOutOfBoundsException e){
//        
//            new DimensionalityError(SDBError.Recoverable,"Coords argument has wrong dimensionality.");
//        }
//        
//        return offset;
//    }

    /**  Map the IndexSpaceID to a single integer that serves as
      *  as an offset into the physical file or network source. 
      *  The second argument should be an array of axis weights computed
      *  the makeAxisWeights() method of this class.
      *
      * The dimensionality of the ISID and the length of the axisWeights argument
      * should match. If the ISID dimensionality is smaller than the axisWeights
      * length, an error will be created. If it is larger, extra dimensions
      * will be ignored.
	  */
    public static final long computeOffset(IndexSpaceID dsid, long [] axisWeights){
        
        int i,d=axisWeights.length;
        long offset=0;
        int [] coords = dsid.refCoords();
                
        try{
            
            for ( i=0;i<d;i++){
                offset += coords[i] * axisWeights[i];
            }
            
        } catch(ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Recoverable,"ID argument has wrong dimensionality.");
        }
        return offset;
    }

    /**  Map the coordinates array to a single integer that serves as
      *  as an offset into the physical file or network source. 
      *  The second argument should be an array of axis weights computed
      *  the makeAxisWeights() method of this class.
      *
      * The lengths of the two arrays should match. 
      * If the coords array is shorter than the axisWeights
      * array, an error will be created. If it is longer, extra dimensions
      * will be ignored.
	  */
    public static final long computeOffset(int [] coords, long [] axisWeights){
        
        int i,d=axisWeights.length;
        long offset=0;
        
        try{
            
            for ( i=0;i<d;i++)
            {
                offset +=  coords[i]*axisWeights[i];
            }
            
        } catch(ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Recoverable,"Coords argument has wrong dimensionality.");
        }
        
        return offset;
    }

    /** Compute the coordinates corresponding to the offset using the given axisWeights.
     *
     *  The lengths of the coords and axisWeights arguments should match. If coords is
     *  longer, extra elements are ignored. If coords is shorter, an error is created.
     *  @deprecated  in favor of the version with long axisWeights
     */
    public final void computeCoords(long offset, int [] coords, int [] axisWeights){
    
        int i, axis, d=axisWeights.length;
        
        try{
        
            for ( i=0; i<d; i++){
                axis=this.ordering[i];
                coords[axis] = (int)(offset / axisWeights[axis]);
                offset -= coords[axis] * axisWeights[axis];
            }
            
        } catch( ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal,"Dimensionalities of coords and axisWeights arguments do not match.");
        }
    }

    /** Compute the coordinates corresponding to the offset using the given axisWeights.
     *
     *  The lengths of the coords and axisWeights arguments should match. If coords is
     *  longer, extra elements are ignored. If coords is shorter, an error is created.
     *  
     */
    public final void computeCoords(long offset, int [] coords, long [] axisWeights){
    
        int i, axis, d=axisWeights.length;
        
        try{
        
            for ( i=0; i<d; i++){
                axis=this.ordering[i];
                coords[axis] = (int)(offset / axisWeights[axis]);
                offset -= coords[axis] * axisWeights[axis];
            }
            
        } catch( ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal,"Dimensionalities of coords and axisWeights arguments do not match.");
        }
    }


    /** Return the axis number stored at the given index. */
    public final int getAxis(int index){
    
        return this.ordering[index];
    }


    /** Returns the length (number of dimensions) of the axisOrdering. */
    public final int dim(){
    
        return this.ordering.length;
    }

    /** Returns the least significant axis for this axisOrdering. This axis
      *  is the one that changes most rapidly. This information
      *  is of particular use for BinaryFileDataSource, which is able
      *  to read a collection of datums lying along this dimension
      *  from the file in a single read operation.
      */
    public final int getLeastSignificantAxis(){
    
        return this.ordering[this.ordering.length-1];
    }


    /** Returns the most significant axis for this axisOrdering. This axis
      * is the one that changes least rapidly. 
      */
    public final int getMostSignificantAxis(){
    
        return this.ordering[0];
    }

    /** Returns the Nth least significant axis for this axisOrdering, counting
      *  from zero. For example, getNthLeastSignificantAxis(1)
      *  will return the axis that changes second most rapidly,
      *  while getNthLeastSignificantAxis(0) is equivalent to calling
      *  getLeastSignificantAxis().
	  */
    public final int getNthLeastSignificantAxis(int n){
    
        try{
        
            return this.ordering[this.ordering.length-(n+1)];
            
        } catch(ArrayIndexOutOfBoundsException e){
            
            new DimensionalityError(SDBError.Recoverable, "Value for N is out of bounds.");
        }
        
		return 0;
    }




    /** Returns a copy of this AxisOrdering. */
    public final AxisOrdering copy(){
    
        return (AxisOrdering) this.clone();
    }

    /** Returns a copy of this AxisOrdering. */
   @Override
    public Object clone(){
    
        AxisOrdering newOrdering=new AxisOrdering();
                
        newOrdering.ordering= ordering.clone();        
        
        return newOrdering;
    }
    
    /** Return a String representing the object. */
   @Override
    public String toString(){
        
        return this.toString("");
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        int i,d=ordering.length;
        StringBuffer result=new StringBuffer(indent+"AxisOrdering: {");
        
        for ( i=0;i<d-1;i++){
            result.append(ordering[i]);
            result.append(",");
        }

        result.append(ordering[i]);
        result.append("}");

        return new String(result);
    }

    public static void main(String []  args){
    
        System.out.println("\n\n***AxisOrdering Tests***\n");

/*        
        ISBounds theBounds=new ISBounds(new IntegerIndexSpaceID(2,2),new IntegerIndexSpaceID(6,4));
        AxisOrdering axisOrdering=new AxisOrdering(2);
        ISIterator iter=new ISIterator(theBounds,axisOrdering);
        
        System.out.println("Ordering="+axisOrdering);
        for ( iter.init();iter.valid();iter.next()){
            System.out.println(iter);
        }
*/
        AxisOrdering testOrdering=new AxisOrdering(new int [] {2,0,1});
        ISBounds testBounds=new ISBounds(new int[]{ 10, 10, 10});
        long [] axisWeights= testOrdering.makeAxisWeights(testBounds);
        
        int [] coords=new int[] {5,4,2};
        long offset=AxisOrdering.computeOffset(coords, axisWeights);
        int [] coords2=new int[3];
        
        testOrdering.computeCoords(offset, coords2, axisWeights);
        
        System.out.println("coords={"+coords[0]+", "+coords[1]+", "+coords[2]+"}");
        System.out.println("coords2={"+coords2[0]+", "+coords2[1]+", "+coords2[2]+"}");
    }

}



  /*  Example
        int [] data={	0, 1, 2,
                        3, 4, 5,
                        6, 7, 8,
                        9,10,11};
        
        IntegerIndexSpaceID shape=new IntegerIndexSpaceID(4,3); // we want to view our data array
                                                                // as a 2D array with 4 rows and 3 columns,
                                                                // even though it is really just a 1D array.
        
        AxisOrdering axisOrdering=new AxisOrdering(2); // Default 2D axisOrdering
        
        int [] axisWeights=axisOrdering.makeAxisWeights(shape); // Now we ask the axisOrdering to produce the 
                                                            // weights array, which we will always pass
                                                            // to computeOffset.
        
        
        int index;
    
        for ( int i=0;i<4;i++){
            for ( int j=0;j<3;j++){
            
                // This is where the mapping from 2D to 1D takes place.  We take our row and
                // column coordinates and make an IndexSpaceID out of them. Along with the
                // axisWeights array, this ISID is passed to AxisOrdering.computeOffset(),
                // which will return the index into the 1D array.
                index=AxisOrdering.computeOffset(new IntegerIndexSpaceID(i,j), axisWeights);
                
                
                // Now, let's go ahead and use the index to access the data array.
                System.out.print(data[index]+ " ");
    
            }
            
            System.out.println();
        }
        
        System.exit(0);
    
 */   

