/*
 BDISBoundsIterator.java
 Copyright Oct. 2002 University of New Hampshire
 Durham, NH 03824

 This program is to code the BDISBoundsIterator class, which traverses 
 the specified IndexSpace with the specified shape and the specified
 steps.

 use the while loop or for loop to traverse:

 BDISBoundsIterator iter = new BDISBoundsIterator(dsBounds, shape, steps);
 
 while (iter.valid()) 
 {
   System.out.println(iter.toString());
   iter.next();
 }

 or 

 for(iter.init(); iter.valid(); iter.next()) 
 {
    System.out.println(iter.toString());
 }
 
 --------------------- Edit history (since 12/14/04) -----------------------
 8/17/07  pjr  Fixed long standing bug in the implementation of isFullShape().
               We now use an array of booleans to record whether a shape is clipped
               on a corresponding axis.
 9/16/06  pjr  Fixed long standing bugs in startpoint code. Renamed the start point
               to anchor, since it really anchors the partitioning that is being
               iterated over. next() now resets fullShape to true, so this doesn't
               "stick" anymore.
 04/26/05 rdb: isFullShape() added. Returns true if the current bounds encompasses
               the complete shape specified for the iterator; false otherwise.
 03/27/05 rdb: Unified constructor code. Added comments. No functional changes.
 
 12/14/04 rdb: Restored getAxisOrdering as synonym for getOrdering.
               This makes it consistent with other classes's methods.
               getOrdering deprecated.
 ----------------------------------------------------------------------------
 */

package edu.unh.sdb.datasource;

//import java.io.*;
import edu.unh.sdb.common.*;

/**
 * The BDISBoundsIterator class traverses the specified IndexSpace with the
 * specified shape and the specified steps.
 * 
 * use the while loop or for loop to traverse:
 * <p>
 * BDISBoundsIterator iter = new BDISBoundsIterator(dsBounds, shape, steps);<br>
 * while (iter.valid()) { System.out.println(iter.toString()); iter.next(); }<br>
 * <br>
 * or<br>
 * <br>
 * for(iter.init(); iter.valid(); iter.next()) {<br>
 * System.out.println(iter.toString()); }<br>
 */
public class BDISBoundsIterator extends ISBoundsIterator implements Cloneable, SDBIterator
{
	static final long serialVersionUID = 1;
	
//   private static final long serialVersionUID = 0; // for serialization inheritance
//   protected AxisOrdering ordering;

//    protected boolean valid = false;
// 
//    protected boolean[] fullShape;
//    
//    protected int[] shape;
//    protected int[] steps;
    protected int[] startPoint;
// 
//    protected int[] ubounds;
//    protected int[] lbounds;
   
   protected boolean[] positiveDirection;

   /**
    * Construct an BDISBoundsIterator that will traverse the given space with an
    * an ISBounds of the given shape (dimensions). The last argument specifies
    * the desired gaps between ISBounds in each dimension. A gaps array of all
    * zeros means the generated ISBounds all abut exactly. If a gap value is
    * positive, then there will be that many rows/columns between the generated
    * ISBounds in that dimension. If any of the gaps are negative, then the
    * generated ISBounds will overlap in the corresponding dimension. ( e.g. a
    * gap of -1 means the Bounds will overlap by 1 row/column.)
    * 
    * Values in the shape array must be >= 1, and values in the gaps array must
    * allow the iterator to make forward progress. In particular, for negative
    * gap values, the magnitude must be less than the shape value of the same
    * dimension.
    * 
    * @param space - space over which iteration is done
    * @param shape - shape of iteration block
    * @param gaps  - gaps between blocks (can be negative for overlapping iteration)
    * @param axisOrdering - AxisOrdering for the iteration 
    */
   public BDISBoundsIterator( ISBounds space, 
                            int shape[], 
                            int gaps[], 
                            AxisOrdering ordering )
   {
      super( space.dim() );
      if ( ordering == null )
         this.ordering = new AxisOrdering( space.dim());
      else
         this.ordering = ordering.copy();
      BDsetupAndCheck( shape, gaps );
      this.init( space );
   }
  
   /**
    * Construct a bounds iterator with default AxisOrdering.
    * 
    * @param space - space over which iteration is done
    * @param shape - shape of iteration block
    * @param gaps  - gaps between blocks (can be negative for overlapping iteration)
    */
   public BDISBoundsIterator( ISBounds space, int shape[], int gaps[] )
   {
      this( space, shape, gaps, new AxisOrdering( space.dim()));
   }

   /**
    * Construct a bounds iterator with no gaps and default AxisOrdering.
    * 
    * @param space - space over which iteration is done
    * @param shape - shape of iteration block
   */
   public BDISBoundsIterator( ISBounds space, int shape[] )
   {
      this( space, shape, null, new AxisOrdering( space.dim()));
   }

   /**
    * Construct a bounds iterator with no gaps and default AxisOrdering.
    * 
    * @param space - space over which iteration is done
    * @param shape - shape of iteration block
    */
   public BDISBoundsIterator( ISBounds space, ISBounds shape )
   {
      this( space, shape.getDimensions( new int[space.dim()] ), null,
            new AxisOrdering( space.dim()) );
   }
   
   /**
    * Construct a bounds iterator with no gaps.
    * 
    * @param space - space over which iteration is done
    * @param shape - shape of iteration block
    * @param axisOrdering - AxisOrdering for the space
    */
   public BDISBoundsIterator( ISBounds space, int shape[], AxisOrdering ordering )
   {
      this( space, shape, null, ordering );
   }

   /**
    * Construct a bounds iterator with no gaps.
    * 
    * @param space - space over which iteration is done
    * @param shape - shape of iteration block
    * @param axisOrdering - AxisOrdering for the space
    */
   public BDISBoundsIterator( ISBounds space, ISBounds shape, AxisOrdering ordering )
   {
      this( space, shape.getDimensions( new int[space.dim()] ), null, ordering );
   }

   /**
    * Construct an BDISBoundsIterator that will traverse the given space with an
    * an ISBounds of the given shape (dimensions).
    * 
    * The second argument allows the user to shift the bounds around in the
    * iteration space. Usually, the iteration begins with an ISBounds with a
    * lower bound equal to the lower bound of the iteration space. However, the
    * anchor argument allows the user to offset the partitioning of the iteration
    * space. In particular, the anchor argument specifies a location in the index
    * space that is also the corner points for some set of ISBounds in the
    * partitioning. Notice that this location could be either inside or outside
    * the iteration space. In the event that the anchor location is outside 
    * the iteration space, only ISBounds that intersect with the iteration space 
    * will be generated by the iteration. In the event
    * that the anchor location is somewhere in the interior of the iteration
    * space, the entire space is still covered by the iteration.
    * 
    * Values in the shape array must be >= 1.
    * 
    * @param space - space over which iteration is done
    * @param anchor - anchor position of partitioning
    * @param shape - shape of iteration block
    * @param gaps  - gaps between blocks (can be negative for overlapping iteration)
    * @param axisOrdering - AxisOrdering for the iteration
    */
   public BDISBoundsIterator( ISBounds space, IntegerIndexSpaceID anchor,
         int shape[], int gaps[], AxisOrdering ordering )
   {
      this( space, shape, gaps, ordering );
      this.init( space, anchor );
   }

   /**
    * Construct a bounds iterator with specified anchor position, no gaps 
    * and default axis axisOrdering.
    * 
    * @param space - space over which iteration is done
    * @param anchor - anchor position of partitioning
    * @param shape - shape of iteration block
    */
   public BDISBoundsIterator( ISBounds space, IntegerIndexSpaceID anchor, int shape[] )
   {
      this( space, anchor, shape, null, null );
   }

   /**
    * Construct a bounds iterator with specified anchor position and no gaps. 
    * 
    * @param space - space over which iteration is done
    * @param anchor - anchor position of partitioning
    * @param shape - shape of iteration block
    * @param axisOrdering - AxisOrdering for the space
    */
   public BDISBoundsIterator( ISBounds space, IntegerIndexSpaceID anchor,
                            int shape[], AxisOrdering ordering )
   {
      this( space, anchor, shape, null, ordering );
   }

   /**
    * Construct a bounds iterator with specified anchor position and default 
    * axis axisOrdering.
    * 
    * @param space - space over which iteration is done
    * @param anchor - anchor position of partitioning
    * @param shape - shape of iteration block
    * @param gaps  - gaps between blocks (can be negative for overlapping iteration)
    */
   public BDISBoundsIterator( ISBounds space, IntegerIndexSpaceID anchor,
                            int shape[], int gaps[] )
   {
      this( space, anchor, shape, gaps, null );
   }

   protected BDISBoundsIterator( int d )
   {
      super( d );
   }

   /*
    * Allocates space for private arrays, and checks the shape and gaps arrays
    * for improper values.
    */
   protected final void BDsetupAndCheck( int shape[], int gaps[] )
   {
      int d = this.dim();
      this.shape = new int[d];
      this.steps = new int[d];
      this.startPoint = new int[d];
      
      this.positiveDirection = new boolean[d];
      
      if ( gaps == null )
      {
         gaps = new int[d];
         for ( int i = 0; i < d; i++ )
            gaps[i] = 0;
      }

      try
      {
         for ( int i = 0; i < d; i++ )
         {
         
            /* check shape and set the value of shape */
            if ( shape[i] <= 0 )
               new IllegalArgumentError( SDBError.Fatal,
                     "Shape dimensions must be > 0." );
            this.shape[i] = shape[i];
            
            
            // Build the steps array, and check for illegal gap values.
            this.steps[i] = this.shape[i] + gaps[i];
//            if ( this.steps[i] <= 0 )
//               new IllegalArgumentError( SDBError.Fatal,
//                     "A value in the gap array is too large." );

            this.positiveDirection[i] = this.steps[i] > 0;
            
         }

         
         for ( int i = 0; i < d; i++ )
         {
         }
      } 
      catch ( ArrayIndexOutOfBoundsException e )
      {
         new DimensionalityError( SDBError.Fatal, "The dimensionality of the " +
         		"shape or gaps array doesn't match the iterator space bounds." );
      }
   }

   //------------------------------------------------------------------------
   /**
    * this is a help function for constructor, which is to set the Geometric
    * bounds and initial positon of the shape to traverse.
    */
   protected void init( ISBounds bounds )
   {
      this.ubounds = bounds.upper.clone();
      this.lbounds = bounds.lower.clone();

      int d = this.lower.length;
      
      
      for ( int i = 0; i < d; i++ ) {
      
         if(this.steps[i] > 0){
          
            this.startPoint[i] = this.lbounds[i];
         } else {
         
            this.startPoint[i] = this.ubounds[i] - this.shape[i] + 1;
         }
      }
      
      this.init();
   }

   //------------------------------------------------------------------------
   /*
    * this is a help function for constructor, which is to set the Geometric
    * bounds and initial positon of the shape to traverse.
    */
   protected void init( ISBounds bounds, IntegerIndexSpaceID anchor )
   {
      this.ubounds = bounds.upper.clone();
      this.lbounds = bounds.lower.clone();

      int d = this.lower.length;
      int rem;
      int[] anchorCoords = anchor.refCoords();

      for ( int i = 0; i < d; i++ )
      {
      
         if(this.steps[i] > 0){
            if ( anchorCoords[i] > this.lbounds[i] )
            {
               rem = ( anchorCoords[i] - this.lbounds[i]) % this.steps[i];
               
               if (rem != 0){
                  this.startPoint[i] = this.lbounds[i] - ( this.steps[i] - rem );
               } else {
                  this.startPoint[i] = this.lbounds[i];
               }
            }
            else
            {
               this.startPoint[i] = this.lbounds[i]
                     - ( this.lbounds[i] - anchorCoords[i]) % this.steps[i];
            }
         } else {
            if ( anchorCoords[i] > this.ubounds[i] )
            {
               rem = ( anchorCoords[i] - this.ubounds[i]) % this.steps[i];
               
               if (rem != 0){
                  this.startPoint[i] = this.ubounds[i] - ( this.steps[i] - rem );
               } else {
                  this.startPoint[i] = this.ubounds[i];
               }
            }
            else
            {
               this.startPoint[i] = this.ubounds[i]
                     - ( this.ubounds[i] - anchorCoords[i]) % this.steps[i];
            }
         }
      }
      this.init();
   }
   
   /** Change the bounds of the space that this object iterates over. 
    *  The iterator is reset to the first iteration value. The
    *  gaps between blocks are unchanged.
    */
   public void setBounds(ISBounds spaceBounds){
   
      set(spaceBounds, this.startPoint, this.shape);
   }

  /**
   * Set space bounds, anchor position, and shape of Iterator after
   * construction. The iterator is reset to the first iteration value. The
   * gaps between blocks are unchanged.
   */
  final void set( ISBounds space, int[] anchorCoords, int[] shape )
  {
     int d = space.dim();
     int rem;
     
     try
     {
        for ( int i = 0; i < d; i++ )
        {
           this.shape[i] = shape[i];
           this.lbounds[i] = space.lower[i];
           this.ubounds[i] = space.upper[i];
//
           if ( anchorCoords[i] > this.lbounds[i] )
           {
              rem = ( anchorCoords[i] - this.lbounds[i]) % this.steps[i];
              
              if (rem != 0){
                 this.startPoint[i] = this.lbounds[i] - ( this.steps[i] - rem );
              } else {
                 this.startPoint[i] = this.lbounds[i];
              }
           
           }
           else
           {
//
              this.startPoint[i] = this.lbounds[i]
                    - ( this.lbounds[i] - anchorCoords[i]) % this.steps[i];
           }
        }
        this.init(space);
     } 
     catch ( Exception e )
     {
        this.lower = new int[d];
        this.upper = new int[d];
//
        this.ordering = new AxisOrdering( d );
//
        this.BDsetupAndCheck( shape, null );
        this.init( space, new IntegerIndexSpaceID( anchorCoords ) );
//
        new IllegalArgumentError( SDBError.Recoverable, "One or more of the " +
        		"arguments had too large a dimensionality for this iterator." );
     }
  }

   //------------------------------------------------------------------------
   /** Initialize this BDISBoundsIterator */
   public void init()
   {
      //this.fullShape = true;
      
      if(this.fullShape == null)
         this.fullShape = new boolean[this.lower.length];

      
      for ( int i = 0; i < this.lower.length; i++ )
      {
         this.fullShape[i]=true;
         
         this.lower[i] = this.startPoint[i];
         this.upper[i] = this.startPoint[i] + this.shape[i] - 1;

         if ( this.lower[i] < this.lbounds[i] )
         {
              this.lower[i] = this.lbounds[i];
              this.fullShape[i] = false;
         }
          
         if ( this.upper[i] > this.ubounds[i] )
         {
              this.upper[i] =  this.ubounds[i];
              this.fullShape[i] = false;
         }
         
      }
      this.valid = true;
   }

   /** Set the iterator value to the last valid value in the iteration. */
   public void end()
   {
//      this.fullShape = true;  // assume fullShape
      for ( int i = 0; i < this.lower.length; i++ )
      {         
         this.fullShape[i]=true;

         this.lower[i] = this.startPoint[i] + this.steps[i]
               * ( ( this.ubounds[i] - this.startPoint[i] ) / this.steps[i] );
         this.upper[i] = this.lower[i] + this.shape[i] - 1;
         
         if ( this.lower[i] < this.lbounds[i] )
         {
            this.lower[i] = this.lbounds[i];
            this.fullShape[i] = false;
         }
         
         if ( this.upper[i] > this.ubounds[i] )
         {
            this.upper[i] = this.ubounds[i]; 
            this.fullShape[i] = false;
         }
      }
      this.valid = true;
   }

//    /**
//     * Return the AxisOrdering object for this iterator.
//     */
//    public final AxisOrdering getAxisOrdering()
//    {
//       return this.ordering;
//    }

//    /**
//     * Return the AxisOrdering object for this iterator.
//     * 
//     * @deprecated
//     */
//    public final AxisOrdering getOrdering()
//    {
//       return this.ordering;
//    }

//    /** Return the length of this iterator's space along the specified axis. */
//    public final int getAxisLength( int axis )
//    {
//       try
//       {
//          return this.ubounds[axis] - this.lbounds[axis] + 1;
//       } 
//       catch ( ArrayIndexOutOfBoundsException e )
//       {     
//          new IllegalArgumentError( SDBError.Recoverable,
//                "Argument specifies an illegal axis." );
//          return -1;
//       }
//    }

//    /** Returns an ISBounds representing the space being iterated over. */
//    public final ISBounds getSpaceBounds()
//    {
//       return new ISBounds( new IntegerIndexSpaceID( this.lbounds ),
//             new IntegerIndexSpaceID( this.ubounds ) );
//    }
// 
//    /**
//     * Returns an ISBounds representing the shape, but not the position, of the
//     * space being iterated over.
//     */
// 
//    public final ISBounds getSpaceDimensions()
//    {
//       return new ISBounds( this.getSpaceDimensionsArray() );
//    }
// 
//    /**
//     * Returns an array of ints representing the dimensions of the space being
//     * iterated over.
//     */
//    public final int[] getSpaceDimensionsArray()
//    {
//       int  d = this.lbounds.length;
//       int[] axisLengths = new int[d];
// 
//       for ( int i = 0; i < d; i++ )
//       {
//          axisLengths[i] = this.ubounds[i] - this.lbounds[i] + 1;
//       }
//       return axisLengths;
//    }
// 
//    /** Return an IndexSpaceID representing the shape of the iterator's blocks. 
//     *  This method does not handle the case in which the current iterator value
//     *  is not a full shape.
//     *
//     */
//    public final IndexSpaceID getBlockShape()
//    {
//       return new IntegerIndexSpaceID( this.shape );
//    }
// 
//    /**
//     * Return this iterator's internal shape array, representing the shape of the
//     * iterator's blocks.
//     */
//    final int[] getBlockShapeArray()
//    {
//       return this.shape;
//    }
// 
//    /**
//     * Return an IndexSpaceID representing the iterator's steps. A step value for
//     * an axis is the distance between corners of consecutive ISBounds.
//     */
//    public final IndexSpaceID getSteps()
//    {
//       return new IntegerIndexSpaceID( this.steps );
//    }
// 
//    /**
//     * Return this iterator's internal steps array. A step value for an axis is
//     * the distance between corresponding corners of consecutive ISBounds.
//     */
//    final int[] getStepsArray()
//    {
//       return this.steps;
//    }
// 
//    /**
//     * Return the gaps array used during construction. This array is not actually
//     * stored, so it is computed within this method. The gaps array represents
//     * the distance between lower and upper coordinates of adjacent bounds in the
//     * iteration. i.e. it is the gaps between the blocks.
//     */
//    final int[] getGapsArray()
//    {
//       int  d = this.lower.length;
//       int[] gaps = new int[d];
// 
//       for ( int i = 0; i < d; i++ )
//          gaps[i] = this.steps[i] - this.shape[i];
//       return gaps;
//    }

   //------------------------------------------------------------------------
   /*
    * this is a help function for next() function, which reset the axis if the
    * axis is outside its bound
    */
   protected void resetAxis( int axisNumber )
   {
      this.fullShape[axisNumber] = true;
      
      this.lower[axisNumber] = this.startPoint[axisNumber];
      this.upper[axisNumber] = this.lower[axisNumber] + this.shape[axisNumber] - 1;

   
      if ( this.lower[axisNumber] < this.lbounds[axisNumber] )
      {
         this.lower[axisNumber] = this.lbounds[axisNumber];
         this.fullShape[axisNumber] = false;
      }
      
      if ( this.upper[axisNumber] > this.ubounds[axisNumber] ) 
      {
         this.upper[axisNumber] = this.ubounds[axisNumber];
         this.fullShape[axisNumber] = false;
      }
      
      
      
   }

   /**
    * this is a help fuction for next() function, which is to set a specified
    * value to the specified axis if the value is in bounds
    */
   protected boolean incrementAxis( int axisNumber )
   {
      
      this.fullShape[axisNumber] = true;
      
//      if(this.positiveDirection[axisNumber]){
//         if(this.lower[axisNumber] == this.lbounds[axisNumber]){
//         
//            // We are at the beginning of a line, so we must use startPoint, not lower
//            if ( this.startPoint[axisNumber] + this.steps[axisNumber] > this.ubounds[axisNumber] )
//                return false;
//            else
//            {
//               this.lower[axisNumber] = this.startPoint[axisNumber] + this.steps[axisNumber];
//               this.upper[axisNumber] += this.steps[axisNumber];
//
//               if ( this.upper[axisNumber] > this.ubounds[axisNumber] ){
//                  
//                  this.upper[axisNumber] = this.ubounds[axisNumber];
//                  this.fullShape[axisNumber] = false;
//               }
//
//               return true;
//            }
//      
//         }  else {
//            
//            // Here, we can use this.lower
//            if ( this.lower[axisNumber] + this.steps[axisNumber] > this.ubounds[axisNumber] )
//                return false;
//            else
//            {
//               this.lower[axisNumber] += this.steps[axisNumber];
//               this.upper[axisNumber] += this.steps[axisNumber];
//
//               if ( this.upper[axisNumber] > this.ubounds[axisNumber] ){
//                  
//                  this.upper[axisNumber] = this.ubounds[axisNumber];
//                  this.fullShape[axisNumber] = false;
//               }
//               
//               return true;
//            }
//         }
//      } else { // negative direction
//         if(this.upper[axisNumber] == this.ubounds[axisNumber]){
//         
//            // We are at the beginning of a line, so we must use startPoint, not upper
//            if ( this.startPoint[axisNumber] + this.steps[axisNumber] + this.shape[axisNumber] < this.lbounds[axisNumber] )
//                return false;
//            else
//            {
//               this.lower[axisNumber] = this.startPoint[axisNumber] + this.steps[axisNumber];
//               this.upper[axisNumber] += this.steps[axisNumber];
//
//               if ( this.lower[axisNumber] < this.lbounds[axisNumber] ){
//                  
//                  this.lower[axisNumber] = this.lbounds[axisNumber];
//                  this.fullShape[axisNumber] = false;
//               }
//
//               return true;
//            }
//      
//         }  else {
//            
//            // Here, we can use this.upper
//            if ( this.upper[axisNumber] + this.steps[axisNumber] + this.shape[axisNumber] < this.lbounds[axisNumber] )
//                return false;
//            else
//            {
//               this.lower[axisNumber] += this.steps[axisNumber];
//               this.upper[axisNumber] = this.lower[axisNumber] + this.shape[axisNumber];
//
//               if ( this.lower[axisNumber] < this.lbounds[axisNumber] ){
//                  
//                  this.lower[axisNumber] = this.lbounds[axisNumber];
//                  this.fullShape[axisNumber] = false;
//               }
//               
//               return true;
//            }
//         }
//      
//      }

      
      
      if(this.positiveDirection[axisNumber]){
         if(this.lower[axisNumber] == this.lbounds[axisNumber]){
         
            // We are at the beginning of a line, so we must use startPoint, not lower
            if ( this.startPoint[axisNumber] + this.steps[axisNumber] > this.ubounds[axisNumber] )
                return false;
            else
            {
               this.lower[axisNumber] = this.startPoint[axisNumber] + this.steps[axisNumber];
               this.upper[axisNumber] = this.lower[axisNumber] + this.shape[axisNumber] - 1;

            }
      
         }  else {
            
            // Here, we can use this.lower
            if ( this.lower[axisNumber] + this.steps[axisNumber] > this.ubounds[axisNumber] )
                return false;
            else
            {
               this.lower[axisNumber] += this.steps[axisNumber];
               this.upper[axisNumber] = this.lower[axisNumber] + this.shape[axisNumber] - 1;

            }
         }
      } else { // negative direction
         if(this.upper[axisNumber] == this.ubounds[axisNumber]){
         
            // We are at the beginning of a line, so we must use startPoint, not upper
            if ( this.startPoint[axisNumber] + this.steps[axisNumber] + this.shape[axisNumber] < this.lbounds[axisNumber] )
                return false;
            else
            {
               this.lower[axisNumber] = this.startPoint[axisNumber] + this.steps[axisNumber];
               this.upper[axisNumber] = this.lower[axisNumber] + this.shape[axisNumber] - 1;

            }
      
         }  else {
            
            // Here, we can use this.upper
            if ( this.upper[axisNumber] + this.steps[axisNumber] < this.lbounds[axisNumber] )
                return false;
            else
            {
               this.lower[axisNumber] += this.steps[axisNumber];
               this.upper[axisNumber] = this.lower[axisNumber] + this.shape[axisNumber] - 1;

            }
         }
      
      }
      
      if ( this.lower[axisNumber] < this.lbounds[axisNumber] ){
         
         this.lower[axisNumber] = this.lbounds[axisNumber];
         this.fullShape[axisNumber] = false;
      }
      

      if ( this.upper[axisNumber] > this.ubounds[axisNumber] ){
         
         this.upper[axisNumber] = this.ubounds[axisNumber];
         this.fullShape[axisNumber] = false;
      }
      
      return true;
      
   }

   //------------------------------------------------------------------------
   /**
    * Return true if the current position is a valid position in the IndexSpace.
    */
   public boolean valid()
   {
      return this.valid;
   }

   //------------------------------------------------------------------------
   /**
    * Return true if the current position defines a bounds that is the full
    * extent of the shape specifed for the Iterator.
    */
   public boolean isFullShape()
   {
      
      for(int i=0; i<this.lower.length; i++){
         
         if ( ! this.fullShape[i]){
            
            return false;
         }
      }
         
      return true;
   }

   //------------------------------------------------------------------------
   /** Moving the BDISBoundsIterator to the next position. */
   public boolean next()
   {
      int axis, d = this.lower.length;


      if ( !this.valid )
         return false;

      for ( int i = 0; i < d; i++ )
      {
         axis = this.ordering.getNthLeastSignificantAxis( i );

         if ( this.incrementAxis( axis ) )
            return true;
         else
         {
            if ( i == d - 1 )
            {
               this.valid = false;
               return false;
            }
            else
               this.resetAxis( axis );
         }
      }
      this.valid = false;
      return false;
   }

   //------------------------------------------------------------------------
   /**
    * Return a simple ISBounds corresponding to the current value of this
    * iterator.
    */
   public ISBounds copyBounds()
   {
      return new ISBounds( this.lower, this.upper );
   }

   //------------------------------------------------------------------------
   /** Return a copy of this BDISBoundsIterator */
   public Object clone()
   {
      int d = this.lower.length;
      ISBounds space = new ISBounds( this.lbounds, this.ubounds );

      BDISBoundsIterator cpy = new BDISBoundsIterator( space, this.shape, this.steps );

      cpy.valid = this.valid;

      for ( int i = 0; i < d; i++ )
      {
         cpy.lower[i] = this.lower[i];
         cpy.upper[i] = this.upper[i];
      }
      return cpy;
   }

   /** Return a String representing the object. */
   public String toString()
   {
      return this.toString( "" );
   }

   //------------------------------------------------------------------------
   /** Return description of this BDISBoundsIterator */
   public String toString( String indent )
   {
//      String indent2 = indent + Globals.toStringIndent;
//      StringBuffer result = new StringBuffer( indent + "{ BDISBoundsIterator:\n" );
//
//      result.append( indent2 + "valid=" + this.valid + "\n" );
//      result.append( indent2 + "fullShape=" + this.isFullShape() + "\n" );
//
//      result.append( indent2 + "Iterator value is: " );
//      result.append( super.toString( "" ) + "\n" );
//
//      result.append( indent2 + "Iterator Bounds are:" );
//      result.append( new ISBounds( new IntegerIndexSpaceID( lbounds ),
//            new IntegerIndexSpaceID( ubounds ) )
//            + "\n" );
//      result.append( indent2 + "Iterator Shape is: " );
//      result.append( new IntegerIndexSpaceID( this.shape ) + "\n" );
//      result.append( indent2 + "Iterator Steps are: " );
//      result.append( new IntegerIndexSpaceID( this.steps ) + "\n" );
//
//      result.append( indent + "}" );
//
//      return new String( result );

      return super.toString(indent);
   }

   /** For debugging. Runs basic tests on the BDISBoundsIterator class. */
   static public void main( String[] args )
   {
   
//      BDISBoundsIterator cacheBlockIterator = new BDISBoundsIterator(
//                                                new ISBounds(new IntegerIndexSpaceID( 896, 896 ),
//                                                             new IntegerIndexSpaceID( 1151, 1151 )
//                                                            ),
//                                                new IntegerIndexSpaceID( 896, 896 ),
//                                                new int[] { 33, 33 }
//                                            );
//
//
//      cacheBlockIterator.setBounds(
//                           new ISBounds(new IntegerIndexSpaceID( 966, 966 ),
//                                        new IntegerIndexSpaceID( 1221, 1221 )
//                                       )
//      );
//
//      cacheBlockIterator.setBounds(
//                           new ISBounds(new IntegerIndexSpaceID( 896, 896 ),
//                                        new IntegerIndexSpaceID( 1151, 1151 )
//                                       )
//      );
//
//      cacheBlockIterator.setBounds(
//                           new ISBounds(new IntegerIndexSpaceID( 897, 897 ),
//                                        new IntegerIndexSpaceID( 1152, 1152 )
//                                       )
//      );
//
//           
//      cacheBlockIterator.setBounds(
//                           new ISBounds(new IntegerIndexSpaceID( 898, 898 ),
//                                        new IntegerIndexSpaceID( 1153, 1153 )
//                                       )
//      );
//   
//      for(cacheBlockIterator.init(); cacheBlockIterator.valid(); cacheBlockIterator.next()){
//      
//         //if(!cacheBlockIterator.isFullShape())
//            System.out.println( cacheBlockIterator );
//      }
//   
//      System.exit(0);
   
   
//      ISBounds space1 = new ISBounds( new IntegerIndexSpaceID( 0, 0, 0 ),
//                                       new IntegerIndexSpaceID( 127, 127, 127 ) );
//      BDISBoundsIterator t1 = new BDISBoundsIterator( space1, new int[] { 3, 128, 128 }, 
//                                                          new int[] { -1, -1, -1 } );
//
//      t1.end();
//      System.out.println( t1 );
//      for ( t1.init(); t1.valid(); t1.next() )
//         System.out.println( t1 );
//
//      System.exit( 0 );

//      ISBounds space2 = new ISBounds( new int[] {1,1,1}, new int[] { 4, 4, 4 } );
//      //BDISBoundsIterator t2=new BDISBoundsIterator(space1, new int[]{3, 4, 4},
//      // new int[]{-1,-1,-1}); // bug here
//      //BDISBoundsIterator t2 = new BDISBoundsIterator( space2, new int[] { 2, 2, 2 }, new int[]{-1,-1,-1} );
//      BDISBoundsIterator t2 = new BDISBoundsIterator( space2, new IntegerIndexSpaceID( new int[]{1,1,1} ), new int[] {  2, 2, 2 }  ); 
//
//      //t2.end();
//      //System.out.println( t2 );
//      for ( t2.init(); t2.valid(); t2.next() )
//         System.out.println( t2 );
//
//      //System.out.println(t2);
//
//      System.exit( 0 );
//
//      System.out.println( "*****Clone Test*****" );
//      ISBounds space = new ISBounds( new int[] { 0, 0 }, new int[] { 30, 30 } );
//      BDISBoundsIterator iter1 = new BDISBoundsIterator( space,
//                                        new int[] { 10, 10 }, new int[] { 0, -1 } );
//      BDISBoundsIterator iter2 = (BDISBoundsIterator) iter1.clone();
//
//      for ( iter1.init(); iter1.valid(); iter1.next() )
//      {
//
//         System.out.println( iter1 );
//      }
//
//      System.out.println( "\nAfter loop:" );
//      System.out.println( "iter1=" + iter1 );
//      System.out.println( "iter2=" + iter2 );
//
//      //System.exit(0);
//
//      int[] lower = { 0, 0, 0 };
//      int[] upper = { 5, 9, 7 };
//
//      int[] shape, steps;
//      shape = new int[3];
//      steps = new int[3];
//
//      shape[0] = 2;
//      shape[1] = 2;
//      shape[2] = 2;
//      steps[0] = 1;
//      steps[1] = 2;
//      steps[2] = 3;
//      System.out.println( "-------shape[0]: " + shape[0] );
//      System.out.println( "-------shape[1]: " + shape[1] );
//      System.out.println( "-------shape[3]: " + shape[2] );
//
//      ISBounds bound = new ISBounds( lower, upper );
//      BDISBoundsIterator iter = new BDISBoundsIterator( bound, shape, steps );
//      BDISBoundsIterator cpy = (BDISBoundsIterator) iter.clone();
//
//      System.out.println( "------------Iterator------------ \n" );
//      System.out.println( "---Iterator dim: " + iter.dim() );
//      System.out.println( "---iterator getBounds: " + iter.copyBounds() );
//      System.out.println( "---iterator lower: " + iter.copyLower().toString() );
//      System.out.println( "---iterator upper: " + iter.copyUpper().toString() );
//
//      for ( iter.init(); iter.valid(); iter.next() )
//      {
//         System.out.println( iter.toString() );
//         System.out.println( "lower: " + iter.copyLower().toString() );
//         System.out.println( "upper: " + iter.copyUpper().toString() );
//      }
//
//      System.out.println();
//      System.out.println( "*************" );
//      System.out.println( "After Loop:" );
//
//      System.out.println( "---iterator getBounds: " + iter.copyBounds() );
//      System.out.println( "---iterator lower: " + iter.copyLower().toString() );
//      System.out.println( "---iterator upper: " + iter.copyUpper().toString() );
//
//      //System.exit(1);
//
//      System.out.println( "" );
//      System.out.println( "--------------copy------------------ \n" );
//      System.out.println( "---copy dim: " + cpy.dim() );
//      System.out.println( "---copy getBounds: " + cpy.copyBounds() );
//      System.out.println( "---copy lower: " + cpy.copyLower().toString() );
//      System.out.println( "---copy upper: " + cpy.copyUpper().toString() );
//
//      for ( cpy.init(); cpy.valid(); cpy.next() )
//      {
//         System.out.println( cpy.toString() );
//         System.out.println( "lower: " + cpy.copyLower().toString() );
//         System.out.println( "upper: " + cpy.copyUpper().toString() );
//      }
//
//      System.out.println( "---copy getBounds: " + cpy.copyBounds() );
//      System.out.println( "---copy lower: " + cpy.copyLower().toString() );
//      System.out.println( "---copy upper: " + cpy.copyUpper().toString() );
//
//      System.out.println( "" );
//      cpy = (BDISBoundsIterator) iter.clone();
//      System.out.println( "--------------copy again---------------- \n" );
//      System.out.println( "---copy dim: " + cpy.dim() );
//      System.out.println( "---copy getBounds: " + cpy.copyBounds() );
//      System.out.println( "---copy lower: " + cpy.copyLower().toString() );
//      System.out.println( "---copy upper: " + cpy.copyUpper().toString() );
      
      
      BDISBoundsIterator fullnessIter = 
      new BDISBoundsIterator( new ISBounds( new int[] {8, 8, 8}), 
                            new IntegerIndexSpaceID(new int[] {-1, 0, -1}),
                            new int[] {4,4,4}
                            );
      
//      for(fullnessIter.init(); fullnessIter.valid(); fullnessIter.next()){
//         
//         System.out.println(fullnessIter);
//      }
      fullnessIter.end();
      fullnessIter.init();
      System.out.println(fullnessIter);
   }
}