/* 
    BlockMapper.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
   ----------------- Edit history as of 01/06/05 -------------------
   06/05/08 rdb: Generic ArrayList
   12/09/05 rdb: Added checks for 0-length components; error, but catch it. 
   10/21/05 rdb: deleted or commented out unused variable declarations
   02/20/05 rdb: updated RecordSpec to FieldIDMapper mapping
   01/30/05 rdb: deprecated datum methods that return Datums -- not needed any more.
   01/06/05 rdb: create Datum objects from RecordDescriptor
   -----------------------------------------------------------------
*/

/** BlockMapper is mainly used by BlockedDatasource. 
    It maps the IndexSpaceID to the block and corresponding IndexSpaceID inside the block.
    It also maps an ISBounds to an array of subblocks of the components.
    It's an immutable class.
*/
package edu.unh.sdb.datasource;
import edu.unh.sdb.common.*;
//import java.io.*;
import java.util.*;

public class BlockManager 
{
    protected ComponentInfo[] components;  // array of blocks, ISIDMappers and FieldIDMappers
    protected RecordDescriptor recordDesc; 
    protected ISBounds bounds;
    protected int dim;
    protected int size;  // number of components
    protected ISBounds[] blockBounds;
    protected boolean[] fmpFlags; // Flag for each component's FieldIDMapper
                                // if the FieldIDMapper is null or an identity
                                // mapper the flag is set true.
    protected int[] accessHistory; // int array stores index information for recent accesses
    protected int[] sortArray;  // this is used to indirect sort components' lower bounds
    
    /** Return true if the data is regular data, i.e. components form a regualar grid*/
    
    public static boolean isRegular(ComponentInfo[] cinfo) {
    
        return false;
        /*int i,j;
        boolean regular=true;
        if (cinfo.length<=1) return regular;
        ISBounds isb;
        long totalVolume=0;
        IndexSpaceID dimension = cinfo[0].getISIDMap().getCompositeExtent().getDimensions();
        
        for (i=0;i<cinfo.length;i++){
            
            isb = cinfo[i].getISIDMap().getCompositeExtent() ;
            if (! dimension.equals(isb.getDimensions())) return false;
            totalVolume += isb.volume();
        
        }
        
        return (BlockManager.buildISBounds(cinfo).volume() == totalVolume);
        */
        
    
    }
    
    public static ISBounds buildISBounds(ComponentInfo[] cinfo) {
        
        IntegerIndexSpaceID lower,upper;
        int LowerBound=2147483647, UpperBound=-2147483648;
        int ubound,lbound;
        int ndim,i,j;
        ndim = cinfo[0].getComponent().dim();
        lower = (IntegerIndexSpaceID) cinfo[0].getISIDMap().compositeExtent.copyUpper().copyID();
        upper = (IntegerIndexSpaceID) lower.copyID();
        
        for (i=0; i<ndim; i++) {
            for (j=0; j<cinfo.length; j++){
                ubound = cinfo[j].getISIDMap().compositeExtent.copyUpper().getCoord(i);

                lbound = cinfo[j].getISIDMap().compositeExtent.copyLower().getCoord(i);
                if (ubound	> UpperBound) {
                    UpperBound = ubound; 
                }
                
                if (lbound < LowerBound) {
                    LowerBound = lbound;
                }
            }
            upper.setCoord(i,UpperBound);
            lower.setCoord(i,LowerBound);
            UpperBound = -2147483648;
            LowerBound = 2147483647;
        }
        return new ISBounds(lower,upper);
    }   
    /** Construct a BlockManager from an array of ComponentInfo.*/
    public BlockManager (ComponentInfo[] cinfo ){
        this.components = cinfo;
        this.initialize();
    }
    
    private void initialize()
    {
        int ndim = 0; 
        int numAttrib = 0;

        // Check to see if each component's dim matches
        if ( this.components == null || this.components.length == 0 )
           new SDBError( SDBError.Recoverable, 
                              "Probable system error: 0-length components arraay.");
           
        ndim = this.components[0].getComponent().dim();

        //compute the max of upperbound for each 
        //dimension, use it for the new upperbound of the
        //BlockedDataSource. Set the lower bounds to 0.
        

        
        this.setBounds(BlockManager.buildISBounds(this.components));
        this.size = this.components.length;
        // set record descriptor for this BlockManager
        if (this.components[0].getFieldIDMap() == null || this.components[0].getFieldIDMap().isIdentityMapper() ) {
            this.recordDesc = this.components[0].getRecordDescriptor();
        }
        else {
            numAttrib = this.components[0].getFieldIDMap().size();
            AttributeDescriptor[] attribs = new AttributeDescriptor[numAttrib] ;
            if (this.components[0].getFieldIDMap() == null || this.components[0].getFieldIDMap().isIdentityMapper() ) 
            this.recordDesc = this.components[0].getRecordDescriptor();
            else {
                int[] compositeFields = this.components[0].getFieldIDMap().getCompositeFields();
                int[] componentFileds = this.components[0].getFieldIDMap().getComponentFields();
                // rdb unused int size = this.components[0].getFieldIDMap().size();
                for ( int i=0; i<numAttrib; i++) {
                    int j = 0;
                    while ( compositeFields[j] != i )
                        j++;
                    attribs[i] = this.components[0].getRecordDescriptor()
                                         .getAttribute(componentFileds[j]);
                }
                this.recordDesc = new RecordDescriptor(attribs);
            }
        }
        
        this.dim = this.bounds.dim();
        // set private parameters
        fmpFlags = new boolean[this.size];
        for ( int i=0; i<this.size; i++){
            fmpFlags[i] = true;
            if (this.components[i].getFieldIDMap() == null) fmpFlags[i] = false;
            else if (this.components[i].getFieldIDMap().isIdentityMapper() && 
                     this.components[i].getFieldIDMap().size() == 
                     this.components[i].getRecordDescriptor().size())
                     fmpFlags[i] = false;
        
        }
        
        int historySize = (int)Math.exp(Math.log(this.size)/ndim);  // average number of blocks on each dimension
        historySize = 0;
        accessHistory = new int[historySize];  // int array that stores the recent access to block.
                                               // this is for faster locating the block when calling the datum
        for ( int i=0; i<accessHistory.length; accessHistory[i]=i++);
        
        
        /* sort the component array by their lower bounds
        int min,temp;
        sortArray = new int[this.components.length];
        for (i=0;i<this.components.length;i++) sortArray[i] = i;
        for (i=0;i<sortArray.length;i++){
            min = sortArray[i];
            for(j=i;j<sortArray.length;j++){
            
                if (components[sortArray[j]].getISIDMap().getCompositeExtent().getLower().lessThanOrEqual(
                    components[sortArray[min]].getISIDMap().getCompositeExtent().getLower())) 
                    min = j;
            
            }
            
            temp = sortArray[i];
            sortArray[i] = sortArray[min];
            sortArray[min] = temp;
        
        }
        */
    }
    
    /* Check the ComponentInfo array for validity of parameters*/
    public boolean check(ComponentInfo[] cinfo)
    {
       if ( cinfo == null || cinfo.length == 0 )
          return false;
       
        int ndim = cinfo[0].getComponent().dim();
        RecordDescriptor rDesc0 = components[0].getRecordDescriptor().
                                 create( components[0].getFieldIDMap() );
        RecordDescriptor rDesc;
        for ( int i=0; i< cinfo.length; i++){
            // Check dimensionalities
            if ( cinfo[i].getComponent().dim() != ndim ) 
            {
                new DimensionalityError(SDBError.Recoverable,
                                     "Component dimesionalities do not match.");
                return false;
            }
            
            //Check ISIDMappers, they can't be null
            if ( cinfo[i].getISIDMap() == null ) 
            { 
                new IllegalArgumentError(SDBError.Fatal,"Component["+i+"].ISIDMapper is null.");
                return false;
            }
            
            //Check if field names match
            if (fmpFlags[i]) 
                rDesc = components[i].getRecordDescriptor().
                        create(components[i].getFieldIDMap());
            else 
                rDesc = components[i].getRecordDescriptor();
            
            if (rDesc.equals(rDesc0) == false) 
               return false;
        }
        
        
        
        //Check number of attribute and attribute type for each component
        int numAttribute = 0;
        if (cinfo[0].getFieldIDMap() == null)
            numAttribute = cinfo[0].getRecordDescriptor().size();
        for ( int i=0; i<cinfo.length; i++) 
        {
           //Fix me : Check the type compatibility
           //rdb  line below is dumb code to avoid compiler warning
           if ( numAttribute < 0 ) numAttribute = 0; 
        }
        return true;
    }
    
    private void setBounds(ISBounds isb){
    
        this.bounds = isb;
    }
    
    /** Return the Bounds of this BlockManager */
    public ISBounds getBounds(){
    
        return (ISBounds) this.bounds.clone();
    
    }
    
    /** Pass by reference datum method. Retunrs the Datum corresponding 
     *  to the given IndexSpaceID and FieldIDMapper.
     */
    public void datum(Datum d, IndexSpaceID id, FieldIDMapper fmap){
        IntegerIndexSpaceID isid = (IntegerIndexSpaceID) id.copyID();
        int i=0,j=0,k=0;
        boolean found=false;
        //search the accessHistory first and try to locate the component
        for(j=0; (j<accessHistory.length)&&(!found) ;j++){
            if ( components[accessHistory[j]].getISIDMap().contains(isid)) {
                found = true;
                i = accessHistory[j];
                for(k=j; k>0; k--) {
                    accessHistory[k] = accessHistory[k-1];
                }
                accessHistory[0] = i;
            }
        }
        
        
        // if not found in access history, use linear search to 
        // locate the component to which the id belongs
        if (!found) {
            //while ( i<this.size && !found ) {
            i = searchBlock(isid);
            if ( i >= 0) {
        //System.out.println("IndexSpaceID:"+isid);
        //System.out.println("in Block " + i +"  "+components[i].getISIDMap().getCompositeExtent());  
                found = true;
                
                if (accessHistory.length > 0) {
                    for(k=accessHistory.length - 1; k>0; k--) {
                        accessHistory[k] = accessHistory[k-1];
                    }
                }
            }
        }
        
        if (!found) {
            d = this.getRecordDescriptor().createDatum();
            return;
        }

        else isid = (IntegerIndexSpaceID) components[i].getISIDMap().toComponent(isid);
        
        
        if (fmap != null) {
            if (components[i].getFieldIDMap() != null) {
                FieldIDMapper directMapper =  
                       FieldIDMapper.createCrossMap( fmap, components[i].getFieldIDMap());
                components[i].getComponent().datum(d,isid ,directMapper);
                return;
            }
            else {
                components[i].getComponent().datum(d,isid ,fmap);
            }
        }
        else{
            if (components[i].getFieldIDMap() != null) {
                components[i].getComponent().datum( d, isid, components[i].getFieldIDMap());
                return;
            }
            else {
                components[i].getComponent().datum( d, isid);
                return;
            }
            
        }
    }

    /** Pass by reference datum method. Retunrs the Datum corresponding 
     *  to the given IndexSpaceID.
     */    
    public void datum(Datum d, IndexSpaceID id){
    
        this.datum(d, id, (FieldIDMapper) null);
       
    }

    /** Pass by reference datum method. Retunrs the Datum corresponding 
     *  to the given IndexSpaceID and RecordSpec.
     */    
    public void datum(Datum d, IndexSpaceID id, RecordSpec spec){
        FieldIDMapper specMapper = spec.getFieldIDMapper();
        this.datum(d, id, specMapper);
    }
    
    /** Return the Datum corresponding to the given IndexSpaceID. 
     * @deprecated
     */
    public Datum datum(IndexSpaceID id) {
    
        Datum d = this.getRecordDescriptor().createDatum();
        this.datum( d, id);
        return d;
    }
    
    /** Return the Datum corresponding to the given IndexSpaceID and RecordSpec.  
     * @deprecated
     */
    public Datum datum(IndexSpaceID id, RecordSpec spec) {
        Datum d = this.getRecordDescriptor().createDatum( spec );
        this.datum( d, id, spec);
        return d;
    }

    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The FieldIDMapper argument determines the mapping between 
        datums in the datasource and datums in the DataBlock. */    
    public void subblock( DataBlock theBlock, ISBounds dstBounds,
         ISBounds srcBounds, FieldIDMapper mp )
   {
      ISIDMapper SrcToDst = new ISIDMapper( srcBounds, dstBounds );
      int totalSize = dstBounds.volume();
      int currentSize = 0;
      ISBounds isb1, isb2;
      int length = this.components.length;

      for ( int i = 0; i < length; i++ )
      {
         isb2 = components[i].getISIDMap().getCompositeExtent();
         isb1 = hasIntersection( srcBounds, isb2 );
         if ( isb1 != null )
         {

            if ( fmpFlags[i] )
            { // if the componentInfo has Field subset info
               if ( mp != null )
               {
                  FieldIDMapper directMapper = FieldIDMapper.createCrossMap(
                                   mp, this.components[i].getFieldIDMap() );
                  this.components[i].getComponent().subblock( 
                               theBlock,
                               SrcToDst.toComposite( isb1 ),
                               components[i].getISIDMap().toComponent( isb1 ),
                                directMapper );
               }
               else
               {
                  this.components[i].getComponent().subblock( 
                        theBlock,
                        SrcToDst.toComposite( isb1 ),
                        components[i].getISIDMap().toComponent( isb1 ),
                        this.components[i].getFieldIDMap() );
               }
            }
            else
            {
               if ( mp != null )
               {
                  this.components[i].getComponent().subblock( theBlock,
                        SrcToDst.toComposite( isb1 ),
                        components[i].getISIDMap().toComponent( isb1 ), mp );
               }
               else
               {
                  this.components[i].getComponent().subblock( theBlock,
                        SrcToDst.toComposite( isb1 ),
                        components[i].getISIDMap().toComponent( isb1 ) );
               }
            }
            currentSize += SrcToDst.toComposite( isb1 ).volume();
            if ( currentSize >= totalSize )
            { // the datablock has been filled completely.
               if ( currentSize > totalSize )
               {
                  new InternalFaultError( SDBError.Fatal,
                        "BlockManager implementation error. -- currentSize("
                              + currentSize + ") > totalSize (" + totalSize  + ")." );
               }
               return;
            }
         }
      }
      return;
   }

    /**
     * Reads a subblock defined by srcBounds into the region of the Datablock
     * argument defined by dstBounds.
     */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds) {
        this.subblock (theBlock,dstBounds,srcBounds, (FieldIDMapper) null);
    }
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The RecordSpec argument specifies a subset of the DataSource
        attributes to be written into the DataBlock datums.*/
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds, RecordSpec spec) {
    
        this.subblock(theBlock,dstBounds,srcBounds, spec.getFieldIDMapper());
    
    }
 

    
    /**Return the ComponentInfo array.*/
    public ComponentInfo[] getComponentInfo(){
        return this.components;
    }
    
    /**Return the ComponentInfo specified by the index.*/
    public ComponentInfo getComponentInfo(int index){
        return this.components[index];
    }
    
    /** Return the RecordDescriptor of this BlockManager */
    public RecordDescriptor getRecordDescriptor(){
    
        return this.recordDesc;
    
    }
    
    /** Return the number of components within this BlockManager */
    
    public int size(){
    
    
        return this.size;
    
    }
    
    /** return the dimensionality of this BlockManager */
    public int dim() {
    
        return this.dim;
    
    }
    
    
    /** Return a ARBlock, which is a AR representation of the requested
        area of specified ISBounds
    */
    
    public ARBlock ARRep(ISBounds bounds) {
        
        ArrayList<ComponentInfo> list =  new ArrayList<ComponentInfo>();
        
        ISBounds isb1,isb2;
        IndexSpaceID offset = bounds.copyLower();
        offset.negate();
        int length = this.components.length;
        for ( int i=0;i<length;i++) {
            isb2 = components[i].getISIDMap().getCompositeExtent();
            isb1 = hasIntersection(bounds,isb2);
            if (isb1 != null) {
                isb2 = isb1.copy();
                isb2.move(offset);
                
                ISIDMapper newMapper = new ISIDMapper(components[i].getISIDMap().toComponent(isb1),
                                                      isb2);
                ComponentInfo comp = new ComponentInfo (components[i].getComponent(),
                                                        components[i].getFieldIDMap(),newMapper);
                list.add(comp);
            
            }
        }
        ComponentInfo[] cinfo = new ComponentInfo[list.size()];
        for( int i=0;i<cinfo.length;i++){
            cinfo[i] = list.get(i);
        }
        ARBlock rep = new ARBlock(cinfo);
        return rep;
    
    
    
    }
    
    // Helper function to find out if two ISBounds have intersection.
    protected ISBounds hasIntersection(ISBounds isb1,ISBounds isb2){
        return isb1.intersection(isb2);
        
    }
    
    // helper function to do binary search the sortArray
    public int searchBlock(IndexSpaceID isid){
        
        for (int i=0; i<this.components.length; i++){
            
            if (components[i].contains(isid)) return i; //linear search
        
        }
        return -1; // not found
    
    /*  int mid;
        boolean found=false;
        mid = (this.sortArray.length - 1) / 2;
        int high = this.sortArray.length-1;
        int low = 0;
        while (!found && (high - low) > 1) {
            
            if (components[sortArray[mid]].getISIDMap().contains(isid)) {
                found = true;
                return mid;
            
            } 
            
            else {
            
                if (isid.lessThanOrEqual(components[sortArray[mid]].getISIDMap().getCompositeExtent().getLower())) {
                    high = mid;              
                }
                
                else {                
                    low = mid;                
                }
                mid = (high - low) /2 + low;
            
            }
            
        
        }
        
        if (components[sortArray[low]].getISIDMap().contains(isid)) return low;
        if (components[sortArray[high]].getISIDMap().contains(isid)) return high;
     
         
        return -1; // not found
    */
    }
    
}
