package edu.unh.sdb.datasource;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.BitSet;
import java.util.Hashtable;

/**
 * Saves and restores a mapping between an iteration ID that uniquely identifies
 * an iteration and iteration information: the block size used in the previous
 * iteration with the same id and a boolean that said if the block id is optimal
 * (a larger block resulted in thrashing)
 * Helps keeping track of how much data was read by the dynamic chunking module.
 * If more data than the iteration subvolume was read, it signals cache thrashing.
 * @author dlipsa
 */
class BlockSizeAdaptiveOptimization {

   public BlockSizeAdaptiveOptimization (
      String iterationId, int defaultBlockEdge)
   {
      this.iterationId = iterationId;
      this.defaultIterationInfo = new IterationInfo (defaultBlockEdge, 
         TypeOfIteration.NON_EXISTENT);
      this.discoverThrashing = new DiscoverThrashingReloadBlock();
   }
   
   
   @SuppressWarnings(value="unchecked")
   public int readBlockEdge ()
   {
      IterationInfo ii;
      try
      {         
         FileInputStream fis = new FileInputStream(fileName);
         ObjectInputStream ois = new ObjectInputStream(fis);      
         mapIterationIdToInfo = (Hashtable<String, IterationInfo>)ois.readObject ();
         IterationInfo iterationInfo = mapIterationIdToInfo.get (iterationId);
         ii = (iterationInfo == null) ? defaultIterationInfo : iterationInfo;
      }
      catch (Exception e)
      {
         DynamicChunking.logger.warning (
            "Error reading block size adaptive optimization data: " +
            e.toString ());
         mapIterationIdToInfo = new Hashtable<String, IterationInfo> ();
         ii = defaultIterationInfo;
      }       
      DynamicChunking.logger.info ("READ: " + iterationId + " " + ii);
      previousTypeOfIteration = ii.typeOfIteration;
      return blockEdgeFromPreviousIterationInfo (ii);      
   }
   
   
   public void writeIterationInfo (int blockEdge, TypeOfIteration typeOfIteration)
   {
      try
      {
         IterationInfo iterationInfo = new IterationInfo (blockEdge, typeOfIteration);
         currentFromPreviousIterationInfo (iterationInfo, previousTypeOfIteration);
         mapIterationIdToInfo.put (iterationId, iterationInfo);
         FileOutputStream fos = new FileOutputStream (fileName);
         ObjectOutputStream oos = new ObjectOutputStream (fos);
         oos.writeObject (mapIterationIdToInfo);
         DynamicChunking.logger.info ("WRITE: " + iterationId + " " + iterationInfo);         
      }
      catch (IOException e)
      {
         DynamicChunking.logger.warning (
            "Error writing block size adaptive optimization data: " + 
            e.toString ());
      }
   }
   
   
   public void initDiscoverThrashing (
      IndexSpaceID subvolumeDims, int blockEdge,
      int numberOfBlocks)
   {
      //discoverThrashing.init(subvolumeDims, blockEdge);      
      discoverThrashing.init(numberOfBlocks);
   }
   
   /**
    * Check if reading 'block' doesn't make the total data read greater
    * than the subvolumeSize
    * @param block
    * @return true if reading 'block' makes the data read greater
    * than subvolumeSize, false otherwise.
    */
   public boolean isThrashingIfReadBlock (IndexSpaceID block, int blockIndex)
   {      
      //return discoverThrashing.isThrashingIfReadBlock(block);
      return discoverThrashing.isThrashingIfReadBlock(blockIndex);
   }
   
   private static int blockEdgeFromPreviousIterationInfo (IterationInfo info)
   {
      switch (info.typeOfIteration)
      {
         case THRASHING:
            return info.blockEdge / 2;
         case NO_THRASHING:
            return info.blockEdge * 2;
         case OPTIMAL:
         case NON_EXISTENT:
            return info.blockEdge;
         default:
            throw new IllegalStateException (
               "Invalid IterationType: " + info.typeOfIteration);
      }
   }
   
   private static void currentFromPreviousIterationInfo (
      IterationInfo current,
      TypeOfIteration previous)
   {
      switch (current.typeOfIteration)
      {
         case THRASHING:            
            switch (previous)
            {
               case THRASHING:
               case NON_EXISTENT:
                  break;
               case NO_THRASHING:
                  current.blockEdge /= 2;
                  current.typeOfIteration = TypeOfIteration.OPTIMAL;
                  break;
               case OPTIMAL:
                  throw new IllegalStateException (
                     "OPTIMAL previous iteration with THRASHING current iteration");
               default:
                  throw new IllegalStateException (
                     "Invalid IterationType: " + previous);
            }
         case NO_THRASHING:
            switch (previous)
            {
               case THRASHING:
                  current.typeOfIteration = TypeOfIteration.OPTIMAL;
                  break;
               case NO_THRASHING:
               case NON_EXISTENT:
                  break;
               case OPTIMAL:
                  current.typeOfIteration = TypeOfIteration.OPTIMAL;
                  break;
               default:
                  throw new IllegalStateException (
                     "Invalid IterationType: " + previous);                  
            }
            break;
         default:
            throw new IllegalStateException (
               "Invalid IterationType: " + current.typeOfIteration);
      }
   }
   
   
   /**
    * Maps between iteration ID and block size
    */
   private Hashtable<String, IterationInfo> mapIterationIdToInfo;
   private String iterationId;
   private IterationInfo defaultIterationInfo;
   private TypeOfIteration previousTypeOfIteration;
   private static final String fileName = "blockSizeOptimizationAdaptive.bin";
   DiscoverThrashingReloadBlock discoverThrashing;
      
   public enum TypeOfIteration 
   {
      THRASHING,
      NO_THRASHING,
      OPTIMAL,
      NON_EXISTENT,
      INVALID
   }
   
   private static class IterationInfo implements Serializable
   {
      IterationInfo (int blockEdge, TypeOfIteration typeOfIteration)
      {
         this.blockEdge = blockEdge;
         this.typeOfIteration = typeOfIteration;
      }
      int blockEdge;
      TypeOfIteration typeOfIteration;
      @Override
      public String toString ()
      {
         return "" + blockEdge + " " + typeOfIteration;
      }
   }
   
   private static class DiscoverThrashingSubvolumeSize
   {
      void init (IndexSpaceID subvolumeDims, int blockEdge)
      {
         subvolumeSize = 1;
         for (int i=0; i < subvolumeDims.dim (); i++)
            subvolumeSize *= Math.ceil ((double)subvolumeDims.getCoord (i) / blockEdge);
         subvolumeSize *= Math.pow (blockEdge, 3);               
      }
      
      public boolean isThrashingIfReadBlock (IndexSpaceID block)
      {      
         dataRead += block.volume ();
         return dataRead > subvolumeSize;        
      }      
      
      private long subvolumeSize;
      private long dataRead;      
   }
   
   private static class DiscoverThrashingReloadBlock
   {
      void init (int numberOfBlocks)
      {
         blockLoaded = new BitSet(numberOfBlocks);
      }
      
      public boolean isThrashingIfReadBlock (int blockIndex)
      {
         if (blockLoaded.get(blockIndex))
            return true;
         else
         {
            blockLoaded.set(blockIndex);
            return false;
         }              
      }
      
      private BitSet blockLoaded;
   }   
}
