package edu.unh.sdb.datasource;

import edu.unh.sdb.common.EulerAngles;
import edu.unh.sdb.common.EulerAnglesYXZ;
import java.util.Date;
import java.util.TreeSet;
import javax.vecmath.Vector3d;

/**
 * A cubic subvolume with edge size 'subvolumeEge' voxels and voxel size in 
 * bytes given by 'voxelSize', is partitioned in cubic blocks. 
 * This function returns the maximum block edge (for a cubic block) 
 * such the blocks touched by a slice traversing the subvolume, fit in
 * the 'availableMemory'. The angle of traversal for the slice is specified
 * using heading and pitch angles. We assume that subvolumeEdge is a power of 2.
 * @author dlipsa
 */
class BlockSizeSliceOptimization {    
   BlockSizeSliceOptimization (EulerAngles eulerAngles, 
      IndexSpaceID subvolumeEdges)
   {
      this.eulerAngle = eulerAngles.create (
         transformAngle (eulerAngles.getHeading ()),
         transformAngle (eulerAngles.getPitch ()),
         transformAngle (eulerAngles.getBank ()));      
      for (int i = 0; i < subvolumeEdges.dim (); i++)
         subvolumeEdges.setCoord (i,
            SliceUtil.nextPowerOf2 (subvolumeEdges.getCoord (i)));
      this.subvolumeEdges = subvolumeEdges;
   }
   
   public void setVoxelSize (int voxelSize)
   {
      this.voxelSize = voxelSize;
   }
   
   public void setAvailableMemory (long availableMemory)
   {
      this.availableMemory = availableMemory;
   }
   
   /**
    * Finds out the maximum block size such that the working set fits into the
    * available memory.
    * @return max block size
    */
   public int optimizeBlockSize ()
   {
      final long MIN_BLOCK_SIZE = 4*1024;
      long estimatedMemory;      
      int maxBlocks;
      int blockEdge = Math.min (
         subvolumeEdges.getCoord (0),
         Math.min (subvolumeEdges.getCoord (1), subvolumeEdges.getCoord (2)));
      long blockSize = blockSize = (long)Math.pow (blockEdge, 3) * voxelSize;
      int tableEntries = (int)(subvolumeEdges.volume () / 
         (long)Math.pow (blockEdge, 3));
      while (blockSize >= MIN_BLOCK_SIZE)
      {
         maxBlocks = maxBlocks (blockEdge, eulerAngle);         
         estimatedMemory = estimateBlockMemory (blockSize) * maxBlocks + 
            estimateBlockTableMemory (tableEntries);
         DynamicChunking.logger.fine ("Available memory = " + availableMemory + 
            ", estimated memory for block " + blockEdge + "^3 = " + estimatedMemory);
         if (estimatedMemory < availableMemory)
            break;
         tableEntries *= 8;
         blockEdge /= 2;
         blockSize = (long)Math.pow (blockEdge, 3) * voxelSize;         
      }
      if (blockSize < MIN_BLOCK_SIZE)
      {
         DynamicChunking.logger.info ("No optimization: last block " + blockEdge + "^3");
         return 0;
      }
      else
      {
         DynamicChunking.logger.info ("Using blocks of size " + blockEdge + "^3");
         return blockEdge;
      }
   }
   
   public static long estimateBlockTableMemory (long count)
   {
      final int REFERENCE_SIZE = 4;
      final int VECTOR_PAYLOAD = 16;      
      return count * REFERENCE_SIZE + VECTOR_PAYLOAD;
   }

   public static long estimateBlockMemory (long blockSize)
   {
      /* determined using SizeOf on Java 1.5 */      
      //final int BLOCK_PAYLOAD = 304;      
      /* determined using SizeOf on Java 1.6*/
      final int BLOCK_PAYLOAD = 360;
      return blockSize + BLOCK_PAYLOAD;
   }      
   
   private static double transformAngle (double angle)
   {
      if (angle < 0) angle = - angle;
      if (angle > 90) angle %= 90;
      if (angle > 45) angle = 90 - angle;
      return angle;
   }
   
   /**
    * A subvolume is partitioned in blocks of size blockEdge where the subvolume
    * has size subvolumeEdges. Each block has the min,max corners at 
    * (i,j,k)*(l/n), (i+1, j+1, k+1)*(l/n) where l is the edge of
    * the subvolume.
    * For all blocks in the subvolume, calculate the projection of two corners
    * on the normal to the iteration plane. We choose the two corners that have
    * the minimum and maximum projection value on the normal. For heading,pitch
    * between 0 and 45 degrees these corners are: (i,j+1,k), (j+1,j,k+1).
    */
   private static int maxBlocksCalculated (
      IndexSpaceID subvolumeEdges, int blockEdge, Vector3d normal)
   {
      assert normal.x >= 0 && normal.y <= 0 && normal.z >= 0;
      TreeSet<BlockProjectionIntervalValue> intervals = 
         new TreeSet<BlockProjectionIntervalValue> ();
      int simultaneousIntervals = 0;
      int max = 0;
      int nx = subvolumeEdges.getCoord (0) / blockEdge,
         ny = subvolumeEdges.getCoord (1) / blockEdge,
         nz = subvolumeEdges.getCoord (2) / blockEdge;
      for (int i = 0; i < nx; i++)
         for (int j = 0; j < ny; j++)
            for (int k = 0; k < nz; k++)
            {
               BlockProjectionIntervalValue begin = 
                  new BlockProjectionIntervalValue (
                  i*normal.x + (j+1)*normal.y + k* normal.z, IntervalMargin.BEGIN);
               intervals.add (begin);
               BlockProjectionIntervalValue end = new BlockProjectionIntervalValue (
                  (i+1)*normal.x + j*normal.y + (k+1)* normal.z, IntervalMargin.END);
               intervals.add (end);
               assert (begin.value <= end.value);
            }
      for (BlockProjectionIntervalValue value : intervals)
         switch (value.margin)
         {
            case BEGIN:
               simultaneousIntervals++;
               if (simultaneousIntervals > max)
                  max = simultaneousIntervals;
               break;
            case END:
               simultaneousIntervals--;
               break;
         }         
      return max;
   }
   
   private int maxBlocks (int blockEdge, EulerAngles angles)
   {
      int power = (int)
         (Math.log(subvolumeEdges.getCoord (0) / blockEdge) / Math.log (2));
      if (
         (angles.getHeading () % 5) == 0 &&
         (angles.getPitch () % 5) == 0 &&
         (angles.getBank () % 5) == 0 &&
         subvolumeEdges.getCoord (0) == subvolumeEdges.getCoord (1) &&
         subvolumeEdges.getCoord (1) == subvolumeEdges.getCoord (2) &&
         power <= maxBlocksCached.length
         )
      {
         if (power == 0)
            return 1;
         else
            return maxBlocksCached (
               power-1, (int)angles.getHeading (), (int)angles.getPitch ());
      }
      else
         return maxBlocksCalculated (subvolumeEdges,
            blockEdge, angles.vectorFromHeadingPitch ());
   }
   
   /**
    * Returns the maximum number of blocks intersected by a slice with the slice
    * at the angular orientation determined by (heading,pitch) and the subvolume
    * partitioned in 2^(partitionIndex+1)
    */
   private static int maxBlocksCached (int partitionIndex, int heading, int pitch)
   {
      int headingIndex = indexValue (heading);
      int pitchIndex = indexValue (pitch);
      int max = maxBlocksCached[partitionIndex][headingIndex][pitchIndex];
      return max;
   }
         
   private static void generateMaxBlocksCached (String[] argv)
   {
      if (argv.length != 3)
      {
         System.err.println ("BlockSizeOptimization twosPower step (java|matlab)");
         System.exit (13);
      }
      int twosPower = Integer.parseInt (argv[0]);
      int step = Integer.parseInt (argv[1]);
      String 
         COLUMNS_FIRST, COLUMNS_SEP, COLUMNS_LAST, 
         ROWS_FIRST, ROWS_SEP, ROWS_LAST,
         PARTITION_FIRST, PARTITION_SEP, PARTITION_LAST;
      if (argv[2].equals ("java"))
      {
         COLUMNS_FIRST = "{";
         COLUMNS_SEP=", ";
         COLUMNS_LAST="}";
         ROWS_FIRST = "{\n";
         ROWS_SEP = ",";
         ROWS_LAST="\n}";
         PARTITION_FIRST="{";
         PARTITION_SEP = ",\n";
         PARTITION_LAST = "}";
      }
      else
      {
         COLUMNS_FIRST="";
         COLUMNS_SEP=", ";
         COLUMNS_LAST="";
         ROWS_SEP=";";
         ROWS_FIRST="[";
         ROWS_LAST = "]";
         PARTITION_FIRST="{";
         PARTITION_SEP = ",";
         PARTITION_LAST="}";
         
      }
      EulerAngles eulerAngles = new EulerAnglesYXZ (0, 0, 0);
      long maxTime = 0;
      System.out.println (PARTITION_FIRST);
      IntegerIndexSpaceID subvolumeEdges = new IntegerIndexSpaceID (
         twosPower*2, twosPower*2, twosPower*2);
      for (int i = 0; i < twosPower; i++)
      {
         System.out.print (ROWS_FIRST);
         for (int heading = 0; heading <= 45; heading += step)
         {
            System.out.print(COLUMNS_FIRST);
            for (int pitch = 0; pitch <= 45; pitch += step)
            {
               eulerAngles.setHeading (heading);
               eulerAngles.setPitch (pitch);
               long beginTime = new Date ().getTime ();
               int calculated = maxBlocksCalculated (subvolumeEdges,
                  (int) Math.pow (2, i+1),
                     eulerAngles.vectorFromHeadingPitch ());
               long endTime = new Date ().getTime ();               
               System.out.print ("" + calculated);               
               if (pitch != 45)
                  System.out.print (COLUMNS_SEP);
               if (maxTime < (endTime - beginTime))
                  maxTime = endTime - beginTime;
            }
            System.out.print(COLUMNS_LAST);
            if (heading != 45)
               System.out.println (ROWS_SEP);
         }
         System.out.print(ROWS_LAST);
         if (i != (twosPower - 1))
            System.out.println (PARTITION_SEP);
      }
      System.out.println (PARTITION_LAST);
      System.err.println ("\nMax time for maxBlocks is: " + maxTime + " ms");      
   }
   
   private static void maxBlocksCalculated (String[] argv)
   {
      if (argv.length != 3)
      {
         System.out.println ("BlockSizeOptimization twosPower heading pitch");
         System.exit (13);
      }
      long beginTime = new Date ().getTime ();
      int twosPower = Integer.parseInt (argv[0]);
      int heading = Integer.parseInt (argv[1]);
      int pitch = Integer.parseInt (argv[2]);
      EulerAngles eulerAngles = new EulerAnglesYXZ (heading, pitch, 0);
      IntegerIndexSpaceID subvolumeEdges = new IntegerIndexSpaceID (
         twosPower*2, twosPower*2, twosPower*2);
      int calculated = maxBlocksCalculated (subvolumeEdges,
         (int) Math.pow (2, twosPower),
         eulerAngles.vectorFromHeadingPitch ());
      long endTime = new Date ().getTime ();      
      System.err.println (
         "Maximum blocks intersected by a slice for partition 2^" + twosPower + 
         " and angles (" + heading + ", " + pitch + ")" +
         "\n\tcalculated=" + Integer.toString (calculated) +
         " in " + (endTime - beginTime) + " ms" + 
         "vector=" + eulerAngles.vectorFromHeadingPitch ());      
   }
   
   public static void main (String[] argv)
   {
      generateMaxBlocksCached (argv);
   }
   
   private static int indexValue (int angle)
   {
      return angle / 5;
   }

   private static class BlockProjectionIntervalValue implements Comparable
   {

      public BlockProjectionIntervalValue (double value, IntervalMargin margin)
      {
         this.value = value;
         this.margin = margin;
      }
      
      public int compareTo (Object o)
      {
         BlockProjectionIntervalValue other = (BlockProjectionIntervalValue) o;
         if (other == null)
            throw new ClassCastException ("Object parameter is of wrong class");
         int compareValue = Double.compare (this.value, other.value);         
         if (compareValue == 0)
         {
            if (this.margin == IntervalMargin.BEGIN &&
               other.margin == IntervalMargin.END)
               return 1;
            else
               return -1;
         }
         else
            return compareValue;               
      }
      
      @Override
      public boolean equals (Object o)
      {
         return false;
      }

      @Override
      public int hashCode ()
      {
         throw new UnsupportedOperationException ("Not yet implemented");
      }
      
      double value;
      IntervalMargin margin;
   }
   
   private static enum IntervalMargin
   {
      BEGIN,
      END
   }
   
   private EulerAngles eulerAngle;
   private IndexSpaceID subvolumeEdges;
   private int voxelSize;
   private long availableMemory;   
   
   /**
    * maxBlocksCached[partition][heading][pitch]
    * Maximum number of blocks required for a slice traversing a subvolume
    * split into (2^('partition'+1))^3 blocks. The slice normal is rotated with
    * 'heading' and 'pitch' angles. This table contains only angles 0,5,10,...,45 
    * degrees. Results for other rotations angles can be obtained through symmetry.
    * These matrices are symmetric about the main diagonal.
    */
   static int maxBlocksCached[][][] = 
   {
 {
{4, 6, 6, 6, 6, 6, 6, 6, 6, 6},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7},
{6, 7, 7, 7, 7, 7, 7, 7, 7, 7}
},

{
{16, 20, 20, 24, 24, 24, 28, 28, 28, 28},
{20, 23, 22, 23, 26, 26, 29, 30, 31, 31},
{20, 23, 23, 24, 26, 28, 29, 31, 31, 31},
{24, 23, 24, 26, 27, 29, 31, 31, 32, 33},
{24, 26, 26, 27, 29, 30, 31, 32, 33, 33},
{24, 26, 28, 29, 31, 31, 33, 34, 33, 33},
{28, 29, 29, 31, 31, 33, 33, 35, 35, 34},
{28, 30, 31, 32, 33, 33, 35, 35, 35, 35},
{28, 31, 31, 33, 34, 35, 35, 35, 35, 34},
{28, 31, 31, 33, 35, 35, 35, 35, 35, 35}
},

{
{64, 72, 80, 88, 88, 96, 104, 112, 120, 120},
{72, 79, 84, 88, 94, 101, 109, 117, 125, 127},
{80, 84, 89, 94, 100, 107, 114, 122, 129, 133},
{88, 88, 94, 100, 107, 115, 121, 127, 133, 137},
{88, 95, 101, 107, 114, 120, 127, 133, 136, 137},
{96, 102, 108, 115, 123, 127, 133, 137, 140, 138},
{104, 109, 116, 122, 127, 134, 138, 140, 142, 140},
{112, 117, 124, 129, 134, 138, 142, 143, 143, 139},
{120, 126, 130, 135, 138, 141, 143, 143, 143, 141},
{120, 127, 133, 140, 143, 143, 145, 145, 145, 142}
},

{
{256, 288, 304, 336, 352, 384, 416, 448, 480, 496},
{288, 307, 326, 352, 374, 401, 430, 460, 497, 525},
{304, 326, 349, 373, 397, 424, 452, 485, 517, 533},
{336, 352, 373, 397, 423, 452, 480, 511, 535, 544},
{352, 376, 399, 424, 451, 480, 508, 535, 548, 548},
{384, 403, 427, 453, 481, 508, 534, 550, 559, 553},
{416, 432, 459, 484, 511, 536, 553, 563, 569, 560},
{448, 465, 492, 519, 540, 554, 565, 569, 570, 562},
{480, 501, 525, 542, 555, 565, 572, 576, 573, 563},
{496, 525, 540, 561, 569, 575, 579, 581, 575, 567}
},

{
{1024, 1120, 1216, 1312, 1408, 1504, 1632, 1760, 1888, 2016},
{1120, 1207, 1298, 1391, 1490, 1594, 1710, 1836, 1978, 2091},
{1216, 1299, 1391, 1486, 1586, 1693, 1807, 1934, 2073, 2136},
{1312, 1395, 1488, 1585, 1687, 1795, 1913, 2042, 2143, 2163},
{1408, 1493, 1592, 1693, 1797, 1917, 2030, 2139, 2196, 2194},
{1504, 1603, 1703, 1806, 1916, 2030, 2135, 2202, 2232, 2216},
{1632, 1722, 1827, 1934, 2046, 2145, 2213, 2249, 2263, 2241},
{1760, 1854, 1963, 2076, 2161, 2219, 2262, 2281, 2282, 2250},
{1888, 2002, 2105, 2173, 2224, 2260, 2289, 2299, 2293, 2258},
{2016, 2108, 2164, 2227, 2261, 2287, 2311, 2317, 2295, 2267}
}    
   };
}
