/*	
 BlockedDataSource.java
 Copyright 2000 University of New Hampshire
 Durham, NH 03824
 */

//---------------------- Edit history (as of 11/04) -----------------------
// 10/21/05 rdb: Deleted unused variables
// 02/14/05 rdb: Added several constructors for convenience and consistency with
//                 the variations available in AJDS:
//               createComponents method also added to facilitate new constructors
// 02/07/05 rdb: initialize -> initializeComponents
//               some cleaning up of constructors.
// 11/23/04 rdb: changed DatumBlock usage to BasicBlock. DatumBlock deprec
//
//-------------------------------------------------------------------------
package edu.unh.sdb.datasource;

//import java.io.*; 
import edu.unh.sdb.common.*;

/**
 * BlockedDataSource does a spatial join on several blocks of DataSources into
 * one DataSource
 */
public class BlockedDataSource extends CompositeDataSource
{

   private BlockManager manager;

   /** Construct a BlockedDataSource from an array of ComponentInfo. */
   public BlockedDataSource( String dsname, ComponentInfo[] compinfo )
   {
      super( dsname, compinfo );
      //rdb called by parent: this.initializeComponents( compinfo );
   }

   /** This constructor joins all the blocks of the input DataSources. 
    *  The attributes of all should be the same. The components are all
    *  mapped to the composite with identity ISIDMapper.
    *  The given DataSource array is parallel to the
    *  array of components (or ComponentInfos).
    */
   public BlockedDataSource( String aname, DataSource[] sources ) 
   {
       //super( aname, sources, null, null );
       super( aname, 
             createComponentInfos( sources, null, null ));
   }

   /** Construct a new BlockedDS by specified by the given ISBounds array and the
    *  corresponding DataSource array. componentPositions are the positions of the
    *  components in the composite. 
    * The given arrays are parallel to
    *  the array of components (or ComponentInfos).
    */
   public BlockedDataSource( String aname, DataSource[] sources, 
                                  ISBounds[] componentLocations )
   {
      super( aname, 
            createComponentInfos( sources, componentLocations, null ));
   }


   /** This constructor joins all the blocks of the input arrays. 
    *  The attributes of the component DataSources should be compatible 
    *  The components are all  mapped to the composite with identity ISIDMapper
    * The given DataSource array is parallel to the array of components 
    * (or ComponentInfos). The ISBounds of this new BlockedDS is the given ISBounds.
    */
   public BlockedDataSource(String aname, DataSource[] sources, 
                                  ISBounds compositeBounds)
   {
       super(aname, 
             createComponentInfos(sources, null, compositeBounds));
   }

   /**
    * Construct a BlockedDataSource from parallel arrays of DataSource,
    * and ISIDMapper
    */
   public BlockedDataSource( String aname, DataSource[] sources,
                     ISIDMapper[] ISIDmap )
   {
      super( aname, sources, null, ISIDmap );
      //rdb: deleted lots of code that is already executed by parent
   }

   /**
    * Construct a BlockedDataSource from parallel arrays of DataSource,
    * FieldIDMapper and ISIDMapper
    */
   public BlockedDataSource( String aname, DataSource[] sources,
         FieldIDMapper[] FIDmap, ISIDMapper[] ISIDmap )
   {
      super( aname, sources, FIDmap, ISIDmap );
      //rdb: deleted lots of code that is already executed by parent
   }

   /** Creates an array of ComponentInfo from the given information for
    *  each component. The given arrays should be parallel to the 
    *  returned array if provided (not null). The "ponentLocations"
    *  specifies the locations of the components in the composite.
    * 
    * The elements in the given
    *  DataSource array may have duplicates.
    * 
    * This version is specifically intended for BlockedDataSource,
    * but could some of it be moved to parent?
    */
  static ComponentInfo[] createComponentInfos( 
                              DataSource[] sources,
                              ISBounds[]   ponentLocations,
                              ISBounds     positeBounds )
  {
     int size = sources.length;
     
     ComponentInfo[] cpnInfos = new ComponentInfo[size];
     ISIDMapper[] ISIDmap = new ISIDMapper[size];
     
     if ( ponentLocations == null )
     {
        ponentLocations = new ISBounds[size];
        for ( int i = 0; i < size; i++ )
           ponentLocations[i] = sources[i].copyBounds();
     }
     if ( positeBounds == null )
     {
        positeBounds = ponentLocations[0].copy();
        for ( int i = 1; i < sources.length; i++ )
           positeBounds.setToBoundingVolume( ponentLocations[i] );
     }
     for ( int i = 0; i < size; i++ )
           ISIDmap[ i ] = new ISIDMapper( ponentLocations[ i ] ); // identity map

     for ( int i = 0; i < size; i++ )
        cpnInfos[i] = new ComponentInfo( sources[i], null, ISIDmap[ i ]);

     return cpnInfos;
  }

   /** Check the validity of the given ComponentInfo array */
   protected boolean check( ComponentInfo[] cinfo )
   {
      BlockManager manager = new BlockManager( cinfo );
      return manager.check( cinfo );
   }

   /** Initialize the component related information on construction */
   protected void initializeComponents( ComponentInfo[] compinfo )
   {
      if ( BlockManager.isRegular( compinfo ))
         this.manager = new RegularGridBlockManager( compinfo );
      else
         this.manager = new BlockManager( compinfo );

      this.recordDesc = manager.getRecordDescriptor();
      this.bounds = manager.getBounds();
   }

   /**
    * Return the BlockManager, from which you can get information for each
    * Component.
    */
   public BlockManager getBlockManager()
   {
      return this.manager;
   }

   /** Activate() method call its component DataSource to activate itself */
   public void activate()
   {
      int i;
      ComponentInfo[] components = this.manager.getComponentInfo();
      for ( i = 0; i < components.length; i++ )
      {
         if ( components[i].getComponent() instanceof DataSource )
         {
            DataSource c = (DataSource) components[i].getComponent();
            c.activate();
         }
      }
   }

   /**
    * Get a block of data of the DataSource. The DataBlock is passed by
    * reference
    */
   public void subblock( DataBlock blk, ISBounds blkisb, ISBounds dsisb,
         RecordSpec spec )
   {
      manager.subblock( blk, blkisb, dsisb, spec );
   }

   /**
    * Get a block of data of the DataSource, pass by reference before this
    * method is called, check to make sure the Dimensionality and dsisb is
    * inside the DataSource's bounds
    */
   public void subblock( DataBlock blk, ISBounds blkisb, ISBounds dsisb )
   {
      manager.subblock( blk, blkisb, dsisb );
   }

   /** Pass by reference subblock method taking FieldIDMapper */
   public void subblock( DataBlock blk, ISBounds blkisb, ISBounds dsisb,
         FieldIDMapper mp )
   {
      manager.subblock( blk, blkisb, dsisb, mp );

   }

   /** get the Datablock and fit in a Datablock which may have different size */
   public DataBlock subblock( ISBounds blk, ISBounds resolution )
   {
      ISIDMapper mapper;
      //DatumBlock ret;
      BasicBlock ret;
      ISIterator iter;
      if ( resolution.volume() == blk.volume())
      {
         return this.subblock( blk );
      }

      ISBounds isb1 = resolution.copy();
      isb1.moveToOrigin();
      /*************************************************************************
       * ****** deleted 11/23/04 ret = new DatumBlock(new
       * FloatDatum[isb1.volume()], isb1,this.recordDesc); /
       ************************************************************************/
      ret = new BasicBlock( isb1, this.recordDesc );
      iter = new ISIterator( isb1 );
      mapper = new ISIDMapper( blk, isb1 );

      if ( resolution.volume() > blk.volume())
      {

         /**********************************************************************
          * deleted 11/23/04 *********** DatumBlock data_blk =
          * (DatumBlock)this.subblock(blk); while (iter.valid()){
          * ret.setDatum(iter,data_blk.datum(mapper.toComponent(iter)));
          * iter.next(); } /
          *********************************************************************/
         BasicBlock data_blk = (BasicBlock) this.subblock( blk );
         while ( iter.valid())
         {
            Datum d = data_blk.datum( mapper.toComponent( iter ));
            //should be: ret.setFloat( iter, d.getFloats());
            ret.setDatum( iter, d );
            iter.next();
         }

         return ret;

      }
      if ( resolution.volume() < blk.volume())
      {
         /**********************************************************************
          * deleted 11/23/04 *********** while (iter.valid()){
          * ret.setDatum(iter,this.datum(mapper.toComponent(iter)));
          * iter.next(); } /
          *********************************************************************/
         while ( iter.valid())
         {
            Datum d = this.datum( mapper.toComponent( iter ));
            //should be: ret.setFloat( iter, d.getFloats());
            ret.setDatum( iter, d );
            iter.next();
         }
         return ret;
      }
      return null;
   }

   /**
    * Return by reference the Datum corresponding to the given IndexSpaceID.
    */
   public void datum( Datum dtm, IndexSpaceID isid )
   {
      manager.datum( dtm, isid );
      return;
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding FieldIDMapper.
    */
   public void datum( Datum dtm, IndexSpaceID isid, FieldIDMapper mp )
   {
      manager.datum( dtm, isid, mp );
      return;
   }

   /**
    * Get the Datum and fill its values in the passed Datum at positions
    * specified by the corresponding RecordSpec.
    */
   public void datum( Datum dtm, IndexSpaceID isid, RecordSpec recSpec )
   {
      manager.datum( dtm, isid, recSpec );
      return;
   }

   /**
    * Return a ARBlock, which is a AR representation of the requested area of
    * specified ISBounds
    */

   public ARBlock ARRep( ISBounds bounds )
   {

      return manager.ARRep( bounds );

   }

   public String toString()
   {

      StringBuffer result = new StringBuffer();
      result.append( super.toString());
      result.append( "\n" );

      result.append( "DataSource Type: Blocked DataSource.\n" );

      result.append( "Number of Components:" );
      result.append( " " + this.manager.size());
      result.append( "\n" );
      return new String( result );
   }

   static public void main( String[] args )
   {
      AttributeDescriptor[] attribs = new AttributeDescriptor[3];
      attribs[0] = new AttributeDescriptor( "float", "Carbon" );
      attribs[1] = new AttributeDescriptor( "float", "Oxygen" );
      attribs[2] = new AttributeDescriptor( "float", "Nitrogen" );
      RecordDescriptor recordDesc = new RecordDescriptor( attribs );

      IntegerIndexSpaceID lower1 = new IntegerIndexSpaceID( 0, 0 );
      IntegerIndexSpaceID upper1 = new IntegerIndexSpaceID( 1, 3 );
      IntegerIndexSpaceID lower2 = new IntegerIndexSpaceID( 3, 0 );
      IntegerIndexSpaceID upper2 = new IntegerIndexSpaceID( 4, 3 );

      ISBounds db1 = new ISBounds( lower1, upper1 );
      ISBounds extent1 = new ISBounds( lower1, upper1 );
      ISBounds extent2 = new ISBounds( lower2, upper2 );
      int orderingarray1[] = { 1, 0 };
      AxisOrdering ordering1 = new AxisOrdering( orderingarray1 );

      TextFileDataSource ds1 = new TextFileDataSource( "", recordDesc, db1,
            ordering1, "testdata/test-8-3.txt" );
      ds1.activate();

      AttributeDescriptor[] attribs2 = new AttributeDescriptor[2];
      attribs2[0] = new AttributeDescriptor( "float", "Carbon" );
      attribs2[1] = new AttributeDescriptor( "float", "Oxygen" );
      RecordDescriptor recordDesc2 = new RecordDescriptor( attribs2 ); // 2
                                                                       // attributes

      TextFileDataSource ds2 = new TextFileDataSource( "", recordDesc2, db1,
            ordering1, "testdata/test-8-2.txt" );
      ds2.activate();

      //FieldID Mapper
      int specarray[] = { 0, 2 };
      RecordSpec rspec1 = new RecordSpec( specarray );
      int fids[] = { 0, 1 };
      FieldIDMapper fidmapper1 = new FieldIDMapper( specarray, fids );

      //ISID Mapper
      ISIDMapper ismapper1 = new ISIDMapper( db1, extent1 );
      ISIDMapper ismapper2 = new ISIDMapper( db1, extent2 );
      ComponentInfo[] components = new ComponentInfo[2];
      components[0] = new ComponentInfo( ds1, fidmapper1, ismapper1 );//fidmapper1
      components[1] = new ComponentInfo( ds2, null, ismapper2 );

      BlockedDataSource bds1 = new BlockedDataSource( "", components );
      System.out.println( bds1.getBounds());

      // Test Datum methods
      System.out.println( "Test BlockedDataSource.datum(IndexSpaceID) method" );
      IntegerIndexSpaceID isid = new IntegerIndexSpaceID( 0, 0 );

      int specarray3[] = { 1 };
      RecordSpec rspec3 = new RecordSpec( specarray3 );
      int specarray2[] = { 2 };
      RecordSpec rspec2 = new RecordSpec( specarray2 );

      Datum d1, d2;
      d1 = bds1.datum( isid );
      d2 = ds1.datum( new IntegerIndexSpaceID( 0, 0 ), rspec1 );
      System.out.println( isid + " :" );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      isid = new IntegerIndexSpaceID( 1, 2 );
      d1 = bds1.datum( isid );
      d2 = ds1.datum( new IntegerIndexSpaceID( 1, 2 ), rspec1 );
      System.out.println( isid + " :" );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      isid = new IntegerIndexSpaceID( 2, 3 );
      d1 = bds1.datum( isid );
      d2 = new FloatDatum( 2 );
      ((FloatDatum) d2 ).setNull( 0 );
      ((FloatDatum) d2 ).setNull( 1 );
      System.out.println( isid + " :" );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      isid = new IntegerIndexSpaceID( 3, 3 );
      d1 = bds1.datum( isid );
      d2 = ds2.datum( new IntegerIndexSpaceID( 0, 3 ));
      System.out.println( isid + " :" );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      isid = new IntegerIndexSpaceID( 4, 0 );
      d1 = bds1.datum( isid );
      d2 = ds2.datum( new IntegerIndexSpaceID( 1, 0 ));
      System.out.println( isid + " :" );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      System.out
            .println( "\nTest BlockedDataSource.datum(RecordSpec,IndexSpaceID) method" );
      isid = new IntegerIndexSpaceID( 4, 0 );
      d1 = bds1.datum( isid, rspec3 );
      d2 = ds2.datum( new IntegerIndexSpaceID( 1, 0 ), rspec3 );
      System.out.println( isid + " :" );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      isid = new IntegerIndexSpaceID( 1, 2 );
      d1 = bds1.datum( isid, rspec3 );
      d2 = ds1.datum( new IntegerIndexSpaceID( 1, 2 ), rspec2 );
      System.out.println( isid + " :" );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      //test subblock methods
      System.out.println( "\nTest BlockedDataSource.subblock(ISBounds) method" );
      ISBounds isb_test = new ISBounds( new IntegerIndexSpaceID( 1, 1 ),
            new IntegerIndexSpaceID( 3, 2 ));
      DataBlock block_test = bds1.subblock( isb_test );
      d1 = block_test.datum( new IntegerIndexSpaceID( 0, 1 ));
      d2 = ds1.datum( new IntegerIndexSpaceID( 1, 2 ), rspec1 );

      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );
      System.out.println( "" );
      d1 = block_test.datum( new IntegerIndexSpaceID( 2, 0 ));
      d2 = ds2.datum( new IntegerIndexSpaceID( 0, 1 ));
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      System.out
            .println( "\nTest BlockedDataSource.subblock(ISBounds,ResordSpec) method" );
      block_test = bds1.subblock( isb_test, rspec3 );
      d1 = block_test.datum( new IntegerIndexSpaceID( 0, 0 ));
      d2 = ds1.datum( new IntegerIndexSpaceID( 1, 1 ), rspec2 );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );
      System.out.println( "" );
      d1 = block_test.datum( new IntegerIndexSpaceID( 2, 1 ));
      d2 = ds2.datum( new IntegerIndexSpaceID( 0, 2 ), rspec3 );
      System.out.println( "Result Values:\t " + d1 );
      System.out.println( "Correct Values:\t " + d2 );

      /*
       * //test subblock(RecordSpec,ISBounds) RecordSpec spec = new
       * RecordSpec(); spec.setIndexedFieldNumber(0, 0); IntegerIndexSpaceID
       * blower = new IntegerIndexSpaceID(0,0); IntegerIndexSpaceID bupper = new
       * IntegerIndexSpaceID(4,3); ISBounds isb =new ISBounds(blower,bupper); //
       * bds1.getBounds() DatumBlock blk = (DatumBlock) bds1.subblock(spec,isb);
       * isb.moveToOrigin(); ISIterator iter = new ISIterator(isb);
       * 
       * for(;iter.valid();iter.next()){ System.out.print(iter); d =
       * blk.datum(iter); if (d == null) System.out.println("Null Datum"); else
       * System.out.println(d);
       *  }
       */
      // test datum(isid,spec)
      /*
       * RecordSpec spec = new RecordSpec(); spec.setIndexedFieldID(0, new
       * FieldID(0)); iter.init(); System.out.print(iter);
       * System.out.println(bds1.datum(iter,spec)); while
       * (iter.hasMoreElements()) { iter.next(); System.out.print(iter);
       * d=bds1.datum(iter,spec); System.out.println(d);
       *  }
       */

      System.out.println( "The End" );

   } //end main

}

