//
//  Cache.java
//  CacheTest
//
//  Created by Philip Rhodes on Sun Nov 16 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//
// ----------------- Edit history as of 01/06/05 -------------------
// 01/06/05 rdb: create Datum objects from RecordDescriptor
// -----------------------------------------------------------------
//

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

/** 
 * The Cache class is the parent class for all Caches. A cache is used to 
 * store data in memory read from an indexable object, presumably a  
 * datasource. 
*/

abstract class Cache {


    // This is the object that the cache will read from.
    //protected Indexable theSource;
    protected PhysicalDataSource theSource;
    protected int dimensionality;
    protected int numAttributes;
    
    protected ISBounds iterationSpaceBounds;
    
    protected int[] lowerIterationSpaceCoords;
    protected int[] upperIterationSpaceCoords;
    
    protected int [] iterationSpaceDimensions;
    
    
    protected ISBounds dsSpaceBounds;
    
    protected int[] lowerDSSpaceCoords;
    protected int[] upperDSSpaceCoords;
    
    protected int [] dsSpaceDimensions;
    
        
    /** This constructor stores the indexable object that is being cached.*/
//    Cache(Indexable ds){
    Cache(PhysicalDataSource ds)
    {
        this.iterationSpaceBounds = ds.copyBounds(); 
        this.theSource = ds;
        this.dimensionality = ds.dim();
        this.numAttributes = ds.getNumAttributes();
        this.lowerIterationSpaceCoords = this.iterationSpaceBounds.copyLowerArray(new int[this.dimensionality]);
        this.upperIterationSpaceCoords = this.iterationSpaceBounds.copyUpperArray(new int[this.dimensionality]);
        
        this.iterationSpaceDimensions = this.iterationSpaceBounds.getDimensions(new int[this.dimensionality]);
        
        
        this.dsSpaceBounds = ds.getBounds();
        
        this.lowerDSSpaceCoords = this.dsSpaceBounds.copyLowerArray(new int[this.dimensionality]);
        this.upperDSSpaceCoords = this.dsSpaceBounds.copyUpperArray(new int[this.dimensionality]);
        
        this.dsSpaceDimensions = this.dsSpaceBounds.getDimensions(new int[this.dimensionality]);
        
        
    }

    /** This constructor stores the indexable object that is being cached.*/
    Cache(Indexable ds, ISBounds iterationSpaceBounds)
    {
    
        this.dsSpaceBounds = ds.getBounds();
        
        this.iterationSpaceBounds = iterationSpaceBounds.copy();
        
        this.iterationSpaceBounds.intersectWith(this.dsSpaceBounds);
        
        if(!this.iterationSpaceBounds.equals(iterationSpaceBounds)){
        
            new SDBError(SDBError.Warning,"Cache constructor: iteration space was clipped to datasource bounds.");
        }
        
        this.theSource = (PhysicalDataSource) ds;
        this.dimensionality = ds.dim();
        this.numAttributes = ds.getNumAttributes();
        
        this.lowerIterationSpaceCoords = this.iterationSpaceBounds.copyLowerArray(new int[this.dimensionality]);
        this.upperIterationSpaceCoords = this.iterationSpaceBounds.copyUpperArray(new int[this.dimensionality]);

        this.iterationSpaceDimensions = this.iterationSpaceBounds.getDimensions(new int[this.dimensionality]);
        
        
        
        this.lowerDSSpaceCoords = this.dsSpaceBounds.copyLowerArray(new int[this.dimensionality]);
        this.upperDSSpaceCoords = this.dsSpaceBounds.copyUpperArray(new int[this.dimensionality]);
        
        this.dsSpaceDimensions = this.dsSpaceBounds.getDimensions(new int[this.dimensionality]);
        
        
        
    }

    /** Return a reference to the object being cached. */
    Indexable getSource()
    {
        return this.theSource;
    }
    
    /** Return the dimensionality of this cache.*/
    final int dim(){
    
        return this.dimensionality;
    }


    /** Create a new datum and fill it with the values found at the 
     *  given location.
     */
    Datum lookup(IndexSpaceID id)
    {
        Datum newDatum= this.theSource.getRecordDescriptor().createDatum();
        this.lookup(newDatum, id);
        return newDatum;
    }
    
    abstract void lookup(Datum theDatum,IndexSpaceID id);
    
    //abstract void lookup(BasicBlock queryBlock, ISBounds queryBounds);
    
    abstract void lookup(BasicBlock queryBlock, ISBounds dstBounds, ISBounds queryBounds);
    
    
  /** Retrieves the requested datum from the Cache, or directly from the datasource, refreshing
   * the cache with a new block. This version provides scratch
   *  space for threaded operation, and exploits coherence where possible to reduce
   *  overhead.
   */
   void lookup(Datum theDatum,IndexSpaceID id, CacheContext context ){
   
      new SDBError(SDBError.Fatal, "Not implemented");
   }


}
