//
//  ChunkedDataSource.java
//        rdb notes 01/11/12: 
//          1. This code does not support attribute order datasources.
//          2. Seems better(conceptually) for chunking to be an
//             attribute of any PhysicalDataSource.
//          3. OR implement like CacheDataSource. I.e., ChunkedDataSource
//             is wrapped around any PhysicalDataSource???
//
//  Created by Xuan Tang  on Sun Feb 1 2004.
//  Copyright(c) 2004 UNH. All rights reserved.
//
//----------------- Edit history as of 12/14/04 -------------------
// 03/04/12 rdb: Added support for double, int and short (in loadChunk method)
//               BUT I HAVE NOT TESTED IT.
// 01/31/12 rdb: Made new constructor accepting a FDLDescriptor; created a
//                 new private method so constructors could share code.
//               Removed extraneous instance variable, spaceBounds and other
//                 instance variables that override inherited variables:
//                   dimensionality, axisWeights
//               Deleted some lines of code that re-execute code done by 
//                 super calls.
// 07/10/06 dlipsa: Fix noGapSubblock and datum to work for bytes(beside floats)
// 12/04/05 rdb: Refactor to move all public subblock methods to parent,
//                 implement here subBasicBlock and subCompositeBlock, though
//                 the latter has not been implemented!
//               Deleted(by comments) subblock( bounds ) and subPartitionedBlock
//                 methods. subPartitionedBlock is identical code to what
//                 is in noGapSubset(which I also renamed from subBasicBlock).
// 10/21/05 rdb: Deleted unused variables to avoid compiler warnings
// 04/26/05 rdb: Fixed problems in psubblock
// 04/23/05 rdb: Added code to test return from allocateBuffer
// 02/19/05 rdb: Removed endian from here -- now located in PhysicalDS
//               Also byteBuffer which gets replaced with getBuffer()
//               Deleted unused variables: attributeByteSize,
//               Deleted variables defined in parent: numAttributes, filename
//
// 02/06/05 rdb: Replaced all allocateDirect method calls to new allocateBuffer
//                  deleted pallocateBuffer; same thing done in allocateBuffer
//               Added new allocBuffer method to parent, so none needed here.
// 01/19/05 rdb: Converted getFloatAttributes calls to getFloats.
//          rdb: There's still lots of code here that is too type specific.
// 01/06/05 rdb: create Datum objects from RecordDescriptor
//-----------------------------------------------------------------

package edu.unh.sdb.datasource;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import edu.unh.sdb.common.*;

/** This class performs chunking I/O operations. The data file should have been 
 *   reorganized according to the spacial chunking.
 *   We assume the disk block size is 4K(4086 bytes )
 *   We also assume that the data file can be equally divided into a number of blocks.
 *   And the size of each block is a multiple of the disk block size
 *
 * NOTE: This class only works for BasicBlock so far.
 */

// rdb note 03/04/12: If the parent were BinaryFileDataSource, the parent would
//    define the physical file.
public class ChunkedDataSource extends PhysicalDataSource
{
   protected ISBounds partitioningBounds;
   protected RegularISPartitioning chunkingPartitioning;
   
   private RandomAccessFile raFile;
   private FileChannel fChannel;
   
   private int chunkByteSize;
   private int fieldByteSize;
   private ISBounds chunkBounds; // refers to the shape of each chunk in Grid Space
   private int chunkVolume; // Added by PJR
   private int numChunkRods; // Added by PJR
   
   private BasicBlock cacheBlock;
   private ISBounds cacheBounds;
   private ISIDMapper cacheIndexMapper;
   
   private int[] steps, chunkSteps;
   private int leastSignificantAxis;
   
   private int chunkRodSize, chunkRodLength;
   private ISIterator iter;
   //private ISIterator chunkIter;
   private long[] partSpaceAxisWeights;
   
   //-------------------------- constructors -----------------------------
   /**
    * Create a ChunkedDataSource from an fdl; 
    *   In this case, the partitioning specifications are determined by
    * the size of each partition, not the size of the partitioning space.
    * I.e., the user specifies that each partition will be defined, so 
    * we have to compute the bounds of the partitioning space (how many
    * partitions are defined along each dimension).
    * 
    * @param name
    * @param fdl 
    */
   public ChunkedDataSource( String name, FDLDescriptor fdl )
   {
      super( name, fdl );
      
      IndexSpaceID chunkDims = fdl.getChunkDimensions();
      this.chunkingPartitioning = new RegularISPartitioning( chunkDims, 
                          bounds, this.getAxisOrdering() );
      initializeCommon( this.chunkingPartitioning.getPartitioningBounds());     
   }
   
   /**
    * Creates a ChunkedDataSource object from 'fileName',
    * how many chunks are on each direction are specified by 'partitioningBounds',
    * the dimensions and placement of the volume given by 'spaceBounds',
    * axis ordering 'order' and records described by 'recordDesc'.
    * It seems that 'order' specifies axis ordering for both an individual
    * chunk and for all chunks.
    * 
    * This constructor must compute the "nominal" size of each partition; 
    * partitions on the borders may be smaller, but this feature is not
    * completely supported right now.
    * 
    * 'datum' method caches one chunk at a time.
    */
   public ChunkedDataSource( String name,
         ISBounds partitioningBounds,
         ISBounds spaceBounds,
         String fileName,
         AxisOrdering order,
         RecordDescriptor recordDesc )
   {
      //super(name, recordDesc, partitionBounds, order,fileName, ByteOrder.BIG_ENDIAN); // temp implementation, have to be trimed
      super( name, recordDesc, spaceBounds, order, fileName ); // temp implementation, have to be trimed

      this.chunkingPartitioning = new RegularISPartitioning( partitioningBounds, bounds );
      
      //this.chunkBounds =this.chunkingPartitioning. getPartitioningBounds();
      this.chunkBounds = this.chunkingPartitioning.getGeometricBounds();
      
      //System.out.println(" chunkSpaceSize " + this.ChunkBounds) ;
      //System.out.println(" ********************************************** " ) ;
      initializeCommon( partitioningBounds ); // initialize rest of fields.
      //this.seekCost = -1 ;
   }
   //--------------- initialize ----------------------------------
   private void initializeCommon( ISBounds partitioningBounds )
   {
      this.partSpaceAxisWeights = this.axisOrdering.makeAxisWeights( partitioningBounds );
      
      this.partitioningBounds = partitioningBounds;
      
      this.chunkVolume = this.chunkBounds.volume();
      //rdb this.endian = ByteOrder.BIG_ENDIAN; // should be set by fdl file
      //rdb this.fieldByteSize = 4; // assume the field byteSize is 4( one float attribute)
      this.fieldByteSize = this.recordDesc.byteSize();
            
      this.chunkByteSize = chunkBounds.volume() * fieldByteSize;
      
      this.axisWeights = this.axisOrdering.makeAxisWeights( this.partitioningBounds );
      //rdb: done by super this.recordDesc = recordDesc;
      this.cacheBlock = new BasicBlock( chunkBounds, this.recordDesc, this.axisOrdering );
      
      //------------ set the step -----------------
      //int i,d=this.dim();
      IndexSpaceID dimensions = partitioningBounds.getDimensions();
      int dimMinus1 = dimensionality - 1;
      
      this.steps = new int[ dimensionality ];
      this.leastSignificantAxis = this.axisOrdering.getLeastSignificantAxis();
      
      this.steps[ this.axisOrdering.getAxis( dimMinus1 )]
            = dimensions.getCoord( this.axisOrdering.getAxis( dimMinus1 ));
      
      for ( int i = 0; i < dimMinus1; i++ )
      {
         this.steps[ this.axisOrdering.getAxis( i ) ] = 1;
      }
      IndexSpaceID chunkDimensions = chunkBounds.getDimensions();
      
      this.chunkSteps = new int[ dimensionality ];
      this.chunkSteps[ this.axisOrdering.getAxis( dimMinus1 ) ]
            = chunkDimensions.getCoord( this.axisOrdering.getAxis( dimMinus1 ));
      
      for ( int i = 0; i < dimMinus1; i++ )
      {
         this.chunkSteps[ this.axisOrdering.getAxis( i ) ] = 1;
      }
      
      // size of the rod of a chunk measured by number of datums
      chunkRodSize = chunkBounds.getDimensions().getCoord( this.leastSignificantAxis );
      
      // length of the chunk rod measured by number of primitive elements
      chunkRodLength = chunkRodSize * this.numAttributes;
      
      //chunkIter = new ISIterator(ChunkBounds,this.ordering,this.chunkSteps);
      this.numChunkRods = this.chunkVolume / chunkRodSize;
      
      iter = new ISIterator( partitioningBounds, this.axisOrdering, this.steps );
      
      //this.gapSize =-1;
      
      //recompute the axisWeights according to the bounds in partitioned space
      this.axisWeights = this.axisOrdering.makeAxisWeights( partitioningBounds );      
   }
   
   /** return the ISBounds of this chunkedDataSource in partitioned Space
    */
   public ISBounds getPartitionSpaceBounds()
   {
      return this.partitioningBounds;
   }
   
   /** Open the binary data file for this datasource. */
   public void activate()
   {
      super.activate();
      try
      {
         this.raFile = new RandomAccessFile( this.getURL(), "r" );
         
         this.fChannel = raFile.getChannel();
      }
      catch( IOException e )
      {
         new FileError( SDBError.Recoverable, "ChunkedDataSource:: " +
               "Error opening RandomAccessFile - " + getDataFileString() + " :: " + e );
      }
   }
   
   /** The DataSource will release any resources used during the
    *	readling or processing of data.
    */
   public void deactivate()
   {
      super.deactivate();
      try
      {
         this.fChannel.close();
         this.raFile.close();
      }
      catch( IOException e )
      {
         new FileError( SDBError.Recoverable, "Error closing RandomAccessFile " + e );
      }
   }
   
   /** set the byte order for the given buffer */
   public void setOrder( ByteBuffer buffer )
   {
      if (( this.endian ).equals( ByteOrder.BIG_ENDIAN ))
         buffer.order( ByteOrder.BIG_ENDIAN );
      else
         buffer.order( ByteOrder.LITTLE_ENDIAN );
   }
   
   /** get the byte size of each chunk */
   public int getChunkByteSize()
   {
      return this.chunkByteSize;
   }
   
   /** 
    * get the dimensions of each chunk in grid space
    */
   public IndexSpaceID getChunkDimensions()
   {
      return this.chunkBounds.getDimensions();
   }
   
   /** 
    * Return an ISBounds indicating the dimensions of a chunk. 
    */ 
   public ISBounds getChunkBounds(){
   
      return this.chunkBounds.copy();
   }
   
   /** 
    * load the data from the data source into the argument datum
    */
   public void datum( Datum datum, IndexSpaceID dataId )
   {
      IndexSpaceID partitionId = chunkingPartitioning.map( dataId );
      
      ISBounds spaceBounds = chunkingPartitioning
            .getContainingPartitionBounds( dataId );
      
      IndexSpaceID cacheindexID;
      if ( inCache( dataId ))
      {
         cacheindexID = cacheIndexMapper.toComponent( dataId );
         cacheBlock.datum( datum, cacheindexID );
      }
      else
      {
         this.cacheBounds = spaceBounds;
         //int chunkoffset = this.computeOffset( partitionID);
         long chunkoffset = AxisOrdering.computeOffset( partitionId, this.axisWeights );
         
         Values data = cacheBlock.getValues();
         this.loadChunk( chunkoffset, data );
         
         //for ( int i = 0; i < data.length; i++ )
         //System.out.println( data[i] );
         //System.out.println("chunk space Size " + ChunkBounds );
         //System.out.println( "CacheBlock is :" + cacheBlock );
         cacheIndexMapper = new ISIDMapper( chunkBounds, spaceBounds );
         //System.out.println( "id Mapper " + cacheIndexMapper );
         cacheindexID = cacheIndexMapper.toComponent( dataId );
         //System.out.println("in cache Block index " + cacheindexID);
         
         cacheBlock.datum( datum, cacheindexID );
      }
   }
   
   private boolean inCache( IndexSpaceID id )
   {
      if (( cacheBounds != null ) &&( cacheBounds.contains( id )) )
         return true;
      else
         return false;
   }
   
   /********** deleted 12/04/05 rdb ** code duplicates what is in noGapSubBlock
    *
    * /** this method uses the concept of chunked rod, an array of chunks that are
    * continuous in the storage order. It reads a rod of data chunks in a single
    * read using a one-dimensional buffer that is big enough to hold the entire
    * chunked rod.  By reading several chunks at a time, we expect a better
    * I/O performance. Of course, there is an optimized maximam buffer size,
    * which we have to determine in the future research. The argument parameter
    * is given in partitioned space, it return a dataBlock in the grid space.
    *
    * private BasicBlock subPartitionedBlock( ISBounds partSpaceQueryBnds )
    * {
    * // Instead of making a new iterator, can we assign to an existing one, stored
    * // as a data member?
    *
    * //ISIterator iter=new ISIterator(partSpaceQueryBnds,this.ordering,this.steps);
    * iter.setBounds( partSpaceQueryBnds );
    *
    * long chunkOffset;
    *
    * int partRodSize = partSpaceQueryBnds.getDimension( this.leastSignificantAxis );
    *
    * if ( ! allocateBuffer( partRodSize * this.chunkByteSize ))
    * partRodSize = getBufferSize() / this.chunkByteSize;
    *
    * FloatBuffer floatBuffer = getBuffer().asFloatBuffer();
    * long fileOffset = 0;
    * int dstOffset = 0;
    *
    * // Let's precompute ChunkBounds.volume()
    * int bufferSize = chunkVolume * this.numAttributes * partRodSize;
    * //int bufferSize = ChunkBounds.volume() * this.numAttributes * partRodSize;
    *
    * float[] datablockArray = new float[
    * partSpaceQueryBnds.volume() * chunkBounds.volume() * this.numAttributes];
    * int n = 0;
    *
    * try
    * {
    * for ( iter.init(); iter.valid(); iter.next())
    * {
    * chunkOffset = AxisOrdering.computeOffset( iter, this.axisWeights );
    *
    * fileOffset = chunkOffset * this.chunkByteSize;
    * if ( fileOffset !=((int) raFile.getFilePointer()) )
    * raFile.seek( fileOffset );
    *
    * getBuffer().clear();
    * fChannel.read( getBuffer() );
    *
    * floatBuffer.clear();
    * //floatBuffer.get(floatData,0,bufferSize);
    * //System.out.println("buffer size :" + bufferSize );
    *
    * //dstOffset = n * bufferSize;
    * for ( int i = 0; i < partRodSize; i++ )
    * {
    * dstOffset = n * bufferSize + chunkRodLength * i;
    * //for ( chunkIter.init(); chunkIter.valid(); chunkIter.next())
    * for ( int j = 0; j < this.numChunkRods; j++ )
    * {
    * //System.out.println("scrOffset : " + srcOffset);
    * // System.out.println("scrOffset : " + floatBuffer.position());
    *
    * //System.out.println("dstOffset : " + dstOffset);
    * //System.out.println(" _______________________ ");
    * //System.arraycopy(floatData,srcOffset,datablockArray,dstOffset, chunkRodSize);
    * floatBuffer.get( datablockArray, dstOffset, chunkRodLength );
    * //srcOffset += chunkRodSize;
    * dstOffset += chunkRodLength * partRodSize;
    * }
    * //System.out.println("******** ");
    * }
    * n++ ;
    * }
    * } catch( IOException e )
    * {
    * System.out.println( e );
    * }
    *
    * ISBounds gridSrcBounds = this.chunkingPartitioning.mapToSpace( partSpaceQueryBnds );
    * return new BasicBlock( gridSrcBounds, this.recordDesc,
    * new FloatValues( datablockArray ));
    * }
    *
    * public DataBlock subblock( ISBounds spaceSrcBounds )
    * {
    * return subPartitionedBlock( this.chunkingPartitioning.map( spaceSrcBounds ));
    * }
    *
    * /************* end delete *****************************/
   
   /** This subblock method will check if gap between two neighboring rods
    *(in data storage order) is smaller than the argument seekcost. The seek
    * cost is messured by the number of index spaces in the partitioned space.
    * If the gap is smaller than the seekcost, a long buffer is allocated
    * and one o/s read operation  is used to continuously read the two rods
    * including the gap bwteen the two rods.
    *
    * This means that given a slice query, if the gaps between neighboring rods
    * are smaller than the seekCost, a long buffer is allocated, and both rods
    * and gaps between rods are read in the buffer using a single O/S read
    * operation.
    *
    * Then only the rod data is copied from the buffer into the dst datablock.
    * The gap data is disregarded.
    *
    *  This method also handles n-d queries in the same fashion.
    * The goal is to improve the I/O performance by reducing the number of O/S reads.
    */
   void gappedSubBlock( BasicBlock dstBlock, ISBounds dstBounds,
         ISBounds   srcBounds )
   {
      ISBounds partSpaceQueryBnds = this.chunkingPartitioning.map( srcBounds );
      //rdb this.pallocateBuffer(partSpaceQueryBnds);
      if ( !allocateBuffer( this.partitioningBounds, partSpaceQueryBnds, this.chunkByteSize ))
      {
         System.out.println( "ChunkedDataSource did not get big enough buffer.");
      }
      
      //        System.out.println("grid queryBounds :" + srcBounds);
      //        System.out.println("part queryBounds :" + partSpaceQueryBnds);
      //        System.out.println("bufferDimension: " + this.bufferDimension );
      
      //this.iter.setBounds( partSpaceQueryBnds );
      
      float[] datablockArray =((FloatValues) dstBlock.getValues()).getDataReference();
      long fileOffset;
      
      int partRodSize = partSpaceQueryBnds.getDimension( this.leastSignificantAxis );
      
      int partRodElemSize = chunkVolume * this.numAttributes * partRodSize;
      // number of primitve element in a partition rod
      FloatBuffer floatBuffer = getBuffer().asFloatBuffer();
      
      //rdb: this really should be a bounds iterator not an ISIterator with steps
      ISIterator bufferIter = new ISIterator( partSpaceQueryBnds,
            this.axisOrdering, this.bufferSteps );
      //ISBoundsIterator bufferIter = new ISBoundsIterator(
      // partSpaceQueryBnds, this.bufferBounds, this.ordering );
      
      
      //ISBounds   bufBounds = new ISBounds( this.bufferBounds );
      //ISIterator inBufIter = new ISIterator( bufBounds, this.ordering, this.steps );
      ISIterator inBufIter = new ISIterator( this.bufferBounds, this.axisOrdering, this.steps );
      
      int dstOffset;
      long rodFileOffset;
      int bufferPosition;
      int n = 0;
      try
      {
         // Iterate over the partition space one buffer block at a time.
         // Each buffer block could have 1 or more chunks in it
         for ( bufferIter.init(); bufferIter.valid(); bufferIter.next())
         {
            fileOffset = this.chunkByteSize
                  * AxisOrdering.computeOffset( bufferIter, partSpaceAxisWeights );
            
            if ( fileOffset !=(int) raFile.getFilePointer() )
               raFile.seek( fileOffset );
            
            getBuffer().clear();
            int bytesRead = fChannel.read( getBuffer() );
            if ( bytesRead <= 0 ) // rdb added to avoid compiler warning, but should check
               new SDBError( SDBError.Warning, "ChunkedDataSource.psubblock: "
                     + "no bytes read from channel???" );
            
            // At this point 1 or more file-contiguous chunks(chunk rods) have been
            // read from the file into the I/O buffer. We now have to "un-chunk" the
            // data into its correct positions in the data array for the destination block.
            
            // "Move" the inBufIter so that its coordinates correspond to the position
            // of the buffer in the 'bufferIter' iteration.
            inBufIter.moveTo( bufferIter );
            
            // Iterate over the chunks in the buffer block
            for ( inBufIter.init(); inBufIter.valid(); inBufIter.next())
            {
               rodFileOffset = AxisOrdering.computeOffset( inBufIter, partSpaceAxisWeights )
               * this.chunkByteSize;
               bufferPosition =(int)( rodFileOffset - fileOffset );
               //                    System.out.println(" buffer position : " + bufferPosition );
               //                    System.out.println(" buffer capacity : " + byteBuffer.capacity());
               getBuffer().position( bufferPosition );
               
               //rdb: I bet the next line is not needed. Setting the underlying
               //     position should also be setting the floatBuffer position.
               floatBuffer.position( bufferPosition / 4 ); // this only works for float implementation
               for ( int i = 0; i < partRodSize; i++ )
               {
                  dstOffset = n * partRodElemSize + chunkRodLength * i;
                  // for ( chunkIter.init(); chunkIter.valid(); chunkIter.next())
                  for ( int j = 0; j < this.numChunkRods; j++ )
                  {
                     //   System.out.println("scrOffset : " + floatBuffer.position());
                     //    System.out.println("dstOffset : " + dstOffset);
                     //    System.out.println(" _______________________ ");
                     //    System.out.println("floatBufferLimit:" + floatBuffer.limit());
                     floatBuffer.get( datablockArray, dstOffset, this.chunkRodLength );
                     dstOffset += this.chunkRodLength * partRodSize;
                  }
               }
               n++ ;
            }
         }// end of buffer iter
      }
      catch( IOException e )
      {
         new FileError( SDBError.Recoverable, "Exception in " +
               "ChunkedDataSource.subblock(DataBlock, ISBounds, ISBounds)::"
               + e );
      }
   }
   
   /** Return the Datum corresponding to the given IndexSpaceID. The
    * Datum will only contain attributes specified by the RecordSpec
    * argument.
    */
   public void datum( Datum datum, IndexSpaceID did, FieldIDMapper fmap )
   {
      new IllegalArgumentError( SDBError.Fatal,
            "ChunkedDataSource.datum( Datum, ISID, FIM) not implemented " );
   }
   
   /**
    * Given a chunk index, read the data from the chunk into a Values object
    */
   protected void loadChunk( long chunkIndex, Values data )
   {
      long fileOffset = this.chunkByteSize * chunkIndex;
      
      allocateBuffer( this.chunkByteSize );
      ByteBuffer byteBuffer = getBuffer();
      int bytesInBuffer = byteBuffer.limit();
      
      int bytesToRead = this.chunkByteSize;
      if (bytesToRead > data.size() * StorageType.byteSize(data.getStorageType()))
      {
         bytesToRead = data.size() * StorageType.byteSize(data.getStorageType());
         if ( bytesToRead < bytesInBuffer )
            bytesInBuffer = bytesToRead;
         new SDBError( SDBError.Warning,
               "loadChunk warning: array smaller than chunk size." );
      }
      else if (bytesToRead < data.size()* StorageType.byteSize(data.getStorageType()))
      {
         new SDBError( SDBError.Recoverable,
               "loadChunk error: array larger than chunk size." );
      }
      
      try
      {
         if ( fileOffset !=((int) raFile.getFilePointer()) )
            raFile.seek( fileOffset );
         int numIndex = 0;
         while( bytesToRead > 0 )   //rdb added while in case io/buffer too small
         {
            byteBuffer.clear();
            int bytesRead = fChannel.read( getBuffer() );
            if ( bytesRead <= 0 ) // rdb added to avoid compiler warning, but should check
               new SDBError( SDBError.Warning, "ChunkedDataSource.loadChunk: "
                     + "no bytes read from channel???" );
            
            byteBuffer.flip();
            if ( bytesInBuffer > bytesToRead ) // for last read, might not be full
               bytesInBuffer = bytesToRead;
            switch(data.getStorageType())
            {
               case StorageType.Byte:
                  byteBuffer.get(data.getBytes(), numIndex, bytesRead);
                  break;
               case StorageType.Float:
                  byteBuffer.asFloatBuffer().get(data.getFloats(), numIndex, bytesRead/4 );
                  // rdb: 03/04/12 : bytesRead was not divided by 4. I don't think
                  //      it ever worked with that. FloatBuffer arg should be #
                  //      floats read, not # bytes read.
                  //byteBuffer.asFloatBuffer().get(data.getFloats(), numIndex, bytesRead );
                  break;
               case StorageType.Double:
                  byteBuffer.asDoubleBuffer().get( data.getDoubles(), numIndex, bytesRead/8 );
                  break;
               case StorageType.Int:
                  byteBuffer.asIntBuffer().get( data.getInts(), numIndex, bytesRead/4 );
                  break;
               case StorageType.Short:
                  byteBuffer.asShortBuffer().get( data.getShorts(), numIndex, bytesRead/2 );
                  break;
               default:
                  new SDBError(SDBError.Fatal, "Type not implemented");
            }
            
            bytesToRead =- bytesInBuffer;
            numIndex += bytesInBuffer;
         }
      }
      catch( IOException e )
      {
         new SDBError(SDBError.Recoverable, e.toString());
      }
   }
   
   /** given a chunk index, return a float array of the data in that chunk.
    */
   /* =======================================
    protected float[] loadChunk( long chunkIndex)
   {
      int floatBufSize = this.chunkBounds.volume() * this.numAttributes;
      float[] floatData = new float[ floatBufSize ];
    
      loadChunk( chunkIndex, floatData );
      return floatData;
   }*/
   
   /** return number of chunks contained by the datasource */
   public int numOfChunks()
   {
      //return this.srcGridSpaceBounds.volume()/ this.partSpaceBounds.volume() ;
      return this.chunkingPartitioning.getPartitioningBounds().volume();
   }
   
   // *******************************    Main Test **********************************************
   public static void main( String[] args )
   {
      /*
       int[] bounds = { 100,100};
       int[] partSpaceBounds = { 1024,1024};
       int[] orderarray = {0,1};
       
       DataSource ds = DataSource.createPhysicalDataSource("/local/tmp/xtang/trav/testdata/file.24.1.bin.xml", "foo");
       ds.activate();
       RecordDescriptor rec = ds.getRecordDescriptor();
       
       AxisOrdering order = new AxisOrdering(orderarray);
       
       //ChunkingDataSource chunkingsource = new ChunkingDataSource(4, new ISBounds(bounds), "data.bin",
       //                                                            4096);
       ChunkedDataSource chunkingsource = new ChunkedDataSource("foo",
       new ISBounds(partSpaceBounds), ds.getBounds(),
       "/local/tmp/xtang/trav/testdata/file.24.1.bin",order,rec );
       float[] data;
       chunkingsource.activate();
       
       long startTime, stopTime;
       
       long startTime = System.currentTimeMillis();
       
       for ( int i =0; i< chunkingsource.numOfChunks(); i++)
       data = chunkingsource.loadChunk(i);
       long stopTime = System.currentTimeMillis();
       
       System.out.println("Subblock traverse Time is " +(stopTime - startTime) / 1000.0 + " seconds.");
       
       System.out.println( "number of chunks " + chunkingsource.numOfChunks());
       */
      
      /*       //--------------------------- other tests ------------------------
       int[] pBounds = {10,10};
       RegularISPartitioning partition = new RegularISPartitioning( pBounds, new ISBounds(bounds));
       System.out.println("_________________________________________");
       System.out.println("partition is :" + partition );
       int[] testid = { 10,0};
       
       int[] lower = { 9,9};
       int[] upper = {11,11};
       
       int[] lower1 = { 1,1};
       int[] upper1 = {2,2};
       
       
       
       ISBounds tbounds = new ISBounds( new IntegerIndexSpaceID(lower), new IntegerIndexSpaceID(upper));
       ISBounds tbounds1 = new ISBounds( new IntegerIndexSpaceID(lower1), new IntegerIndexSpaceID(upper1));
       
       IntegerIndexSpaceID testIndex = new IntegerIndexSpaceID( testid );
       System.out.println(partition.map(testIndex));
       System.out.println(partition.map(tbounds));
       System.out.println(partition.mapToSpace(tbounds1));
       
       
       FloatDatum fdatum = new FloatDatum(1 );
       //ISIterator iter = new ISIterator(ds.getBounds());
       ISIterator iter = new ISIterator( tbounds );
       
       startTime = System.currentTimeMillis();
       for (iter.init(); iter.valid(); iter.next())
       {
       chunkingsource.datum( fdatum, iter);
       System.out.println(fdatum);
       }
       stopTime = System.currentTimeMillis();
       System.out.println("datum traverse Time is " +(stopTime - startTime) / 1000.0 + " seconds.");
       
       
       int[] lower2 = { 1,1};
       int[] upper2 = {2,2};
       ISBounds tbounds2 = new ISBounds( new IntegerIndexSpaceID(lower2), new IntegerIndexSpaceID(upper2));
       BasicBlock smallBlock;
       
       startTime = System.currentTimeMillis();
       // for (iter.init(); iter.valid(); iter.next())
       // {
       smallBlock = chunkingsource.subPartitionedBlock( tbounds2);
       System.out.println(smallBlock);
       //  }
       stopTime = System.currentTimeMillis();
       System.out.println("subblock Time is " +(stopTime - startTime) / 1000.0 + " seconds.");
       
       
       int[] lower3 = { 0,0};
       int[] upper3 = {3,3};
       
       
       
       ISBounds tbounds3 = new ISBounds( new IntegerIndexSpaceID(lower3), new IntegerIndexSpaceID(upper3));
       
       
       BasicBlock passrefBlock = new BasicBlock( tbounds3 , smallBlock.getRecordDescriptor());
       
       System.out.println(" before subblock" + passrefBlock );
       chunkingsource.subblock(passrefBlock, passrefBlock.getBounds(), passrefBlock.getBounds());
       
       System.out.println( passrefBlock);
       
       
       System.out.println(" ******************************************* ");
       System.out.println(" chunked data Source is " + chunkingsource );
       
       int[] lower4 = { 0,0};
       int[] upper4 = {999,999};
       ISBounds tbounds4 = new ISBounds( new IntegerIndexSpaceID(lower4), new IntegerIndexSpaceID(upper4));
       
       IntegerIndexSpaceID out_bufferDimension = new IntegerIndexSpaceID(new int[2]);
       int[] out_steps = new int[2];
       chunkingsource.computeBestBufferSize( chunkingsource.getPartitionSpaceBounds(), tbounds4,
       100, out_bufferDimension, out_steps);
       
       System.out.println(" ******************************************* ");
       
       DataSource ds1 = DataSource.createPhysicalDataSource("/local/tmp/xtang/trav/testdata/file.24.1.bin.3d.xml", "foo");
       ds1.activate();
       RecordDescriptor rec1 = ds1.getRecordDescriptor();
       System.out.println("rec1 " + rec1);
       System.out.println("ds1 " + ds1);
       
       AxisOrdering order1 =((BinaryFileDataSource)ds1).getAxisOrdering();
       int[] partSpaceBounds1 = { 64,64,64};
       ChunkedDataSource chunkingsource1 = new ChunkedDataSource("foo",
       new ISBounds(partSpaceBounds1), ds1.getBounds(),
       "file.24.1.bin",order1,rec1 );
       System.out.println(" new chunked Datasource is : " + chunkingsource1 );
       int[] lower5 = { 0,0,0};
       int[] upper5 = {59,59,59};
       tbounds4 = new ISBounds( new IntegerIndexSpaceID(lower5), new IntegerIndexSpaceID(upper5));
       
       
       out_bufferDimension = new IntegerIndexSpaceID(new int[3]);
       out_steps = new int[3];
       chunkingsource1.computeBestBufferSize( chunkingsource1.getPartitionSpaceBounds(), tbounds4, 100,
       out_bufferDimension, out_steps);
       */
      
      ISBounds tbounds4;
      
      System.out
            .println( "********************************************************\n\n\n\n" );
      
      PhysicalDataSource ds2 =(PhysicalDataSource) DataSource
            .create( "foo", "/home/xtang/file/simplefile.bin.xml" );
      ds2.activate();
      RecordDescriptor rec2 = ds2.getRecordDescriptor();
      System.out.println( "rec1 " + rec2 );
      System.out.println( "ds1 " + ds2 );
      
      AxisOrdering order2 =((BinaryFileDataSource) ds2 ).getAxisOrdering();
      int[] partSpaceBounds2 = { 4, 4, 4 };
      ChunkedDataSource chunkingsource2 = new ChunkedDataSource( "foo",
            new ISBounds( partSpaceBounds2 ), ds2.getBounds(),
            //"/home/csg/xtang/file/simplefile.bin",order2,rec2 );
            "/home/xtang/file/simplefile.bin", order2, rec2 );
      System.out.println( " new chunked Datasource is : " + chunkingsource2 );
      int[] lower6 = { 0, 0, 0 };
      int[] upper6 = { 2, 2, 2 };
      tbounds4 = new ISBounds( new IntegerIndexSpaceID( lower6 ),
            new IntegerIndexSpaceID( upper6 ));
      
      chunkingsource2.activate();
      
      System.out
            .println( " ******************************************\n********** chunking tests**********************\n" );
      
      int[] partSpaceBounds3 = { 4, 4, 4 };
      ChunkedDataSource chunkingsource3 = new ChunkedDataSource( "foo",
            new ISBounds( partSpaceBounds3 ), ds2.getBounds(),
            //"/local/tmp/xtang/trav/testdata/file.24.1.bin",order2,rec2 );
            //"/home/csg/xtang/file/simplefile.bin",order2,rec2 );
            "/home/xtang/file/simplefile.bin", order2, rec2 );
      chunkingsource3.activate();
      System.out.println( " chunkingsource3 " + chunkingsource3 );
      
      int[] lower7 = { 0, 0, 0 };
      int[] upper7 = { 5, 5, 5 };
      tbounds4 = new ISBounds( new IntegerIndexSpaceID( lower7 ),
            new IntegerIndexSpaceID( upper7 ));
      
      BasicBlock theBlock = new BasicBlock( tbounds4, rec2 );
      BasicBlock theBlock1 = new BasicBlock( tbounds4, rec2 );
      BasicBlock theBlock2 = new BasicBlock( tbounds4, rec2 );
      BasicBlock theBlock3 = new BasicBlock( tbounds4, rec2 );
      BasicBlock theBlock4 = new BasicBlock( tbounds4, rec2 );
      
      chunkingsource3.setMaxGapSize( 20 );
      //chunkingsource3.allocateBuffer(tbounds4);
      chunkingsource3.subBasicBlock( theBlock, tbounds4, tbounds4 );
      System.out.println( theBlock );
      System.out.println( " ************ number of reads :"
            + chunkingsource3.getReadCounter());
      System.out
            .println( " ******************************************\n************************************\n" );
      
      chunkingsource3.resetReadCounter();
      chunkingsource3.setMaxGapSize( 2 );
      //chunkingsource3.allocateBuffer(tbounds4);
      chunkingsource3.subBasicBlock( theBlock1, tbounds4, tbounds4 );
      System.out.println( theBlock1 );
      System.out.println( " ************ number of reads :"
            + chunkingsource3.getReadCounter());
      chunkingsource3.subBasicBlock( theBlock4, tbounds4, tbounds4 );
      System.out.println( theBlock4 );
      
      System.out
            .println( " ******************************************\n********** bfds tests**************************\n" );
      ds2.setMaxGapSize( 20 );
      //ds2.allocateBuffer(tbounds4);
      ds2.subBasicBlock( theBlock2, theBlock.getBounds(), tbounds4 );
      System.out.println( theBlock2 );
      System.out.println( " number of reads :" + ds2.getReadCounter());
      
      ISIterator iter = new ISIterator( ds2.getBounds(), ds2.getAxisOrdering());
      iter.init();
      for ( int i = 0; i < 10; i++ )
      {
         iter.next();
         System.out.println( ds2.datum( iter ));
      }
      
      System.out
            .println( " ******************************************\n************************************\n" );
      
      ds2.resetReadCounter();
      ds2.setMaxGapSize( 0 );
      //ds2.allocateBuffer(tbounds4);
      ds2.subBasicBlock( theBlock3, theBlock.getBounds(), tbounds4 );
      System.out.println( theBlock3 );
      System.out.println( " number of reads :" + ds2.getReadCounter());
      
      ds2.resetReadCounter();
      ds2.setMaxGapSize( 0 );
      //ds2.allocateBuffer(tbounds4);
      ds2.subblock( theBlock4, theBlock.getBounds(), tbounds4 );
      System.out.println( theBlock4 );
      System.out.println( " number of reads :" + ds2.getReadCounter());
   }
   
   /** This is a preliminary implementation.  Assume that the dstBounds
    * is the bounds of dstBlock. That means the bounds of the query block
    * MUST be the same as the bounds of the dstBlock. Also, the dstBlock
    * has to be a basicBlock.  The dstBounds and srcBounds messured in
    * the GRID space, so that it matches the interface of existing
    * CacheDataSource
    *
    * If this.setGapSize(int) has been called, and gapSize have been
    * set to a nonnegative value, a smart subblock function is called.
    * The smart subblock function eliminates unnecessary reads if gap
    * between two contiguous read is smaller that the gap size specified by
    * this.setGapSize.  if this.gapSize is a negative value, subblock
    * function reads one partitionSpaceRod at a time, and skip the gaps
    * between contiguous rods.
    */
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      if ( this.maxGapSize <= 0 )
         this.noGapSubBlock( dstBlk, dstBnds, srcBnds );
      else
         this.gappedSubBlock( dstBlk, dstBnds, srcBnds );
   }
   
   private void noGapSubBlock( BasicBlock dstBlock, ISBounds dstBounds,
         ISBounds srcBounds )
   {
      ISBounds partSpaceQueryBnds = this.chunkingPartitioning.map( srcBounds );
      
      this.iter.setBounds( partSpaceQueryBnds );
      
      long chunkOffset;
      int storageType = dstBlock.getStorageType();
      
      int partRodSize = partSpaceQueryBnds.getDimension( this.leastSignificantAxis );
      
      if ( ! allocateBuffer( partRodSize * this.chunkByteSize ))
         partRodSize = getBufferSize() / this.chunkByteSize;
           
      long fileOffset = 0;
      int dstOffset = 0;
      
      // Let's precompute ChunkBounds.volume()
      int bufferSize = chunkVolume * this.numAttributes * partRodSize;
      
      int n = 0;
      try
      {
         for ( iter.init(); iter.valid(); iter.next())
         {
            chunkOffset = AxisOrdering.computeOffset( iter, this.partSpaceAxisWeights );
            
            fileOffset = chunkOffset * this.chunkByteSize;
            if ( fileOffset !=((int) raFile.getFilePointer()) )
               raFile.seek( fileOffset );
            getBuffer().clear();
            int bytesRead = fChannel.read( getBuffer() );
            
            //*************** testing util, take me out after test is done !!! *************
            readCounter++ ;
            readLength += bytesRead;
            //*****************************************************************************
            
            getBuffer().clear();
            for ( int i = 0; i < partRodSize; i++ )
            {
               dstOffset = n * bufferSize + chunkRodLength * i;
               //for ( chunkIter.init(); chunkIter.valid(); chunkIter.next())
               for ( int j = 0; j < this.numChunkRods; j++ )
                  
               {
                  //System.out.println("scrOffset : " + floatBuffer.position());
                  //System.out.println("dstOffset : " + dstOffset);
                  //System.out.println(" _______________________ ");
                  switch ( storageType )
                  {
                     // rdb 03/04/12: unlike loadChunk code, this code does not
                     //        need to divide the chunkRodLength by bytes per value
                     //        because chunkRodLength is already in units of float
                     //        or int or double, etc.
                     case StorageType.Byte:
                        getBuffer().get( dstBlock.getBytes(), dstOffset,
                                            chunkRodLength );
                        break;
                     case StorageType.Float:
                        getBuffer().asFloatBuffer().get( dstBlock.getFloats(),
                                            dstOffset, chunkRodLength );
                        break;
                     case StorageType.Double:
                        getBuffer().asDoubleBuffer().get( dstBlock.getDoubles(),
                                            dstOffset, chunkRodLength );
                        break;
                     case StorageType.Int:
                        getBuffer().asIntBuffer().get( dstBlock.getInts(), 
                                            dstOffset, chunkRodLength );
                        break;
                     case StorageType.Short:
                        getBuffer().asShortBuffer().get( dstBlock.getShorts(),
                                            dstOffset, chunkRodLength );
                        break;
                     default:
                        new SDBError( SDBError.Fatal, "Type not implemented: " + 
                                            storageType );
                  }
                  dstOffset += chunkRodLength * partRodSize;
               }
            }
            n++ ;
         }
      }
      catch( Exception e )
      {
         new SDBError(SDBError.Recoverable, e.toString());
      }
   }
   
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         RecordSpec recSpec )
   {
      new SDBError( SDBError.Warning, "RecordSpec subblock not implemented." );
   }
   
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         FieldIDMapper mp )
   {
      new SDBError( SDBError.Warning, "FieldIDMapper subblock not implemented." );
   }
   
  /** Reads a subblock defined by srcBounds into the region of the Datablock argument
  defined by dstBounds. The FieldIDMapper argument determines the mapping between 
  datums in the datasource and datums in the DataBlock. Note that the DataBlock is
  assumed to already have a valid data array. the dataBlock can be basicBlock or compositeBlock
  */
 public void subblock( DataBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds, 
                               FieldIDMapper mp )
 {
    checkActivated(); 
    if ( mp == null && dstBlock instanceof BasicBlock)
       //mp = new FieldIDMapper( this.getNumAttributes() );
       subBasicBlock((BasicBlock) dstBlock, dstBounds, srcBounds);
       
    else if ( dstBlock instanceof BasicBlock )
       subBasicBlock((BasicBlock) dstBlock, dstBounds, srcBounds, mp );
    else if ( dstBlock instanceof CompositeBlock )
       subCompositeBlock((CompositeBlock) dstBlock, dstBounds, srcBounds, mp );
    else
    {
       new SDBError( SDBError.Recoverable, "Unsupported DataBlock type: " 
                                     + dstBlock.getClass().getName() );
    }
 }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      new SDBError( SDBError.Warning, "RecordSpec subblock not implemented." );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         RecordSpec recSpec )
   {
      new SDBError( SDBError.Warning, "RecordSpec subblock not implemented." );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         FieldIDMapper mp )
   {
      new SDBError( SDBError.Warning, "FieldIDMapper subblock not implemented." );
   }
}