/*    
 CompositeDataSource.java
 last modified 1/16/2002
 Copyright 2000 University of New Hampshire
 Durham, NH 03824
 
 ----------- Edit history as of 01/19/05 ---------------------------
 11/29/05 rdb: Added initialization of FDLDescriptor
 10/21/05 rdb: Deleted unused variables to avoid compiler warnings
 02/14/05 rdb: allowed constructor to take null arguments, created defaults
               deleted the multiple ByteBuffers - each component has its own.
 02/07/05 rdb: added initializeBuffers() and some cleaning up.
                  
 */

package edu.unh.sdb.datasource;

//import java.io.*;

import edu.unh.sdb.common.FieldIDMapper;
import edu.unh.sdb.common.IllegalArgumentError;
import edu.unh.sdb.common.SDBError;

/** This class is a parent class for AttributeJoinDataSoure and 
 *  BlockedDataSource. It provides access methods concerned with arrays
 *  of ComponentInfo objects. 
 */
abstract public class CompositeDataSource extends DataSource
{
   /** Information for every component. */
   protected ComponentInfo[] componentInfos;

   /**
    * Create a CompositeDataSource defined by the ComponentInfo array.
    *
    */
   
   public CompositeDataSource( String aname, ComponentInfo[] cinfo )
   {
      super( aname );

      this.check( cinfo );
      this.componentInfos = cinfo; //clone()?
      this.dimensionality = cinfo[0].getISIDMap().getCompositeExtent().dim();
      /* set this.recordDesc, this.numAttributes and this.bounds. */
      initializeComponents( cinfo );
      this.numAttributes = this.recordDesc.getNumAttributes();
      this.fdlDescriptor = new FDLDescriptor(
                                 this.recordDesc, this.bounds, this.axisOrdering );
   }

   /** Create a new CompositeDataSource object from the given arrays 
    *  that are all parallel to the array this.componentInfos.
    */
   CompositeDataSource( String aname, DataSource[] sources,
         FieldIDMapper[] FIDmap, ISBounds[] componentBounds,
         ISBounds[] compositeBounds )
   {
      super( aname );
      int size = sources.length;
      ISIDMapper[] ISIDmap = new ISIDMapper[size]; //parallel to sources[]
      this.componentInfos = new ComponentInfo[size];

      // create ISIDMappers and componentInfos 
      for ( int i = 0; i < size; i++ )
      {
         ISIDmap[i] = new ISIDMapper( componentBounds[i], compositeBounds[i] );
         this.componentInfos[i] = new ComponentInfo( sources[i], FIDmap[i],
               ISIDmap[i] );
      }
      this.check( this.componentInfos );
      this.name = aname;
      this.dimensionality = componentBounds[0].dim();
      initializeComponents( this.componentInfos );
      this.fdlDescriptor = new FDLDescriptor(
                                    this.recordDesc, this.bounds, this.axisOrdering );
   }

   /** Create a new CompositeDataSource object from the given arrays 
    *  that are all parallel to the array this.componentInfos.
    */
   CompositeDataSource( String aname, DataSource[] sources,
         FieldIDMapper[] FIDmap, ISIDMapper[] ISIDmap )
   {
      super( aname );
      int len = sources.length;
      if ( FIDmap == null )
      {
         FIDmap = new FieldIDMapper[ len ];
         for ( int i = 0; i < len; i++ )
            /*** rdb: Is it better to have default FIDMapper or null?
             *   default mapping is good fro BlockDS, not AJDS.
             * I think I'll try to fix the "null" situation. 
             */ 
            //FIDmap[i] = new FieldIDMapper( sources[0].getNumAttributes() );
            FIDmap[ i ] = null;
      }

      if ( ISIDmap == null )
      {
         ISIDmap = new ISIDMapper[ len ];
         for ( int i = 0; i < len; i++ )
            ISIDmap[i] = new ISIDMapper( sources[i].getBounds() );
      }

      //Check to see if the array length matches and create component info
      if (( len == FIDmap.length ) && ( len == ISIDmap.length ))
      {
         this.componentInfos = new ComponentInfo[ len ];
         for ( int i = 0; i < len; i++ )
            this.componentInfos[i] = new ComponentInfo( sources[i], FIDmap[i], ISIDmap[i] );
      }
      else
      {
         new IllegalArgumentError( SDBError.Fatal, "DataSource array's length " +
         		"doesn't match FieldIDMapper array's or ISIDmapper array's length" );
      }

      this.check( this.componentInfos );
      this.dimensionality = ISIDmap[0].getCompositeExtent().dim();
      initializeComponents( this.componentInfos );
      this.fdlDescriptor = new FDLDescriptor(
                                    this.recordDesc, this.bounds, this.axisOrdering );
   }

   /* Initialize Component infromation: recordDesc, bounds, numAttributes. */
   abstract protected void initializeComponents( ComponentInfo[] cpnInfo );
   
   /** Check the array of componentInfo */
   abstract protected boolean check( ComponentInfo[] cpnInfo );

   /** Returns the reference to the array this.componentInfos[].
    */
   public ComponentInfo[] getComponentInfos()
   {
      return this.componentInfos;
   }

   /** When this method is called, the subblock methods of this collection will 
    *  read data as bytes, without performing any type conversion. This is meant 
    *  to be used by servers, since they send raw bytes over the network anyway,
    *  and any conversion is wasted effort.
    */
   public void useBytesOnly(){

      int i, len = this.componentInfos.length;
      DataSource ds;

      for ( i = 0; i < len; i++ )
      {
         ds = (DataSource) this.componentInfos[i].getComponent();

         ds.useBytesOnly();
      }

   }



   /** Calls the activate method of all component datasources.*/
   public void activate()
   {
      super.activate();

      int i, len = this.componentInfos.length;
      DataSource ds;

      for ( i = 0; i < len; i++ )
      {
         ds = (DataSource) this.componentInfos[i].getComponent();

         if ( !ds.isActivated() )
         {
            ds.activate();
         }
      }
   }

   /** Calls the deactivate method of all component datasources.*/
   public void deactivate()
   {

      super.deactivate();

      int i, len = this.componentInfos.length;
      DataSource ds;

      for ( i = 0; i < len; i++ )
      {
         ds = (DataSource) this.componentInfos[i].getComponent();

         if ( ds.isActivated() )
         {
            ds.deactivate();
         }

      }
   }

   /** Returns the array of component DataSources that are corresponding to
    *  the array this.componentInfos[]. The DataSources may or may not
    *  be distinct.
    */
   public DataSource[] getComponentDataSources()
   {

      int size = this.componentInfos.length;
      DataSource[] DSarr = new DataSource[size];

      for ( int i = 0; i < size; i++ )
      {
         DSarr[i] = (DataSource) this.componentInfos[i].getComponent();
      }
      return DSarr;
   }

   /** Returns the array of FieldIDMappers that are corresponding to
    *  the array this.componentInfos[]. 
    */
   public FieldIDMapper[] getFIDMappers()
   {
      int size = this.componentInfos.length;
      FieldIDMapper[] fidmap = new FieldIDMapper[size];

      for ( int i = 0; i < size; i++ )
      {
         fidmap[i] = this.componentInfos[i].getFieldIDMap();
      }
      return fidmap;
   }

   /** Returns the array of componentFieldNumbers for each ComponentInfo's
    *  FieldIDMapper.
    */
   public int[][] getComponentFieldIDs()
   {

      int fids[][];
      int i, size = this.componentInfos.length;

      fids = new int[size][];
      for ( i = 0; i < size; i++ )
      {
         fids[i] = this.componentInfos[i].getFieldIDMap().getComponentFields();
      }
      return fids;

   }

   /** Returns the array of compositeFieldNumbers for each ComponentInfo's
    *  FieldIDMapper.
    */
   public int[][] getCompositeFieldIDs()
   {

      int fids[][];
      int i, size = this.componentInfos.length;

      fids = new int[size][];
      for ( i = 0; i < size; i++ )
      {
         fids[i] = this.componentInfos[i].getFieldIDMap().getCompositeFields();
      }
      return fids;

   }

   /** Returns the array of component ISBoundses corresponding to the array
    *  this.componentInfos[].
    */
   public ISBounds[] getComponentBounds()
   {

      int size = this.componentInfos.length;
      ISBounds[] isbs = new ISBounds[size];
      ISIDMapper isidm;

      for ( int i = 0; i < size; i++ )
      {
         isidm = this.componentInfos[i].getISIDMap();
         isbs[i] = isidm.getComponentSubset();
      }
      return isbs;
   }

   /** Returns the array of destination ISBoundses corresponding to the 
    *  array this.componentInfos[].
    */
   public ISBounds[] getCompositeBounds()
   {

      int size = this.componentInfos.length;
      ISBounds[] isbs = new ISBounds[size];
      ISIDMapper isidm;

      for ( int i = 0; i < size; i++ )
      {
         isidm = this.componentInfos[i].getISIDMap();
         isbs[i] = isidm.getCompositeExtent();
      }
      return isbs;
   }

   /** Return true if parameter equals this object, false otherwise. */
   public boolean equals( Object obj )
   {

      int i, size = this.componentInfos.length;
      CompositeDataSource cpsDS = (CompositeDataSource) obj;
      ComponentInfo[] cinfos = cpsDS.getComponentInfos();

      if ( size != cinfos.length )
         return false;

      for ( i = 0; i < size; i++ )
      {
         if ( !( this.componentInfos[i].equals( cinfos[i] ) ) )
            return false;
      }

      return true;
   }

   /** Print the data members of the DataSource in an easy to read fashion.*/
   public String toString()
   {

      int i;
      int numComponents = this.componentInfos.length;
      StringBuffer result = new StringBuffer( "   CompositeDataSource:\n" );
      DataSource component;

      for ( i = 0; i < numComponents; i++ )
      {
         result.append( "            component name=" );
         component = (DataSource) this.componentInfos[i].getComponent();
         result.append( component.getName() );
         result.append( "\n" );
      }

      result.append( super.toString() );

      return new String( result );
   }

   //-------------------- main --------------------------------------
   public static void main( String args[] )
   {
   }
}