//
//  DSArrayStencil.java
//  ObliqueCacheTest
//
//  Created by Philip Rhodes on 6/3/06.
//  Copyright 2006 Philip J Rhodes. All rights reserved.
//

package edu.unh.sdb.datasource;

import java.util.*;
//import java.io.*;
import edu.unh.sdb.common.*;



/** This is an implementation of a partitioned stencil that uses an array to
 *  maintain the record of active elements. As a result, the set and get
 *  methods are extremely fast, but iteration might be a little less efficient
 *  than the list implementation. However, it does not tax the garbage collector,
 *  so it may be more appropriate for repeated use.
 */
public class DSArrayStencil extends DSPartitionedStencil{
   
   protected boolean [] inStencil;
   
   protected int size = 0;
   
   protected int  lastIndexRequested=-1;  // last index given to a getIndexed()
   protected int lastActiveOffset=-1; // last offset of an active element
                                         // found by getIndexed()

   protected int dataBytesUsed =0;  // Data bytes used by last call to getData()
   protected DataBlock block; // used in getData()



    // Old
   /** Construct a stencil that identifies areas of interest using a partitioning.
    *  The first argument is the dimensions of the partitioning space, meaning
    *  that the index space bounded by the second argument will be divided up into
    *  a matching number of partitions in each dimension. For example, if we build
    *  an 8x8 partitioning of a 128x128 index space, each partition will have
    *  dimensions 16x16.  In this situation, the first argument for this constructor
    *  should be [8,8] and the second argument should be {<0,0>, <128,128>}.
    */
    
   /** Construct a stencil that identifies areas of interest using a partitioning.
    *  The first argument is the dimensions of an element of the partitioning, 
    *  meaning that the index space bounded by the second argument will be divided
    *  up into pieces of the shape described by the first argument. For example, 
    *  if we build an 8x8 partitioning of a 128x128 index space, each element
    *  of the partitioning will have dimensions 16x16. In this situation, 
    *  the first argument for this constructor should be [16,16] and the second
    *  argument should be {<0,0>, <127,127>}.
    */   
//   public DSArrayStencil( int [] elementDimensions, ISBounds space){
//
//      super(new RegularISPartitioning(elementDimensions, space));
//      this.inStencil = new boolean[(int) this.partitioning.volume()];
//   }


   public DSArrayStencil( RegularISPartitioning partitioning){

      super(partitioning);
      
      this.inStencil = new boolean[(int) this.partitioning.volume()];
   }
   
   /** Fill the given byte array with the data contained in the ChomboDataSource
    *  given to the method.  The ChomboDataSource is able to calculate the number
    *  of bytes used in the data buffer, and stores a reference to the most
    *  recently used buffer which contains the appropriate data.  This method
    *  also stores the number of data bytes used in the array, and may be retrieved
    *  by calling getDataBytesUsed().
    * 
    * @param data The byte buffer to fill
    * @param cds  ChomboDataSource object reference
    * @return The byte buffer
    */
   public byte[] getData(byte[] data, ChomboDataSource cds)
   {
       // Make byte buffer and write header based on this stencil.
       int offset = 0;
       data = cds.makeByteBuffer(data, this);
       offset = cds.writeHeader(data, this);
       for(int i = 0; i < this.size(); i++)
       {
           offset = cds.server_subblock(data, this.getIndexedBounds(i), offset);
       }
       
       this.dataBytesUsed = cds.getBufferSize();
       data = cds.getBuffer();
       return data;
   }

   /** Fill the given byte array with the data contained in the active elements
    *  of this stencil. If the given array is not long enough to contain the stencil
    *  data, a new array is allocated and returned. The number of bytes actually used
    *  in the array can be retrieved by calling getDataBytesUsed().
    */
   public byte [] getData(byte [] data, DataSource ds){
   
      int requiredCapacity = this.size * (int) this.partitionVolume();
      int dataOffset = 0;
      int volume;
      ISBounds blockBounds;
      

      if (this.block == null){
      
         blockBounds = new ISBounds(this.getPartitionDimensions());
         this.block=new BasicBlock(blockBounds, ds.getRecordDescriptor());
      }
      
      volume = block.volume();
      
      if(data == null || data.length < requiredCapacity) {
      
         data = new byte[requiredCapacity];
      }
      
      for(int i = 0; i< this.size; i++){
      
         blockBounds = this.getIndexedBounds(i);
         
         ds.subblock(block, blockBounds);
      
         System.arraycopy(block.getValues().getBytes(), 0, data, dataOffset, volume);
         
         dataOffset += volume;
         
      }
   
      this.dataBytesUsed=dataOffset;
      
      return data;
   }


   /** Return the number of bytes actually used in the last call to getData().
    */
   public int getDataBytesUsed()
   {
      return this.dataBytesUsed;
   }



   
   /** Set the stencil element containing the id to "active". If it is already
     * active, then the method returns false. Returns true otherwise.
     */
   public boolean set(IndexSpaceID id){
   
   
      int offset = (int) this.partitioning.containingOffset(id);
      
      boolean wasntSet = !this.inStencil[offset];
      
      if(wasntSet) {
      
         this.inStencil[offset] = true;         
         this.size++;
      }
   
      return wasntSet;
   }


   /** Set the stencil element with index corresponding to id to "active.  id 
     * should already have been mapped to the partitioning space. If it is already
     * active, then the method returns false. Returns true otherwise.
     */
   public  boolean setElement(IndexSpaceID pid){
   
   
      int offset = (int) this.partitioning.elementOffset(pid);
      
      boolean wasntSet = !this.inStencil[offset];
      
      if(wasntSet) {
      
         this.inStencil[offset] = true;         
         this.size++;
      }
   
      return wasntSet;
   }




   /** Return true if the given id maps to an active stencil element, false otherwise.*/
   public boolean get(IndexSpaceID id){

      int offset = (int) this.partitioning.containingOffset(id);
      
      return this.inStencil[offset];
   }




   /** 
   *   Sets the argument ISBounds to the bounds of the specified stencil element.
   *   The ordering of the stencil elements is arbitrary. If the specified index
   *   is not less than the number of elements, then the bounds argument is
   *   not changed, and the method returns null. Otherwise, the bounds are 
   *   modified and a reference to the argument bounds is returned.
   */
   public ISBounds getIndexedBounds( ISBounds stencilBounds, int index )
   {
      long offset;
      
      if( index < this.size){
      
         offset = this.getIndexedOffset( index );
         this.partitioning.getOffsetBounds( stencilBounds, offset );
         return stencilBounds;
      } else {
      
         return null;
      }
      
   }

   /** 
    * Return the partition offset corresponding to the given index.
    */
   protected  final long getIndexedOffset( int index ){

      int offset;
      int startIndex;
      
      if( index > this.lastIndexRequested ) {
      
         offset = this.lastActiveOffset + 1;
         startIndex = this.lastIndexRequested + 1; // for loop runs (index - lir) times
      } else if( index == this.lastIndexRequested ){
      
         offset = this.lastActiveOffset;
         startIndex = index + 1; // for loop won't run
      } else {
      
         offset = 0;
         startIndex = 0; // for loop runs (index + 1) times
      }
      
      
            
      for(int i=startIndex; i<=index; i++){
      
         while( ! inStencil[offset]){
         
            offset++;
         }
      }
      
      this.lastIndexRequested = index;
      this.lastActiveOffset = offset;
      
      
      
      return offset;
   
   }



   /** Set the element corresponding to the 1D offset to be active.*/
   protected void setElementFromOffset(int offset){
   
      this.inStencil[offset] = true;
      this.size++;
   }



   /** return number of active elements contained by this stencil
   */
   public int size()
   {
     return this.size;
   }

   /** Clear this stencil of all elements.
   */
   public void clear(){

      Arrays.fill(this.inStencil, false);
      this.size = 0;
      
      this.lastIndexRequested = -1;
      this.lastActiveOffset = -1;
   }


   /** Return a String representing the object. */
   public String toString ()
   {
      
      return this.toString ("");
   }

   public String toString(String indent){
   
      String indent2=indent+Globals.toStringIndent;
      int length = this.inStencil.length;
      
      StringBuffer result = new StringBuffer(indent + "{ DSArrayStencil:\n");
      result.append(super.toString(indent2)+"\n");
      
      result.append(indent2 + "Stencil size =" + size +"\n");
      
      result.append(indent2 + "array= ");
      
      for (int i=0; i< length; i++){
      
         if(this.inStencil[i]){
            result.append("1 ");
         } else {
            result.append("0 ");
         }
      
      }
      
      result.append("\n");
      
      result.append(indent + "}");
      
      return new String(result);
   
   }

   //_____________________________________________________________



    public static void  main(String [] argv){

        ISBounds spaceBounds=new ISBounds(new int [] {10,10});

        //DSArrayStencil stencil = new DSArrayStencil(new int[] {5,5}, spaceBounds);
        DSArrayStencil stencil = new DSArrayStencil(new RegularISPartitioning(new IntegerIndexSpaceID(new int[] {5,5}), spaceBounds));


        stencil.set(new IntegerIndexSpaceID(2,2));

        stencil.set(new IntegerIndexSpaceID(8,8));


        System.out.println("2,2 yields "+stencil.get(new IntegerIndexSpaceID(2,2)));
        System.out.println("2,8 yields "+stencil.get(new IntegerIndexSpaceID(2,8)));
        System.out.println("8,2 yields "+stencil.get(new IntegerIndexSpaceID(8,2)));
        System.out.println("8,8 yields "+stencil.get(new IntegerIndexSpaceID(8,8)));

    }


}
