//
//  DSPartitionedStencil.java
//  StencilTest
//
//  Created by Philip Rhodes on 6/6/06.
//  Copyright 2006 Philip J Rhodes. All rights reserved.
//

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;


public abstract class DSPartitionedStencil extends DSStencil{

   protected int[] partitioningDimensionsArray;
   protected RegularISPartitioning partitioning;
   protected int partitionListBytesLength = 0;
   protected int headerLength = 4; // length of the header when writing to a byte array.


//   public DSPartitionedStencil( int [] partitioningDimensions, ISBounds space){
//   
//      super(space);
//      
//      this.partitioningDimensionsArray = partitioningDimensions;
//      this.partitioning = new RegularISPartitioning(partitioningDimensions, space);
//   }


   public DSPartitionedStencil( RegularISPartitioning partitioning){
   
      super(partitioning.getSpaceBounds());
      
      this.partitioningDimensionsArray = 
               partitioning.getPartitioningDimensions(new int[partitioning.dim()]);
               
      this.partitioning = partitioning;
   }




   /** Fill the given array with a representation of the partition list.
    *  This representation is meant to be parsed by setPartitionListFromBytes().
    *  If the array argument is null or refers to an array of insufficient
    *  length, a new byte array will be allocated and a reference returned.
    *  After this method returns, the number of bytes consumed by the representation
    *  can be retrieved by a call to getPartitionListBytesLength().
    */
   public byte [] getPartitionListBytes(byte [] bytes){


      int len = this.size();
      //int d = this.partitioning.dim();
      //int [] coords;
      int bytesPerValue = this.chooseBytesPerValue();


      // Set the length data member so it can be retrieved using
      // getPartitionListBytesLength().

      this.partitionListBytesLength = len * bytesPerValue + this.headerLength;

      // If the argument was either null, or points to an array
      // that is too short, we will allocate a new array.
      if(bytes == null || bytes.length < (len * bytesPerValue + this.headerLength)) {

         bytes = new byte [len * bytesPerValue + this.headerLength];
      }

      // Write the header.
      // First, we write the number of active elements to the beginning
      // of the array as a short.
      bytes[0] = (byte) ((len & 0x0000FF00) >>> 8);
      bytes[1] = (byte) ((len & 0x000000FF));

      // Next, we write the number of bytes per coordinate as a short.
      // This is too much space for a small value, but we can preserve
      // fullword alignment if we use two bytes.
      
      bytes[2] = (byte) (0);
      bytes[3] = (byte) bytesPerValue;

      // Now write the partition coordinates to the array.
      
      int offset;
      
      for(int i=0; i<len; i++){
      
         offset = (int) this.getIndexedOffset(i);

         for(int k = bytesPerValue - 1; k >= 0; k--){

            bytes[this.headerLength + bytesPerValue * i + k ] =
                  (byte) (offset & 0x000000FF);

            offset >>>= 8;
         }
      }


//      System.out.print("getPartitionListBytes(): bytes=");
//      for(int j=0; j< bytes.length; j++){
//
//         System.out.print(" "+bytes[j]);
//      }
//      System.out.println();

      return bytes;
   }
   

   
      
   /** Set the partition list according to the data found in the
    *  given array of bytes.
    */
   public byte [] setPartitionListFromBytes(byte [] bytes){

//      System.out.print("setPartitionListFromBytes(): bytes=");
//      for(int j=0; j< bytes.length; j++){
//
//         System.out.print(" "+bytes[j]);
//      }
//      System.out.println();


      this.clear();
      int i=-5, k=-5,bytesPerValue=-5, size = -5;

      try{
         // First, read the number of active partitions from the array.
         size = ( (bytes[0] & 0xFF) << 8 | (bytes[1] & 0xFF));

         // Now read the number of bytes per offset value
         bytesPerValue = ( (bytes[2] & 0xFF) << 8 | (bytes[3] & 0xFF));  

         
         int offset;

         for(i=0; i < size; i++){
         
            offset = bytes[this.headerLength + bytesPerValue * i + 0] & 0xFF;

            for(k=1; k< bytesPerValue; k++){

               offset <<= 8;

               offset |= bytes[this.headerLength + bytesPerValue * i + k] & 0xFF;

            }        

            this.setElementFromOffset(offset);
         }
      } catch (ArrayIndexOutOfBoundsException e){

         System.out.println("EXCEPTION: OOB in setPartitionListFromBytes:");
         System.out.println("                  size: " + size);
         System.out.println("     this.headerLength: " + this.headerLength);
         System.out.println("    this.bytesPerValue: " + bytesPerValue);
         System.out.println("                     i: " + i);
         System.out.println("                     k: " + k);
         System.out.println("this.headerLength + bytesPerValue * i + k: " + (this.headerLength + bytesPerValue * i + k));
         System.out.println("bytes.length: " + bytes.length);
         System.out.println();
         
         new SDBError(SDBError.Recoverable, "Array out of bounds while setting stencil from byte array.");

      }
      return bytes;
   }
   
   

   /** Return the number of bytes used by the last call to
    *  getPartitionListBytes().
    */
   public int getPartitionListBytesLength(){


      return this.partitionListBytesLength;
   }





   /** determine how many bytes to use per offset when representing the
    *  stencil as an array of bytes.
    */
   protected byte chooseBytesPerValue(){
   
      int length = (int) this.partitioning.volume();
   
      if ( (length & 0xFFFFFF00) == 0) {
      
         return 1;
      }

      if ( (length & 0xFFFF0000) == 0) {
      
         return 2;
      }

      if ( (length & 0xFF000000) == 0) {
      
         return 3;
      }
      
      return 4;      
   }


   /** Return the partition offset corresponding to the given index.*/
   protected abstract long getIndexedOffset(int index);


         
         
   /** Set the element corresponding to the 1D offset to be active.*/
   protected abstract void setElementFromOffset(int offset);
         
         
   
   /** Clear this stencil of all elements.
   */
   public abstract void clear();   
   

  /** Returns the dimensions of an element as an IndexSpaceID.
   */
   public IndexSpaceID getPartitionDimensions(){

     return new IntegerIndexSpaceID(this.partitioning.getPartitionDimensionsArray());
   }

  /** Returns the dimensions of an element as an int array.
   */
  public int [] getPartitionDimensionsArray(){
  
     return this.partitioning.getPartitionDimensionsArray();
  }
   
   
   
   /** Return an array of integers indicating the dimensions of the 
    *  partitioning. This represents the number of divisions per dimension
    *  of the space.
    */
   public int [] getPartitioningDimensions(){
   
      return this.partitioningDimensionsArray.clone();
   }


   /** Return true if the argument represents a partition coordinate that is 
    *  within the bounds of the stencil partitioning.  This is a partitioning space operation,
    *  not an index space operation.
    */
   public final boolean inPartitioningBounds(IndexSpaceID pid){
   
      return this.partitioning.inBounds(pid);
   }

   /** Return true if the argument represents a partition coordinate that is 
    *  within the bounds of the stencil partitioning.  This is a partitioning space operation,
    *  not an index space operation.
    */
   public final boolean inPartitioningBounds(int [] coords){
   
      return this.partitioning.inBounds(coords);
   }
   
   /** Computes the volume of the stencil partitioning space. i.e. the total
    *  number of elements.
    */
   public long volume()
   {
      return this.partitioning.volume();
   }


   /** Return the volume of a single element in the stencil partitioning.
    */
   public long partitionVolume()
   {

      return this.partitioning.partitionVolume();
   }


   /** return the stencil's partitioning object. */
   ISPartitioning getPartitioning(){

     return this.partitioning;
   }

   /** Return a String representing the object. */
   public String toString ()
   {
      
      return this.toString ("");
   }


   public String toString(String indent){
   
      String indent2=indent+Globals.toStringIndent;
      
      StringBuffer result = new StringBuffer(indent + "{ DSPartitionedStencil:\n");
      //result.append(super.toString(indent2)+"\n");
      
      result.append(this.partitioning.toString(indent2));
      result.append("\n");
      
      result.append(indent2+"partitionListBytesLength="+partitionListBytesLength+"\n");
      
      
      result.append(indent + "}");
      
      return new String(result);
   
   }
   
   
}
