/*
 * @author rdb
 *
 * Created on Nov 25, 2004 by rdb
 *   This is intended to be a class parent to DataSource and DataBlock
 * so that implementation of utility convenience functions can be shared.
 * For example, 
 *       float[] getFloats( isid, attrIndex );
 * returns a single 1-d array of all of the values for a particular attr.
 * This is very efficient if the data is stored in this way, but less so
 * if it is not. However, we want all DataSources and DataBlocks to have 
 * the same interface, so we can put generic solutions here which are 
 * overridden in the children that can support a much more efficient
 * implementation.
 * 
 * Since Indexable is an interface, of course, we can't put code there. 
 * 
 * ------------------------- Edit history ----------------------------
 * 02/10/12 rdb: Made getUpperBoundsArray public (getLowerBoundsArray already was
 * 08/05/09 jpf: Added method for retrieving attribute names
 * 03/26/09 jpf: removed final from setAxisOrdering in order to override
 *               the method in BinaryFileDataSource
 * 08/25/08 kjr: Changed datum(Datum, IndexSpaceID) to account for non-zero
 *                lower bounds
 * 01/09/08 pjr: Rewrote subblock(block, bounds) to directly call subBasicBlock
 *                when appropriate. Added subBasicBlock() implementation that should
 *                be overridden by child classes.
 * 01/10/06 rdb: Changed lots of copyBounds calls to getBounds. Assume that subblock
 *               methods don't modify the bounds.
 * 07/15/05 rdb: made getValues() abstract
 * 03/31/05 rdb: made getLowerArray and getUpperArray public access
 * 03/31/05 rdb: made volume() NOT final
 * 02/20/05 rdb: updated RecordSpec to FieldIDMapper mapping
 * 02/07/05 rdb: getBounds() -> copyBounds()
 *               getBoundsReference() -> getBounds()
 * 02/01/05 rdb: Added reference versions of all array return methods
 * 01/30/05 rdb: Added subblock( DataBlock, ISBounds src )
 *                     subblock( DataBlock, ISBounds src, FieldIDMapper )
 *                     subblock( DataBlock, ISBounds src, RecordSpec )
 *                     subblock( DataBlock, RecordSpec )
 *                     subblock( DataBlock, FieldIDMapper )
 *                 and their implementations.
 * 
 * 01/26/05 rdb: Added getStorageType()
 * 01/25/05 rdb: Moved DataBlock subblock( ... ) implementations to here from
 *                children.
 * 01/24/05 rdb: Added get<type>sByAttribute
 *                     get<type>sByPoint
 *                     getValues( int dataOrder )
 *                     getValuesForAttribute( int attr )
 *                     getDataOrder()
 *                     setDataOrder()
 *                     
 *               Defined term DataOrder to describe attribute vs point order. It 
 *                 refers to the "storage" order for a physical data source or a
 *                 a data block that has a uniform storage order; otherwise it
 *                 refers to the order that is most likely to provide more efficient
 *                 access. 
 * 01/12/05 rdb Defined 6 createDatum methods; just pass through to RecordDescriptor.
 * 
 * 12/05/04 rdb Moved common stuff from DataBlock, DataSource, BasicBlock, 
 *                    PhysicalDataSource (and more?) to here
 * 12/03/04 rdb Added getAxisOrdering, so available for all DS and DataBlocks
 *              Added setAxisOrdering and a default implementation that almost
 *                    certainly should be overridden by children OR the implementation
 *                    needs to be changed to included computed variables such as is
 *                    generated in PhysicalDataSource.
 * 11/27/04 rdb Added notion of getValues( attr ) and getValues()
 *              Makes default implementations of get<type>  much more efficient, 
 *              with little additional effort on part of child classes.
 * 11/25/04 rdb Created. Preliminary implementation -- only implementing
 * 				the new functions. Should eventually move common code from
 *               DataSource and DataBlock up to here: 
 * 					dim()
 * 					getBounds()
 * 					getNumAttributes()
 * 					getRecordDescriptor()
 */
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

public abstract class DataCollection implements Indexable
{
   //--------------------- Data members: --------------------------------------
   //     11/27/04 note: There are lots of things common to DataBlock and
   //                    DataSource that should be moved here, like dim, bounds, etc.
   //                    That should happen sometime.

   /** Denotes extent of Index Space */
   protected    ISBounds    bounds; 
    
   /** Describes fields of Datums returned by this DataSource. */
   protected   RecordDescriptor    recordDesc;

   /** This value is taken from this.bounds, but is duplicated here for speed.*/
   protected int dimensionality;
   
   /** This value is taken from this.recordDesc, but is duplicated here for speed.*/
   protected int numAttributes;
   

   /** Defines the preferred data order -- either by attribute or by point. In 
    *  normal cases, this reflects the actual storage axisOrdering, but for some 
    *  complex formats, it's what is likely to provide most efficient mass
    *  data access. This is especially an issue for the methods:
    *  
    *     getDoubles(), getFloats(), getInts(), getShorts(), getBytes()
    */
   protected int dataOrder = pointOrder;
   
   public static final int pointOrder     = 0;
   public static final int attributeOrder = 1;

   // AxisOrdering variables 
   /** AxisOrdering describes how axis indexes map to the physical data storage */
   protected  AxisOrdering  axisOrdering    = null;
   /** axisWeights define stride values to find the neighbor in index space in a
    *  given axis.
    */
   protected  long[]        axisWeights = null;  
   
      
   //---------------------- Constructors -------------------------------------
   /** For now, prevent construction without knowledge of bounds, exc by children  */
   protected DataCollection() {}
   
   /** Create a DataCollection with given bounds and record descriptor. */
   protected DataCollection( ISBounds inBounds, RecordDescriptor rd )
   {
      this.bounds = new ISBounds (inBounds);
      this.dimensionality = this.bounds.dim();
      setRecordDescriptor( rd );
      setAxisOrdering( null );  // generate axis weights
   }

   /** Create a DataCollection with given bounds, record descriptor, and axisordering */
   protected DataCollection( ISBounds inBounds, RecordDescriptor rd, AxisOrdering ordering )
   {
      this.bounds = new ISBounds (inBounds);
      this.dimensionality = this.bounds.dim();
      setRecordDescriptor( rd );
      setAxisOrdering( ordering );  // generate axis weights
   }
   
   
   //----------------------- datum access methods -----------------------------
   /** Given an IndexSpaceID, return the corresponding datum.*/
   public  Datum datum( IndexSpaceID did )
   {
      Datum newDatum = this.createDatum();
      datum( newDatum, did );
      return newDatum;
   }
   
   /** Given an IndexSpaceID, return the corresponding datum with attributes
     * specified in the RecordSpec. */
   public  Datum datum( IndexSpaceID did, RecordSpec spec )
   {
      return datum( did, spec.getFieldIDMapper() );
   }

   /** Given an IndexSpaceID, return the corresponding datum with attributes
    * specified in the FieldIDMapper. 
    */
   public  Datum datum( IndexSpaceID did, FieldIDMapper fmap )
   {
      Datum newDatum = this.createDatum( fmap );
      datum( newDatum, did, fmap );
      return newDatum;
   }

   /** Given an IndexSpaceID, copy the values of the Datum at that position to the 
    *  argument Datum. 
    * TODO 
    * Define error conditions
    */
//   public abstract void datum( Datum dtm, IndexSpaceID did );   
   
//  TODO: Eventually, we should modify Granite so that for all datum and subblock
//  methods, there is a public method that does argument validation. These methods
//  then call the corresponding pdatum or psubblock method, which performs no checking
//  and does the actual access. These methods are package access only.Code inside 
//  the granite packages will generally call the package version only. This avoids
//  a lot of unnecessary checking of arguments.  -PJR   
//
   /** Given an IndexSpaceID, copy the values of the Datum at that position to the 
    *  argument Datum. 
    * TODO 
    * Define error conditions
    */
   public void datum( Datum dtm, IndexSpaceID did ){
   
      if (dtm.size() != this.getNumAttributes()){
         new IllegalArgumentError(SDBError.Recoverable,
         "Datum is not the right size for this datasource.");
         return;
      }
      
      // 8/25/08 kjr: changed to account for off-center bounds
      // pdatum(dtm, did);
      IndexSpaceID shifted = (IndexSpaceID) did.clone();
      did.subtract(shifted, getBounds().getLower());
      pdatum(dtm, shifted);
   }


   /** Given an IndexSpaceID, copy the values of the Datum at that position to the 
    *  argument Datum. This is a package access only method, and does no checking
    *  of the arguments.
    */
    void pdatum( Datum dtm, IndexSpaceID did ){
    
      new SDBError(SDBError.Fatal, "Not implemented");
    }



   /** Given an IndexSpaceID and FieldIDMapper, copy the values of the Datum at 
    *  that position to the argument Datum. 
    * TODO Define error conditions
    */
   public abstract void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp );

      
   /** Given an IndexSpaceID and RecordSpec, copy the values of the Datum at 
    *  that position to the argument Datum.
    *  
    * TODO Define error conditions
    * 
    * Definition below as in DataSource. 
    * rdb: I thought RecordSpec "selected" fields from source, not from
    *      the destination. 
    * PBR: Return by reference the Datum corresponding to the given 
    *  IndexSpaceID. Only datum fields referred to by the RecordSpec
    *  are modified.
    *
    */
   public void datum( Datum dtm, IndexSpaceID did, RecordSpec recSpec )
   {
       datum( dtm, did, recSpec.getFieldIDMapper() );
   }

   /** Copy the Datum argument into the position indicated by the values in
    *  the int coordArray interpreted as an IndexSpaceID. Some children may
    *  find it convenient to override this to avoid the construction of the
    *  IndexSpaceID. 
    */
   public void datum( Datum d, int[] coordArray )
   {
      datum( d, new IntegerIndexSpaceID( coordArray ));
   }
   
   //----------------------- subblock access methods ---------------------------
   /** Return a new DataBlock that is a subset of this DataCollection.
    *  The data may be shared. 
    */
   public  DataBlock subblock( ISBounds srcBounds )
   {
      DataBlock dstBlock = this.createDataBlock( srcBounds.copyToOrigin());
      this.subblock( dstBlock, dstBlock.getBounds(), srcBounds);
      return dstBlock;            
   }
   
   /** Return a new DataBlock that is an attribute projection of this DataCollection 
    * based on the attributes as defined by the RecordSpec. It is legal to 
    * pass a null for the RecordSpec argument which means that all attributes are
    * to be extracted.
    */
   public  DataBlock subblock( RecordSpec rs )
   {
      DataBlock dstBlock = this.createDataBlock( this.bounds, rs );
      this.subblock( dstBlock, this.bounds, rs.getFieldIDMapper() );
      return dstBlock;            
   }
   
   /** Return a new DataBlock that is an attribute projection of this DataCollection 
    * based on the attributes as defined by the FieldIDMapper. It is legal to 
    * pass a null for the FieldIDMapper argument which means that all attributes are
    * to be extracted.
    */
   public  DataBlock subblock( FieldIDMapper fmap )
   {
      DataBlock dstBlock = this.createDataBlock( this.bounds, fmap );
      this.subblock( dstBlock, this.bounds, fmap );
      return dstBlock;            
   }
   
   /** Return a new DataBlock that is a subset of this DataCollection based
    *  on bounds and selective attributes as defined by the RecordSpec. It is legal to 
    * pass a null for the RecordSpec argument which means that all attributes are
    * to be extracted.
    */
   public  DataBlock subblock( ISBounds srcBounds, RecordSpec rs )
   {
      DataBlock dstBlock = this.createDataBlock( srcBounds.copyToOrigin(), rs );
      this.subblock( dstBlock, dstBlock.getBounds(), srcBounds, rs );
      return dstBlock;            
   }
   
   /** Return a new DataBlock that is a subset of this datablock based on bounds
    * and selective attributes as defined by the FieldIDMapper. It is legal to 
    * pass a null for the FieldIDMapper argument which means that all attributes are
    * to be extracted.
    */
   public  DataBlock subblock( ISBounds srcBounds, FieldIDMapper fmap )
   {
      DataBlock dstBlock = this.createDataBlock( srcBounds.copyToOrigin(), fmap );
      this.subblock( dstBlock, dstBlock.getBounds(), srcBounds, fmap );
      return dstBlock;            
   }
    
   
   /** Reads a subblock defined by srcBounds into the region of the Datablock argument
    defined by dstBounds. */
   public abstract void subblock( DataBlock theBlock, 
                                  ISBounds dstBounds, ISBounds srcBounds );
   
   /** Reads a subblock defined by srcBounds into the region of the Datablock argument
    * defined by dstBounds. The FieldIDMapper argument determines the mapping between 
    * datums in this DataBlock and datums in the argument DataBlock. It is legal to 
    * pass a null for the FieldIDMapper argument which means that all attributes are
    * to be extracted.
    */      
   public abstract void subblock( DataBlock theBlock, 
						         ISBounds dstBounds, ISBounds srcBounds, 
						         FieldIDMapper mp );
   /******
   {
      System.err.println( "DataCollection.subblock( DataBlock, ISB, ISB, fim )" +
            " should be abstract and should not be called." );
   };
   /***************************************/
   
   /** Reads a subblock defined by srcBounds into the Datablock argument. If the
    *  size and shape of the DataBlock argument does not match the source bounds,
    *  the source bounds will be modified to match the destination shape and size,
    *  if there is enough data in the source; otherwise, only the portion of the 
    *  destination is changed for which there is source data.
    */
   public void subblock( DataBlock theBlock, ISBounds srcBounds )
   {
      if (theBlock instanceof BasicBlock) {
         subBasicBlock((BasicBlock)theBlock, theBlock.getBounds(), srcBounds);
      } else {  
         subblock( theBlock, srcBounds, (FieldIDMapper)null );
      }
   }

   /** should be overridden by child classes for efficiency.*/
   void subBasicBlock(BasicBlock theBlock, ISBounds dstBounds, ISBounds srcBounds){
      subblock(theBlock, dstBounds, srcBounds);
   }
   
   /** Reads a subblock defined by srcBounds into the Datablock argument. If the
    *  size and shape of the DataBlock argument does not match the source bounds,
    *  the source bounds will be modified to match the destination shape and size,
    *  if there is enough data in the source; otherwise, only the portion of the 
    *  destination is changed for which there is source data. Only the attributes
    *  identified by the RecordSpec are included. It is valid for the RecordSpec
    *  argument to be null; this specifies that all attributes are to be extracted.
    */
   public void subblock( DataBlock theBlock, ISBounds srcBounds, RecordSpec rs )
   {
      subblock( theBlock, srcBounds, rs.getFieldIDMapper() );
   }
   
   /** Reads a subblock defined by srcBounds into the Datablock argument. If the
    *  size and shape of the DataBlock argument does not match the source bounds,
    *  the source bounds will be modified to match the destination shape and size,
    *  if there is enough data in the source; otherwise, only the portion of the 
    *  destination is changed for which there is source data. Only the attributes
    *  identified by the FieldIDMapper are included. It is valid for the FieldIDMapper
    *  argument to be null; this specifies that all attributes are to be extracted.
    */
   public void subblock( DataBlock theBlock, ISBounds srcBounds, FieldIDMapper fmap )
   {
      ISBounds dstBnds = theBlock.getBounds();
      if ( srcBounds.equalShape( dstBnds ))
         this.subblock( theBlock, dstBnds, srcBounds, fmap );
      else
      {
         ISBounds newSrcBnds = dstBnds.copyTo( srcBounds.copyLower() );
         if ( !this.bounds.contains( newSrcBnds ))
            newSrcBnds = this.bounds.intersection( newSrcBnds );
         this.subblock( theBlock, dstBnds, newSrcBnds, fmap );
     }
   }

   /** Reads a subblock defined by srcBounds into the region of the Datablock argument
    * defined by dstBounds. The RecordSpec argument specifies a subset of the attributes
    * of this DataSource to be written into the DataBlock datums. It is legal to 
    * pass a null for the RecordSpec argument which means that all attributes are
    * to be extracted.
    */
   public void subblock( DataBlock theBlock, 
                         ISBounds dstBounds, ISBounds srcBounds, 
                         RecordSpec spec )
   {
      this.subblock( theBlock, dstBounds, srcBounds, spec.getFieldIDMapper() );
   }
   
//   /** Fill the datablocks in the DataBlockCollection with data read from
//    *  the bounds specified in the stencil. This method is meant to
//    *  save latency costs across a network by filling several datablocks
//    *  with a single query.  
//    *  @deprecated
//    */
//   public void subblock(DSStencil stencil, DataBlockCollection dstBlocks) {
//      
//      ISBounds sBounds = new ISBounds(this.dimensionality);
//            
//      int size = stencil.size();
//      
//      for( int i=0; i<size; i++){
//      
//         this.subblock(dstBlocks.get(i), stencil.getIndexedBounds(sBounds, i));
//      }
//      
//   }

   /** Fill the datablocks specified as "active" in the DSBlockStencil. 
    *  This method is meant to save latency costs across a network by filling
    *  several datablocks with a single query.  
    */
   public void subblock(DSBlockStencil stencil) {
      
      ISBounds sBounds = new ISBounds(this.dimensionality);
      DataBlock db;
            
      int size = stencil.size();
      
      for( int i=0; i<size; i++){
      
         db = stencil.getIndexedBlockAndBounds(sBounds, i);
      
         this.subblock(db, sBounds);
      }
      
   }



   //++++++++++++++++++++++ base methods ++++++++++++++++++++++++++++++++++
   // should work for children
   
   /**
    * Returns the dimensionality of the Index Space for this DataSource.
    */
   public final int dim()
   {
      return this.dimensionality;
   }

   /**
    * Returns a copy of the bounds of the Index Space for this DataSource.
    *
    * Can make deprecated to find out who is using to make sure they shouldn't
    * be using getBounds() (the reference version).
    */
   public final ISBounds copyBounds()
   {
      return (ISBounds) this.bounds.clone();
   }

   /**  Return a reference to the ISBounds for this DataBlock. */
   public final ISBounds getBounds()
   {
       return bounds;
   }

   /** Return a reference to an array containing 
     * the dimensions of this data collection. 
     */
   public final int[] getDimensionsArray()
   {
       return this.bounds.getDimensionsArray();
   }

   /** Return a reference to an array containing 
     * the lower bounds of this collection. 
     */
   public final int[] getLowerBoundsArray()
   {
       return this.bounds.getLowerArray();
   }

   /**  Return a reference to an array containing 
     * the upper bounds of this collection. 
     */
   public final int[] getUpperBoundsArray()
   {
       return this.bounds.getUpperArray();
   }

   /** Returns the data storage order. Normally data will be stored 
    * either in attribute order (all values of same attribute are contiguous) 
    * or in point order (all values of a single point are contiguous). 
    * In a physical file, arbitrary combinations are possible although we
    * currently only support attribute order or point order. Even if we 
    * later support other file orderings, we might want to retain the 
    * "view" of everything internally to be one of these two.
    * 
    * The value returned will be one of:
    *    DataCollection.pointOrder
    *    DataCollection.attributeOrder
    */
   final public int getDataOrder()
   {
      return this.dataOrder;
   }

   /** Sets the attribute storage order. 
    * 
    * @param int attrDataOrder. Must be one of
    *    DataCollection.pointOrder
    *    DataCollection.attributeOrder
    * @return void
    */
   final void setDataOrder( int storageOrder )
   {
      switch ( storageOrder )
      {
         case DataCollection.pointOrder:
         case DataCollection.attributeOrder:
              this.dataOrder = storageOrder;
              break;
         default:
            new SDBError( SDBError.Recoverable, 
                  "Invalid attribute storage order value.");
      }
   }
   
   /** Maps an n-dimensional IndexSpaceID to a 1d offset.*/
   protected final int computeOffset( IndexSpaceID dsid )
   {   
      return computeOffset( dsid.refCoords() );
   }

   /** Maps an n-dimensional array of coordinates to a 1d offset*/
   protected final int computeOffset( int coordArray[] )
   {
       int offset = 0;
       try {
           for ( int i=0; i < this.dimensionality; i++ )
               offset += coordArray[i] * axisWeights[i];
           return offset;
       } catch ( ArrayIndexOutOfBoundsException e )
	       {
	           new DimensionalityError(SDBError.Fatal,
	                  "computeOffset argument has the wrong dimensionality.");
	           return 0;
	       }
   }

   // rdb: why both size and volume?
   //      size cannot be final. Overridden by AR classes and VRDataBlock
   /** Return the number of datums . 
    *  @deprecated
    *  Use volume() method.
    */
   public int size()
   {
      return this.bounds.volume();
   }

   /** Returns the number of elements in this datasource index space. */
   public int volume()
   {
      try {
         return this.bounds.volume();
      } catch ( SDBException ex )
      {
         System.err.println( "DataCollection.volume(): "+ ex.getMessage());
      }
      return 0;
   }

   /** Returns a long with the number of elements in this datasource index space. */
   public final long longVolume()
   {
      return this.bounds.longVolume();
   }

   /**
    * Sets the bounds of the Index Space for this DataSource. This method is
    * meant to be called by the constructors of certain child classes, e.g.
    * BlockedDataSource. If a null reference is passed, then an
    * IllegalArgumentException is thrown.
    */
   protected final void setBounds( ISBounds bounds )
   {
      if ( bounds != null )
      {
         this.bounds = (ISBounds) bounds.clone();
         this.dimensionality = this.bounds.copyLower().dim();
         this.axisWeights = this.axisOrdering.makeAxisWeights( this.bounds );
      }
      else
      {
         new IllegalArgumentError( SDBError.Recoverable,
               "Datasource.setBounds() was passed a null reference" );
      }
   }

   /** return the number of attributes at each position of the index space.*/
   public final int getNumAttributes()
   {
       return this.numAttributes;
      // return this.recordDesc.size();
   }

   /** return the name of each attribute */
   public final String[] getAttributeNames()
   {
       return (this.recordDesc.getFieldNames());
   }



   
   /** return the storage type for this collection */
   public final int getStorageType()
   {
       return this.recordDesc.getStorageType();
   }
   
   /** Returns a copy of the RecordDescriptor*/
   public final RecordDescriptor getRecordDescriptor()
   {
       return recordDesc.copy();
   }
   
   /** Return the number of bytes used to represent a datum.*/
   public final int byteSize()
   {
       return recordDesc.byteSize();
   }

   /** When this method is called, this subblock methods of this collection will 
    *  read data as bytes, without performing any type conversion. This is meant 
    *  to be used by servers, since they send raw bytes over the network anyway,
    *  and any conversion is wasted effort.
    */
   public void useBytesOnly(){
   
      this.recordDesc=new BytesOnlyRecordDescriptor(this.recordDesc);
      this.numAttributes = this.recordDesc.size();
   }
   
   
   
   
   /** Set the DataSource RecordDescriptor to a copy
       of the argument. This method is meant to be called
       by the constructors of child classes,
       e.g. AttributeJoinDataSource. If a null reference is
       passed, then an IllegalArgumentException is thrown.
    */
   protected final void setRecordDescriptor( RecordDescriptor r )
   {
      if ( r!= null )
      {
         this.recordDesc = r.copy();
         this.numAttributes = recordDesc.size();
      } 
      else 
         new IllegalArgumentError(SDBError.Recoverable, 
               "Datasource.setRecordDescriptor() was passed a null reference");
   } 

   /** Return a copy of the AxisOrdering .*/
   public AxisOrdering getAxisOrdering()
   {
       return (AxisOrdering) this.axisOrdering.clone();
   }

   /** Set a new the AxisOrdering for this datasource.
    *  Note that this operation will have no affect on previously constructed
    *  Iterators or DataBlocks constructed for this DataSource. Note that children
    *  may need to override if there is something more than the axisWeights
    *  that needs to be recomputed.
    */
   public void setAxisOrdering( AxisOrdering newOrder )
   {
      if ( newOrder == null )
         this.axisOrdering = new AxisOrdering( this.dimensionality );
      else
         this.axisOrdering = newOrder.copy();
      this.axisWeights = this.axisOrdering.makeAxisWeights( this.bounds );
   }
   
   /** Return the Weight of the specified axis. */
   public final long getAxisWeight(int axis)
   {
       return this.axisWeights[axis];
   }

   /** Return the Weight of the specified axis. */
   public final long [] getAxisWeights()
   {
       return this.axisWeights.clone();       
   }

   //+++++++++++++++++++++++++ createDatum methods +++++++++++++++++++++++++++
   // All these methods just pass on the responsibility to the RecordDescriptor
   //
   /** create a Datum matching RecordDescriptor. */
   public final Datum createDatum() 
                                 { return this.recordDesc.createDatum(); }
   /** create a Datum matching RecordDescriptor, adjusted by RecordSpec. */
   public final Datum createDatum( RecordSpec rs ) 
                                 { return this.recordDesc.createDatum( rs ); }
   /** create a Datum matching RecordDescriptor, adjusted by FieldIDMapper. */
   public final Datum createDatum( FieldIDMapper fm ) 
                                 { return this.recordDesc.createDatum( fm ); }
   /** create a Datum matching RecordDescriptor, initialized to Datum value. */
   public final Datum createDatum( Datum d ) 
                                 { return this.recordDesc.createDatum( d ); }
   /** create a Datum matching RecordDescriptor, based or RecordSpec, init to Datum. */
   public final Datum createDatum( Datum d, RecordSpec rs ) 
                                 { return this.recordDesc.createDatum( d, rs ); }
   /** create a Datum matching RecordDescriptor, based on FieldIDMapper, init to Datum. */
   public final Datum createDatum( Datum d, FieldIDMapper fm ) 
                                 { return this.recordDesc.createDatum( d, fm ); }
 
   //+++++++++++++++++++++++++ createBlock methods +++++++++++++++++++++++++++
   // All these methods just pass on the responsibility to the RecordDescriptor
   //
   /** create a DataBlock matching RecordDescriptor. */
   public final DataBlock createDataBlock( ISBounds bnds ) 
                         { return this.recordDesc.createDataBlock( bnds ); }
   /** create a DataBlock matching RecordDescriptor, adjusted by RecordSpec. */
   public final DataBlock createDataBlock( ISBounds bnds, RecordSpec rs ) 
                         { return this.recordDesc.createDataBlock( bnds, rs ); }
   /** create a DataBlock matching RecordDescr adjusted by FieldIDMapper. */
   public final DataBlock createDataBlock( ISBounds bnds, FieldIDMapper fm ) 
                         { return this.recordDesc.createDataBlock( bnds, fm ); }
   /** create a DataBlock matching RecordDescriptor, initialized to value. */
   //public final DataBlock createDataBlock( ISBounds bnds, DataBlock blk ) 
   //                   { return this.recordDesc.createDataBlock( bnds, blk ); }
   /** create a DataBlock matching RecordDescriptor, based or RecordSpec, init to Datum. */
   //public final DataBlock createDataBlock( ISBounds bnds, DataBlock blk, RecordSpec rs ) 
   //                              { return this.recordDesc.createDataBlock( bnds, blk, rs ); }
   /** create a DataBlock matching RecordDescriptor, based on FieldIDMapper, init to Datum. */
   //public final DataBlock createDataBlock( ISBounds bnds, DataBlock blk, FieldIDMapper fm ) 
   //                              { return this.recordDesc.createDataBlock( bnds, blk, fm ); }
 
   //+++++++++++++++++++++++++ template methods +++++++++++++++++++++++++++
   // Utilize child specific implementations.
   // Some of these may be overridden by some children for efficiency reasons.
   
   //---------------- get<type> array methods ------------------
   // These are "template methods". They utilize child-specific methods
   // but they work for children -- however, for performance reasons the
   // child classes might want to override.
   
   //----------------------- attribute access methods -----------------------------
   /** Return the field specified by the arguments as a double.*/
   public double getDouble( IndexSpaceID isid, int attr )
   {
      return this.datum( isid ).getDouble( attr );
   }

	/** Return the field specified by the arguments as a float.*/
   public float getFloat( IndexSpaceID isid, int attr )
   {
      return this.datum( isid ).getFloat( attr );
   }

	/** Return the field specified by the arguments as an int.*/
   public int getInt( IndexSpaceID isid, int attr )
   {
      return this.datum( isid ).getInt( attr );
   }

	/** Return the field specified by the arguments as a short.*/
   public short getShort( IndexSpaceID isid, int attr )
   {
      return this.datum( isid ).getShort( attr );
   }
		    
	/** Return the field specified by the arguments as a byte.*/
   public byte getByte( IndexSpaceID isid, int attr )
   {
      return this.datum( isid ).getByte( attr );
   }

   //----------------------- get<type>( ISID ) -----------------------------
   // for a given ISID get all attributes in point order as a given type.
   //
   /** Return all attributes of the Datum at the as a 1-d double array */
   public double[] getDoubles( IndexSpaceID isid )
   {
      return this.datum( isid ).getDoubles( );
   }
 
   /** Reference version of getDoubles( isid )
    *  The passed array is returned for convenience, but if it is
    *  null or not the right length a new array is created and returned. */
   public double[] getDoubles( IndexSpaceID isid, double[] fill )
   {
      return this.datum( isid ).getDoubles( fill );
   }

   /** Return all attributes of the Datum as a 1-d float array */
   public float[] getFloats( IndexSpaceID isid )
   {
      return this.datum( isid ).getFloats( );
   }
   
   /** Reference version of getFloats( isid )
    *  The passed array is returned for convenience, but if it is
    *  null or not the right length a new array is created and returned. */
   public float[] getFloats( IndexSpaceID isid, float[] fill )
   {
      return this.datum( isid ).getFloats( fill );
   }

   /** Return all attributes of the Datum as a 1-d int array */
   public int[] getInts( IndexSpaceID isid )
   {
      return this.datum( isid ).getInts( );
   }
   
   /** Reference version of getInts( isid )
    *  The passed array is returned for convenience, but if it is
    *  null or not the right length a new array is created and returned. */
   public int[] getInts( IndexSpaceID isid, int[] fill )
   {
      return this.datum( isid ).getInts( fill );
   }

   /** Return all attributes of the Datum as a 1-d short array */
   public short[] getShorts( IndexSpaceID isid )
   {
      return this.datum( isid ).getShorts( );
   }
   
   /** Reference version of getShorts( isid )
    *  The passed array is returned for convenience, but if it is
    *  null or not the right length a new array is created and returned. */
   public short[] getShorts( IndexSpaceID isid, short[] fill )
   {
      return this.datum( isid ).getShorts( fill );
   }

   /** Return all attributes of the Datum as a single byte array */
   public byte[] getBytes( IndexSpaceID isid )
   {
      return this.datum( isid ).getBytes( );
   }

   /** Reference version of getBytes( isid )
    *  The passed array is returned for convenience, but if it is
    *  null or not the right length a new array is created and returned. */
   public byte[] getBytes( IndexSpaceID isid, byte[] fill )
   {
      return this.datum( isid ).getBytes( fill );
   }

   //----------------------- get<type>( int ) -----------------------------
   // for a given attribute get all values from the collection in a single
   // 1D array -- 
   //
   /** Return a 1-d double array containing all values of the specified 
    *    attribute using the existing axis axisOrdering. */
   public double[] getDoubles( int attr )
   {
      return getValues( attr ).getDoubles();
   }
   
   /** Fill the passed 1-d double array with all values of the specified 
    *    attribute using the existing axis axisOrdering. The passed array is
    *    returned, but if it is null or not the right size a new array 
    *    is created and returned. 
    */
   public double[] getDoubles( int attr, double[] fill )
   {
      return getValues( attr ).getDoubles( fill );
   }
   
   /** Return a 1-d float array containing all values of the specified 
    *    attribute using the existing axis axisOrdering. */
   public float[] getFloats( int attr )
   {
      return getValues( attr ).getFloats();
   }
   
   /** Fill the passed 1-d float array with all values of the specified 
    *    attribute using the existing axis axisOrdering. The passed array is
    *    returned, but if it is null or not the right size a new array 
    *    is created and returned. 
    */
   public float[] getFloats( int attr, float[] fill )
   {
      return getValues( attr ).getFloats( fill );
   }
   
   /** Return a 1-d int array containing all values of the specified 
    *    attribute using the existing axis axisOrdering. */
   public int[] getInts( int attr )
   {
      return getValues( attr ).getInts();
   }
   
   /** Fill the passed 1-d int array with all values of the specified 
    *    attribute using the existing axis axisOrdering. The passed array is
    *    returned, but if it is null or not the right size a new array 
    *    is created and returned. 
    */
   public int[] getInts( int attr, int[] fill )
   {
      return getValues( attr ).getInts( fill );
   }
   
   /** Return a 1-d short array containing all values of the specified
    *    attribute using the existing axis axisOrdering. */
   public short[] getShorts( int attr )
   {
      return getValues( attr ).getShorts();
   }
   
   /** Fill the passed 1-d short array with all values of the specified 
    *    attribute using the existing axis axisOrdering. The passed array is
    *    returned, but if it is null or not the right size a new array 
    *    is created and returned. 
    */
   public short[] getShorts( int attr, short[] fill )
   {
      return getValues( attr ).getShorts( fill );
   }
   
   /** Return a 1-d byte array containing all values of the specified
    *    attribute using the existing axis axisOrdering. */
   public byte[] getBytes( int attr )
   {
      return getValues( attr ).getBytes();
   }

   /** Fill the passed 1-d byte array with all values of the specified 
    *    attribute using the existing axis axisOrdering. The passed array is
    *    returned, but if it is null or not the right size a new array 
    *    is created and returned. 
    */
   public byte[] getBytes( int attr, byte[] fill )
   {
      return getValues( attr ).getBytes( fill );
   }
   
   //----------------------- get<type>s( ) -----------------------------
   // for all attributes get all values from the collection in a single
   // 1D array -- 
   //   
   /** Return a 1-d double array containing all values of all 
    *    attributes using the existing axis axisOrdering and data axisOrdering
    *    (attribute vs. point). 
    *  The returned array may be a reference to the actual data.
    */
   public double[] getDoubles()
   {
      return this.getValues().getDoubles();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public double[] getDoubles( double[] filler )
   {
      return this.getValues().getDoubles( filler );
   }
   
   /** Return a 1-d double array containing all values of all 
    *    attributes in attribute order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public double[] getDoublesByAttribute()
   {
      return this.getValuesByAttribute().getDoubles();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public double[] getDoublesByAttribute( double[] filler )
   {
      return this.getValuesByAttribute().getDoubles( filler );
   }
   
   /** Return a 1-d double array containing all values of all 
    *    attributes in point order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public double[] getDoublesByPoint()
   {
      return this.getValuesByPoint().getDoubles();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and attribute data axisOrdering. The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public double[] getDoublesByPoint( double[] filler )
   {
      return this.getValuesByPoint().getDoubles( filler );
   }
   
   /** Return a 1-d float array containing all values of all 
    *    attributes using the existing axis axisOrdering and data axisOrdering
    *    (attribute vs. point). 
    *  The returned array may be a reference to the actual data.
    */
   public float[] getFloats()
   {
      return this.getValues().getFloats();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
    public float[] getFloats (float[] filler )
   {
      return this.getValues().getFloats( filler );
   }
   
   /** Return a 1-d float array containing all values of all 
    *    attributes in attribute order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public float[] getFloatsByAttribute()
   {
      return this.getValuesByAttribute().getFloats();
   }
   
   /** Return a 1-d float array containing all values of all 
    *    attributes in point order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public float[] getFloatsByPoint()
   {
      return this.getValuesByPoint().getFloats();
   }  
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public float[] getFloatsByAttribute( float[] filler )
   {
      return this.getValuesByAttribute().getFloats( filler );
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and attribute data axisOrdering. The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public float[] getFloatsByPoint( float[] filler )
   {
      return this.getValuesByPoint().getFloats( filler );
   }
   
   /** Return a 1-d int array containing all values of all 
    *    attributes using the existing axis axisOrdering and data axisOrdering
    *    (attribute vs. point). 
    *  The returned array may be a reference to the actual data.
    */
   public int[] getInts()
   {
      return this.getValues().getInts();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public int[] getInts( int[] filler )
   {
      return this.getValues().getInts( filler );
   }
   
   /** Return a 1-d int array containing all values of all 
    *    attributes in attribute order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public int[] getIntsByAttribute()
   {
      return this.getValuesByAttribute().getInts();
   }
   
   /** Return a 1-d int array containing all values of all 
    *    attributes in point order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public int[] getIntsByPoint()
   {
      return this.getValuesByPoint().getInts();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public int[] getIntsByAttribute( int[] filler )
   {
      return this.getValuesByAttribute().getInts( filler );
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and attribute data axisOrdering. The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public int[] getIntsByPoint( int[] filler )
   {
      return this.getValuesByPoint().getInts( filler );
   }
   
   /** Return a 1-d short array containing all values of all 
    *    attributes using the existing axis axisOrdering and data axisOrdering
    *    (attribute vs. point). 
    *  The returned array may be a reference to the actual data.
    */
   public short[] getShorts()
   {
      return this.getValues().getShorts();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public short[] getShorts( short[] filler )
   {
      return this.getValues().getShorts( filler );
   }
   
   /** Return a 1-d short array containing all values of all 
    *    attributes in attribute order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public short[] getShortsByAttribute()
   {
      return this.getValuesByAttribute().getShorts();
   }
   
   /** Return a 1-d short array containing all values of all 
    *    attributes in point order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public short[] getShortsByPoint()
   {
      return this.getValuesByPoint().getShorts();
   }
     
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public short[] getShortsByAttribute( short[] filler )
   {
      return this.getValuesByAttribute().getShorts( filler );
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and attribute data axisOrdering. The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public short[] getShortsByPoint( short[] filler )
   {
      return this.getValuesByPoint().getShorts( filler );
   }
   
   /** Return a 1-d byte array containing all values of all 
    *    attributes using the existing axis axisOrdering and data axisOrdering
    *    (attribute vs. point). 
    *  The returned array may be a reference to the actual data.
    */
   public byte[] getBytes()
   {
      return this.getValues().getBytes();
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public byte[] getBytes( byte[] filler )
   {
      return this.getValues().getBytes( filler );
   }
   
   /** Return a 1-d byte array containing all values of all 
    *    attributes in attribute order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public byte[] getBytesByAttribute()
   {
      return this.getValuesByAttribute().getBytes();
   }
   
   /** Return a 1-d byte array containing all values of all 
    *    attributes in point order using the existing axis axisOrdering.
    *  The returned array may be a reference to the actual data. 
    */
   public byte[] getBytesByPoint()
   {
      return this.getValuesByPoint().getBytes();
   }
     
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and data axisOrdering (attribute vs. point). The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public byte[] getBytesByAttribute( byte[] filler )
   {
      return this.getValuesByAttribute().getBytes( filler );
   }
   
   /** Fill a 1-d array containing all values of all attributes using the 
    * existing axis axisOrdering and attribute data axisOrdering. The passed 
    * array is returned for convenience. If the passed array is null or is not long 
    * of correct length, a warning is issued, and a new array of correct length
    * is created and returned.
    */
   public byte[] getBytesByPoint( byte[] filler )
   {
      return this.getValuesByPoint().getBytes( filler );
   }
   
   /** Return a Values array of all attributes in whatever type and order
    *  they are stored internally (either attribute order or point order 
    *  for Granite-defined objects). The default implementation is often
    *  overridden for efficiency.
    */
   abstract public Values getValues();
   /***
   {
      return this.subblock( getBounds() ).getValues();
   } /**/
   
   /** Return a Values array of all attributes in their internal type in
    *  the attribute order. The returned result may be a reference to the
    *  actual data if it is possible, or a copy of the data reorganized
    *  to match the desired order. 
    */
    public Values getValuesByAttribute()
    { 
       if ( this.getDataOrder() == DataCollection.attributeOrder )
          return this.getValues();
       else // is by point, need by attr
          return this.getValues().reorder( this.getNumAttributes() );
    }
   
    /** Return a Values array of all attributes in their internal type in
     *  the point order. The returned result may be a reference to the
     *  actual data if it is possible, or a copy of the data reorganized
     *  to match the desired order. 
     */
     public Values getValuesByPoint()
     { 
        if ( this.getDataOrder() == DataCollection.pointOrder )
           return this.getValues();
        else    // is by attr, need by point
           return this.getValues().reorder( this.volume() );
     }
    
   /** Return a Values array of the specified attribute in whatever type
    *  it has internally. This could be a reference to the actual data.
    *  The implementation in this class is a default that is expected to
    *  be overridden in children classes for performance.
    */
    public Values getValues( int attr )
    { 
       if ( attr < 0 || attr >= getNumAttributes() )
       {
          new SDBError(SDBError.Recoverable, "getValues: invalid attribute index: " + attr  );
          return null;
       }

       RecordSpec rs   =  new RecordSpec( new int[] { attr });
       return this.subblock( this.bounds, rs ).getValues();
    }
}
