/* 
    FDLDescriptor.java
    Author:  Denise D. Mitchell 2002
    University of New Hampshire
    Durham, NH 03824
    
  ----------------- Edit history as of 1/23/05 --------------------------------
  02/12/12 rdb: Added blockSize field for SparseDataSource instead of using
                   chunking information field; also get/setBlockDimensions
  01/11/12 rdb: Added support for sparse data sources: block mapping
                Fixed writeFDL method to output the chunking information
                Deprecated setFileFormat( boolean ); it is implied by 
                    setChunkDimensions and can be turned off by passing null
                Changed isChunkedFileFormat to test for non-null chunkDimensions
  08/25/08 kjr: Added support for metadata (MetadataCollection).
                Added 'fdlFileName' to hold actual .fdl file name.
                Added methods to write data to files in xml form.
  06/05/08 rdb: Made Vectors generic
  01/31/06 dlipsa: Added support for creating chunked data sources through 
           an xfdl file
  11/28/05 rdb: Changed getDataFileURL so that it tests existence of raw name and
                  if it does not exist, we prepend the fdl file directory location 
                  and check for the existence of that file. This was done in 
                  DataSource.create, but there are other places that use the method.
                Added getDataFileString() to support simple retrieval of the URL string.
  06/01/05 rdb: fixed minor mistake in trying to locate .xfdl file
                added getRootDir and setRootDir
  05/31/05 rdb: added 'clone' method and Cloneable interface
  04/21/05 rdb: create method changes to allow fdl extension to be omitted
                in argument. Adds .xfdl, .fdl, .xml first; if all fails,
                tries to open without the extension -- but that is usually
                the data file itself.
  03/14/05 rdb: added static create( String fdlFileName )
                This seems a more logical place to do the work than DataSource
                Also added a variable to save and store the directory location
                of the fdl file and a method to extract it, getFDLRoot();
  01/23/05 rdb: Replaced PhysicalDataSource.pointOrder and attributeOrder constants
                whit DataCollection.pointOrder and attributeOrder
*/

package edu.unh.sdb.datasource;
import edu.unh.sdb.common.*;
//import java.io.*;
import java.util.*;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.*;

/**
 * The FDLDescriptor class holds data read from a file descriptor .fdl file
 * The data will be read, validated and properly instantiated for use by the
 * DataSource classes.
*/

 public class FDLDescriptor implements Cloneable
 {
    /** directory of fdl file used to create this descriptor (if created by file) */
    protected    String    fdlFileRoot = null;
    
    /** name of fdl file used to create this descriptor */
    protected    String    fdlFileName = null;
    
    /** Denotes extent of Index Space */
    protected    ISBounds    bounds;
    
    /** Denotes the chunk (partition) dimensions for ChunkedDataSource */
    protected IndexSpaceID chunkDimensions = null;

    /** Denotes the block (partition) dimensions for SparseDataSource */
    protected IndexSpaceID blockDimensions = null;

    /** Declares the file name (URL) where the data is located */
    protected String dataFileURL;

    /** Declares the file type */
    protected String dataFileType;

    /** Declares the record size **/
    protected int recordSize;
        
    /** declares boolean for point order **/
    //private boolean pointOrder;
    private int fieldLayout;
    
    /**
     * Specifies if the file is chunked (it is stored in cubes on disk instead of
     * being stored in rows and planes).
     */
    private boolean chunkedFileFormat = false;    

    /** Declares a vector to hold the field name strings **/
    protected Vector<String> fieldNames;
        
    /** Declares a vector to hold the field type strings**/
    protected Vector<String> fieldTypes;
    
    //added by Gang Lu
    /** Declares the axis axisOrdering */
    protected AxisOrdering axisOrdering;
    
    /** Declares the file offset */
    protected long fileOffset;

    // Added by PJR -- default endianness is big
    protected ByteOrder endianness = ByteOrder.BIG_ENDIAN;
    
    // Added by PJR -- For now, we only support one attribute when reading
    // from a Chombo file. This next field specifies that attribute.
    protected int chomboDesiredAttribute=0;

    
    /** Holds the metadata declared in this FDL*/
    private MetadataCollection metadata = null;

    /** Holds the information for mapping blocks in a sparse data source */
    protected String[] sparseMapData = null;
	
    
    /** Denotes the Sparse Data  bounds **/
    protected ISBounds sparseDataBounds;
    
    /** 
     * Default Constructor sets the bounds and recordDesc members to default values.
    */
    public FDLDescriptor()
    {
       bounds = null;
       dataFileURL = "";
       dataFileType = "unknown";

       fieldNames = new Vector<String>();
       fieldTypes = new Vector<String>();
       fieldLayout = DataCollection.pointOrder;
       axisOrdering = null;
       endianness = ByteOrder.BIG_ENDIAN;
       metadata = null;    
       sparseDataBounds = null;
    }
    
    /** 
     * Create an FDLDescriptor from a RecordDescriptor, Bounds, and AxisOrdering.
     * Intended for Composite datasource creation. 
     * 
     */
    public FDLDescriptor( RecordDescriptor rd, ISBounds bnds, AxisOrdering axes )
    {
        this.bounds = bnds.copy();
        dataFileURL = "";
        dataFileType = "unknown";
        fieldNames = new Vector<String>();
        fieldNames.copyInto( rd.getFieldNames() );
        fieldTypes = new Vector<String>();
        fieldTypes.copyInto( rd.getFieldTypes() );
        fieldLayout = DataCollection.pointOrder;
        axisOrdering = axes;
        endianness = ByteOrder.BIG_ENDIAN;
        metadata = null;
    }
    
    /** 
     * Create an FDLDescriptor from a RecordDescriptor, Bounds, AxisOrdering
     * and Metadata.
     */
    public FDLDescriptor( RecordDescriptor rd, ISBounds bnds, AxisOrdering axes,
            MetadataCollection metadata )
    {
        this.bounds = bnds.copy();
        dataFileURL = "";
        dataFileType = "unknown";
        fieldNames = new Vector<String>();
        fieldNames.copyInto( rd.getFieldNames() );
        fieldTypes = new Vector<String>();
        fieldTypes.copyInto( rd.getFieldTypes() );
        fieldLayout = DataCollection.pointOrder;
        axisOrdering = axes;
        endianness = ByteOrder.BIG_ENDIAN;
        this.metadata = metadata;
    }
    
    /** 
     * Factory method pattern to create.
    */
    static public FDLDescriptor create( String fdlFileName )
    {
       FDLDescriptor fdld = null;
	   String[] typicalExtensions = new String[] { ".xfdl", 
	 	                                           ".fdl", 
	 	                                           ".xml" };
	   String fdlLoc  = null;
	   String newName = new String( fdlFileName );

	   // check if file name is "standard"
	   File test = null;
	   for ( int i = 0; i < typicalExtensions.length && test == null; i++ )
	   {
	      if ( newName.endsWith( typicalExtensions[ i ] ))
	         test = new File( newName );
	   }

	   if ( test == null || !test.exists() )   // try to find a standard name
	   {     
	      for ( int i = 0; 
	          (test == null || !test.exists() ) && i < typicalExtensions.length; 
	          i++)
	      {
	         newName = fdlFileName + typicalExtensions[ i ] ;
	         test = new File( newName );
	      }
	      // if appending standard extensions doesn't work, maybe original name works 
	      if ( test == null || !test.exists() )
          {
             newName = fdlFileName;
	         test = new File( newName );
          }
	   }
	   
	   if ( test.exists() )
	   {
	      XFDLReader reader = null;
	      try
	      {
	         reader = new XFDLReader( newName );
	      }
	      catch ( Exception ex )
	      {
	         new SDBError( SDBError.Fatal, "Unable to open file as .xfdl: " + newName );
	      }
	      // save location of fdl file
	      fdlLoc = test.getAbsoluteFile().getParentFile().getAbsolutePath();
	      fdld = reader.getFDLDescriptor();
	      fdld.fdlFileRoot = fdlLoc;
         fdld.fdlFileName = test.getName();
	   }	   
	   else
	      new IllegalArgumentError( SDBError.Fatal,
	                                    "Couldn't find fdl file = " + fdlFileName);
	   return fdld;
    }
    
    /**
     * Create a clone of this object
     */
    public Object clone()
    {
       FDLDescriptor fdlCopy = new FDLDescriptor( );
       if(this.axisOrdering != null){
       fdlCopy.axisOrdering  = (AxisOrdering) this.axisOrdering.clone();
       }
       
       fdlCopy.bounds = (ISBounds) this.bounds.clone();
       fdlCopy.chomboDesiredAttribute = this.chomboDesiredAttribute;
       fdlCopy.chunkDimensions        = this.chunkDimensions;
       fdlCopy.blockDimensions        = this.blockDimensions;
       fdlCopy.dataFileType  = this.dataFileType;
       fdlCopy.dataFileURL   = this.dataFileURL;
       fdlCopy.endianness    = this.endianness;
       fdlCopy.fdlFileName   = this.fdlFileName;
       fdlCopy.fdlFileRoot   = this.fdlFileRoot;
       fdlCopy.fieldLayout   = this.fieldLayout;
       fdlCopy.fieldNames    = this.fieldNames;
       fdlCopy.fieldTypes    = this.fieldTypes;
       fdlCopy.fileOffset    = this.fileOffset;
       fdlCopy.metadata      = this.metadata;
       fdlCopy.recordSize    = this.recordSize;
       fdlCopy.sparseMapData = this.sparseMapData;
       fdlCopy.sparseDataBounds = this.sparseDataBounds;
       return fdlCopy;
    }
    
    //+++++++++++++++++++++++++++++ utility methods +++++++++++++++++++++++++++++
    /**
     * Return the directory that is considered to the 'root' directory for searching
     * for the data file. This is normally the directory that contains the
     * fdl file used to create this FDLDescriptor
     * 
     * @return fdlFile location
     */
    public String getRootDirectory()
    {
       return this.fdlFileRoot; 
    }

    /**
     * Set the root directory for searching for the data file.
     */
    public void setRootDirectory( String newRoot )
    {
       this.fdlFileRoot = newRoot; 
    }

    /**generate a record descriptor using the fieldNames and fieldTypes vectors
     * if there are no field names/types, then an empty RecordDescriptor will be returned
     * 
     */

    public RecordDescriptor getRecordDescriptor()
    {
      RecordDescriptor    recordDesc;
      int elements = fieldNames.size();   //assume that the parser only gave us pairs of names/types

      AttributeDescriptor[] attrDescs= new AttributeDescriptor[elements];
      for ( int i=0; i<elements; i++ )
      {
        attrDescs[i]=new AttributeDescriptor( fieldTypes.elementAt(i), 
                                              fieldNames.elementAt(i));
      }

      recordDesc = new RecordDescriptor(attrDescs);  
      return recordDesc;
    }

    /** Sets the bounds of the Index Space for this 
        FDLDescriptor.  If a null reference is
        passed, then an IllegalArgumentException is thrown.
    */
     
    /** Sets the Index Space bounds **/
    public void setBounds(ISBounds bounds)
    {
        if (bounds!=null){
            this.bounds=bounds.copy();
        } else {
            new IllegalArgumentError(SDBError.Recoverable,"FDLDescriptor.setBounds() was passed a null reference");
        }
    }
        
    /** Returns the bounds of the Index Space for this 
        FDLDescriptor.    */
    public ISBounds getBounds()
    {
        if (bounds != null)
          return this.bounds.copy();
        else return null;
    }

    /** Sets the URL of the XML input file**/
    public void setDataFileURL( String dfURL )
    {
        if ( dfURL != null )
            dataFileURL = dfURL;
        else 
        {
            new IllegalArgumentError( SDBError.Recoverable,
               "FDLDescriptor.setDataFileURL() was passed a null reference");
        }
    }
    /**
     * This method uses the argument as a potential file name and checks for
     * its existence. If the file does not exist, we attempt to "find" it.
     * For now we try to see if it is located the same place the fdl input
     * file was located (if we are given an fdl filename). 
     * 
     * If the file doesn't exist, we return the passed in argument.
     * 
     * In the future we could have some environment variables and installation
     * specified data that could define a DATAPATH search strategy.
     * 
     */
    private String checkDataFileName( String fileName )
    {
       File tester = new File( fileName );
       if ( !tester.exists()  
             &&  this.fdlFileRoot != null 
             &&  fileName.indexOf( File.separatorChar ) < 0 )  // no separators in name
       {
          // find path prefix 
          String fdlRootedName = this.fdlFileRoot + File.separator + fileName;
          tester = new File( fdlRootedName );

          if ( tester.exists() )
             return fdlRootedName;
       }
       return fileName;
    }
    
    /** Returns the file URL for this FDLDescriptor as a String.
     *  If the URL refers to a file and it is not fully qualified, we try to
     *  find it from the current working directory. if that fails we look for
     *  it beginning at the location of the fdl file;  if we find it there, we
     *  return the full path name, otherwise we return the original url. 
     *  
     *  Currently, this does not truly support real URLs.
     */
    public String getDataFileURL()
    {
       return checkDataFileName( dataFileURL );
       //return dataFileURL;
    }

    /** Returns the file specification string for this FDLDescriptor as a String.
     *  No checks are made. 
     */
    public String getDataFileString()
    {
       return dataFileURL;
       //return dataFileURL;
    }

    /** Sets the DataFile Type as a String*/
    public void setDataFileType( String dfType )
    {
      if ( dfType != null )
          dataFileType = dfType;
      else 
      {
          new IllegalArgumentError(SDBError.Recoverable,
             "FDLDescriptor.setDataFileType() was passed a null reference");
      }
    }

    /** Returns the DataFile Type as a string */
    public String getDataFileType()
    {
       return dataFileType;
    }
        
    /** Sets the Record Size (not used)*/
    public void setRecordSize( int rs )
    {
       recordSize = rs;
    }
        
    /** Returns the Record Size as an integer(not used)*/
    public int getRecordSize()
    {
       return recordSize;
    }

    /** Returns the the dimensionality for this FDLDescriptor as an int*/
    public int getDim()
    {
        return bounds.dim();
    }
        
    /** Adds a field name to the current record descriptor **/
    public void addFieldName( String n )
    {
      fieldNames.addElement(n);
    }
        
    /** Adds a field type to the current record descriptor **/
    public void addFieldType( String t )
    {
      fieldTypes.addElement(t);
    }	
	
    /** Sets the attribute format. An argument value of 'true'
		indicates the file is in point order, while 'false' indicates
		attribute order. */
    public void setFieldLayout( int fieldLayout )
    {
        this.fieldLayout = fieldLayout;
    }   

    /**
     * @deprecated implied by setChunkedDimensions( isid )
     * 
     * @param chunkedFileFormat 
     */
    public void setFileFormat( boolean chunkedFileFormat )
    {
        this.chunkedFileFormat = chunkedFileFormat;
    }
    
    
    public void setChunkDimensions (IndexSpaceID id)
    {
        chunkDimensions = new IntegerIndexSpaceID (id);
    }
        
    /** Returns a boolean if Point Order has been selected */
    public int getFieldLayout()
    {
        return this.fieldLayout;
    }
        
    public boolean isChunkedFileFormat ()
    {
        return chunkDimensions != null;
    }
    
    public IndexSpaceID getChunkDimensions ()
    {
        return chunkDimensions;
    }
      
    /** Returns the number of fields in a record descriptor as an integer */
    public int getNumberOfFields()
    {
        return fieldNames.size();
    }

    /** Sets the axisOrdering */
    public void setAxisOrdering( AxisOrdering axisOrdering )
    {
        this.axisOrdering = axisOrdering;
    }
    
    /** Returns the axisOrdering */
    public AxisOrdering getAxisOrdering()
    {
        return axisOrdering;
    }
    
    /** Sets the file offset */
    public void setFileOffset( long fileOffset )
    {
        this.fileOffset = fileOffset;
    }
    
    /** Returns the file offset */
    public long getFileOffset()
    {
        return fileOffset;
    }
    
	/** Returns a ByteOrder indicating whether the file
		byte order is bigendian.*/
    public ByteOrder getByteOrder()
    {
    	   return this.endianness;
    }
    
	
    /** Tell the FDLDescriptor the byte order for the file. */
    public void setByteOrder( ByteOrder endianness )
    {
        this.endianness = endianness;
    }
    
    public boolean isBigEndian()
    {
        if ( endianness == ByteOrder.BIG_ENDIAN )
            return true;
        else
            return false;
    }

    
    /** For Chombo files, we currently support reading only a single
     * attribute. This method allows the fdld to specify which attribute
     * should be read. Currently, this value is specified in the fdl file.
     */
    public void setChomboDesiredAttribute( int desiredAttribute ){
    
      this.chomboDesiredAttribute = desiredAttribute;
    }

    /** Return the attribute being read from the Chombo file. */
    public int getChomboDesiredAttribute(){
    
      return this.chomboDesiredAttribute;
    }
    
    //++++++++++++ SparseDataSource methods +++++++++++++++++
    public void setBlockDimensions( IndexSpaceID id )
    {
        blockDimensions = new IntegerIndexSpaceID( id );
    }
    public IndexSpaceID getBlockDimensions()
    {
        return blockDimensions;
    }
      
    /**
	 * returns this objects Sparse block mapper information
	 * 
	 * @return this objects Sparse block mapper information
	 */
	public String[] getSparseBlockMap() {
		return this.sparseMapData;
	}

	/**
	 * sets this object's sparse block mapping info
	 * 
	 * @param mapInfo String>[] of form srcBlk(blkOrigin):dstBlk
	 */
	public void setSparseBlockMap( String[] mapInfo ) 
   {
		this.sparseMapData = mapInfo;
	}
   /**
    * 
    * @return true if the data source is stored with sparse blocks
    */
   public boolean isSparse()
   {
      return sparseMapData != null && sparseMapData.length > 0;
   }
   
   /** Sets the Index Space bounds **/
   public void setSparseDataBounds(ISBounds sparseBounds)
   {
      this.sparseDataBounds=sparseBounds.copy();
    
   }
   
   /** Returns the bounds of the Index Space for this
    FDLDescriptor.    */
   public ISBounds getSparseDataBounds()
   {
       if (sparseDataBounds != null)
           return this.sparseDataBounds.copy();
       else return null;
   }

   /**
	 * returns this objects Metadata object
	 * 
	 * @return this objects Metadata object
	 */
	public MetadataCollection getMetadata() {
		return this.metadata;
	}

	/**
	 * sets this object's MetadataCollection object
	 * 
	 * @param meta the MetadataCollection
	 */
	public void setMetadata(MetadataCollection meta) {
		this.metadata = meta;
	}

	/**
	 * Write a .xfdl formatted file to the file that is represented by this
	 * FDLDescriptor. The file will contain all the attributes that are
	 * represented in this FDLDescriptor.
	 */
	public void writeFDL() {
		writeFDL(new File(fdlFileRoot + File.separatorChar + fdlFileName));
	}

	/**
	 * Write a .xfdl formatted file to the file specified which includes all the
	 * attributes in this FDLDescriptor.
	 * 
	 * TODO: add "Chunked" data
	 */
	public void writeFDL(File f) {
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(f));
			// write header
			writer.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
			writer
					.write("<!DOCTYPE FileDescriptor PUBLIC \"-//SDB//DTD//EN\"  \"fdl.dtd\">\n\n");

			String axisOrdering = "";
			int[] order = getAxisOrdering().ordering;
			for (int i = 0; i < order.length - 1; i++) {
				axisOrdering += "" + i + ' ';
			}
			axisOrdering += order[order.length - 1];

			String fieldLayout;
			if (getFieldLayout() == DataCollection.pointOrder) {
				fieldLayout = "byPoint";
			} else {
				fieldLayout = "byAttribute";
			}

			String fileFormat;
			if ( chunkedFileFormat ) {
				fileFormat = "chunked";
			} else {
				fileFormat = "native";
			}

			String byteOrder;
			if (getByteOrder() == ByteOrder.BIG_ENDIAN) {
				byteOrder = "big";
			} else {
				byteOrder = "little";
			}

			// write FileDescriptor
			writer.write("<FileDescriptor\n");
			writer.write("\tfileName=      \"" + getDataFileURL() + "\"\n");
			writer.write("\tfileType=      \"" + getDataFileType() + "\"\n");
			writer.write("\trecordSize=    \"" + getRecordSize() + "\"\n");
			writer.write("\taxisOrdering=  \"" + axisOrdering + "\"\n");
			writer.write("\tfileOffset=    \"" + getFileOffset() + "\"\n");
			writer.write("\tfieldLayout=   \"" + fieldLayout + "\"\n");
			writer.write("\tfileFormat=    \"" + fileFormat + "\"\n");
			writer.write("\tbyteOrder=     \"" + byteOrder + "\"\n");
			writer.write("                >\n\n");

			// write Bounds
			ISBounds b = getBounds();
			if (b != null) {
				int[] lower = b.getLowerArray();
				int[] upper = b.getUpperArray();
				for (int i = 0; i < upper.length; i++) {
					writer.write("\t<Bounds lower= \"" + lower[i]
							+ "\" upper= \"" + upper[i] + "\"/>\n");
				}
				writer.write('\n');
			}

			// write Fields
			String[] fieldNames = getRecordDescriptor().getFieldNames();
			String[] fieldTypes = getRecordDescriptor().getFieldTypes();
			for (int i = 0; i < fieldNames.length; i++) {
				writer.write("\t<Field fieldName= \"" + fieldNames[i]
						+ "\" fieldType= \"" + fieldTypes[i] + "\"/>\n");
			}
			writer.write('\n');
         
         // write Chunk dimensions, if specified
         if ( chunkDimensions != null )
         {
            int[] isid = chunkDimensions.getCoords();
            for ( int i = 0; i < isid.length; i++ )
            {
               writer.write( "\t<Chunk dimension=\"" + isid[ i ] + "\" />\n" );
            }
            writer.write( "\n" );
         }
         
         // write sparse block dimensions, if specified
         if ( blockDimensions != null )
         {
        	 int totalDatumInBlock = 1 ;
            int[] isid = blockDimensions.getCoords();
            for ( int i = 0; i < isid.length; i++ )
            {
               writer.write( "\t<SparseBlock dimension=\"" + isid[ i ] + "\" />\n" );
               totalDatumInBlock = totalDatumInBlock * isid[ i ];
            }
            writer.write( "\n" );
            
            // abi, 04/28/14 : SparseData bounds in the sparsed File
            ISBounds sp_bounds = getSparseDataBounds();
            if (sp_bounds != null) {
                int[] lower = sp_bounds.getLowerArray();
                int[] upper = sp_bounds.getUpperArray();
                for (int i = 0; i < upper.length; i++) {
                    writer.write("\t<SparseDataBounds lower= \"" + lower[i]
                                 + "\" upper= \"" + upper[i] + "\"/>\n");
                }
                writer.write('\n');
         }
         }
         
         // write sparse map info, if specified
         if ( sparseMapData != null )
         {
            for ( String s: sparseMapData )
            {
               writer.write( "\t<Map map=\"" + s + "\" />\n" );
            }
         }

			// write Metadata
			if (getMetadata() != null) {
				writer.write(getMetadata().toXML("\t"));
			}

			// close FileDescriptor
			writer.write("</FileDescriptor>");
			writer.close();
		

		} catch (IOException e) {
			e.printStackTrace();
		}
		
		
	}
	

  /** Returns a String representation of this class*/
    public String toString()
    {
      String result = "{ URL: " + dataFileURL + ", Type: " + dataFileType;

      result = result + " RecordSize: " + getRecordSize();
      if ( bounds != null )
          result = result + ",\nBounds: " + bounds.toString();
        
      result = result + ",\nRecordDesciptor: " +  getRecordDescriptor().toString();
      
      result = result + ",\nPointOrder: " + getFieldLayout();
      
      if ( axisOrdering != null )
        result = result + ",\nAxisOrdering: " + getAxisOrdering().toString() + "}";
      
      if(metadata != null)
    	result = result + ",\nMetadata: " + getMetadata().toString() + "}";
      
      return result;
    }
}
