/* 
    PackedIndexSpaceID.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
*/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;

/** This class implements the IndexSpaceID class using a single integer to 
    represent the dimensions of the ID. The leftmost 2 bits are used to 
    represent the dimensionality, which can range from 1 to 4.
*/

public class PackedIndexSpaceID extends IndexSpaceID implements Cloneable {

    private static final int LowerDimensionalityLimit=1; 
    private static final int UpperDimensionalityLimit=4; 
    protected int packedID=0;



    /** Constructs a one dimensional PackedIndexSpaceID with axis value set 
        to the argument.
    */
    public PackedIndexSpaceID(int i) {

        this.packedID= (i & 0x3FFFFFFF);
        this.setDim(1);
    }

    /** Constructs a two dimensional PackedIndexSpaceID with axis values set 
        to the given values.
    */
    public PackedIndexSpaceID(int i, int j){

        this.packedID=((i & 0x00007FFF) << 15) | (j & 0x00007FFF);
        this.setDim(2);
    }


    /** Constructs a three dimensional PackedIndexSpaceID with axis values set 
        to the given values.
    */
    public PackedIndexSpaceID(int i, int j, int k){

        this.packedID=((i & 0x000003FF) << 20) | ((j & 0x000003FF) << 10) | (k & 0x000003FF);
        this.setDim(3);
    }


    /** Constructs a PackedIndexSpaceID with dimensionality matching the length of the 
        parameter array. The axes will be set to the contents of the array.
    */
    public PackedIndexSpaceID(int [] axes){
    
        int i,dim=axes.length;
        int bitsPerAxis=30/dim;
        int mask= (1 << bitsPerAxis) - 1;
        
        this.packedID=0;
        
        for(i=0;i<dim;i++){
            this.packedID <<= bitsPerAxis;
            this.packedID |= (axes[i] & mask);
        }
        
        this.setDim(axes.length);
    }
    
    
    /** Constructs a PackedIndexSpaceID from the contents of the IndexSpaceID argument. 
        If the dimensionality of the argument is too large, an exception will be thrown.
    */
    public PackedIndexSpaceID(IndexSpaceID id){
    
        int i,dim=id.dim();
        int bitsPerAxis=30/dim;
        int mask= (1 << bitsPerAxis) - 1;
        
        this.packedID=0;
            
        for(i=0;i<dim;i++){
            this.packedID <<= bitsPerAxis;
            this.packedID |= (id.getCoord(i) & mask);
        
        }
        
        this.setDim(dim);
    }
    
    
    /** A private constructor used by the clone() method. */
    private PackedIndexSpaceID(){
    
    }


    /** Return a copy of this IndexSpaceID. */
    public final IndexSpaceID copyID(){
        
        return (IndexSpaceID) this.clone();
    }


    /** Return a copy of this IndexSpaceID. */
    public Object clone(){
        
        PackedIndexSpaceID id=new PackedIndexSpaceID();
    
        id.packedID=this.packedID;
        
        return id;
    }


    /**Set the axes of this id to the axes of the parameter id.
       The assign method may elect to refer to the data in
       the parameter id, and not make a copy of this information.*/
    public final void assign(PackedIndexSpaceID id){

        this.packedID=id.packedID;
    }

    /**Set the coords of this id equal to the coords of the parameter id.
       The assign method may elect to refer to the data in
       the parameter id, and not make a copy of this information.*/
    void assign(int[] coords){
        int i,dim=this.dim();
        
        for(i=0;i<dim;i++){
            
            this.setCoord(i,coords[i]);
        }
    }
    

    /** Get the value corresponding to the specified axis. If the specified axis 
        is too large, a DimensionalityError is created. */
    public int getCoord(int axisnumber){

        int dim=this.dim();
        
        if (axisnumber<0 || axisnumber >= dim){ 
            new DimensionalityError(SDBError.Recoverable,"Argument specifies a non-existent axis.");
        }
        
        int bitsPerAxis=30/dim;
        int shiftdistance=bitsPerAxis*axisnumber+2;
        
            
        return (this.packedID << shiftdistance) >> (32-bitsPerAxis);     
    }
    
    
    
    /** Get an array of axis values.*/
    int[] refCoords(){
        
        int i,dim=this.dim();
        int [] axes=new int[dim];
        
        for(i=0;i<dim;i++){
            
            axes[i]=this.getCoord(i);
        
        }
    
        return axes;
    
    
    /*  FiXME: Doesn't work for negatives. Can't be bothered with it right now.
    
        int i,dim=this.dim();
        int bitsPerAxis=30/dim;
        int [] axes=new int[dim];
        int bits=this.packedID & 0x3FFFFFFF;
        
        for(i=0;i<dim;i++){
            
            axes[i]=bits >> (bitsPerAxis*(dim-1));
            bits=(bits << bitsPerAxis) & 0x3FFFFFFF;
        }
        
        return axes;
        
   */     
    }


    /** Fill the array argument with the coordinate values for this ISID. For convenience,
        a reference to the argument array is returned. */
    public int[] getCoords(int[] coords){
    
        int i, d=this.dim();
        
        if( coords.length != d){
            new DimensionalityError(SDBError.Recoverable,"Argument array is the wrong length.");
        }
        
        for(i=0; i< d; i++){
        
            coords[i]=this.getCoord(i);
        }
        
        return coords;
    }

    
    
    /** Set the specified axis to the specified value. If the specified axis 
        is too large, a DimensionalityError is created. */ 
    public void setCoord(int axisnumber,int value){
    
        int dim=this.dim();
        
        if (axisnumber<0 || axisnumber >= dim){ 
            new DimensionalityError(SDBError.Recoverable,"Argument specifies a non-existent axis.");
        }
        
        int bitsPerAxis=30/dim;
        int mask= (1 << bitsPerAxis) - 1;
        int shiftdistance=bitsPerAxis*(dim-1-axisnumber);
        
        value&=mask;
        
        this.packedID= (this.packedID & ~(mask << shiftdistance)) | (value << shiftdistance);
    }

    
    /** Returns the dimensionality of this IndexSpaceID. */
    public int dim(){
    
        return ((0xC0000000 & packedID) >>> 30)+1;
    }
    
    /** Return a String representing the object. */
    public String toString(){
        int i;
        int dim=this.dim();
        
        String retval= new String("<PackedISID (");
        
        try {
        
            for(i=0;i<dim-1;i++){
                retval+=this.getCoord(i)+",";
            }
            
            retval+=this.getCoord(i)+")>";
        } catch(DimensionalityException e) {
            System.err.println("***PackedIndexSpaceID.toString(): DimensionalityException");
        }
        
        return retval;
    }    

    /** A private method that sets the leftmost 2 bits to the specifield dimensionality value.
        This value must currently be between 1 and 4, otherwise, a DimensionalityError is created.
    */
    private void setDim(int d){
        
        if (d > PackedIndexSpaceID.UpperDimensionalityLimit 
        		|| d < PackedIndexSpaceID.LowerDimensionalityLimit)
            new DimensionalityError(SDBError.Recoverable,"Specified dimensionality is out of range.");
        
        this.packedID = (this.packedID & 0x3FFFFFFF) | ((d-1) << 30);
    }
    
    public static void main(String[] args){
        
        System.out.println("*****PackedIndexSpaceID Tests*****");
        
        PackedIndexSpaceID first=new PackedIndexSpaceID(10,10);
        
        System.out.println("First ID is "+first);
        
        PackedIndexSpaceID theClone=(PackedIndexSpaceID) first.clone();

        System.out.println("The clone is "+theClone);

        theClone.setCoord(1,5);
        
        System.out.println("After theClone.setCoord: The clone is "+theClone);
        System.out.println("                        and the original is "+first);

        first=new PackedIndexSpaceID(1,-2,3);
        System.out.print("getCoords() for "+first+": ");
        
        int [] axes=first.refCoords();
        
        for(int i=0;i<3;i++){
            System.out.print(axes[i]+" ");
        }
        System.out.println();
    
    }
}