/* 
    PhysicalDataSource.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
    
    -------------------- Edit history as of 06/22/07 -------------------
 *  11/21/13 rdb/can: change access to "directMemoryLimit"
 *  02/12/12 rdb: Made subBasicBlock( ISB, ISB, ISB ) abstract; the old code
 *                called subblock( ISB, ISB, ISB ), which just calls subBasicBlock
 *                again. This used to be abstract and needs to be again.
    01/31/12 rdb: Added constuctor using FDLDescriptor; added setupBuffers method
                  so that constructor could share code.
    07/27/09 jpf: Added method to set byteOrder for a data source.  This is part of new
                  solution to prevent granite from byte swapping data that is stored as
                  little endian.  We temporarily change byteOrder to big endian, perform
                  retrieval of data and then reset it to little endian.
    06/22/07 pjr: set maxBufferSize to be the value specified by -XX:MaxDirectMemorySize.
                  This may change when we need more than one BFDS at a time, but it'll
                  do for now.  Also, allocateBuffer() returns true iff a buffer large
                  enough to contain a complete rod of the queryBounds was allocated, and
                  false otherwise.
    12/07/05 rdb: Added number argument to initializeBuffers.
    12/04/05 rdb: Added checkActivated() method to be used prior to data access.
                    This version does an activate(). If we want to force user
                    to activate, we can put an warning message here.
                    I've put this at the PDS level since I'm not sure it is needed
                    at Composite level -- nor do we want it, I think.
                  Promoted most subblock methods from children to here with Template
                    Pattern methods based on subBasicBlock and subCompositeBlock. 
                 -renamed psubblock to psubBasicBlock; this is first step to 
                  promoting public subblock methods up to PDS -- lots of shared code
                  between AOBFDS and POBFDS.
    11/28/05 rdb: Revised getURL() to access  FDLDescriptor.getDataFileURL()
                  Renamed getDataFileName to getDataFileString to be consistent with
                     same method in FDLDescriptor()
    10/21/05 rdb: Deleted unused variables
                  Added dummy "main" to reference debug methods so no compiler warnings
    06/01/05 rdb: modified to match FDLDescriptor refactoring
    04/26/05 rdb: added  this.bufferBounds for convenience
                  ********* realized that attribute order stuff will never work
                  *         with caching (and maybe other things). We have an
                  *         array of ByteBuffers, but only 1 set of variables to
                  *         define dimensions and steps and bounds. 
                  *         We need a Buffer class, with perhaps a child for
                  *         BufferIterator.
    04/24/05 rdb: added ByteBuffer allocateByteBuffer( ByteBuffer, size )
                    internal convenience function.
    04/23/05 rdb: revised allocateBuffer:
                  -return boolean if fully satisfied
                  -changed to accept long as argument for byte count version
                  -added getBufferSize methods
    04/22/05 rdb: added getMaxBufferSize()
                  changed allocateBuffer() to return size of allocated buffer
                      and to allocate max if request is more than that.
                    ** this is a stopgap, temporary solution, could create
                       problems in other places (such as psubblock).
    04/09/05 rdb: - fixed makeBestBuffer bufferDimensions (problem when gap != 0)
    02/22/05 rdb: - complete rewrite of makeBestBuffer
    02/20/05 rdb: - renamed gapSize to maxGapSize and setGapSize to setMaxGapSize
                  - revised setMaxBufferSize( long ) to set a instance variable
                       allows different sizes for each PDS. Default will be the
                       static variable, now called defaultMaxBufferSize
                  - getBestBufferSize changed:
                     change name to makeBestBuffer
                     added "unitSize" arg so it works for ChunkedDataSources
                     deleted input arguments: 
                         srcBounds - use this.bounds
                         gapSize   - use this.getMaxGapSize
                     deleted the 2 output args; now it directly updates 
                         this.bufferDimension
                         this.bufferSteps
                         
    02/15/05 rdb: disable gapsize = 0 hack in allocateBuffer()
    02/14/05 rdb: added deactivate() that deletes ByteBuffers. There are some
                     problems with BFDS, that are hidden by re-use of data in buffers.
    02/07/05 rdb: consolidated all ByteBuffer definition and support here.
                  Support an array of ByteBuffers. Simple DS only needs 1 so
                  that is the default. 
                  For now, CompositeDataSources let their components make buffers,
                  but these multiple buffer features might be useful in the future.
               implemented buffer support methods:
                  allocateBuffer( i, size );
                  allocateBuffer( i, ISBounds );
                  allocateBuffer( i, ISBounds, unitSize )
                  ByteBuffer getBuffer( i );
                and convenience functions for 1 buffer DataSources:
                  ByteBuffer getBuffer();  
                  allocateBuffer( size );
                  allocateBuffer( ISBounds );
                  allocateBuffer( ISBounds, unitSize )
    02/06/05 rdb: moved endian variable, byte order get/set methods and 
                      allocateBuffer methods from BFDS to here. The ByteBuffer
                      variable is here; its methods should be, too. byte order is
                      key for that, so I moved it as well.
    12/06/04 rdb: moved AxisOrdering to DataCollection
                  moved attribute order constants to DataCollection
*/

package edu.unh.sdb.datasource; 

import edu.unh.sdb.common.*; 
//import java.io.*; 
import java.nio.*;
import java.util.*;


abstract public class PhysicalDataSource extends DataSource
{
    // We redefine the ByteOrder constants here for convenience. 
    public static final ByteOrder BIG_ENDIAN    = ByteOrder.BIG_ENDIAN; 
    public static final ByteOrder LITTLE_ENDIAN = ByteOrder.LITTLE_ENDIAN; 
    protected           ByteOrder endian = ByteOrder.BIG_ENDIAN;
    
    
    Properties properties=System.getProperties();
    
    // This value is set by using a JVM argument such as -XX:MaxDirectMemorySize=128m
    //public static long directMemoryLimit = sun.misc.VM.maxDirectMemory();
    public static long directMemoryLimit = Runtime.getRuntime().maxMemory();

    //rdb: Note that "long" on the buffer size does not really make sense
    //     in the current context. The allocateDirect java method only allows
    //     in 'int' for size. Some day we might provide multiple physical 
    //     buffers for one larger logical buffer, but lots of other things
    //     would need to be changed before that will work!
    
    
    protected long maxBufferSize = PhysicalDataSource.directMemoryLimit;

    protected String url; 

    //used by computeBestBufferSize( )
    //rdb: bufferDimension should define the sizes in each axis of the buffer.
    //     02/21/05: 1. I don't think its correct and 
    //               2. It's not used consistently. POBFDS uses it, but not
    //                  BFDS or AOBFDS or ChunkedDS.
    //     bufferSteps defines how an iteration through the space would
    //        proceed using the buffer.
    protected IntegerIndexSpaceID bufferDimension;
    protected ISBounds            bufferBounds;     
    protected int[]               bufferSteps; 
    
    protected long                maxGapSize = 0;  // largest allowed gap for reading unneeded data

    protected ByteBuffer[]        byteBuffers; 

    // used by computeBestBufferSize, if the datasource is binaryFileDataSource
    // unitByteSize == this.recordByteSize. If the datasource is chunkedDataSource
    // unitByteSize == chunkByteSize; 
    // by default it is the recordByteSize  

    //********** testing util variables **************************************
    protected long readCounter; 
    protected long readLength; 
    //************************************************************************
    
    //----------------------- Constructors ------------------------------
   /**
    * Create a PhysicalDataSource from an fdl
    * 
    * @param name
    * @param fdl 
    */
   public PhysicalDataSource( String name, FDLDescriptor fdl )
   {
      this( name, 
            fdl.getRecordDescriptor(), 
            fdl.getBounds(), 
            fdl.getAxisOrdering(), 
            null   //URL     
          );
   }
   /**
    * Create a Physical data source from specs
    * @param name
    * @param rd
    * @param isb
    * @param ordering
    * @param URL 
    */
   public PhysicalDataSource( String           name,  
                              RecordDescriptor rd,        
                              ISBounds         isb, 
                              AxisOrdering     ordering, 
                              String           URL )
    {
       super( name, rd, isb );

       url = URL;
       setAxisOrdering( ordering );

       // by default the gapSize is 0. This means the datasource
       // does not want to read any extra data to eliminate a read 
       // operation
       this.maxGapSize = 0;
       setupBuffers();
    }

    private void setupBuffers()
    {
       //used by makeBestBuffer     
       this.bufferDimension = new IntegerIndexSpaceID( new int[ bounds.dim() ] );
       this.bufferBounds = bounds.copy();
       this.bufferSteps = new int[ bounds.dim() ];
       // by default there is 1 buffer. Children must recall if need more
       initializeBuffers( 1 );
    }

   //--------------- initializeBuffers ----------------------------------
   /** 
    * Initialize buffers for i/o
    */
   protected void initializeBuffers( int num )
   {
      // allocate the ByteBuffer array
      this.byteBuffers = new ByteBuffer[ num ];
      for ( int i = 0; i < num; i++ )
         this.byteBuffers[ i ] = null;
   }
   /** The DataSource will release any resources used during the 
       *	readling or processing of data. It is important
    *  that all child classes call super.deactivate()
    */    
   public void deactivate()
   {
      try{     
         deleteBuffers();
      } catch(NullPointerException e){     
         new SDBError(SDBError.Recoverable, "NullPointerException in PhysicalDataSource.deactivate()");
      }
      
      super.deactivate();
   }
	
    /** gap means the number of extra data, the datasource willing to read
        to elimilate a read operation. If the user does not call this method
        the default gap size is 0
    */  
    public void setMaxGapSize( long gap )
    {
         this.maxGapSize = gap; 
    }

    /** Returns true if the binary file is big-endian format or little endian 
     * */
    public boolean isBigEndian()
    {
        return endian == ByteOrder.BIG_ENDIAN;
    }
   
    /** Return the byteOrder of this binary file */
    public ByteOrder getByteOrder()
    {
        return endian ;
    }     
    
    /** set the byteOrder of this binary file.  Useful when you want to
        prevent Granite from byte swapping little endian data.  This is
        done by telling Granite that the data is big endian even though it isn't. */
    public void setByteOrder(ByteOrder newOrder){
        endian = newOrder;
    }

    /** set the byte order for the given buffer */
    protected void setBufferByteOrder( ByteBuffer buffer )
     {
         if( (this.endian).equals( ByteOrder.BIG_ENDIAN ) )
           buffer.order( ByteOrder.BIG_ENDIAN );
         else
           buffer.order( ByteOrder.LITTLE_ENDIAN );
      }

    /** set the maximum length, in bytes, of the buffer used directly
     *  with file I/O operations.
     */
    public void setMaxBufferSize( int numBytes )
    {
        this.maxBufferSize = numBytes; 
    }
    
    /** get the maximum length, in bytes, of the buffer used directly
     *  with file I/O operations.
     */
    public int getMaxBufferSize()
    {
        return (int) this.maxBufferSize; 
    }
    
    /** get the current length, in bytes, of the buffer used directly
     *  with file I/O operations.
     */
    public int getBufferSize()
    {
        return getBufferSize( 0 ); 
    }
    
    /** get the current length, in bytes, of the buffer used directly
     *  with file I/O operations.
     */
    public int getBufferSize( int bufNum )
    {
       try
       {
          return this.byteBuffers[ bufNum ].limit();
       }
       catch ( IndexOutOfBoundsException aiex )
       {
          new SDBError( SDBError.Recoverable, "Invalid buffer index: " + bufNum );
          return 0;
       }
    }
    
    /** This method (re)allocates the specified I/O byte buffer to the byte size 
     * given by the second argument. If the size of the existing buffer is correct,
     * it is simply cleared. Otherwise, a new ByteBuffer is allocated. 
     * 
     * This method returns true if a buffer of the requested size was allocated
     * and false otherwise.
     * (rdb 03/2012: called from this class twice and from AttributeOrderBFDS twice)
     */
    protected boolean allocateBuffer( int buf, long byteSize )
    {
       long requestSize = byteSize;
       try
       {
          //this.byteBuffers[ buf ] = allocateBuffer( this.byteBuffers[ buf ], byteSize );
          //return this.byteBuffers[ buf ].capacity() == byteSize;
                    
          if ( byteBuffers[ buf ] != null 
                 && byteSize <= byteBuffers[ buf ].capacity() ) // need new buf?
          {
              // if old buffer is big enough, just keep using it.
              // setting the limit here doesn't do much good since every call to
              // clear() will reset to capacity.
              byteBuffers[ buf ].clear();
              byteBuffers[ buf ].limit( (int) byteSize );
          }
          else 
          {
            if ( byteSize > this.maxBufferSize )
            {
               new SDBError( SDBError.Warning, "PhysicalDataSource.allocateBuffer::" 
                                             + " Trying to allocate buffer (" 
                                             + byteSize + ") larger than max (" 
                                             + this.maxBufferSize + ")." );
               byteSize = (int) this.maxBufferSize;
            }
            else if ( byteSize < 0 )
            {
               new SDBError( SDBError.Warning, " Tried to allocate a negative " +
                                   "length buffer. Setting length to maxBufferSize" );
               byteSize = (int) this.maxBufferSize;
            }
            if ( byteSize > Integer.MAX_VALUE ) 
            {
               new SDBError( SDBError.Warning, "Buffer size > MAXINT " +
                                                          "not supported by yet." );
               byteSize = Integer.MAX_VALUE;
            }
            byteBuffers[ buf ] = null; // try to show that old one is gone
            byteBuffers[ buf ] = ByteBuffer.allocateDirect( (int) byteSize );
            //byteBuffers[ buf ] = ByteBuffer.allocate( (int) byteSize );
            //byteBuffers[ buf ] = ByteBuffer.wrap( new byte[ (int) byteSize ]);
            this.setBufferByteOrder( byteBuffers[ buf ] );
         }
       } 
       catch ( IndexOutOfBoundsException ex )
       {
          new SDBError( SDBError.Recoverable,
                                        "System error: invalid buffer index"
                                        + " passed to allocateBuffer: " + buf );
          return false;
       }
       return this.byteBuffers[ buf ].limit() == requestSize;
   }
    /** This method (re)allocates the argument I/O byte buffer to the byte size 
     * given by the second argument. If the size of the existing buffer is correct,
     * it is simply cleared. Otherwise, a new ByteBuffer is allocated. First arg
     * is allowed to be null.
     * 
     * (rdb 03/2012: not used)
     */
    protected ByteBuffer allocateBuffer( ByteBuffer iobuf, long byteSize )
    {  
       if ( iobuf != null && byteSize <= iobuf.capacity() ) // need new buf?
       {
          iobuf.clear();
          iobuf.limit( (int) byteSize ); // re-use old if it is big enough
       }
       else 
       {
          if ( byteSize > this.maxBufferSize )
          {
             new SDBError( SDBError.Warning, "PhysicalDataSource.allocateBuffer::" 
                                           + " Trying to allocate buffer (" 
                                           + byteSize + ") larger than max (" 
                                           + this.maxBufferSize + ")." );
             byteSize = (int) this.maxBufferSize;
          }
          else if ( byteSize < 0 )
          {
             new SDBError( SDBError.Warning, " Tried to allocate a negative " +
                                 "length buffer. Setting length to maxBufferSize" );
             byteSize = (int) this.maxBufferSize;
          }
          if ( byteSize > Integer.MAX_VALUE ) 
          {
             new SDBError( SDBError.Warning, "Buffer size > MAXINT " +
                                                        "not supported by yet." );
             byteSize = Integer.MAX_VALUE;
          }
          iobuf = ByteBuffer.allocateDirect( (int) byteSize );
          //iobuf = ByteBuffer.allocate( (int) byteSize );
          this.setBufferByteOrder( iobuf );
       }
       return iobuf;
    }

    /** This method allocates a proper I/O bytebuffer for the queryBounds passed in as 
     * an argument. The query bounds is measured in grid Space. If the existing 
     * ByteBuffer is not the right size a new ByteBuffer is created.
     * 
     * This method returns true if a buffer long enough to contain a complete
     * rod of the queryBounds was allocated, and false otherwise.
     * 
     * (rdb 03/2012: called from this class)
     */
    final protected boolean allocateBuffer( int buf,  ISBounds queryBounds )
    {
       return allocateBuffer( buf, this.bounds, queryBounds, this.recordDesc.byteSize() );
    }
    
    /** This method allocates a proper I/O bytebuffer for the queryBounds passed in as 
     * an argument. The query bounds is measured in grid Space. The third argument
     * represents the size needed for each point in the bounds. If the existing 
     * ByteBuffer is not the right size a new ByteBuffer is created.
     * 
     * This method returns true if a buffer long enough to contain a complete
     * rod of the queryBounds was allocated, and false otherwise.
     * 
     * (rdb 03/2012: called from this class)
     */
    final protected boolean allocateBuffer( int buf, ISBounds queryBounds, int unitSize )
    {
       return allocateBuffer( buf, this.bounds, queryBounds, unitSize );
    }
    
    /** This method allocates a proper I/O bytebuffer for the specified grid space
     * domain (gridBounds) and queryBounds passed in as arguments.
     * The query bounds is measured in grid Space. The third argument
     * represents the size needed for each point in the bounds. If the existing 
     * ByteBuffer is not the right size a new ByteBuffer is created.
     * This is especially useful for allocating buffers by chunking.
     * 
     * This method returns true if a buffer long enough to contain a complete
     * rod of the queryBounds was allocated, and false otherwise.
     * 
     * (rdb 03/2012: called 3 times from this class)
     */
    protected boolean allocateBuffer( int buf, 
                                         ISBounds gridSpace, 
                                         ISBounds queryBounds, int unitSize )
    {
       //System.out.println("************ allocateBufferCalled");
       boolean completeSuccess = true; // true if  buffer size request satisfied

       long bufferSize =  this.makeBestBuffer( gridSpace, queryBounds, unitSize );
       
       long bufferByteSize = bufferSize * unitSize;
       
       if ( bufferByteSize > this.maxBufferSize )
       {
          new SDBError( SDBError.Warning, "PhysicalDataSource.allocateBuffer::" +
          		" Trying to allocate buffer larger than max. " );
          bufferByteSize = maxBufferSize;
          completeSuccess = false;
       }
//       completeSuccess &= allocateBuffer( buf, (int) bufferByteSize );

       //PJR  Let's return true if the buffer is long enough to contain
       //     an entire rod of the queryBounds, and false otherwise.
       allocateBuffer( buf, (int) bufferByteSize );
       completeSuccess &= (bufferSize >= queryBounds.getDimension(this.axisOrdering.getLeastSignificantAxis()));
       
       //System.out.println( "allocateBuffer: " + bufferByteSize );
       return completeSuccess;
    }
    
    /** Allocate buffer 0.
     * This is a convenience function for most DataSources
     * 
     * This method returns true if a buffer long enough to contain a complete
     * rod of the queryBounds was allocated, and false otherwise.
     * 
     * (rdb 03/2012: called by ChunkedDataSource, PointOrderBFDS, SparseDataSource)
     */
    final protected boolean allocateBuffer( int size )
    {
       return allocateBuffer( 0, size );
    }

    /** Allocate buffer 0.
     * This is a convenience function for most DataSources
     * 
     * This method returns true if a buffer long enough to contain a complete
     * rod of the queryBounds was allocated, and false otherwise.
     * (rdb 03/2012: called by PointOrderBFDS)
     */
    final protected boolean allocateBuffer( ISBounds bounds )
    {
       return allocateBuffer( 0, bounds );
    }

    /** Allocate buffer 0.
     * This is a convenience function for most DataSources
     * 
     * This method returns true if a buffer long enough to contain a complete
     * rod of the queryBounds was allocated, and false otherwise.
     * (rdb 03/2012: not used)
     */
    final protected boolean allocateBuffer( ISBounds bounds, int unitSize )
    {
       return allocateBuffer( 0, bounds, unitSize );
    }

    /** Allocate buffer 0.
     * This is a convenience function for ChunkedDataSource
     * 
     * This method returns true if a buffer long enough to contain a complete
     * rod of the queryBounds was allocated, and false otherwise.
     * (rdb 03/2012: called by ChunkedDataSource)
     */
    final protected boolean allocateBuffer( ISBounds grid, ISBounds bounds, int unitSize )
    {
       return allocateBuffer( 0, grid, bounds, unitSize );
    }

    /** delete a byte buffer and garbage collect. 
     * 
     */
    protected void deleteBuffer( int buf )
    {
        this.byteBuffers[ buf ] = null; 
        System.gc( ); 
    }

    /** Delete all byte buffers used for data storage for this DataSource
     * and garbage collect. 
     * 
     */
    public void deleteBuffers( )
    {
       for ( int i = 0; i < this.byteBuffers.length; i++ )
          this.byteBuffers[ i ] = null; 
       System.gc( ); 
    }
    
    /** Delete all byte buffers used for data storage for this DataSource
     * and garbage collect.  Convenience name for 1-buffer datasources. 
     * 
     */
    public void deleteBuffer( )
    {
       deleteBuffers();
    }
    
    /** Return a reference to the specified byte buffer */
    final protected ByteBuffer getBuffer( int i )
    {
       try {
          return this.byteBuffers[ i ];
       } catch ( IndexOutOfBoundsException ex )
       {
          new SDBError( SDBError.Recoverable, "Invalid data buffer index: " + i );
          return null;
       }
    }

    /** Return a reference to byte buffer 0, if there is only one. 
     * This is a convenience function for most DataSources
     */
    final protected ByteBuffer getBuffer()
    {
       return this.byteBuffers[ 0 ];
    }

    //******************************* testing util ***************************
    public void resetReadCounter( )
    {
        readCounter = 0; 
        readLength  = 0; 
    }
    public long getReadCounter( )
    {
        return readCounter; 
    }
    public long getReadLength( )
    {
        return readLength; 
    }
    //*************************** end testing util ***************************
 
    //----------------------- subblock and datum methods ----------------------
    /**
     * 
     */
    abstract void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds );
    abstract void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                                  RecordSpec recSpec );
    abstract void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                                  FieldIDMapper mp );
    abstract void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds );
    abstract void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds, 
                                  RecordSpec recSpec );
    abstract void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,  
                                  FieldIDMapper mp );
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
    defined by dstBounds. The BasicBlock is assumed to be created and  a valid data array is already allocated. 
    NOTE: dstBlock can be  BasicBlock or CompositeBlock. data is copied from file to BasicBlock or CompositeBlock 
    based on its data type.
    */
   public void subblock( DataBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds )
   {
      checkActivated(); 
      if ( dstBlock instanceof BasicBlock )
         subBasicBlock( (BasicBlock) dstBlock, dstBounds, srcBounds ); 
      else if ( dstBlock instanceof CompositeBlock )
         subCompositeBlock((CompositeBlock) dstBlock, dstBounds, srcBounds );
      else
      {
         new SDBError( SDBError.Recoverable, "Unsupported DataBlock type: " 
                                       + dstBlock.getClass().getName() );
      }
   }

   /** Reads a subblock defined by srcBounds into the region of the Datablock argument
   defined by dstBounds. The RecordSpec argument specifies a subset of the DataSource
   attributes to be written into the DataBlock datums. Note that the DataBlock is 
   assumed to already have a valid data array. the DataBlock can be BasicBlock 
   or CompositeBlock */
  public void subblock( DataBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds, 
                                RecordSpec recSpec )
  {
     checkActivated(); 
     if ( recSpec == null )
        recSpec = new RecordSpec( this.getNumAttributes() );
     if ( dstBlock instanceof BasicBlock )
        subBasicBlock( (BasicBlock) dstBlock, dstBounds, srcBounds, recSpec );
     else if ( dstBlock instanceof CompositeBlock )
         subCompositeBlock( (CompositeBlock) dstBlock, dstBounds, srcBounds, recSpec );
     else
     {
        new SDBError( SDBError.Recoverable, "Unsupported DataBlock type: " 
                                      + dstBlock.getClass().getName() );
     }
  }

  /** Reads a subblock defined by srcBounds into the region of the Datablock argument
  defined by dstBounds. The FieldIDMapper argument determines the mapping between 
  datums in the datasource and datums in the DataBlock. Note that the DataBlock is
  assumed to already have a valid data array. the dataBlock can be basicBlock or compositeBlock
  */
 public void subblock( DataBlock dstBlock, ISBounds dstBounds, ISBounds srcBounds, 
                               FieldIDMapper mp )
 {
    checkActivated(); 
    if ( mp == null )
       mp = new FieldIDMapper( this.getNumAttributes() );
       
    if ( dstBlock instanceof BasicBlock )
       subBasicBlock( (BasicBlock) dstBlock, dstBounds, srcBounds, mp );
    else if ( dstBlock instanceof CompositeBlock )
       subCompositeBlock( (CompositeBlock) dstBlock, dstBounds, srcBounds, mp );
    else
    {
       new SDBError( SDBError.Recoverable, "Unsupported DataBlock type: " 
                                     + dstBlock.getClass().getName() );
    }
 }
 
   /** Package access form of subblock called by Caches that require a BasicBlock. 
    * If it does not need to be overridden, we just convert the call to the regular
    * subblock method.
    * 
    * rdb 02/12/12: This sure looks like an infinite loop! This has to be
    *               over-ridden by children.
    */
   /******
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      this.subblock( dstBlk, dstBnds, srcBnds );
   }
   /*****************************/
    /** Return a copy of the URL for this datasource.*/
    public String getURL( )
    {
        //rdb return new String( this.url );
       if ( this.fdlDescriptor != null )
          return this.fdlDescriptor.getDataFileURL();
       else
          return this.url;
    }

    /** Return a copy of the URL for this datasource.
     */
    public String getDataFileString( )
    {
       if ( this.fdlDescriptor != null )
          return this.fdlDescriptor.getDataFileString();
       else 
          return this.url;
    }
    /**/

    /** Return a String representing the object. */
    public String toString( )
    {
         return this.toString( "" ); 
    }    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString( String indent ){

        StringBuffer result=new StringBuffer( indent+"PhysicalDataSource:\n" ); 

        result.append( indent+"url="+this.url+"\n" ); 
        
        result.append( indent+"axisOrdering=" ); 
        result.append( this.axisOrdering+"\n" ); 
        
        result.append( super.toString( indent ) ); 

        return new String( result ); 
    }


    /** Returns true if this PhysicalDataSource has the same URL
        and AxisOrdering as the argument PhsyicalDataSource, as
        well as meeting inherited criteria for equality. */
    public boolean equals( Object o ){

        if( o instanceof PhysicalDataSource ){
            PhysicalDataSource d=( PhysicalDataSource ) o; 
    
            return	this.url.equals( d.url ) &&
                    this.axisOrdering.equals( d.axisOrdering ) &&
                    super.equals( d ); 
        } else {
            return false; 
        }
    }


  /** this method computes the best i/o buffer size. As a default the buffer 
   *  size will not be bigger than PhysicalDataSource.maxBufferSize because 
   *  bigger buffers take a lot of memory and reduce the performance. 
   *  
   * This is intended to work for both "normal" and chunked data sets.
   * The first argument is the bounds in units (partition space for chunked,
   * datum space for regular).
   * The second argument is the query within the spaceBounds;
   * The third argument is the size of each unit in bytes.
   * 
   *  It returns the number of units included in the buffer
   * 
   *  It tries to create a buffer that extends as far as possible along
   *  as many less significant axes as will fit in the maximum as long as
   *  what will be read, but not needed (gap) does not get too large.
   *  In other words if the cost of reading the gap is larger than a 
   *  subsequent seek over that gap, we'll stop increasing the buffer size.
   * 
   *  rdb: Complete rewrite 02/21/05
   */
    protected long makeBestBuffer( ISBounds spaceBounds, 
                                   ISBounds queryBounds, int unitSize )
   {
       // first avoid all axisOrdering issues, by mapping to a 0 1 .. d-1 order.
       // We also translate the spaceBounds to origin so we can avoid lower complic.
       //System.out.println( "QBnds: " + queryBounds );
       
       int      dim     = this.dim();
       int[]    dsLengths  = new int[ dim ];
       int[]    qLengths   = new int[ dim ];
       
       long[]   dsWeightsOrig = this.axisOrdering.makeAxisWeights( spaceBounds );
       long[]   qWeightsOrig  = this.axisOrdering.makeAxisWeights( queryBounds );
       
       long[]   dsWeights = new long[ dim ];
       long[]   qWeights  = new long[ dim ];
       
       // translate through axis axisOrdering just once
       for ( int i = 0; i < dim; i++ )
       {
          int        axis = this.axisOrdering.getAxis( i );
          dsWeights[ i ]  = dsWeightsOrig[ axis ];
          qWeights[ i ]   = qWeightsOrig[ axis ];
          dsLengths[ i ]  = spaceBounds.getDimension( axis );
          int  lower    = queryBounds.getLower( axis ) - spaceBounds.getLower( axis );
          int  upper    = queryBounds.getUpper( axis ) - spaceBounds.getLower( axis );
          qLengths[ i ] = upper - lower + 1;
       }
       
       // dsSizes[i] represents the "size" of a DS subblock of full extent
       //            in all axis >= i and of length 1 in all axes < i.
       // qSizes[i]  is the same for the query block
       int[] dsSizes = new int[ dim ];
       int[] qSizes  = new int[ dim+1 ];   // it's convenient to have an extra dimension
       
       dsSizes[ dim-1 ] = dsLengths[ dim-1 ]; 
       qSizes[ dim-1 ]  = qLengths[ dim-1 ];
       qSizes[ dim ]    = 1;
       
       for ( int i = dim-2; i >= 0; i-- )
       {
          dsSizes[ i ] = dsSizes[ i+1 ] * dsLengths[ i ]; // dsWeight[ i+1 ]
          qSizes[ i ]  = qSizes[ i+1 ]  * qLengths[ i ];  // qWeight[ i+1 ]
       }

       // gaps[i] -  # records to skip when you reach the end of query boundd while
       //         traversing the i axis -- when increment the i axis, where is next?
       //         gap for least significant axis not needed, but we just set to 0.
       //    The gap for axis i just the datasource weight[i] - query weight[i]
       long[] gaps  = new long[ dim ];
       gaps[ dim - 1 ] = 0;
       for ( int i = dim-2; i >= 0; i-- )
          gaps[ i ] =  dsWeights[ i ] - qWeights[ i ] ;
 
       // for each axis from least to most significant axis
       //   set the axis range to the number of entries that "fit" in buffer
       //
       
       int lastAxis  = 0;
       int maxUnitsInBuffer = (int) ( this.maxBufferSize / unitSize );
       int[] whatFits = new int[ dim ];
       for ( int i = 0; i < dim; i++ )
          whatFits[ i ] = 1;
       
       for ( int axis = dim-1; axis >= 0 && lastAxis == 0; axis-- )
       {
          if ( qSizes[ axis ] <= maxUnitsInBuffer )  // if entire axis fits
          {
             if ( gaps[ axis ] <= this.maxGapSize )  
             {  
                whatFits[ axis ] = qLengths[ axis ];    // it all fits
                //*************** following line is not really correct *************
                //  whether to quit or not depends on allowed gap size, but we don't
                //  really support gaps yet and I need to get the code working!
                // rdb 04/28/05
                if ( qSizes[ axis ] < dsSizes[ axis ] ) // This is last axis because of subset
                   lastAxis = axis;
             }
             else  // exit loop with this whatFits == 1
                lastAxis = axis;
          }
          else // fit as much of this axis as possible and exit
          {
             whatFits[ axis ] = maxUnitsInBuffer / qSizes[ axis + 1 ];
             if ( whatFits[ axis ] > qLengths[ axis ] ) // but no more than we have
                whatFits[ axis ] = qLengths[ axis ];
             lastAxis = axis;
          }
       }
       
       long bufferSize = whatFits[ lastAxis ] * qSizes[ lastAxis+1 ];
       
       //System.out.print( "Buffer size in units: " + bufferSize + " " + bufferBytes + "  whatfits: [" );
       //for ( int i = 0; i < dim; i++ ) System.out.print( whatFits[i] +  ", ");
       //System.out.println( "]" );
       
       // now build bufferDimension and bufferSteps arrays.
       // bufferDimension describes the shape of the buffer itself which is
       //             related to the datasource shape
       // 
       // bufferSteps identifies how much of the srcBounds is stored in
       //             the buffer. It does this by defining the "steps"
       //             array that is needed to get the position of the
       //             next data in the buffer that is needed by the next
       //             iteration through the srcBounds.
       // This is just "what fits" in each dimension.
       
       for ( int i = 0; i < dim; i++ )
       {
          this.bufferSteps[ this.axisOrdering.getAxis( i )] = whatFits[i];
          
          // bufferDimensions are the same as whatFits until we get past
          //   "lastAxis" which is the last partial (or full) one that
          //    fits.
          if ( i <= lastAxis )
             this.bufferDimension.setCoord( this.axisOrdering.getAxis( i ), whatFits[i] );
          else
             this.bufferDimension.setCoord( this.axisOrdering.getAxis( i ), dsLengths[i] );
       }
       this.bufferBounds = new ISBounds( this.bufferDimension );
              
       /**
       bufferSize = oldmakeBestBuffer( spaceBounds, queryBounds, unitSize );
       /*************************/
       return bufferSize;
   }

/************************* rdb rewrote. 02/22/05 ****************************
 * Leave it here for now, the overall code is still not working in all but
 * the simplest cases, but the old code is just as bad as the new.
 * In particular, the use of this.bufferSteps is just a mystery.
 * 
 * This comes up in ChunkedDataSource.palloc where this.BufferSteps is
 * used to create bufferIterator -- I don't understand how it works -- and of 
 * course, it doesn't throw an exception when all the data is in one buffer.
 *
  protected  long oldmakeBestBuffer( ISBounds spaceBounds, ISBounds queryBounds, 
                                  int      unitSize )   
    {
         long[] srcAxisWeights=this.axisOrdering.makeAxisWeights( spaceBounds ); 
         int leastSignificantAxis = this.axisOrdering.getLeastSignificantAxis( ); 
         IndexSpaceID dimensions  = spaceBounds.getDimensions( ); 

         IntegerIndexSpaceID lowerID = queryBounds.copyLower( ); 
         int[] lower = lowerID.getCoords( ); 
         int[] upper = queryBounds.copyUpper( ).getCoords( ); 

         long startingOffset = AxisOrdering.computeOffset( lower, srcAxisWeights ); 

         int numOfDim = spaceBounds.dim( ); 
         int startOfAxis[][] = new int[numOfDim][numOfDim]; 
         int endOfAxis [][] = new int[numOfDim][numOfDim]; 
         int newEndOfAxis[][] = new int[numOfDim][numOfDim]; 

         //AxisOrdering axisOrdering = this.getAxisOrdering( ); 
         long[] gapSize= new long[ numOfDim-1 ]; //number of datums between seeks.

         for ( int i =0; i<numOfDim-1; i++ )
          {  
             for ( int j =0; j<numOfDim; j++ )
              {
                 startOfAxis[i][j]=lower[j]; 
                 endOfAxis[i][j] = lower[j]; 
                 newEndOfAxis[i][j] = lower[j]; 
              }
          }

         int n = 1; 
         int axisNum = 0; 
         long endOffset, nextEndOffset; 

         long newStartOffset; 
         long bufferSize = queryBounds.getDimensions( ).getCoord( leastSignificantAxis ); 
         long newbufferSize =0; 
         IntegerIndexSpaceID bufferDim; 
         
         for ( int i=0; i<numOfDim; i++ )
             this.bufferSteps[ this.axisOrdering.getAxis( i ) ] = 1;  
         
         int     axis = numOfDim - 1;
         for ( int i = 0; 
                  i < numOfDim-1; 
                  i++, axis-- )
          {
             startOfAxis[ i ][ this.axisOrdering.getNthLeastSignificantAxis( n++ )] += 1; 
             for ( int j = 0; j<=i; j++ )
              {
                 axisNum = axisOrdering.getNthLeastSignificantAxis( j ); 
                 endOfAxis[i][axisNum]=upper[axisNum]; 
                 newEndOfAxis[i][axisNum] = upper[axisNum]; 
              }
                          
             endOffset = AxisOrdering.computeOffset( endOfAxis[i], srcAxisWeights ); 
             newStartOffset = AxisOrdering.computeOffset( startOfAxis[i], srcAxisWeights ); 
         
             //compute the next end point, and see if we should enlarge the buffer 
             axisNum = axisOrdering.getNthLeastSignificantAxis( i+1 ); 
             //System.out.println( " i " + i ); 
             newEndOfAxis[i][ axisNum] = upper[ axisNum ]; 
             nextEndOffset = AxisOrdering.computeOffset( endOfAxis[i], srcAxisWeights ); 
             newbufferSize = ( nextEndOffset - startingOffset +1 ); 
           
             // compute the buffer size 
             gapSize[i]= ( newStartOffset - endOffset -1 ); 
             this.bufferSteps[ this.axisOrdering.getAxis( axis ) ]
                               = dimensions.getCoord( this.axisOrdering.getAxis( axis )); 
//             
//             System.out.println( "gapSize =" +gapSize[i] ); 
//             System.out.println( "bufferSize:" + bufferSize ); 
//             System.out.println( "RodSize:" +queryBounds.getDimensions( ).getCoord( leastSignificantAxis )  ); 
             if( ( gapSize[i] > this.maxGapSize )  
                   || ((long)( newbufferSize * unitSize ) > this.maxBufferSize ))
              {
                bufferDim =  new IntegerIndexSpaceID( endOfAxis[i] ) ; 
                bufferDim.subtract( lowerID ); 
                for ( int m =0 ; m < this.dim( ); m++ )
                  this.bufferDimension.setCoord( m, bufferDim.getCoord( m ) + 1 ); // PJR added + 1

                //System.out.println( " return bufferDimension :" + out_bufferDimension ); 
                //System.out.print( " return step size :" ); 
                //printArray( out_steps ); 
                //rdb return bufferSize; 
                
                bufferSize = (int)howMuch( this.axisOrdering.getAxis( axis ), srcAxisWeights, unitSize );
                System.out.println( "Buffersize: " + bufferSize );
                return bufferSize;
              }
              else
                 bufferSize = newbufferSize;
         } 
         //this.printArray( startOfAxis ); 
         //this.printArray( endOfAxis ); 
         //out_bufferDimension =  new IntegerIndexSpaceID( upper ) ; 
         for ( int i = 0; i < this.dim( ); i++ )
            this.bufferDimension.setCoord( i, upper[i] + 1 ); // PJR added + 1

         this.bufferDimension.subtract( lowerID ); 
         
         int axis0 = this.axisOrdering.getAxis( 0 );

         this.bufferSteps[ axis0 ] = 
                                dimensions.getCoord( axis0 ); 
//         System.out.println( "return BufferSize :"+ ( AxisOrdering.computeOffset( upper, srcAxisWeights ) - startingOffset +1 ) ); 
//         System.out.println( "return bufferDimension :" + out_bufferDimension ); 
//         System.out.print( "return step size :" ); 
//         printArray( out_steps ); 

         this.bufferBounds = new ISBounds( this.bufferDimension );
         
         //rdb: I think this is not correct. Above for loop does 1 less than dimension.
         //     If we don't exit in that loop, then we know that at least 1 entire
         //     "layer" of the DS fits in buffer, but less than the full range
         //     might fit. The code below assumes that it will ALL fit.
         //
         // return AxisOrdering.computeOffset( upper, srcAxisWeights ) - startingOffset +1; 
         //
         // rdb replaced above with code below:
        return (int) howMuch( axis0, srcAxisWeights, unitSize );
    }
    /*****************************************************************/
    
    // Find how much of a particular axis (and all the axes of lesser significance
    // can be fit into the available buffer memory. 
    //
    private long howMuch( int axis, long[] srcAxisWeights, int unitSize )
    {
       long recsPerLayer = srcAxisWeights[ axis ];
       int maxLayers = (int) (this.maxBufferSize / ( recsPerLayer * unitSize ));
       if ( maxLayers > this.bufferDimension.getCoord( axis ))
          maxLayers = this.bufferDimension.getCoord( axis );
       return maxLayers * recsPerLayer;           
    }
    /********************/

    //debugging tool
    //********************************************************************
    private static void printArray( int[][] a )
      {
        for ( int i=0; i<a.length; i++ )
          {
             System.out.print( "[" ); 
             for ( int j =0; j<a[0].length; j++ )
               System.out.print( a[i][j] +" " ); 
             System.out.println( "]" ); 
          }
      }
    private static void printArray( int[] a )
    {
     System.out.print( "[" ); 
     for ( int i=0; i<a.length; i++ )
           System.out.print( a[i] +" " ); 
     System.out.println( "]" ); 
    }
    
    private static void printArray( long[] a )
    {
     System.out.print( "[" ); 
     for ( int i=0; i<a.length; i++ )
           System.out.print( a[i] +" " ); 
     System.out.println( "]" ); 
    }
    //******************************************************************
    public static void main( String[] args )
    {
       printArray( (int[][]) null );
       printArray( (int[]) null );
       printArray( (long[]) null );
    }
}
