/************************************************************ 
 PointOrderBFDS.java
 Copyright 2000 University of New Hampshire
 Durham, NH 03824

  Xuan Tang xtang@cs.unh.edu
 last update: 11/1/03
 update comments: 1. using byteOrder for endian
 ************************************************************/
/*
* ----------------- Edit history (started 11/28/04) ---------
* 03/04/12 rdb: Added constructor using FDLDescriptor as argument
* 03/28/09 jpf: Removed commented out if-else code, that had been replaced with a switch
*               statement just above.
* 03/05/09 rdb: Corrected bug in subBasicBlock where a non 000 origin in destination
*                was not properly handled. 
* 01/10/06 rdb: Deleted subBasicBlock( dstBlock, dstBounds, srcBounds ) - same as parent
* 01/08/06 rdb: Problems with subblock, noted but not fixed:
*               1. No check for mismatch between passed in destination and source blocks
*               2. No method with FieldIDMapper argument (or RecordSpec)!
* 12/04/05 rdb; added checkActivate() calls.
* 10/21/05 rdb: removed unused variables; deleted unused methods setByteArray
*                 and toBinaryString. If ever want them, use parent versions.
* 10/18/05 dlipsa: changed datum(IndexSpaceID) to use the field recordDesc
*                  instead of getRecordDescriptor() which does a clone().
* 06/01/05 rdb: rewrote datum methods so they all use datumBuffer ByteBuffer
*                 this should fix the byte order problem with datum queries.
* 05/13/05 rdb: -fixed a bug in allocateBuffer dealing with fullBufLimit variable
* 04/26/05 rdb: -made substantial changes to psubblock and other related code
*                in order to properly handle requests that are not equal 
*                multiples of the cache block size. 
*              - consolidated the RecordSpec/FieldIDMapper code
* 02/07/05 rdb: modified allocateBuffer, moved to PhysicalDataSource and used
*                 it to replace all other direct allocations of the buffer.
* 01/30/05 rdb: deleted subblock( ISBounds ). Now handled by DataCollection
* 01/08/05 rdb: removed explicit creation of Datum children
* 11/28/04 rdb: Changed to reflect introduction of MixedValues
*                 class to support mixed type records. Also explicitly
*                 listed more StorageType values.
*
***************************************************************/

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

import java.io.*;
import java.nio.*;

/** This class will read from a file of binary data, providing an n-dimensional
 datasource view of that file. this class represents data file that is in point order
 ( attributes are seperated in chunks )
 */

public class PointOrderBFDS extends BinaryFileDataSource
{
   private ByteBuffer  datumBuffer;    
   
   // Need some information about how to read the data for a rod. It can be
   // that the rod is bigger than the i/o buffer, so we may need multiple
   // reads to fill the rod
   int readsPerRod  = 1;    // how  many i/o buffers does it take to fill a rod
   int fullBufLimit = 0;    // how many bytes in a full buffer are devoted to data
   int lastBufLimit = 0;    // how many bytes of the buffer are for the last read

  //------------------ Constructors ------------------------------------
  /**
    * Create a BinaryFileDataSource from an fdl
    * 
    * @param name
    * @param fdl 
    */
   public PointOrderBFDS( String name, FDLDescriptor fdl )
   {
      this( name, 
            fdl.getRecordDescriptor(), 
            fdl.getBounds(), 
            fdl.getAxisOrdering(), 
            null,   // String URL 
            fdl.endianness
           );
   }
   /** create a PointOrderBFDS, set the initial file offset at 0 
    */
   public PointOrderBFDS( String name, RecordDescriptor recordDesc,
         ISBounds theBounds, AxisOrdering ordering, String URL, ByteOrder endian )
   {
      super( name, recordDesc, theBounds, ordering, URL, endian ); // call inherited constructor
      this.dataOrder = DataCollection.pointOrder;
   }

   /** create a PointOrderBFDS, set the initial file offset at the given value 
    */
   public PointOrderBFDS( String name, RecordDescriptor recordDesc,
         ISBounds theBounds, AxisOrdering ordering, String URL,
         ByteOrder endian, int offset )
   {
      super( name, recordDesc, theBounds, ordering, URL, endian, offset ); // call inherited constructor
      this.dataOrder = DataCollection.pointOrder;
   }

   /** Return the Datum corresponding to the given IndexSpaceID. */
   @Override
   public Datum datum( IndexSpaceID did )
   {
      Datum returnVal = recordDesc.createDatum();

      this.datum( returnVal, did );
      return returnVal;

   }

   /**
    * Fill the Datum ByteBuffer with a datum
    */
   private void readDatumBuffer( IndexSpaceID did )
   {
      checkActivated(); 
      long fileOffset;
      
      if ( this.datumBuffer == null )
      {
         this.datumBuffer = ByteBuffer.allocateDirect( this.recordByteSize );
         this.datumBuffer.order( this.endian );
      }

      try
      {
         //Position file stream
         fileOffset = AxisOrdering.computeOffset( did, this.axisWeights )
               * this.recordByteSize + this.iniOffset;
         if ( fileOffset != ( (int) raFile.getFilePointer() ) )
            raFile.seek( fileOffset );

         this.datumBuffer.clear();
         fChannel.read( this.datumBuffer );
         
         this.datumBuffer.position( 0 ); // PJR
         
         

         //*************** testing util 
         this.readCounter++;
         this.readLength += this.recordByteSize; 
         //*************************************
      } 
      catch ( IOException e )
      {
         new FileError( SDBError.Recoverable,
               "BinaryFileDataSource.datum(Datum, IndexSpaceID):: IOException: " + e );
      }      
   }

   /**
    * Fill Datum with values from the DataSource at the specified position. No
    * argument validation is performed. Package Access only.
    */
   @Override
   void pdatum( Datum datum, IndexSpaceID did )
   {
      int storageType;
      readDatumBuffer( did );
      
      for ( int i = 0; i < this.recordNumFields; i++ )
      {
         storageType = recordDesc.getStorageType( i );
         setDatum( i, storageType, datum, this.datumBuffer );
      }
   }

   /** Get the Datum and fill its values in the passed Datum at positions
    *  specified by the corresponding RecordSpec.
    */
   @Override
   public void datum( Datum datum, IndexSpaceID did, RecordSpec recSpec )
   {
      int   storageType;
      int   byteOffset;
      int[] fieldNumbers = recSpec.getFieldNumbers();
      readDatumBuffer( did );

      for ( int i = 0; i < recSpec.size(); i++ )
      {
         storageType = recordDesc.getAttribute( fieldNumbers[i] ).getStorageType();
         byteOffset  = recordDesc.byteOffset( fieldNumbers[i] );
         this.datumBuffer.position( byteOffset );  // position buffer at read point
         setDatum( i, storageType, datum, this.datumBuffer );
      }
   }

   /** Get the Datum and fill its values in the passed Datum at positions
    *  specified by the corresponding FieldIDMapper.
    */
   public void datum( Datum datum, IndexSpaceID did, FieldIDMapper mp )
   {
      int   storageType;
      int   byteOffset;
      int[] componentFields = mp.getComponentFields();
      int[] compositeFields = mp.getCompositeFields();
      readDatumBuffer( did );

      for ( int i = 0; i < mp.size(); i++ )
      {
         storageType = recordDesc.getAttribute( componentFields[i] ).getStorageType();
         byteOffset  = recordDesc.byteOffset( componentFields[i] );
         this.datumBuffer.position( byteOffset );  // position buffer at read point
         setDatum( compositeFields[i], storageType, datum, this.datumBuffer );
      }
   }

   /** Return the Datum corresponding to the given IndexSpaceID. The 
    Datum will only contain attributes specified by the RecordSpec
    argument. 
    */
   @Override
   public Datum datum( IndexSpaceID did, RecordSpec spec )
   {
      checkActivated(); 
      int specsize = spec.size();
      Datum returnVal = this.getRecordDescriptor().createDatum( spec );

      this.datum( tempDatum, did );

      for ( int i = 0; i < specsize; i++ )
         returnVal.setFloat( i, tempDatum.getFloat( spec.getFieldNumber( i )));

      return returnVal;
   }

   /**
    * allocate the I/O buffer and set up information for multiple reads if
    * needed.
    */
   private void allocateIOBuffer( int numUnits, int unitSize )
   {
      int rodByteSize = numUnits * unitSize;
      if ( allocateBuffer( rodByteSize ))
      {
         this.fullBufLimit = this.lastBufLimit =getBufferSize();
         this.readsPerRod = 1;         
      }
      else
      {
         // we'll need multiple reads of this one buffer.
         int bufSize = getBufferSize();
         
          // first make sure that we only deal with complete records in each buffer read
         if ( bufSize % this.recordByteSize != 0 )
         {
            this.fullBufLimit = ( bufSize / this.recordByteSize ) * this.recordByteSize;
            getBuffer().limit( this.fullBufLimit );
         }
         else 
            this.fullBufLimit = bufSize;

         this.readsPerRod = rodByteSize / this.fullBufLimit;
         
         if ( rodByteSize / this.fullBufLimit != 0 )
         {
            this.lastBufLimit = rodByteSize % this.fullBufLimit;
            this.readsPerRod++;
         }
      }
   }

   // basicBlock implementation java 1.4 
   // dstBlock is considered as composite, the calling binaryFileDataSource is considered as component
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
                                         ISBounds inputSrcBounds, FieldIDMapper mp )
   {
   
      //System.out.println("inputSrcBounds="+inputSrcBounds);
      
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );
      
      int[] intData = null;
      float[] floatData = null;
      double[] doubleData = null;
      short[] shortData = null;
      byte[] byteData = null;

      int dataArrayOffset;
      long fileOffset;
      int compositeNumFields = dstBlock.getRecordDescriptor().size();
      int numFields = mp.size();
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis ); 
      
      // file data
      long dataArrayStep = dstBlock.getAxisWeight( this.leastSignificantAxis );
      ISIterator iter = new ISIterator( srcBounds, this.axisOrdering, this.steps );

      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );
      int type;
      int datumPos, fieldPos;
      int datumByteSize = this.recordDesc.byteSize();
      int[] componentFields;
      int[] compositeFields;

      // get the data array of dstblock
      int dataType = dstBlock.getStorageType();
      
      switch(dataType){
      
         case StorageType.Int:
            intData = ( (IntValues) dstBlock.getValues() ).getDataReference();
            break;
            
         case StorageType.Float:
            floatData = ( (FloatValues) dstBlock.getValues() ).getDataReference();
            break;
            
         case StorageType.Short:
         case StorageType.Unsigned_Short:
            shortData = ( (ShortValues) dstBlock.getValues() ).getDataReference();
            break;
            
         case StorageType.Double:
            doubleData = ( (DoubleValues) dstBlock.getValues() ).getDataReference();
            break;
            
         case StorageType.Record:
            byteData = ( (MixedValues) dstBlock.getValues() ).getDataReference();
            break;
            
         case StorageType.Byte:
         case StorageType.Char:
         case StorageType.Unsigned_Byte:
         case StorageType.Boolean:
            byteData = ( (ByteValues) dstBlock.getValues() ).getDataReference();
            break;
            
         default: 
            new InternalFaultError( SDBError.Recoverable,
            "Unknown or unsupported Storage Type." );
            break;

      }

      //rdb getBuffer() = ByteBuffer.allocateDirect( rodSize * this.recordByteSize );
      allocateIOBuffer( rodSize,  this.recordByteSize );
      
      try
      {
         componentFields = mp.getComponentFields();
         compositeFields = mp.getCompositeFields();

         for ( iter.init(); iter.valid(); iter.next() )
         {

            fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                  * this.recordByteSize + this.iniOffset;

            iterCopy = new IntegerIndexSpaceID( iter );
            dataArrayOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy ));

            //int datumsInBuff = this.fullBufLimit / this.recordByteSize;
            
            for ( int b = 0; b < this.readsPerRod; b++ )
            {
               if ( fileOffset != ( (int) raFile.getFilePointer() ) )
                  raFile.seek( fileOffset );
               
               getBuffer().clear();
               
               // have to set the buffer limit after the clear operation
               if ( b < this.readsPerRod - 1 ) 
                  getBuffer().limit( this.fullBufLimit );  // not last read                
               else  // last read
               {
                  //datumsInBuff = this.lastBufLimit / this.recordByteSize;
                  getBuffer().limit( this.lastBufLimit );                  
               }

               int bytesRead = fChannel.read( getBuffer() );

               //*************** testing util 
               readCounter++;
               readLength += bytesRead; 
               //******************************************
             
               fieldPos = 0;
               datumPos = 0;

               for ( int i = 0; i < rodSize; i++ , dataArrayOffset += dataArrayStep )
               {
                  for ( int j = 0; j < numFields; j++ )
                  {
                     fieldPos = recordDesc.byteOffset( componentFields[j] );
                     type = recordDesc.getStorageType( componentFields[j] );
                     // or type=recordDesc.getStorageType(j);

                     int offset = dataArrayOffset * compositeNumFields
                        + compositeFields[j];
                        
                     switch(type){
                     
                        case StorageType.Int:
                           intData[offset] = getBuffer().getInt( datumPos + fieldPos );
                           break;
                           
                        case StorageType.Short:
                           shortData[offset] = getBuffer().getShort( datumPos + fieldPos );
                           break;
                           
                        case StorageType.Double:
                           doubleData[offset] = getBuffer().getDouble( datumPos + fieldPos );
                           break;
                           
                        case StorageType.Float:
                           floatData[offset] = getBuffer().getFloat( datumPos + fieldPos );
                           break;
                           
                        case StorageType.Byte:
                        case StorageType.Char:
                        case StorageType.Unsigned_Byte:
                           byteData[offset] = getBuffer().get( datumPos + fieldPos );
                           break;
                           
                        case StorageType.Record:
                           new InternalFaultError( SDBError.Recoverable,
                           "Records are not yet implemented." );
                           break;
                           
                        default:
                           new InternalFaultError( SDBError.Recoverable,
                           "Unknown Storage Type." );
                           break;
                     
                     
                     }
                     
                  }
                  datumPos += datumByteSize;
               }
            }
         }
      } catch ( IOException e )
      {
         new FileError( SDBError.Recoverable,  "IO Exception in " +
         		"BinaryFileDataSource.subblock(DataBlock, ISBounds, ISBounds,FieldIDMapper)::"
                     + e );
      }
   }

   // compositeBlock implementation 
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
                                   ISBounds inputSrcBounds, FieldIDMapper mp )
   {
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );
      

      int datumArrayOffset;
      long fileOffset;
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis ); 
      int datumArrayStep = (int) dstBlock.getAxisWeight( this.leastSignificantAxis );

      ISIterator iter = new ISIterator( srcBounds, this.axisOrdering, this.steps );
      
      //rdb getBuffer() = ByteBuffer.allocateDirect( rodSize * this.recordByteSize );
      allocateIOBuffer( rodSize, this.recordByteSize );

      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );

      try
      {
         for ( iter.init(); iter.valid(); iter.next() )
         {
            iterCopy.assign( iter );
            fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                                     * this.recordByteSize + this.iniOffset;
            datumArrayOffset = dstBlock.computeOffset( boundsMapper.toComposite( iterCopy ) );

            if ( fileOffset != ( (int) raFile.getFilePointer() ) )
               raFile.seek( fileOffset );
                        
            int bufUnitSize = this.fullBufLimit / this.recordByteSize;
            
            for ( int b = 0; b < this.readsPerRod; b++ )
            {  
               getBuffer().clear();
               
               if ( b < this.readsPerRod - 1 )
                  getBuffer().limit( this.fullBufLimit ); // not last read
               else                // last read
               {
                   getBuffer().limit( this.lastBufLimit );
                   bufUnitSize = this.lastBufLimit / this.recordByteSize;
               }
                  
               fChannel.read( getBuffer() );
               
               dstBlock.setrod( getBuffer(), bufUnitSize, datumArrayOffset,
                                             datumArrayStep, mp, this.recordDesc );
            }
         }
      } catch ( IOException e )
      {
         new FileError(
               SDBError.Recoverable,
               "IO Exception in BinaryFileDataSource.subblock(DataBlock, ISBounds, ISBounds,FieldIDMapper)::"
                     + e );
      }
   }

   // this is based on a new java 1.4 
   void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
         ISBounds inputSrcBounds, RecordSpec recSpec )
   {
      subBasicBlock( dstBlock, dstBounds, inputSrcBounds, 
                                    new FieldIDMapper( recSpec ));
      
      /*********************** rdb refactor *************************
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );
      
      int dataType = dstBlock.getStorageType();

      // get the data array of dstBlock
      int[] intData = null;
      float[] floatData = null;
      double[] doubleData = null;
      short[] shortData = null;
      byte[] byteData = null;

      if ( dataType == StorageType.Int )
         intData = ( (IntValues) dstBlock.getValues() ).getDataReference();
      else if ( dataType == StorageType.Float )
         floatData = ( (FloatValues) dstBlock.getValues() ).getDataReference();
      else if ( dataType == StorageType.Short )
         shortData = ( (ShortValues) dstBlock.getValues() ).getDataReference();
      else if ( dataType == StorageType.Double )
         doubleData = ( (DoubleValues) dstBlock.getValues() ).getDataReference();
      else // the data must be store in byte array
      byteData = ( (ByteValues) dstBlock.getValues() ).getDataReference();

      int dataArrayOffset, i, j, componentIndex;
      long fileOffset;
      int numFields = recSpec.size();
      int dst_numFields = dstBlock.getNumAttributes();

      int rodSize = srcBounds.getDimensions().getCoord(
            this.leastSignificantAxis ); // a rod is a sequence of contiguous file data
      long dataArrayStep = dstBlock.getAxisWeight( this.leastSignificantAxis );
      ISIterator iter = new ISIterator( srcBounds, this.ordering, this.steps );

      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );
      int type;
      int datumPos, fieldPos;

      int[] fieldNumbers = recSpec.getFieldNumbers();

      //rdb getBuffer() = ByteBuffer.allocateDirect( rodSize * this.recordByteSize );
      allocateBuffer( rodSize * this.recordByteSize );
      
      try
      {
         for ( iter.init(); iter.valid(); iter.next() )
         {
            fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                  * this.recordByteSize + this.iniOffset;
            iterCopy.assign( iter );

            dataArrayOffset = dstBlock.computeOffset( boundsMapper
                  .toComposite( iterCopy ) );

            //System.out.println("     Reading a rod of "+rodSize+" datums from location "+iter);
            if ( fileOffset != ( (int) raFile.getFilePointer() ) )
               raFile.seek( fileOffset );
            getBuffer().clear();
            fChannel.read( getBuffer() );

            fieldPos = 0;
            datumPos = 0;

            for ( i = 0; i < rodSize; i++ , dataArrayOffset += dataArrayStep )
            {
               //read the datum
               for ( j = 0; j < numFields; j++ )
               {
                  fieldPos = recordDesc.byteOffset( fieldNumbers[j] );
                  type = recordDesc.getStorageType( fieldNumbers[j] );
                  int offset = dataArrayOffset * dst_numFields + j;

                  if ( type != StorageType.Record )
                  {
                     if ( type == StorageType.Int )
                        intData[offset] = getBuffer().getInt( datumPos
                              + fieldPos );
                     else if ( type == StorageType.Short )
                        shortData[offset] = getBuffer().getShort( datumPos
                              + fieldPos );
                     else if ( type == StorageType.Double )
                        doubleData[offset] = getBuffer().getDouble( datumPos
                              + fieldPos );
                     else if ( type == StorageType.Float )
                        floatData[offset] = getBuffer().getFloat( datumPos
                              + fieldPos );
                     else if ( type == StorageType.Byte )
                        byteData[offset] = getBuffer().get( datumPos + fieldPos );
                  }
                  else
                  {
                     new InternalFaultError( SDBError.Recoverable,
                           "Records are not yet implemented." );
                  }
               }
               datumPos += recordByteSize;
               // System.out.println("Datum is "+data[dataArrayOffset]);
            }
         }
      } catch ( IOException e )
      {
         new FileError(
               SDBError.Recoverable,
               "IO Exception in BinaryFileDataSource.subblock(DataBlock, ISBounds, ISBounds,FieldIDMapper)::"
                     + e );
      }
      /****************************************************************/
   }

   // composite implementation
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
         ISBounds inputSrcBounds, RecordSpec recSpec )
   {
      // rdb; for now will map to FieldIDMapper, though it is not entirely correct
      subCompositeBlock( dstBlock, dstBounds, inputSrcBounds, new FieldIDMapper( recSpec ));
      
      /*********************** rdb refactor *************************
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );
      

      int datumArrayOffset;
      long fileOffset;
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis );
      int datumArrayStep = (int) dstBlock.getAxisWeight( this.leastSignificantAxis );

      ISIterator iter = new ISIterator( srcBounds, this.ordering, this.steps );
      //rdb getBuffer() = ByteBuffer.allocateDirect( rodSize * this.recordByteSize );
      allocateBuffer( rodSize * this.recordByteSize );

      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );

      try
      {
         for ( iter.init(); iter.valid(); iter.next() )
         {

            iterCopy.assign( iter );
            fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                  * this.recordByteSize + this.iniOffset;
            datumArrayOffset = dstBlock.computeOffset( boundsMapper
                  .toComposite( iterCopy ) );

            if ( fileOffset != ( (int) raFile.getFilePointer() ) )
               raFile.seek( fileOffset );
            getBuffer().clear();
            fChannel.read( getBuffer() );

            dstBlock.setrod( getBuffer(), rodSize, datumArrayOffset,
                  datumArrayStep, recSpec, this.recordDesc );
         }
      } catch ( IOException e )
      {
         new FileError(
               SDBError.Recoverable,
               "IO Exception in BinaryFileDataSource.subblock(DataBlock, ISBounds, ISBounds,FieldIDMapper)::"
                     + e );
      }
      /********************************************************************/
   }
   


   /** This function is used by SDBServer, it extracts a byte array from disk 
    * regardless the type of attributes. To improve the performance, the user has 
    * to allocate the byte array in advance according to the size of the query bounds. 
    * The query block has the same axisOrdering as the datasource. 
    * 
    * In this function, we assume that the bounds of dstBlock equals the srcBounds. 
    * Therefore, we only need to copy rods from source, and put them contiguously 
    * in the dst array one rod after another, there is no gaps between rods in the 
    * dst array.      
    */   
   protected void rsubblock( byte[] subSet, ISBounds inputSrcBounds){
   
      server_subblock(subSet, inputSrcBounds, 0);
   }


   /** Starting at the position indicated by bytesOffset, fill the byte array with the
    *  the data corresponding to the given ISBounds. Data is written to the array in
    *  a contiguous manner using the DataSource AxisOrdering.
    */
   public void old_server_subblock( byte[] subSet, ISBounds inputSrcBounds, int bytesOffset )
   {     
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );

      long fileOffset;
      int dstIndex = bytesOffset;

      // a rod is a sequence of contiguous file data
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis ); 
      int rodByteSize = rodSize * this.recordByteSize;
      //byte[] rodArray=new byte[rodByteSize];
      //rdb getBuffer() = ByteBuffer.allocateDirect( rodByteSize );
      allocateIOBuffer( rodSize, this.recordByteSize );
      


      ISIterator iter = new ISIterator( srcBounds, this.axisOrdering, this.steps );
      try
      {
         for ( iter.init(); iter.valid(); iter.next() )
         {
            fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                  * this.recordByteSize + this.iniOffset;

            if ( fileOffset != ( (int) raFile.getFilePointer() ) )
            {
               //System.out.println(" fileOffset is : " + fileOffset );
               raFile.seek( fileOffset );
            }
            
            int bufSize = this.fullBufLimit / this.recordByteSize;
            
            for ( int b = 0; b < this.readsPerRod; b++ )
            {  
               getBuffer().clear();
               if ( b < this.readsPerRod - 1 )
                  getBuffer().limit( this.fullBufLimit );  // not last read
               else
               {
                   getBuffer().limit( this.lastBufLimit );
                   bufSize = this.lastBufLimit;
               }
               
               fChannel.read( getBuffer() );
               getBuffer().rewind();
                   
               getBuffer().get( subSet, dstIndex, rodByteSize );

               dstIndex += bufSize;
            }
         }
      } catch ( IOException e )
      {
         new FileError( SDBError.Recoverable, "IO Exception::" + e );
      }
      //float[] testf = new float[subSet.length/4];
      //RemoteBFDS.bytesToFloats( testf, subSet);
      //for ( int i =0 ; i< testf.length; i++)
      //System.out.println(testf[i]);
   }


   public void server_subblock( byte [] bytes, ISBounds inputSrcBounds, int bytesoffset )
   {
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );

      int dataArrayOffset= bytesoffset;
      long fileOffset;
      int numFields = this.recordDesc.getNumAttributes();
      
      // a rod is a sequence of contiguous file data
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis ); 
      
      if ( ! allocateBuffer( srcBounds ))
      {
         System.err.println( "POBFDS.psubblock: buffer too small -- unpredictable results");
         rodSize = getBuffer().capacity() / this.recordByteSize;
      }
      
      int rodElementSize = rodSize * numFields;
      long rodFileOffset;
      int bufferPosition;

      
      ISBoundsIterator bufBndsIter =  new ISBoundsIterator( srcBounds, 
                                    this.bufferBounds, this.axisOrdering );

      ISIterator inBufIter = new ISIterator( this.bounds, this.axisOrdering, this.steps );
     
      int fullBufDataByteSize = this.recordByteSize * bufBndsIter.volume();
      try
      {          
         for ( bufBndsIter.init(); bufBndsIter.valid();  bufBndsIter.next() )
         {               
            long datumOffset =
                  AxisOrdering.computeOffset( bufBndsIter.getLower(), this.axisWeights );
            
            fileOffset = datumOffset * this.recordByteSize + this.iniOffset;

            //if( fileOffset != ( (int)raFile.getFilePointer()) )
            if ( fileOffset != ( raFile.getFilePointer() ) )
               raFile.seek( fileOffset );
            
            getBuffer().clear();
            
            if ( bufBndsIter.isFullShape() )
               getBuffer().limit( fullBufDataByteSize );
            else
               getBuffer().limit( bufBndsIter.volume() * this.recordByteSize );
            
            int bytesRead = fChannel.read( getBuffer() );

            if ( bytesRead <= 0 )
               new SDBError( SDBError.Warning, "PointOrderBFDS.server_sublock: " +
                                 bytesRead + " bytes read on fChannel.read." );
            
 
            //*************** testing util 
            readCounter++;
            readLength += bytesRead; 
            //******************************************

            inBufIter.setBounds( bufBndsIter );

            for ( inBufIter.init(); inBufIter.valid(); inBufIter.next() )
            {
               rodFileOffset = AxisOrdering.computeOffset( inBufIter, this.axisWeights )
                                             * this.recordByteSize + this.iniOffset;
               bufferPosition = (int) ( rodFileOffset - fileOffset );

               getBuffer().position( bufferPosition );

               getBuffer().get( bytes, dataArrayOffset, rodElementSize );

               dataArrayOffset += rodElementSize;
            }
         }// end of buffer iter
      } catch ( IOException e )
      {
//         System.out.println( "Float buffer position " + floatBuffer.position() );
//         System.out.println( "Float buffer limit " + floatBuffer.limit() );
//         System.out.println( "Float buffer capacity " + floatBuffer.capacity() );
         new FileError( SDBError.Recoverable,
               "IO Exception in PointOrderBFDS.psubblock(DataBlock, ISBounds, ISBounds)::"
                     + e );
      }
   }





   /**
    * This subblock method will check if gap between two neighboring rods (in
    * data storage order) is smaller than the argument seekcost. The seek cost
    * is measured by the number of index spaces in the grid space. If the gap is
    * smaller than the seekcost, a long buffer is allocated and one o/s read
    * operation is used to continuously read the two rods including the gap
    * between the two rods.
    * 
    * This means that given a slice query, if the gaps between neighboring rods
    * are smaller than the seekCost, a long buffer is allocated, and both rods
    * and gaps between rods are read in the buffer using a single O/S read
    * operation. Then only the rod data is copied from the buffer into the dst
    * datablock. The gap data is disregarded.
    * 
    * This method also handles n-d queries in the same fashion. The goal is to
    * improve the I/O performance by reducing the number of O/S reads.
    */

   @Override
   protected void subBasicBlock( BasicBlock dstBlock, ISBounds dstBounds,
                             ISBounds   inputSrcBounds )
   {

      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );   
      int dataType = dstBlock.getStorageType();
      int[] intData = null;
      float[] floatData = null;
      double[] doubleData = null;
      short[] shortData = null;
      byte[] byteData = null;

      int dataArrayOffset;
      long fileOffset;
      int numFields = this.recordDesc.getNumAttributes();
      
      // a rod is a sequence of contiguous file data
      int rodSize = srcBounds.getDimensions().getCoord( this.leastSignificantAxis ); 
      
      //rdb this.allocateBuffer( srcBounds );
      //allocateIOBuffer( rodSize, this.recordByteSize );
      if ( ! allocateBuffer( srcBounds ))
      {
         System.err.println( "POBFDS.psubblock: buffer too small -- unpredictable results");
         rodSize = getBuffer().capacity() / this.recordByteSize;
      }
      
      int rodElementSize = rodSize * numFields;
      FloatBuffer floatBuffer = null;
      ShortBuffer shortBuffer = null;
      IntBuffer intBuffer = null;
      DoubleBuffer doubleBuffer = null;

      //long dataArrayStep = dstBlock.getAxisWeight( this.leastSignificantAxis );
      //System.out.println(" DataArrayStem is :" + dataArrayStep);
      long rodFileOffset;
      int bufferPosition;

      // rdb: 03/05/09 added copyToOrigin() to account for non 0 destination origin
      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds.copyToOrigin() );

      // get the dataArray from the BasicBlock
      // let's use a switch here, since they are faster than ifs--PJR
      switch ( dataType )
      {
         case StorageType.Int:
            intBuffer = getBuffer().asIntBuffer();
            intData = ( (IntValues) dstBlock.getValues() ).getDataReference();
            break;
         case StorageType.Float:
            floatBuffer = getBuffer().asFloatBuffer();
            floatData = ( (FloatValues) dstBlock.getValues() ).getDataReference();
            break;
         case StorageType.Short:
         case StorageType.Unsigned_Short:
            shortBuffer = getBuffer().asShortBuffer();
            shortData = ( (ShortValues) dstBlock.getValues() ).getDataReference();
            break;
         case StorageType.Double:
            doubleBuffer = getBuffer().asDoubleBuffer();
            doubleData = ( (DoubleValues) dstBlock.getValues() ).getDataReference();
            break;
         case StorageType.Boolean:
         case StorageType.Char:
         case StorageType.Byte:
         case StorageType.Unsigned_Byte:
            byteData = ( (ByteValues) dstBlock.getValues() ).getDataReference();
            break;
         case StorageType.Record:
            byteData = ( (MixedValues) dstBlock.getValues() ).getDataReference();
            break;
         default:
            new InternalFaultError( SDBError.Fatal,
                  "Unknown storage type in PointOrderBFDS.pssublock()" );
      }

      //rdb: would it be better for bufferIter to be an ISBounds iterator?
      //ISIterator bufferIter = new ISIterator( srcBounds, this.ordering, this.bufferSteps );
      
      ISBoundsIterator bufBndsIter =  new ISBoundsIterator( srcBounds, 
                                    this.bufferBounds, this.axisOrdering ); 

      ISIterator inBufIter = new ISIterator( this.bounds, this.axisOrdering, this.steps );     

      int fullBufDataByteSize = this.recordByteSize * bufBndsIter.volume();
      try
      {          
         //for ( bufferIter.init(); bufferIter.valid(); bufferIter.next() )
         for ( bufBndsIter.init(); bufBndsIter.valid();  bufBndsIter.next() )
         {     
            //rdb: Need to check if the bufferIterator specification covers
            //     an entire I/O buffer. (That's why bufferIter as ISBIterator
            //     might be simpler?
            long datumOffset =
                  AxisOrdering.computeOffset( bufBndsIter.getLower(), this.axisWeights );
  
            fileOffset = datumOffset * this.recordByteSize + this.iniOffset;

            //if( fileOffset != ( (int)raFile.getFilePointer()) )
            if ( fileOffset != ( raFile.getFilePointer() ) )
               raFile.seek( fileOffset );
            
            getBuffer().clear();
            
            if ( bufBndsIter.isFullShape() )
               getBuffer().limit( fullBufDataByteSize );
            else
               getBuffer().limit( bufBndsIter.volume() * this.recordByteSize );
            
            int bytesRead = fChannel.read( getBuffer() );
            if ( bytesRead <= 0 )
               new SDBError( SDBError.Warning, "PointOrderBFDS.psubblock: " +
                             bytesRead + " bytes read on fChannel.read." );
             
            //*************** testing util 
            readCounter++;
            readLength += bytesRead; 
            //******************************************

            //bufBounds.moveTo( bufferIter );
            //iter.setBounds( bufBounds );
            
            inBufIter.setBounds( bufBndsIter );
          
            for ( inBufIter.init(); inBufIter.valid(); inBufIter.next() )
            {
               rodFileOffset = AxisOrdering.computeOffset( inBufIter, this.axisWeights )
                                             * this.recordByteSize + this.iniOffset;
               bufferPosition = (int) ( rodFileOffset - fileOffset );
               getBuffer().position( bufferPosition );                                             

               dataArrayOffset = this.numAttributes * 
                   dstBlock.computeOffset( boundsMapper.toComposite( inBufIter )); 
  
               //int byteDataLength = byteData.length; // debug

               switch ( dataType )
               {
                  case StorageType.Float:
                     floatBuffer.position( bufferPosition / 4 );
                     floatBuffer.get( floatData, dataArrayOffset, rodElementSize );
                     break; 
                  case StorageType.Short:
                     shortBuffer.position( bufferPosition / 2 );
                     shortBuffer.get( shortData, dataArrayOffset, rodElementSize );
                     break;
                  case StorageType.Int:
                     intBuffer.position( bufferPosition / 4 );
                     intBuffer.get( intData, dataArrayOffset, rodElementSize );
                     break;
                  case StorageType.Double:
                     doubleBuffer.position( bufferPosition / 8 );
                     doubleBuffer.get( doubleData, dataArrayOffset, rodElementSize );
                     break;
                  case StorageType.Byte:
                     if ( dataArrayOffset + rodElementSize <=  byteData.length )
                        getBuffer().get( byteData, dataArrayOffset, rodElementSize );
                     else
                     {
                        new SDBError(SDBError.Warning, "ERROR in psubblock: data array " +
                        		                 "exceeded on copy from buffer");
                     }
                     break;

                  default:
                     new InternalFaultError( SDBError.Fatal,
                           "Unimplemented storage type in PointOrderBFDS.pssublock()" );
                     break;
               }

               //System.out.println("************ floatBuffer limit is :" + floatBuffer.limit() );              
               //floatBuffer.clear();

            }
            //System.out.println( "bad#, lastBad: " + bad + " " + lastBad );
         }// end of buffer iter
      } catch ( IOException e )
      {
         // System.out.println( "Float buffer position " + floatBuffer.position() );
         // System.out.println( "Float buffer limit " + floatBuffer.limit() );
         // System.out.println( "Float buffer capacity " + floatBuffer.capacity() );
         new FileError( SDBError.Recoverable,
               "IO Exception in PointOrderBFDS.psubblock(DataBlock, ISBounds, ISBounds)::"
                     + e );
      }
      //        catch (Exception e){

      //           System.out.println("Exception in pssublock:"+e);
      //       }
      //System.out.println("Exited psubblock()." );
   }



   /** Reads a subblock defined by srcBounds into the region of the Datablock argument
    defined by dstBounds. Note that the DataBlock is assumed to already have a valid data
    array.
    NOTE: DataBlock must be a CompositeBlock 
    */
   void subCompositeBlock( CompositeBlock dstBlock, ISBounds dstBounds,
         ISBounds inputSrcBounds )
   {
      //rdb: for now map to RecordSpec method which just maps to FieldIDMapper
      //     however, the rs to fim mapping isn't right yet. If it's not
      //     fixed we'll have to go back to separate methods for rs and fim.
      //     when that happens this is the right code.
      subCompositeBlock( dstBlock, dstBounds, inputSrcBounds, 
                                    new RecordSpec( this.getNumAttributes() ));
     
      /******************* rdb refactor ***************************
      // need to translate the src specification so it is relative to the 
      // "conceptual" origin of the DataSource. 
      ISBounds srcBounds = inputSrcBounds.copy().makeRelativeTo( this.bounds.lower );
      
      int datumArrayOffset;
      long fileOffset;

      // a rod is a sequence of contiguous file data
      int rodSize = srcBounds.getDimensions().getCoord(
            this.leastSignificantAxis );
      int datumArrayStep = (int) dstBlock
            .getAxisWeight( this.leastSignificantAxis );

      ISIterator iter = new ISIterator( srcBounds, this.ordering, this.steps );
      //rdb getBuffer() = ByteBuffer.allocateDirect( rodSize * this.recordByteSize );
      allocateBuffer( rodSize * this.recordByteSize );

      ISIDMapper boundsMapper = new ISIDMapper( srcBounds, dstBounds );

      try
      {
         for ( iter.init(); iter.valid(); iter.next() )
         {

            iterCopy.assign( iter );
            fileOffset = AxisOrdering.computeOffset( iter, this.axisWeights )
                  * this.recordByteSize + this.iniOffset;
            datumArrayOffset = dstBlock.computeOffset( boundsMapper
                  .toComposite( iterCopy ) );

            if ( fileOffset != ( (int) raFile.getFilePointer() ) )
               raFile.seek( fileOffset );
            getBuffer().clear();
            fChannel.read( getBuffer() );

            dstBlock.setrod( getBuffer(), rodSize, datumArrayOffset,
                  datumArrayStep, this.recordDesc );
         }
      } catch ( IOException e )
      {
         new FileError( SDBError.Recoverable,
               "IO Exception: BinaryFileDataSource.subblock(DataBlock, ISBounds, ISBounds,FieldIDMapper)::"
                     + e );
      }
      /************************************************************************/
   }


   /** Print the data members of the DataSource in an easy to read fashion.*/
   @Override
   public String toString()
   {

      StringBuffer result = new StringBuffer( "PointOrderBinaryFileDataSource:\n" );

      result.append("readsPerRod ="); result.append(readsPerRod); result.append("\n");
      result.append("fullBufLimit ="); result.append(fullBufLimit); result.append("\n");
      result.append("lastBufLimit ="); result.append(lastBufLimit); result.append("\n");

      result.append( super.toString() );

      return new String( result );
   }

   /** Returns true if this BinaryFileDataSource has the same
    pointOrder as the argument BinaryFileDataSource, as
    well as meeting inherited criteria for equality. */
   @Override
   public boolean equals( Object o )
   {

      if ( o instanceof PointOrderBFDS )
      {

         PointOrderBFDS d = (PointOrderBFDS) o;

         return this.endian.equals( d.endian ) && super.equals( d );

      }
      else
      {
         return false;
      }

   }

   @Override
   public int hashCode ()
   {
      throw new UnsupportedOperationException ("Not yet implemented");
   }

   /*  
    static public void main(String[] args){
    
    
    System.out.println("\n******************************\n***BinaryFileDataSource Tests***\n******************************");
    
    // Let's use the first test file to start with
    AttributeDescriptor[] attrDescriptors=new AttributeDescriptor[]{new AttributeDescriptor(StorageType.Float,"Carbon")};
    RecordDescriptor recordDesc=new RecordDescriptor(attrDescriptors); 

    

    AttributeDescriptor attrDesc;
    IndexSpaceID dimensions;
    AxisOrdering axisOrdering;
    ISBounds ISBounds;
    BinaryFileDataSource ds1;
    
    int i,j,k;
    Datum d;
    int numAttributes;

    //attrDesc=new AttributeDescriptor(StorageType.Float,"Carbon");
    //recordDesc.add(attrDesc);
    

    // One D Test
    System.out.println("\n\n***One Dimensional Test***\n");
    
    axisOrdering=new AxisOrdering(1);
    ISBounds=new ISBounds(new IntegerIndexSpaceID(0),new IntegerIndexSpaceID(15));
    

    ds1=new BinaryFileDataSource("foo",recordDesc, ISBounds,axisOrdering,"testdata/test2.16.1.bin",true);

    
    ds1.activate();
    
    
    ISIterator iter=new ISIterator(ISBounds);

    System.out.println("ds1.datum(<0>) should be {0.0}-->"+ds1.datum(new IntegerIndexSpaceID(0)));    
    System.out.println("ds1.datum(<1>) should be {1.0}-->"+ds1.datum(new IntegerIndexSpaceID(1)));  
    System.out.println("ds1.datum(<7>) should be {3.1}-->"+ds1.datum(new IntegerIndexSpaceID(7)));    
    System.out.println("ds1.datum(<15>) should be {3.3}-->"+ds1.datum(new IntegerIndexSpaceID(15)));    

    //One D Multiple Attributes
    System.out.println("\n\n***One Dimensional Test Multiple Attributes***\n");
    
    //Syntax example        
    attrDescriptors=new AttributeDescriptor[]{	new AttributeDescriptor(StorageType.Float,"Carbon"),
    new AttributeDescriptor(StorageType.Float,"Nitrogen"),
    new AttributeDescriptor(StorageType.Float,"Oxygen"),
    new AttributeDescriptor(StorageType.Float,"Bogusium")
    };
    
    RecordDescriptor fourFloats=new RecordDescriptor(attrDescriptors); 

    ds1=new BinaryFileDataSource("bar",fourFloats, ISBounds,axisOrdering,"testdata/simple.6.4.bin",true);
    ds1.activate();
    
    
    int[] specs=new int[]{3,2,0,1};
    RecordSpec rspec=new RecordSpec(specs);
    
    
    
    System.out.println("ds1.datum(<0>) should be {300.0,200.0,0.0,100.0}-->"+ds1.datum(new IntegerIndexSpaceID(0),rspec));    
    System.out.println("ds1.datum(<7>) should be {307.0,207.0,7.0,107.0}-->"+ds1.datum(new IntegerIndexSpaceID(7),rspec));    
    System.out.println("ds1.datum(<15>) should be {315.0,215.0,15.0,115.0}-->"+ds1.datum(new IntegerIndexSpaceID(15),rspec));    
    
    

    
    
    // Two D Test
    System.out.println("\n\n***Two Dimensional Test***\n");
    
    ISBounds=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(7,7));
    //int TwoD_Ordering[]={0,1};
    axisOrdering=new AxisOrdering(2);
    
    attrDescriptors=new AttributeDescriptor[]{new AttributeDescriptor(StorageType.Float,"Carbon")};
    recordDesc=new RecordDescriptor(attrDescriptors); 
    
    ds1=new BinaryFileDataSource("baz",recordDesc, ISBounds,axisOrdering,"testdata/simple.6.1.bin",true);
    
    ds1.activate();
    
    System.out.println("DS Ordering="+axisOrdering);
    
    System.out.println("ds1.datum(<0,0>) should be {0.0}-->"+ds1.datum(new IntegerIndexSpaceID(0,0)));    
    System.out.println("ds1.datum(<0,1>) should be {1.0}-->"+ds1.datum(new IntegerIndexSpaceID(0,1)));    
    System.out.println("ds1.datum(<1,0>) should be {8.0}-->"+ds1.datum(new IntegerIndexSpaceID(1,0)));    

    iter=new ISIterator(ISBounds, new AxisOrdering(2));

    System.out.println();
    
    System.out.println(ds1.dumpData());

    
    System.out.println("2D test of datum(Datum, IndexSpaceID, FieldIDMapper)");
    
    RecordDescriptor rd2=new RecordDescriptor(new AttributeDescriptor[]{
    new AttributeDescriptor(StorageType.Float,"Carbon"),
    new AttributeDescriptor(StorageType.Float,"Oxygen"),
    new AttributeDescriptor(StorageType.Float,"Nitrogen"),
    new AttributeDescriptor(StorageType.Float,"Helium") }
    ); 
    
    
    DataSource ds2=new BinaryFileDataSource("boom",rd2, ISBounds,axisOrdering,"testdata/simple.6.4.bin",true);
    Datum d1=new Datum(3);
    
    ds2.activate();
    
    FieldIDMapper fidm=new FieldIDMapper( new int[]{2,0,3}, new int[]{0,2,1} );
    
    ds2.datum(d1, new IntegerIndexSpaceID(0,1), fidm);
    
    System.out.println(ds2.dumpData());
    
    System.out.println("ds2.datum(<0,1>) should be {201.0, 301.0, 1.0}-->"+d1);    
    
    ds2.datum(d1, new IntegerIndexSpaceID(1,1), fidm);
    System.out.println("ds2.datum(<1,1>) should be {209.0, 309.0, 9.0}-->"+d1);    
    
    
    
    
    // TwoD subblock() Tests

    System.out.println("\n\n***Two Dimensional subblock() Test***\n");
    
    ISBounds bounds=new ISBounds(new IntegerIndexSpaceID(2,2),new IntegerIndexSpaceID(6,4));
    DataBlock dblock=ds1.subblock(bounds);
    System.out.println("\n Bounds are "+bounds);
    
    
    
    System.out.println(dblock);
    
    ISBounds blockBounds=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(5,5));
    ISBounds srcBounds=new ISBounds(new IntegerIndexSpaceID(2,2),new IntegerIndexSpaceID(3,3));
    ISBounds dstBounds=new ISBounds(new IntegerIndexSpaceID(1,1),new IntegerIndexSpaceID(2,2));

    //DataBlock sblock=new DatumBlock(blockBounds,recordDesc);
    DataBlock sblock=new BasicBlock(blockBounds,recordDesc);
    
    
    System.out.println("Test of subblock(DataBlock,ISBounds,ISBounds)");
    
    
    
    ds1.subblock(sblock,dstBounds,srcBounds);
    
    System.out.println();
    System.out.println("blockBounds are "+blockBounds);
    System.out.println("srcBounds are "+srcBounds);
    System.out.println("dstBounds are "+dstBounds);
    
    System.out.println(sblock);
    
    System.out.println("Test of subblock(DataBlock,ISBounds,ISBounds,FieldIDMapper)");
    //sblock=new DatumBlock(blockBounds,rd2.createBlock(fidm));
    sblock=new BasicBlock(blockBounds,rd2.createBlock(fidm));
    ds2.subblock(sblock,dstBounds,srcBounds,fidm);
    
    System.out.println(sblock);
    
    
    

    // Three D Test
    System.out.println("\n\n***Three Dimensional Test***\n");
    
    ISBounds=new ISBounds(new IntegerIndexSpaceID(0,0,0),new IntegerIndexSpaceID(3,3,3));
    int ThreeD_Ordering[]={0,1,2};
    axisOrdering=new AxisOrdering(ThreeD_Ordering);
    
    recordDesc=new RecordDescriptor(new AttributeDescriptor[]{new AttributeDescriptor(StorageType.Float,"Carbon")});
    
    ds1=new BinaryFileDataSource("bang",recordDesc, ISBounds,axisOrdering,"testdata/simple.6.1.bin",true);
    
    ds1.activate();
    

    System.out.println(ds1.dumpData());

    

    // ThreeD subblock() Tests

    System.out.println("\n\n***Three Dimensional subblock() Test***\n");
    
    bounds=new ISBounds(new IntegerIndexSpaceID(0,2,1),new IntegerIndexSpaceID(3,3,2));
    dblock=ds1.subblock(bounds);
    System.out.println("\n Bounds are "+bounds);
    

    System.out.println(dblock);
    }
    */

}
