package edu.unh.sdb.datasource;

import edu.unh.sdb.common.Datum;
import edu.unh.sdb.common.SDBIterator;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Reads data from a {@link DataSource} along a line segment.
 * @author dlipsa
 */
public class RayIterator implements SDBIterator
{   
   /** Reads data from dataSource along the line seqment between p1 and p2.
    */
   public RayIterator (DataSource dataSource, Point3d begin, Point3d end,
      IndexSpaceID blockSize, double step)
   {
      this.dataSource = dataSource;
      if (! dataSource.isActivated ())
         dataSource.activate ();
      transferFunction = TransferFunction.createDefault (dataSource);      
      
      this.begin = begin;
      this.end = end;      
      this.blockSize = blockSize;
      if (
         blockSize.getCoord(0) == 1 && 
         blockSize.getCoord(1) == 1 && 
         blockSize.getCoord(2) == 1
         )
         voxelBlock = true;
      normal = calculateNormal (begin.x, begin.y, begin.z, end.x, end.y, end.z);
      this.step = step;
      this.stepVector = new Vector3d (end.x - begin.x, end.y - begin.y, end.z - begin.z);
      this.stepVector.normalize ();
      this.stepVector.scale (step);
      int blocksCount = (int)Math.floor (begin.distance (end) / step) + 1;
      if (voxelBlock)
         this.blocks = new Datum[blocksCount];
      else         
         this.blocks = new TransferFunction.Block[blocksCount];
   }
   
   public void init ()
   {
      current = new Point3d (begin);      
      read ();
   }
   
   public boolean next ()
   {
      current.add (stepVector);
      boolean valid = valid ();
      if (valid)
         read ();
      return valid;
   }
   
   public boolean valid ()
   {
      return index < (blocks.length - 1);
   }
   
   public void end ()
   {}
   
   public final TransferFunction getTransferFunction ()
   {
      return transferFunction;
   }
      
   /**
    * Returns the begin, current and end point or block coordinates. The 
    * 'original' versions for the begin and end coordinates are necessar 
    * because the sometimes the algorithm switches the original begin and 
    * end points.
    */
   public Point3d getBeginPoint ()
   {
      return begin;
   }
   
   public Point3d getOriginalBeginPoint ()
   {
      return begin;
   }
   
   public Point3d getCurrentPoint ()
   {
      return current;
   }
   
   public Point3d getEndPoint ()
   {
      return end;
   }
   
   public Point3d getOriginalEndPoint ()
   {
      return end;
   }

   public Vector3d getLineNormal ()
   {
      return normal;
   }
   
   public int getCurrentElementIndex ()
   {
      return index;
   }
      
   public Object getCurrentElement ()
   {
      return blocks[index];
   }
   
   public int getNumberOfElements ()
   {
      return blocks.length;
   } 
   
   public double getStep ()
   {
      return step;
   }
   
   /**
    * Calculate the normal in the plane XZ on the line that conects the two points
    * line eq: _x = _p0 + t * _d
    * plane perpendic. on the line that goes through origin: _x * _d = 0
    * the normal is in plane XZ: y = 0
    */
   private static Vector3d calculateNormal (
      double beginX, double beginY, double beginZ,
      double endX, double endY, double endZ)
   {
      if (beginZ != endZ)
         return new Vector3d (1, 0, -(endX - beginX)/(endZ - beginZ));
      else
         return new Vector3d (0, 0, 1);
   }
   
   private void read ()
   {
      index++;
      if (voxelBlock)
         readDatum();
      else
         readBlock ();
   }
   
   private void readDatum ()
   {
      IntegerIndexSpaceID id = new IntegerIndexSpaceID (
         (int)Math.round((current.x)), (int)Math.round(current.y), 
         (int)Math.round(current.z));
      blocks[index] = dataSource.datum(id);
   }
   
   private void readBlock ()
   {
      TransferFunction.Block block = new TransferFunction.Block ();
      blocks[index] = block;
      ISBounds bounds = new ISBounds (blockSize);
      block.setData (dataSource.createDataBlock (bounds));

      int[] lower = new int[] {
         getStart (current.x, blockSize.getCoord (0)), 
         getStart(current.y, blockSize.getCoord (1)),
         getStart(current.z, blockSize.getCoord (2))
      };
      int[] upper = new int[] {
         lower[0] + blockSize.getCoord (0) - 1, 
         lower[1] + blockSize.getCoord (1) - 1, 
         lower[2] + blockSize.getCoord (2) - 1
      };
      bounds.assign (lower, upper);
      dataSource.subblock (block.getData (), bounds);
      block.setRatios(
         current.x - Math.floor (current.x),
         current.y - Math.floor (current.y),
         current.z - Math.floor (current.z));      
   }
   
   
   private int getStart (double val, int size)
   {
      if ((size & 1) == 1)
         // odd
         return ((int)Math.round (val)) - (size >> 1);
      else
         // even
         return ((int)Math.floor (val)) - (size >> 1) + 1;
   }
   
   private Point3d begin, end, current;
   private int index = -1;
   private Vector3d normal, stepVector;
   private double step;
   private IndexSpaceID blockSize;
   private boolean voxelBlock;
   private Object blocks[];
   private DataSource dataSource;   
   private TransferFunction transferFunction;      
}
