/* 
    BlockMapper.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
*/

/** RegularGridBlockMapper is mainly used by BlockedDatasource. It focuses on regular grid
    blocks and provides fast searching through the blocks. 
    It maps the IndexSpaceID to the block and corresponding IndexSpaceID inside the block.
    It also maps an ISBounds to an array of subblocks of the components.
    It's an immutable class.
*/
//--------------------------- Edit history --------------------------
//
// 10/21/05 rdb: deleted unused variables
// 11/23/04 rdb: replaced deprecated DatumBlock usage
//
// -------------------------------------------------------------------
package edu.unh.sdb.datasource;
import edu.unh.sdb.common.*;

import java.io.*;

public class RegularGridBlockManager extends OverlapBlockManager {
    
    private int[] steps;
    private long[] axisWeights;
    private int[] sortArray;  // indirect sort
    
    public RegularGridBlockManager (ComponentInfo [] cinfo) {
        super(cinfo);
        AxisOrdering ordering;
        int [] axes, origin;
        int i,j;
        
        ordering = new AxisOrdering( this.dim());
        this.steps = new int[this.dim()];
        axes = new int[this.dim()];
        origin = new int[this.dim()];
        for (i=0;i< steps.length; i++){
        
            this.steps[i] = this.components[0].getISIDMap().getCompositeExtent().getDimensions().getCoord(i);
            axes[i] = (this.getBounds().getDimensions().getCoord(i)) / this.steps[i] - 1;
            origin[i] = 0;
        }
        this.axisWeights = ordering.makeAxisWeights(new ISBounds(new IntegerIndexSpaceID(origin),new IntegerIndexSpaceID(axes)));
        
        // order the components array indirectly
        int min,temp;
        this.sortArray = new int[this.components.length];
        for (i=0;i<this.components.length;i++) sortArray[i] = i;
        for (i=0;i<sortArray.length;i++){
            min = sortArray[i];
            for(j=i;j<sortArray.length;j++){
            
                if (components[sortArray[j]].getISIDMap().getCompositeExtent().copyLower().lessThanOrEqual(
                    components[sortArray[min]].getISIDMap().getCompositeExtent().copyLower())) 
                    min = j;
            
            }
            
            temp = sortArray[i];
            sortArray[i] = sortArray[min];
            sortArray[min] = temp;
        
        }
        
    }
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
    defined by dstBounds. The FieldIDMapper argument determines the mapping between 
    datums in the datasource and datums in the DataBlock.*/
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds, FieldIDMapper mp) {
        int i;    
        ISBounds isb1,isb2;
        int [] blocksInRange;
        int numBlocks; // number of involved blocks on each dimension
        IntegerIndexSpaceID lower = srcBounds.copyLower();
        IntegerIndexSpaceID upper = srcBounds.copyUpper();
        int [] newlower = new int[this.dim];
        int [] newupper = new int[this.dim];
        int patch1,patch2;
        int totalBlocks=1;
        
        for (i=0;i<this.dim;i++){
            patch1 = ( lower.getCoord(i) - this.bounds.copyLower().getCoord(i) ) % steps[i];
            patch2 = ( this.bounds.copyUpper().getCoord(i) - upper.getCoord(i) ) % steps[i];
            newlower[i] = lower.getCoord(i) - patch1;
            newupper[i] = upper.getCoord(i) + patch2;
            numBlocks = (int)(((upper.getCoord(i) - lower.getCoord(i) + patch1+patch2 + 1) / steps[i]) +0.5 );
            totalBlocks *= numBlocks;
            
        }
        blocksInRange = new int[totalBlocks];
        ISIterator iter = new ISIterator(new ISBounds(new IntegerIndexSpaceID(newlower),
                                          new IntegerIndexSpaceID(newupper)),steps);
        i=0;
        while (iter.valid()){
            blocksInRange[i] = searchBlock(iter);
            iter.next();
            i++;
        
        }
        
        ISIDMapper SrcToDst = new ISIDMapper(srcBounds,dstBounds);
        
        
        int length = totalBlocks;
        for (i=0;i<length;i++) {
            isb2 = components[blocksInRange[i]].getISIDMap().getCompositeExtent();
            isb1 = srcBounds.intersection(isb2);
            if (isb1 != null) {
                
                if (fmpFlags[i]) {  // if the componentInfo has Field subset info
                    if (mp != null) {
                    
                        FieldIDMapper directMapper = 
                              FieldIDMapper.createCrossMap( mp, this.components[blocksInRange[i]].getFieldIDMap());
                        this.components[blocksInRange[i]].getComponent().subblock(theBlock,
                        SrcToDst.toComposite(isb1), components[blocksInRange[i]].getISIDMap().toComponent(isb1),
                        directMapper );
                    
                    }
                    else {
                    
                        this.components[blocksInRange[i]].getComponent().subblock(theBlock,
                        SrcToDst.toComposite(isb1), components[blocksInRange[i]].getISIDMap().toComponent(isb1),
                        this.components[blocksInRange[i]].getFieldIDMap());
                    
                    }
                }
                else{
                    if (mp != null) {
                    
                        this.components[blocksInRange[i]].getComponent().subblock( theBlock,
                        SrcToDst.toComposite(isb1), components[blocksInRange[i]].getISIDMap().toComponent(isb1),
                        mp );
                    
                    }
                    else {
                    
                        this.components[blocksInRange[i]].getComponent().subblock( theBlock,
                        SrcToDst.toComposite(isb1), components[blocksInRange[i]].getISIDMap().toComponent(isb1));
                    
                    }
                }
            
            }
            else {
                System.out.println(components[blocksInRange[i]].getISIDMap().getCompositeExtent());
                System.out.println(srcBounds);
            } 
            
        }
        
        return ;
        
    }
    
    
    // helper function to do immediate access to the block
    public int searchBlock(IndexSpaceID isid){
        
        int i, offset = 0;
        int d=this.dim();
        int [] axes = new int[d];
        if (d!=isid.dim()){
            new DimensionalityError(SDBError.Recoverable,"ID Argument has wrong dimensionality.");
        }
        
        for(i=0;i<d;i++){
            axes[i] = isid.getCoord(i) / this.steps[i];
            offset+=axes[i]*this.axisWeights[i];
        }
        return sortArray[offset];

        //return -1; // not found
    }
    
    public static void main(String[] args) throws IOException{
    
        RegularGridBlockManager mgr;
        ComponentInfo[] comp = new ComponentInfo[20];
        int x=0,y=0;
        ISBounds bounds;
        
        AttributeDescriptor attrs[]=new AttributeDescriptor[2];
        attrs[0]=new AttributeDescriptor("float","Carbon");
        attrs[1]=new AttributeDescriptor("float","Oxygen");
        RecordDescriptor recordDesc=new RecordDescriptor(attrs); // 2 attributes
        int dsorderingarray[]={0,1};
        AxisOrdering dsordering=new AxisOrdering(dsorderingarray);
        bounds = new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(1,1));
        ISBounds db1 = new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(1,1));
        TextFileDataSource ds1=new TextFileDataSource("",recordDesc, db1,dsordering,"testdata/test-8-2.txt");
        ds1.activate();
        IntegerIndexSpaceID lower,upper;
        lower = new IntegerIndexSpaceID(x,y);
        upper = new IntegerIndexSpaceID(x+1,y+1);
        ISBounds isb = new ISBounds(lower,upper);
        
        int k = 0;
        for ( int i=0; i<4; i++){
            for ( int j=0; j<5; j++){
                
                lower = new IntegerIndexSpaceID(x,y);
                upper = new IntegerIndexSpaceID(x+1,y+1);                
                isb = new ISBounds(lower,upper);
                comp[k] = new ComponentInfo(ds1,null,new ISIDMapper(bounds,isb));
                k ++;
                y += 2;

            }
            y = 0;
            x += 2;
        }
        
        mgr = new RegularGridBlockManager(comp);
        BlockedDataSource bds = new BlockedDataSource("",comp);
        //DatumBlock blk =(DatumBlock) bds.subblock(new ISBounds(new IntegerIndexSpaceID(3,5),new IntegerIndexSpaceID(6,9)));
        BasicBlock blk =(BasicBlock) bds.subblock(new ISBounds(new IntegerIndexSpaceID(3,5),new IntegerIndexSpaceID(6,9)));
        if ( blk == null )
           System.out.println( "Got null block for subblock creation." );
        IntegerIndexSpaceID id1 = new IntegerIndexSpaceID(1,0);
        System.out.println(id1);
        System.out.print("Mapped to Block: " );
        System.out.println(comp[mgr.searchBlock(id1)].getISIDMap().getCompositeExtent());
        System.out.println(" New ISID: "+comp[mgr.searchBlock(id1)].getISIDMap().toComponent(id1));
        
        id1 = new IntegerIndexSpaceID(3,8);
        System.out.println(id1);
        System.out.print("Mapped to Block: " );
        System.out.println(comp[mgr.searchBlock(id1)].getISIDMap().getCompositeExtent());
        System.out.println(" New ISID: "+comp[mgr.searchBlock(id1)].getISIDMap().toComponent(id1));
        
        id1 = new IntegerIndexSpaceID(4,5);
        System.out.println(id1);
        System.out.print("Mapped to Block: " );
        System.out.println(comp[mgr.searchBlock(id1)].getISIDMap().getCompositeExtent());
        System.out.println(" New ISID: "+comp[mgr.searchBlock(id1)].getISIDMap().toComponent(id1));        
        
        id1 = new IntegerIndexSpaceID(7,4);
        System.out.println(id1);
        System.out.print("Mapped to Block: " );
        System.out.println(comp[mgr.searchBlock(id1)].getISIDMap().getCompositeExtent());
        System.out.println(" New ISID: "+comp[mgr.searchBlock(id1)].getISIDMap().toComponent(id1));        

        id1 = new IntegerIndexSpaceID(7,9);
        System.out.println(id1);
        System.out.print("Mapped to Block: " );
        System.out.println(comp[mgr.searchBlock(id1)].getISIDMap().getCompositeExtent());
        System.out.println(" New ISID: "+comp[mgr.searchBlock(id1)].getISIDMap().toComponent(id1));      

        id1 = new IntegerIndexSpaceID(5,9);
        System.out.println(id1);
        System.out.print("Mapped to Block: " );
        System.out.println(comp[mgr.searchBlock(id1)].getISIDMap().getCompositeExtent());
        System.out.println(" New ISID: "+comp[mgr.searchBlock(id1)].getISIDMap().toComponent(id1));        
                                
        System.out.println("End.");
    
    }
    
 
}
