//
//  RegularISPartitioning.java
//  Unstructured
//
//  Created by Philip Rhodes on Sat Oct 05 2002.
//  Copyright( c) 2002 University of New Hampshire. All rights reserved.
//
//------------------- Edit history --------------------------------------
// 03/02/12 rdb: Added some more convenience mapping functions (for SparseDS)
//               <Perhaps these and existing mappers should be moved to parent?>
//                  
// 02/12/12 rdb: -Added constructor that takes partition size as parameter,
//               rather than the partitioning bounds; reorganized all the
//               constructor code to define shared code; moved some code to
//               the parent class and added instance variables to simplify
//               constructor and other code; fixed error in creating 
//               partitionDimensions if partitioning doesn't evenly divide
//               space -- other things also probably don't work in this case,
//               at least this part is ok.
//              -Renamed instance variables that represent the "sizes" of
//               blocks from "Dimensions" variables to "Shape" variables. 
//               The term "dimension" is so over-loaded, I like to use it as
//               little as possible.
//              -Made new methods with "Shape" in the anme to return the 
//               "shape" of various components -- moved all these to parent.
// 02/01/12 rdb: -Added new method that converts a space ISBounds to a list
//                of space ISBounds
//                   ISBounds[] mapToSpaceParts( ISBounds spaceBnds ) 
//               -Removed deprecation on one of constructors; it is the one used
//                by ChunkedDataSource and only differs from an un-deprecated
//                constructor in that it uses default axisOrdering.
// 06/04/08 rdb: added comments, some minor reformatting.
// 10/21/05 rdb: removed unused variables
//-----------------------------------------------------------------------


package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.SparseDataSource.MapInfo;

import java.util.*;

/** 
 * This class implements a regular partitioning of a bounded index space. A
 *  partitioning divides the space into some number of partitions. For regular
 *  partitionings, these partitions are all the same size and shape, except for
 *  those that span an edge of the index space.( That is, it is possible that
 *  partitions near the edge may be clipped by the index space bounds.) A regular
 *  partitioning can be specified either by giving the dimensions of a partition
 *  to the constructor, or by specifying how many partitions each axis should
 *  be divided into.
 */

public class RegularISPartitioning extends ISPartitioning
{
   //----------------- instance variables -----------------------------
   protected long[]       axisWeights;
   protected int[]        tempCoords; // Scratch space.
   protected AxisOrdering ordering;
   
   /*-------- key instance variables inherited from parent
	protected int      dimensionality;
   
   protected ISBounds partitioningBounds;
   protected int      partitioningVolume; 
   
	protected int[]    partitioningLowerBounds;
	protected int[]    partitioningUpperBounds;
	protected int[]    partitioningShape;
   
   protected int[]    partitionShape; // size of each partition

   protected ISBounds spaceBounds;
	protected int[]    spaceLowerBounds;
	protected int[]	 spaceUpperBounds;
   protected int[]    spaceShape;
   ----------------------------------*/
   /**
    * Create a regular partition over space, given the desired partitioning space
    * 
    * @param partitioningBounds
    * @param spaceBounds 
    */
   public RegularISPartitioning( ISBounds partitioningBounds, ISBounds spaceBounds )
   {  
      super( partitioningBounds, spaceBounds );
      initialize( new AxisOrdering( this.dimensionality ));
   }

   /**
    * Create a regular partition over space, given the desired partitioning space
    *        and an AxisOrdering
    * 
    * @param partitioningBounds
    * @param spaceBounds
    * @param ordering 
    */
   public RegularISPartitioning( ISBounds partitioningBounds, ISBounds spaceBounds, 
		                         AxisOrdering ordering )
   {      
      super( partitioningBounds, spaceBounds );
      initialize( ordering.copy() );
   }
    /**
    * Create a regular partition over space, given the desired block size for
    * each partition.
    * 
    * @param partitionSize
    * @param spaceBounds 
    */
   public RegularISPartitioning( IndexSpaceID partitioningShape, ISBounds spaceBounds )
   {  
      super( partitioningShape, spaceBounds );
      initialize( new AxisOrdering( this.dimensionality ));
   }

  /** 
    * Create a regular partition over space, given the shape of each partition, 
    *         not the dimensions of the partitioning Space.
    *      Also specifies AxisOrdering
    * 
    * @param partitioningBounds
    * @param spaceBounds
    * @param ordering 
    */
   public RegularISPartitioning( IndexSpaceID partitioningShape, 
                                 ISBounds spaceBounds, 
		                           AxisOrdering ordering)
   {      
      super( partitioningShape, spaceBounds );
      initialize( ordering.copy() );    
   }
   
  //----------------- initialize -----------------------------
   /**
    * Initialize the fields that are not set by super class:
    *   AxisOrdering information
    *   partitionShape information
    * 
    * @param ordering 
    */
   private void initialize( AxisOrdering axisOrdering )
   {      
      tempCoords = new int[ dimensionality ];
      ordering = axisOrdering;
      axisWeights = ordering.makeAxisWeights( partitioningBounds );
      partitionShape = new int[ dimensionality ];
      
      // figure out partition block size -- the "nominal" one for all but the
      //   outer borders.
      for ( int d = 0; d < dimensionality; d++ )
      {
         partitionShape[ d ] 
                   = ( spaceShape[d] - 1 ) / partitioningShape[d] + 1;        
         //rdb old code only worked if partitioning has all equal size partitions
         //             = spaceShape[ i ] / partitioningShape[ i ];
         if ( partitionShape[d] <= 0 )
         {
            new SDBError( SDBError.Fatal, "Got bad partition shape value: " + 
                                partitionShape[ d ] + " at index " + d );
         }           
      }
   }
   
    /**
     * return the int array describing the shape of each partition
     * @return 
     */
    int[] getPartitionShape()
    {
       return this.partitionShape;
    }

    /**
     * return the int array describing the shape of each partition
     * @return 
     */
    final int[] getPartitionDimensionsArray()
    {
       return getPartitionShape();
    }
     
//   public RegularISPartitioning( int[] partitioningShape, ISBounds spaceBounds )
//   {     
//      super( partitioningShape, spaceBounds );
//      
//      int d = this.dimensionality;
//      
//      this.partitionShape = new int[d];
//      this.tempCoords = new int[d];
//      this.ordering = new AxisOrdering( d);
//      this.axisWeights = ordering.makeAxisWeights( partitioningShape );
//      
//      for(  int i = 0; i < d; i++ )
//      {        
//         this.partitionShape[i] = spaceBounds.getDimension(i) / partitioningShape[i];
//      }      
//   }


   public RegularISPartitioning project( int axis )
   {
      return new RegularISPartitioning( 
                          new IntegerIndexSpaceID( 
                                              partitionShape).project(axis), 
    		                                     getPartitioningBounds().project( axis ) );  
   }
   
   /** Return a reference to the partitioning ordering axis weights. */
   long [] getAxisWeights()
   {      
      return this.axisWeights;
   }
   
   /** Return a reference to the partitioning ordering. */
   AxisOrdering getOrdering()
   {      
      return this.ordering;
   }
   
   /** 
    * Map the id onto the partitioning.
    */
   public final IndexSpaceID map( IndexSpaceID dataPoint )
   {     
      //return this.mapLow(p); // Let's use mapLow as the default mapping.
      
      IndexSpaceID id = new IntegerIndexSpaceID( dataPoint );
      
      int [] scoords = dataPoint.refCoords();
      int [] dcoords = id.refCoords();
      
      for(  int i=0; i<this.dimensionality; i++ )
      {        
         dcoords[i] =( scoords[i] - this.spaceLowerBounds[i]) / this.partitionShape[i];
      }      
      return id;      
   }

   /** Map the first argument onto the partitioning, placing the result
    *  in the second argument. A reference to the second id is returned
    *  as a convenience. 
    */
   public final IndexSpaceID map( IndexSpaceID s, IndexSpaceID d )
   {           
      int [] scoords = s.refCoords();
      int [] dcoords = d.refCoords();
      
      for(  int i=0; i < this.dimensionality; i++ )
      {        
         dcoords[i] =( scoords[i] - this.spaceLowerBounds[i]) / this.partitionShape[i];
      }
      
      return d;      
   }

   /** 
    * Map the coords array onto the partitioning, replacing the old value. 
    * Returns the partitioning isid (bounds.lower()) of the partition containing
    *   the space location passed in in the array.
    */
   final void map( int [] coords )
   {            
      for(  int i=0; i < this.dimensionality; i++ )
      {        
         coords[i] =( coords[i] - spaceLowerBounds[i]) / partitionShape[i];
      }     
   }

   /** Map the bounds onto the partitioning, returning a new ISBounds. */
   public final ISBounds map( ISBounds bounds )
   {
      ISBounds newBounds = new ISBounds( bounds );
      
      int [] lower = newBounds.getLowerArray();
      int [] upper = newBounds.getUpperArray();
      
      this.map( lower );
      this.map( upper );
      
      return newBounds;
   }
   
   
   /** 
    * Map the index space id onto the partitioning space and return the 1D
    * offset of the containing partition. Currently, the default axis ordering 
    * is assumed.
    */
   public final long containingOffset( IndexSpaceID id )
   {     
      int [] coords = id.refCoords();
      
      for(  int i = 0; i < this.dimensionality; i++ )
      {        
         tempCoords[i] =(  coords[i] - this.spaceLowerBounds[i]) / this.partitionShape[i];        
      }
      
      return AxisOrdering.computeOffset( tempCoords, this.axisWeights );     
   }
      
   /** 
    * Return the 1D offset of the element specified by the id. The id should
    *  refer to the partitioning space, not the underlying index space.
    */
   public final long elementOffset( IndexSpaceID id )
   {     
      return AxisOrdering.computeOffset( id, this.axisWeights );     
   }     
      
   /** add by xuan */
   public ISBounds getGeometricBounds()
   {
      return new ISBounds(  this.partitionShape );
   }
   
   //-------------------add by xuan -------------------------------------------
   /** 
    * map a partition ISBounds to space ISBounds
    */
   public ISBounds mapToSpace(  ISBounds partitionBounds)
   {
      IndexSpaceID partitionLow = partitionBounds.copyLower();
      IndexSpaceID partitionUpper = partitionBounds.copyUpper();
      return new ISBounds(  this.mapToSpaceLower( partitionLow) , this.mapToSpaceUpper(  partitionUpper) );
   }
   
   /** 
    * map partition ID to space ISBounds
    */
   public ISBounds mapToSpace(  IndexSpaceID partitionID )
   {
      int[] lower = new int[ this.dimensionality ];
      int[] upper = new int[ this.dimensionality ];
      int[] id = new int[ this.dimensionality ];
      partitionID.getCoords( id );
      for(  int i = 0 ; i < this.dimensionality; i++ )
      {
         lower[i] = partitionShape[i] * id[i];
         upper[i] = lower[i] + partitionShape[i] - 1;
      }
      return new ISBounds( new IntegerIndexSpaceID( lower ), new IntegerIndexSpaceID( upper ));
   }
   //--------------- mapToSpaceParts( ISBounds ) ------------------------
   /**
    * Map a bounds in space to the set of spatial bounds that represent
    * the pieces of the bounds that are each only in exactly one partition.
    * This was motivated by needs of SparseDataSource, but could have other uses.
    * 
    * @param ISBounds: spaceBnds -- portion of space domain to be mapped
    * @return ISBounds[]:  -- space bounds corresponding to separate partitions
    */
   public ISBounds[] mapToSpaceParts( ISBounds spaceBnds )
   {
      ISBoundsIterator subBlockIter = getPartitionIterator( spaceBnds );
      ArrayList<ISBounds> bndsList = new ArrayList<ISBounds>();
      while ( subBlockIter.valid())
      {
         bndsList.add( ((ISBounds)subBlockIter).copy() );
         subBlockIter.next();
      }
      ISBounds[] bndsArray = new ISBounds[ bndsList.size() ];
      for ( int b = 0; b < bndsArray.length; b++ )
         bndsArray[ b ] = bndsList.get( b );
      return bndsArray;
   }
   
   //----------- mapToSpaceLower( IndexSpaceID ) -----------------------  
   private IndexSpaceID mapToSpaceLower( IndexSpaceID partitionID )
   {
      int[] lower = new int[ this.dimensionality ];
      int[] id = new int[ this.dimensionality ];
      partitionID.getCoords( id );
      for(  int i = 0; i < this.dimensionality; i++ )
      {
         lower[i] = partitionShape[i] * id[i];
      }
      return new IntegerIndexSpaceID( lower );
   }
   
   //----------- mapToSpaceUpper( IndexSpaceID ) -----------------------
   private IndexSpaceID mapToSpaceUpper( IndexSpaceID partitionID )
   {
      int[] lower = new int[this.dimensionality];
      int[] upper = new int[this.dimensionality];
      int[] id = new int[this.dimensionality];
      partitionID.getCoords( id);
      for(int i =0 ; i< this.dimensionality; i++)
      {
         lower[i] = partitionShape[i]* id[i];
         upper[i] = lower[i]+partitionShape[i]-1;
      }
      return new IntegerIndexSpaceID( upper);
   }

   /** If upper is true, return the upper value of the specified axis 
    *  of the partition specified by the partition id, otherwise, return the lower value.  */
   public final int getPartitionBoundary( IndexSpaceID partitionID, int axis, boolean upper)
   {      
   
      int [] coords=partitionID.refCoords();
      int lowerBoundary = spaceLowerBounds[axis] + coords[axis] * this.partitionShape[axis];
   
      if( upper){
      
         return lowerBoundary + this.partitionShape[axis];
      } else {
      
         return lowerBoundary;
      }
      
   
   }

   /** If upper is true, return the upper value of the specified axis 
    *  of the partition containing the id, otherwise, return the lower value.  */
   public final int getContainingPartitionBoundary( IndexSpaceID id, 
                                                    int axis, 
                                                    boolean upper)
   {         
      int [] coords=id.refCoords();
   
      if ( upper )
      {     
         //return( coords[axis] / this.partitionShape[axis]) * this.partitionShape[axis] + this.partitionShape[axis] - 1;
         return(( coords[axis] - spaceLowerBounds[axis]) / partitionShape[axis]) 
                             * this.partitionShape[axis] + this.partitionShape[axis] - 1;
      } 
      else 
      {     
         //return( coords[axis] / partitionShape[axis]) * partitionShape[axis];
         return(( coords[axis] - spaceLowerBounds[axis]) / partitionShape[axis]) 
                             * this.partitionShape[axis];
      }   
   }
   
   /** Return the bounds of the partition that contains the given id. */
   public final ISBounds getContainingPartitionBounds( IndexSpaceID id)
   {      
      return getContainingPartitionBounds( id, new ISBounds( id));
   }
   
   /** 
    * return the partition index (1D order position of the partitions). This
    * can be easily mapped to a seek address, for example.
    */
   public int getPartitionIndex( IndexSpaceID dataPoint )
   {
      IndexSpaceID blockLoc = map( dataPoint ); // map space pos to partitioning
      
      return (int) this.partitionOffset( blockLoc );
   }
   
   /** 
    * Sets the ISBounds argument to the bounds of the partition in the data space
    * that contains the given id, specified in the data space. 
    */
   public final ISBounds getContainingPartitionBounds( IndexSpaceID id, 
                                                       ISBounds bounds)
   {                  
      int d=this.dimensionality;
      int [] coords=id.refCoords();
      int [] lower = bounds.getLowerArray();
      int [] upper = bounds.getUpperArray();
      
      for(  int i=0; i<d; i++ )
      {        
         //lower[i] =( coords[i] / this.partitionShape[i]) * this.partitionShape[i];
         //upper[i] = lower[i] + this.partitionShape[i] - 1;
         lower[i] =((coords[i] - spaceLowerBounds[i]) / this.partitionShape[i]) 
                             * this.partitionShape[i];
         upper[i] = lower[i] + this.partitionShape[i] - 1;      
      }     
      return bounds;
   }
     
//   /** From RegularCellPartitioning.java */
//   protected final IndexSpaceID mapLow( IndexSpaceID p)
//   {
//      
//      int []  cellIndices=new int[this.dimensionality];
//      float cellIndexFloat;
//      //IndexSpaceID tDimensions=this.indexBounds.getDimensions();
//      int upperBound,lowerBound;
//      
//      for(  int i=0; i<this.dimensionality; i++)
//      {
//         
//         cellIndexFloat=(p.getCoord( i)-this.spaceLowerBounds[i])/
//               this.partitionShape[i] + this.partitioningLowerBounds[i];
//         
//         //cellIndices[i] =( cellIndexFloat < 1.0) ? 0 :( int)Math.ceil(cellIndexFloat-1.0);
//         
//         //upperBound=tDimensions.getCoord(i);
//         upperBound=this.partitioningUpperBounds[i];
//         //cellIndices[i] =(  cellIndices[i] >= upperBound ) ? upperBound  : cellIndices[i];
//         cellIndices[i] =(int)( (  cellIndexFloat >= upperBound ) ? upperBound  : cellIndexFloat);
//         
//         
//         lowerBound=this.partitioningLowerBounds[i];
//         cellIndices[i] =(  cellIndices[i] <= lowerBound ) ? lowerBound  : cellIndices[i];
//         
//      }
//      
//      return new IntegerIndexSpaceID( cellIndices);
//   }
//   
//   
//   /** Map the point onto the grid. If p falls on a partition boundary, we choose the
//    *  partition with the Higher index. */
//   protected final IndexSpaceID mapHigh( IndexSpaceID p)
//   {
//      
//      int i;
//      int []  cellIndices=new int[this.dimensionality];
//      int upperBound,lowerBound;
//      
//      for(i=0;i<this.dimensionality;i++)
//      {
//         
//         cellIndices[i]=(int)( (p.getCoord( i)-this.spaceLowerBounds[i])/
//               this.partitionShape[i] +
//               this.partitioningLowerBounds[i]
//               );
//         
//         
//         upperBound=this.partitioningUpperBounds[i];
//         cellIndices[i] =(  cellIndices[i] >= upperBound ) ? upperBound  : cellIndices[i];
//         
//         
//         lowerBound=this.partitioningLowerBounds[i];
//         cellIndices[i] =(  cellIndices[i] <= lowerBound ) ? lowerBound  : cellIndices[i];
//         
//      }
//      
//      return new IntegerIndexSpaceID( cellIndices);
//   }
//   
//  

   public  ISBoundsIterator getPartitionIterator()
   {     
      return new ISBoundsIterator( new ISBounds( spaceShape ), 
                                   new IntegerIndexSpaceID( this.spaceLowerBounds ), 
                                   this.partitionShape );
   }
   
   
   public  ISBoundsIterator getPartitionIterator( ISBounds subSpace )
   {     
      return new ISBoundsIterator( subSpace, 
                                   new IntegerIndexSpaceID( this.spaceLowerBounds ), 
                                   this.partitionShape );
   }
   
   public  ISBoundsIterator getPartitionIterator( ISBounds subSpace, 
                                                  ISBoundsIterator iter )
   {      
      iter.set( subSpace, this.spaceLowerBounds, this.partitionShape );      
      return iter;
   }
   
   
   /** 
    * Sets the argument ISBounds to the space bounds corresponding to the 
    * partition defined by the partitioning ID and returns the bounds as well. 
    */
   public ISBounds getIndexedBounds( ISBounds     bounds, 
                                     IndexSpaceID partitioningIndex )
   {
      int d = this.dimensionality;
      int [] coords = partitioningIndex.refCoords();
      int [] lower = bounds.getLowerArray();
      int [] upper = bounds.getUpperArray();
     
      for(  int i=0; i<d; i++ )
      {
         lower[i] = this.spaceLowerBounds[i] + coords[i] * partitionShape[i];
         upper[i] = lower[i] + partitionShape[i] - 1;
      }
      
      return bounds;
   }
   
   
   /** 
    * Sets the argument ISBounds to the bounds corresponding to the partition 
    * offset. The ISBounds parameter is returned as a convenience.
    */
   public ISBounds getOffsetBounds( ISBounds bounds, long offset )
   {      
      int d = this.dimensionality;
      
      this.ordering.computeCoords( offset, tempCoords, axisWeights);
      
      int [] lower = bounds.getLowerArray();
      int [] upper = bounds.getUpperArray();
      
      
      for(int i=0; i<d; i++)
      {        
         lower[i] = spaceLowerBounds[i] + tempCoords[i] * partitionShape[i];
         upper[i] = lower[i] + partitionShape[i] - 1;        
      }
      
      return bounds;
   }
      
   /** 
    * Returns a new ISID corresponding to the 1D partition offset. The ISID
    *  refers to the partitioning space, not the index space.
    */
   public IndexSpaceID getOffsetID( long offset)
   {     
      return this.getOffsetID( 
                 new IntegerIndexSpaceID( new int[this.dimensionality]), offset);
   }
   
   
   /** 
    * Sets the argument ISID to the value corresponding to the 1D partition offset.
    *  The ISID refers to the partitioning space, not the index space. The 
    * parameter is returned as a convenience.
    */
   public IndexSpaceID getOffsetID( IndexSpaceID id, long offset)
   {      
      this.ordering.computeCoords( offset, id.refCoords(), this.axisWeights);
      
      return id;
   }
   
   
   /** 
    * Map the partitioning space id to its 1D offset.  It uses the partitioning's
    * AxisOrdering to do the mapping. The parameter identifies a position in the
    * partitioning space, not the base space.
    */
   public final long partitionOffset( IndexSpaceID partitionId)
   {     
      return AxisOrdering.computeOffset( partitionId.refCoords(), this.axisWeights);     
   }
   
   /** 
    * Return a String representing the object. 
    */
   public String toString()
   {     
      return this.toString( "");
   }
   
   /** 
    * Return a String representing the object indented by the argument string. 
    */
   public String toString( String indent)
   {     
      String indent2 = indent + Globals.toStringIndent;
      int d = this.dimensionality;
      
      StringBuffer result=new StringBuffer( indent + "{ RegularISPartitioning:\n" );
      result.append( super.toString( indent2) + "\n" );
      
      result.append( indent2 + "geometric partition dimensions=<" );
      for( int i = 0; i < d-1; i++ )
      {
         result.append( this.partitionShape[ i ] + ", " );
      }
      result.append( this.partitionShape[ d-1 ] + ">\n"  );
      result.append( indent + "}" );
      return new String( result );
   }
    
   public static void main( String[] argv)
   {   
      //testConstruction();
      testMapping();
   }
   private static void testMapping()
   {
      ISBounds pBounds = new ISBounds( new int[] {2, 2, 2} );
      ISBounds sBounds = new ISBounds( new int[] {8, 8, 8} );
      RegularISPartitioning rp = new RegularISPartitioning( pBounds, sBounds );
     
      int b = 0;
      IndexSpaceID[] partitionPos = { 
                                  new IntegerIndexSpaceID( b, b, b ),
                                  new IntegerIndexSpaceID( b, b, b+1 ),
                                  new IntegerIndexSpaceID( b, b+1, b ),
                                  new IntegerIndexSpaceID( b+1, b, b ),
                                  new IntegerIndexSpaceID( b+1, b+1, b+1 ) };
      System.out.println( "Testing that partitionOffset returns 'block' #" );
      int[] p = rp.partitionShape;
      System.out.printf( "Partition shape: %d %d %d\n", p[0], p[1], p[2] );
      for ( int i = 0; i < partitionPos.length; i++ )
      {
         System.out.println( i + ": " + rp.partitionOffset( partitionPos[i] ) );
      }
   }
   private static void testConstruction()
   {
      int spaceSizeMean = 20;
      int partitioningSizeMean = 5;
      int maxSizeOffset = 3;
      int maxSizeDifference = 2;
      int[] sSize = new int[ 3 ];
      int[] pSize = new int[ 3 ];
      ISBounds  pBounds;
      ISBounds  spaceBounds;
      
      for ( int s = spaceSizeMean-maxSizeOffset; 
                          s < spaceSizeMean + maxSizeOffset; s++ )
      {
         sSize[0] = s;
         for ( int d = -maxSizeDifference; d < maxSizeDifference; d++ )
         {
            sSize[1] = s - d;
            sSize[2] = s + d;
         }
         spaceBounds= new ISBounds( sSize );
         pSize[0] = 3;
         for ( int pSoff = -1; pSoff < 1; pSoff++ )
         {
            pSize[1] = pSize[0] + pSoff;
            pSize[2] = pSize[0] - pSoff;
            pBounds=new ISBounds( pSize );
       
            // test constructor with partitioning bounds
            RegularISPartitioning p = new RegularISPartitioning( pBounds,
                             spaceBounds );

            System.err.printf( "-------------> %d %d %d X %d %d %d\n", 
                                sSize[0], sSize[1], sSize[2], 
                                pSize[0], pSize[1], pSize[2] );
            System.err.println( "Partitioning bounds: " + p );
            long vol = p.volume();
            System.err.println( "volume: " + vol ); 
            
            // now build equivalent using the shape of the partitions

            int[] shape = p.getPartitionShape();
            IndexSpaceID shapeISID = new IntegerIndexSpaceID( shape );

            RegularISPartitioning ps = new RegularISPartitioning( shapeISID,
                             spaceBounds );
            System.err.println( "Partition shape: " + ps );
            vol = ps.volume();
            System.err.println( "volume: " + vol ); 
           
            if ( !p.partitioningBounds.equals( ps.partitioningBounds))
               System.err.println( "--- inconsistent partitioning shape" );
         }
      }
       
      /**** test pre-Feb 2012
      // test constructor with partition shape
      IndexSpaceID partitionShape = new IntegerIndexSpaceID( 5, 5 );
      RegularISPartitioning ps=new RegularISPartitioning( 
                                      partitionShape, spaceBounds);
      System.out.println( "Partitioning 2: " + ps );
      vol = ps.volume();
      }
      ISBounds  pBounds=new ISBounds( new int[]{2,2});
      ISBounds  spaceBounds= new ISBounds( new int[]{11,11});
      
      // test constructor with partitioning bounds
      RegularISPartitioning p=new RegularISPartitioning( pBounds, spaceBounds);
      
      System.out.println( "Partitioning 1: " + p );
      long vol = p.volume();

      int[] shape = p.getPartitionShape();
      System.out.println( "Partition shape: " + ( new IntegerIndexSpaceID( shape )));
      
      ISBounds bounds2 = p.getContainingPartitionBounds( 
                          new IntegerIndexSpaceID( new int[] {0,0}));
      
      System.out.println( bounds2);
      
      p.getContainingPartitionBounds( new IntegerIndexSpaceID( 
                          new int[] {5,5}), bounds2);
      
      System.out.println( bounds2);
         
      // test constructor with partition shape
      IndexSpaceID partitionShape = new IntegerIndexSpaceID( 5, 5 );
      RegularISPartitioning ps=new RegularISPartitioning( 
                                      partitionShape, spaceBounds);
      System.out.println( "Partitioning 2: " + ps );
      vol = ps.volume();
      /***************************/
      /*
        ISBoundsIterator iter=p.getPartitionIterator(spaceBounds);
         
         
        for(iter.init(); iter.valid(); iter.next()){
         
            System.out.println(iter);
        }
         
         
        System.out.println("\n*********\n");
        pBounds = new ISBounds(new int[]{2,2});
        spaceBounds = new ISBounds(new int[]{10,10});
         
        p=new RegularISPartitioning(pBounds, spaceBounds);
         
        ISBounds subSpace=new ISBounds(new int[] {3,3}, new int[]{8,8});
         
        p.getPartitionIterator(subSpace, iter);
         
         
        for(iter.init(); iter.valid(); iter.next()){
         
            System.out.println(iter);
        }
         */
   }      
}



