/*
 * RegularPartitionISID - represents an index space id as a partition/offset form; 
 *                     it retains a copy of the original "space" representation
 * 
 * History
 * 
 * 03/02/12 rdb: re-create with simplicity as goal -- don't extend 
 *               IntegerIndexSpaceID
 */
package edu.unh.sdb.datasource;

//import edu.unh.sdb.common.*;


/**
 * Barely started; don't trust anything; 
 * 
 * Perhaps shouldn't try to extend IntegerIndexSpaceID
 *
 * @author rdb
 */
public class RegularPartitionISID //extends IntegerIndexSpaceID
{
   //------------------- instance variables ----------------------------
   protected RegularISPartitioning partitioning;
   protected long         blockNum;
   protected int[]        blockShape;
   protected int[]        blockLoc;   // loc of block in partitioning space
   protected IndexSpaceID blockISID;  // ISID of block in partitioning space

   protected int[]        offset;     // the position in block of location
   protected IndexSpaceID offsetISID; // block location as IndexSpaceID
   protected IndexSpaceID spaceISID;  // original space location
   
   //------------- inherited from IntegerIndexSpaceID
   // protected int[] axes;        // id in the non-block space of this loc

   //---------------- constructors --------------------------------------
   public RegularPartitionISID( RegularISPartitioning part, int[] spacePos )
   {
      //super( pos );
      this( part, new IntegerIndexSpaceID( spacePos ));
   }

   /** 
    * Constructs a RegularPartitionISID for the spacePos in the partitioning, part
    */
   public RegularPartitionISID( RegularISPartitioning part, IndexSpaceID spacePos )
   {
      partitioning = part;
      blockShape   = part.getPartitionShape();
      
      spaceISID    = spacePos.copyID();
      
      computePartitionFields( spacePos.getCoords() );
      
      offsetISID   = new IntegerIndexSpaceID( offset );
      blockNum     = part.partitionOffset( blockISID );
   }
 
   //------------- computePartitionFields ------------------------------------
   /*
    * compute internal variables
    * @param pos
    */
   private void computePartitionFields( int[] pos )
   {
      offset = new int[ pos.length ];
      blockLoc = new int[ pos.length ];
      
      for ( int d = 0; d < pos.length; d++ )
      {
         offset[ d ]   = pos[ d ] % blockShape[ d ];
         blockLoc[ d ] = pos[ d ] / blockShape[ d ];
      }
      blockISID = new IntegerIndexSpaceID( blockLoc );
   }
   /**
    * Map this RegularPartitionISID back to an ISID in the original space
    * 
    * @return IndexSpaceID that is location of this object in original space
    */
   public IndexSpaceID mapToSpace()
   {
      ISBounds partitionBnds = partitioning.mapToSpace( blockISID );
      IndexSpaceID spaceLoc = partitionBnds.copyLower();
      spaceLoc.add( offsetISID );
      return spaceLoc;
   }
   
   /**
    * return integer (long) index of the partition, which is the 1D location of
    * this partition in the partitioning space (not the data space).
    * 
    * @return 
    */
   public long getPartitionIndex()
   {
      return blockNum;
   }
   
   /**
    * Change the parameter to describe the bounds in space coordinates 
    * of the partition containing this position. Return the bounds parameter
    * as a convenience.
    * 
    * @param partBounds
    * @return 
    */
   public ISBounds getPartitionBounds( ISBounds partBounds )
   {
      partBounds.assign(  partitioning.mapToSpace( blockISID ) );
      return partBounds;
   }

   /**
    * Return the bounds in space coordinates of the partition containing 
    * this position. 
    * 
    * @return 
    */
   public ISBounds getPartitionBounds()
   {
      ISBounds partBounds = partitioning.mapToSpace( blockISID ).copy();
      return partBounds;
   }
   
   /**
    * return the ISID of the location relative to the partition it is in.
    */
   public IndexSpaceID getOffsetISID()
   {
      return offsetISID;
   }

   /** Return a copy of this RegularPartitionISID. */
   @Override
   public Object clone()
   {
      return new RegularPartitionISID( partitioning, spaceISID );
   }


   /** Return a String representing the object. */
   @Override
   public String toString()
   {
      return "<" + blockNum + ":" + offsetISID + ">";
   }


   //---------------- main unit test -------------------------
   public static void main( String[] args )
   {
      System.out.println( "*****RegularPartitionISID Tests*****" );
      // Create a RegularISPartitioning
      ISBounds[] spaceBnds =
      {  new ISBounds( new IntegerIndexSpaceID( 10, 10, 10 )),
         new ISBounds( new IntegerIndexSpaceID(  8, 16, 16 )),
         new ISBounds( new IntegerIndexSpaceID(  8, 16,  6 ))
      };
      IntegerIndexSpaceID[] partitionSizes =
      {  new IntegerIndexSpaceID( 2, 2, 2 ),
         new IntegerIndexSpaceID( 2, 4, 4 ),
         new IntegerIndexSpaceID( 4, 2, 2 )
      };
      RegularISPartitioning risp;
      
      for ( int s = 0; s < spaceBnds.length; s++ )
      {
         for ( int p = 0; p < partitionSizes.length; p++ )
         {
            risp = new RegularISPartitioning( partitionSizes[p], spaceBnds[s] );
            testPartitioning( risp );
         }
      }
   }
   
   // for a specific partitioning,
   //    1. map a set of spatial ISIDs to a RegularPartition ISID
   //    2. map the RegularPartitionISID back to spatial ISID
   //    3. make sure they match
   //
   //    4. 
   private static void testPartitioning( RegularISPartitioning regisp )
   {
      IntegerIndexSpaceID[] spacePts =
      {  new IntegerIndexSpaceID( 0, 0, 0 ),
         new IntegerIndexSpaceID( 2, 2, 2 ),
         new IntegerIndexSpaceID( 4, 4, 4 ),
         new IntegerIndexSpaceID( 2, 0, 2 ),
         new IntegerIndexSpaceID( 4, 4, 0 ),
         new IntegerIndexSpaceID( 0, 2, 2 )
      };
      System.out.println( "+++++++++++++++++++++++++++++++++++++++++++++++" );
      System.out.println( regisp );
      for ( int p = 0; p < spacePts.length; p++ )
      {
         IndexSpaceID pt = spacePts[p];
         System.out.print( "--------------" + pt + " --> " );
         
         RegularPartitionISID regISID = new RegularPartitionISID( regisp, pt );
         IndexSpaceID mappedISID = regISID.mapToSpace();
         System.out.println( mappedISID );
         if ( !mappedISID.equals( pt ) )
            System.out.println( "*** Failed. inverse map: " + mappedISID );         
      }
   }
}
