package edu.unh.sdb.datasource;

import edu.unh.sdb.common.Datum;
import java.io.FileOutputStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;
import javax.vecmath.Tuple2d;
import javax.vecmath.Tuple3d;
import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;


/**
 * A class that knows how to read data for the intersection between a
 * {@link Volume} over a {@link DataSource} and a {@link Plane}. The data is
 * read from a {@link DataSource}.
 * @author dlipsa
 */
//------------------- Edit history --------------------------------------
// 06/06/08 rdb: Made ArrayList and Comparator methods generic
// 11/09/05 rdb: Deleted display method. Shouldn't have openGL in core Granite
// 10/25/05 dlipsa: removed copy, fillBorder, index, scanConversionAxisOrder
// 10/22/05 rdb: Tried check in which failed and check out.
//               copy and fillBorder are still unused
//               index and scanConversionAxisOrder are new unused methods.
//                   I made them all "protected" to get rid of warnings.
// 10/21/05 rdb: Made correct static references to GL constants (GL. instead of gl.)
//               Commented out "fillBorder" and "copy" methods. They're not used.
//-----------------------------------------------------------------------


public class Slice
{   
   /**
    * Creates a new instance of Slice and sets its parameters
    */
   public Slice (byte[] outsideColor, byte[] insideColor,
      Volume volume, Plane plane, DataSource dataSource)
   {
      setOutsideColor (outsideColor);
      setInsideColor (insideColor);
      setVolume (volume);
      setPlane (plane);
      setDataSource (dataSource);      
   }
   
   /**
    * @return How many bytes do we use per texel, when we store the texture.
    * Currently this is 4 (rgba)
    */
   public int getBytesPerTexel ()
   {
      return bytesPerTexel;
   }
   
   /**
    * Specifies how do we convert the Datum read from the file into a rgba tupple.
    */
   public void setTransferFunction (TransferFunction transferFunction)
   {
      this.transferFunction = transferFunction;
   }
   
   /**
    * Specifies the volume we iterate over
    */
   public void setVolume (Volume volume)
   {
      this.volume = volume;
   }
   
   /**
    * @return volume we iterate over
    */
   public Volume getVolume ()
   {
      return volume;
   }
   
   /**
    * Specifies the plane we want to intersect with the volume
    */
   public void setPlane (Plane plane)
   {
      this.plane = plane;
   }
   
   /**
    * @return plane we want to intersect with the volume
    */
   public Plane getPlane ()
   {
      return plane;
   }
   
   /**
    * Sets the data  source where the volume is stored.
    */
   public void setDataSource (DataSource dataSource)
   {
      this.dataSource = dataSource;
   }
   
   /**
    * @return The data  source where the volume is stored.
    */
   public DataSource getDataSource ()
   {
      return dataSource;
   }
   
   
   /**
    * Sets the color for the texture ouside the bounds of the slice projection
    */
   public void setOutsideColor (byte[] color)
   {
      outsideColor = color;
   }
   
   /**
    * Sets the color for the texture inside the bounds of the slice projection
    */
   public void setInsideColor (byte[] color)
   {
      insideColor = color;
   }
   
   
   public static void setGoodScanConversionOrder (boolean good)
   {
      if (good)
         SCAN_CONVERSION_ORDER = new int[][]{{2,1}, {2, 0}, {1, 0}};
      else
         SCAN_CONVERSION_ORDER = new int[][] {{1,2}, {2, 0}, {0, 1}};
      if (SliceUtil.logger.isLoggable (Level.INFO))
      {
         String s = new String();
         if (SCAN_CONVERSION_ORDER[0][0] == 2)
            s += "Good scan conversion order: ";
         else
            s += "Bad scan conversion order: ";
         s += Arrays.deepToString (SCAN_CONVERSION_ORDER);
      SliceUtil.logger.info (s);      
      }      
   }
   
   /**
    * @return a byte array containing the texture.
    */
   public byte[] getTextureData ()
   {
      return textureData;
   }
   
   /**
    * @return the x,y dimensions of the texture. These are a power of 2.
    */
   public int[] getTextureDimensions ()
   {
      return textureDimensions;
   }
   
   /**
    * @return bounds[0] contains the lower and upper bounds for the x
    * coordinate bounds[1] contains the same for the y coordinate. These
    * coordinates delimit the rectangle that includes the intersection poligon
    * projected to maximum projection orthogonal plane.
    */
   public int[][] getBounds ()
   {
      return bounds;
   }
   
   /**
    * @return the 3d points that form the intersection polygon.
    */
   public Point3d[] getIntersection3d ()
   {
      return intersection3d;
   }
   
   /**
    * @return the 2d points that form the intersection polygon seen from
    * a direction normal to the plane of the polygon.
    */
   public Point2d[] getIntersection2d ()
   {
      return intersection2d;
   }
   
   /**
    * @return the axis perpendicular on the plane
    *         for which the slice has the maximum projection area
    */
   public int getMaxProjectionAxis ()
   {
      return maxProjectionAxis;
   }
   
   /**
    * @return the 2d points that form the intersection polygon, translated
    * and scaled such that they are between [0,1].
    */
   public Point2d[] getTextureCoordinates ()
   {
      return textureCoordinates;
   }
   
   
   /**
    * Initialize the slice by allocating memory for the data to be read,
    * seting the default transfer function and performing other initialisations
    * This initialisation can be done once per iteration. In an iteration the slice will
    * slide along its normal until the entire volume will be traversed.
    */
   public void init ()
   {   
      this.dataSource.activate();
   
      // set the default texture function
      if (transferFunction == null)
         setTransferFunction (TransferFunction.createDefault (dataSource));
      
      // allocate texture data
      int diagonal = SliceUtil.nextPowerOf2 (
              (int)Math.ceil (volume.getMaxDiagonal ()));
      textureData = new byte[diagonal*diagonal*bytesPerTexel];      
      calculateSliceTranslation ();
   }
      
   /**
    * Calculates the intersection between the {@link Plane} and the
    *  {@link Volume} and fills the the intersection poligon with data
    */
   public void read ()
   {
      ArrayList<Edge> edgeList;
      
      numberOfReads = 0;
      long begin = 0;
      if (transferFunction == null)
      {
         SliceUtil.logger.warning ("Transfer function is not set");
         return;
      }
      if (SliceUtil.logger.isLoggable (Level.FINE))
         begin = new Date ().getTime ();
      if ((intersection3d = volume.calculateIntersectionPoints (plane)) == null)
         return;      
      intersection2d = faceAxisZeroVerticalAxisOne (
         plane.getNormal (), vertical, intersection3d);
      SliceUtil.orderPointsForSimpleClosedPath (
              intersection2d, intersection3d);
      translatePoligonAndBoundingBox (intersection2d);
      edgeList = getEdgeList (intersection3d);
      read (edgeList);
      textureCoordinates = calculateTextureCoordinates (
              intersection3d, getMaxProjectionAxis (),
              textureDimensions, bounds);
      if (SliceUtil.logger.isLoggable (Level.FINE))
      {
         long end = new Date ().getTime ();
         SliceUtil.logger.fine (
                 "Slice\n" + toString () + "\n" +
                 "time for " + numberOfReads + " reads " +
                 " is " + ((double)(end - begin))/1000 + "s\n");
         log ("slice.out");
      }
   }
   
   /**
    * Specifies a vector in the intersection plane. The
    * intersection poligon is displayed rotated such that this vector is vertical.
    */
   public void setVertical (Vector3d vertical)
   {
      this.vertical = vertical;
   }
   
   public ISBounds getSpaceBounds ()
   {
      return new ISBounds (getSpaceShapeArray ());
   }
   
   public int[] getSpaceShapeArray ()
   {
//      maxProjectionAxis = maxProjectionAxis (plane.getNormal ());      
//      return volume.getSpaceShapeArray (plane, maxProjectionAxis);

      // PJR: The above code may be broken, so I'll try something crude instead.
      
      return this.dataSource.getBounds().getDimensions().getCoords();

   }

   /**
    * Required for ObliquePlaneIteratorInterface
    */
   public float[] getNormalArray ()
   {
      Vector3d normal = plane.getNormal ();
      
      return new float[] {(float)normal.x, (float)normal.y, (float)normal.z};
   }
   
   public int[] getPlaneShapeArray ()
   {
      int[] planeShapeArray = new int[2];
      double[] lower = volume.getLowerArray ();
      double[] upper = volume.getUpperArray ();
      for (int i = 0; i < 2; i++)
         planeShapeArray[i] = (int)
         (upper[SCAN_CONVERSION_ORDER[maxProjectionAxis][i]] -
                 lower[SCAN_CONVERSION_ORDER[maxProjectionAxis][i]]);
      return planeShapeArray;
   }
   
   /**
    * Calculates the translation for the 2D projection of intersection polygon
    * such that the 2D polygon is stationary as the polygon move through the volume.
    * This function needs to be called each time the normal of the plane changes.
    */
   public void calculateSliceTranslation ()
   {
      Point3d l = volume.getLower ();
      Point3d u = volume.getUpper ();
      Point3d[] volumeCenter = new Point3d[1];
      volumeCenter[0] = new Point3d ((u.x+l.x)/2, (u.y+l.y)/2, (u.z+l.z)/2);
      Point2d[] v = faceAxisZeroVerticalAxisOne (plane.getNormal (), vertical, volumeCenter);
      SliceUtil.logger.fine ("intersection2dVertical=" + vertical);
      volumeCenterProjection = new Vector2d (-v[0].x, -v[0].y);      
   }
   
   @Override
   public String toString ()
   {
      NumberFormat nf = NumberFormat.getInstance ();
      nf.setMaximumFractionDigits (1);
      nf.setMinimumFractionDigits (1);
      StringBuffer msg =  new StringBuffer (
              "volume " + volume + "\n" +
              "plane " + plane + "\n" +
              "textureDimensions " + nf.format (textureDimensions[0]) + " " +
              nf.format (textureDimensions[1]) + "\n" +
              "bounds " + nf.format (bounds[0][0]) + " " +
              nf.format (bounds[0][1]) + " " +
              nf.format (bounds[1][0]) + " " + nf.format (bounds[1][1]) + "\n");
      msg.append ("intersection3d ");
      for (Point3d intersection: intersection3d)
         msg.append (intersection + " ");
      msg.append ("\nintersection2d ");
      for (Point2d intersection: intersection2d)
         msg.append (intersection + " ");
      msg.append ("\ntextureCoordinates ");
      for (Point2d textureCoordinate: textureCoordinates)
         msg.append (textureCoordinate + " ");
      msg.append ("\nDataSource " + dataSource.getAxisOrdering () +
              "\nProjection Plane " +
              Arrays.toString (SCAN_CONVERSION_ORDER[maxProjectionAxis]));
      return msg.toString ();
   }
   
   /**
    * Project the 'source' on a plane perpendicular on 'axis' and store the
    * result in 'dest'.
    */
   public static void projectTuple (Tuple3d source, Tuple2d dest, int axis)
   {
      double[] s = new double[3];
      source.get (s);
      dest.set (s[SCAN_CONVERSION_ORDER[axis][0]], s[SCAN_CONVERSION_ORDER[axis][1]]);
   }
      
   /**
    * Calculate the area of a poligon using Discrete Green Theorem.
    * http://www.mathpages.com/home/kmath201.htm
    */
   private static double calculateSliceProjectionArea (List edgeList, int axis)
   {
      Point2d left = new Point2d (), right = new Point2d ();
      double area = 0;
      for(Iterator it = edgeList.iterator (); it.hasNext ();)
      {
         Edge edge = (Edge)it.next ();
         projectTuple (edge.p1, left, axis);
         projectTuple (edge.p2, right, axis);
         area += (right.x*left.y - left.x*right.y);
      }
      return Math.abs (area / 2);
   }
      
   /**
    * 3D Scan-Conversion Algorithms for Voxel-Based Graphics,
    * Arie Kaufman
    */
   private void read (ArrayList<Edge> edgeList)
   {
      maxProjectionAxis = maxProjectionAxis (plane.getNormal ());
      Comparator<Edge> edgeComparatorYIncreasing =
              new Comparator<Edge> ()
      {
         public int compare ( Edge  e1, Edge e2)
         {
            //Edge e1 = (Edge)o1, e2 = (Edge)o2;
            double v =
                    SliceUtil.get (e1.p1, SCAN_CONVERSION_ORDER[maxProjectionAxis][1]) -
                    SliceUtil.get (e2.p1, SCAN_CONVERSION_ORDER[maxProjectionAxis][1]);
            return (v > 0) ? 1 : ((v == 0) ? 0 : -1);
         }
      };
      double A = plane.get (SCAN_CONVERSION_ORDER[maxProjectionAxis][0]),
              B = plane.get (SCAN_CONVERSION_ORDER[maxProjectionAxis][1]),
              C = plane.get (maxProjectionAxis), D = plane.get (3);
      double dz_x = - A/C;
      double dz_y = -B/C;
      initEdgeList (SCAN_CONVERSION_ORDER[maxProjectionAxis],
              dz_x, edgeList, bounds);
      initCurrentSlice ();
      Collections.sort (edgeList, edgeComparatorYIncreasing);
      processScanLines (edgeList, dz_x, dz_y, D/C);
   }
   
   public void log (String fileName)
   {
      try
      {
         FileOutputStream sliceOut = new FileOutputStream (fileName);
         sliceOut.write (textureData, 0,
                 textureDimensions[0]*textureDimensions[1]*getBytesPerTexel ());
         sliceOut.close ();
      }
      catch(Exception e)
      {
         e.printStackTrace ();
      }
   }
      
   /**
    * Returns the index in the slice, where
    * 'bounds[0][0]' <= x <= 'bounds[0][1]' and
    * 'bounds[1][0]' <= y <= 'bounds[1][1]'
    */
   private int index (int[][] bounds, int[] dimensions, int x, int y)
   {
      return ((y - bounds[1][0]) * dimensions[0] +
              (x - bounds[0][0])) * getBytesPerTexel ();
   }
   
   private void set (int[][] bounds, int[] dimensions, int x, int y, byte[] b)
   {
      int index = index (bounds, dimensions, x, y);
      System.arraycopy (b, 0, textureData, index, getBytesPerTexel ());
   }
   
   private void logLine (int[][] bounds, int[] dimensions, int y)
   {
      //int index = index (bounds, dimensions, bounds[0][0], y);
      StringBuffer line = new StringBuffer (1024);
      for (int i = 0; i < (bounds[0][1] - bounds[0][0]); i++)
      {
         line.append (Byte.toString (textureData[i]));
         line.append (' ');
      }
      SliceUtil.logger.finest ("line " + y + ": " + line);
   }
   
   
   /**
    * Calculates the 'dimensions' of the texture (must be powers of 2)
    * Fills the texture with 'insideColor' within the 'bounds' and with
    * 'outsideColor' in the rest of it.
    */
   private void initCurrentSlice ()
   {
      textureDimensions[0] = SliceUtil.nextPowerOf2 (
              bounds[0][1] - bounds[0][0] + 1);
      textureDimensions[1] = SliceUtil.nextPowerOf2 (
              bounds[1][1] - bounds[1][0] + 1);
      int xSize = bounds[0][1] - bounds[0][0] + 1;
      int ySize = bounds[1][1] - bounds[1][0] + 1;
      // the slice is blue by default
      for(int y = 0; y < textureDimensions[1]; y++)
         for(int x = 0; x < textureDimensions[0]; x++)
         {
         int index = (y*textureDimensions[0] + x)*3;
         if (x > xSize || y > ySize)
            System.arraycopy (outsideColor,
                    0, getTextureData (), index, getBytesPerTexel ());
         else
            System.arraycopy (insideColor, 0,
                    getTextureData (), index, getBytesPerTexel ());
         }
   }
   
   private void processScanLines (ArrayList<Edge> edgeList, double dz_x,
           double dz_y, double DslashC)
   {
      Comparator<Edge> edgeComparatorXIntersectionIncreasing =
              new Comparator<Edge> ()
      {
         public int compare (Edge e1, Edge e2)
         {
            return (int) ( e1.x - e2.x );
         }
      };
      ArrayList<Edge> activeEdgeList = new ArrayList<Edge> (edgeList.size ());
      int indexEdgesCurrentY = 0;
      Edge first = null;
      double z0 = 0, z;
      for (int y = bounds[1][0]; y <= bounds[1][1]; y++)
      {
         for(Iterator it = activeEdgeList.iterator (); it.hasNext ();)
         {
            Edge l = (Edge)it.next ();
            l.x += l.dx;
         }
         indexEdgesCurrentY = addNewEdges (
                 y, edgeList, SCAN_CONVERSION_ORDER[maxProjectionAxis],
                 activeEdgeList, indexEdgesCurrentY);
         removeOldEdges (y, SCAN_CONVERSION_ORDER[maxProjectionAxis],
                 activeEdgeList);
         
         Collections.sort (
                 activeEdgeList, edgeComparatorXIntersectionIncreasing);
         Iterator currentIt = activeEdgeList.iterator ();
         Edge last = activeEdgeList.get (activeEdgeList.size () - 1);
         Edge current = (Edge)currentIt.next ();
         if (current == first)
            z0 += current.dxz + dz_y;
         else
         {
            first = current;
            z0 = current.x*dz_x + y*dz_y - DslashC + 0.5;
         }
         z = z0;
         processScanLine (y, last, dz_x, z, current, currentIt);
      }
   }
   
   private void processScanLine (int y, Edge last,
           double dz_x, double z, Edge current, Iterator currentIt)
   {
      IndexSpaceID position = new IntegerIndexSpaceID (0,0,0);
      boolean fill = false;
      for (int x = (int)current.x; x <= last.x; x++, z+= dz_x)
      {
         // both left and right integer coordinates are "inside"
         if (!fill && current.x <= x ||
                 fill && current.x < x)
         {
            fill = ! fill;
            if (currentIt.hasNext ())
               current = (Edge)currentIt.next ();
         }
         
         if (fill)
         {
            position.setCoord (SCAN_CONVERSION_ORDER[maxProjectionAxis][0], x);
            position.setCoord (SCAN_CONVERSION_ORDER[maxProjectionAxis][1], y);
            position.setCoord (maxProjectionAxis, (int)z);
            Datum datum = dataSource.datum (position);
            numberOfReads++;
            transferFunction.convert (datum, textureData,
                    index (bounds, textureDimensions, x, y));
         }
      }
      if (SliceUtil.logger.isLoggable (Level.FINEST))
         logLine (bounds, textureDimensions, y);
   }
   
   /**
    * Returns the bounds of the rectangle containing the data as a two
    * dimensional array 'bounds': [0][0,1]=x0, xf; [1][0,1]=y0, yf
    * Removes horizontal edges from the edge list.
    */
   private static void initEdgeList (int[] projectionCoord, double dz_x,
           ArrayList edgeList, int[][] bounds)
   {
      int x0 = Integer.MAX_VALUE, xf = Integer.MIN_VALUE;
      int y0 = Integer.MAX_VALUE, yf = Integer.MIN_VALUE;
      double current;
      Edge l;
      for(Iterator it = edgeList.iterator (); it.hasNext ();)
      {
         l = (Edge)it.next ();
         
         if (SliceUtil.get (l.p1, projectionCoord[1]) >
                 SliceUtil.get (l.p2, projectionCoord[1]))
         {  Point3d t = l.p1; l.p1 = l.p2;l.p2 = t;}
         else if (SliceUtil.get (l.p1, projectionCoord[1]) ==
                 SliceUtil.get (l.p2, projectionCoord[1]))
         {
            it.remove ();
            continue;
         }
         
         x0 = l.minValue (x0, projectionCoord[0]);
         y0 = l.minValue (y0, projectionCoord[1]);
         xf = l.maxValue (xf, projectionCoord[0]);
         yf = l.maxValue (yf, projectionCoord[1]);
         
         if (yf < (current = Math.floor (
                 SliceUtil.get (l.p1, projectionCoord[1]))))
            yf = (int)current;
         if (yf < (current = Math.floor (
                 SliceUtil.get (l.p2, projectionCoord[1]))))
            yf = (int)current;
         
         l.dx = (SliceUtil.get (l.p2, projectionCoord[0]) -
                 SliceUtil.get (l.p1, projectionCoord[0])) /
                 (SliceUtil.get (l.p2, projectionCoord[1]) -
                 SliceUtil.get (l.p1, projectionCoord[1]));
         l.x = SliceUtil.get (l.p1, projectionCoord[0]) +
                 (Math.ceil (SliceUtil.get (l.p1, projectionCoord[1])) -
                 SliceUtil.get (l.p1, projectionCoord[1]))*l.dx;
         l.dxz = l.dx * dz_x;
      }
      bounds[0][0] = x0; bounds[0][1] = xf;
      bounds[1][0] = y0; bounds[1][1] = yf;
   }
   
   private static int addNewEdges (
           int y, ArrayList<Edge> edgeList, int[] projectionCoord,
           ArrayList<Edge> activeEdgeList, int indexEdgesCurrentY)
   {
      Edge l;
      while(indexEdgesCurrentY < edgeList.size () &&
              ((int)Math.ceil (SliceUtil.get (
              (l = edgeList.get (indexEdgesCurrentY)).p1,
              projectionCoord[1]))) == y)
      {
         activeEdgeList.add (l);
         indexEdgesCurrentY++;
      }
      //System.err.println(activeEdgeList);
      return indexEdgesCurrentY;
   }
   
   private static void removeOldEdges (int y, int[] projectionCoord,
           ArrayList activeEdgeList)
   {
      Edge l;
      // remove old edges from active edge list
      for(Iterator it = activeEdgeList.iterator (); it.hasNext ();)
      {
         l = (Edge)it.next ();
         if (SliceUtil.get (l.p2, projectionCoord[1]) < y)
            it.remove ();
      }
      //System.err.println(activeEdgeList);
   }
   
   /**
    * @return -1 if there is no intersection, or the axis that gets the maximum
    *          poligon projection on the plane perpendicular to it.
    */
   private static int maxProjectionAxis (List edgeList)
   {
      double maxArea = 0;
      int axis = -1;
      for(int i = 0; i < 3; i++)
      {
         double area = calculateSliceProjectionArea (edgeList, i);
         if (area - maxArea > SliceUtil.ZERO)
         {
            maxArea = area;
            axis = i;
         }
      }
      return axis;
   }
   
   private static int maxProjectionAxis (Vector3d n)
   {
      Vector3d result = new Vector3d ();
      double min = Double.POSITIVE_INFINITY;
      int minIndex = -1;
      for (int i = 0; i < 3; i++)
      {
         result.set (PRINCIPAL_AXIS[i]);
         result.cross (result, n);
         double value = result.lengthSquared ();
         if (value < min)
         {
            min = value;
            minIndex = i;
         }
      }
      return minIndex;
   }
   
   
   
   
   /**
    * Sets axisOrder like this:
    * axisOrder[0], axisOrder[1] are set based on 'dataSourceAxisOrder',
    * from right to left such that, axisOrder[0] is the coordinate that changes
    * more often, and then axisOrder[1].
    */
    /*
    protected static void scanConversionAxisOrder (
                   AxisOrdering dataSourceAxisOrder, int axis, int[] axisOrder)
    {
        for(int i = 0, j = 0; i < dataSourceAxisOrder.dim(); i++)
        {
            int newAxis = dataSourceAxisOrder.getNthLeastSignificantAxis (i);
            if(newAxis != axis)
                axisOrder[j++] = newAxis;
        }
    }
     */
   
   private static ArrayList<Edge> getEdgeList (Point3d[] intersection3d)
   {
      ArrayList<Edge> edgeList = new ArrayList<Edge>(intersection3d.length);
      for (int i = 0; i < intersection3d.length; i++)
         edgeList.add (new Edge (
                 intersection3d[i],
                 intersection3d[(i+1)%intersection3d.length]));
      return edgeList;
   }
   
   private static Point2d[] calculateTextureCoordinates (
           Point3d[] intersection3d,
           int axis, int[] dimensions, int[][] bounds)
   {
      Point2d[] texture = new Point2d[intersection3d.length];
      for(int i = 0; i < intersection3d.length; i++)
      {
         texture[i] = new Point2d ();
         Point3d point3d = intersection3d[i];
         projectTuple (point3d, texture[i], axis);
         texture[i].x = (texture[i].x - bounds[0][0])/(dimensions[0] - 1);
         texture[i].y = (texture[i].y - bounds[1][0])/(dimensions[1] - 1);
      }
      return texture;
   }
   
   /**
    * Translate the intersection poligon such that the projection of the
    * center of the volume is centered in the origin.
    */
   private void translatePoligonAndBoundingBox (Point2d[] intersection2d)
   {
      for(int i = 0; i < intersection2d.length; i++)
      {
         Point2d p = intersection2d[i];
         p.add (volumeCenterProjection);
      }
   }
   
   /**
    * Calculate the rotation matrix so that you face axisZero and axisOne is vertical,
    * and rotate all points in intersection3d with that rotation matrix. 
    * axisZero and axisOne should be perpendicular before the execution of this 
    * function. Points in intersection3d are in a plane perpendicular on axisZero 
    * before the execution of this function.
    */
   private static Point2d[] faceAxisZeroVerticalAxisOne (
      Vector3d axisZero,
      Vector3d axisOne,
      final Point3d[] intersection3d)
   {
      Point3d result = new Point3d ();            
      Point2d[] intersection2d = new Point2d[intersection3d.length];
      
      Matrix4d rotation = SliceUtil.calculateRotation (axisZero, axisOne, 
         SliceUtil.Z_AXIS, SliceUtil.Y_AXIS);
      for(int i = 0; i < intersection3d.length; i++)
      {
         rotation.transform (intersection3d[i], result);
         intersection2d[i] = new Point2d (result.x, result.y);
      }
      return intersection2d;
   }

   
   private Volume volume;
   private Plane plane;
   private DataSource dataSource;
   private byte[] textureData;
   private int[] textureDimensions = new int[2];
   private int[][] bounds = new int[2][2];
   /**
    * axisOrder[0], axisOrder[1] = x,y coordinate along which we do scan conversion
    * Because x changes more often, we have to choose these two coordinates
    * from rigth to left from dataSource.getAxisOrdering().
    */
   //private int[] axisOrder = new int[2];
   
   /**
    * maxProjectionAxis coordinate that has the perpendicular plane with the
    * the maximum projection area of the intersection polygon on it.
    */
   private int maxProjectionAxis;
   private Point3d[] intersection3d;
   private Point2d[] intersection2d;
   private Point2d[] textureCoordinates;
   private Vector3d vertical;
   
   private int bytesPerTexel = 4;
   private byte[] outsideColor;
   private byte[] insideColor;
   private long numberOfReads;
   private Vector2d volumeCenterProjection;
   private TransferFunction transferFunction;
   
   /**
    * PERPENDICULAR_PLANE[i] stores the axes of the plane perpendicular on
    * axis i. The first index specifies the axis along which scan lines are calculated. (x)
    * The second index specifies the axis along which you read the next scan line. (y)
    */
   /* *
    * good scan conversion order
    */
   //private static int[][] SCAN_CONVERSION_ORDER = {{2,1}, {2, 0}, {1, 0}};
   /**
    * bad scan conversion order
    */
   private static int[][] SCAN_CONVERSION_ORDER = {{1,2}, {2, 0}, {0, 1}};
   private static final double[][] PRINCIPAL_AXIS = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}};
}
