/*
 * SliceUtil.java
 * 
 *------------------- edit history -----------------------------
 *06/06/08 rdb: Made ArrayList, Comparator generic
 * 
 */
package edu.unh.sdb.datasource;
import java.util.Arrays;
import java.util.Comparator;
import java.util.logging.Logger;
import javax.vecmath.AxisAngle4d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;
import javax.vecmath.Point3i;
import javax.vecmath.Tuple3d;
import javax.vecmath.Vector3d;


/**
 * Utility functions for use in {@link Slice}.
 * @author dlipsa
 */
public class SliceUtil
{
   
   public static int nextPowerOf2 (int value)
   {
      return (int)Math.pow (2, Math.ceil (Math.log (value)/Math.log (2)));
   }
   
   public static double get (Tuple3d t, int axis)
   {
      switch(axis)
      {
         case 0: return t.x;
         case 1: return t.y;
         case 2: return t.z;
         default: throw new IllegalArgumentException (new Integer (axis).toString ());
      }
   }
      
   public static void set (Tuple3d t, int axis, double value)
   {
      switch(axis)
      {
         case 0: t.x = value;break;
         case 1: t.y = value;break;
         case 3:t.z = value;break;
         default: throw new IllegalArgumentException (new Integer (axis).toString ());
      }
   }
   
   /**
    * Intersection between a ray segment and a plane
    * section 13.9 3D Math Primer for
    * Graphics and Game Development by Dunn&Parberry
    */
   public static Point3d segmentPlaneIntersection (
         Vector3d p0, Vector3d dv,
         double leftLimit, double rightLimit,
         Vector3d n, double d)
   {
      double dot = dv.dot (n);
      if (Math.abs (dot) >= SliceUtil.ZERO)
      {
         double t = (d - p0.dot (n)) / dot;
         if (t >= leftLimit && t <= rightLimit)
         {
            Point3d intersection = new Point3d ();
            intersection.scaleAdd (t, dv, p0);
            return intersection;
         }
      }
      return null;
   }
   
   public static void swap (Object[] a, int i, int j)
   {
      Object t = a[i];
      a[i] = a[j];
      a[j] = t;
   }
   
   
   /**
    * Order the points to draw a simple closed path, from
    * Algorithms in C++, Sedgewick, chapter 24.
    */
   public static void orderPointsForSimpleClosedPath (
         Point2d[] intersection2d, Point3d[] intersection3d)
   {
      if (intersection2d.length == 0)
         return;
      
      Integer[] index = new Integer[intersection2d.length];
      for (int i = 0; i < index.length; i++) index[i] = new Integer (i);
      /**
       * angle between the point, the anchor point and
       * the positive horizontal direction.
       */
      final double[] angle = new double[index.length];
      
      // otherwise the algorithm given in Sedgewick doesn't work.
      lowestYFirst (intersection2d, intersection3d);
      Point2d anchor = intersection2d[0];
      
      for(int i = 1; i < intersection2d.length; i++)
         angle[i] = theta (anchor, intersection2d[i]);
      Arrays.sort (index,
            new Comparator<Integer>()
      {
    	 public int compare( Integer i1, Integer i2 )
    	 {
    		 int i = i1.intValue();
    		 int j = i2.intValue();
    		 if ( angle[ i ] < angle[ j ] )
    			return -1;
    		 else if ( angle[ i ] == angle[ j ] )
    			return 0;
    		 else 
    			return 1;
    	 }
    	 /****
         public int compare (Object o1, Object o2)
         {
            int i = ((Integer)o1).intValue ();
            int j = ((Integer)o2).intValue ();
            if (angle[i] < angle[j]) return -1;
            else if (angle[i] == angle[j]) return 0;
            else return 1;
         }
         /*******************/
      });
      insitu (intersection2d, index.clone ());
      insitu (intersection3d, index);
   }               
   
   /**
    * Calculate the rotation to get from 'srcZero' and 'srcOne' to
    * 'destZero' and 'destOne' .
    * srcZero and srcOne should be perpendicular before the execution of this 
    * function and destZero and destOne should be perpendicular as well.
    */
   public static Matrix4d calculateRotation (
      Vector3d srcZero, Vector3d srcOne,
      Vector3d destZero, Vector3d destOne)
   {
      Matrix4d rotationZero = new Matrix4d ();
      Matrix4d rotationOne = new Matrix4d ();
      
      AxisAngle4d axisAngle = new AxisAngle4d ();
      calculateRotation (srcZero, destZero, axisAngle);
      SliceUtil.logger.fine ("zero=" + axisAngle + " degrees=" + axisAngle.angle * 180 / Math.PI);
      rotationZero.set (axisAngle);
      
      Vector3d axis = new Vector3d (srcOne);
      rotationZero.transform (axis);         
      calculateRotation (axis, destOne, axisAngle);
      SliceUtil.logger.fine ("one=" + axisAngle + " degrees=" + axisAngle.angle * 180 / Math.PI);         
      rotationOne.set (axisAngle);           
      rotationOne.mul (rotationZero);
      return rotationOne;
   }
      
   static public Point3d toPoint3d (Point3i p)
   {
      return new Point3d (p.x, p.y, p.z);
   }
   
   public static void calculateRotation (Vector3d src, Vector3d dest, Matrix4d m)
   {
      AxisAngle4d axisAngle = new AxisAngle4d ();
      calculateRotation (src, dest, axisAngle);
      m.set (axisAngle);
   }
   
   /**
    * Calculate the rotation to get from 'src' to 'dest' and store that in 
    * 'axisAngle'
    */
   public static void calculateRotation (
      Vector3d src,
      Vector3d dest,
      AxisAngle4d axisAngle)
   {
      if (src.length() < ZERO || dest.length() < ZERO)
      {
         axisAngle.set (new Vector3d (0, 0, 1), 0);
         return;
      }
      Vector3d normal = new Vector3d ();
      normal.cross (src, dest);
      double angle;
      if (normal.length () < ZERO)
      {
         // src and dest are parallel.
         getANormal (src, normal);
         angle = 0;
      }
      else
      {
         double sinAngle = normal.length () /
            src.length () / dest.length();

         // find the angle, which will be in [-pi, pi]
         angle = Math.asin (sinAngle); /* angle in [-pi/2, pi/2] */
      }
      
      double dotProduct = src.dot (dest);
      if (dotProduct < -ZERO) /* cosAngle < 0 => angle in [-pi,-pi/2] or  [3*pi/2,pi] */
      {
         if (angle < 0)       /* angle in [-pi/2,0] */
            angle = - Math.PI - angle; /* angle in [-pi, -pi/2]*/
         else                /* angle in [0, pi/2]*/
            angle = Math.PI - angle; /* angle in [pi/2, pi] */
      }

      // for some reason when you set a rotation matrix from a AxisAngle
      // the axis needs to be normalized otherwise you get the wrong result.
      normal.normalize ();
      axisAngle.set (normal, angle);      
   }
   
   
   /**
    *  Modifies 'n' to contain a vector perpendicular on 'value'. 'n' should be
    *  allocated and 'value' should be non 0.
    */      
   private static void getANormal (Vector3d v, Vector3d n)
   {
      if (Math.abs(v.x) > ZERO && Math.abs(v.y) > ZERO)
      {
         n.x = 1;n.y = - v.x / v.y;n.z = 0;                 
         return;
      }
      if (Math.abs(v.x) > ZERO && Math.abs(v.z) > ZERO)
      {
         n.x = 1;n.y = 0;n.z = - v.x / v.z;         
         return;
      }      
      if (Math.abs(v.y) > ZERO && Math.abs(v.z) > ZERO)
      {
         n.x = 0;n.y = 1;n.z = - v.y / v.z;         
         return;
      }                  
      if (Math.abs(v.x) > ZERO)
      {
         n.x = 0;n.y = 1;n.z = 0;                 
         return;
      }
      if (Math.abs(v.y) > ZERO)
      {
         n.x = 0;n.y = 0;n.z = 1;
         return;
      }      
      if (Math.abs(v.z) > ZERO)
      {
         n.x = 1;n.y = 0;n.z = 0;
         return;
      }
      assert false;
      n.x = 1; n.y = 0; n.z = 0;
   }
   
   
   /**
    * Rearange an array insitu, when you have the correct indexes
    * Sedgewick, Elementary Sorting
    */
   private static void insitu (Object[] a, Integer index[])
   {
      int i, j, k; Object t;
      for (i = 0; i < index.length; i++)
         if (index[i].intValue () != i)
         {
         t = a[i]; k = i; 	/*k is the hole that has to be filled*/
         do
         {
            /*we fill hole k, the next hole is index[k]*/
            j = k; a[j] = a[index[j].intValue ()];
            k = index[j].intValue (); index[j] = new Integer (j); /*we mark that we filled hole k*/
         }
         while (k != i);
         a[j] = t;
         }
   }
   
   private static void lowestYFirst (Point2d[] intersection2d, Point3d[] intersection3d)
   {
      int anchorIndex = 0;
      Point2d anchor = intersection2d[0];
      for(int i = 1; i < intersection2d.length; i++)
      {
         if (intersection2d[i].y < anchor.y)
         {
            anchorIndex = i;
            anchor = intersection2d[anchorIndex];
         }
      }
      SliceUtil.swap (intersection2d, 0, anchorIndex);
      SliceUtil.swap (intersection3d, 0, anchorIndex);
   }
   
   static private double theta (Point2d p1, Point2d p2)
   {
      double dx, dy, ax, ay;
      double t;
      dx = p2.x - p1.x; ax = Math.abs (dx);
      dy = p2.y - p1.y; ay = Math.abs (dy);
      t = (ax+ay < SliceUtil.ZERO) ? 0 : dy/(ax+ay);
      if (dx < 0) t = 2-t;
      else if (dy < 0) t = 4+t;
      return t*90.0;
   }

   public static Vector3d X_AXIS = new Vector3d (1, 0, 0);
   public static Vector3d Y_AXIS = new Vector3d (0, 1, 0);
   public static Vector3d Z_AXIS = new Vector3d (0, 0, 1);
   
   public static double ZERO = 0.000001;
   static Logger logger = Logger.getLogger ("edu.unh.sdb.datasource.Slice");
}
