/*
 * SparseDataSource -- partitions a data source; some partitions are allowed
 *                     to be "missing". In principle, all points in a missing
 *                     partition receive a default value (0). In practice
 *                     we expect it to be used for AR representations.
 * 
 * Pending issues:
 *    byteOrder (may work)
 *    dataOrder -- only supports point order
 *    subCompositeBlock not implemented
 * 
 * Thought of storing the sparse data in a 1D ChunkedDataSource; problem is
 * that the chunks have to turn into an array of Datums which really should
 * then be organized in 3D (e.g.) I don't think that will gain us much over 
 * just using a file. 
 * 
 * On the other hand, 
 * 1. how do we deal with the attribute order vs point order datasources
 * 2. The existing approach should be able to handle  CompositeDataSources -- 
 *    by converting them to single-file datasources. Is that acceptable in the 
 *    long run?
 * 3. Could handle all these (though not necessarily uniformly) by requiring
 *    the original ds to be accessible and always accessing the data from
 *    the original -- however, I think we'll lose performance and distribution
 *    efficiency.
 * 
 * 
 * Created
 *    12/22/11 to 03/05/12 rdb Initial implementation.  
 * 
 * History
 *    03/05/12 Seems to work when everything is full blocks and when all
 *             blocks are present -- this is essentially equiv to a ChunkedDS
 *    02/22/14 Can changed private DataBlock getDataBlock( IndexSpaceID spaceLoc )
 *    		   by using mapInfo to find mapped data block id from sparse to origin
 *    		   added "public int mappedBlockIndex(int originBlockIndex)" 
 */
 
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

import java.util.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;

/**
 * Implements a data source that only contains data for a subset of its bounds.
 * The parent class partitions the bounds space into fixed size blocks except
 * for the border blocks that can be smaller if the block size does not divide
 * exactly into the data source bounds. 
 * 
 * This version maintains a cache of blocks and an LRU replacement algorithm.
 * The default # partitions is 64.
 * 
 * IMPORTANT: the physical blocks in the sparse data source are all the same 
 * size, so that the location of any physical block can easily be converted
 * to a seek address in the file. This means that partial blocks will be null
 * padded in the physical file. The bounds of the SparseDataSource define the
 * bounds of the true data, so the code can easily figure out what blocks are 
 * partial; for the time being, the mapper data puts the size of all non-
 * standard blocks in the xfdl file.
 * 
 * This is intended for use with the simulation interface to AR data, but
 * could become a standard AR tool.
 * 
 * The xfdl file for a sparse data source includes a mapping of blocks in the
 * virtual DataSource to blocks in the physical file.
 * 
 * For now, access to non-existent locations in the virtual DataSource geometry
 * are considered to be an error. 
 * 
 * Later, we can add some methods that allow program to specify that missing
 * data is to be replaced by zero or some other value -- probably attribute 
 * specific.
 * 
 * @author rdb
 * 
 * Created between 12/22/11 and 03/05/12
 */
// BinaryFileDataSource creates and opens the physical file for us
public class SparseDataSource extends BinaryFileDataSource
{
   //------------------ class variables ---------------------------------
   
   //------------------ instance variables ------------------------------
   protected ISBounds blockBounds = null;
   protected HashMap<Integer,Integer> blockMapper = null; 
   
   protected DataSource origDS = null;  // Not always known or accessible
   protected String[]   mappingStrings = null;
   protected MapInfo[]  mapInfo = null;
   
   //----- variables for partition management
   protected ISBounds              partitioningBounds;
   protected RegularISPartitioning blockPartitioning;
   protected IndexSpaceID          blockSizes;  // not an ID, but size specification
   protected int                   partitionByteSize;  
   
   //------ variables for the cache
   protected HashMap<Integer, DBCache>  dbCache = null;   
   protected int     maxCacheSize = 0; 
   protected int     maxCacheSizeDefault = 64;
   protected long    cacheRefTime = 0; // used to order cache references
   
   //------------------ internal MapInfo class ---------------------------------
   public static class MapInfo
   {
      int   origBlkId;
      int   sparseBlkId;
      int[] lower = null;  // lower bound of block in orig ds
      int[] upper = null;  // upper bound of block in 
      int[] size  = null;  // size in each dimension of a block; this is chunk
                           //    size for all but (some) border blocks
      MapInfo( String mapString )
      {
         parseMapString( this, mapString );
      }
      @Override
      public String toString()
      {
         return sparseBlkId + 
                arrayToString( lower ) + "=" + 
                arrayToString( size  ) + "->" + origBlkId;
      }
   }
   //------------------- internal DBCache class --------------------------
   private static class DBCache
   {
      int       partitionNum; // 1D index of the partition
      DataBlock db;
      long      refTime; // position of DataBlock in reference history array
   }
   
   //------------------ constructors ------------------------------------
   /*
    * Maybe we should have other constructors allowing user to dynamically
    * create a SparseDataSource. For now, we'll settle for building a 
    * SparseDataSource from an fdl file.
    * 
    * There is a utility class, SparseDataTestGen, that generates the xfdl
    * files needed for a SparseDataSource by picking random blocks. 
    * Eventually, we'll need to morph that into a more general purpose
    * utility that will do the same based on an AR fdl file and an error 
    * tolerance -- or integrate it into the AR generation code.
    * 
    * If an app wants to create its own SparseDataSource, for now you would
    * have create an FDLDescriptor and pass it to the constructor that 
    * accepts an FDLDescriptor, which will require the same information 
    * that we would need to ask for in a SparseDataSource constructor.
    */
   /**
    * This is the mechanism for creating a sparse ds from an xfdl file, but
    * it is not normally used by an app, which should use Datasource.create, 
    * which calls this method. This is public unlike most equivalent 
    * Granite DataSource constructors -- this is probably a better
    * model to provide an app with the opportunity to create a specific
    * kind of DataSource dynamically -- compared to the complex and inconsistent
    * alternative public constructors that we provide elsewhere.
    * 
    * @param name - informal name for the ds
    * @param fdl  - fdl describing characteristics -- usually extracted from
    *               xfdl file.
    */
   public SparseDataSource( String name, FDLDescriptor fdl )
   {
      super( name, fdl );
      this.fdlDescriptor = fdl;
      initialize();
   }
   /**
    * Define the number of partitions to be kept in the cache memory.
    * 
    * @param nPartitions 
    */
   public void setCacheSize( int nPartitions )
   {
      if ( maxCacheSize == 0 || maxCacheSize > nPartitions )
      {
         // empty the cache
         //-- initialize hash map for cache
         dbCache = new HashMap<Integer, DBCache>();
         cacheRefTime = 0;
      }
      maxCacheSize = nPartitions;
   }
   //--------------------- initialize -------------------------------------
   protected final void initialize()
   {
      //-- set up block and partitioning variables
      blockSizes = fdlDescriptor.getBlockDimensions();
      if ( blockSizes == null )
      {
         System.err.println("SparseDataSource error: blocksizes null. " +
                       "Using test default: 4x4x4" );
         blockSizes = new IntegerIndexSpaceID( 4,4,4 );
      }

      setCacheSize( maxCacheSizeDefault );
      
      blockBounds = new ISBounds( blockSizes );
      blockPartitioning = new RegularISPartitioning( 
                             new IntegerIndexSpaceID( blockSizes ), 
                             getBounds(),
                             fdlDescriptor.getAxisOrdering() );
      partitioningBounds = blockPartitioning.getPartitioningBounds();
      
      //-- set up mapping information
      mappingStrings = fdlDescriptor.getSparseBlockMap();
      buildMapper( mappingStrings );
      
      //-- invoke parent code to allocate I/O buffer. For the time being,
      //   we will only worry about reading one partition at a time.
      partitionByteSize = blockBounds.volume() * recordDesc.byteSize();
      if ( !allocateBuffer( partitionByteSize ))
         new SDBError( SDBError.Fatal, "SparseDataSource can't allocate io buffer "
                             + partitionByteSize );
   }
   //--------------------- setBlockBounds( ISBounds ) ----------------------
   /**
    * set/reset the size of the blocks; only do this if the physical file
    * changes correspondingly -- not a public method.
    */
   protected void setBlockBounds( ISBounds blkBounds )
   {
      blockBounds = blkBounds.copy();
   }
   //++++++++++++++++++++++ Hash table methods ++++++++++++++++++++
   //--------------------- buildMapper() -------------------------
   /**
    * generate the block mapping hash table
    */
   protected void buildMapper( String[] mapStrings )
   {
      blockMapper = new HashMap<Integer,Integer>(); 
      mapInfo = new MapInfo[ mapStrings.length ];
      for ( int m = 0; m < mapInfo.length; m++ )
      {
         mapInfo[ m ] = new MapInfo( mapStrings[ m ]);
         addMap( mapInfo[m].origBlkId, mapInfo[m].sparseBlkId );
      }
   }
   //--------------------- addMap( int, int ) ----------------------
   /**
    * addMap - add another block map entry
    * 
    */
   protected void addMap( int original, int sparse )
   {
      blockMapper.put( original, sparse );
   }
   //--------------------- getMap( int ) ----------------------
   /**
    * getMap - get the sparse block id for the specified original block
    * 
    */
   public int getMap( int original )
   {
      Integer target = blockMapper.get( original );
      if ( target == null )
         return -1;
      else
         return target.intValue();
   }
   //------------------- getBlock( IndexSpaceID ) -------------------------
   /**
    * Return the block index (id) for the partition that contains the "loc" 
    *    IndexSpaceID.
    * 
    * @param loc
    * @return 
    */
   public int getBlock( IndexSpaceID loc )
   {
      System.err.println( "SparseDataBlock.getBlock( ISID ) not implemented!");
      return -1;
   }
   //---------------- getParts( ISBounds ) ---------------------------
   /**
    * returns an array of ISBounds representing the pieces of the input bounds
    * that have data in this sparse data source.
    */
   public ArrayList<ISBounds> getParts( ISBounds bnds )
   {
      ISBounds[] parts = blockPartitioning.mapToSpaceParts( bnds );
      ArrayList<ISBounds> goodBlocks = new ArrayList<ISBounds>();
      
      int missingParts = 0;
      for ( ISBounds part: parts )
      {
         long blockIndex = blockPartitioning.partitionOffset( part.getLower() );
         Integer fileIndex = blockMapper.get( (int) blockIndex );
         if ( fileIndex != null )
            goodBlocks.add( part );
         else
            missingParts++;
      }
      System.err.println( "SparseDS.getParts missing parts: " + missingParts );
      return goodBlocks;
      
   }
   
   //---------------- getParts( ISBounds ) ---------------------------
   /**
    * returns an array of ISBounds representing the pieces of the input bounds
    * that have data in this sparse data source.
    */
   public ISBounds getPart( IndexSpaceID loc )
   {
      ISBounds part = blockPartitioning.getContainingPartitionBounds( loc );
      long blockIndex = blockPartitioning.partitionOffset( part.getLower() );
      Integer fileIndex = blockMapper.get( (int) blockIndex );
      if ( fileIndex != null )
         return part;
      else
         return null;
   }
   //----------- abstract DataCollection datum methods to be overridden -------
   
   /** Given an IndexSpaceID and FieldIDMapper, copy the values of the Datum at 
    *  that position to the argument Datum. 
    * TODO Define error conditions
    */
   public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
   {
      RegularPartitionISID rpISID;
      DataBlock part = getDataBlock( did );
      if ( part == null )
      { 
         // this is debugging code;  final code should probably throw exception
         System.err.println( "SparseDataSource.datum( DataBlock, ISID )" +
            " isid not in data source: " + did ); 
      }
      else
      {
         rpISID = new RegularPartitionISID( blockPartitioning, did );
         part.datum( dtm, rpISID.offsetISID, mp );
      }
   }
   
   /** 
    * Given an IndexSpaceID and FieldIDMapper, copy the values of the Datum at 
    *  that position to the argument Datum. 
    * Don't need to explicitly call super, but I'm doing that because I
    * think we'll want to have our own code here.
    */
   @Override
   public void datum( Datum dtm, IndexSpaceID did )
   {  
      RegularPartitionISID rpISID;
      DataBlock part = getDataBlock( did );
      //System.out.println("get part in datum method!!"); // not getting here
      if ( part == null )
      { 
         // this is debugging code;  final code should probably throw exception
         System.err.println( "SparseDataSource.datum( DataBlock, ISID )" +
            " isid not in data source: " + did ); 
      }
      else
      {
         rpISID = new RegularPartitionISID( blockPartitioning, did );
         part.datum( dtm, rpISID.offsetISID );
      }
   }
   //+++++++++++++ override abstract subBasicBlock methods +++++++++++++++
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      subParts( dstBlk, dstBnds, srcBnds, null, null );
   }
   
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                       RecordSpec recSpec )
   {
      subParts( dstBlk, dstBnds, srcBnds, recSpec, null );
   }
   
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                       FieldIDMapper mp )
   {
      subParts( dstBlk, dstBnds, srcBnds, null, mp );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      new SDBError( SDBError.Warning, "RecordSpec subblock not implemented." );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         RecordSpec recSpec )
   {
      new SDBError( SDBError.Warning, "RecordSpec subblock not implemented." );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         FieldIDMapper mp )
   {
      new SDBError( SDBError.Warning, "FieldIDMapper subblock not implemented." );
   }
   
   //----------------------- SparseDataBlock utility methods ----------------
   /**
    * Read all parts of a subblock from the sparse ds
    * 
    */
   private void subParts( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                      RecordSpec rs, FieldIDMapper fm )
   {
      ISBounds[] parts = blockPartitioning.mapToSpaceParts( srcBnds );
      for ( ISBounds part: parts )
      {
         subPart( dstBlk, dstBnds, srcBnds, part, rs, fm );
      }
   }
   /**
    * Read a subblock portion of a partition needed for this subblock request
    * 
    * This version does NOT correctly handle edge partitions that are not
    * full size.
    */
   private void subPart( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                       ISBounds partBnds, RecordSpec rs, FieldIDMapper fm )
   {
      RegularPartitionISID rpISIDlow, rpISIDup;
      IndexSpaceID spaceLow = partBnds.getLower();
      IndexSpaceID spaceUp  = partBnds.getUpper();
      
      int[]  dstLow = dstBnds.copyLowerArray();
      int[]  dstUp  = dstBnds.copyUpperArray();
      
      IndexSpaceID dstLowISID = dstBnds.copyLower();
      IndexSpaceID dstUpISID  = dstBnds.copyUpper();
      
      DataBlock partDB = getDataBlock( spaceLow );
      
      if ( partDB == null )
      { 
    	  // set by position, iterator, index based ID, NOT need to create datum, just set a array 
    	  int size = partDB.numAttributes;
    	  //for ()
    	    //float[] array = { Float.NaN, Float.NaN, Float.NaN, Float.NaN }; 
    	  
    	  // setFloats, iterate over the partDB, 
    	  // for index base id
    	  
         // this is debugging code;  final code should probably throw exception
         System.err.println( "SparseDataSource.subblock ISBounds " +
            " is not in data source: " + partBnds + "\nvolume/datum of space: " + dstBlk.volume() ); //TODO
      }
      else
      {
         rpISIDlow = new RegularPartitionISID( blockPartitioning, spaceLow );
         //System.err.println( rpISIDlow );
         rpISIDup = new RegularPartitionISID( blockPartitioning, spaceUp );
         ISBounds partSrc = new ISBounds( rpISIDlow.getOffsetISID(),
                                          rpISIDup.getOffsetISID() );
         
         for ( int d = 0; d < dstBlk.dim(); d++ )
         {
            dstLow[ d ] += spaceLow.getCoord( d ) - srcBnds.getLower( d );
            dstUp [ d ] = dstLow[ d ] + spaceUp.getCoord( d ) - spaceLow.getCoord( d );
         }
                  
         ISBounds adjDstBnds = new ISBounds( dstLow, dstUp );
         
         if ( rs == null && fm == null )
            partDB.subblock( dstBlk, adjDstBnds, partSrc );
         else if ( rs == null && fm != null )
            partDB.subblock( dstBlk, adjDstBnds, partSrc, fm );
         else if ( rs != null && fm == null )
            partDB.subblock( dstBlk, adjDstBnds, partSrc, rs );
         else
            new SDBError( SDBError.Fatal, "SparseDataSource.subblock internal "
                         + "error; both RecordSpec and FieldIDMapper given. " );
      }                             
   }
   /**
    * If the desired block is in the cache, return it.
    * else if there is room in the cache to add a new block, add it.
    * else remove the least recently referenced
    * @param blockIndex
    * @return 
    */
   private DataBlock getCacheBlock( int blockIndex )
   {
      DBCache dbc = dbCache.get( blockIndex );
      if ( dbc == null )
         dbc = readCacheBlock( blockIndex );
      else
         dbc.refTime = cacheRefTime++;
      return dbc.db;
   }
   
   /**
    * Create a new CacheBlock entry and put it in the cache hash. If there
    *    is no room, remove the LRU cache entry.
    * Read the real data into the cache DataBlock
    * 
    * @param blockIndex
    * @return 
    */
   private DBCache readCacheBlock( int blockIndex )
   {
      DBCache cacheEntry;
      if ( dbCache.size() < maxCacheSize )
      {
         cacheEntry = new DBCache();
         cacheEntry.db = this.createDataBlock( this.blockBounds );
      }
      else  // need to replace least recently referenced entry
      {
         cacheEntry = removeLRUentry();
      }
      dbCache.put( blockIndex, cacheEntry );
      cacheEntry.refTime = cacheRefTime++;
      cacheEntry.partitionNum = blockIndex;
      
      readPartition( cacheEntry.db, blockIndex );
      return cacheEntry;
   }
   
   /**
    * find the LRU cache entry and remove it from the cache hash and return
    * the DBCache entry.
    */
   private DBCache removeLRUentry()
   {
      long oldestRef = cacheRefTime;
      int keyToGo = -1;
      DBCache entry = null;
      
      for ( int key: dbCache.keySet() )
      {
         entry = dbCache.get( key );
         if ( entry.refTime < oldestRef )
         {
            keyToGo = entry.partitionNum;
            oldestRef = entry.refTime;
         }
      }
      if ( keyToGo >= cacheRefTime )
         new SDBError( SDBError.Fatal, "SparseDataSource.removeLRUentry: System failure.");
      entry = dbCache.remove( keyToGo );
      return entry;
   }
   
   /**
    * read the specified partition into the provided DataBlock, Not sure
    * we really need it; the hard work is donein loadPartition.
    * 
    * @param db
    * @param partitionIndex 
    */
   private void readPartition( DataBlock db, int partitionIndex )
   {
      Values values = db.getValues();
      loadPartition( partitionIndex, values );      
   }
   /**
    * Determine the partition (block) containing the specified space location.
    * If the block does not exist, print an error message (maybe should throw exc)
    * else read the block from the file into a cache.
    * 
    * @param spaceLoc
    * @return 
    */
   private DataBlock getDataBlock( IndexSpaceID spaceLoc )
   {
//	  if ( this.fdlDescriptor.isSparse() )
//		  System.out.println( "SparseDS has mapinfo!!" );
	  // boolean isSparse = this.getFDLDescriptor().isSparse();
	  // read blockIndex = 4 which does not exist
      int blockIndex = blockPartitioning.getPartitionIndex( spaceLoc ); 
     
      int mapSparseBlockIndex = -1;
      mapSparseBlockIndex = this.mappedBlockIndex( blockIndex );
      if ( mapSparseBlockIndex == -1 ) // didn't find block index in sparse file
    	  return null;
      
      blockIndex = mapSparseBlockIndex; // found, use sparse file block index
      
      DataBlock db = getCacheBlock( blockIndex );  
      readPartition( db, blockIndex );  // if not found block, throw SDBError here.
      return db;
   }
   /**
    * Get the mapped origin block index from sparse
    * 
    * @param spaceLoc
    * @return 
    */
	public int mappedBlockIndex(int originBlockIndex) 
	{
		int ret = -1;
		if (this.mapInfo != null) {
			for (int m = 0; m < mapInfo.length; m++) {
				if (originBlockIndex == mapInfo[m].sparseBlkId) {
					System.out.println("sparsed:" + mapInfo[m].origBlkId
							+ "--->origin:" + mapInfo[m].sparseBlkId);
					ret = mapInfo[m].origBlkId;
				}
			}
		}
		return ret;
	}
   
   /** 
    * Return a Values array of all attributes in whatever type and order
    *  they are stored internally (either attribute order or point order 
    *  for Granite-defined objects). The default implementation is often
    *  overridden for efficiency.
    */
   @Override
   public Values getValues()
   {
      System.err.println( "SparseDataBlock.getValues()" + " not implemented." );
      return null;
   }
   
   /***************************************/
   /**
    * copyFields: copy all the relevant DataSource fields from the parent DS
    * 
    */
   private void copyFields( DataSource base )
   {
      this.axisOrdering = base.getAxisOrdering().copy();
      this.axisWeights  = base.getAxisWeights().clone();
      this.dataOrder    = base.getDataOrder();
      this.fdlDescriptor = (FDLDescriptor) base.getFDLDescriptor().clone();
      this.recordDesc   = base.getRecordDescriptor().copy();      
   }
   //------------------ parseMapString( MapInfo, String ) -------------------
   /**
    * parseMapString: convert string rep of map information to fill a MapInfo object
    * 
    * String format:
    *   k@(s,r,c)=(s0,s1,s2)->n
    * 
    * where n is the block index in the sparse data set
    *       (s,r,c) (lower coord) is the "origin" of the block in original space
    *       (s0,s1,s2) sizes of block (this might become optional, so make it so
    *       n is the block index in the original space
    */
   private static void parseMapString( MapInfo mi, String mapString )
   {
      int at = mapString.indexOf( "@" );
      mi.sparseBlkId = strToInt( mapString.substring( 0, at ));
      
      int lparen = mapString.indexOf( "(" );
      int rparen = mapString.indexOf( ")" );
      mi.lower   = strToIntArray( mapString.substring( lparen+1, rparen ));
      
      int eq = mapString.indexOf( "=" );     
      if ( eq >= 0 )
      {
         lparen = mapString.indexOf( "(", eq ); // get next parens
         rparen = mapString.indexOf( ")", eq );
         mi.size = strToIntArray( mapString.substring( lparen+1, rparen ));
      }
      
      int arrow = mapString.indexOf( "->" );
      mi.origBlkId = strToInt( mapString.substring( arrow + 2 ));
   }
   //+++++++++++ utility string methods (put into a Utilities class? ++++++++
   //--------------- strToInt utility --------------------------
   private static int strToInt( String istr )
   {
      istr = istr.trim();
      
      //int lastBlank = istr.lastIndexOf( " " );
      //if ( lastBlank >= 0)
         //istr = istr.substring( lastBlank + 1 );
      try {
         return Integer.parseInt( istr );
      }
      catch (NumberFormatException nfe ) {
         System.err.println( "Internal error: SparseDataSource.strToInt. "
                             + nfe.getMessage() );
      }
      return -1;
   }
   //---------------- strToIntArray utility -------------------------
   private static int[] strToIntArray( String istr )
   {
      String[] values = istr.split( "," );
      int[] ivals = new int[ values.length ];
      
      for ( int i= 0; i < values.length; i++ )
      {
         ivals[ i ] = strToInt( values[ i ] );
      }
      return ivals;
   }
   //---------------- arrayToString utility --------------------------
   private static String arrayToString( int[] arr )
   {
      if ( arr == null )
         return "";
      StringBuilder sb = new StringBuilder();
      sb.append( "(" );
      for ( int i = 0; i < arr.length; i++ )
      {
         sb.append( arr[i] );
         sb.append( "," );
      }
      sb.setCharAt( sb.length() - 1, ')' );
      return sb.toString();
   }  
   /**
    * Given a partition index, read the data from the partition into a 
    * Values object. 
    * 
    * This code has been copied from ChunkedDataSource.loadChunk
    * Since Partitions are likely to be much larger than chunks often are,
    * I've used int instead of long for the partition index value.
    * 
    * Also, I took out the loop that was in loadChunk just in case the buffer
    * was not big enough -- since we are always reading exactly one partition,
    * the buffer size should always be ok.
    */
   protected void loadPartition( int partitionIndex, Values data )
   {
      long fileOffset = this.partitionByteSize * partitionIndex;
      
      ByteBuffer byteBuffer = getBuffer();
      int bytesInBuffer = byteBuffer.limit();
      
      int bytesToRead = this.partitionByteSize;
      if (bytesToRead > data.size() * StorageType.byteSize(data.getStorageType()))
      {
         bytesToRead = data.size() * StorageType.byteSize(data.getStorageType());
         if ( bytesToRead < bytesInBuffer )
            bytesInBuffer = bytesToRead;
         new SDBError( SDBError.Warning,
               "SparseDataSource.loadpartition warning: array smaller than partition size." );
      }
      else if (bytesToRead < data.size()* StorageType.byteSize(data.getStorageType()))
      {
         new SDBError( SDBError.Recoverable,
               "SparseDataSource.loadpartition error: array larger than partition size." );
      }
      
      try
      {
         if ( fileOffset !=((int) raFile.getFilePointer()) )
            raFile.seek( fileOffset );
         //int numIndex = 0;
         if( bytesToRead == 0 )   //rdb, a while in loadChunk
         {
            new SDBError( SDBError.Warning, "SparseDataSource.loadPartition: read count is 0");
         }
         else
         {
            byteBuffer.clear();
            int bytesRead = fChannel.read( getBuffer() );
            if ( bytesRead <= 0 ) // rdb added to avoid compiler warning, but should check
               new SDBError( SDBError.Fatal, "SparseDataSource.loadpartition: "
                     + "no bytes read from channel???" );
            else if ( bytesRead < bytesToRead ) // didn't get everything???
               new SDBError( SDBError.Warning, "SparseDataSource.loadpartition: "
                     + "too few bytes read from channel: " + bytesRead 
                     + " rather than " + bytesToRead );
            
            byteBuffer.flip();
            //if ( bytesInBuffer > bytesToRead ) // for last read, might not be full
               //bytesInBuffer = bytesToRead;
            
            
            int sType = data.getStorageType();
            switch( sType )
            {
               case StorageType.Byte:
                  byteBuffer.get(data.getBytes() );
                  break;
               case StorageType.Float:
                  //byteBuffer.asFloatBuffer().get( fData, numIndex, bytesRead/4 );
                  byteBuffer.asFloatBuffer().get( data.getFloats() );
                  break;
               case StorageType.Double:
                  byteBuffer.asDoubleBuffer().get( data.getDoubles() );
                  break;
               case StorageType.Int:
                  byteBuffer.asIntBuffer().get( data.getInts() );
                  break;
               case StorageType.Short:
                  byteBuffer.asShortBuffer().get( data.getShorts() );
                  break;
               default:
                  new SDBError( SDBError.Fatal, "Type not implemented:" + sType );
            }
         }
      }
      catch( IOException e )
      {
         new SDBError(SDBError.Recoverable, e.toString());
      }
   }
   
   //----------------- main tests parsing/printing for mapping info
   public static void main( String[] args )
   {
      // testMapString();
   }
   private static void testMapString()
   {
      String[] maps = { "2@(0,0,0)=(4,4,4)->9",
                        "23@(100 , 100, 50 ) = ( 20, 5 , 2)   ->  22",
                        "1456 @ (99,89,4332 )  ->  5666"
                      };
      for ( int i = 0; i < maps.length; i++ )
      {
         SparseDataSource.MapInfo mi = new SparseDataSource.MapInfo( maps[i] );
         System.out.println( "---->" + maps[i] );
         System.out.println( mi );
      }
      
   }
   
}
