/*
 * SparseDataSource -- partitions a data source; some partitions are allowed
 *                     to be "missing". In principle, all points in a missing
 *                     partition receive a default value (0). In practice
 *                     we expect it to be used for AR representations.
 * 
 * Pending issues:
 *    byteOrder (may work)
 *    dataOrder -- only supports point order
 *    subCompositeBlock not implemented
 * 
 * Thought of storing the sparse data in a 1D ChunkedDataSource; problem is
 * that the chunks have to turn into an array of Datums which really should
 * then be organized in 3D (e.g.) I don't think that will gain us much over 
 * just using a file. 
 * 
 * On the other hand, 
 * 1. how do we deal with the attribute order vs point order datasources
 * 2. The existing approach should be able to handle  CompositeDataSources -- 
 *    by converting them to single-file datasources. Is that acceptable in the 
 *    long run?
 * 3. Could handle all these (though not necessarily uniformly) by requiring
 *    the original ds to be accessible and always accessing the data from
 *    the original -- however, I think we'll lose performance and distribution
 *    efficiency.
 * 
 * 
 * Created
 *    12/22/11 to 03/05/12 rdb Initial implementation.  
 * 
 * History
 *    03/05/12 Seems to work when everything is full blocks and when all
 *             blocks are present -- this is essentially equiv to a ChunkedDS
 *    02/22/14 Can changed private DataBlock getDataBlock( IndexSpaceID spaceLoc )
 *    		   by using mapInfo to find mapped data block id from sparse to origin
 *    		   added "public int mappedBlockIndex(int originBlockIndex)" 
 */
 
package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

import java.util.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;

/**
 * Created between 12/22/11 and 03/05/12
 * 
 * edited by AbinayaPitchai
 *  
 */
// BinaryFileDataSource creates and opens the physical file for us
public class SparseStorage
{
   //------------------ class variables ---------------------------------
   
   //------------------ instance variables ------------------------------
   protected ISBounds blockBounds = null;
   protected HashMap<Integer,Integer> blockMapper = null; 
   
   protected DataSource origDS = null;  // Not always known or accessible
   protected String[]   mappingStrings = null;
   protected SparseBlockInfo sparseBlockInfo;
   protected SparseDataTable sparseTable;
   
   
   //----- variables for partition management
   protected ISBounds              partitioningBounds;
   protected RegularISPartitioning blockPartitioning;
   protected IndexSpaceID          blockSizes;  // not an ID, but size specification
   protected int                   partitionByteSize;  
   
   //------ variables for the cache
   protected HashMap<Integer, DBCache>  dbCache = null;   
   protected int     maxCacheSize = 0; 
   protected int     maxCacheSizeDefault = 64;
   protected long    cacheRefTime = 0; // used to order cache references
   
  
   //------------------- internal DBCache class --------------------------
   private static class DBCache
   {
      int       partitionNum; // 1D index of the partition
      DataBlock db;
      long      refTime; // position of DataBlock in reference history array
   }
   
   //------------------ constructors ------------------------------------
   /*
    * Maybe we should have other constructors allowing user to dynamically
    * create a SparseDataSource. For now, we'll settle for building a 
    * SparseDataSource from an fdl file.
    * 
    * There is a utility class, SparseDataTestGen, that generates the xfdl
    * files needed for a SparseDataSource by picking random blocks. 
    * Eventually, we'll need to morph that into a more general purpose
    * utility that will do the same based on an AR fdl file and an error 
    * tolerance -- or integrate it into the AR generation code.
    * 
    * If an app wants to create its own SparseDataSource, for now you would
    * have create an FDLDescriptor and pass it to the constructor that 
    * accepts an FDLDescriptor, which will require the same information 
    * that we would need to ask for in a SparseDataSource constructor.
    */
   /**
    * This is the mechanism for creating a sparse ds from an xfdl file, but
    * it is not normally used by an app, which should use Datasource.create, 
    * which calls this method. This is public unlike most equivalent 
    * Granite DataSource constructors -- this is probably a better
    * model to provide an app with the opportunity to create a specific
    * kind of DataSource dynamically -- compared to the complex and inconsistent
    * alternative public constructors that we provide elsewhere.
    * 
    * @param name - informal name for the ds
    * @param fdl  - fdl describing characteristics -- usually extracted from
    *               xfdl file.
    */
   
   final   DataSource dataSource; 
   public SparseStorage( String name, String fdlName )
   {
	  dataSource = DataSource.create(name, fdlName);
      initialize();
   }
   /**
    * Define the number of partitions to be kept in the cache memory.
    * 
    * @param nPartitions 
    */
   public void setCacheSize( int nPartitions )
   {
      if ( maxCacheSize == 0 || maxCacheSize > nPartitions )
      {
         // empty the cache
         //-- initialize hash map for cache
         dbCache = new HashMap<Integer, DBCache>();
         cacheRefTime = 0;
      }
      maxCacheSize = nPartitions;
   }
   //--------------------- initialize -------------------------------------
   protected final void initialize()
   {
      //-- set up block and partitioning variables
      blockSizes = dataSource.getFDLDescriptor().getBlockDimensions();
      if ( blockSizes == null )
      {
         System.err.println("SparseDataSource error: blocksizes null. " +
                       "Using test default: 4x4x4" );
         blockSizes = new IntegerIndexSpaceID( 4,4,4 );
      }

      setCacheSize( maxCacheSizeDefault );
      
      blockBounds = new ISBounds( blockSizes );
      blockPartitioning = new RegularISPartitioning( 
                             new IntegerIndexSpaceID( blockSizes ), 
                             dataSource.getBounds(),
                             dataSource.getFDLDescriptor().getAxisOrdering() );
      partitioningBounds = blockPartitioning.getPartitioningBounds();
      
      //-- set up mapping information
      mappingStrings = dataSource.getFDLDescriptor().getSparseBlockMap();
      addBlocksInfoToTable( mappingStrings );
      
      //-- invoke parent code to allocate I/O buffer. For the time being,
      //   we will only worry about reading one partition at a time.
      partitionByteSize = blockBounds.volume() * dataSource.getRecordDescriptor().byteSize();
//      if ( !allocateBuffer( partitionByteSize ))
//         new SDBError( SDBError.Fatal, "SparseDataSource can't allocate io buffer "
//                             + partitionByteSize );
   }
   //--------------------- setBlockBounds( ISBounds ) ----------------------
   /**
    * set/reset the size of the blocks; only do this if the physical file
    * changes correspondingly -- not a public method.
    */
   protected void setBlockBounds( ISBounds blkBounds )
   {
      blockBounds = blkBounds.copy();
   }
   //++++++++++++++++++++++ Hash table methods ++++++++++++++++++++
   /**
    * generate the block mapping hash table
    * 
    * @author AbinayaPitchai
    */
   protected void addBlocksInfoToTable( String[] mapStrings )
   {
	   sparseTable = new SparseDataTable();
	   for(int m = 0; m < mapStrings.length; m++){
    	  sparseBlockInfo = new SparseBlockInfo(mapStrings[m]);
    	  sparseTable.addToTable(sparseBlockInfo.getOriginalBlockId(), sparseBlockInfo);
      }
      
   }
   
   //------------------- getBlock( IndexSpaceID ) -------------------------
   /**
    * Return the block index (id) for the partition that contains the "loc" 
    *    IndexSpaceID.
    * 
    * @param loc
    * @return 
    */
   public int getBlock( IndexSpaceID loc )
   {
      System.err.println( "SparseDataBlock.getBlock( ISID ) not implemented!");
      return -1;
   }
   //---------------- getParts( ISBounds ) ---------------------------
   /**
    * returns an array of ISBounds representing the pieces of the input bounds
    * that have data in this sparse data source.
    */
   public Map<Integer, ISBounds> getParts( ISBounds bnds )
   {
      ISBounds[] parts = blockPartitioning.mapToSpaceParts( bnds );
    //  ArrayList<ISBounds> goodBlocks = new ArrayList<ISBounds>();
      Map<Integer, ISBounds> partsMap = new HashMap<Integer, ISBounds>() ; 
      
      int missingParts = 0;
      for ( ISBounds part: parts )
      {
         //long blockIndex = blockPartitioning.partitionOffset( part.getLower() );
    	  
         //Integer fileIndex = blockMapper.get( (int) blockIndex ); -- entering sparse block Id and getting original block id.
    	  long blockIndex = blockPartitioning.getPartitionIndex(part.getLower());
         SparseBlockInfo blockInfo = sparseTable.getSparseBlockInfo((int) blockIndex);
        // Integer fileIndex = blockInfo.getOriginalBlockId(); 
         if (blockInfo != null ){
          //  goodBlocks.add( part );
         	partsMap.put( blockInfo.getOriginalBlockId(), part) ; 
         } else{
            missingParts++;
         }
      }
      System.err.println( "SparseDS.getParts missing parts: " + missingParts );
     
      return partsMap ;
      
   }
   
   //---------------- getParts( ISBounds ) ---------------------------
   /**
    * returns an array of ISBounds representing the pieces of the input bounds
    * that have data in this sparse data source.
    */
   public ISBounds getPart( IndexSpaceID loc )
   {
      ISBounds part = blockPartitioning.getContainingPartitionBounds( loc );
      //long blockIndex = blockPartitioning.partitionOffset( part.getLower() );
      long blockIndex = blockPartitioning.getPartitionIndex(part.getLower());
      //Integer fileIndex = blockMapper.get( (int) blockIndex );
      Integer fileIndex = sparseTable.getSparseBlockInfo((int) blockIndex).getOriginalBlockId();
      if ( fileIndex != null ){
    	  // Given index space resides sparse region, so a sparse block index is returned.
         return part;
      }else{
    	 // Given index space doesn't exist in the sparse region.
         return null;
      }
   }
   //----------- abstract DataCollection datum methods to be overridden -------
   
   /** Given an IndexSpaceID and FieldIDMapper, copy the values of the Datum at 
    *  that position to the argument Datum. 
    * TODO Define error conditions
    */
   public void datum( Datum dtm, IndexSpaceID did, FieldIDMapper mp )
   {
      RegularPartitionISID rpISID;
      DataBlock part = getDataBlock( did );
      if ( part == null )
      { 
    	  dtm = null ; 
         // this is debugging code;  final code should probably throw exception
         System.err.println( "SparseDataSource.datum( DataBlock, ISID )" +
            " isid not in data source: " + did ); 
      }
      else
      {
         rpISID = new RegularPartitionISID( blockPartitioning, did );
         part.datum( dtm, rpISID.offsetISID, mp );
      }
   }
   
   /** 
    * Given an IndexSpaceID and FieldIDMapper, copy the values of the Datum at 
    *  that position to the argument Datum. 
    * Don't need to explicitly call super, but I'm doing that because I
    * think we'll want to have our own code here.
    */
   
   public void datum( Datum dtm, IndexSpaceID did )
   {  
      RegularPartitionISID rpISID;
      DataBlock part = getDataBlock( did );
      //System.out.println("get part in datum method!!"); // not getting here
      if ( part == null )
      { 
    	  dtm = null ; 
         // this is debugging code;  final code should probably throw exception
         System.err.println( "SparseDataSource.datum( DataBlock, ISID )" +
            " isid not in data source: " + did ); 
      }
      else
      {
         rpISID = new RegularPartitionISID( blockPartitioning, did );
         part.datum( dtm, rpISID.offsetISID );
      }
   }
//   //+++++++++++++ override abstract subBasicBlock methods +++++++++++++++
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      subParts( dstBlk, dstBnds, srcBnds, null, null );
   }
   
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                       RecordSpec recSpec )
   {
      subParts( dstBlk, dstBnds, srcBnds, recSpec, null );
   }
   
   void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                       FieldIDMapper mp )
   {
      subParts( dstBlk, dstBnds, srcBnds, null, mp );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
   {
      new SDBError( SDBError.Warning, "RecordSpec subblock not implemented." );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         RecordSpec recSpec )
   {
      new SDBError( SDBError.Warning, "RecordSpec subblock not implemented." );
   }
   
   void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
         FieldIDMapper mp )
   {
      new SDBError( SDBError.Warning, "FieldIDMapper subblock not implemented." );
   }
   
   //----------------------- SparseDataBlock utility methods ----------------
   /**
    * Read all parts of a subblock from the sparse ds
    * 
    */
   private void subParts( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                      RecordSpec rs, FieldIDMapper fm )
   {
      ISBounds[] parts = blockPartitioning.mapToSpaceParts( srcBnds );
      for ( ISBounds part: parts )
      {
         subPart( dstBlk, dstBnds, srcBnds, part, rs, fm );
      }
   }
   /**
    * Read a subblock portion of a partition needed for this subblock request
    * 
    * This version does NOT correctly handle edge partitions that are not
    * full size.
    */
   private void subPart( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                       ISBounds partBnds, RecordSpec rs, FieldIDMapper fm )
   {
      RegularPartitionISID rpISIDlow, rpISIDup;
      IndexSpaceID spaceLow = partBnds.getLower();
      IndexSpaceID spaceUp  = partBnds.getUpper();
      
      int[]  dstLow = dstBnds.copyLowerArray();
      int[]  dstUp  = dstBnds.copyUpperArray();
      
      IndexSpaceID dstLowISID = dstBnds.copyLower();
      IndexSpaceID dstUpISID  = dstBnds.copyUpper();
      
      DataBlock partDB = getDataBlock( spaceLow );
      
      if ( partDB == null )
      { 
    	  // set by position, iterator, index based ID, NOT need to create datum, just set a array 
    	  int size = partDB.numAttributes;
    	  //for ()
    	    //float[] array = { Float.NaN, Float.NaN, Float.NaN, Float.NaN }; 
    	  
    	  // setFloats, iterate over the partDB, 
    	  // for index base id
    	  
         // this is debugging code;  final code should probably throw exception
         System.err.println( "SparseDataSource.subblock ISBounds " +
            " is not in data source: " + partBnds + "\nvolume/datum of space: " + dstBlk.volume() ); //TODO
      }
      else
      {
         rpISIDlow = new RegularPartitionISID( blockPartitioning, spaceLow );
         //System.err.println( rpISIDlow );
         rpISIDup = new RegularPartitionISID( blockPartitioning, spaceUp );
         ISBounds partSrc = new ISBounds( rpISIDlow.getOffsetISID(),
                                          rpISIDup.getOffsetISID() );
         
         for ( int d = 0; d < dstBlk.dim(); d++ )
         {
            dstLow[ d ] += spaceLow.getCoord( d ) - srcBnds.getLower( d );
            dstUp [ d ] = dstLow[ d ] + spaceUp.getCoord( d ) - spaceLow.getCoord( d );
         }
                  
         ISBounds adjDstBnds = new ISBounds( dstLow, dstUp );
         
         if ( rs == null && fm == null )
            partDB.subblock( dstBlk, adjDstBnds, partSrc );
         else if ( rs == null && fm != null )
            partDB.subblock( dstBlk, adjDstBnds, partSrc, fm );
         else if ( rs != null && fm == null )
            partDB.subblock( dstBlk, adjDstBnds, partSrc, rs );
         else
            new SDBError( SDBError.Fatal, "SparseDataSource.subblock internal "
                         + "error; both RecordSpec and FieldIDMapper given. " );
      }                             
   }
   /**
    * If the desired block is in the cache, return it.
    * else if there is room in the cache to add a new block, add it.
    * else remove the least recently referenced
    * @param blockIndex
    * @return 
    */
   private DataBlock getCacheBlock( int blockIndex )
   {
      DBCache dbc = dbCache.get( blockIndex );
      if ( dbc == null )
         dbc = readCacheBlock( blockIndex );
      else
         dbc.refTime = cacheRefTime++;
      return dbc.db;
   }
   
   /**
    * Create a new CacheBlock entry and put it in the cache hash. If there
    *    is no room, remove the LRU cache entry.
    * Read the real data into the cache DataBlock
    * 
    * @param blockIndex
    * @return 
    */
   private DBCache readCacheBlock( int blockIndex )
   {
      DBCache cacheEntry;
      if ( dbCache.size() < maxCacheSize )
      {
         cacheEntry = new DBCache();
         cacheEntry.db = dataSource.createDataBlock( this.blockBounds );
      }
      else  // need to replace least recently referenced entry
      {
         cacheEntry = removeLRUentry();
      }
      dbCache.put( blockIndex, cacheEntry );
      cacheEntry.refTime = cacheRefTime++;
      cacheEntry.partitionNum = blockIndex;
      
     // readPartition( cacheEntry.db, blockIndex );
      return cacheEntry;
   }
   
   /**
    * find the LRU cache entry and remove it from the cache hash and return
    * the DBCache entry.
    */
   private DBCache removeLRUentry()
   {
      long oldestRef = cacheRefTime;
      int keyToGo = -1;
      DBCache entry = null;
      
      for ( int key: dbCache.keySet() )
      {
         entry = dbCache.get( key );
         if ( entry.refTime < oldestRef )
         {
            keyToGo = entry.partitionNum;
            oldestRef = entry.refTime;
         }
      }
      if ( keyToGo >= cacheRefTime )
         new SDBError( SDBError.Fatal, "SparseDataSource.removeLRUentry: System failure.");
      entry = dbCache.remove( keyToGo );
      return entry;
   }
   
   /**
    * read the specified partition into the provided DataBlock, Not sure
    * we really need it; the hard work is done in loadPartition.
    * 
    * @param db
    * @param partitionIndex 
    */
   private DataBlock readPartition( SparseBlockInfo sparseBlockInfo)
   {
	
	
//	Integer[] sparseLower = sparseBlockInfo.getStartPosition(); 
//	Integer[] sparseSize = sparseBlockInfo.getSize();
//	int[] sparseUpper = new int[sparseLower.length];
//	int[] sparseLower2 = new int[sparseLower.length];
//	int[] size = new int[sparseSize.length];
//	for(int m = 0 ; m < sparseLower.length ; m++){
//		sparseUpper[m] = sparseLower[m] + sparseSize[m]  - 1 ;
//		sparseLower2[m] = sparseLower[m];
//		size[m] = sparseSize[m];
//		
//	}
//	
//	ISBounds spBounds = new ISBounds(new IntegerIndexSpaceID(size[0], size[1], size[2]));
//	DataBlock db = dataSource.subblock(spBounds);
//	dataSource.subblock(db, new ISBounds(sparseLower2, sparseUpper));
	
	 DataBlock dbBlock = null ;
	 RandomAccessFile sparse_raf;
	 Integer[] blockSize = sparseBlockInfo.getSize(); 
	 int read ; 
	 try {
		sparse_raf = new RandomAccessFile(dataSource.getFDLDescriptor().getDataFileURL(), "rw");
		Values dsVals = dataSource.getValues();
		long pos = sparseBlockInfo.getSeekAddress() ; 
		int byteVal = getByteValue(dsVals);
		int totalsize = getTotalByteSizeOfBlock(blockSize, dsVals); // 1 datum - 4 float vals, 1 float = 4 byte, total 6 datums so 16 x 6 
		float[] floatVals = new float[ byteVal * getBlockValuesCount(blockSize)];
		int add = 0 ;
		for(int i = 0; i < totalsize ; i= i + byteVal){
			byte[] buffer = new byte[byteVal];
			sparse_raf.seek(pos);
			pos = pos + byteVal ; 
			read = sparse_raf.read(buffer);
			ByteBuffer byteBuffer = ByteBuffer.wrap(buffer);
			floatVals[add] = byteBuffer.getFloat() ; // currently having as Float type, need to change it according to the data Type.
			add++;
	  }
	  
	  dbBlock = dataSource.createDataBlock(new ISBounds(convertIntegerArrayToIntArray(blockSize)));
	  ISIterator iter2 = new ISIterator( dbBlock.getBounds() );
	  int start = 0 ;
	  for( ; iter2.valid(); iter2.next()){
		  Datum dat = dataSource.createDatum() ; 
		  dat.setFloats(floatVals, start , dbBlock.getAttributeNames().length);
		  dbBlock.setDatum(iter2, dat);
		  start = start + dbBlock.getAttributeNames().length ; 
	  }
	 
	
	} catch (FileNotFoundException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	} catch (IOException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
	
	 return dbBlock;
	
	
   }
   
   public DataBlock readPartition(Integer sparseBlockIndex){
	   return readPartition(sparseTable.getSparseBlockTable().get(sparseBlockIndex)) ; 
	   
   }
   /**
    * Determine the partition (block) containing the specified space location.
    * If the block does not exist, print an error message (maybe should throw exc)
    * else read the block from the file into a cache.
    * 
    * @param spaceLoc
    * @return 
    */
   private DataBlock getDataBlock( IndexSpaceID spaceLoc )
   {
	 
      int blockIndex = blockPartitioning.getPartitionIndex( spaceLoc ); 
      SparseBlockInfo blockInfo = this.mappedSparseBlockInfo(blockIndex);
      DataBlock db = null; 
      if(blockInfo != null){
       db = readPartition(blockInfo)  ;
      }// if not found block, throw SDBError here.
      return db;
   }
   /**
    * Get the mapped origin block index from sparse
    * 
    * @param spaceLoc
    * @return 
    */
   public int mappedBlockIndex(int originBlockIndex){
	   SparseBlockInfo blockInfo =  mappedSparseBlockInfo(originBlockIndex) ;
	   int sparseBlockIndex = -1 ; 
	   if(blockInfo!= null){
		   sparseBlockIndex = blockInfo.getSparseBlockId();
	   }
	   
	   return sparseBlockIndex;
   }
   
   
	public SparseBlockInfo mappedSparseBlockInfo(int originBlockIndex) 
	{

		SparseBlockInfo ret = null;
		if(this.sparseTable != null){
			SparseBlockInfo block = sparseTable.getSparseBlockInfo(originBlockIndex);
			if(block!= null){
				System.out.println("sparsed:" + block.getSparseBlockId()
						+ "--->origin:" + block.getOriginalBlockId());
				ret = block;
			}
		}
		return ret;
	}
   
   /** 
    * Return a Values array of all attributes in whatever type and order
    *  they are stored internally (either attribute order or point order 
    *  for Granite-defined objects). The default implementation is often
    *  overridden for efficiency.
    */
   public Values getValues()
   {
      System.err.println( "SparseDataBlock.getValues()" + " not implemented." );
      return null;
   }
   
   /***************************************/
   /**
    * copyFields: copy all the relevant DataSource fields from the parent DS
    * 
    */
   private void copyFields( DataSource base )
   {
      dataSource.axisOrdering = base.getAxisOrdering().copy();
      dataSource.axisWeights  = base.getAxisWeights().clone();
      dataSource.dataOrder    = base.getDataOrder();
      dataSource.fdlDescriptor = (FDLDescriptor) base.getFDLDescriptor().clone();
      dataSource.recordDesc   = base.getRecordDescriptor().copy();      
   }
   /**
    * Given a partition index, read the data from the partition into a 
    * Values object. 
    * 
    * This code has been copied from ChunkedDataSource.loadChunk
    * Since Partitions are likely to be much larger than chunks often are,
    * I've used int instead of long for the partition index value.
    * 
    * Also, I took out the loop that was in loadChunk just in case the buffer
    * was not big enough -- since we are always reading exactly one partition,
    * the buffer size should always be ok.
    */
   protected void loadPartition( int partitionIndex, Values data )
   {
//      long fileOffset = this.partitionByteSize * partitionIndex;
//      
//      ByteBuffer byteBuffer = getBuffer();
//      int bytesInBuffer = byteBuffer.limit();
//      
//      int bytesToRead = this.partitionByteSize;
//      if (bytesToRead > data.size() * StorageType.byteSize(data.getStorageType()))
//      {
//         bytesToRead = data.size() * StorageType.byteSize(data.getStorageType());
//         if ( bytesToRead < bytesInBuffer )
//            bytesInBuffer = bytesToRead;
//         new SDBError( SDBError.Warning,
//               "SparseDataSource.loadpartition warning: array smaller than partition size." );
//      }
//      else if (bytesToRead < data.size()* StorageType.byteSize(data.getStorageType()))
//      {
//         new SDBError( SDBError.Recoverable,
//               "SparseDataSource.loadpartition error: array larger than partition size." );
//      }
//      
//      try
//      {
//         if ( fileOffset !=((int)raFile.getFilePointer()) )
//            raFile.seek( fileOffset );
//         //int numIndex = 0;
//         if( bytesToRead == 0 )   //rdb, a while in loadChunk
//         {
//            new SDBError( SDBError.Warning, "SparseDataSource.loadPartition: read count is 0");
//         }
//         else
//         {
//            byteBuffer.clear();
//            int bytesRead = fChannel.read( getBuffer() );
//            if ( bytesRead <= 0 ) // rdb added to avoid compiler warning, but should check
//               new SDBError( SDBError.Fatal, "SparseDataSource.loadpartition: "
//                     + "no bytes read from channel???" );
//            else if ( bytesRead < bytesToRead ) // didn't get everything???
//               new SDBError( SDBError.Warning, "SparseDataSource.loadpartition: "
//                     + "too few bytes read from channel: " + bytesRead 
//                     + " rather than " + bytesToRead );
//            
//            byteBuffer.flip();
//            //if ( bytesInBuffer > bytesToRead ) // for last read, might not be full
//               //bytesInBuffer = bytesToRead;
//            
//            
//            int sType = data.getStorageType();
//            switch( sType )
//            {
//               case StorageType.Byte:
//                  byteBuffer.get(data.getBytes() );
//                  break;
//               case StorageType.Float:
//                  //byteBuffer.asFloatBuffer().get( fData, numIndex, bytesRead/4 );
//                  byteBuffer.asFloatBuffer().get( data.getFloats() );
//                  break;
//               case StorageType.Double:
//                  byteBuffer.asDoubleBuffer().get( data.getDoubles() );
//                  break;
//               case StorageType.Int:
//                  byteBuffer.asIntBuffer().get( data.getInts() );
//                  break;
//               case StorageType.Short:
//                  byteBuffer.asShortBuffer().get( data.getShorts() );
//                  break;
//               default:
//                  new SDBError( SDBError.Fatal, "Type not implemented:" + sType );
//            }
//         }
//      }
//      catch( IOException e )
//      {
//         new SDBError(SDBError.Recoverable, e.toString());
//      }
   }
   
   public Hashtable<Integer, SparseBlockInfo> getSparseDataTable(){
	   return sparseTable.getSparseBlockTable();
   }
   
   
   public DataSource getSparseDS(){
	   return dataSource;
   }
   
  private int getTotalByteSizeOfBlock(Integer[] blockSize, Values vals ){
	  
	   int byteVal = getByteValue(vals) ; 
	   int noOfAttr = dataSource.getNumAttributes(); 
	   return ( byteVal * noOfAttr * getBlockValuesCount(blockSize));
}
   
   private int getByteValue(Values vals ) {
	  int retByte = 1 ;  // Mixed or Byte Type Values
	  if(vals instanceof FloatValues || vals instanceof IntValues || vals instanceof ShortValues){
		  retByte = 4  ;
	  }else if (vals instanceof DoubleValues){
	   retByte = 8 ;
   }
	  return retByte;
	   
   }
   
   private int getBlockValuesCount(Integer[] blockSize){
	   int blockValsCount = 1 ; 
	   for(int i = 0 ; i < blockSize.length ; i ++){
		   blockValsCount = blockValsCount * blockSize[i];
	   }
	   return blockValsCount;
   }
   
   private int[] convertIntegerArrayToIntArray(Integer[]blockSize){
	  
	   int[] intArr = new int[blockSize.length];
	   for(int i = 0 ; i < blockSize.length ; i ++){
		  intArr[i] = blockSize[i];
	   }
	   return intArr;
   }
 //----------------- main tests parsing/printing for mapping info
   public static void main( String[] args ){
      testMapString();
	}
   
   
   private static void testMapString(){

      String[] maps = { "2@(0,0,0)=(4,4,4)->9",
                        "23@(100 , 100, 50 ) = ( 20, 5 , 2)   ->  22",
                        "1456 @ (99,89,4332 )  ->  5666"
                      };
      
      SparseDataTable sparseblktable = new SparseDataTable();
      for ( int i = 0; i < maps.length; i++ ){
         SparseBlockInfo  mi = new SparseBlockInfo( maps[i] );
         System.out.println( "---->" + maps[i] );
         System.out.println("mi.getSparseBlockId(): " + mi.getSparseBlockId() );
         sparseblktable.addToTable(mi.getSparseBlockId(), mi);
       
      }
      
      SparseBlockInfo info = sparseblktable.getSparseBlockInfo(5666);
      System.out.println("info's sparseBlock :" + info.getSparseBlockId());
      System.out.println("info's originalBlock :" + info.getOriginalBlockId());
      
      
      
      
      
   }
   
}
