/* 
    TextFileDataSource.java
    Copyright 2000 University of New Hampshire
    Durham, NH 03824
*/

//---------------------- Edit history (as of 11/04) -----------------------
// 02/12/12 rdb: Added impl of subBasicBlock( ISB, ISB, ISB ) (which
//               just call subblock), because this is now an abstract
//               class in parent.
// 12/05/05 rdb: replaced subblock methods with subBasicBlock and subCompositeBlock.
// 10/21/05 rdb: removed unused variables
// 01/10/05 rdb: removed direct allocation of children of Datum
// 11/29/04 rdb: added support for Short and Byte types
// 11/23/04 rdb: replaced deprecated DatumBlock usage 
//
//-------------------------------------------------------------------------


package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;

import java.io.*;


/** This class will read from a file of numeric ASCII data, providing an n-dimensional
    datasource view of that file. At the moment, it should only be used for small files, since
    it reads the entire file into an array. 
*/

public class TextFileDataSource extends PhysicalDataSource{


    protected StreamTokenizer dataTokens;
    protected Datum theData[]; 
	 protected FileInputStream dataFile;
	
    public TextFileDataSource(	String name,
                                RecordDescriptor recordDesc, 
                                ISBounds ISBounds,
                                AxisOrdering ordering, 
                                String URL){
        
        super(name,recordDesc,ISBounds,ordering,URL); // call inherited constructor
    }
        
    
    /** Reads the contents of the file into an array. */
    public void activate(){

        int i,j;
        int numelements=this.bounds.volume();
        int numAttributes=recordDesc.size();

        try {
            this.dataFile=new FileInputStream(this.getURL());
        } catch (IOException e){
            new FileError(SDBError.Recoverable,"IOException in FileDataSource activate(): "+e);
        }

        Reader dr=new BufferedReader(new InputStreamReader(this.dataFile));
        dataTokens=new StreamTokenizer(dr);
        
        theData=new Datum[numelements]; 
        
       
       for(i=0;i<numelements;i++){ 
           theData[i]=this.getRecordDescriptor().createDatum();
           for(j=0;j<numAttributes;j++){
               //theData[i].setFloat(j,this.readValue(recordDesc.getIndexedAttribute(j).type()));
               theData[i].setFloat(j,this.readValue(recordDesc.getStorageType(j)));

           }
       }            

    }


	/** Releases the datum array.*/
	public void deactivate(){
		
		//super.deactivate();
		
		this.theData=null;
	
	}

    
    /** Return the Datum corresponding to the given IndexSpaceID. */
    public    Datum datum( IndexSpaceID did )
    {
       return this.getRecordDescriptor().createDatum( 
             theData[(int) AxisOrdering.computeOffset(did, this.axisWeights)] );
    }

    /** Return the Datum corresponding to the given IndexSpaceID. The 
        Datum will only contain attributes specified by the RecordSpec
        argument. 
    */    
    public Datum datum( IndexSpaceID did, RecordSpec spec )
    {
       return theData[(int) AxisOrdering.computeOffset(did, this.axisWeights)].subset( spec );
    }

    /** Return by reference the Datum corresponding to the given 
     *  IndexSpaceID.
     */
    public void datum( Datum dtm, IndexSpaceID did) 
    {
        Datum thisdtm = theData[(int) AxisOrdering.computeOffset(did, this.axisWeights)];
        int size = thisdtm.size();

        /* fill values of thisdtm in dtm */
        for (int i=0; i<size; i++) {
            dtm.setFloat(i, thisdtm.getFloat(i));
        }
    }

    
    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum(Datum dtm, IndexSpaceID did, FieldIDMapper mp) {

        Datum thisdtm = theData[(int) AxisOrdering.computeOffset(did, this.axisWeights)];
        int [] cpnFids = mp.getComponentFields();
        int [] cpsFids = mp.getCompositeFields();
        int size = cpsFids.length;

        /* fill values of thisdtm in dtm */
        for (int i=0; i<size; i++) {
            dtm.setFloat(cpsFids[i], 
                                thisdtm.getFloat(cpnFids[i]));
        }

    }


    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec) {

        Datum thisdtm = theData[(int) AxisOrdering.computeOffset(did, this.axisWeights)];
        int size = recSpec.size();
        int [] spec = recSpec.getFieldNumbers();

        /* fill values of thisdtm in dtm */
        for (int i=0; i<size; i++) {
            dtm.setFloat(i, thisdtm.getFloat(spec[i]));
        }

    }
    
    
    // dsisb=srcBounds       blkisb=dstBounds
    // dstBounds, srcBounds

    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, ISBounds srcBounds)
    {
        ISIterator iter1 = new ISIterator(srcBounds);
        ISIterator iter2 = new ISIterator(dstBounds);
        
        for( ; iter1.valid(); iter1.next(), iter2.next())
        {
            theBlock.setDatum( iter2, this.datum(iter1) );
        }
    }
    
    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The FieldIDMapper argument determines the mapping between 
        datums in the datasource and datums in the DataBlock. */    
    public void subblock( DataBlock theBlock, 
                          ISBounds dstBounds, 
                          ISBounds srcBounds, 
                          FieldIDMapper mp)
    {
        ISIterator iter1 = new ISIterator(srcBounds);
        ISIterator iter2 = new ISIterator(dstBounds); 
        Datum d = this.getRecordDescriptor().createDatum();
        //Datum [] data = ( (DatumBlock)theBlock).data();
            
        for( ; iter1.valid(); iter1.next(), iter2.next())
        {
            //d = data[theBlock.computeOffset(iter2)];
            //d = theBlock.datum( iter2 );
            this.datum( d, iter1, mp );
            theBlock.setDatum( iter2, d );
        }
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock argument
        defined by dstBounds. The RecordSpec argument specifies a subset of the DataSource
        attributes to be written into the DataBlock datums.*/
    public void subblock( DataBlock theBlock, 
                          ISBounds  dstBounds, 
                          ISBounds srcBounds, 
                          RecordSpec rs)
   {
        ISIterator iter1 = new ISIterator(srcBounds);
        ISIterator iter2 = new ISIterator(dstBounds);
        Datum d;
        for(;iter1.valid();iter1.next(),iter2.next()){
         
            d = this.getRecordDescriptor().createDatum( rs );
            this.datum( d, iter1, rs );
            theBlock.setDatum( iter2, d );
        }
    }


    /** Returns a spatial subset of the DataSource as delimited by the 
     *  ISBounds argument. The DataSource data is copied when filling the DataBlock. 
    */
    public    DataBlock subblock( ISBounds srcdb )
    {
       /************* deleted 11/23/04 ***************
        DatumBlock  bb   = new BasicBlock( srcdb, this,recordDesc );
        ISIterator srcIter = new ISIterator(srcdb);
        FloatDatum[] data=new FloatDatum[srcdb.volume()];
        
        int i;
        for( srcIter.init(), i=0; srcIter.valid(); srcIter.next())
        {
           data[i++]=(FloatDatum) this.datum(srcIter);
        }
        return new DatumBlock(data,srcdb,recordDesc);
       /*****************************************************/
        ISBounds   dstdb   = srcdb.copyToOrigin();
        ISIterator srcIter = new ISIterator( srcdb );
        ISIterator dstIter = new ISIterator( dstdb );

        BasicBlock  bb   = new BasicBlock( srcdb, this.recordDesc );
        
        for( srcIter.init(), dstIter.init(); 
             srcIter.valid(); 
             srcIter.next(), dstIter.next())
        {
           bb.setDatum( dstIter, this.datum( srcIter ));
        }
        return bb;
    }


    public  DataBlock subblock( FloatDatum [] data, ISBounds srcdb )
    {
       /************* deleted 11/23/04 ***************
        DatumBlock  bb   = new BasicBlock( srcdb, this,recordDesc );
        ISIterator srcIter = new ISIterator(srcdb);
        
        int i;
        for( srcIter.init(), i=0; srcIter.valid(); srcIter.next())
        {
           data[i++]=(FloatDatum) this.datum(srcIter);
        }
        return new DatumBlock(data,srcdb,recordDesc);
       /*****************************************************/
        ISBounds   dstdb   = srcdb.copyToOrigin();
        ISIterator srcIter = new ISIterator( srcdb );
        ISIterator dstIter = new ISIterator( dstdb );

        BasicBlock  bb   = new BasicBlock( srcdb, this.recordDesc );
        
        int i = 0;
        for( srcIter.init(), dstIter.init(), i = 0; 
             srcIter.valid(); 
             srcIter.next(), dstIter.next(), i++ )
        {
           bb.setDatum( dstIter, this.datum( srcIter ));
           data[i++]=(FloatDatum) this.datum(srcIter);
        }
        return bb;
     }


    public    DataSource subset(RecordDescriptor rd){
    
        return null;
    }
    
    public    DataSource subset(ISBounds db){
    
        return null;
    }
    
    public    DataSource subset(RecordDescriptor rd, ISBounds db){
    
        return null;
    }
    

    private float readValue(int type){
        
        try {
        
            dataTokens.nextToken();
            
            switch (type){
                case StorageType.Float:
                    return (float) (dataTokens.nval);
                case StorageType.Double:
                    return (float) (dataTokens.nval);
                // rdb: in order to get consistent handling of types, we need to 
                //      make sure that the input is actually turned into its "native"
                //      type and then cast to float. Although this loses information,
                //      it is the correct interpretation of reading this file as a ints
                //      or shorts or bytes. Besides, I need this in order to generate
                //      test programs that automatically compare outputs of different
                //      variations of data sources.
                case StorageType.Int:
                    return (int) (dataTokens.nval);
                case StorageType.Short:
                   return(short) (dataTokens.nval);
                case StorageType.Byte:
                   return (byte) (dataTokens.nval);
                default:
                    new IllegalArgumentError(SDBError.Recoverable,"Bad type in TextFileDataSource.readValue()");                                        					return Float.NaN;
            }
        } catch(IOException e){
            new FileError(SDBError.Recoverable,"Error reading file in TextFileDataSource.readValue(): "+e);
            return Float.NaN;
        }
        
    }

    /************* Following methods should never get called, since
     * TextFileDataSource overrides the subblock methods of PhysicalDataSource, 
     * which is where these methods are called.
     */
    private String errMsg = "TextFileDataSource.subBasicBlock: should NOT get called.";
    private String errMsg2 = "TextFileDataSource.subCompositeBlock: should NOT get called. ";
    void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds ) 
    {
       new SDBError( SDBError.NeverThrow, errMsg + "(bb, isb, isb)");
    }
    void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                                  RecordSpec recSpec ) 
    {
       new SDBError( SDBError.NeverThrow, errMsg + "(bb, isb, isb, rs)");
    }
    void subBasicBlock( BasicBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,
                                  FieldIDMapper mp ) 
    {
       new SDBError( SDBError.NeverThrow, errMsg + "(bb, isb, isb, fidmp" );
    }
    
    void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds )
    {
       new SDBError( SDBError.NeverThrow, errMsg + "(cb, isb, isb)");
    }
    void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds, 
                                  RecordSpec recSpec )
    {
       new SDBError( SDBError.NeverThrow, errMsg2 + "(cb, isb, isb, rs)" );
    }
    void subCompositeBlock( CompositeBlock dstBlk, ISBounds dstBnds, ISBounds srcBnds,  
                                  FieldIDMapper mp )
    {
       new SDBError( SDBError.NeverThrow, errMsg2 + "cb, isb, isb, fidmp");
    }
    

    /** Print the data members of the DataSource in an easy to read fashion.*/
    public String toString(){

        StringBuffer result=new StringBuffer("TextFileDataSource: theData.length=");
        
        if(theData!=null){
            result.append(theData.length);
        } else {
            result.append("null");
        }
        
        
        result.append("\n");
        result.append(super.toString());

        return new String(result);
    }
    
    
    static public void main(String[] args){
    
    
        System.out.println("\n******************************\n" 
                         + "***TextFileDataSource Tests***\n"
                         + "******************************" );
        
        AttributeDescriptor[] attrDescriptors=new AttributeDescriptor[]{
                           new AttributeDescriptor(StorageType.Double,"Carbon"),
                           new AttributeDescriptor(StorageType.Double,"Nitrogen"),
                           new AttributeDescriptor(StorageType.Double,"Oxygen"),
                           new AttributeDescriptor(StorageType.Double,"Bogosity")
                                                 };
                                                 
        RecordDescriptor recordDesc=new RecordDescriptor(attrDescriptors); 
        
        
        AxisOrdering ordering;
        ISBounds ISBounds;
        TextFileDataSource ds1;
         
        // One D Test
        String file = "testdata/file.16-4.txt";
        System.out.println("\n\n***One Dimensional Test*** : " + file + "\n");
        
        ordering=new AxisOrdering(1);
        ISBounds=new ISBounds(new IntegerIndexSpaceID(0),new IntegerIndexSpaceID(63));
    

        // 1-d file, 16 datums of 4 attributes
        ds1=new TextFileDataSource("foo",recordDesc, ISBounds,ordering, file );
        
        ds1.activate();

        System.out.println(ds1);
        
        //System.out.println(ds1.dumpData());
    
        System.out.println("ds1.datum(<0>) should be {0.0, 0.1, 0.2, 0.3}-->"
                           + ds1.datum(new IntegerIndexSpaceID(0)));    
        System.out.println("ds1.datum(<7>) should be {7.0, 7.1, 7.2, 7.3}-->"
                           + ds1.datum(new IntegerIndexSpaceID(7)));    
        System.out.println("ds1.datum(<15>) should be {15.0, 15.1, 15.2, 15.3}-->"
                           +ds1.datum(new IntegerIndexSpaceID(15)));    

    
        // Two D Test
        file = "testdata/colrow.16.4-1.txt";
        System.out.println("\n\n***Two Dimensional Test*** " + file + "\n");
        
        // 2-d file 15x3 of 1 attribute, 1-0 axis axisOrdering
        ISBounds=new ISBounds( new IntegerIndexSpaceID(0,0),
                               new IntegerIndexSpaceID(15,3));
        int TwoD_Ordering[]={1,0};
        ordering=new AxisOrdering(TwoD_Ordering);

        attrDescriptors=new AttributeDescriptor[]{
                   new AttributeDescriptor(StorageType.Double,"Carbon")
                                                 };
                                                 
        recordDesc=new RecordDescriptor(attrDescriptors); 
         
        ds1=new TextFileDataSource("bar",recordDesc, ISBounds,ordering, file );
        
        ds1.activate();
        
        System.out.println("DS Ordering="+ordering);
        
        System.out.println("ds1.datum(<0,0>) should be {0.0}-->"
              +ds1.datum(new IntegerIndexSpaceID(0,0)));    
        System.out.println("ds1.datum(<0,1>) should be {100}-->"
              +ds1.datum(new IntegerIndexSpaceID(0,1)));    
        System.out.println("ds1.datum(<1,0>) should be {1}-->"
              +ds1.datum(new IntegerIndexSpaceID(1,0)));    

        System.out.println("ds1.datum(<8,1>) should be {108}-->"
              +ds1.datum(new IntegerIndexSpaceID(8,1)));    
        System.out.println("ds1.datum(<8,2>) should be {208}-->"
              +ds1.datum(new IntegerIndexSpaceID(8,2)));    
        System.out.println("ds1.datum(<8,3>) should be {308}-->"
              +ds1.datum(new IntegerIndexSpaceID(8,3)));    
        
        System.out.println("ds1.datum(<15,2>) should be {215}-->"
              +ds1.datum(new IntegerIndexSpaceID(15,2)));    
        System.out.println("ds1.datum(<15,3>) should be {315}-->"
              +ds1.datum(new IntegerIndexSpaceID(15,3)));    
        System.out.println("ds1.datum(<14,3>) should be {314}-->"
              +ds1.datum(new IntegerIndexSpaceID(14,3)));    
        
        
        //System.out.println(ds1.dumpData());
        
        
        
        // TwoD subblock() Tests

        System.out.println("\n\n***Two Dimensional subblock() Test*** " + file + "\n");
        
        ISBounds bounds=new ISBounds(new IntegerIndexSpaceID(2,2),
                                     new IntegerIndexSpaceID(10,3));
        DataBlock dblock=ds1.subblock(bounds);
        System.out.println("\n Bounds are "+bounds);

        //bounds.moveToOrigin();
        
        System.out.println(ds1.dumpData());
        
        System.out.println("dblock.datum(<0,0>) should be {202}-->"
              +dblock.datum(new IntegerIndexSpaceID(0,0)));
        System.out.println("dblock.datum(<0,1>) should be {302}-->"
              +dblock.datum(new IntegerIndexSpaceID(0,1)));
        System.out.println("dblock.datum(<1,0>) should be {203}-->"
              +dblock.datum(new IntegerIndexSpaceID(1,0)));
        
        System.out.println("dblock.datum(<4,0>) should be {206}-->"
              +dblock.datum(new IntegerIndexSpaceID(4,0)));
        System.out.println("dblock.datum(<4,1>) should be {306}-->"
              +dblock.datum(new IntegerIndexSpaceID(4,1)));

        System.out.println("dblock.datum(<8,0>) should be {210}-->"
              +dblock.datum(new IntegerIndexSpaceID(8,0)));
        System.out.println("dblock.datum(<8,1>) should be {310}-->"
              +dblock.datum(new IntegerIndexSpaceID(8,1)));
        System.out.println("dblock.datum(<7,1>) should be {309}-->"
              +dblock.datum(new IntegerIndexSpaceID(7,1)));
        
        
        //System.out.println(dblock);
        
        
        file = "testdata/file.8.8-4.txt";
        System.out.println("\n\n***Two Dimensional subblock(DataBlock, "
                   + " ISBounds, ISBounds, FieldIDMapper) Test***" + file + "\n");
        
        bounds=new ISBounds( new IntegerIndexSpaceID(0,0),
                             new IntegerIndexSpaceID(7,7));
        RecordDescriptor rd=new RecordDescriptor(new AttributeDescriptor[] { 
                        new AttributeDescriptor(StorageType.Float,"carbon"),
                        new AttributeDescriptor(StorageType.Float,"nitrogen"),
                        new AttributeDescriptor(StorageType.Float,"naugahyde"),
                        new AttributeDescriptor(StorageType.Float,"freon")
                    }
                );
                             
                                    
        TextFileDataSource tfds4=
                 new TextFileDataSource("4 attributes", rd, bounds, null, file );
        tfds4.activate();
        
        //System.out.println(tfds4);
        //System.out.println(tfds4.dumpData());
        
        FieldIDMapper fidm=new FieldIDMapper(new int[] {1,0,3,2},new int[] {0,3,1,2});
        //rdb: I think the rs below does the same as the fidm above.
        RecordSpec    rs  = new RecordSpec( new int[]  {1,3,2,0});
        
        bounds=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(3,2));
        //DataBlock dblock4=new DatumBlock( bounds, rd.createBlock(fidm));
        //rdb DataBlock dblock4=new BasicBlock( bounds, rd.create(fidm));
        DataBlock dblock4=new BasicBlock( bounds, rd.create(rs));
        
        ISBounds dstBounds=new ISBounds(new IntegerIndexSpaceID(1,1),
                                        new IntegerIndexSpaceID(2,2));
        ISBounds srcBounds=new ISBounds(new IntegerIndexSpaceID(4,4),
                                        new IntegerIndexSpaceID(5,5));
        
        tfds4.subblock( dblock4, dstBounds, srcBounds, fidm );
        //tfds4.subblock( dblock4, dstBounds, srcBounds );

        System.out.println("Should be:");
        
        System.out.println("{NaN, NaN, NaN, NaN}	{NaN, NaN, NaN, NaN}	{NaN, NaN, NaN, NaN}");	
        System.out.println("{NaN, NaN, NaN, NaN}	{44.2, 44.4, 44.3, 44.1}	{45.2, 45.4, 45.3, 45.1}");		
        System.out.println("{NaN, NaN, NaN, NaN}	{54.2, 54.4, 54.3, 54.1}	{55.2, 55.4, 55.3, 55.1}");		
        System.out.println("{NaN, NaN, NaN, NaN}	{NaN, NaN, NaN, NaN}	{NaN, NaN, NaN, NaN}");	
        System.out.println();

        
        System.out.println(dblock4);
        
        //System.exit(0);
        
        
        
        

        // Three D Test
        file = "testdata/file.4.4.4-1.txt";
        System.out.println("\n\n***Three Dimensional Test*** " + file + "\n");
        
        ISBounds=new ISBounds(new IntegerIndexSpaceID(0,0,0),new IntegerIndexSpaceID(3,3,3));
        int ThreeD_Ordering[]={0,2,1};
        ordering=new AxisOrdering(ThreeD_Ordering);
        ds1=new TextFileDataSource("baz", recordDesc, ISBounds, ordering, file );
        
        ds1.activate();
        
        
        System.out.println("ds1.datum(<0,0,0>) should be {0}-->"
              +ds1.datum(new IntegerIndexSpaceID(0,0,0)));    
        System.out.println("ds1.datum(<1,0,0>) should be {100}-->"
              +ds1.datum(new IntegerIndexSpaceID(1,0,0)));    
        System.out.println("ds1.datum(<0,1,0>) should be {1}-->"
              +ds1.datum(new IntegerIndexSpaceID(0,1,0)));    
        System.out.println("ds1.datum(<0,0,1>) should be {10}-->"
              +ds1.datum(new IntegerIndexSpaceID(0,0,1)));    
        
        System.out.println("ds1.datum(<1,1,1>) should be {111}-->"
              +ds1.datum(new IntegerIndexSpaceID(1,1,1)));    
        System.out.println("ds1.datum(<2,1,1>) should be {211}-->"
              +ds1.datum(new IntegerIndexSpaceID(2,1,1)));    
        System.out.println("ds1.datum(<1,2,1>) should be {112}-->"
              +ds1.datum(new IntegerIndexSpaceID(1,2,1)));    
        System.out.println("ds1.datum(<2,2,1>) should be {212}-->"
              +ds1.datum(new IntegerIndexSpaceID(2,2,1)));    

        System.out.println("ds1.datum(<1,1,2>) should be {121}-->"
              +ds1.datum(new IntegerIndexSpaceID(1,1,2)));    
        System.out.println("ds1.datum(<2,1,2>) should be {221}-->"
              +ds1.datum(new IntegerIndexSpaceID(2,1,2)));    
        System.out.println("ds1.datum(<1,2,2>) should be {122}-->"
              +ds1.datum(new IntegerIndexSpaceID(1,2,2)));    
        System.out.println("ds1.datum(<2,2,2>) should be {222}-->"
              +ds1.datum(new IntegerIndexSpaceID(2,2,2)));    
        
        System.out.println("ds1.datum(<3,3,3>) should be {333}-->"
              +ds1.datum(new IntegerIndexSpaceID(3,3,3)));    
        System.out.println("ds1.datum(<2,3,3>) should be {233}-->"
              +ds1.datum(new IntegerIndexSpaceID(2,3,3)));    
        System.out.println("ds1.datum(<3,2,3>) should be {332}-->"
              +ds1.datum(new IntegerIndexSpaceID(3,2,3)));    
        System.out.println("ds1.datum(<3,3,2>) should be {323}-->"
              +ds1.datum(new IntegerIndexSpaceID(3,3,2)));    
        
        System.out.println( "++++++++++++ end tests ++++++++++++" );
    }
}
