//
//  ThreadedBlockCacheDataSource.java
//  CacheTest
//
//  Created by Philip Rhodes on Fri Dec 12 2003.
//  Copyright (c) 2003 University of New Hampshire. All rights reserved.
//


package edu.unh.sdb.datasource;

//import edu.unh.sdb.common.*;

/** 
 *  The ThreadedBlockCacheDataSource provides cache service for an underlying datasource. CacheDataSources
 *   are typically created by the CacheMaker class, which tunes the size of the cache blocks to
 *   a particular iteration and underlying datasource. As its name implies, the ThreadedBlockCacheDataSource
 *   uses an IO thread to read cache blocks from disk before they are actually needed by the user. To accomplish
 *   this, this class actually creates two cache blocks. One is being accessed by the user, and the other
 *   is being preloaded. This scheme is especially effective if the user is doing substantial computation, so
 *   that the IO thread has time to completely read its block before it is needed.
 *     The IO thread is triggered whenever the lower bounds of a user query crosses into a new cache block. When
 *   this happens, it is assumed that the user is done with the old block, so new data can be loaded into it.
 *   Notice that if a subblock query straddles two cache blocks, the I/O thread will not be triggered, because
 *   the lower bounds of the query block are still within the old block.
 *      When a query is received, it will be satisfied from the two cache blocks if possible.
 *    Otherwise, the query is satisfied directly from the datasource, and the cache blocks remain unchanged.
 */
public class ThreadedBlockCacheDataSource extends BlockCacheDataSource{

   protected AxisOrdering iteratorOrdering;

    /** Create a ThreadedBlockCacheDataSource with a single CacheBlock connected to the DataSource
      * argument. The CacheBlock dimensions are specified by the last argument. The memory 
      * for the cache itself is not allocated until this datasource is activated. 
      */
	public ThreadedBlockCacheDataSource(String name, DataSource ds, ISBounds iterationSpace, IndexSpaceID blockDimensions, AxisOrdering iteratorOrdering){
	
		super(name, ds, iterationSpace, blockDimensions);
      
      this.iteratorOrdering = iteratorOrdering;
	}

    /** Create a ThreadedBlockCacheDataSource with a single CacheBlock connected to the DataSource
      * argument. The CacheBlock dimensions are specified by the last argument. The memory 
      * for the cache itself is not allocated until this datasource is activated. 
      */
	public ThreadedBlockCacheDataSource(String name, DataSource ds,  IndexSpaceID blockDimensions, AxisOrdering iteratorOrdering){
	
		super(name, ds, ds.getBounds(), blockDimensions);

      this.iteratorOrdering = iteratorOrdering;
	}



	/** Calls activate() for the underlying datasource, and creates the actual cache,
      * which allocates memory for the CacheBlock.
	  */	
	public void activate(){
	
	    super.activate();
        
       //theCache=new ThreadedBlockCache(this.theSource,  this.iterationSpace, this.blockDimensions, ((PhysicalDataSource)this.theSource).getAxisOrdering());
      theCache=new ThreadedBlockCache(this.theSource,  this.iterationSpace, this.blockDimensions, this.iteratorOrdering);
        
    }

	/** Calls super.deactivate() for the underlying datasource, and waits for any lingering IO threads
	 *  to terminate.
	 */	
	public void deactivate(){
	
		this.halt();
		
	    super.deactivate();
	}



	/** Any lingering IO threads will have terminated after this method completes.
	 */	
	protected void halt(){
	
		if (theCache != null){
      
			((ThreadedBlockCache)theCache).waitOnIOThread();
		}
	}
	
	
	/** For debugging. Runs basic tests on the ThreadedBlockCacheDataSource class. */
    static public void main(String[] args){

        System.out.println("\n******************************\n***ThreadedBlockCacheDataSource Tests***\n******************************");
           
    }
		
}
