package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.vecmath.Vector3d;

/**
 * @author dlipsa
 * Converts a datum read from a RGB data source into rgba array of 4 bytes ready
 * to be stored into a texture.
 * Converts a datum read from a CT data source (which contains ints in the
 * range 0-4096) into a rgba array of 4 bytes. First it scales the CT value to
 * a value in the range 0-255 and then duplicates that value for all components.
 */
public abstract class TransferFunction
{
   // do not change the order of values in the enum. If you do, change 
   public static enum Feature
   {
      SKIN,
      BONE,
      SKIN_BONE
   };
   
   public static enum Type
   {
      INVALID,
      RGB,
      CT,
      VORTICITY_MAGNITUDE 
   }
   
   public abstract void convert (Datum datum, byte[] texture, int index);
   public abstract void convert (Values values, byte[] texture);
   public abstract boolean isFirstBigger (byte[] t1, int i1, byte[] t2, int i2);
   public abstract Datum createDatum ();
   public abstract void trilinearInterpolation (Block block, Datum datum,
      IndexSpaceID lower);
   public abstract Type getType ();   
   public abstract void classifyVoxel (DataBlock src, IndexSpaceID srcId,
      DataBlock dest, IndexSpaceID destId, Feature feature);
   public abstract void shadeVoxel (DataBlock src, IndexSpaceID srcId,
      DataBlock dest, IndexSpaceID destId, Vector3d L, Vector3d H, 
      double distFromPicturePlane);   
   
   public void shadeVoxels (DataBlock src, ISBounds srcBounds, DataBlock dest, 
      Vector3d L, Vector3d H, double distFromPicturePlane)
   {
      ISIterator it = new ISIterator (srcBounds);
      for (it.init (); it.valid (); it.next ())
      {
         IndexSpaceID destId = it.relativeLocation ();
         shadeVoxel (src, it, dest, destId, L, H, distFromPicturePlane);
      }
   }
   
   public void classifyVoxels (DataBlock src, ISBounds srcBounds, DataBlock dest, Feature feature)
   {
      ISIterator it = new ISIterator (srcBounds);
      for (it.init (); it.valid (); it.next ())
      {
         IndexSpaceID destId = it.relativeLocation ();
         classifyVoxel (src, it, dest, destId, feature);
      }    
   }
      
   public static TransferFunction createDefault (DataSource dataSource)
   {
      Type dataSourceType = type (dataSource);
      if (dataSourceType == Type.RGB)
         return new TransferFunctionRGB ();
      else if (dataSourceType == Type.CT)
         return new TransferFunctionCT ();
      else if (dataSourceType == Type.VORTICITY_MAGNITUDE)
         return new TransferFunctionVM();
         
      else
         return null;
   }

   static public double trilinearInterpolation (TransferFunction.Block block,
      IndexSpaceID lower, int attr)
   {
      if (logger.isLoggable (Level.FINE))
         logger.log (Level.FINE,"trilinear: ");
      // interpolation along X
      double tyz = block.getDouble (lower, attr) * (1-block.ratioX);
      lower.add (0, 1);
      tyz += block.getDouble (lower, attr) * block.ratioX;
      
      lower.subtract (0, 1);lower.add (1, 1);
      double ty1z = block.getDouble (lower, attr) * (1 - block.ratioX);
      lower.add (0, 1);
      ty1z += block.getDouble (lower, attr) * block.ratioX;
      
      lower.subtract (0, 1);lower.add (2, 1);
      double ty1z1 = block.getDouble (lower, attr) * (1 - block.ratioX);
      lower.add (0, 1);
      ty1z1 += block.getDouble (lower, attr) * block.ratioX;
      
      lower.subtract (0, 1);lower.subtract (1, 1);
      double tyz1 = block.getDouble (lower, attr) * (1 - block.ratioX);
      lower.add (0, 1);
      tyz1 += block.getDouble (lower, attr) * block.ratioX;
      
      // restore lower
      lower.subtract (0, 1);lower.subtract (2, 1);
      
      //interpolation along Y
      double tz = tyz * (1- block.ratioY) + ty1z * block.ratioY;
      double tz1 = tyz1 * (1 - block.ratioY) + ty1z1 * block.ratioY;
      
      // interpolation along Z
      double val = tz * (1 - block.ratioZ) + tz1 * block.ratioZ;
      if (logger.isLoggable (Level.FINE))
         logger.log (Level.FINE," -> " + val);
      return val;
   }
   
   /**
    * Block of data that is read for a datasource.
    */
   public static class Block
   {
      public Block ()
      {}
      
      public Block (DataBlock data, double ratioX, double ratioY, double ratioZ)
      {
         this.data = data;
         this.ratioX = ratioX;
         this.ratioY = ratioY;
         this.ratioZ = ratioZ;
      }
      
      public double getDouble ( IndexSpaceID isid, int attr )
      {
         double val = data.getDouble (isid, attr);
         if (logger.isLoggable (Level.FINE))
            logger.log (Level.FINE, " " + val);
         return val;
      }
      
      public void setRatios (double x, double y, double z)
      {
         ratioX = x;
         ratioY = y;
         ratioZ = z;
      }
      
      public void setRatioX (double ratioX)
      {  this.ratioX = ratioX;}
      
      public void setRatioY (double ratioY)
      {  this.ratioY = ratioY;}
      
      public void setRatioZ (double ratioZ)
      {  this.ratioZ = ratioZ;}
      
      public double getRatioX ()
      {  return ratioX;}
      
      public double getRatioY ()
      {  return ratioY;}
      
      public double getRatioZ ()
      {  return ratioZ;}
      
      public DataBlock getData ()
      {  return data;}
      
      public void setData (DataBlock data)
      {
         this.data = data;
      }
      
      private DataBlock data;
      private double ratioX, ratioY, ratioZ;      
   }
      

   /* See: Display of surfaces from volume data, Mark Levoy
    */   
   protected static double shadeVoxel (DataBlock src, IndexSpaceID srcId, int attr, 
      Vector3d L, Vector3d H, double d)
   {
      final double cp = 1;
      final double ka = 0.1;
      final double kd = 0.45;
      final double ks = 0.5;
      final double n = 27.0;
      // Atmosferic attenuation 16.1.3 Foley et. al.
      final double k1 = .1;
      final double k2 = .9/256;
      final Vector3d zero = new Vector3d (0, 0, 0);
                  
      Vector3d N = gradient (src, srcId, attr);
      double color;
      if (N.epsilonEquals (zero, .000001))
         color = cp * ka;
      else
      {
         N.normalize ();
         color = cp * ka + (cp * (k1 + k2 * d)) * 
            (kd * Math.abs (N.dot (L)) + ks * Math.pow (Math.abs (N.dot (H)), n));
         color = (color <= 1) ? color : 1;
      }
      return color;
   }
      
   private static Type type (DataSource dataSource)
   {
      RecordDescriptor recordDescriptor = dataSource.getRecordDescriptor ();
      int[] storageTypes = recordDescriptor.getStorageTypes ();
      if (storageTypes.length == 3 &&
         storageTypes[0] == StorageType.Byte &&
         storageTypes[1] == StorageType.Byte &&
         storageTypes[2] == StorageType.Byte)
         return Type.RGB;
      
      else if (storageTypes.length == 1 &&
         storageTypes[0] == StorageType.Short)
         return Type.CT;
         
      else if (storageTypes.length == 1 &&
         storageTypes[0] == StorageType.Double)
         return Type.VORTICITY_MAGNITUDE;
         
      else
         return Type.INVALID;
   }
   
   protected static Vector3d gradient (DataBlock block, IndexSpaceID id, int attr)
   {
      double[] g = new double[3];
      for (int i = 0; i < 3; i++)
      {
         id.add (i, 1);
         g[i] = block.getDouble (id, attr) / 2;
         id.subtract (i, 2);
         g[i] -= block.getDouble (id, attr) / 2;
         id.add (i, 1);
      }      
      return new Vector3d (g);      
   }
   
   static Logger logger = Logger.getLogger ("edu.unh.sdb.datasource.TransferFunction");
   
}
