/*
    URCellBlock.java
    Copyright 2004 University of New Hampshire
    Durham, NH 03824
    Programmer: LI YE
 */

// -------------------- Edit history as of 12/05/04 ---------------------------
// 12/09/05 rdb: all createDataBlock calls changed to copy bounds to origin first.
// 11/13/05 rdb: Removed all uses of CompositeBlock
// 10/24/05 rdb: added  DataBlock urCellBlock( ISBounds, ISBounds ) convenience method
//                cleaned up unused variables and deprecated code
// 10/21/05 rdb: removed unused variables.
// 12/05/04 rdb: changed name of recordDesc to recordDesc
//               deleted size(); it is inherited from DataCollection
//------------------------------------------------------------------------------
/**
 * The URCellBlock class implements an interface to access blocks of cell data
 * in an uniform cell space. It contains a point datablock corresponding to
 * the uniform cell space.
 */


package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
import edu.unh.sdb.lattice.*;
//import java.io.*;


public class URCellBlock extends RectCellBlock implements CellIndexable {
    private DimResolution resolution;
        //store the resolution of a cell block
    //rdb private CompositeBlock ptBlock;
    private DataBlock ptBlock;
        //store the cell data block
    private int[] path;
        //store the path of an uniform region cell block in a ARRS tree

    /** Construct a URCellBlock without arguments */
    public URCellBlock()
    { 
       super();
       this.resolution = null;
       this.ptBlock = null;
       this.path = null;
       //this.resBounds = null;
    }

    /** Construct a URCellBlock with the given bounds, axis axisOrdering and
      * RecordDescriptor. The bounds will be moved to the origin.
      */
    public URCellBlock(ISBounds isb, AxisOrdering ordering,
                     RecordDescriptor recordDesc)
    {
       super(isb, ordering, recordDesc);
       this.resolution = null;
       this.ptBlock = null;
       this.path = null;
       //this.resBounds = null;
    }

    /** Construct a URCellBlock with the given bounds and number of attributes
      */
    public URCellBlock(ISBounds isb, RecordDescriptor recordDesc)
    {
       super(isb, recordDesc);
       this.resolution = null;
       this.ptBlock = null;
       this.path = null;
       //this.resBounds = null;
    }
 
    /** Construct a URCellBlock with a given URCellBlock */
    public URCellBlock(DataBlock b)
    {
       super(b.getBounds(), b.getRecordDescriptor());
       //rdb this.ptBlock = (CompositeBlock)((URCellBlock)b).getPointDataBlock();
       this.ptBlock = ((URCellBlock)b).getPointDataBlock();
       this.resolution = ((URCellBlock)b).getDimResolution();
       this.path = ((URCellBlock)b).getPath();
    }

    /** Set the resolution */
    public void setDimResolution(DimResolution resolution)
    {
       this.resolution = resolution;
       this.bounds = this.resolution.getCellBlockBounds();
    }

    /** Set the path of a uniform region cell block in a ARRS tree */
    public void setPath( int[] path )
    {
       if ( path == null )
          this.path = null;
       else
          this.path =  path.clone();
    }


    /** Set a point data block to a uniform cell block */
    public void setPointDataBlock( DataBlock block )
    {
       //rdb this.ptBlock = (CompositeBlock)block;
       this.ptBlock = block;
    }

    /** Get the path of an uniform region cell block in a ARRS tree */
    public int[] getPath()
    {
       return this.path;
    }

    /** Get the dimensional resolution */
    public DimResolution getDimResolution()
    {
       return this.resolution;
    }

    /** Get the base bounds */
    public ISBounds getBaseBounds()
    {
       if (this.resolution != null)
          return this.resolution.getBaseBounds().copy();

       return this.ptBlock.copyBounds();
    }

    /** Return a String representing the object. */
    public String toString(){
       return this.toString("");
    }
                                                                                
    /** Return a String representing the object indented by the argument string.
      */
    public String toString(String indent){
       int i;
       String indent2=indent+Globals.toStringIndent;
       StringBuffer result=new StringBuffer(indent+"{ URCellBlock:\n");
       result.append(indent2+"Bounds: ");
       result.append(this.bounds.toString()+"\n");
       result.append(indent2+"Point bounds: ");
       result.append(this.ptBlock.getBounds().toString()+"\n");
       if (this.resolution != null) {                                    
          result.append(this.resolution.toString(indent2)+"\n");
       }
       if (this.path != null) {
          result.append(indent2+"Path is:     <PATH (");
          for (i = 0; i < this.path.length-1; i++) {
             result.append(this.path[i]+", ");
          }
          result.append(this.path[this.path.length-1]+")>\n");
       }
       result.append(indent+"}");
       return new String(result);
    }

    /** Given an IndexSpaceID, return the corresponding datum.*/
    public  Datum datum(IndexSpaceID did) 
    {
       return null;
    }

    /** Given an array of coordinates, copy the corresponding datum
        into the first parameter.*/
    public void datum(Datum d, int[] coordArray) 
    { 
       return;
    }

    /** Given an IndexSpaceID, return the corresponding datum with attributes
        specified in the RecordSpec. */
    public  Datum datum(IndexSpaceID did, RecordSpec spec)
    { 
        return null;
    }

    /** Set the location specified by the IndexSpaceID to the given datum.*/
    public void setDatum(IndexSpaceID isid, Datum d)
    {
        //this.block.setDatum(isid, d);
        return;
    }

    /** Return a new DataBlock that is a subset of this URCellBlock. */
    public  DataBlock subblock(ISBounds db) 
    { 
       if (this.ptBlock == null)
          return null;
       //get the point bounds corresponding to the cell bounds
       ISBounds bd = db.copyIncreasedBy(1);
       
       if (!this.ptBlock.getBounds().contains(bd)) {
          new IllegalArgumentError(SDBError.Recoverable,"The value of the bounds is invalid.");
          return null;
       }

       //get the point data block corresponding to the cell bounds
       //rdb CompositeBlock ptblock;
       DataBlock ptblock;
       ISBounds bd2;
       //rdb ptblock = new CompositeBlock(bd, this.ptBlock.getRecordDescriptor());
       ptblock = this.ptBlock.getRecordDescriptor().createDataBlock( bd.copyToOrigin() );
       bd2 = ptblock.getBounds();
      // System.out.println("The bounds of CompositeBlock: " + bd2.toString());
       this.ptBlock.subblock(ptblock, bd2, bd);
      // System.out.println(ptblock.toString());
                                                                                
       //create a uniform cell block
       URCellBlock ublock = new URCellBlock(db, this.recordDesc);
       //set a point data block to the uniform cell block
       ublock.setPointDataBlock(ptblock);
      
       return ublock; 
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock
      * argument defined by dstBounds. 
      */
    public void subblock( DataBlock theBlock, ISBounds dstBounds, 
                                  ISBounds srcBounds )
    {
       ISBounds pointSrcBounds, pointDstBounds;
       //rdb CompositeBlock ptDataBlock;
       DataBlock ptDataBlock;
       URCellBlock cb;
                                                                                
       if ( this.ptBlock == null )
          return;

       pointSrcBounds = srcBounds.copyIncreasedBy(1);
       pointDstBounds = dstBounds.copyIncreasedBy(1);
       try {
          cb = (URCellBlock)theBlock;
          if ( cb.getPointDataBlock() == null ) 
          {
             //rdb ptDataBlock = new CompositeBlock(pointSrcBounds,
             //rdb                       this.ptBlock.getRecordDescriptor());
             ptDataBlock = this.ptBlock.getRecordDescriptor()
                               .createDataBlock( pointSrcBounds.copyToOrigin() );
             pointDstBounds = ptDataBlock.getBounds();
          }                                                                                
          else
             //rdb ptDataBlock = (CompositeBlock)cb.getPointDataBlock();
             ptDataBlock = cb.getPointDataBlock();

          this.ptBlock.subblock( ptDataBlock, pointDstBounds, pointSrcBounds );
          cb.setPointDataBlock( ptDataBlock );

       } catch (ClassCastException e){
            new IllegalArgumentError(SDBError.Recoverable,
                  "Non-CellBlock passed to CellDataSource.subblock().");
         }
    }


    /** Reads a subblock defined by srcBounds into the region of the Datablock 
      * argument defined by dstBounds. The FieldIDMapper argument determines 
      * the mapping between datums in this DataBlock and datums in the argument
      * DataBlock. 
      */
    public void subblock( DataBlock theBlock, ISBounds dstBounds, 
                          ISBounds srcBounds, FieldIDMapper mp )
    { 
       ISBounds pointSrcBounds, pointDstBounds;
       //rdb CompositeBlock ptDataBlock;
       DataBlock ptDataBlock;
       URCellBlock cb;
                 
       if (this.ptBlock == null)
          return;

       pointSrcBounds = srcBounds.copyIncreasedBy(1);
       pointDstBounds = dstBounds.copyIncreasedBy(1);
       try{
          cb = (URCellBlock)theBlock;
          if (cb.getPointDataBlock() == null) {
             //rdb ptDataBlock = new CompositeBlock(pointSrcBounds,
             //rddb                           this.ptBlock.getRecordDescriptor());
             ptDataBlock = this.ptBlock.getRecordDescriptor()
                               .createDataBlock( pointSrcBounds.copyToOrigin() );
             pointDstBounds = ptDataBlock.getBounds();
          }
          else
             ptDataBlock = cb.getPointDataBlock();
             //rdb ptDataBlock = (CompositeBlock)cb.getPointDataBlock();

          this.ptBlock.subblock( ptDataBlock, pointDstBounds, pointSrcBounds, mp );
          cb.setPointDataBlock( ptDataBlock );
                 
       } catch (ClassCastException e){
            new IllegalArgumentError(SDBError.Recoverable,
                  "Non-CellBlock passed to CellDataSource.subblock().");
         }
    }
  
 
    /** Reads a subblock defined by srcBounds into the region of the Datablock 
      * argument defined by dstBounds. The RecordSpec argument specifies a 
      * subset of the attributes of this DataSource to be written into the 
      * DataBlock datums.
      */
    public void subblock(DataBlock theBlock, ISBounds dstBounds, 
                                    ISBounds srcBounds, RecordSpec spec)
    { 
       ISBounds pointSrcBounds, pointDstBounds;
      //rdb CompositeBlock ptDataBlock;
       DataBlock ptDataBlock;
       URCellBlock cb;
                                                                                
       if (this.ptBlock == null)
          return;

       pointSrcBounds = srcBounds.copyIncreasedBy(1);
       pointDstBounds = dstBounds.copyIncreasedBy(1);
       try{
          cb = (URCellBlock)theBlock;
          if ( cb.getPointDataBlock() == null ) 
          {
             //rdb ptDataBlock = new CompositeBlock(pointSrcBounds,
             //rdb                           this.ptBlock.getRecordDescriptor());
             ptDataBlock = this.ptBlock.getRecordDescriptor()
                                       .createDataBlock( pointSrcBounds.copyToOrigin() );
             pointDstBounds = ptDataBlock.getBounds();
          }
          else
             ptDataBlock = cb.getPointDataBlock();
             //rdb ptDataBlock = (CompositeBlock)cb.getPointDataBlock();

          this.ptBlock.subblock( ptDataBlock, pointDstBounds, pointSrcBounds, spec );
          cb.setPointDataBlock(ptDataBlock);
           
       } catch (ClassCastException e){
            new IllegalArgumentError(SDBError.Recoverable,
                  "Non-CellBlock passed to CellDataSource.subblock().");
         }
    }

    /** Return a subblock corresponding to the given RecordSpec and the given
     *  ISBounds.
     */
    public DataBlock subblock(RecordSpec rs, ISBounds db)
    {
       if (this.ptBlock == null)
          return null;

       //get the point bounds corresponding to the cell bounds
       ISBounds bd = db.copyIncreasedBy(1);
                                                                                
       if ( !this.ptBlock.getBounds().contains( bd )) 
       {
          new IllegalArgumentError(SDBError.Recoverable,
                                        "The value of the bounds is invalid.");
          return null;
       }
                                                                                
       //get the point data block corresponding to the cell bounds
       //rdb CompositeBlock ptblock;
       DataBlock ptblock;
       ISBounds bd2;
       //rdb ptblock = new CompositeBlock(bd, this.ptBlock.getRecordDescriptor());
       ptblock = this.ptBlock.getRecordDescriptor().createDataBlock( bd.copyToOrigin() );
       bd2 = ptblock.getBounds();
       this.ptBlock.subblock(ptblock, bd2, bd);
                                                                                
       //create a uniform cell block
       URCellBlock ublock = new URCellBlock( db, this.recordDesc );
       //set a point data block to the uniform cell block
       ublock.setPointDataBlock( ptblock );
                                                                                
       return ublock;
    }
                                                                                
    /** Return a reference to the actual array containing the data. */
    //public Datum [] data(){ return null; }

    /** Return by reference the Datum corresponding to the given
     *  IndexSpaceID.
     */
    public void datum(Datum dtm, IndexSpaceID did) 
    { 
       return;
    }
   

    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding FieldIDMapper.
     */
    public void datum(Datum dtm, IndexSpaceID did, FieldIDMapper mp)
    {  
       return;
    }
   

    /** Get the Datum and fill its values in the passed Datum at positions
     *  specified by the corresponding RecordSpec.
     */
    public void datum(Datum dtm, IndexSpaceID did, RecordSpec recSpec)
    { 
       return;
    }

    
    /** Get the data source */
    //rdb public CompositeBlock getPointDataBlock()
    public DataBlock getPointDataBlock()
    {
       return this.ptBlock;
    }


    /** Return a cell corresponding to the specified position */
    public Cell getCell(IntegerIndexSpaceID position)
    {
       ISBounds baseB, baseB2;

       //get the base bounds of the specified cell 
       baseB = new ISBounds((IntegerIndexSpaceID)position.clone(), 
                            (IntegerIndexSpaceID)position.clone());
       baseB = baseB.copyIncreasedBy(1);

       //create a cell id object
       ARRSCellID cellid = new ARRSCellID((IntegerIndexSpaceID)position.clone());
       //rdb CompositeBlock pblock = new CompositeBlock(baseB,
       //rdb                              this.ptBlock.getRecordDescriptor()); 
       DataBlock pblock = this.ptBlock.getRecordDescriptor()
                                      .createDataBlock( baseB.copyToOrigin() );
       baseB2 = pblock.getBounds();
       this.ptBlock.subblock(pblock, baseB2, baseB);
       //System.out.println("pblock : \n"); //--------
       //System.out.println(pblock.toString()); //--------
       Datum[] data = new Datum[baseB2.volume()];
       Point[] point = new Point[baseB2.volume()];
       ISIterator iter = new ISIterator(baseB2);
       int i = 0;
       for (iter.init(); iter.valid(); iter.next()) {
          data[i] = pblock.datum(iter);
          //pblock.datum(data[i], iter.copyID());
          point[i] = new Point(data[i]);
          i++;
       }

       ARRSCell arcell = new ARRSCell(cellid, baseB);
       arcell.setCellData(data);
       arcell.setCellVertices(point);

       return arcell;
    }


    /** Return a cell block with the specified bounds and the specified
     *  resolution 
     */
    public DataBlock urCellBlock( ISBounds b, ISBounds res )
    {
       // create a Cell Block of the right size and call the in place version
       DataBlock newCB = new URCellBlock( res, this.recordDesc );
       this.urCellBlock( newCB, b, res );
       return newCB;
    }

   /** 
    * Fill a cell block with the specified resolution and the specified
    *  bounds 
    */
    public void urCellBlock( DataBlock ublock, ISBounds b, ISBounds cellRes )
    {
       //Indexable pblock;
       //rdb: Why CompositeBlock??
       //rdb CompositeBlock pblock, pblock2;
       DataBlock pblock, pblock2;
       ISBounds pbounds, pbounds2;
       DSApproximator approx = new DSApproximator();
       if ( cellRes.equals( this.bounds )) 
          this.subblock( ublock, ublock.getBounds(), b );
       else 
       {
          pbounds = cellRes.copyIncreasedBy(1);
          pbounds2 = this.ptBlock.getBounds();
          ISIDMapper ptIDMap = new ISIDMapper( pbounds, pbounds2 );
          pbounds = b.copyIncreasedBy(1); 
          //pbounds.moveToOrigin();
          pbounds2 = ptIDMap.toComposite( pbounds );
          //rdb: why a CompositeBlock???
          //pblock2 = new CompositeBlock( pbounds2, this.recordDesc );
          pblock2 = this.recordDesc.createDataBlock( pbounds2.copyToOrigin() );
          this.ptBlock.subblock( pblock2, pblock2.getBounds(), pbounds2 ) ;
          pbounds.moveToOrigin();
          //rdb pblock = (CompositeBlock)approx.approximate( pblock2, pbounds );
          pblock = (DataBlock) approx.approximate( pblock2, pbounds );

          /*create a uniform cell block object corresponding to the specified
            resolution and the specified bounds*/
          ((URCellBlock)ublock).setPointDataBlock( pblock );
       }
    }
}

