/* 
	UniBlockCacheDataSource.java
	Copyright 2000 University of New Hampshire
	Durham, NH 03824
*/
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables; commented out rootId uses
//-----------------------------------------------------------------------

package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;
import java.nio.*;


/** 
 *  The UniBlockCacheDataSource provides cache service for an underlying datasource. CacheDataSources
 *   are typically created by the CacheMaker class, which tunes the size of the cache blocks to
 *   a particular iteration and underlying datasource. As its name implies, the UniBlockCacheDataSource
 *   uses a cache with a single block. When a query is received, it will be satisfied from this cache
 *   block if possible. Otherwise, one of two things will be done to satisfy the query. If the query can
 *   be satisfied entirely by reading a different block into the cache, then the new block is read, and 
 *   the query satisfied. This is always the course of action for datum queries. For subblock queries, if 
 *   the query region is entirely contained in some cache block, then that block is read and the query
 *   satisfied from its contents. If the query region spans more than one cache block, the query is 
 *   satisfied directly from the datasource, and the cache block remains unchanged.
 */
public class UniBlockCacheDataSource extends BlockCacheDataSource{


    /** Create a UniBlockCacheDataSource with a single CacheBlock connected to the DataSource
      * argument. The CacheBlock dimensions are specified by the last argument. The memory 
      * for the cache itself is not allocated until this datasource is activated. 
      */
	public UniBlockCacheDataSource(String name, DataSource ds, ISBounds iterationSpace, IndexSpaceID blockDimensions){
	
		super(name, ds, iterationSpace, blockDimensions);
	}

    /** Create a UniBlockCacheDataSource with a single CacheBlock connected to the DataSource
      * argument. The CacheBlock dimensions are specified by the last argument. The memory 
      * for the cache itself is not allocated until this datasource is activated. 
      */
	public UniBlockCacheDataSource(String name, DataSource ds, IndexSpaceID blockDimensions){
	
		super(name, ds, ds.getBounds(), blockDimensions);
	}


  /** Return the cache reference for possible need to assignment operation. This is only used
    * when separate cache loading timeing is needed. It also saves memory and copy cost in such case if we 
    * use only one copy of the cache.
    *
    * Added by Baoqiang Yan on 11/30/2007
    */
	public CacheDataBlock getCacheBlockReference(){
		return ((UniBlockCache)theCache).getCacheDataBlockReference();
	}
	
  /** Read the data into the theCache which is referenced by the cache block 
    * Added by Baoqiang Yan on 11/30/2007
    */
  public void nextCacheBlock(ISBounds cacheSrcBounds)
  {
      System.out.println("Entering nextCacheBlock of UniBlockCacheDataSource ");
      ((UniBlockCache)theCache).loadCacheBlock(cacheSrcBounds);
      System.out.println("Leaving nextCacheBlock of UniBlockCacheDataSource ");
  }


	/** Calls activate() for the underlying datasource, and creates the actual cache,
      * which allocates memory for the CacheBlock.
	  */	
	public void activate(){
	
	    super.activate();
        
        theCache=new UniBlockCache(this.theSource, this.iterationSpace, this.blockDimensions, ((PhysicalDataSource)this.theSource).getAxisOrdering());
        
    }
	
	
	/** For debugging. Runs basic tests on the UniBlockCacheDataSource class. */
    static public void main(String[] args){

        System.out.println("\n******************************\n***UniBlockCacheDataSource Tests***\n******************************");
        
        AttributeDescriptor[] attrDescriptors=new AttributeDescriptor[]{
                                                    new AttributeDescriptor(StorageType.Float,"Carbon"),
                                                    new AttributeDescriptor(StorageType.Float,"Nitrogen"),
                                                    new AttributeDescriptor(StorageType.Float,"Oxygen"),
                                                    new AttributeDescriptor(StorageType.Float,"Bogosity")
                                                 };
                                                 
        RecordDescriptor recordDesc=new RecordDescriptor(attrDescriptors); 
 
        AxisOrdering ordering;
        ISBounds bounds;
        BinaryFileDataSource ds1;
        UniBlockCacheDataSource cds;       

        // One D Test
        System.out.println("\n\n***One Dimensional Test***\n");
        
        ordering=new AxisOrdering(1);
        bounds=new ISBounds(new IntegerIndexSpaceID(0),new IntegerIndexSpaceID(63));
    
        IndexSpaceID blockSizes=new IntegerIndexSpaceID(8);
        //IndexSpaceID numBlocks=new IntegerIndexSpaceID(2);
        
        ds1=new PointOrderBFDS("foo",recordDesc, bounds,ordering,"testdata/file.6.4.bin", ByteOrder.BIG_ENDIAN );
        cds=new UniBlockCacheDataSource("bar",ds1,blockSizes);
        
        cds.activate();
        
        
    
        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    
        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    

        System.out.println("ds1.datum(<7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(7)));    
        System.out.println("ds1.datum(<8>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(8)));    
        
        System.out.println("ds1.datum(<16>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(16)));    

        System.out.println("ds1.datum(<62>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(62)));    
        System.out.println("ds1.datum(<54>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(54)));    

        System.out.println();
        System.out.println("       Round Two");



        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    
        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    

        System.out.println("ds1.datum(<7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(7)));    
        System.out.println("ds1.datum(<8>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(8)));    
        
        System.out.println("ds1.datum(<16>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(16)));    

        System.out.println("ds1.datum(<62>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(62)));    
        System.out.println("ds1.datum(<54>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(54)));    
   
        // Two-D  Tests
        System.out.println();
        System.out.println("Two Dimensional Tests");
        
        ordering=new AxisOrdering(2);
        bounds=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(7,7));
    
        blockSizes=new IntegerIndexSpaceID(4,4); //blocks are 4x4
        //numBlocks=new IntegerIndexSpaceID(2,2);
        
        ds1=new PointOrderBFDS("foo",recordDesc, bounds,ordering,"testdata/file.6.4.bin", ByteOrder.BIG_ENDIAN );
        cds=new UniBlockCacheDataSource("bar",ds1,blockSizes);
        
        cds.activate();

        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));    
        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));  

        System.out.println("ds1.datum(<0,3>) should be {103.1, 103.2, 103.3, 103.4}-->"+cds.datum(new IntegerIndexSpaceID(0,3)));    
        System.out.println("ds1.datum(<0,4>) should be {104.1, 104.2, 104.3, 104.4}-->"+cds.datum(new IntegerIndexSpaceID(0,4)));    

        System.out.println("ds1.datum(<0,7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(0,7)));    
        System.out.println("ds1.datum(<1,0>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(1,0)));    

        System.out.println("ds1.datum(<2,0>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(2,0)));    
        System.out.println("ds1.datum(<3,0>) should be {124.1, 124.2, 124.3, 124.4}-->"+cds.datum(new IntegerIndexSpaceID(3,0)));
        System.out.println("ds1.datum(<4,0>) should be {132.1, 132.2, 132.3, 132.4}-->"+cds.datum(new IntegerIndexSpaceID(4,0)));
        
        System.out.println("ds1.datum(<7,6>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(7,6)));    
        System.out.println("ds1.datum(<6,6>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(6,6)));    
   
        System.out.println();
        System.out.println("       Round Two");



        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));    
        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));  

        System.out.println("ds1.datum(<0,3>) should be {103.1, 103.2, 103.3, 103.4}-->"+cds.datum(new IntegerIndexSpaceID(0,3)));    
        System.out.println("ds1.datum(<0,4>) should be {104.1, 104.2, 104.3, 104.4}-->"+cds.datum(new IntegerIndexSpaceID(0,4)));    

        System.out.println("ds1.datum(<0,7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(0,7)));    
        System.out.println("ds1.datum(<1,0>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(1,0)));    

        System.out.println("ds1.datum(<2,0>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(2,0)));    
        System.out.println("ds1.datum(<3,0>) should be {124.1, 124.2, 124.3, 124.4}-->"+cds.datum(new IntegerIndexSpaceID(3,0)));
        System.out.println("ds1.datum(<4,0>) should be {132.1, 132.2, 132.3, 132.4}-->"+cds.datum(new IntegerIndexSpaceID(4,0)));
        
        System.out.println("ds1.datum(<7,6>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(7,6)));    
        System.out.println("ds1.datum(<6,6>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(6,6)));    
   
    }
		
}
