/* 
    XFDLReader.java
    Author:  Denise D. Mitchell 2002
    University of New Hampshire
    Durham, NH 03824
    
    Modified in 6/03 by Gang Lu, support AxisOrdering
    ---------------- Edit history since 03/14/05 -----------------------
    03/04/12 rdb: Add Map entry processing (completed it, actually)
    02/12/12 rdb: Added handling of block specs instead of using chunk
    02/10/12 rdb: Replalced "obsolete" collection, Vector, with ArrayList
    01/11/12 rdb: added support for sparse datasources: a block mapper that
                    maps a block # (or its "lower" bound) in original file to 
                    a block in the sparse file; intended for use in AR reps.
                  Made sure that no chunk dimensions specified results in 
                    null chunkDimensions, rather than (0); then null field
                    indicates not chunked. 
    01/07/09 pjr: added support for chomboDesiredAttribute
    08/25/08 kjr: added support for metadata (MetadataCollection)
    06/05/08 rdb: made ArrayList and Vectors generic
    01/31/06 dlipsa: Added support for creating chunked data sources through 
           an xfdl file
    11/29/05 rdb: Seems like missing attribute can now come as empty string rather
                    than null reference; added tests for that in process code.
    10/21/05 rdb: removed many blank lines; removed unused variables
                  forced references in main to unused variables we want to keep
*/

package edu.unh.sdb.datasource;

import  edu.unh.sdb.common.*;

import java.util.*;
import java.io.IOException;
import java.nio.*;
import org.w3c.dom.*;

public class XFDLReader extends XMLReader 
{
/** the FDLDescriptor that will hold all data for this FDL */
private FDLDescriptor fdlDescriptor; 

/** interim lower bounds specifications */
private ArrayList<Integer> lowerAxes;

/** interim chunk specifications */
private ArrayList<Integer> chunkDimensions;

/** interim block specifications for sparse data source */
private ArrayList<Integer> blockDimensions;

/** interim upper bounds specifications */
private ArrayList<Integer> upperAxes;

/** interim axis axisOrdering specification */
private AxisOrdering axisOrdering;

/** SparseData lower bounds specifications **/
private ArrayList<Integer> lowerSparseAxes;

/** SparseData upper bounds specifications **/
private ArrayList<Integer> upperSparseAxes;


/** interim map specifications */
private ArrayList<String> mapEntries;

/** holds metadata in a java.util.Properties object in key,value pairs. */
private MetadataCollection metaCollection;

  /** XML version of a File Description Language Reader.  Takes an argument name of a file or 
  URL as a String.  Assumes that verbose mode is off so no debugging messages will be
    displayed*/
/** Used by BinaryFileDataSource, it is the starting point of the file
 *  by Default, it is 0
 */

private long fileOffset;
private int fieldLayout;
private boolean chunkedFileFormat;
private ByteOrder endianness;
private int chomboDesiredAttribute;

//--------------------- Constructors --------------------------------
//
//-------------- XFDLReader --------------------------
/**
 * Create an XFDLReader for the file designated by the argument
 */
  public XFDLReader( String infile) 
  {
    super( infile ); //start the parser with the verbose flag turned off
  }
    
  /** XML version of a File Description Language Reader.  Takes an argument name 
   * of a file or URL as a String and a boolean to turn on/off verbose mode for 
   * debugging
   */
  public XFDLReader( String infile, boolean v) 
  {
    super( infile, v ); //start the parser with the verbose flag argument
  }
  
   /** Instantiates the FDLDescriptor object to hold new parser info*/
  protected void setDefaults()
  {
    this.fdlDescriptor = new FDLDescriptor();
    lowerAxes = new ArrayList<Integer>();
    upperAxes = new ArrayList<Integer>();
    chunkDimensions = new ArrayList<Integer>();
    blockDimensions = new ArrayList<Integer>();
    mapEntries      = new ArrayList<String>();
    lowerSparseAxes = new ArrayList<Integer>();
    upperSparseAxes = new ArrayList<Integer>();
  }
  
  /** main method used for testing the class */
  static public void main( String[] argv )
  {        
    if (argv.length != 1) {
      System.err.println( "Missing fileName." );
      System.exit(1);
      }
        
    try {
      XFDLReader theApp = new XFDLReader(argv[0], true); // instantiate the application
        if ( theApp == null )
           System.out.println( "Got null XFDLReader???" );
    } catch (Exception e)
      {
        e.printStackTrace();
        System.exit(1);
      }
        System.exit(0);
    } 
    
  protected boolean processDoc()
  {        
       
      //Document doc = parser.getDocument();
        // Document is well-formed
            
        // Traverse all the children of the root element
        
      this.fdlDescriptor.setDataFileURL(
                 doc.getDocumentElement().getAttribute( "fileName" ));
      this.fdlDescriptor.setDataFileType
                 (doc.getDocumentElement().getAttribute( "fileType" ));
      this.fdlDescriptor.setRecordSize(
                  sToi(doc.getDocumentElement().getAttribute( "recordSize" )) );
    
      // added by Gang Lu
      processAxisOrdering(doc.getDocumentElement().getAttribute( "axisOrdering" ));
      
      fileOffset = sToi(doc.getDocumentElement().getAttribute( "fileOffset" ));
      this.fdlDescriptor.setFileOffset(fileOffset);
      
      processFieldLayout(
            (doc.getDocumentElement().getAttribute( "fieldLayout" )).trim().toLowerCase());
      this.fdlDescriptor.setFieldLayout(fieldLayout);

      processFileFormat(
            (doc.getDocumentElement().getAttribute( "fileFormat" )).trim().toLowerCase());
      //rdb this.fdlDescriptor.setFileFormat(chunkedFileFormat);      
      
      processByteOrder(
            (doc.getDocumentElement().getAttribute( "byteOrder" )).trim().toLowerCase());
      this.fdlDescriptor.setByteOrder(endianness);

      processChomboDesiredAttribute(
            (doc.getDocumentElement().getAttribute( "chomboDesiredAttribute" )).trim().toLowerCase());
      this.fdlDescriptor.setChomboDesiredAttribute(chomboDesiredAttribute);

     
      for (Node child = doc.getDocumentElement().getFirstChild();
               child != null; child = child.getNextSibling())
      {
        // When child is an element
         if ( child instanceof Element )
         {
            if ( verbose )
               System.err.println( child.getNodeName() );
            if ( child.getNodeName().equals( "Bounds" ) )
            {
               processBoundsPair( child );
            }
            else if ( child.getNodeName().equals( "Field" ) )
            {
               processField( child );
            }
            else if ( child.getNodeName().equals( "Chunk" ) )
            {
               processChunk( child );
            }
            else if ( child.getNodeName().equals( "SparseBlock" ) )
            {
               processSparseBlock( child );
            } else if (child.getNodeName().equals("SparseDataBounds")){
                processSparseDataBounds(child);
            }
            else if ( child.getNodeName().equals( "Metadata" ) )
            {
               processMetadata( child );
            }
            else if ( child.getNodeName().equals( "Map" ) )
            {
               processMap( child );
            }
         }
            
       } // end "for" loop, no more children nodes
       
        processFinalBounds();
        processFinalChunks ();
        processFinalSparseBlocks();
     
        processFinalMetadata();
        processFinalMap();
        processFinalSparseDataBounds();
        
        this.fdlDescriptor.setAxisOrdering(getAxisOrdering());
        if ( verbose)
        {
          System.out.println( "the FDLDescriptor: "
                + getFDLDescriptor().toString());
        }
        
        return true;
}
 
    private void processField(Node n)
    {
       this.fdlDescriptor.addFieldName( 
            makeChildrenText( n.getAttributes().getNamedItem( "fieldName" )) );
       this.fdlDescriptor.addFieldType( 
            makeChildrenText( n.getAttributes().getNamedItem( "fieldType" )) );
    }
    
    private void processBoundsPair(Node n)
    {
      lowerAxes.add(makeChildrenInteger(
            n.getAttributes().getNamedItem( "lower" )));
      upperAxes.add(makeChildrenInteger(
            n.getAttributes().getNamedItem( "upper" )));
    }

    private void processChunk (Node n)
    {
       chunkDimensions.add( makeChildrenInteger( 
                          n.getAttributes().getNamedItem( "dimension" )));
    }

    private void processSparseBlock (Node n)
    {
       blockDimensions.add( makeChildrenInteger( 
                          n.getAttributes().getNamedItem( "dimension" )));
    }
    
    private void processSparseDataBounds(Node n){
    	lowerSparseAxes.add(makeChildrenInteger(
                                                n.getAttributes().getNamedItem( "lower" )));
        upperSparseAxes.add(makeChildrenInteger(
                                                n.getAttributes().getNamedItem( "upper" )));
    	
    }
    
    private void processMetadata(Node n) {

		String name = null;
		if (n.hasAttributes()) {
			Node nm = n.getAttributes().getNamedItem("name");
			if (nm != null)
				name = nm.getNodeValue();
		}

		Metadata m = metaCollection.getMetadata(name);
		if (m == null) {
			m = new Metadata(name);
		}

		try {
			m.load(n.getTextContent());
		} catch (DOMException e) {
			e.printStackTrace();
			return;
		} catch (IOException e) {
			// it is unlikely that this will happen since it is making a stream
			// from a String.
			e.printStackTrace();
			return;
		}

		metaCollection.update(m);
	}

    private void processMap (Node n)
    {      
       mapEntries.add(makeChildrenText(
            n.getAttributes().getNamedItem( "map" )));
    }
    
    private void processFinalBounds()
    {
        this.fdlDescriptor.setBounds( new ISBounds(
            new IntegerIndexSpaceID( vectorToIntArray( lowerAxes ) ), 
            new IntegerIndexSpaceID( vectorToIntArray( upperAxes )) ));
    }
    
    private void processFinalChunks()
    {
       if ( chunkDimensions.size() > 0 )
       {
          this.fdlDescriptor.setChunkDimensions(
              new IntegerIndexSpaceID( vectorToIntArray( chunkDimensions )));
       }
    }
    
    private void processFinalSparseBlocks()
    {
       if ( blockDimensions.size() > 0 )
       {
          int[] blockArray = vectorToIntArray( blockDimensions );
          this.fdlDescriptor.setBlockDimensions(
              new IntegerIndexSpaceID( blockArray ));
       }
    }
    private void processFinalSparseDataBounds(){
        this.fdlDescriptor.setSparseDataBounds( new ISBounds(
                                                   new IntegerIndexSpaceID( vectorToIntArray( lowerSparseAxes ) ),
                                                   new IntegerIndexSpaceID( vectorToIntArray( upperSparseAxes )) ));
    }
    
    private void processFinalMetadata()
    {
		this.fdlDescriptor.setMetadata(metaCollection);
	 }

    private void processFinalMap()
    {
       if ( mapEntries != null )
       {
          String[] mapArray = new String[ mapEntries.size() ];
          for ( int i = 0; i < mapArray.length; i++ )
             mapArray[ i ] = mapEntries.get( i );
          this.fdlDescriptor.setSparseBlockMap( mapArray );
       }
    }
   
    
    /** The AxisOrdering specified at FDL.dtd as an attribute of FileDescriptor
     *  Usage: For an AxisOrdering with three dimensions and the sequence is 2, 1, 0
     *         <FileDescriptor axisOrdering="2 1 0" >
     */
    private void processAxisOrdering( String axisString )
    {
        axisOrdering = null;
        
        if ( axisString != null && !axisString.equals( "") )
        {
            int[] axis=processArrayOfInt(axisString);
            /*
            int len;
            ArrayList axisList = new ArrayList();
            StringTokenizer st= new StringTokenizer(axisString);
        
            while(st.hasMoreTokens())
            {
                axisList.add(st.nextToken());
            }
        
            len = axisList.size();
            if ( len > 0)
            {
            int [] axis = new int[len];
            boolean exceptionOccur=false;
        
            for ( int i=0; i < len; i++)
            {
                try{
                    axis[i] = sToi(((String)axisList.elementAt(i)).trim());
                
                }
                catch(NumberFormatException e)
                {
                    String message = "Error: The value for Axis " + i + " is not an integer: " + axisList.elementAt(i);
                    new XMLReaderError(SDBError.Recoverable, message);
                    exceptionOccur=true;
                    break;
                }
            }
             
            
            */
            //if error occur, use the default axisOrdering
            if ( axis!=null)
                axisOrdering = new AxisOrdering(axis);  
            else
                axisOrdering = null;
        }
    
        else
            axisOrdering = null;         
    }
    
    
        
    private void processFieldLayout( String orderingStr )
    { 
        fieldLayout=DataCollection.pointOrder;
      
        if ( orderingStr != null && !orderingStr.equals( "" ))
        {
            if ( orderingStr.equals( "byattribute" ))
                fieldLayout=DataCollection.attributeOrder;
            else if ( !orderingStr.equals( "bypoint" ))
                new XMLReaderError(SDBError.Recoverable, 
                      "The content for fieldLayout: " 
                      + orderingStr + " is not valid" );
        }
    }

    private void processFileFormat( String fileFormat )
    { 
        chunkedFileFormat = false;
        if ( fileFormat != null && fileFormat.equalsIgnoreCase( "chunk" ))
        {
           System.err.println( "FileFormat: " + fileFormat );
           new XMLReaderError( SDBError.Warning, "Chunk attribute deprecated:"
                                + " Chunking is implied by the Chunk parameters." );
           /*** rdb 01/11/12
            if ( fileFormat.equals( "chunked" ))
                chunkedFileFormat = true;
            else if ( !fileFormat.equals( "native" ))
                new XMLReaderError(SDBError.Recoverable, 
                      "The content for fileFormat: " + fileFormat + 
                      " is not valid" );
            /********/
        }
    }        
     
     private void processByteOrder( String endiannessStr )
    { 
        endianness=ByteOrder.BIG_ENDIAN;
      
        if ( endiannessStr !=null && !endiannessStr.equals( "" ))
        {
            if ( endiannessStr.equals( "little" ))
                endianness=ByteOrder.LITTLE_ENDIAN;
            else if ( !endiannessStr.equals( "big" ))
                new XMLReaderError(SDBError.Recoverable, 
                      "The content for byteOrder: " 
                      + endiannessStr + " is not valid" );
          
        }
    }

   private void processChomboDesiredAttribute( String in )
   { 
        chomboDesiredAttribute=0;
      
        if ( in !=null && !in.equals( "" ))
        {
            chomboDesiredAttribute = sToi(in);
        }
    }

    
  /** returns an FDLDescriptor object containing all of the files attributes */
  public FDLDescriptor getFDLDescriptor()
  {
    return this.fdlDescriptor;
  }
    
  /** Sets the Axis Ordering object*/
  public void setAxisOrdering(AxisOrdering ao)
  {
    axisOrdering = ao;
  }
  
  /** Returns the Axis Ordering object.  If no axis axisOrdering has been requested
    the default order of [0,1, n-1] will be returned.*/
  public AxisOrdering getAxisOrdering()
  {
    if (axisOrdering != null)
      return axisOrdering;
    else
      return createDefaultAxisOrder();
  }
    
  private AxisOrdering createDefaultAxisOrder()
  {
    int len;
    if (getFDLDescriptor() != null)  //Is there a recent FDLDescriptor?
      len = getFDLDescriptor().getDim();
    else 
      len = 1;  //assume dim is 1.
        
    int [] axis = new int[len];
        
    for (int i = 0; i < len; i++)
      axis[i] = i;
        
    return new AxisOrdering(axis);
  }
  
  public long getFileOffset()
  {
      return fileOffset;
  }
  
    public int getFieldLayout()
  {
      return fieldLayout;
  }
    
     public ByteOrder getByteOrder()
  {
      return endianness;
  }
}
