/* 
    GBounds.java
    Copyright 2003 University of New Hampshire
    Durham, NH 03824
    
    This is the array implementation of GBounds.
*/
//------------------- Edit history --------------------------------------
// 10/21/05 rdb: removed unused variables
//-----------------------------------------------------------------------

package edu.unh.sdb.lattice;

import edu.unh.sdb.common.*;
//import java.io.*;


/** The GBounds class is meant to represent n-d hypercubes in 
  *  the data space of a lattice. 
  */
public class GBounds  extends Bounds implements Cloneable{

    protected float [] lower,upper;

    /** This constructor takes two corner points (SpaceIDs)
      *  that define a hypercube of arbitrary dimensionality.
      *   
      *  Notice that the getLower() and getUpper() methods may
      *  or may not return the same Point values given in this
      *  constructor. 
      *
      */
    public GBounds(Point corner1,Point corner2){
        
        int d=corner1.dim();
        int i;
        float lowervalue,uppervalue;
    
        if(d == corner2.dim()){
        
            this.lower= corner1.getCoords().clone(); 
            this.upper= corner2.getCoords().clone(); 
            
            
            for(i=0;i<d;i++){
                
                lowervalue=this.lower[i];
                uppervalue=this.upper[i];
            
                if(lowervalue > uppervalue){
                
                    this.lower[i]=uppervalue;
                    this.upper[i]=lowervalue;
                
                } 
// Uncommenting the lines below would generate an exception when we
// try to generate an Bounds with zero length in some dimension.
/*                else {
                    if (lowervalue==uppervalue) {
                        throw new BoundsException("Can't create Bounds with zero extent in some dimension.");
                    }
                }
*/                
            }
        } else {
            new DimensionalityError(SDBError.Recoverable, "Can't create GBounds with two corners of different dimensionality.");
        }
            
    }


    /** This constructor takes a single Point
      * that defines the dimensions of a hypercube with
	  *	lower coordinates at the origin. This constructor
	  *	 takes the absolute value of all coordinates in the argument.
      */
    public GBounds(Point dimensions){
        
        int d=dimensions.dim();
        int i;
        
    
		this.lower = new float[d];
		this.upper = dimensions.getCoords(new float[d]);
            
		for(i=0;i<d;i++){
			this.lower[i]=0;
			this.upper[i]=Math.abs(this.upper[i]);
		}
            
    }

    /** This constructor takes a single float array
      * that defines the dimensions of a hypercube with
      * lower coordinates at the origin. This constructor 	  
      *	takes the absolute value of all coordinates in the argument.
      */
    public GBounds(float [] dimensions){
        
        int d=dimensions.length;
        int i;
        
    
		this.lower = new float[d];
		this.upper = new float[d];
            
		for(i=0;i<d;i++){
			this.lower[i]=0;
			this.upper[i]=Math.abs(dimensions[i]);
		}
            
    }




    /** This constructor takes two corner points expressed as int arrays
      *  that define a hypercube of arbitrary dimensionality.
      *  
      *  The arrays may be used directly in the constructed GBounds. That
      *  is, GBounds may choose not to clone the arrays.
      *   
      *  Notice that the getLower() and getUpper() methods may
      *  or may not return the same Point values given in this
      *  constructor. 
      *  
      *  In the interest of speed, no explicit check of the array lengths is
      *  made. If the first array is longer than the second, an error
      *  will be created, but not if the second array is longer than the
      *  second.
      *  
      */
    GBounds(float [] corner1,float [] corner2){
        
        int d=corner1.length;
        int i;
        float lowervalue,uppervalue;

        try{
            for(i=0;i<d;i++){
                
                lowervalue = corner1[i];
                uppervalue = corner2[i];
            
                if(lowervalue > uppervalue){
                    corner1[i] = uppervalue;
                    corner2[i] = lowervalue;
                } 
            }

            this.lower=corner1;
            this.upper=corner2;
            
        } catch (ArrayIndexOutOfBoundsException e) {
            new DimensionalityError(SDBError.Recoverable, "Can't create GBounds with two corners of different dimensionality.");
        }
    }


    /** Construct a copy of the argument bounds. 
    */
    public GBounds(GBounds bounds){
        
        this.lower=bounds.lower.clone();
        this.upper=bounds.upper.clone();
    }

    /** This constructor just allocates space for the lower and upper arrays.*/
    protected GBounds(int dimensionality){
    
        this.lower = new float[dimensionality];
        this.upper = new float[dimensionality];
    }

    
    
    /**  Create a new GBounds with lower bounds at the origin, and
     *   upper bounds with the specified Point.
     *   
     *   This implementation assumes that a freshly created array of ints
     *   will be initialized to zero.
     */
/*    public static GBounds newBoundsAtOrigin(Point id){

        // This implementation assumes that a freshly created array of ints will be initialized to zero. 
        return new GBounds(new float[id.dim()] ,(float [])(id.getCoords().clone()));
    }
*/ 
    
    /** Create a new GBounds with lower bounds at the origin, and
      *  upper bounds with coordinates corresponding
      *  to the integer array argument. 
      *  
      *  This method may elect to use the array argument directly, 
      *  without making a copy.
      *  
      *  This implementation assumes that a freshly created array of ints
      *  will be initialized to zero.
      */
/*     static GBounds newBoundsAtOrigin(float [] coords){

        return new GBounds(new float[coords.length] ,coords);
    }
*/          
    
    /** Return a copy of this GBounds object
    */
    public final GBounds copy(){
    
        return new GBounds( this.lower.clone(), this.upper.clone());
    }

    /** Return a copy of this GBounds object. Here, this method is the same as
	 *	copy, but child classes use this method to allow the copying of
	 *	just the bounds information.
     */
    public GBounds copyBounds(){
    
        return new GBounds( this.lower.clone(), this.upper.clone());
    }


    
    /** Return a copy of this GBounds object
    */
    public Object clone(){
        
        return new GBounds( this.lower.clone(), this.upper.clone());
    }
    
    /** Set the lower and upper bounds of this object to refer
     *   to the lower and upper bounds of the parameter. No 
     *   copy of the bounds are made.
     */
    final void assign(GBounds b){
    
        this.lower=b.lower;
        this.upper=b.upper;
    }
     
    /** Set the lower and upper bounds of this object to refer
      *  to the argument arrays. No copy of the arrays are made.
      *  
      *  In the interest of speed, no dimensionality checking is done,
      *  so make sure that the argument arrays have the same length.
      *  
      *  Package access only.
    */    
    final void assign(float [] lower, float[] upper){

        this.lower=lower;
        this.upper=upper;
    }

    /** Set the lower bound of this object to refer
      *  to the parameter. If the Point contains
      *  an integer array, the GBounds may 
      *  refer directly to that array.
      *  
      *  In the interest of speed, no dimensionality checking is done,
      *  so make sure that the argument has a length matching
      *  the dimensionality of the upper bounds.
      *          
      *  Package access only.
      */
    final void assignLower(Point lower){ 
    
        this.lower=lower.getCoords();
    }
    
    /** Set the upper bound of this object to refer
     *   to the parameter. If the Point contains
     *   an integer array, the GBounds may 
     *   refer directly to that array.
     *   
     *   In the interest of speed, no dimensionality checking is done,
     *   so make sure that the argument has a length matching
     *   the dimensionality of the lower bounds.
     *
     *   Package access only.        
     *
     */
    final void assignUpper(Point upper){
    
        this.upper=upper.getCoords();
    }
 
    /** Set the lower bound of this object to refer
      *  to the parameter. No copy of the array arguments are made.
      *  
      *  In the interest of speed, no dimensionality checking is done,
      *  so make sure that the argument array has a length matching
      *  the dimensionality of the upper bounds.
      *  
      *  Package access only.        
      */
    final void assignLower(float [] lower){ 
    
        this.lower=lower;
    }
    
    /** Set the upper bound of this object to refer
      *  to the parameter. No copy of the array arguments are made.
      *  
      *  In the interest of speed, no dimensionality checking is done,
      *  so make sure that the argument array has a length matching
      *  the dimensionality of the lower bounds.
      *  
      *  Package access only.        
      *  
      */
    final void assignUpper(float [] upper){
    
        this.upper=upper;
    }

    /** Set the lower bounds to the coordinates represented
      *  by the Point argument.
      */
    public final void setLower(Point lower){ 
    
        if (this.lower.length == lower.dim()){
        
            this.lower=lower.getCoords().clone();
            
        } else {
        
            new DimensionalityError(SDBError.Recoverable, "Argument has wrong dimensionality.");
        }
    }
    
    /** Set the upper bounds to the coordinates represented
      *  by the Point argument.
      */
    public final void setUpper(Point upper){

        if (this.lower.length == upper.dim()){
        
        	this.upper= upper.getCoords().clone();
        } else {
        
            new DimensionalityError(SDBError.Recoverable, "Argument has wrong dimensionality.");
        }
        
    }

    /** Set the lower bounds to the coordinates represented
      *  by the integer array argument.
      */
    public final void setLower(float [] lower){ 

        if (this.lower.length == lower.length){
        
            this.lower= lower.clone();
        } else {
        
            new DimensionalityError(SDBError.Recoverable, "Argument has wrong dimensionality.");
        }
        
    }
    
    /**  Set the upper bounds to the coordinates represented
      *  by the integer array argument.
      */
    public final void setUpper(float [] upper){

        if (this.lower.length == upper.length){
        
            this.upper= upper.clone();
        } else {
        
            new DimensionalityError(SDBError.Recoverable, "Argument has wrong dimensionality.");
        }

    }
       
             
    /** Return an Point representing the upper bounds.*/
    public final Point getLower(){
    
        return new Point(this.lower);
    }
    
    
    /** Return an Point representing the lower bounds.*/
    public final Point getUpper(){
    
        return  new Point(this.upper);
    }

    /**  Return a reference to the integer array
      *  representing the upper bounds.
      */
    final float [] getLowerArray(){
    
        return this.lower;
    }

    /** Copy the lower bound coordinates into the
	  *	array argument. A reference to the argument
	  *	is returned. 
	  */
    public final float [] getLowerArray(float [] arr){
    
		int i, d=this.lower.length;
		
		for(i=0; i<d; i++){
		
			arr[i] = this.lower[i];
		}
	
        return arr;
    }
    
    
    /** Return a reference to the integer array
      *  representing the upper bounds.
      */
    final float [] getUpperArray(){
    
        return  this.upper;
    }

    /** Copy the upper bound coordinates into the
	  *	array argument. A reference to the argument
	  *	is returned. 
	  */
    public final float [] getUpperArray(float [] arr){
    
		int i, d=this.lower.length;
		
		for(i=0; i<d; i++){
		
			arr[i] = this.upper[i];
		}
	
        return arr;
    }
	

    
    
    /** Return dimensionality of Bounds. */
    public final int dim(){
    
        return this.lower.length;
    }

        
    /** Return true if parameter is equal to this object, false otherwise.*/
    public boolean equals(GBounds b){
    
        int i, d=this.dim();
        
        try{
            for(i=0; i<d; i++){
                
                if(this.lower[i] != b.lower[i] || this.upper[i] != b.upper[i]){
                    return false;
                }
            }
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
            return false;
        }

        
        return true;
    }
    
    
    /** Return the total volume of the GBounds as an integer.
      *  "Volume" is defined to be the total number of elements
      *  within the dataspace enclosed by the GBounds.  
      */
    public final float volume(){
        
        int d=this.lower.length;
        int i;
		float volume=1.0f;
        float [] dimensions=this.getDimensions(new float[d]);
        
        for(i=0;i<d;i++){
            volume*=dimensions[i];
        }
        
        return volume;
    }



    /** Return a Point with axis values denoting the size
      *  of each corresponding dimension of the GBounds.
      */
    public final Point getDimensions(){
        
        return new Point(this.getDimensions(new float[this.lower.length]));
    }


    /** Fill the argument array with coordinate values denoting the size
      *  of each corresponding dimension of the GBounds. A reference to the argument
      *  array is returned as a convenience.
      *  
      *  The array argument should have a length corresponding to the
      *  dimensionality of the bounds object. However, if the argument
      *  array is too long, the extra elements will be ignored. If it is
      *  too short,  a DimensionalityError will will be created.
      *
      */
    public final float [] getDimensions(float[] coords){
    

        int i,d=lower.length;
        
        try {
            
            for(i=0;i<d;i++){
                coords[i] = this.upper[i] - this.lower[i];
            }
            
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal, "Argument array has too few dimensions.");
        }
        
        return coords; 
            
    }

    /** Return the extent of this GBounds along the specified axis.*/
    public final float getDimension(int axis){
        
        try {

            return this.upper[axis] - this.lower[axis];
            
        } catch (ArrayIndexOutOfBoundsException e){
        
            new DimensionalityError(SDBError.Fatal, "Axis argument array is too large.");
            return -1;
        }
    }




    /** Translate the GBounds object 
      *  in space by the value of the parameter Point. Each
      *  coordinate of the Point is added to the coordinates of the
      *  upper and lower bounds of this GBounds.
      *          
      *  This is a *relative* move.
      */
    public final void move(Point distance){

        this.move(distance.getCoords());
    }


    /** Translate the GBounds object 
      *  in space by the coordinates in the integer array argument. Each
      *  coordinate is added to the coordinates of the
      *  upper and lower bounds of this GBounds.
      *  
      *  This is a *relative* move.      
      *    
      *  The array argument should have a length corresponding to the
      *  dimensionality of the bounds object. However, if the argument
      *  array is too long, the extra elements will be ignored. If it is
      *  too short,  a DimensionalityError will will be created.
      */
    public final void move(float [] distance){
        
        int i, d=this.lower.length;
        
        try{
            
            for(i=0; i<d ; i++){
                
                this.lower[i]+=distance[i];
                this.upper[i]+=distance[i];
            }
            
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Argument array has too few dimensions.");
        }

    }



    /** Translate the GBounds object in space so that the
      *  lower left corner is moved to the location specified 
      *  by the parameter Point. 
      *  
      *  This is an *absolute* move.
      *  
      *  The Point should have the same dimensionality as the bounds
      *  object. However, if the argument array is too long, the extra 
      *  elements will be ignored. If it is too short,
      *  a DimensionalityError will will be created.
      *  
      */
    public void moveTo(Point location){

        moveTo(location.getCoords());
    }
    
    

    /** Translate the GBounds object in space so that the
      *  lower left corner is moved to the coordinates specified 
      *  int the integer array argument. 
      *  
      *  This is an *absolute* move.
      *  
      *  The array argument should have a length corresponding to the
      *  dimensionality of the bounds object. However, if the argument
      *  array is too long, the extra elements will be ignored. If it is
      *  too short, a DimensionalityError will be created.
      *  
      */
    public void moveTo(float[] location){

        int i, d=this.lower.length;

        try{
            for(i=0; i< d; i++){
            
                this.upper[i] += (location[i] - lower[i]);
                this.lower[i] = location[i];
            }
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Argument array has too few dimensions.");
        }

    }




    
    /** Translate this GBounds in space so that the lower
     *    point is now at the origin.
     */
    public void moveToOrigin(){
    
        
        int i, d=this.lower.length;
    
        for(i=0; i<d; i++){
        
            this.upper[i] -= this.lower[i];
            this.lower[i]=0;
        }
    }

	/** Returns a new GBounds of the same shape as the argument
	 *	bounds, but moved to the origin.
	 */
	public static GBounds copyAndMoveToOrigin(GBounds bounds){
	
        int i, d=bounds.lower.length;
		GBounds result=new GBounds(d);
		
        for(i=0; i<d; i++){
        
            result.upper[i] = bounds.upper[i] - bounds.lower[i];
            result.lower[i]=0;
        }
		
		return result;
	}

    /** Updates this GBounds object to be the 
     *   bounding volume of this GBounds and the parameter
     *   GBounds. If the dimensionality of the parameter GBounds
     *   does not match the dimensionality of this GBounds, then
     *   a DimensionalityException is thrown.
     */
    public void boundingVolume(GBounds db){
        
        int i,d=this.lower.length;
    
        try{          
            for(i=0;i<d;i++){ // for each axis
            
                // Only change the lower value for the current axis
                // if the value from db is smaller.
                this.lower[i] = (db.lower[i] < this.lower[i]) ? db.lower[i] : this.lower[i];
                                                

                // Only change the upper value for the current axis
                // if the value from db is larger.
                this.upper[i] = (db.upper[i] > this.upper[i]) ? db.upper[i] : this.upper[i];
                
            }
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
        }
    }


    /** Updates this object to be the intersecton of itself with the GBounds argument. If the intersection
      *  exists the method returns true, and false otherwise. This object is unchanged.
      */
    public boolean intersects(GBounds b2){
    
        int i,d=this.lower.length;
                
        try{
            for(i=0;i<d;i++){
                
                if( (this.lower[i] < b2.lower[i] && this.upper[i] < b2.upper[i]) || 
                    (b2.lower[i] < this.lower[i] && b2.upper[i] < this.upper[i])){
                    
                    return false;
                }
            }    
        
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
            return false;
        }
        
        return true; // The bounds intersect for all dimensions, so return true.
    }




    /** Returns the intersecton of this object with the GBounds argument. If the intersection
      *  doesn't exist, then return null. 
      */
    GBounds intersection(GBounds b2){
    
        int i,d=this.lower.length;
        
        float max[]=new float[d];
        float min[]=new float[d];
     
        try{
            for(i=0;i<d;i++){
                
                if( (this.lower[i] < b2.lower[i] && this.upper[i] < b2.lower[i]) 
                    || (b2.lower[i] < this.lower[i] && b2.upper[i] < this.lower[i]) 
                ){
                    
                    return null;
                    
                } else {
    
                    max[i] = ( this.lower[i] > b2.lower[i] ) ? this.lower[i] : b2.lower[i];
                    min[i] = ( this.upper[i] > b2.upper[i] ) ? b2.upper[i] : this.upper[i];
                }
                
            }  
        
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
            return null;
        }

        
        
        return new GBounds(max,min);
    }


    /** Returns the intersecton of this object with the GBounds argument. If the intersection
      *  doesn't exist, then return null. This form of the intersection method takes two integer
      *  arrays as arguments, which it uses as internal "scratch" space. It does not allocate
      *  any new objects if no intersection is found. These arrays must have length equal to the dimensionality of the 
      *  bounds, but their contents are ignored and overwritten. Of the two intersection methods,
      *  this one is about half as fast as the other with a non-null intersection, but is
      *  about twice as fast as the other with a null intersection.
      */
    GBounds intersection(GBounds b2, float max[], float min[]){
    
        int i,d=this.lower.length;
        
     
        try{
            for(i=0;i<d;i++){
                
                if( (this.lower[i] < b2.lower[i] && this.upper[i] < b2.lower[i]) 
                    || (b2.lower[i] < this.lower[i] && b2.upper[i] < this.lower[i]) 
                ){
                    
                    return null;
                    
                } else {
    
                    max[i] = ( this.lower[i] > b2.lower[i] ) ? this.lower[i] : b2.lower[i];
                    min[i] = ( this.upper[i] > b2.upper[i] ) ? b2.upper[i] : this.upper[i];
                }
                
            }    
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
            return null;
        }

        
        return new GBounds( max.clone(), min.clone());
    }

/*
    public boolean intersects(TriangleCell c){
    
        Point[] cellVertices=c.getVertexCoords();
        int i,j;
        byte [] outcodes=new byte[3];
    
        if(this.lower.length!=2){
            new IllegalArgumentError(SDBError.Fatal,"GBounds.intersects(): GBounds is not 2D.");
        }

        // We could make this faster with a single function call,
        // which avoids getting the bounds axis values repeatedly.
        for(i=0;i<3;i++){
            outcodes[i]=outcode(cellVertices[i]);
        }

        byte oc1, oc2, ocOR;
        int numClearlyOutside=0;
        
        for(i=0;i<3;i++){
        
            j=(i+1)%3;
            
            oc1=outcodes[i];
            oc2=outcodes[j];
            
            // Trivially accept triangles with vertices inside the
            // rectangle
            if(oc1==0 || oc2==0){
                return true;
            }
        
            // Test for lines that cross straight through from 
            // left to right, or top to bottom.
            ocOR=(byte) (oc1 | oc2);
            if(ocOR==0x03 || ocOR==0x0C){
                return true;
            }
            
            // Count the number of edges that are clearly
            // outside the rectangle
            if((oc1 & oc2) !=0) {
                numClearlyOutside++;
            } else {
                // The edge is not clearly outside, so we'll have to do
                // some line intersections. 
            
                // If one vertex has the center left outcode, then we
                // can try intersecting the edge with the left rectangle 
                // boundary.
                if(oc1 == 0x01 || oc2 ==0x01) { // left edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(this.lower),
                                    new Point(this.lower[0],this.upper[1]),
                                    vertical)
                    ){
                        return true;
                    }
                } else if(oc1==0x02 || oc2==0x02){ // right edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(this.upper[0],this.lower[1]),
                                    new Point(this.upper),
                                    vertical)
                    ){
                        return true;
                    }
                } else if(oc1==0x04 || oc2==0x04){ // bottom edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(this.lower),
                                    new Point(this.upper[0],this.lower[1]),
                                    horizontal)
                    ){
                        return true;
                    }
                } else if(oc1==0x08 || oc2==0x08){ // top edge
                    if(intersection_helper(	cellVertices[i],
                                    cellVertices[j],
                                    new Point(this.lower[0],this.upper[1]),
                                    new Point(this.upper),
                                    horizontal)
                    ){
                        return true;
                    }
                } else if(ocOR==0x0F) { //diagonals
                
                    // Here, we handle the case where the vertices lie in opposite 
                    // corner outcodes.  First, we'll see if the edge intersects with 
					// the left side.
					
					
					if(intersection_helper(	cellVertices[i],
									cellVertices[j],
									new Point(this.lower),
									new Point(this.lower[0], this.upper[1]),
									vertical)						
					){
						return true;
					}
					
					// Now, we have to see which of the two kinds of diagonals we have,
					// and either test the top or bottom sides of the rectangle.
					
					if (oc1==0x09 || oc2==0x09){
						// The line must go from upper left to lower right,
						// so we'll test the top side.
						if(intersection_helper( cellVertices[i],
										cellVertices[j],
										new Point(this.lower[0], this.upper[1]),
										new Point(this.upper),
										horizontal)
							
						){
							return true;
						}
					} else {
						// The line must go from lower left to upper right,
						// so we'll test the bottom side.
						if(intersection_helper( cellVertices[i],
										cellVertices[j],
										new Point(this.lower),
										new Point(this.upper[0], this.lower[1]),
										horizontal)
							
						){
							return true;
						}
					}
					
				
                }
            } //if
        }


        if(numClearlyOutside==3){ 
            return false;
        } else {
            //return inside(this.lower,cellVertices);
            return c.inside(new Point(this.lower));

        }

    }
*/    
    /*  intersection_helper()
    
        This private helper function will return true if the line v1-v2 intersects the line b1-b2, and 
        false otherwise. The last argument specifies whether b1-b2 is horizontal or vertical, using
        the contants defined below.
          It is important that b1 < b2 for the axis being checked. If not, incorrect results will be 
        generated. Also, b1-b2 is meant to be one side of a GBounds, so it is currently assumed to be
        either horizontal or vertical, depending on the last argument. However, v1-v2 can have any
        orientation. 
    */
    /*
    private static final int vertical=0;
    private static final int horizontal=1;
    
    private final boolean intersection_helper(Point v1, Point v2, Point b1, Point b2,int vh){
    
        float t;
        
		
        if(vh==vertical){
            float x0=v1.getCoord(0);
            float x1=v2.getCoord(0);
                    
            t=(b1.getCoord(0)-x0)/(x1-x0);
            
            if( t < 0 || t > 1){
				System.out.println("Foo");
                return false;
            } else {
                float y0=v1.getCoord(1);
                float y1=v2.getCoord(1);
                float y=y0+t*(y1-y0);
				
                if(y < b1.getCoord(1) || y > b2.getCoord(1)){ // Problem here // seems alright
                    return false;
                } else {
                    return true;
                }
            }
        } else { //Horizontal

            float y0=v1.getCoord(1);
            float y1=v2.getCoord(1);
                    
            t=(b1.getCoord(1)-y0)/(y1-y0);
            
            if( t < 0 || t > 1 ){
				System.out.println("Foo");
                return false;
            } else {
                float x0=v1.getCoord(0);
                float x1=v2.getCoord(0);
                float x=x0+t*(x1-x0);
            
                if(x < b1.getCoord(0) || x > b2.getCoord(0)){
                    return false;
                } else {
                    return true;
                }
            }
        
        
        }
		
    }
*/ 
 
	/* 	1001	1000	1010
		0001	0000	0010
		0101	0100	0110
	*/

/*     
    private final byte outcode(Point cellVertex){
    
        float x=cellVertex.getCoord(0);
        float y=cellVertex.getCoord(1);
            
        return (byte)(((y > this.upper[1]?1:0) << 3) |
                      ((y < this.lower[1]?1:0) << 2) |  
                      ((x > this.upper[0]?1:0) << 1) |
                      ((x < this.lower[0]?1:0)     ) );
    }

*/



    /** Returns true if this Bounds completely contains the argument bounds. */
    public boolean contains(GBounds b2){
    
        int i,d=this.lower.length;
       
        try{
            for(i=0;i<d;i++){
                
                if( ! ( this.lower[i] <= b2.lower[i] && this.upper[i] >= b2.upper[i]) ){        
                    return false;            
                }
            }    
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Dimensionality of Bounds argument does not match this Bounds object.");
            return false;
        }

        return true;
    }


    /** Returns true if this Bounds contains the argument Point. */
    public boolean contains(Point id){

        return this.contains(id.getCoords());
    }
    
    /** Returns true if this Bounds contains the coordinates contained in the integer array.
      *
      *  The array argument should have a length corresponding to the
      *  dimensionality of the bounds object. However, if the argument
      *  array is too long, the extra elements will be ignored. If it is
      *  too short,  a DimensionalityError will will be created.
      */
    public boolean contains(float [] coords){
    
        int i, d=this.lower.length;

        try{
            for(i=0; i<d; i++){
                
                if ( (coords[i] < this.lower[i] ) ||
                    (coords[i] > this.upper[i] ) 
                ) {
                    
                    return false;
                }
            }
        } catch (ArrayIndexOutOfBoundsException e){
            new DimensionalityError(SDBError.Recoverable, "Argument array has too few dimensions.");
            return false;
        }

        return true;
    }
    
    
    
    /** Return a String representing the object. */
    public String toString(){
        
        return this.toString("");
    }

    
    
    /** Return a String representing the object indented by the argument string. */
    public String toString(String indent){
    
        int i, d=this.lower.length;
        StringBuffer result=new StringBuffer(indent+"{GBounds: l=<");
        

        for(i=0; i<d-1; i++){
        
            result.append(lower[i]+", ");
        }
        
        result.append(lower[i]+">, u=<");

        for(i=0; i<d-1; i++){
        
            result.append(upper[i]+", ");
        }
        
        result.append(upper[i]+">}");

        return new String(result);
    }




    /** For debugging. Runs basic tests on the GBounds class. */
    static public void main(String[] args){
    
        System.out.println("\n**********************\n***GBounds Tests***\n**********************");
    
        Point temp;
        
        Point id1=new Point(0);
        Point id2=new Point(7);
        
        Point id3=new Point(1,1);
        Point id4=new Point(6,6);
        
        Point id5=new Point(2,3);
        Point id6=new Point(0,7);
        
        Point id7=new Point(1,2,3);
        Point id8=new Point(4,1,7);
        Point id9=new Point(8,1,9);
        
        Point id10=new Point(3);
        
        System.out.println("\n\n***GBounds Constructor Tests***");
        
        GBounds db1=new GBounds(id1,id2);
        GBounds db2=new GBounds(id3,id4);
        GBounds db3=new GBounds(id5,id6);
        GBounds db4=new GBounds(id7,id8);
        GBounds db5=new GBounds(new Point(1,1,1), new Point(4,5,6));
        GBounds db6=new GBounds(new Point(1,-1,0), new Point(6,5,4));
        GBounds theCopy;
    
        System.out.print(" 1D GBounds: Should be {(0),(7)}-->"+db1); 
        System.out.print("\n 2D GBounds: Should be {(1,1),(6,6)}-->"+db2); 
        System.out.print("\n 2D GBounds with coord swapping: Should be {(0,3),(2,7)}-->"+db3);
        System.out.print("\n 3D GBounds with coord swapping: Should be {(1,1,3),(4,2,7)}-->"+db4);
        System.out.println();
        
        try {
            GBounds foo=new GBounds(id1,id3);
            System.out.println(" ***Error: A DimensionalityException should have been"
                               + " thrown due to mismatched constructor arguments.: "
                               + foo );
        } catch (DimensionalityException e){
            System.out.println(" DimensionalityException correctly thrown due to mismatched constructor arguments.");
        }
        
        /*
        try {
            GBounds bar=new GBounds(id8,id9);
            System.out.println(" ***Error: A BoundsException should have been thrown due to a degenerate volume definition.");
        } catch (BoundsException e){
            System.out.println(" BoundsException correctly thrown due to a degenerate volume definition.");
        }
        */
        
        System.out.println("\n\n***Copy Tests***");
        
        theCopy=(GBounds) db1.clone();
        System.out.print(" 1D GBounds clone(): Should be {(0),(7)}-->"+theCopy); 
        theCopy=(GBounds) db2.clone();
        System.out.print("\n 2D GBounds clone(): Should be {(1,1),(6,6)}-->"+theCopy);
        theCopy=(GBounds) db4.clone();
        System.out.print("\n 3D GBounds clone(): Should be {(1,1,3),(4,2,7)}-->"+theCopy);
        

        System.out.println("\n\n***getLower Tests***");

        temp=db1.getLower();
        System.out.print(" 1D GBounds getLower(): Should be (0)-->"+temp); 
        temp=db2.getLower();
        System.out.print("\n 2D GBounds getLower(): Should be (1,1)}-->"+temp); 
        temp=db4.getLower();
        System.out.print("\n 3D GBounds getLower(): Should be (1,1,3)-->"+temp); 

        System.out.println("\n\n***getUpper Tests***");

        temp=db1.getUpper();
        System.out.print(" 1D GBounds getUpper(): Should be (7)-->"+temp);
        temp=db2.getUpper();
        System.out.print("\n 2D GBounds getUpper(): Should be (6,6)-->"+temp);
        temp=db4.getUpper();
        System.out.print("\n 3D GBounds getUpper(): Should be (4,2,7)-->"+temp);


        System.out.println("\n\n***dim Tests***");

        System.out.print(" 1D GBounds dim(): Should be 1-->"+db1.dim()); 
        System.out.print("\n 2D GBounds dim(): Should be 2-->"+db2.dim()); 
        System.out.print("\n 3D GBounds dim(): Should be 3-->"+db4.dim()); 


        System.out.println("\n\n***volume Tests***");

        System.out.print(" 1D GBounds volume(): Should be 8-->"+db1.volume()); 
        System.out.print("\n 2D GBounds volume(): Should be 36-->"+db2.volume()); 
        System.out.print("\n 3D GBounds volume(): Should be 40-->"+db4.volume()); 


        System.out.println("\n\n***getDimensions Tests***");

        System.out.print(" 1D GBounds getDimensions(): Should be (8)-->"+db1.getDimensions()); 
        System.out.print("\n 2D GBounds getDimensions(): Should be (6,6)-->"+db2.getDimensions()); 
        System.out.print("\n 3D GBounds getDimensions(): Should be (4,2,5)-->"+db4.getDimensions()); 


        System.out.println("\n\n***move Tests***");
        db1.move(id10);
        System.out.print(" 1D GBounds move(): Should be {(3),(10)}-->"+db1);
        db3.move(id5);
        System.out.print("\n 2D GBounds move(): Should be {(2,6),(4,10)}-->"+db3);
        db4.move(id9);
        System.out.print("\n 3D GBounds move(): Should be {(9,2,12),(12,3,16)}-->"+db4);
        id8.negate();
        db4.move(id8);
        System.out.print("\n 3D GBounds move() in negative direction: Should be {(5,1,5),(8,2,9)}-->"+db4);
        db4.move(id8);
        System.out.print("\n 3D GBounds move() in negative direction: Should be {(1,0,-2),(4,1,2)}-->"+db4);


        System.out.println("\n\n***moveTo Tests***");
        db1.move(id10);
        System.out.print(" 1D GBounds move(): Should be {(6),(13)}-->"+db1);


        db3.move(id5);
        System.out.print("\n 2D GBounds move(): Should be {(4,9),(6,13)}-->"+db3);
        
        db4.move(id9);
        System.out.print("\n 3D GBounds move(): Should be {(9,1,7),(12,2,11)}-->"+db4);



        System.out.println("\n\n***BoundingVolume Tests***");
        
        db2.boundingVolume(db3);
        
        System.out.println(" 2D BoundingVolume Test: Should be {(1,1),(6,13)}-->"+db2);
        
        theCopy=(GBounds) db5.clone();
        theCopy.boundingVolume(db6);
        System.out.println(" 3D BoundingVolume Test: Should be {(1,-1,0),(6,5,6)}-->"+theCopy);
        
        
        System.out.println("\n\n***Intersection Tests***");
        
        System.out.println("Using public boolean intersection(GBounds b2)");
        
        db2.intersection(db3);
        
        System.out.println(" 2D Intersection Test: Should be {(4,9),(6,13)}-->"+db2);
        
        
        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(10,10));
                
        theCopy.intersection(db2);
        System.out.println(" 2D Intersection Test: Should be {(0,0),(0,0)}-->"+theCopy);
        
        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(1,2));
                
        theCopy.intersection(db2);
        System.out.println(" 2D Intersection Test: Should be {(5,11),(6,13)}-->"+theCopy);
        
        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(-1,-2));
                
        theCopy.intersection(db2);
        System.out.println(" 2D Intersection Test: Should be {(4,9),(5,11)}-->"+theCopy);
        
        theCopy=new GBounds(new Point(1,1),new Point(2,2));
        GBounds db7 = new GBounds(new Point(0,3),new Point(0,3));
        
        db7.intersection(theCopy);
        System.out.println(" 2D Intersection Test: Should be {(0,0),(0,0)}-->"+db7);
        
        db7=new GBounds(new Point(0,0),new Point(3,3));
        theCopy.intersection(db7);
        System.out.println(" 2D Intersection Test: Should be {(1,1),(2,2)}-->"+theCopy);
        
        
        theCopy=(GBounds) db6.clone();
                
        theCopy.intersection(db5);
        System.out.println(" 3D Intersection Test: Should be {(1,1,1),(4,5,4)}-->"+theCopy);


        /************/
        System.out.println("Using GBounds intersection2(GBounds b2)");
        
        db2=new GBounds(new Point(4,9),new Point(6,13));
        db3=new GBounds(new Point(4,9),new Point(6,13));
        db5=new GBounds(new Point(1,1,1), new Point(4,5,6));
        db6=new GBounds(new Point(1,-1,0), new Point(6,5,4));

        
        GBounds result=db2.intersection(db3);
        
        System.out.println(" 2D Intersection Test: Should be {(4,9),(6,13)}-->"+result);

        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(10,10));
                
        result=theCopy.intersection(db2);
        System.out.println(" 2D Intersection Test: Should be null-->"+result);
        
        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(1,2));
                
        result=theCopy.intersection(db2);
        System.out.println(" 2D Intersection Test: Should be {(5,11),(6,13)}-->"+result);
        
        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(-1,-2));
                
        result=theCopy.intersection(db2);
        System.out.println(" 2D Intersection Test: Should be {(4,9),(5,11)}-->"+result);
        
        theCopy=new GBounds(new Point(1,1),new Point(2,2));
        db7 = new GBounds(new Point(0,3),new Point(0,3));
        
        result=db7.intersection(theCopy);
        System.out.println(" 2D Intersection Test: Should be null-->"+result);
        
        db7=new GBounds(new Point(0,0),new Point(3,3));
        result=theCopy.intersection(db7);
        System.out.println(" 2D Intersection Test: Should be {(1,1),(2,2)}-->"+result);
        
        
        theCopy=(GBounds) db6.clone();
                
        result=theCopy.intersection(db5);
        System.out.println(" 3D Intersection Test: Should be {(1,1,1),(4,5,4)}-->"+result);
        /*************/
        System.out.println("Using GBounds intersection3(GBounds b2, int max[], int min[])");
        
        float [] min2 = new float[2];
        float [] max2 = new float[2];
        
        db2=new GBounds(new Point(4,9),new Point(6,13));
        db3=new GBounds(new Point(4,9),new Point(6,13));
        db5=new GBounds(new Point(1,1,1), new Point(4,5,6));
        db6=new GBounds(new Point(1,-1,0), new Point(6,5,4));

        
        result=db2.intersection(db3,min2,max2);
        
        System.out.println(" 2D Intersection Test: Should be {(4,9),(6,13)}-->"+result);

        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(10,10));
                
        result=theCopy.intersection(db2,min2,max2);
        System.out.println(" 2D Intersection Test: Should be null-->"+result);
        
        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(1,2));
                
        result=theCopy.intersection(db2,min2,max2);
        System.out.println(" 2D Intersection Test: Should be {(5,11),(6,13)}-->"+result);
        
        theCopy=(GBounds) db2.clone();
        theCopy.move(new Point(-1,-2));
                
        result=theCopy.intersection(db2,min2,max2);
        System.out.println(" 2D Intersection Test: Should be {(4,9),(5,11)}-->"+result);
        
        theCopy=new GBounds(new Point(1,1),new Point(2,2));
        db7 = new GBounds(new Point(0,3),new Point(0,3));
        
        result=db7.intersection(theCopy,min2,max2);
        System.out.println(" 2D Intersection Test: Should be null-->"+result);
        
        db7=new GBounds(new Point(0,0),new Point(3,3));
        result=theCopy.intersection(db7,min2,max2);
        System.out.println(" 2D Intersection Test: Should be {(1,1),(2,2)}-->"+result);
        
        float [] min3=new float[3];
        float [] max3=new float[3];
        
        theCopy=(GBounds) db6.clone();
        result=theCopy.intersection(db5,min3,max3);
        System.out.println(" 3D Intersection Test: Should be {(1,1,1),(4,5,4)}-->"+result);
        
        /**********/
                System.out.println();
        System.out.println("Speed Test:  intersection() vs. intersection2() vs. intersection3() [in 3D]");
        System.out.println(" With non-null intersection:");
        int i;
        long startTime,stopTime;
        GBounds dst=new GBounds(new Point(1,2,3),new Point(10,11,12));
        GBounds src=new GBounds(new Point(4,5,6),new Point(12,13,14));
        
        System.gc();
        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.intersects(src);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("  Time for one Million intersects() calls: " + (stopTime-startTime) + " milliseconds.");
        
        System.gc();
        
        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.intersection(src);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("  Time for one Million intersection(GBounds) calls:" + (stopTime-startTime) + " milliseconds.");

        System.gc();

        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.intersection(src,min3,max3);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("  Time for one Million intersection(GBounds, int[], int[]) calls:" + (stopTime-startTime) + " milliseconds.");
        

        System.out.println(" With null intersection:");


        dst=new GBounds(new Point(1,2,3),new Point(10,11,12));
        src=new GBounds(new Point(9,12,11),new Point(15,16,17));
        
        System.gc();
        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.intersects(src);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("  Time for one Million intersects() calls: " + (stopTime-startTime) + " milliseconds.");
        
        System.gc();
        
        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.intersection(src);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("  Time for one Million intersection(GBounds) calls:" + (stopTime-startTime) + " milliseconds.");

        System.gc();

        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.intersection(src,min3,max3);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("  Time for one Million intersection(GBounds, int[], int[]) calls:" + (stopTime-startTime) + " milliseconds.");
        
        System.gc();


        System.out.println();
        System.out.println("Speed Test:  assign() vs. copy()");
        

    
        dst=new GBounds(new Point(1,2,3),new Point(10,11,12));
        src=new GBounds(new Point(4,5,6),new Point(12,13,14));
        

        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst=  src.copy();
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("Time for one Million copies: " + (stopTime-startTime) + " milliseconds.");
        
        
        
        startTime=System.currentTimeMillis();

        for(i=0;i<1000000;i++){
            dst.assign(src);
        }
        
        stopTime=System.currentTimeMillis();
        
        System.out.println("Time for one Million assigns: " + (stopTime-startTime) + " milliseconds.");
        
 
    }


    
    
}

