/* 
    LatticeBuilder.java
    Created by Denise Mitchell October 2002.
    University of New Hampshire
    Durham, NH 03824

    constructs immutable Lattices
  ------------------- Edit history as of 02/07/05 ---------------------
  02/07/05 rdb: renamed getDimensionality() to dim() for consistency with other classes.
*/

package edu.unh.sdb.lattice;
//import java.io.*;
import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;


//--------------------------LatticeBuilder class ----------------------
/** Manufactures Logical Lattice objects Hold relevant data and validates
before instantiating the appropriate Lattice objects*/


public class LatticeBuilder
{
  // table/list to hold the available lattices
  protected LatticeCollection latticeList;
  
  // the name of the pending lattice
  private String name;

  // the topology datasource
  private DataSource topologyDS;
    
  // the lattice geometry bounds
  private GBounds latticeGBounds;
    
  // the lattice bounds (for use in building partition within geometry);
  private ISBounds latticeISBounds;
    
  // the topology geometry bounds
  private GBounds topologyGBounds;
    
  // the topology geometry bounds (for use in building partition within geometry);
  private ISBounds topologyISBounds;
  
  /** declare the geometry type for this lattice */
  private int latticeGeometryType;
  
  /** declare the geometry type for topology */
  private int topologyGeometryType;
    
  /** declare constants to denote default topology */
  static final public int DEFAULT_GEOMETRY = 0;
    
  /** declare constants to denote unstructured topology */
  static final public int UNSTRUCTURED_GEOMETRY = 1;
     
  /** declare constants to denote recti-linear topology */
  static final public int REGULARCELL_GEOMETRY = 2;
  
  /** declare constants to denote recti-linear topology */
  static final public int REGULARPOINT_GEOMETRY = 3;
  
  // the RecordSpec for SplitUnstructuredTopology
  private RecordSpec vertexIndicesRecordSpec;
  
  // the RecordSpec for SplitUnstructuredTopology
  private RecordSpec coordinateRecordSpec;
  
  // the RecordSpec for SplitUnstructuredTopology
  private RecordSpec dataRecordSpec;
  
  //FDL File name for SplitUnstructuredTopology
  private String pointInfoFDLFileName;
  
  //FDL File name for SplitUnstructuredTopology
  private String vIndicesFDLFileName;
  
  //FDL File name for SplitUnstructuredTopology
  private String coordinateFDLFileName;
  
  // array of FDLGroup for SplitUnstructuredTopology
  private FDLGroup[] fdlGroupArray;
  
  /** declare the approximator type for this lattice */
  private int approximatorType;
    
  /** declare constants to denote default approximator */
  static final public int DEFAULT_APPROXIMATOR = 0;
    
  /** declare constants to denote array approximator */
  static final public int ARRAY_APPROXIMATOR = 1;
     
  /** declare constants to denote triangle approximator */
  static final public int AREA_TRIANGLE_APPROXIMATOR = 2;

  /** declare the topology dimension for this lattice */
  private int dimensionality;
  
  /** declare the topology type for this lattice */
  private int topologyType;
    
  /** declare constants to denote default topology */
  static final public int DEFAULT_TOPOLOGY = 0;
    
  /** declare constants to denote unstructured topology */
  static final public int UNSTRUCTURED_TOPOLOGY = 1;
     
  /** declare constants to denote recti-linear topology */
  static final public int RECTILINEAR_TOPOLOGY = 2;
    
  /**flag to set if lattice is not to be instantiated */
  protected boolean discardFlag;
    
    
  /** Default Constructor 
  */
  public LatticeBuilder(String n)
  {
    name = n;
    topologyGBounds=null;
    topologyISBounds=null;
    latticeGBounds=null;
    latticeISBounds=null;
  }
   
    /** Sets the lattice name to the String provided*/
    public void setName(String s)
    {
      name = s;
    }
  
    /** Returns the lattice name*/
    public String getName()
    {
      return name;
    }
    
    /** Sets the rood datasource **/
    public void setTopologyDS(DataSource ds)
    {
      if (ds == null)
      {
        System.err.println ("Lattice: " + getName() +  
                " Topology DataSource: has not been specified");
      }
        
      topologyDS = ds;
    }
    
    /** Sets the approximator type for this Lattice */
    public void setApproximatorType(int t)
    {
      approximatorType = t;
    }
    
    /** Returns the approximator type as an integer*/
    public int getApproximatorType()
    {
      return approximatorType;
    }
    
    /** returns a new Approximator object based on current builder data**/
    private Approximator makeApproximator()
    {
      Approximator retVal = null;  //ArrayApproximator is default value
        
      switch(approximatorType)
      {
        case AREA_TRIANGLE_APPROXIMATOR:
            retVal = new AreaTriangleCellApproximator();
            break;
        case ARRAY_APPROXIMATOR:
        case DEFAULT_APPROXIMATOR:
            retVal = new ArrayApproximator();
      }
        return retVal;
    }
    
   

    /** Sets the topology type for this Lattice */
    public void setTopologyType(int t)
    {
      topologyType = t;
    }
    
    /** Returns the topology type as an integer*/
    public int getTopologyType()
    {
      return topologyType;
    }
    

    /** Sets the topology dimension for this Lattice */
    public void setDimensionality(int t)
    {
      dimensionality = t;
    }
    
    /** Returns the topology dimensionality as an integer*/
    public int getDimensionality()
    {
      return dimensionality;
    }
    
    /** Sets the vertexIndicesRecordSpec for SplitUnstructuredTopology */
    public void setVertexIndicesRecordSpec(RecordSpec rs)
    {
        vertexIndicesRecordSpec = rs;
    }
    
    /** Returns the vertexIndicesRecordSpec for SplitUnstructuredTopology */
    public RecordSpec getVertexIndicesRecordSpec()
    {
        return vertexIndicesRecordSpec;
    }
    
    /** Sets the coordinateRecordSpec for SplitUnstructuredTopology */
    public void setCoordinateRecordSpec(RecordSpec rs)
    {
        coordinateRecordSpec = rs;
    }
    
    /** Returns the coordinateRecordSpec for SplitUnstructuredTopology */
    public RecordSpec getCoordinateRecordSpec()
    {
        return coordinateRecordSpec;
    }
    
    
    /** Sets the dataRecordSpec for SplitUnstructuredTopology */
    public void setDataRecordSpec(RecordSpec rs)
    {
        dataRecordSpec = rs;
    }
    
    /** Returns the dataRecordSpec for SplitUnstructuredTopology */
    public RecordSpec getDataRecordSpec()
    {
        return dataRecordSpec;
    }
    
    /** Sets the array of FDLGroup for SplitUnstructuredTopology */
    public void setFDLGroup(FDLGroup[] fdlGroup)
    {
        fdlGroupArray = fdlGroup;
    }
    
    /**Returns the vertexIndicesName */
    public String getVertexFileIndicesName(int i)
    {
        if(fdlGroupArray != null)
            return fdlGroupArray[i].vertexIndicesFileName;
        return null;
    }
    
    /**Returns the coordinateName */
    public String getCoordinateFileName(int i)
    {
        if(fdlGroupArray != null)
            return fdlGroupArray[i].coordinateFileName;
        return null;
    }
    
    /**Returns the dataName */
    public String getDataFileName(int i)
    {
        if(fdlGroupArray != null)
            return fdlGroupArray[i].dataFileName;
        return null;
    }
    
    
    /** Sets the FDL File name for SplitUnstructuredTopology */
    public void setPointInfoFDLFileName(String name)
    {
        pointInfoFDLFileName = name;
    }
    
    /**Returns the pointInfoFDLFileName */
    public String getPointInfoFDLFileName()
    {
        return pointInfoFDLFileName;
    }
    
    
     /** Sets the FDL File name for SplitUnstructuredTopology */
    public void setVIndicesFDLFileName(String name)
    {
        vIndicesFDLFileName = name;
    }
    
    /**Returns the vIndicesFDLFileName */
    public String getVIndicesFDLFileName()
    {
        return vIndicesFDLFileName;
    }
    
    
     /** Sets the FDL File name for SplitUnstructuredTopology */
    public void setCoordinateFDLFileName(String name)
    {
        coordinateFDLFileName = name;
    }
    
    /**Returns the coordianteFDLFileName */
    public String getCoordinateFDLFileName()
    {
        return coordinateFDLFileName;
    }
    
    
    
    /** returns a new Topology object based on current builder data**/
    private Topology makeTopology()
    {
      Topology retVal = null;  //ArrayApproximator is default value
        
      switch(topologyType)
      {
        case UNSTRUCTURED_TOPOLOGY:
          retVal = new SplitUnstructuredTopology(pointInfoFDLFileName, vIndicesFDLFileName, 
                    new UnstructuredGeometry(topologyISBounds, topologyGBounds), dimensionality);
           break;
        case RECTILINEAR_TOPOLOGY:
        case DEFAULT_TOPOLOGY:
            retVal = new RectilinearTopology(topologyDS, makeGeometry("topology"));
      }
        return retVal;
    }
    
    
    /** Sets the geometry bounds */
    public void setGBounds(GBounds b, String type)
    {
      if (b != null)
      {
          if(type.equals("lattice"))
            latticeGBounds=b;
          else if(type.equals("topology"))
            topologyGBounds=b;
      }
      else
        new IllegalArgumentError(SDBError.Recoverable,
                "LatticeBuilder.setLatticeGBounds() was passed a null reference");
    }
    
    /** Sets the topology bounds */
    public void setISBounds(ISBounds b, String type)
    {
      if (b != null)
      {
          if(type.equals("lattice"))
            latticeISBounds=b;
          else if(type.equals("topology"))
              topologyISBounds=b;
      }
      else
        new IllegalArgumentError(SDBError.Recoverable,
                "LatticeBuilder.setTopologyBounds() was passed a null reference");
    }

    /** returns the Index Space Bounds from this builder or defaults to the topologyDS bounds */
    public ISBounds getLatticeISBounds()
    {
      //if (latticeISBounds != null)   //user specified bounds, so use these
            return latticeISBounds;
    }
    
      /** returns the Index Space Bounds from this builder or defaults to the topologyDS bounds */
    public ISBounds getTopologyISBounds()
    {
      //if (topologyISBounds != null)   //user specified bounds, so use these
            return topologyISBounds;
    }
    
    /** returns the Index Space Bounds from this builder or defaults to the topologyDS bounds */
    public GBounds getLatticeGBounds()
    {
      //if (latticeGBounds != null)   //user specified bounds, so use these
            return latticeGBounds;
    }
    
      /** returns the Index Space Bounds from this builder or defaults to the topologyDS bounds */
    public GBounds getTopologyGBounds()
    {
      //if (topologyGBounds != null)   //user specified bounds, so use these
            return topologyGBounds;
    }
    
     public int getLatticeGeometryType()
    {
            return latticeGeometryType;
    }
    
    public int getTopologyGeometryType()
    {
            return topologyGeometryType;
    }
    
     public void setGeometryType(int gtype, String type)
    {
            if(type.equals("lattice"))
                latticeGeometryType = gtype;
            if(type.equals("topology"))
                topologyGeometryType = gtype;
    }
    
    /** check Topology's Geometry, if it is omitted, set it to the lattice's Geometry */
    public void checkTopology()
    {
        if(topologyGBounds == null && topologyISBounds == null)
        {
            topologyGBounds = latticeGBounds;
            topologyISBounds = latticeISBounds;
            topologyGeometryType=latticeGeometryType;
        }
    }
    
    
     /** returns a new Geometry object based on current builder data
	
	@FIXME: This needs to make either a RegularPointGeometry or RegularCellGeometry.
    */ 
     private Geometry makeGeometry(String type)
    {  
        int gType=0;
        ISBounds isBounds=null;
        GBounds gBounds=null;
        
        if(type.equals("lattice"))
        {
           gType=latticeGeometryType;
           isBounds=latticeISBounds;
           gBounds=latticeGBounds;
        }
        if(type.equals("topology"))
        {
            gType=topologyGeometryType;
            isBounds=topologyISBounds;
            gBounds=topologyGBounds;
        }
            
        switch(gType)
        {
            case UNSTRUCTURED_GEOMETRY:
                return new UnstructuredGeometry(isBounds, gBounds);
            case REGULARCELL_GEOMETRY:
                return new RegularCellGeometry(isBounds, gBounds);
            case DEFAULT_GEOMETRY:
            case REGULARPOINT_GEOMETRY:
                return new RegularPointGeometry(isBounds, gBounds);
        }
      
        return null;
    }
     
     
     
    /** instantiates an actual lattice **/
    public Lattice makeLattice()
    {
       return new Lattice(getName(), 
                          makeTopology(), 
                          makeGeometry("lattice"), 
                          makeApproximator());
    }
    

 /** diagnostic utility method for dumping this class to a string **/
    public String toString()
    {
      String s = "LatticeBuilder name: " ; 
      s = s + getName();
      s = s + " Approx: " + approximatorType;
      return s;
    }
 }
