/* 
LatticeCollection.java
Copyright 2002 University of New Hampshire
Durham, NH 03824
Authored by Denise Mitchell
-----------------------edit history -------------------------------------
06/05/08 rdb: made Vector generic

*/

package edu.unh.sdb.lattice;

//import java.io.*;
import java.util.*;

/**
 * The LatticeCollection class provides for a list of Lattices
 * with customary list functionality.  The list was intended to be
 * populated either by reading a database or by processing some Data
 * Description Language.  The list is only allowed to grow.  Elements
 * are never deleted from the list.
 */

public class LatticeCollection 
{
    /** The Lattice list */
    private Vector<Lattice> latticeList;
    
    /** holds index of next element in stored in the vector */
    private int next;
    
    /** default constructor **/
    public LatticeCollection()
    {
        latticeList = new Vector<Lattice>();
        next = 0;
    }
    
    /** returns the next Lattice object or null Lattice object if no more exists */
    public Lattice next()
    {
        next++;
        if (next <= latticeList.size())
          return getElementAt(next-1);
        else
          return  null;
    }
    
    /** Returns the Lattice at the designated index or null Lattice object if no more exists.
    Returns a null Lattice object if index does not exist. */
    public Lattice getElementAt(int i)
    {
        if (i < latticeList.size())
            return latticeList.elementAt(i);
        else 
            return null;
    }
    
    /** Returns the Lattice at the designated index or null Lattice object if no more exists.
    Returns a null Lattice object if index does not exist. This is a shorthand method to getElementAt() 
    added for compatibility with SDB naming conventions.*/
    public Lattice get(int i)
    {
        return getElementAt(i);
    }
    
    /** Add a Lattice to the end of the list */
    public void add(Lattice lattice)
    {
        if (lattice != null)
        {
            latticeList.addElement(lattice);
        }
        else
            System.err.println("Request to add a null data source to collection is being disregarded ");
    }
    
    /** Resets the index to beginning of the list */
    public void rewind()
    {
        next = 0;
    }
    
    /** Returns the integer index value for the Lattice denoted by string. Returns
    -1 if the Lattice name does not exist**/
    public int indexOfName(String s)
    {
        Lattice lattice;
        
        int i;
        for (i = 0; i < latticeList.size(); i++)
        {
            lattice = latticeList.elementAt(i);
            if (lattice.getName().equals(s)) //names are the same so...
                return i; //return the matching index
        }
        return -1;
    }
    
    /** Returns true if Lattice named by String is in the list */
    public boolean nameExists(String s)
    {
        /* if it exists, return true else */
        int size = latticeList.size();
        
        for (int i = 0; i < size; i++)
        {
            if (s.equals( getElementAt(i).getName()) )
                return true; 
        }
        
        return false; 
    }
    
    /** Returns the integer index of the Lattice corresponding to the 
    Lattice named by the string.  If name does not exist, a null Lattice will
    be returned**/
    public int indexOf(String s)
    {
        return latticeList.indexOf(s);
    }
    
    /** Returns the Lattice object corresponding to the Lattice named
    by the string **/
    public Lattice getLattice(String s)
    {
        int size = latticeList.size();
        Lattice lattice;
        
        int i;
        for (i = 0; i < size; i++)
        {
            lattice = latticeList.elementAt(i);
            if (lattice.getName().equals(s)) //names are the same so...
                return lattice; //return the matching index
        }
        return null;
    }
    
    /** Returns the Lattice object corresponding to the Lattice named
    by the string.  This is a shorthand for getLattice(String) to comply
    with SDB naming conventions.**/
    public Lattice get(String s)
    {
        return getLattice(s);
    }
    
    /** Returns the number of elements stored this Lattice collection as an integer
    **/
    public int size()
    {
        return latticeList.size();
    }
    
    /** adds an entire new collection onto the existing list **/
    /** if name exists in both collections, the new one is discarded **/
    public void merge(LatticeCollection latticeCollection)
    {
        Lattice lattice;
        
        for (int i = 0; i < latticeCollection.size(); i++)
        {
            lattice = latticeCollection.get(i);
            if (nameExists(lattice.getName() ))
            {
              System.out.println("Lattice : " + lattice.getName() +
                    " already exists.  Ignoring subsequent definition: ");
            }
            else
            add(lattice);
        }
    }
    
    /** for debugging - prints the total number of Lattice collections stored.
    Later will give names of each Lattice */
    public String toString()
    {
        String s = "Size of Lattice collection: " + size();
        for (int i = 0; i < size(); i++)
        {
            s = s + "\n" + i + ". " + (latticeList.elementAt(i)).getName(); 
        }
        return s;
        
    }
    
} // end LatticeCollection class

