//++++++++++++++ CommandSpec ++++++++++++++++++++++++
// a package class to store the command name/parm specifications
//    Should be extended to allow variable arg specification for each
//    argument rather than just the last.
//
// Edited:
//   6/05/08 -- made Vector generic
//   8/10/01 -- added ability for commands to have multiple words
//   6/11/01 -- change public variables into package visibility
//   5/29/01 -- added + var arg spec and numReqParms as public data member
//   4/26/01 -- added hasNumberArg method
//   2/ 6/01 -- made . a legal argument type => any argument type is legal

package edu.unh.sdb.util.CommandProcessor;
import java.util.Vector;

class CommandSpec
{
  // package visibility for all these variables
  String        parms;   // specif of arguments
  CommandAction action;  // Action to be executed
  char          status; // '+', '0' or '-': used checking args
  int           numMatch;     // # correctly matched args
  int           numPromote;   // # actual parameters "promoted" to more
                              // general match. Used to disambiguate.
  int           numReqParms;  // number required args
  char          varArgs;      // set to '+' or '*' if variable args
  Vector<String> cmdWords;     // for multiword commands

  //--------- CommandSpec( String, String, CommandAction ) -----------
  public      CommandSpec( String cmd, String parmSpec, CommandAction act )
  {
     boolean goodCmd = true;
     String  goodSpec = new String( ".sndif*+" );
     char    lastChar = ' ';     // last parameter specification

     // really need to check that * or + only allowed at the end
     for ( int i = 0; goodCmd && ( i < parmSpec.length()); i++ )
       goodCmd = goodCmd &&  goodSpec.indexOf( parmSpec.charAt( i ) ) >= 0;
     if ( !goodCmd )
        throw new CommandDefException( "Parameter spec bad: " + parmSpec );

     this.cmdWords = new Vector<String>();
     cmd = cmd.trim();
     int space = cmd.indexOf( ' ' );
     while ( space >= 0 )
     {
        this.cmdWords.add( cmd.substring( 0, space ) );
        cmd = cmd.substring( space ).trim();
        space = cmd.indexOf( ' ' );
     }
     if ( cmd.length() > 0 ) // if there's anything left, add it in
        this.cmdWords.add( cmd );

     this.action = act;
     this.varArgs = ' ';  // assume no variable args
     if ( parmSpec == null  ||  parmSpec.length() == 0 )
        this.parms = new String();
     else 
     {
        lastChar = parmSpec.charAt( parmSpec.length() - 1 );
        if (  ( lastChar != '*' )
            &&( lastChar != '+' ) )
           this.parms = parmSpec;
        else 
        {
           this.parms = parmSpec.substring( 0, parmSpec.length() - 1 );
           this.varArgs = lastChar;
        }
     }
     this.numReqParms = this.parms.length();
     if ( this.varArgs == '*' ) 
        this.numReqParms--; // last parm spec not required
     init();
  }
  //----------- init( ) ---------------
  public void init( ) 
  {
     this.status = '0';    // initialize to partially complete
     this.numMatch = 0;    //    and no matched args
     this.numPromote = 0;
  }
  //----------- toString( ) ---------------
  public String toString( ) 
  {
     if ( this.varArgs == ' ' )
        return getCommand() + "(" + this.parms + ")";
     else
        return getCommand() + "(" + this.parms + this.varArgs + ")";
  }
  //+++++++++++++++ accessors ++++++++++++++++++++++++++++++++
  //--------------- getCommand -------------------------------
  // return space separated words that compose the command
  //
  public String getCommand()
  {
     String cmd = new String();
     for ( int i = 0; i < this.cmdWords.size(); i++ )
        cmd = cmd + this.cmdWords.get( i ).toString() + ' ';
     return cmd.trim();
  }
  //--------------- getCommandWord ---------------------------
  public String getCommandWord( int i )
  {
     return this.cmdWords.get( i );
  }
  //--------------- numCmdWords ---------------------------
  public int numCmdWords( )
  {
      return this.cmdWords.size();
  }
  //
  //----------- hasNumberArg( ) -----------------
  // returns true if one or more arguments are a Number
  //         or Float or Double
  public boolean hasNumberArg()
  {
     return ( ( this.parms.length() > 0 )
                       && 
            (     this.parms.indexOf( 'n' ) >= 0 
               || this.parms.indexOf( 'f' ) >= 0
               || this.parms.indexOf( 'd' ) >= 0
               || this.parms.indexOf( 'i' ) >= 0
             ));
  }
  //----------- partialMatch( String ) -----------------
  // is argument the first word of the command
  public boolean partialMatch( String word )
  {
      return word.equals( cmdWords.get( 0 ) );
  }
  //----------- fullMatch( Arguments args ) -----------------
  // is argument the first word of the command
  public boolean fullMatch( Arguments args )
  {
      boolean match = true;
      for ( int i = 0; match && i < numCmdWords(); i++ )
         match = args.get( i ).toString().equals( cmdWords.get( i ));
      return match;
  }
  //
  //----------- parmType( int ) -----------------
  // returns the character representing the desired type for
  // argument specified in parameter.
  public char   parmType( int i )
  {
     if ( i < this.parms.length() )
        return this.parms.charAt( i );
     else if ( this.varArgs != ' ' )
        return this.parms.charAt( this.parms.length() - 1 );
     else
        return '?';
  }
  //------------- checkArg( int parmNum, char argType ) ----------------
  // check if specification of parameter 'parmNum' is compatible
  // with the argument type, argType.

  public char checkArg( int       parmNum,
                        char      argType )    // input type
  {
     char     retCode;
     boolean  match = false;
     char parmType = this.parmType( parmNum );
     switch ( parmType )
     {
     case '.':
              match = true;
              this.numPromote++;   // we'll count this as a double promotion
              this.numPromote++;
              break;
     case 'n':
              match = argType == 'f' || argType == 'i';
              if ( match ) 
                 this.numPromote++;
              break;
     case 'f':
              if ( argType == 'f' )
                 match = true;
              else if ( argType == 'i' )
              {
                 match = true;
                 this.numPromote++;
              }
              break;
     case 'd':
              if ( argType == 'd' )
                 match = true;
              else if ( argType == 'f'  || argType == 'i' )
              {
                 match = true;
                 this.numPromote++;
              }
              break;
     case 'i':
              match = argType == 'i';
              break;
     case 's':
              match = argType == 's';
              break;
     case '?':
              match = false;
              break;
     }
     if ( match )
        this.numMatch++;
     // test if we had a complete match previously, but there are 
     // extra arguments. this may still be the only valid match
     if ( this.numMatch >= this.numReqParms )
        retCode = '+';    // extra args does not nec. mean failure
     else if ( !match )
        retCode = '-';
     else if ( parmNum + 1  >= this.numReqParms )
        retCode = '+';
     else
        retCode = '0';
     return retCode;
  }
  //===================== main ======================================
  public static void main( String argv[] )
  {
     CommandSpec cs1 = new CommandSpec( "name", "ii*", null );
     CommandSpec cs2 = new CommandSpec( "name", "", null );
     CommandSpec cs3 = new CommandSpec( "name", "i", null );
     CommandSpec cs4 = new CommandSpec( "arb", "..", null );
     CommandSpec cs5 = new CommandSpec( "arb", "..*", null );

     System.out.println( "cs1 parms: " + cs1.parms.length() );
     System.out.println( "cs2 parms: " + cs2.parms.length() );
     System.out.println( "cs3 parms: " + cs3.parms.length() );
     System.out.println( "cs4 parms: " + cs4.parms.length() );
     System.out.println( "cs5 parms: " + cs5.parms.length() );

     System.out.print( "cs1: " );
     for ( int i=0; i<5; i++ )
        System.out.print( cs1.parmType( i ) );
     System.out.println();

     System.out.print( "cs2: " );
     for ( int i=0; i<5; i++ )
        System.out.print( cs2.parmType( i ) );
     System.out.println();

     System.out.print( "cs3: " );
     for ( int i=0; i<5; i++ )
        System.out.print( cs3.parmType( i ) );
     System.out.println();

     System.out.println("-----cs1 ii ---------------");
        System.out.print( cs1.checkArg( 0 , 'i' ));
        System.out.print( cs1.checkArg( 1 , 'i' ));
        System.out.println();
     System.out.println("+++++++++++++++++++++++++++");

     System.out.println("-----cs1 iii ---------------");
        System.out.print( cs1.checkArg( 0 , 'i' ));
        System.out.print( cs1.checkArg( 1 , 'i' ));
        System.out.print( cs1.checkArg( 2 , 'i' ));
        System.out.println();
     System.out.println("+++++++++++++++++++++++++++");
     
     System.out.println("-----cs1 iis ---------------");
        System.out.print( cs1.checkArg( 0 , 'i' ));
        System.out.print( cs1.checkArg( 1 , 'i' ));
        System.out.print( cs1.checkArg( 2 , 's' ));
        System.out.println();
     System.out.println("+++++++++++++++++++++++++++");
     
     System.out.println("-----cs2 i ---------------");
        System.out.print( cs2.checkArg( 0 , 'i' ));
        System.out.println();
     System.out.println("+++++++++++++++++++++++++++");
     
     System.out.println("-----cs3 i ---------------");
        System.out.print( cs3.checkArg( 0 , 'i' ));
        System.out.println();
     System.out.println("+++++++++++++++++++++++++++");
     
     System.out.println("-----cs3 ii ---------------");
        System.out.print( cs3.checkArg( 0 , 'i' ));
        System.out.print( cs3.checkArg( 1 , 'i' ));
        System.out.println();
     System.out.println("+++++++++++++++++++++++++++");
  }
}
