#include "RemoteDataSource.h"



/******************************************************************
 * Constructor
 *
 * args: RemoteConnection* rc
 *            - A connection to the granite client must exist
 *
 * Purpose: Registers callback methods for sim code
 *          Initializes variables/metadata
 *
 *****************************************************************/
RemoteDataSource::RemoteDataSource(GraniteRemoteIF *remote, char* name, 
                                   char* path)
{
   remoteConn = remote;

   dsID = remoteConn->createDataSource(name, path);

   int* arr = remoteConn->getDimensions(dsID, &numDims);

   dim_sizes = (int*)malloc( sizeof(int) * numDims );
   for (int i=0;i<numDims;i++) {
      dim_sizes[i]=arr[i];
   }

   /* domains represent how many (parallel processing) sections this has been
      broken into */
   numdomains = 1;
    
   numAttributes = 0;

   par_rank=0;
   par_size=1;

   md=NULL; 


   dsEndian =  remoteConn->getByteOrder(dsID);
   
    int val = 1;

    char* ptr = (char*)&val;

    if ( *ptr == 0) 
       osEndian = BIG;
    else 
       osEndian = LITTLE;
 
   InitializeMD(15);     
   InitializeMesh();
   InitializeScalarData();
}


/******************************************************************
 * Destructor
 *
 *
 *****************************************************************/
RemoteDataSource::~RemoteDataSource(){

}

/******************************************************************
 * GetMesh
 *
 * args: int domain       - each chunk of data represents a domain
 *       const char* name - 
 *
 * Purpose: Sets mesh values and returns it for visit to use
 *
 *****************************************************************/
VisIt_MeshData* RemoteDataSource::GetMesh(int domain, const char* name){

    int i=0;

    VisIt_MeshData *mesh = (VisIt_MeshData*) malloc(sizeof(VisIt_MeshData));
    mesh->meshType = VISIT_MESHTYPE_RECTILINEAR;
    mesh->rmesh = (VisIt_RectilinearMesh*)malloc(sizeof(VisIt_RectilinearMesh));

    mesh->rmesh->ndims = numDims;

    mesh->rmesh->baseIndex[0] = 0;
    mesh->rmesh->baseIndex[1] = 0;
    mesh->rmesh->baseIndex[2] = 0;

    mesh->rmesh->minRealIndex[0] = 0;
    mesh->rmesh->minRealIndex[1] = 0;
    mesh->rmesh->minRealIndex[2] = 0;

    mesh->rmesh->maxRealIndex[0] = dim_sizes[0];
    mesh->rmesh->maxRealIndex[1] = dim_sizes[1];
    mesh->rmesh->maxRealIndex[2] = dim_sizes[2];

    mesh->rmesh->dims[0] = dim_sizes[0];
    mesh->rmesh->dims[1] = dim_sizes[1];
    mesh->rmesh->dims[2] = dim_sizes[2];

    mesh->rmesh->xcoords = VisIt_CreateDataArrayFromFloat( VISIT_OWNER_SIM, 
                                                           p_xcoords );
    mesh->rmesh->ycoords = VisIt_CreateDataArrayFromFloat( VISIT_OWNER_SIM, 
                                                           p_ycoords );
    mesh->rmesh->zcoords = VisIt_CreateDataArrayFromFloat( VISIT_OWNER_SIM, 
                                                           p_zcoords );
    return mesh;
}
    

/******************************************************************
 * GetMetaData
 *
 * args: None
 *
 * Purpose: Sets and returns metadata
 *
 *****************************************************************/
VisIt_SimulationMetaData* RemoteDataSource::GetMetaData(void){

    fprintf(stderr, "  RemoteDataSource::GetMetaData about to return\n");
    return md;
}
    
/******************************************************************
 * GetScalar
 *
 * args: int domain
 *       char* name
 *
 * Purpose: Sets and returns scalar data for a specific attribute  
 *
 * Responisibility:     
 *          This method needs to check the endianness of the OS and the
 *          DataSource.  Based on their values it may need to turn
 *          off byte swapping on the remote conenction side or temporarilly
 *          fool granite by swapping the byte order for the xfer.
 *
 *
 *
 * Comment: Code currently ignores domain
 *
 *****************************************************************/
VisIt_ScalarData* RemoteDataSource::GetScalar(int domain, const char *name){

   if (osEndian == BIG) {
      remoteConn->disableRemoteByteSwapping();
   }
   else if ( ( osEndian == LITTLE ) && ( dsEndian == BIG  )) {
      remoteConn->enableRemoteByteSwapping();
   }
   else if ( ( osEndian == LITTLE ) && ( dsEndian == LITTLE )) {
      remoteConn->disableRemoteByteSwapping();
      remoteConn->setByteOrder(dsID, "BIG_ENDIAN");    
   }

   VisIt_ScalarData* sd = NULL;
   for (int i=0; i<md->numScalars;i++){
      if (strcmp(name,md->scalars[i].name)==0)
      {
         fprintf(stderr," getting % data!\n",name);
         sd = (VisIt_ScalarData*)malloc(sizeof(VisIt_ScalarData));
         sd->len  = (dim_sizes[0])*(dim_sizes[1])*(dim_sizes[2]);
         sd->data = VisIt_CreateDataArrayFromFloat(VISIT_OWNER_SIM, 
                                                   this->scalarData[i]);
      }
   }
   if ( ( osEndian == LITTLE ) && ( dsEndian == LITTLE )) {
      remoteConn->setByteOrder(dsID, "LITTLE_ENDIAN");    
   }
   return sd;
}
    
/******************************************************************
 * GetDomainList
 *
 * args: None
 *
 * Purpose: Sets and returns domains
 *
 *****************************************************************/   
VisIt_DomainList* RemoteDataSource::GetDomainList(){
    int i;
    VisIt_DomainList *dl = (VisIt_DomainList*)malloc(sizeof(VisIt_DomainList));

    dl->nTotalDomains = numdomains;

    dl->nMyDomains = numdomains;
    dl->myDomains = VisIt_CreateDataArrayFromInt(VISIT_OWNER_VISIT, 
                                                (int*)malloc(sizeof(int)));
    dl->myDomains.iArray[0] = par_rank;
    dl->myDomains.iArray[1] = par_rank;

    return dl;
}


/******************************************************************
 * InitializeScalarData
 *
 * args: None
 *
 * Purpose:
 *
 *****************************************************************/
void RemoteDataSource::InitializeScalarData()
{
    int total_alloc = 0;

    /* Allocate space for scalar data for each attribute */
    total_alloc += sizeof(float*) * this->md->numScalars;
    this->scalarData = (float**)malloc(sizeof(float*) * this->md->numScalars);
    if(this->scalarData == NULL)
    {
       fprintf(stderr," RemoteDataSource::InitializeVariables: Memory "
                      "allocation failed for " 
                      "scalarData. Exiting....");
       exit(0);
    }


    int numElements = dim_sizes[0]*dim_sizes[1]*dim_sizes[2];

    /* Allocate a 1D array and fill with scalar data for ALL attributes */
    int mallocSize = sizeof(float) * this->md->numScalars * numElements;

    float* bigArray = (float*)malloc(mallocSize);
  
    if(bigArray == NULL)
    {
       fprintf(stderr," RemoteDataSource::InitializeVariables: Memory "
                         "allocation failed for " 
                         "scalarData[%d]. Exiting....");
       exit(0);
    }
    remoteConn->getFloatsAttrOrder(dsID, bigArray, mallocSize );

    /*  Assign starting point of each array */
    for (int i=0;i<md->numScalars;i++) {
       this->scalarData[i] = &bigArray[ i * numElements ];
    }

}


/******************************************************************
 * InitializeMD
 *
 * args: int MaxNumCustCMD
 *
 * Purpose: initializesd metadata
 *
 *****************************************************************/
void RemoteDataSource::InitializeMD(int MaxNumCustCMD)
{
    md = (VisIt_SimulationMetaData*)malloc(sizeof(VisIt_SimulationMetaData));

    md->currentCycle = 0;
    md->currentTime  = 0;
    md->currentMode  = VISIT_SIMMODE_STOPPED;

    md->numMeshes = 1;
    md->meshes = (VisIt_MeshMetaData*)malloc(sizeof(VisIt_MeshMetaData) 
                                                  * md->numMeshes);

    md->meshes[0].name = strdup("mesh");
    md->meshes[0].meshType = VISIT_MESHTYPE_RECTILINEAR;
    md->meshes[0].topologicalDimension = numDims;
    md->meshes[0].spatialDimension = numDims;
    md->meshes[0].numBlocks = numdomains;
    md->meshes[0].blockTitle = strdup("Domains");
    md->meshes[0].blockPieceName = strdup("domain");
    md->meshes[0].numGroups = 0;
    md->meshes[0].units = NULL;
    md->meshes[0].xLabel = NULL;
    md->meshes[0].yLabel = NULL;
    md->meshes[0].zLabel = NULL;

    /* num attributes */
    md->numScalars = remoteConn->getNumAttributes(dsID);
    numAttributes = md->numScalars;
    md->scalars = (VisIt_ScalarMetaData*)malloc(sizeof(VisIt_ScalarMetaData) 
                                                 * md->numScalars);

    int sz=0;
    char** strArr = remoteConn->getAttributeNames(dsID, &sz);

    for (int i=0; i<sz && i<md->numScalars; i++) {
       md->scalars[i].name = strdup(strArr[i]);
    }
    // free unneeded array
    free(strArr);

    for (int i=0; i<md->numScalars; i++){
       md->scalars[i].meshName = strdup("mesh");
       md->scalars[i].centering = VISIT_VARCENTERING_NODE;
       md->scalars[i].treatAsASCII = 0;
    }

    md->numMaterials   = 0;
    md->numCurves      = 0;
    md->numExpressions = 0;
    md->expressions = 0;

    /* this will set up the generic and custom commands*/
    VisItInitAllCMD(md, MaxNumCustCMD);
}                    



/******************************************************************
 * InitializeMesh
 *
 * args: None
 *
 * Purpose: 
 *
 *****************************************************************/
void RemoteDataSource::InitializeMesh(){


  if (!parseGridFile("/local/data/jpo39/jcd0004.392x112x112.rr.grid")) {
      fprintf(stderr, "No mesh file located! Creating default uniform mesh.\n");
      int i,j;

      // allocate x coords
      p_xcoords = (float*)malloc(dim_sizes[0] * sizeof(*p_xcoords));
      if(p_xcoords == NULL)
      {
         printf("Memory allocation failed(1). Exiting....");
         exit(0);
      }
  

      // allocate y coords
      p_ycoords = (float*)malloc(dim_sizes[1] * sizeof(*p_ycoords));
      if(p_ycoords == NULL)
      {
         printf("Memory allocation failed(1). Exiting....");
         exit(0);
      }
 
      // allocate z coords
      p_zcoords = (float*)malloc(dim_sizes[2] * sizeof(*p_zcoords));
      if(p_zcoords == NULL)
      {      
         printf("Memory allocation failed(1). Exiting....");
         exit(0);
      }

      // assign axes values linearly
     for (i=0; i<dim_sizes[0]; i++)
     {
        p_xcoords[i] = i*i;
     }
     for (i=0; i<dim_sizes[1]; i++)
     {
        p_ycoords[i] = i*i;
     }
     for (i=0; i<dim_sizes[2]; i++)
     {
       p_zcoords[i] = i*i;
     }
   }
}








bool RemoteDataSource::parseGridFile(const char* filename)
{
    fprintf(stderr, "reading grid from file '%s'\n",filename);
    FILE* infile = NULL;


    if(filename != NULL)
       infile = fopen( filename, "r" );

    char junk[1024];
    int numXvals, numYvals, numZvals;
    float* xc = NULL, *yc = NULL, *zc = NULL; // x|y|z coordinates

    if(infile != NULL) {
        // HACK -- need to make more robust, 
        //         skip to DIMENSIONS, then X_COORDINATES, etc.
        // Format:
        //     DATASET RECTILINEAR_GRID
        //     DIMENSIONS 392 112 112
        //     X_COORDINATES 392 float
        //     ...
        int width,height,depth;
        fscanf(infile, "%s %s %s", junk, junk, junk);  // skip 3 tokens
        fscanf(infile, "%d %d %d", &width, &height, &depth);


        // read in the x coords
        fscanf(infile, "%s %d %s", junk, &numXvals, junk); 
        p_xcoords = (float*)malloc(numXvals * sizeof(*p_xcoords));       
        fprintf(stderr, "\n\n\n xcoords: \n\n\n");
        for(int i=0; i<numXvals; i++) {
            fscanf(infile, "%f", &p_xcoords[i]);
            fprintf(stderr, "%13.6f\n", p_xcoords[i]);
        }
        fprintf(stderr,"\n\n\n");



        // read in the y coords
        fscanf(infile, "%s %d %s", junk, &numYvals, junk);   
        p_ycoords = (float*)malloc(numYvals * sizeof(*p_ycoords));
        for(int i=0; i<numYvals; i++) {
            fscanf(infile, "%f", &p_ycoords[i]);
        }

        // read in the z coords
        fscanf(infile, "%s %d %s", junk, &numZvals, junk); 
        p_zcoords = (float*)malloc(numZvals * sizeof(*p_zcoords));  
        for(int i=0; i<numZvals; i++)
            fscanf(infile, "%f", &p_zcoords[i]);

        fclose(infile);
    }
    else {
       fprintf(stderr,"unable to open coordinate info file '%s'", filename);
    }
    return infile;
}
