///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//          U N I V E R S I T Y    O F    N E W    H A M P S H I R E         //
//                                                                           //
//                       COMPUTER  SCIENCE  DEPARTMENT                       //
//                                                                           //
//                    Graphics And Visualization Research                    //
//                                                                           //
//          --------------------------------------------------------         //
//                                                                           //
//  Professor:   Dan Bergeron <rdb@cs.unh.edu>                               //
//  Programmer:  Andy Foulks <rafoulks@cs.unh.edu>                           //
//  File:        RemoteConnection.cpp                                        //
//                                                                           //
//                                                                           //
//  History:                                                                 //
//      7/10/09 - (Frain) Adding commands that can be handled by this class  //
//      7/27/09 - (Frain) Added flag to turn on/off byte swapping            //
//                - not the most efficient way but preprocessor is too       //
//                  restricting                                              //
///////////////////////////////////////////////////////////////////////////////


// local includes
#include "RemoteConnection.h"

// C includes
#include <stdarg.h>
#include <errno.h>
#include <netinet/in.h>
#include <unistd.h>

#include <netdb.h>
#include <sys/socket.h>
#include <resolv.h>
#include <arpa/inet.h>

// Frain added for string tokenizing
#include <stdio.h>  
#include <string.h>


// Frain
float endianSwap( float value );


/* ========================================================================= */
/**
 *  Constructs a RemoteConnection to connect to the hostname and
 *  portnum specified.
 */

RemoteConnection::RemoteConnection(const char* hostname, int port)
{
    byteSwapFlag = true;

    TRACE(hostname,port);

    DEBUG("hostname=%s, port=%d",hostname,port);

    mConnectionEstablished = false;
    mSocketId = connectToServer(hostname, port);

    if(mSocketId > 0)
        mConnectionEstablished = true;
    else
        exit(fprintf(stderr,"Unable to connect to '%s:%d'\n\n",hostname,port));
}

/* ========================================================================= */
/**
 *  Sends a string message (using a printf style format string) to 
 *  the server and waits for a reply.  Returns the reply to the 
 *  caller as an integer.
 */

int RemoteConnection::sendStringGetInt(const char* format, ...)
{
    TRACE(format);
    char message[BUFSIZE+1];
    int value = -99;

    if(mConnectionEstablished)
    {
        va_list args;
        va_start(args, format);
        int size = vsnprintf(message, BUFSIZE, format, args);
        va_end(args);

        if(size < BUFSIZE)
            sendMessage(mSocketId, message);
        else
            error("Message '%s' NOT sent, size=%d TOO LARGE\n", size);

        receiveReply((void*)&value, sizeof(int), mSocketId);
    }
    else
    {
        error("Unable to sendStringGetInt: No connection established");
    }

    return value;
}

/* ========================================================================= */
/**
 *  Sends a string message (using a printf style format string) to 
 *  the server and waits for a reply.  Returns the reply to the 
 *  caller as a float.
 */

float RemoteConnection::sendStringGetFloat(const char* format, ...)
{
    TRACE(format);
    static char message[BUFSIZE+1];
    float value = 0.0;

    if(mConnectionEstablished)
    {
        va_list args;
        va_start(args, format);
        int size = vsnprintf(message, BUFSIZE, format, args);
        va_end(args);

        if(size < BUFSIZE)
            sendMessage(mSocketId, message);
        else
            error("Message '%s' NOT sent, size=%d TOO LARGE\n", size);

        receiveReply((void*)&value, sizeof(float), mSocketId);
    }
    else
    {
        error("Unable to sendStringGetFloat: No connection established");
    }
    return value;
}

/* ========================================================================= */
/**
 *  Sends a string message (using a printf style format string) to 
 *  the server and waits for a reply.  Returns the reply to the 
 *  caller in the first parameter 'buf', up to the size specified 
 *  in the second parameter 'bufsize'.  The returned data is NOT
 *  byte swapped.
 *
 *  @param  buf         Buffer to hold the data returned from the 
 *                      server.  The memory pointed to should be
 *                      allocated before calling this function. 
 *  @param  bufsize     The maximum number of BYTES to put into the buffer
 *  @param  format      printf style string command to send to the server
 */

void RemoteConnection::sendStringGetRawData(void* buf, int bufsize, 
                                             const char* format, ...)
{
    TRACE(buf,bufsize,format,);
    char message[BUFSIZE+1];

    if(mConnectionEstablished) {
        va_list args;
        va_start(args, format);
        int size = vsnprintf(message, BUFSIZE, format, args);
        va_end(args);

        if(size < BUFSIZE)
            sendMessage(mSocketId, message);
        else
            error("Message '%s' NOT sent, size=%d TOO LARGE\n", size);

        receiveReply(buf, bufsize, mSocketId);
    }
    else {
        error("Unable to sendStringGetRawData: No connection established");
    }
}


/* ========================================================================= */
/**
 *  Sends a string message (using a printf style format string) to 
 *  the server and waits for a reply.  Returns the reply to the 
 *  caller in the first parameter 'buf', up to the size specified 
 *  in the second parameter 'bufsize'.  The returned data is NOT
 *  byte swapped.
 *
 *  @param  bufsize     The maximum number of BYTES to put into the buffer
 *  @param  format      printf style string command to send to the server
 */

char* RemoteConnection::sendStringGetString(const char* format, ...)
{
    TRACE(format,);
    char message[BUFSIZE+1];

    if(mConnectionEstablished) {
        va_list args;
        va_start(args, format);
        int size = vsnprintf(message, BUFSIZE, format, args);
        va_end(args);

        if(size < BUFSIZE)
            sendMessage(mSocketId, message);
        else
            error("Message '%s' NOT sent, size=%d TOO LARGE\n", size);

         
        char* temp = receiveStringReply(mSocketId);
        fprintf(stderr,"   sendStringGetString returning: <%s>\n",temp);
        return temp;
    }
    else {
        error("Unable to sendStringGetStrings: No connection established");
    }
    return NULL;
}




/* ========================================================================= */
/* ==                           PRIVATE METHODS                           == */
/* ========================================================================= */
/**
 *  Method to communicate with server.
 *
 *  General routine to send a string message to the server.
 *  Protocol for sending messages is to send ascii text strings
 *  with a newline (\n) at the end.  The server will read from the 
 *  socket until newline.
 */

void RemoteConnection::sendMessage(int socketId, const char* message)
{
    TRACE(socketId,message);
    REQUIRE(strlen(message)<(unsigned)BUFSIZE,
            ("message len=%d, BUFSIZE=%d",(int)strlen(message),BUFSIZE));
    DEBUG("Sending message '%s' to server", message);

    int length = strlen(message);

    // message must contain a newline or it won't work
    if(message[length-1] != '\n') {
        if(length+1 > BUFSIZE)
            exit(fprintf(stderr,"message does not contain newline. "
                        "Adding newline will overflow send buffer.  Message "
                        "length is %d, BUFSIZE=%d\n",length,BUFSIZE));

        warning("Message '%s' does not contain newline.  Adding newline "
                "before sending.", message);

        char fixedMessage[BUFSIZE+1];
        sprintf(fixedMessage,"%s\n", message);

        writebytes(socketId, fixedMessage, length+1);
    }
    else {
        writebytes(socketId, message, length);
    }
}

/* ========================================================================= */
/**
 *  Method to communicate with server.
 *
 *  The protocol for receiving messages from the server is to 
 *  break each message into three parts:  
 *      The first message is a 4 byte integer value representing an
 *          error code.  0 is success.
 *      The second message from the server is a single 32 bit 
 *          integer containing the number of bytes of data being sent.  
 *      The third message is the data itself.
 *
 *  In this procotol, we assume the server is always sending us
 *  little endian data.
 *  
 *  The data is returned to the user via the first parameter, 
 *  up to the number of bytes specified int he second parameter.
 */

void RemoteConnection::receiveReply(void* buffer, int bufsize, int socketId)
{
    TRACE(buffer,bufsize,socketId);
    int size = 0;
    int errorcode = 0;

    // first message is an error code
    readbytes(socketId, (char*) &errorcode, 4);
    errorcode = ntohl(errorcode);  // big endian to little endian

    DEBUG("got errorcode='%d'",errorcode);

    // get the first 4 byte message == size of the data/text to follow
    readbytes(socketId, (char*) &size, 4);

    size = ntohl(size);  // convert big endian to little (network to host)

    DEBUG("size of message to follow is '%d' bytes", size);

    if( !errorcode ) {
        if(bufsize >= size)
            readbytes(socketId, (char*)buffer, size);
        else
            error("receiveReply: size of buffer='%d' is not large enough "
                   "to hold size of message='%d' !!", bufsize, size);      

        //  Frain added flag check
        if (byteSwapFlag) { 
           // Byte swapping routine
           int* bufptr = (int*) buffer;
           for(int i=0; i<size/4; i++)
              bufptr[i] = ntohl(bufptr[i]);
        }
    }
    else {
        error("receiveReply: error '%d' received", errorcode);

        // currently, we assume that on an error condition, no other 
        // data will be sent (size should be 0).
        REQUIRE(size==0,("Error received but size='%d' (should be 0)",size));
    }
}


/* ========================================================================= */
/**
 *  08-04-09 - Frain Added this method
 *           - I need to request strings of unknown size!
 *             This method allocates memory itself
 * 
 *  Method to communicate with server.
 *
 *  The protocol for receiving messages from the server is to 
 *  break each message into three parts:  
 *      The first message is a 4 byte integer value representing an
 *          error code.  0 is success.
 *      The second message from the server is a single 32 bit 
 *          integer containing the number of bytes of data being sent.  
 *      The third message is the data itself.
 *
 *  In this procotol, we assume the server is always sending us
 *  little endian data.
 *  
 *  The data is returned to the user via the first parameter, 
 *  up to the number of bytes specified int he second parameter.
 */

char* RemoteConnection::receiveStringReply(int socketId)
{
    TRACE(buffer,bufsize,socketId);
    int size = 0;
    int errorcode = 0;

    // first message is an error code
    readbytes(socketId, (char*) &errorcode, 4);
    errorcode = ntohl(errorcode);  // big endian to little endian

    DEBUG("got errorcode='%d'",errorcode);

    // get the first 4 byte message == size of the data/text to follow
    readbytes(socketId, (char*) &size, 4);

    size = ntohl(size);  // convert big endian to little (network to host)

    DEBUG("size of message to follow is '%d' bytes", size);

    fprintf(stderr, "  ----- size = %d\n",size);
    char* buffer = (char*) malloc(sizeof(char) * (size + 1));
    buffer[ size ] = '\0'; 
 
    if( !errorcode ) {
       readbytes(socketId, buffer, size);
    }
    else {
        error("receiveReply: error '%d' received", errorcode);

        // currently, we assume that on an error condition, no other 
        // data will be sent (size should be 0).
        REQUIRE(size==0,("Error received but size='%d' (should be 0)",size));
    }
    fprintf(stderr,"  RemoteConnection::receiveStringReply: returning <%s>\n",buffer);
    return buffer;
}

/* ========================================================================= */
/**  
 *  Low level function to read bytes from a socket.  This routine 
 *  guarantees that nbytes will be read from the socket,
 *  except for EOF. What happens is that sometimes you cannot read all
 *  of the data that you wanted to because of network latencies.
 *  This function loops until the specified number of bytes is read.
 */

int RemoteConnection::readbytes(int sd, char *buffer, int nbytes)
{
    TRACE(sd,*buffer,nbytes);
    int nreceived;                   /* number of bytes received in the read */
    int totalreceived;               /* total number of bytes read so far    */
 
    for(totalreceived=0; totalreceived < nbytes; totalreceived+=nreceived) {
        nreceived = read(sd, &buffer[totalreceived],nbytes-totalreceived);
 
        if(nreceived < 0) {
            if(errno == EINTR) {
                nreceived = 0;
            } 
            else {
                perror("readbytes: read from socket failure");
                return(-1);
            }
        } 
        else if(nreceived == 0) {               /* this means EOF received */
            break;
        }
    }
 
    return totalreceived;
}

/* ========================================================================= */
/** 
 *  Low level function to send bytes over a socket.  This routine 
 *  guarantees that nbytes will be written to the socket.
 *  What happens is that sometimes you cannot write all of the data that
 *  you wanted to because of network latencies.  This function loops
 *  until the specified number of bytes is written.
 */

int RemoteConnection::writebytes(int sd, const char *buffer, int nbytes)
{
    TRACE(sd,*buffer,nbytes);
    int nsent;                          /* number of bytes sent in the write */
    int totalsent;                      /* total number of bytes sent so far */
 
    for(totalsent=0; totalsent < nbytes; totalsent += nsent) {
        nsent = write(sd, &buffer[totalsent], nbytes-totalsent);
 
        if(nsent < 0) {
            if(errno == EINTR) {
                nsent = 0;
            } 
            else {
                perror("writebytes: write to socket failure");
                return(-1);
            }
        }
    }

    // flushes all written bytes to I/O
    fsync(sd);
 
    return totalsent;
}

/* ========================================================================= */
/**
 *  Method to establish a connection with the server.  Returns
 *  the socketId for the connection, if successful, or -1 if the
 *  connection failed.
 */

int RemoteConnection::connectToServer(const char* hostname, int port)
{
    TRACE(hostname,port);
    int sd = -1;
    struct sockaddr_in dest;
                                                                            
    // open socket for streaming
    if((sd = socket(AF_INET, SOCK_STREAM, 0)) >= 0) {
        // initialize server address/port struct
        memset(&dest, 0, sizeof(dest));
        dest.sin_family = AF_INET;
        // ???
        dest.sin_addr.s_addr = htonl(INADDR_ANY);
        dest.sin_port = htons(port);

        // convert hostname to an internal, binary form
        if(inet_aton(hostname, &dest.sin_addr) != 0) {
            // connect to server
            
            //if(bind(sd, (struct sockaddr*)&dest, sizeof(dest)) == 0) // DGRAM
            if(connect(sd, (struct sockaddr*)&dest, sizeof(dest)) == 0) {
                info("Successfully connected to remote server! (%s:%d)", 
                        hostname, port);
            }
            else {
                error("RemoteConnection::connectToServer: connect() failed");
                perror("perror: ");
                sd = -1;
            }
        }
        else {
            error("RemoteConnection::connectToServer: inet_aton() failed");
            sd = -1;
        }
    }
    else {
        error("RemoteConnection:connectToServer: socket() failed");
        sd = -1;
    }

    return sd;
}

/* ========================================================================= */

RemoteConnection::~RemoteConnection()
{
    TRACE();
    close(mSocketId);
}

/* ========================================================================= */
/* ==                        STAND ALONE TEST CODE                        == */
/* ========================================================================= */

#ifdef STAND_ALONE_REMOTE_CONNECTION



//////////////////////////////////////////////////////////////////////
// endianSwap
//
//  - added Frain
//////////////////////////////////////////////////////////////////////
float endianSwap( float value )
{
    unsigned char swapped[4];
    unsigned char* ptr = (unsigned char*) &value;

    swapped[3] = ptr[0];
    swapped[2] = ptr[1];
    swapped[1] = ptr[2];
    swapped[0] = ptr[3];

    float f = *((float*)(&swapped[0]));

    return f;
}



/*
 *  To compile:
 *      g++ -g -I.. -DSTAND_ALONE_REMOTE_CONNECTION RemoteConnection.cpp
 */

#include <string.h>
#include <iostream>
using namespace std;

int main(int argc, char** argv)
{
    FILE* infile = stdin;
    char hostname[128];
    char buf[128];
    int port;


    printf(" argv[1] = %s\n", argv[1]);

    if(argc >= 2) {
        if(strstr(argv[1],"-server")!=NULL) {
            strcpy(hostname, argv[2]);
            port = atoi(argv[3]);
        }
        
        if (strstr(argv[1],"-file")!=NULL) {
            infile = fopen(argv[2], "r");
            if(infile == NULL) {
                fprintf(stderr,"Unable to open file '%s'\n",argv[2]);
                infile = stdin;
            }
        }

        if (strstr(argv[4],"-file")!=NULL) {
            infile = fopen(argv[5], "r");
            if(infile == NULL) {
                fprintf(stderr,"Unable to open file '%s'\n",argv[5]);
                infile = stdin;
            }
        }        

        if( (strstr(argv[1],"-test") != NULL) || 
            (strstr(argv[4],"-test") != NULL) ) 
        {
          printf(" processing -test\n");

            FILE* outfile = fopen("foo", "w");
            if(outfile != NULL) {
                fprintf(outfile, 
                  "dsCreate test /local/data/jpo39/Increment64.xfdl\n"
                  "getFloat 0 0 0 0 0\n"  
                  "getFloat 0 0 1 2 3\n"
                   // need a space before \n on following line
                  "getFloatsByPoint 0 \n"

                  "subblock 0 1 1 1 3 3 3\n"
                  "getFloat 1 0 0 0 0\n"
                   // need a space before \n on following line
                  "getFloatsByPoint 1 \n"

                  "dsCreate test2 /local/data/jpo39/Increment64_2attr.xfdl\n"
                  "getFloat 2 0 0 0 0\n"  
                  "getFloat 2 1 0 0 0\n"  
                 
                  "getFloat 2 0 1 1 7\n" 
                  "getFloat 2 0 0 0 1\n" 

                   // need a space before \n on following line
                  "getFloatsByPoint 2 \n"
                  "getFloatsByAttr 2 \n"
                 );
            }

            fclose(outfile);
            infile = fopen("foo", "r");
            if(infile == NULL)
                infile = stdin;
        }
    }
    else
    {
        cout << "Enter server address (using numbers-and-dots notation): "; 
        cin >> hostname;

        cout << "Enter port: ";
        cin >> port;

        fgets(buf, 1024, stdin);  // get the extra newline
    }

    RemoteConnection* r = new RemoteConnection(hostname, port);

    if(infile == stdin)
    {
        fprintf(stderr, "\nCommands:\n    msg <string>\n");
        fprintf(stderr, "    dsCreate name xfdlfile\n");
        fprintf(stderr, "    getFloat dc_id attr i j k\n");

        //fprintf(stderr, "    getFloat stepId attr i j k\n\n");
    }

    while(1)
    {
        char message[1024] = {0};

        if(infile == stdin)
            printf("Send command: ");
        fgets(message, 1023, infile);

        if(strstr(message,"quit")!=NULL)
            break;

        if(feof(infile))
            break;

        fprintf(stderr, "Sending: %s", message);
        fprintf(stderr, "    Response from server was: ");

        if(strstr(message,"msg") != NULL)
        {
            char reply[1024];
            r->sendStringGetRawData((void*)reply, sizeof(reply), message);

            fprintf(stderr, "    as string: '%s'\n", reply);
        }
        else if ( (strstr(message,"dsc") != NULL) || 
             (strstr(message,"dsCreate") != NULL))
        {
            int reply = r->sendStringGetInt(message);
            fprintf(stderr, "******** Datasource id = : '%d'  **********\n", 
                    reply);
        }
        /*
        else if(strstr(message, "getFloats") != NULL)
        {
            float array[392*112*112];
            r->sendStringGetRawData((void*)array, sizeof(array), message);

            for(int i=0; i<10; i++)
                fprintf(stderr, "    [%d]=%f\n", i, array[i]);

            fprintf(stderr, "    [%d]=%f\n", 1229312, array[1229312]);
        }
        */
        else if ( strstr(message, "getFloatsByAttr") != NULL)
        {
            int i=0, dc_id=0;
            char *token;
           
            token = strtok (message," ");
            while (token != NULL)
            {
               token = strtok (NULL, " ");

               if (i == 0) {
                 dc_id = atoi(token); 
                   
               }
               i++;
            }
            
            float arr[64];
        
            r->sendStringGetRawData(arr, 64*4, "getFloatsByAttr %d\n",dc_id);
            fprintf(stderr, "    as floats: ");
            // loop here
            for (int i=0; i<64;i++) {
               if (( i % 4 ) == 0 )
                  fprintf(stderr, "\n");
               fprintf(stderr, "%6.2f ", arr[i]);
            }
            fprintf(stderr,"\n\n");
        }
        // getFloatsByPoint
        else if ( strstr(message, "getFloatsByPoint") != NULL)
        {

          fprintf(stderr, " message = :    %s\n",message);

            int i=0, dc_id=0;
            char *token;
           
            token = strtok (message," ");
            while (token != NULL)
            {
               token = strtok (NULL, " ");

               if (i == 0)
                 dc_id = atoi(token); 
                 
               i++;
            }
            
            float arr[64];
     
            r->sendStringGetRawData(arr, 64*4, "getFloatsByPoint %d\n",dc_id);

            fprintf(stderr, "    as floats: ");
            // loop here
            for (int i=0; i<64;i++) {
               if (( i % 4 ) == 0 )
                  fprintf(stderr, "\n");
               fprintf(stderr, "%6.2f ", arr[i]);
            }
            fprintf(stderr,"\n\n");
        }
        // getFloat
        else if(strstr(message, "getFloat") != NULL)
        {
            float reply = r->sendStringGetFloat(message);
            fprintf(stderr, " '%f'\n", reply);
        }
        // getInt
        else if ( strstr(message, "getInt") != NULL)
        {
            int reply = r->sendStringGetInt(message);
            fprintf(stderr, " '%d'\n", reply);
        } 
        else if ( strstr(message, "getByteOrder") != NULL)
        {
            int reply = r->sendStringGetInt(message);
            fprintf(stderr, " '%d'\n", reply);
        } 
        // subblock
        else if ( strstr(message, "subblock") != NULL)
        {
            int reply = r->sendStringGetInt(message);
            fprintf(stderr, " '%d'\n", reply);
        }

    }

    //remove("foo");
    fclose(infile);
    delete r;
    return 0;
}






/* ========================================================================= */

#endif // STAND_ALONE_REMOTE_CONNECTION
