///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//          U N I V E R S I T Y    O F    N E W    H A M P S H I R E         //
//                                                                           //
//                       COMPUTER  SCIENCE  DEPARTMENT                       //
//                                                                           //
//                    Graphics And Visualization Research                    //
//                                                                           //
//          --------------------------------------------------------         //
//                                                                           //
//  Professor:   Dan Bergeron <rdb@cs.unh.edu>                               //
//  Programmer:  Andy Foulks <rafoulks@cs.unh.edu>                           //
//  File:        StarObject.h                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef _OBJECT_H_
#define _OBJECT_H_

//#define DEBUG_ON
#define ASSERTIONS


//#define DEBUG_TO_STDERR // added Frain - not sure where andy normally defines it


// C includes
#include <signal.h>    // to catch SIGSEGV segmentation faults
#include <cstdio>
#include <cstdarg>     // for va_args, etc
#include <cstring>
#include <cstdlib>
//#include <execinfo.h>  // for backtrace and backtrace_symbols

// C++ includes
#include <typeinfo>
#include <iostream>
#include <stdexcept>
#include <vector>

using std::vector;
using std::string;

/**
 *  @b STARview @b Project @b Description
 *
 *  A Framework for Volume Visualization of Hierarchical 
 *  Multidimensional Data
 *
 *  The purpose of the following project is to provide the scientist
 *  with tools necessary to examine and query data output from fluid
 *  flow simulations.  Although tailored specifically to fluid flow
 *  data, the tool will also be designed to provide a generalized
 *  framework from which the scientist can use volume rendering to
 *  examine any kind of data sets.
 *
 *  Class StarObject is the base class of all classes in the project.
 *  It includes commonly used functions such as string tokenization
 *  (@see split()), convertion utilities (@see toInt(), toFloat()),
 *  and static methods to print messages to the user (@see info(),
 *  warning(), and error()).  It includes assertions and facilities
 *  to trace and debug the code. 
 */

class StarObject
{
public:  // constants

public:  // constructors
        StarObject()  {}
virtual ~StarObject() {}

public:  // api

/* ========================================================================= */
/**
 *  Returns the name of the class (implemented 
 *  using C++'s bizarre RTTI syntax).
 */

virtual const char* className() 
{ 
    return typeid(this).name(); 
}

/* ========================================================================= */
/**
 *  Reimplement this function in your subclass.  By default
 *  returns the address and the name of the class.
 */
virtual const char* toString()
{
    static char buf[1024];

    sprintf(buf, "[%s@%p]", className(), this);

    return( (const char*)buf );
}

/* ========================================================================= */
/* ==                                STATIC                               == */
/* ========================================================================= */
/**
 *  Splits a string into tokens.  Inspired by java.  Returns the number
 *  of tokens, but will only split up to the max specified.
 *
 *  @param line     string to split
 *  @param delim    string of delimiters
 *  @param tokens   array to hold individual tokens
 *  @param max      maximum number of elements to put into the array
 */

static int split( const char* line, const char* delim, string* tokens, int max )
{
    int numTokens = 0;

    if( max > 0 )
    {
        char* saveptr = NULL;
        char* copy = new char[strlen(line)+1];

        strcpy( copy, line );

        char* token = strtok_r( copy, delim, &saveptr );

        while( token != NULL && numTokens < max )
        {
            tokens[numTokens] = token;
            numTokens++;

            token = strtok_r( NULL, delim, &saveptr );
        }

        if( token != NULL )
            warning("line '%s' has more than max=%d tokens",line,max);

        delete [] copy;
    }
    else
    {
        warning( "could not split '%s', max='%d'",line,max );
    }

    return numTokens;
}

/* ========================================================================= */

static vector<string> split( const char* line, const char* delim=" \t\n" )
{
    vector<string> tokens;

    char* saveptr = NULL;
    char* copy = new char[strlen(line)+1];

    strcpy( copy, line );

    char* token = strtok_r( copy, delim, &saveptr );

    while( token != NULL )
    {
        tokens.push_back( token );

        token = strtok_r( NULL, delim, &saveptr );
    }

    delete [] copy;

    return tokens;
}

/* ========================================================================= */
/**
 *  For removing comments from lines.  This removes everything 
 *  from the endmark to the newline by replacing the first 
 *  instance of the endmark character with the null byte.  
 *  For example, the string
 *  
 *      "file = /local/tmp/data/filename.tsd  # this is a comment"
 *
 *  Will become
 *
 *      "file = /local/tmp/data/filename.tsd  "
 *
 *  If the endmark character passed in is '#'.
 */

static char* strip( char* line, int size, char endmark )
{
    char* ptr = line;

    for( int i=0; i<size; i++, ptr++ )
    {
        if( ptr == NULL )
        {
            break;
        }
        else if( *ptr == '\0' )
        {
            break;
        }
        else if( *ptr == endmark )
        {
            *ptr = '\0';
            break;
        }
    }

    return line;
}

/* ========================================================================= */
/**
 *  Methods to convert string (C++ or C) to 64 bit integer (long long),
 *  errors in parsing will return 0 and print a warning message.
 */

static long long toLong( string& str )
{
    return toLong( str.c_str() );
}
static long long toLong( const char* txt )
{
    long long value = 0;
    char* endptr = NULL;

    if( txt != NULL )
    {
        value = strtoll( txt, &endptr, 10 );
        
        if( endptr == NULL || endptr[0] != '\0' || txt[0] == '\0' )
        {
            warning( "Unable to convert '%s' to long long", txt );
            value = 0;
        }
    }

    return value;
}

/* ========================================================================= */
/**
 *  Methods to convert string (C++ or C) to integer, errors in parsing
 *  will return 0 and print a warning message.
 */

static int toInt( string& str )
{
    return toInt( str.c_str() );
}
static int toInt( const char* txt )
{
    int value = 0;
    char* endptr = NULL;

    if( txt != NULL )
    {
        value = strtol( txt, &endptr, 10 );
        
        if( endptr == NULL || endptr[0] != '\0' || txt[0] == '\0' )
        {
            warning( "Unable to convert '%s' to int", txt );
            value = 0;
        }
    }

    return value;
}

static bool isInt( const char* txt )
{
    bool isInteger = false;
    char* endptr = NULL;

    if( txt != NULL )
    {
        strtol( txt, &endptr, 10 );
        
        if( endptr == NULL || endptr[0] != '\0' || txt[0] == '\0' )
            isInteger = false;
        else
            isInteger = true;
    }

    return isInteger;
}

/* ========================================================================= */
/**
 *  Methods to convert string (C++ or C) to float, errors in parsing
 *  will return 0.0 and print a warning message.
 */

static float toFloat( string& str )
{
    return toFloat( str.c_str() );
}
static float toFloat( const char* txt )
{
    float value = 0.0;
    char* endptr = NULL;

    if( txt != NULL )
    {
        value = strtod( txt, &endptr );
        
        if( endptr == NULL || endptr[0] != '\0' || txt[0] == '\0' )
        {
            warning( "Unable to convert '%s' to float", txt );
            value = 0.0;
        }
    }

    return value;
}

/* ========================================================================= */
/**
 *  Tries to open  the file for reading. 
 *
 *  First, tries to open the file in the current directory.
 *  If that fails, uses the default search path to create new
 *  filenames to try.  If that fails, tries to use the search 
 *  path defined by the environment variable STARPATH to create
 *  new filenames to try to opn.  If all attempts fail this returns
 *  NULL if unable to open the file.  Like other unix paths,
 *  the directories are separated by a colon.  The STARPATH
 *  can be up to (currently) 2048 characters in size, and there
 *  can be a maximum of 64 paths separated by colons in the 
 *  STARPATH.
 */

static FILE* openFile( const char* name, const char* path=NULL )
{
    static const int MAX_NUM_PATHS = 64;
    static bool firstTime = true;
    static char* DEFAULT_PATH= ".:..:../..:../metadata:../../metadata:"
                               "../../../metadata:./data:../data:../../data:"
                               "/local:/local/data:/data";
    static char* STARPATH = NULL;
    static char* HOME_DIR = NULL;
                
    if( firstTime )
    {
        STARPATH = getenv("STARPATH");
        HOME_DIR = getenv("HOME");

        firstTime = false;
    }

    FILE* fileptr = NULL;

    fileptr = fopen( name, "r" );

    if( fileptr == NULL )
    {
        const int NUM_SEARCH_PATHS = 4;
        const char* searchPaths[NUM_SEARCH_PATHS];

        searchPaths[0] = path;
        searchPaths[1] = DEFAULT_PATH;
        searchPaths[2] = HOME_DIR;
        searchPaths[3] = STARPATH;

        for( int i=0; i<NUM_SEARCH_PATHS && fileptr==NULL; i++ )
        {
            const char* searchPath = searchPaths[i];

            if( searchPath != NULL )
            {
                string tokens[MAX_NUM_PATHS];
                int numTokens = split(searchPath, ":", tokens, MAX_NUM_PATHS);

                for( int i=0; i<numTokens && fileptr==NULL; i++ )
                {
                    string filename = tokens[i] + "/" + name;
                    fileptr = fopen( filename.c_str(), "r" );
                    //printf("tried '%s', FILE=%p\n",filename.c_str(),fileptr);
                }
            }
        }
    }

    return fileptr;
}

/* ========================================================================= */

static void info( const char* format, ... ) 
{
    static const char* infoLog = "info.log";
    static FILE* infoOutfile = stdout;

    static bool firstTime = true;

    if( firstTime )
    {
        if( infoOutfile == NULL ) 
            infoOutfile=fopen( infoLog, "w" );

        firstTime = false;
    }

    if( infoOutfile == NULL ) 
        infoOutfile = stdout;

    fprintf( infoOutfile,"[%s] ", moduleName() );
    //fprintf( infoOutfile,"[%s] ", "[info]" );

    va_list args;
    va_start( args, format );
    vfprintf(infoOutfile, format, args);
    va_end( args );

    fprintf(infoOutfile,"\n"); 
    fflush(infoOutfile);
}

/* ========================================================================= */

static void warning( const char* format, ... ) 
{
    static const char* warnLog = "warning.log";
    static FILE* warnOutfile = stderr;

    static bool firstTime = true;

    if( firstTime )
    {
        if( getenv("LOG_WARNINGS") != NULL ) 
            warnOutfile = fopen( warnLog, "w" );

        firstTime = false;
    }

    if( warnOutfile == NULL ) 
        warnOutfile = stderr;

    fprintf( warnOutfile,"%-*s ",(int)strlen(moduleName())+2, "[WARNING]" );

    va_list args;
    va_start( args, format );
    vfprintf(warnOutfile, format, args);
    va_end( args );

    fprintf(warnOutfile,"\n"); 
    fflush(warnOutfile);
}

/* ========================================================================= */

static int error( const char* format, ... ) 
{
    fprintf( stderr,"%-*s ",(int)strlen(moduleName())+2, "[ERROR]" );

    va_list args;
    va_start( args, format );
    vfprintf(stderr, format, args);
    va_end( args );

    fprintf(stderr,"\n"); 
    fflush(stderr);

    return -1;
}

/* ========================================================================= */

static char* moduleName()
{
   return "starview";
}

/* ========================================================================= */

static void segFaultHandler( int signal )
{
    fprintf( stderr,"\n\n---SEGMENTATION FAULT---\n" );
#ifdef TRACE_ON
    Debug::printCallStack();
#endif 
    exit( -1 );
}

/* ========================================================================= */

static void catchSegmentationFaults( bool catchSegFaults )
{
    if( catchSegFaults )
        signal( SIGSEGV, &segFaultHandler );
}

/* ========================================================================= */
/* ==                              ASSERTIONS                             == */
/* ========================================================================= */

#ifdef ASSERTIONS

static void require(bool expr, char* file, int line, char* format, ...) 
{
    if( !(expr) )
    {
        fprintf(stderr, "\nASSERTION FAILURE: in %s, line %d", file, line);
        fprintf(stderr, "\nError Message: ");

        va_list args;
        va_start( args, format );
        vfprintf( stderr, format, args );
        va_end( args );

        fprintf( stderr, "\n" );
        fflush( stderr );

        exit(-1);
    }
}


#define REQUIRE( expression, print_statement ) \
    if( !(expression) ) \
    { \
        printf("\nASSERTION FAILURE: in %s, line %d",__FILE__, __LINE__);\
        printf("\nError Message: ");\
        printf print_statement; \
        printf("\n"); \
        fflush(stdout); \
        abort(); \
    }
#define NOTNULL(ptr) REQUIRE((ptr!=NULL),(#ptr " is null!"))

/*  
// Here are the new REQUIRE and NOTNULL macros with cleaner syntax.

#define REQUIRE( expression, format, ... ) StarObject::require(\
                 (expression), __FILE__, __LINE__, format, ## __VA_ARGS__)
#define NOTNULL(ptr) REQUIRE((ptr!=NULL), #ptr " is null!")
*/

#define BOUNDS_CHECK(i,min,max) REQUIRE( ((i>=(int)min)&&(i<(int)max)),\
            ("out of bounds: %s=%d, valid=[%d...%d]", #i,i,(int)min,(int)max-1))

#define RANGE_CHECK(f,min,max) REQUIRE( ((f>=(float)min)&&(f<=(float)max)),\
            ("out of range: %s=%0.3f, valid=[%0.3f...%0.3f]", #f,f,min,max))

#else

#define REQUIRE(expression, format, ...) /* nothing */
#define NOTNULL(ptr) /* nothing */

#endif // ASSERTIONS

/* ========================================================================= */
/* ==                                TRACE                                == */
/* ========================================================================= */

// By default, trace does nothing.  Turning it on REALLY SLOWS THINGS DOWN.

// This version of trace ignores the parameters and just records a
// stack trace.  (I don't know how to incorporate __VA_ARGS__ yet).
#ifdef TRACE_ON
#define TRACE(...)\
    StarObject::Debug __trace_obj__(__PRETTY_FUNCTION__,__FILE__,__LINE__)
                //, ## __VA_ARGS__ )
#else
#define TRACE(...)
#endif

/* ========================================================================= */
/* ==                                DEBUG                                == */
/* ========================================================================= */

#ifdef DEBUG_ON

// DEBUG macro calls the StarObject::debug function which automatically
// prints out the name of the function nicely formatted.
//
// if DEBUG_TO_STDERR is set, forces all debug statements to the screen

#ifdef DEBUG_TO_STDERR 
#define DEBUG(fmt,...) fprintf(stderr,"[%s] " fmt "\n",__func__,## __VA_ARGS__)
#else
#define DEBUG(format, ...) StarObject::Debug::debug(\
    __PRETTY_FUNCTION__,__FILE__,__LINE__,format, ## __VA_ARGS__)
#endif

class Debug
{
public:
    char* currentFunc;
    char buf[256];
    Debug(const char* funcInfo, char* file, int line) 
    {
        sprintf(buf, "%s, %s:%d", debugFuncName(funcInfo), file, line);

        updateCallStack( buf );

        currentFunc = buf;
    }

    /* ===================================================================== */

    virtual ~Debug() 
    {
        updateCallStack();
    }

    /* ===================================================================== */

    static void updateCallStack( char* frameInfo=NULL )
    {
        static const int STACK_SIZE = 50;
        static string callStack[STACK_SIZE];
        static int current;

        if( frameInfo != NULL )
        {
            if( strcmp(frameInfo,"@print")==0 )
            {
#ifdef LL_DEBUG
                fprintf(stderr,"about to print stack, current=%d\n",current);
#endif

                for( int j=current-1, i=0; j>=0; j--,i++ )
                    fprintf(stderr,"[%d] %s\n", i, callStack[j].c_str());
            }
            else // otherwise, add frame to stack
            {
                if(current < STACK_SIZE)
                {
                    char buf[256];
                    sprintf(buf, "%s", frameInfo);

                    callStack[current] = buf;
                    current++;
#ifdef LL_DEBUG
                    fprintf(stderr,"added '%s' to stack, current=%d\n",
                            buf,current);
#endif
                }
                else
                {
                    error( "call stack overflow" );
                }
            }
        }
        else
        {
#ifdef LL_DEBUG
            fprintf(stderr,"removing '%s' from stack, current=%d\n",
                     callStack[current-1].c_str(),current);
#endif
            callStack[current] = "";
            current--;
        }
    }

    /* ===================================================================== */

    static void printCallStack()
    {
        updateCallStack( "@print" );
    }

    /* ===================================================================== */

    static void debug(const char* func, char* file, int line, char* format, ...)
    {
        static const char* debugLog = "debug.log";
        static FILE* debugOutfile = NULL;

        static bool firstTime = true;

        if( firstTime )
        {
            if( getenv("DEBUG_TO_STDOUT") != NULL ) 
                debugOutfile = stdout;
            else
                debugOutfile = fopen( debugLog, "w" );

            firstTime = false;
        }

        if( debugOutfile == NULL ) 
            debugOutfile = stdout;

        if(debugOutfile==stdout || debugOutfile==stderr) 
            fprintf( debugOutfile,"[debug] " );

        if( func != NULL )
            fprintf( debugOutfile, "%s: ", debugFuncName(func) );
        else if( file != NULL )
            fprintf( debugOutfile, "%s: ", file );
        else
            fprintf( debugOutfile, "[debug] " );

        va_list args;
        va_start( args, format );
        vfprintf( debugOutfile, format, args );
        va_end( args );

        fprintf( debugOutfile, "\n" );
        fflush( debugOutfile );
    }

    /* ===================================================================== */

    static const char* debugFuncName( const char* prettyfunc )
    {   
        // 'prettyfunc' is assumed to be gcc's __PRETTY_FUNCTION__
        // returns just classname::funcname without all the extra crap
        int i,j,k;
        static string func;

        func = prettyfunc;

        // remove the function parameters
        if( (i=func.find('(')) >= 0 )
        {
            func.erase( i, func.size()-1 );

            // remove extra namespaces, searching backwards
            if( (j=func.rfind("::")) >= 0 )
                if( (k=func.rfind("::", j-2)) >= 0 )
                    func.erase( 0, k+2 );

            // remove return type, if its still there
            if( (k=func.rfind( " " )) >=0 )
                func.erase( 0, k+1 );
        }

        return func.c_str();
    }

    /* ===================================================================== */

    static char* debugFormat(const char* format, ...)
    {
        static char buf[1024];
        va_list args;
        va_start( args, format );
        vsnprintf(buf, 1024, format, args);
        va_end( args );
        return buf;
    }

    /* ===================================================================== */
}; // end inner class Debug

#else

#define DEBUG(format, ...)

#endif // DEBUG_ON

/* ========================================================================= */

private: // helper methods

private: // members

private: // disable copy constructor and operator=
        StarObject(StarObject&);
        StarObject& operator=(StarObject&);
};

#endif // _OBJECT_H_

/* ========================================================================= */
/* ==                                MACROS                               == */
/* ========================================================================= */

#ifdef SHARED_OBJECT
#define REGISTER_SHARED_OBJECT(name) \
    extern "C" \
    { \
        StarObject* getInstance##name() \
        { \
            return new name(); \
        } \
    } 
#else
#define REGISTER_SHARED_OBJECT(name)  // nothing
#endif

#define NOTIMPL fprintf(stderr,"%s NOT IMPLEMENTED YET\n",__PRETTY_FUNCTION__)

/* ========================================================================= */

