package edu.unh.sdb.util.VisitGraniteFramework.GraniteServer;

/* =========================================================================
 * GraniteServerAction.java
 *
 * Modifications: 
 *                 08-04-09 - (Frain) Added code for retrieving # attributes
 *                 07-29-09 - (Frain) Added code for retrieving DS dimensions
 *                 07-27-09 - (Frain) Added code for getting/setting byteOrder 
 *                 07-13-09 - (Frain) Added debug code
 * ========================================================================= */

import java.util.Vector;
import edu.unh.sdb.util.CommandProcessor.CommandAction;
import edu.unh.sdb.datasource.*;
import java.nio.*; // Frain added for ByteOrder

/* ========================================================================= */
/**
 *  Base class for all granite server actions.
 *
 *  Each action is encapsulated as a single class 
 *  re-implementing implementing the executeCmd() method.
 *
 *  The base clase for this class (CommandAction) contains
 *  one method, 'execute'.  This delegated to the
 *  proper executeCmd() method.
 *
 *  All GraniteServerAction subclasses are in this
 *  file as nested classes of this class.
 *  The puropose of this is to keep the code 
 *  in a single place and to prevent so much cut and 
 *  paste of command class code.
 */

public class GraniteServerAction extends CommandAction
{
    protected GraniteServer gServer = null; 
    public GraniteServerAction(GraniteServer gServer) {
       super();
       this.gServer = gServer;
    }
  
    /**  
     *   Process a command with specified arguments.
     *   args[0] is the command as a String
     *   args[ 1 ... ] are the actual arguments to the command
     */
  
    public boolean execute(Vector args) {
        try {
           executeCmd(args);
        }
        catch(Exception e) {
           gServer.logError(args, "Error: '" + e + "'");
        }

        return true;
    }

    public boolean executeCmd(Vector args) {
        System.err.printf("GraniteServerAction: This is the BASE CLASS!!!  " +
                              "You shouldn't be here\n");

        return true;
    }

/* ========================================================================= */

public static class DSCreate extends GraniteServerAction
{
    public DSCreate(GraniteServer g) {
        super(g);
    }

    public boolean executeCmd(Vector args)
    {
     int idPos = 1;
     int fdlPos = 2;
     
     String     dsName  = (String) args.get(idPos);
     String     fdlName = (String) args.get(fdlPos);
     DataSource ds      = null;

     try
     {
        ds = DataSource.create(dsName, fdlName);
        if (ds != null) {
           ds.activate();
        }
     }
     catch (Exception ex)
     {
        gServer.logError(args, "Unable to open: " + ex);
     }

     int graniteID = -1;
     if (ds != null)
        graniteID = gServer.addGraniteObject(ds);
     
     this.gServer.sendData((byte)(graniteID >= 0 ? 0 : 1), graniteID);

     return true;
  }
}

/* ========================================================================= */

public static class GetBounds extends GraniteServerAction
{
    public GetBounds(GraniteServer g) {
        super(g);
    }

    public boolean executeCmd(Vector args)
    {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte   retVal = 0;
     int[]  bounds = null;

     try
     {
        if (grObject instanceof DataCollection)
        {
           bounds = ((DataCollection) grObject).getDimensionsArray();
        }
        else
        {
           this.gServer.logError(args, "id is not a DataCollection:" + grId);
           retVal = 1;
        }

     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, bounds);
     return true;
  }
}

/* ========================================================================= */

public static class GetByte extends GraniteServerAction
{
    public GetByte(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte  retCode = 0;
     byte  value = 0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getByte(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getByte(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getByte(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, value);
     return true;
  }
}

/* ========================================================================= */

public static class GetBytes extends GraniteServerAction
{
    public GetBytes(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos  = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte    retCode = 0;
     byte[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getBytes(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getBytes(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getBytes(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetBytesByAttr extends GraniteServerAction
{
    public GetBytesByAttr(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte    retCode  = 0;
     byte[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getBytesByAttribute();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getBytesByAttribute();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retCode  = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode  = 2;
     }
     this.gServer.sendData(retCode, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetBytesByPoint extends GraniteServerAction
{
    public GetBytesByPoint(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte    retCode  = 0;
     byte[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getBytesByPoint();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getBytesByPoint();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retCode  = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode  = 2;
     }
     this.gServer.sendData(retCode , values);
     return true;
  }
}

/* ========================================================================= */

public static class GetDouble extends GraniteServerAction
{
    public GetDouble(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte   retVal = 0;
     double value = 0.0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getDouble(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getDouble(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getDouble(loc, attr);
        }
        else
		{
		   this.gServer.logError(args, "Id refers to object with no datum access.");
		   retVal = 1;
		}
	     } catch (Exception ex)
	     {
		this.gServer.logError(args, ex.getMessage());
		retVal = 1;
	     }
	     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class GetDoubles extends GraniteServerAction
{
    public GetDoubles(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte      retCode = 0;
     double[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getDoubles(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getDoubles(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getDoubles(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retCode = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetDoublesByAttr extends GraniteServerAction
{
    public GetDoublesByAttr(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte      retVal = 0;
     double[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getDoublesByAttribute();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getDoublesByAttribute();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetDoublesByPoint extends GraniteServerAction
{
    public GetDoublesByPoint(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte      retVal = 0;
     double[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getDoublesByPoint();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getDoublesByPoint();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetFloat extends GraniteServerAction
{
    public GetFloat(GraniteServer g) {
        super(g);
    }

   public boolean executeCmd(Vector args)
   {      
      int idPos   = 1;  
      int attrPos = 2;
      int locPos  = 3; 
      
      int grId        = ((Integer) args.get(idPos)).intValue();
      Object grObject = this.gServer.findGraniteObject(grId);
      int attr        = ((Integer) args.get(attrPos)).intValue();

      DataCollection dc        = null;
      DCIterator     iter      = null;

      byte  retVal = 0;
      float value = 0.0f;

      try
      {
         if (grObject instanceof DCIterator)
         {
            iter = (DCIterator) grObject;
            value = iter.getDataCollection().getFloat(iter, attr);
         }
         else if (grObject instanceof DCBoundsIterator)
         {
            DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
            DataBlock db = bndsIter.getDataBlock();
            IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
            value = db.getFloat(loc, attr);
         }
         else if (grObject instanceof DataCollection)
         {
            dc = (DataCollection) grObject;
            IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
            value = dc.getFloat(loc, attr);
         }
         else
         {
            this.gServer.logError(args, "Id '" + grObject.getClass().getName() + 
                                        "' refers to object with no datum access.");
            retVal = 1;
         }
      }
      catch (NullPointerException npe)
      {
         this.gServer.logError(args, "Unable to find Granite Object: " + grId);
      }
      catch (Exception ex)
      {
         this.gServer.logError(args, ex.getClass().getName() + " : " + ex.getMessage());
         ex.printStackTrace();
         retVal = 2;
      }
      this.gServer.sendData(retVal, value);

      return true;
   }
}

/* ========================================================================= */

public static class GetFloats extends GraniteServerAction
{
    public GetFloats(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     float[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getFloats(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getFloats(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getFloats(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (ClassCastException ex)
     {
        this.gServer.logError(args, "GetFloats.executeCmd", ex);
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetFloatsByAttr extends GraniteServerAction
{
    public GetFloatsByAttr(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     float[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getFloatsByAttribute();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getFloatsByAttribute();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        ex.printStackTrace();
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetFloatsByPoint extends GraniteServerAction
{
    public GetFloatsByPoint(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     float[]  values = null;


     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getFloatsByPoint();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getFloatsByPoint();
           System.out.println("   values.length = " + values.length);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetInt extends GraniteServerAction
{
    public GetInt(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte  retVal = 0;
     int value = 0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getInt(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getInt(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getInt(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     
     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class GetInts extends GraniteServerAction
{
    public GetInts(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte   retVal = 0;
     int[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getInts(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getInts(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getInts(loc);
        }
        else
        {
           this.gServer.logError(args, "id is not a DataCollection:" + grId);
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetIntsByAttr extends GraniteServerAction
{
    public GetIntsByAttr(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte   retVal = 0;
     int[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getIntsByAttribute();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getIntsByAttribute();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetIntsByPoint extends GraniteServerAction
{
    public GetIntsByPoint(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte   retVal = 0;
     int[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getIntsByPoint();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getIntsByPoint();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetShort extends GraniteServerAction
{
    public GetShort(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte  retVal = 0;
     short value = 0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getShort(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getShort(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getShort(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class GetShorts extends GraniteServerAction
{
    public GetShorts(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     short[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getShorts(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getShorts(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getShorts(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetShortsByAttr extends GraniteServerAction
{
    public GetShortsByAttr(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     short[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getShortsByAttribute();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getShortsByAttribute();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetShortsByPoint extends GraniteServerAction
{
    public GetShortsByPoint(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     short[]  values = null;

     try
     {
        if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           values = bndsIter.getDataBlock().getShortsByPoint();
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           values = dc.getShortsByPoint();
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class GetStep extends GraniteServerAction
{
    public GetStep(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int stepPos = 2;
     
     int    grId     = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int    stepNum  = ((Integer) args.get(stepPos)).intValue();
     int    stepId   = -1;       
     
     byte   retVal = 0;
     try
     {
        if (grObject instanceof TimeSeries)
        {
           DataSource ds = (DataSource) ((TimeSeries) grObject).getStep(stepNum);
           ds.activate();
           stepId = this.gServer.addGraniteObject(ds);
        }
        else
        {
           this.gServer.logError(args, "Not a valid iterator object.");
           retVal = 1;
        }
     }
     catch (ArrayIndexOutOfBoundsException aiob )
     {
        this.gServer.logError(args, "No step: " + stepNum + " " + aiob.getMessage());
        retVal = 2;
     }
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 3;
     }
     
     this.gServer.sendData(retVal, stepId);

     return true;
  }
}

/* ========================================================================= */

public static class GetStorageTypes extends GraniteServerAction
{
    public GetStorageTypes(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte   retVal = 0;
     int[]  storageTypes = null;

     try
     {
        if (grObject instanceof DataCollection)
        {
           storageTypes = ((DataCollection) grObject)
                               .getRecordDescriptor().getStorageTypes();
        }
        else
        {
           this.gServer.logError(args, "id is not a DataCollection:" + grId);
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, storageTypes);
     return true;
  }
}

/* ========================================================================= */

public static class GetTime extends GraniteServerAction
{
    public GetTime(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int stepPos = 2;
     
     int    grId     = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int    stepNum  = ((Integer) args.get(stepPos)).intValue();
     
     byte    retVal = 0;
     float   time  = -1.0f;
     try
     {
        if (grObject instanceof TimeSeries)
        {
           time = ((TimeSeries) grObject).getTime(stepNum);
        }
        else
        {
           this.gServer.logError(args, "Not a valid iterator object.");
           retVal = 1;
        }
     }
     catch (ArrayIndexOutOfBoundsException aiob )
     {
        this.gServer.logError(args, "No step: " + stepNum + " " + aiob.getMessage());
        retVal = 2;
     }
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 3;
     }
     
     this.gServer.sendData(retVal, time);
     return true;
  }
}

/* ========================================================================= */

public static class Init extends GraniteServerAction
{
    public Init(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int    grId     = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     try
     {
        if (grObject instanceof DCIterator)
           ((DCIterator) grObject).init();
        else if (grObject instanceof DCBoundsIterator)
           ((DCBoundsIterator) grObject).init();
        else
        {
           this.gServer.logError(args, "Not a valid iterator object.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }

     this.gServer.logSendAction(retVal, "InitAction", 0);
     return true;
  }
}

/* ========================================================================= */

public static class InterpByte extends GraniteServerAction
{
    public InterpByte(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte  retVal = 0;
     byte  value = 0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getByte(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getByte(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getByte(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class InterpBytes extends GraniteServerAction
{
    public InterpBytes(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos  = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte    retVal = 0;
     byte[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getBytes(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getBytes(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getBytes(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class InterpDouble extends GraniteServerAction
{
    public InterpDouble(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte   retVal = 0;
     double value = 0.0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getDouble(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getDouble(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getDouble(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class InterpDoubles extends GraniteServerAction
{
    public InterpDoubles(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte      retVal = 0;
     double[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getDoubles(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getDoubles(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getDoubles(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class InterpFloat extends GraniteServerAction
{
    public InterpFloat(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     DataCollection dc        = null;
     DCIterator     iter      = null;

     byte  retVal = 0;
     float value = 0.0f;

     try
     {
        if (grObject instanceof DCIterator)
        {
           iter = (DCIterator) grObject;
           value = iter.getDataCollection().getFloat(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getFloat(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = dc.getFloat(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     }
     catch (NullPointerException npe)
     {
        this.gServer.logError(args, "Unable to find Granite Object: " + grId);
        retVal = 2;
     }
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getClass().getName() + " : " + ex.getMessage());
        retVal = 3;
     }
     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class InterpFloats extends GraniteServerAction
{
    public InterpFloats(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     float[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getFloats(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getFloats(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getFloats(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (ClassCastException ex)
     {
        this.gServer.logError(args, "GetFloats.executeCmd", ex);
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class InterpInt extends GraniteServerAction
{
    public InterpInt(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte  retVal = 0;
     int value = 0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getInt(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getInt(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getInt(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     
     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class InterpInts extends GraniteServerAction
{
    public InterpInts(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte  retVal = 0;
     int[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getInts(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getInts(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getInts(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class InterpShort extends GraniteServerAction
{
    public InterpShort(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
     int attrPos = 2;
     int locPos  = 3;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     int attr        = ((Integer) args.get(attrPos)).intValue();

     byte  retVal = 0;
     short value = 0;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           value = iter.getDataCollection().getShort(iter, attr);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           value = db.getShort(loc, attr);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           value = ((DataCollection) grObject).getShort(loc, attr);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, value);
     return true;
  }
}

/* ========================================================================= */

public static class InterpShorts extends GraniteServerAction
{
    public InterpShorts(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     int locPos = 2;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte     retVal = 0;
     short[]  values = null;

     try
     {
        if (grObject instanceof DCIterator)
        {
           DCIterator iter = (DCIterator) grObject;
           values = iter.getDataCollection().getShorts(iter);
        }
        else if (grObject instanceof DCBoundsIterator)
        {
           DCBoundsIterator bndsIter = (DCBoundsIterator) grObject;
           DataBlock db = bndsIter.getDataBlock();
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, db.dim(), locPos);
           values = db.getShorts(loc);
        }
        else if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           IndexSpaceID loc = this.gServer.getISIDFromArgs(args, dc.dim(), locPos);
           values = ((DataCollection) grObject).getShorts(loc);
        }
        else
        {
           this.gServer.logError(args, "id refers to object with no datum access.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     this.gServer.sendData(retVal, values);
     return true;
  }
}

/* ========================================================================= */

public static class NewBoundsIterator extends GraniteServerAction
{
    public NewBoundsIterator(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int dcPos    = 1;
     int bndsPos = 2;
          
     int    dcId  = ((Integer) args.get(dcPos)).intValue();
     Object grObj = this.gServer.findGraniteObject(dcId);
     
     int[]         shape   = null;     
     ISBounds      bounds = null; 
     int[]         steps  = null;
     AxisOrdering  axisOrder = null;
     
     int        iterId = -1;
     
     byte  retVal = 0;
     try
     {
        if (grObj instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObj;
           int            dim = dc.dim();
           int            shapePos = bndsPos + 2*dim;
           int            stepPos  = shapePos + dim;
           int            axisOrderPos = stepPos + dim;
           
           bounds = this.gServer.getBoundsFromArgs(args, dim, bndsPos);
           shape  = this.gServer.getIntsFromArgs(args, dim, shapePos);
           
           if (stepPos < args.size())
              steps = this.gServer.getIntsFromArgs(args, dim, stepPos);
           if (axisOrderPos < args.size())
           {
              axisOrder = new AxisOrdering(
                            this.gServer.getIntsFromArgs(args, dim, axisOrderPos));
           }
           
           DCBoundsIterator dsbIter = new DCBoundsIterator (dc, bounds, shape, 
                                                             steps, axisOrder);
           iterId = this.gServer.addGraniteObject(dsbIter);
        }
        else
        {
           this.gServer.logError(args, "Not a valid iterator object.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, "NewBoundsIterator", ex);
        retVal = 2;
     }
     
     this.gServer.sendData(retVal, iterId);
     return true;
  }
}

/* ========================================================================= */

public static class NewIterator extends GraniteServerAction
{
    public NewIterator(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int dcPos   = 1;
     int bndsPos = 2;
          
     int    dcId  = ((Integer) args.get(dcPos)).intValue();
     Object grObj = this.gServer.findGraniteObject(dcId);
     
     ISBounds     bounds = null;
     
     int    iterId = -1;
     
     byte  retVal = 0;
     try
     {
        if (grObj instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObj;
           bounds = this.gServer.getBoundsFromArgs(args, dc.dim(), bndsPos);
           iterId = this.gServer.addGraniteObject(new DCIterator(dc, bounds));
        }
        else
        {
           this.gServer.logError(args, "Not a valid iterator object.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }
     
     this.gServer.sendData(retVal, iterId);
     return true;
  }
}

/* ========================================================================= */

public static class Next extends GraniteServerAction
{
    public Next(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int grId = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte  retVal = 0;
     try
     {
        if (grObject instanceof DCIterator)
           ((DCIterator) grObject).next();
        else if (grObject instanceof DCBoundsIterator)
           ((DCBoundsIterator) grObject).next();
        else
        {
           this.gServer.logError(args, "Not a valid iterator object.");
           retVal = 1;
        }
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }

     this.gServer.logSendAction(retVal, "NextAction", 0);
     return true;
  }
}

/* ========================================================================= */

public static class NumAttributes extends GraniteServerAction
{
    public NumAttributes(GraniteServer g) {
        super(g);
    }

   public boolean executeCmd(Vector args)
   {
      int idPos = 1;
      
      int grId = ((Integer) args.get(idPos)).intValue();
      Object grObject = this.gServer.findGraniteObject(grId);
      
      int numAttrs = -1;

      byte  retVal = 0;
      try
      {
         if (grObject instanceof DataCollection)
            numAttrs = ((DataCollection) grObject).getNumAttributes();
         else
         {
            this.gServer.logError(args, "Not a DataCollection: " + grId);
            retVal = 1;
         }
      } 
      catch (Exception ex)
      {
         this.gServer.logError(args, ex.getMessage());
         retVal = 2;
      }

      this.gServer.sendData(retVal, numAttrs);
      return true;
   }
}

/* ========================================================================= */

public static class NumSteps extends GraniteServerAction
{
    public NumSteps(GraniteServer g) {
        super(g);
    }

   public boolean executeCmd(Vector args)
   {
      int idPos = 1;
      
      int grId = ((Integer) args.get(idPos)).intValue();
      Object grObject = this.gServer.findGraniteObject(grId);
      
      int numSteps = -1;

      byte  retVal = 0;
      try
      {
         if (grObject instanceof TimeSeries)
            numSteps = ((TimeSeries) grObject).getNumSteps();
         else
         {
            this.gServer.logError(args, "Not a valid iterator object.");
            retVal = 1;
         }
      } 
      catch (Exception ex)
      {
         this.gServer.logError(args, ex.getMessage());
         retVal = 2;
      }

      this.gServer.sendData(retVal, numSteps);
      return true;
   }
}

/* ========================================================================= */

public static class Subblock extends GraniteServerAction
{
    public Subblock(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos    = 1;
     int bndsPos  = 2;
     
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     byte      retVal = 0;
     DataBlock db    = null;
     int       dbid  = -1;

     try
     {
        if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           ISBounds bnds = this.gServer.getBoundsFromArgs(args, dc.dim(), bndsPos);
           db = dc.subblock(bnds);
           dbid = this.gServer.addGraniteObject(db);
        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum access.");
           retVal = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retVal = 2;
     }

     this.gServer.sendData(retVal, dbid);
     return true;
  }
}

/* ========================================================================= */

public static class TSCreate extends GraniteServerAction
{
    public TSCreate(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
      System.err.printf("entered TSCreate.executeCmd\n");

     int idPos = 1;
     int fdlPos = 2;
     
     String     tsName  = (String) args.get(idPos);
     String     fdlName = (String) args.get(fdlPos);
     TimeSeries ts      = null;
     int        grId    = -1;
     int tsid = -1;
     
     try
     {
        ts = new TimeSeries(tsName, fdlName);
     }
     catch (Exception ex)
     {
        gServer.logError(args, "Unable to open: " + ex);
     }
     
     if (ts != null)
     {
         tsid = gServer.addGraniteObject(ts);

         this.gServer.sendData((byte) 0, tsid);
     }
     else 
     {
        this.gServer.sendData((byte) 1, grId);
     }

     return true;
  }
}

/* ========================================================================= */

public static class Valid extends GraniteServerAction
{
    public Valid(GraniteServer g) {
        super(g);
    }

  public boolean executeCmd(Vector args)
  {
     int idPos = 1;
     
     int    grId     = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     
     boolean valid = false;
     try
     {
        if (grObject instanceof DCIterator)
           valid = ((DCIterator) grObject).valid();

        else if (grObject instanceof DCBoundsIterator)
           valid = ((DCBoundsIterator) grObject).valid();

        else
           this.gServer.logError(args, "Not a valid iterator object.");
     } 
     catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
     }

     if (valid)
        this.gServer.sendData((byte) 0, 1);
     else 
        this.gServer.sendData((byte) 1, 0);
     return true;
  }
}

/* ========================================================================= 
 *  class GetByteOrder
 *
 * (Frain) Added 7-27-09
 *
 * returns:
 *          0 - Big Endian
 *          1 - Little Endian
 *
 * ========================================================================= */
public static class GetByteOrder extends GraniteServerAction
{
  public GetByteOrder(GraniteServer g) {
      super(g);
  }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;

     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte  retCode = 0;
     int  value = 0;

     try
     {
        if (grObject instanceof PhysicalDataSource)
        {
           PhysicalDataSource pds = (PhysicalDataSource) grObject;
           String order = pds.getByteOrder().toString();
          
           if (order.equals("BIG_ENDIAN"))
               value = 0;
           else
               value = 1;

        }
        else
        {
           this.gServer.logError(args, "Id refers to object with no datum " +
                                       "access.");           
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, value);
     return true;
  }
} // end class GetByteOrder

/* ========================================================================= 
 *  class SetByteOrder
 *
 * (Frain) Added 7-27-09
 *
 * ========================================================================= */
public static class SetByteOrder extends GraniteServerAction
{
  public SetByteOrder(GraniteServer g) {
      super(g);
  }

  public boolean executeCmd(Vector args)
  {
     int idPos    = 1;
     int orderPos = 2;

     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);
     String byteOrder = ( args.get(orderPos).toString() );

     byte  retCode = 0;
     int  value = 0;

     try
     {
        if (grObject instanceof PhysicalDataSource)
        {
           PhysicalDataSource pds = (PhysicalDataSource) grObject;
            
           if (byteOrder.equals("BIG_ENDIAN")) {
              System.out.println(" Setting byte order to Big Endian");
              pds.setByteOrder( ByteOrder.BIG_ENDIAN);
           }
           else if (byteOrder.equals("LITTLE_ENDIAN")) {
              System.out.println(" Setting byte order to Little Endian");
              pds.setByteOrder( ByteOrder.LITTLE_ENDIAN);
           }
           else
              System.out.println(" <ERROR> Unrecognized Byte Order of <" +
                                 byteOrder + ">!");
 
        }
        else
        {
           this.gServer.logError(args, "Id refers to non Physical Data Source object ");           
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, value);
     return true;
  }
} // end class SetByteOrder



/* ========================================================================= 
 *  class GetDimensions
 *
 * (Frain) Added 7-29-09
 *
 * returns: An array of ints, each element is size of corresponding dimension
 *         
 * ========================================================================= */
public static class GetDimensions extends GraniteServerAction
{
  public GetDimensions(GraniteServer g) {
      super(g);
  }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
    
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte  retCode = 0;
     
     int arr[]=null;

     try
     {
        if (grObject instanceof PhysicalDataSource)
        {
           PhysicalDataSource pds = (PhysicalDataSource) grObject;
           arr = pds.getDimensionsArray();  
        }
        else
        {
           this.gServer.logError(args, "Id refers to object that isn't " +
                                       "a Physical Data Source"); 
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, arr);
     return true;
  }
} // end class GetDimensions


/* ========================================================================= 
 *  class GetNumDimensions
 *
 * (Frain) Added 7-29-09
 *
 * returns: integer number of dimensions of the data source
 *         
 * ========================================================================= */
public static class GetNumDimensions extends GraniteServerAction
{
  public GetNumDimensions(GraniteServer g) {
      super(g);
  }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
    
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte  retCode = 0;
     
     int arr[]=null;
     int numDimensions=0;

     try
     {
        if (grObject instanceof PhysicalDataSource)
        {
           PhysicalDataSource pds = (PhysicalDataSource) grObject;
           arr = pds.getDimensionsArray();  

           if (arr == null)
               numDimensions = 0;
           else
               numDimensions = arr.length;                   
        }
        else
        {
           this.gServer.logError(args, "Id refers to object that isn't a Physical Data Source"); 
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, numDimensions);
     return true;
  }
} // end class GetNumDimensions


/* ========================================================================= 
 *  class GetNumAttributes
 *
 * (Frain) Added 8-04-09
 *
 * returns: integer number of attributes in the data source
 *         
 * ========================================================================= */
public static class GetNumAttributes extends GraniteServerAction
{
  public GetNumAttributes(GraniteServer g) {
      super(g);
  }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
    
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte  retCode = 0;
     
     int arr[]=null;
     int numAttributes=0;

     try
     {
        if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           numAttributes = dc.getNumAttributes();             
        }
        else
        {
           this.gServer.logError(args, "Id refers to object that isn't a Physical Data Source"); 
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }
     this.gServer.sendData(retCode, numAttributes);
     return true;
  }
} // end class GetNumAttributes


/* ========================================================================= 
 *  class GetAttributeName
 *
 * (Frain) Added 08-04-09
 *
 * returns: String attribute name
 *         
 * ========================================================================= */
public static class GetAttributeNames extends GraniteServerAction
{
  public GetAttributeNames(GraniteServer g) {
      super(g);
  }

  public boolean executeCmd(Vector args)
  {
     int idPos   = 1;
 
     int grId        = ((Integer) args.get(idPos)).intValue();
     Object grObject = this.gServer.findGraniteObject(grId);

     byte  retCode = 0;
     
     String names[] = null;

     try
     {
        if (grObject instanceof DataCollection)
        {
           DataCollection dc = (DataCollection) grObject;
           names = dc.getAttributeNames();             
        }
        else
        {
           this.gServer.logError(args, "Id refers to object that isn't a Physical Data Source"); 
           retCode = 1;
        }
     } catch (Exception ex)
     {
        this.gServer.logError(args, ex.getMessage());
        retCode = 2;
     }

     String returnString = new String(names[0]);

     for (int i=1;i<names.length;i++){
        returnString += " " + names[i];   
     }
     System.out.println("  returnString = " + returnString);
     byte[] byteArr = returnString.getBytes();

     this.gServer.sendData(retCode,byteArr);
     return true;
  }
} // end class GetAttributeNames



} // end class GraniteServerAction

/* ========================================================================= */
/*
 * Notation conventions:
 * dc    - DataCollection id
 * ds    - DataSource id (also a dc)
 * ts    - TimeSeries id
 * it    - iterator id
 * 
 * <ipos> - d integers where d is dimension
 * <fpos> - d floats where d is dimension
 *  
 * longName  | method             | arguments
 * 
 * Constructors
 * ---------------------------
 * dsCreate  | DataSource.create  |   name fdl    
 *                                      returns: int ds ID  <0 => failure
 *                                      or
 *                                      returns: metadata array, mdb[0] = ds ID
 * tsCreate  | new TimeSeries     |   name fdl 
 *                                          returns ts
 *                 
 * getStep   | ts.getStep(n)      |    ts n   
 *                                     returns: int ds ID, <0 => failure
 *                                     or
 *                                     returns: metadata array, mdb[0] = ds ID
 * getTime   | ts.getTime(n)      |   ts n
 *                                          returns float
 * 
 * subblock  |   c.subblock       |    dc   <ipos>     <ipos>   
 *                                          # lowerBound upperBound
 *                                          # returns dc id
 * delete    | remove id from list | graniteID
 *                 
 * Data access commands
 * --------------------
 * # get 1 attribute at position 
 * gd  getDouble         | double getDouble                 | dc attr <ipos>
 * gf  getFloat          | float  getFloat                  | dc attr <ipos>
 * gi  getInt            | int    getInt                    | dc attr <ipos>
 * gs  getShort          | short  getShort                  | dc attr <ipos>
 * gs  getByte           | byte   getByte                   | dc attr <ipos>

 * # get all attributes at position    
 * gds getDoubles        | double[] getDoubles              | dc <ipos>
 * gfs getFloats         | float[]  getFloats               | dc <ipos>
 * gis getInts           | int[]    getInts                 | dc <ipos>
 * gss getShorts         | short[]  getShort                | dc <ipos>
 * gbs getBytes          | byte[]   getByte                 | dc <ipos>
 *    
 * # get all data in attribute order
 * gda getDoublesByAttr  | double[] getDoublesByAttribute   | dc 
 * gfa getFloatsByAttr   | float[]  getFloatsByAttribute    | ds
 * gia getIntsByAttr     | int[]    getIntsByAttribute      | ds
 * gsa getShortsByAttr   | short[]  getShortByAttribute     | ds
 * gba getBytesByAttr    | byte[]   getBytesByAttribute     | ds
 *
 * # get all data in point order     
 * gdp getDoublesByPoint | double[] getDoublesByPoint       | dc
 * gfp getFloatsByPoint  | float[]  getFloatsByPoint        | ds
 * gip getIntsByPoint    | int[]    getIntsByPoint          | ds
 * gsp getShortsByPoints | short[]  getShortByPoint         | ds
 * gbp getBytesByPoint   | byte[]   getBytesByPoint         | ds
 * 
 * # interpolate attr at gpos      
 * id  interpDouble      | double   getDouble               | dc attr <fpos>  
 * if  interpFloat       | float    getFloat                | dc attr <fpos>
 * ii  interpInt         | int      getInt                  | dc attr <fpos>
 * is  interpShort       | short    getShort                | dc attr <fpos>
 * ib  interpByte        | byte     getByte                 | dc attr <fpos>
 *
 * # interp all attrs at gpos      
 * ids interpDoubles     | double[] getDoubles              | dc <fpos> 
 * ifs interpFloats      | float[]  getFloats               | dc <fpos>
 * iis interpInts        | int[]    getInts                 | dc <fpos>
 * iss interpShorts      | short[]  getShorts               | dc <fpos>
 * isb interpBytes       | byte[]   getByte                 | dc <fpos>
 *       
 *
 * Iterator commands
 * -----------------
 * ni  newIterator    | new DSIterator(ISIterator) | dc <ipos> <ipos> [<ipos>]
 *                                                    # lower  upper   step
 *                                                     returns it
 * 
 * next                  | it.next()                        | it
 * init                  | it.init()                        | it
 * valid                 | it.valid()                       | it   
 *                                                   0 if invalid, else valid
 * resetIter             | it.reset                 | it <ipos> <ipos> [<ipos>]
 * setDC                 | it.setDataCollection             | it dc
 * 
 * gd  getDouble         | double  getDouble( attr, it )    | it attr 
 * gf  getFloat          | float   getFloat ( attr, it )    | it attr 
 * gi  getInt            | int     getInt   ( attr, it )    | it attr 
 * gs  getShort          | short   getShort ( attr, it )    | it attr 
 * gb  getByte           | byte    getByte  ( attr, it )    | it attr 
 * 
 * gds getDoubles        | double[] getDoubles( it )        | it
 * gfs getFloats         | float[]  getFloats ( it )        | it
 * gis getInts           | int[]    getInts   ( it )        | it
 * gss getShorts         | short[]  getShorts ( it )        | it
 * gbs getBytes          | byte[]   getBytes  ( it )        | it
 * 
 * gbo getByteOrder      | int      getByteOrder            | dc
 * sbo setByteOrder      | int      setByteOrder            | dc <BIG_ENDIAN or
 *                                                                LITTLE_ENDIAN>
 * 
 * 
 *
 * nbi | new DSIterator(ISBIterator) | dc <ipos> <ipos>  <ipos> [<ipos>]
 * newBoundsIterator 
 *                                     #  lower  upper  shape   step
 *                                     # Note "step" is the "extra" step taken
 *                                     #  beyond what is implied by shape; can
 *                                     #  be < 0.
 *                                          
 * block                | bIter.getDataBlock               | boundsIter
 *
 * resetIter            | it.reset         | it <ipos> <ipos>  <ipos> [<ipos>]
 * setDC                | it.setDataCollection             | it dc
 * 
 * Metadata commands
 * -----------------
 * b   bounds           | int[]   getDimensionsArray()     | dc
 * na  numAttr          | int     getNumAttributes()       | dc
 * st  storageTypes     | int[]   getRecordDescriptor().getStorageTypes() | dc
 * 
 *             StorageType:  0 Undefined, 1 Boolean, 2 Byte, 3 UnsignedByte
 *                           4 Char,      5 Short,   6 UnsignedShort, 7 Int
 *                           8 Float,     9 Double, 10 Long,         12 Record
 *                           
 *             Note: Only Byte, Short, Int, Float, Double are really supported.
 *             
 */


 /*
  *  Command syntax:
  *
  *  dsCreate             ss     DsCreateAction
  *  dsc                  ss     DsCreateAction
  *  tsCreate             ss     DsCreateAction
  *  tsc                  ss     DsCreateAction
  *  numSteps             i      NumStepsAction
  *  ns                   i      NumStepsAction
  *  getStep              ii     GetStepAction
  *  step                 ii     GetStepAction
  *  getTime              ii     GetTimeAction    
  *  time                 ii     GetTimeAction    
  *  subblock             ii*    SubblockAction
  *  sub                  ii*    SubblockAction
  *  getDouble            iii*   GetDoubleAction
  *  gd                   iii*   GetDoubleAction
  *  getFloat             iii*   GetFloatAction
  *  gf                   iii*   GetFloatAction
  *  getInt               iii*   GetIntAction
  *  gi                   iii*   GetIntAction
  *  getShort             iii*   GetShortAction
  *  gs                   iii*   GetShortAction
  *  getByte              iii*   GetByteAction
  *  gb                   iii*   GetByteAction
  *  getDoubles           ii*    GetDoublesAction
  *  gds                  ii*    GetDoublesAction
  *  getFloats            ii*    GetFloatsAction
  *  gfs                  ii*    GetFloatsAction
  *  getInts              ii*    GetIntsAction
  *  gis                  ii*    GetIntsAction
  *  getShorts            ii*    GetShortsAction
  *  gss                  ii*    GetShortsAction
  *  getBytes             ii*    GetBytesAction
  *  gbs                  ii*    GetBytesAction
  *  getDoublesByAttr     i      GetDoublesByAttrAction
  *  gda                  i      GetDoublesByAttrAction
  *  getFloatsByAttr      i      GetFloatsByAttrAction
  *  gfa                  i      GetFloatsByAttrAction
  *  getIntsByAttr        i      GetIntsByAttrAction
  *  gia                  i      GetIntsByAttrAction
  *  getShortsByAttr      i      GetShortsByAttrAction
  *  gsa                  i      GetShortsByAttrAction
  *  getBytesByAttr       i      GetBytesByAttrAction
  *  gba                  i      GetBytesByAttrAction
  *  getDoublesByPoint    i      GetDoublesByPointAction
  *  gdp                  i      GetDoublesByPointAction
  *  getFloatsByPoint     i      GetFloatsByPointAction
  *  gfp                  i      GetFloatsByPointAction
  *  getIntsByPoint       i      GetIntsByPointAction
  *  gip                  i      GetIntsByPointAction
  *  getShortsByPoints    i      GetShortsByPointsAction
  *  gsp                  i      GetShortsByPointsAction
  *  getBytesByPoint      i      GetBytesByPointAction
  *  gbp                  i      GetBytesByPointAction
  *  getByteOrder         i      GetByteOrderAction
  *  gbo                  i      GetByteOrderAction
  *  getNumDimensions     i      GetNumDimensions  
  *  getDimensions        i      GetDimensions
  *  getNumAttributes     i      GetNumAttributes
  *  getAttributeNames    i      GetAttributeName
  *  interpDouble         iif*   InterpDoubleAction
  *  id                   iif*   InterpDoubleAction
  *  interpFloat          iif*   InterpFloatAction
  *  if                   iif*   InterpFloatAction
  *  interpInt            iif*   InterpIntAction
  *  ii                   iif*   InterpIntAction
  *  interpShort          iif*   InterpShortAction
  *  is                   iif*   InterpShortAction
  *  interpByte           iif*   InterpByteAction
  *  ib                   iif*   InterpByteAction
  *  interpDoubles        if*    InterpDoublesAction
  *  ids                  if*    InterpDoublesAction
  *  interpFloats         if*    InterpFloatsAction
  *  ifs                  if*    InterpFloatsAction
  *  interpInts           if*    InterpIntsAction
  *  iis                  if*    InterpIntsAction
  *  interpShorts         if*    InterpShortsAction
  *  iss                  if*    InterpShortsAction
  *  interpBytes          if*    InterpBytesAction
  *  ibs                  if*    InterpBytesAction
  *  newIterator          ii*    NewIteratorAction
  *  ni                   ii*    NewIteratorAction
  *  next                 i      NextAction
  *  init                 i      InitAction
  *  valid                i      ValidAction
  *  newBoundsIterator    ii*    NewBoundsIteratorAction
  *  nbi                  ii*    NewBoundsIteratorAction
  *
  */
