/* 
	MultiBlockCacheDataSource.java
	Copyright 2000 University of New Hampshire
	Durham, NH 03824
	
	-------------------- Edit history as of 01/30/05 -------------------------
	01/30/05 rdb: Deleted datum( IS ) and datum( is, RecordSpec ) - super versions ok
*/


package edu.unh.sdb.datasource;

import edu.unh.sdb.common.*;
//import java.io.*;


/** 
  *	The MultiBlockCacheDataSource provides caching service using a MultiBlockCache. 
  */

public class MultiBlockCacheDataSource extends BlockCacheDataSource
{
    IndexSpaceID cacheDimensions;


	/** Creates a MultiBlockCacheDataSource that caches the specified DataSource. The blocks used have
	  *	dimensions specified by the blockDimensions argument. The cacheDimensions argument specifies
	  *	the dimensions of the array containing the blocks. Only the region specified by the iterationSpace
     *   parameter is cached when this constructor is used. 
	  */
	public MultiBlockCacheDataSource( String name, DataSource ds,  
	                                  ISBounds iterationSpace, 
	                                  IndexSpaceID blockDimensions, 
	                                  IndexSpaceID cacheDimensions)
	{
        super(name, ds, iterationSpace, blockDimensions);
        this.cacheDimensions=cacheDimensions;
	}

	/** Creates a MultiBlockCacheDataSource that caches the specified DataSource. The blocks used have
	  *	dimensions specified by the blockDimensions argument. The cacheDimensions argument specifies
	  *	the dimensions of the array containing the blocks. The entire datasource index space is cached
     *   when this constructor is used. 
	  */
	public MultiBlockCacheDataSource( String name, DataSource ds,  
	                                  IndexSpaceID blockDimensions, 
	                                  IndexSpaceID cacheDimensions)
	{
	   this( name, ds, ds.getBounds(), blockDimensions, cacheDimensions );
	}
   
   
   
   
   /** Return a partitioning corresponding to the manner in which the iteration
    *  space is divided into cache block regions by the cache.
    */
   public ISPartitioning getPartitioning(){
   
      return new RegularISPartitioning(this.blockDimensions, this.iterationSpace);
   
   }

	/** Calls activate() for the underlying datasource, and creates the actual cache,
	  *	which allocates memory for the CacheBlocks.
	  */	
	public void activate()
	{
	     super.activate();
        theCache = new MultiBlockCache( this.theSource, 
                                        this.iterationSpace, 
                                        this.blockDimensions, 
                                        this.cacheDimensions);
	}
   
   
		
   /** Load the data at the given index into the provided datum. */
    public	void datum( Datum datum, IndexSpaceID did )
    {
        if (datum.size() != this.getNumAttributes()){
            new IllegalArgumentError(SDBError.Recoverable,
                  "Datum is not the right size for this datasource.");
            return;
        }
        
        theCache.lookup(datum,did);
    }




   /** Load the data at the given index into the provided datum. No argument
    *  validation is performed. Package Access only. This version provides scratch
    *  space for threaded operation.
    */
   void pdatum(Datum datum, IndexSpaceID did, int [] sQBLowerInts, int [] sQBUpperInts, int [] datumCoords1, ISBounds sourceQueryBounds )
    {

        ((MultiBlockCache)theCache).lookup( datum, did, sQBLowerInts, sQBUpperInts, datumCoords1, sourceQueryBounds );
    }


//   /** Load the data at the given index into the provided datum. No argument
//    *  validation is performed. Package Access only. This version provides scratch
//    *  space for threaded operation, and exploits coherence where possible to reduce
//    *  overhead.
//    */
//   void pdatum(Datum datum, IndexSpaceID did, CacheContext context )
//    {
//
//        ((MultiBlockCache)theCache).lookup( datum, did, context );
//    }


   /** If the data for the specified ISID is not present in the cache, then
    *  add the location to the stencil. This information can later be used for a bulk read.
    *  Package Access Only.
    */
   void addToStencil(DSBlockStencil stencil, IndexSpaceID id){
   
   
      ((MultiBlockCache)theCache).addToStencil(stencil, id);
   }


   /** Ask the datasource to retrieve the data contained in each bounds specified
    *  by the stencil, and place it in the corresponding datablocks.
    */
   void prefetch(DSBlockStencil stencil){
   
      
      ((MultiBlockCache)theCache).prefetch(stencil);
   }



    
   public IndexSpaceID getCacheArrayDimensions(){
      
      return new IntegerIndexSpaceID(this.cacheDimensions);
   }
   
    
	
	/** For debugging. Runs basic tests on the MultiBlockCacheDataSource class. */
    static public void main(String[] args){

        System.out.println("\n******************************\n***MultiBlockCacheDataSource Tests***\n******************************");
        
        AttributeDescriptor[] attrDescriptors=new AttributeDescriptor[]{
                                                    new AttributeDescriptor(StorageType.Double,"Carbon"),
                                                    new AttributeDescriptor(StorageType.Double,"Nitrogen"),
                                                    new AttributeDescriptor(StorageType.Double,"Oxygen"),
                                                    new AttributeDescriptor(StorageType.Double,"Bogosity")
                                                 };
                                                 
        RecordDescriptor recordDesc=new RecordDescriptor(attrDescriptors); 
 
        AxisOrdering ordering;
        ISBounds bounds;
        TextFileDataSource ds1;
        MultiBlockCacheDataSource cds;
        
        // One D Test
        System.out.println("\n\n***One Dimensional Test***\n");
        
        ordering=new AxisOrdering(1);
        bounds=new ISBounds(new IntegerIndexSpaceID(0),new IntegerIndexSpaceID(63));
    
        IndexSpaceID blockSizes=new IntegerIndexSpaceID(8);
        IndexSpaceID numBlocks=new IntegerIndexSpaceID(2);
        
        ds1=new TextFileDataSource("foo",recordDesc, bounds,ordering,"testdata/file.6.4.txt");
        cds=new MultiBlockCacheDataSource("bar",ds1,blockSizes,numBlocks);
        
        cds.activate();
        
        
    
        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    
        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    

        System.out.println("ds1.datum(<7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(7)));    
        System.out.println("ds1.datum(<8>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(8)));    
        
        System.out.println("ds1.datum(<16>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(16)));    

        System.out.println("ds1.datum(<62>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(62)));    
        System.out.println("ds1.datum(<54>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(54)));    

        System.out.println();
        System.out.println("       Round Two");



        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    
        System.out.println("ds1.datum(<0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0)));    

        System.out.println("ds1.datum(<7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(7)));    
        System.out.println("ds1.datum(<8>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(8)));    
        
        System.out.println("ds1.datum(<16>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(16)));    

        System.out.println("ds1.datum(<62>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(62)));    
        System.out.println("ds1.datum(<54>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(54)));    
   
        // Two-D  Tests
        System.out.println();
        System.out.println("Two Dimensional Tests");
        
        ordering=new AxisOrdering(2);
        bounds=new ISBounds(new IntegerIndexSpaceID(0,0),new IntegerIndexSpaceID(7,7));
    
        blockSizes=new IntegerIndexSpaceID(4,4); //blocks are 4x4
        numBlocks=new IntegerIndexSpaceID(2,2);
        
        ds1=new TextFileDataSource("foo",recordDesc, bounds,ordering,"testdata/file.6.4.txt");
        cds=new MultiBlockCacheDataSource("bar",ds1,blockSizes,numBlocks);
        
        cds.activate();

        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));    
        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));  

        System.out.println("ds1.datum(<0,3>) should be {103.1, 103.2, 103.3, 103.4}-->"+cds.datum(new IntegerIndexSpaceID(0,3)));    
        System.out.println("ds1.datum(<0,4>) should be {104.1, 104.2, 104.3, 104.4}-->"+cds.datum(new IntegerIndexSpaceID(0,4)));    

        System.out.println("ds1.datum(<0,7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(0,7)));    
        System.out.println("ds1.datum(<1,0>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(1,0)));    

        System.out.println("ds1.datum(<2,0>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(2,0)));    
        System.out.println("ds1.datum(<3,0>) should be {124.1, 124.2, 124.3, 124.4}-->"+cds.datum(new IntegerIndexSpaceID(3,0)));
        System.out.println("ds1.datum(<4,0>) should be {132.1, 132.2, 132.3, 132.4}-->"+cds.datum(new IntegerIndexSpaceID(4,0)));
        
        System.out.println("ds1.datum(<7,6>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(7,6)));    
        System.out.println("ds1.datum(<6,6>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(6,6)));    
   
        System.out.println();
        System.out.println("       Round Two");



        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));    
        System.out.println("ds1.datum(<0,0>) should be {100.1, 100.2, 100.3, 100.4}-->"+cds.datum(new IntegerIndexSpaceID(0,0)));  

        System.out.println("ds1.datum(<0,3>) should be {103.1, 103.2, 103.3, 103.4}-->"+cds.datum(new IntegerIndexSpaceID(0,3)));    
        System.out.println("ds1.datum(<0,4>) should be {104.1, 104.2, 104.3, 104.4}-->"+cds.datum(new IntegerIndexSpaceID(0,4)));    

        System.out.println("ds1.datum(<0,7>) should be {107.1, 107.2, 107.3, 107.4}-->"+cds.datum(new IntegerIndexSpaceID(0,7)));    
        System.out.println("ds1.datum(<1,0>) should be {108.1, 108.2, 108.3, 108.4}-->"+cds.datum(new IntegerIndexSpaceID(1,0)));    

        System.out.println("ds1.datum(<2,0>) should be {116.1, 116.2, 116.3, 116.4}-->"+cds.datum(new IntegerIndexSpaceID(2,0)));    
        System.out.println("ds1.datum(<3,0>) should be {124.1, 124.2, 124.3, 124.4}-->"+cds.datum(new IntegerIndexSpaceID(3,0)));
        System.out.println("ds1.datum(<4,0>) should be {132.1, 132.2, 132.3, 132.4}-->"+cds.datum(new IntegerIndexSpaceID(4,0)));
        
        System.out.println("ds1.datum(<7,6>) should be {162.1,162.2,162.3,162.4}-->"+cds.datum(new IntegerIndexSpaceID(7,6)));    
        System.out.println("ds1.datum(<6,6>) should be {154.1,154.2,154.3,154.4}-->"+cds.datum(new IntegerIndexSpaceID(6,6)));    
   
    }
		
}